/*  Copyright (C) 2010-2024 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "bwbDefaultData.h"

#include <runtimeInfo/runtimeInfo.h>
#include <moduleBasics/module.h>

#include <algorithm>
#include <exception>
#include <iomanip>
#include <string>
#include <vector>
#include <utility>  // make_pair

using std::endl;

bwbDefaultData::bwbDefaultData(const std::shared_ptr<RuntimeIO>& rtIO)
    :
    rtIO_(rtIO),
    delta_isa(EndnodeReadOnly<double>("requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission/delta_ISA").read(rtIO->acxml).value()),
    engineType(EndnodeReadOnly<std::string>("type").read(rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion/propulsor")).value()),
    initialCruiseMachNumber(EndnodeReadOnly<double>("initial_cruise_mach_number").read(
        rtIO->acxml.at("requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission")).value()),
    bestCruiseCL("C_LoptimumCruise", "optimum cruise CL", 0, "1", 0, 2),
    designCoGX("x", "x component", 0, "1"),
    designCoGY("y", "y component", 0, "1"),
    designCoGZ("z", "z component", 0, "1"),
    forwardCoGX("x", "x component", 0, "1"),
    forwardCoGY("y", "y component", 0, "1"),
    forwardCoGZ("z", "z component", 0, "1"),
    aftCoGX("x", "x component", 0, "1"),
    aftCoGY("y", "y component", 0, "1"),
    aftCoGZ("z", "z component", 0, "1") {
    // I am not responsible for this bloated code :( (LN 2024)
    std::string subPath = "analysis/aerodynamics/lift_coefficients";
    if (!(rtIO_->acxml.find(subPath))) {
        rtIO_->acxml[subPath];
        rtIO_->acxml.at(subPath).addAttrib("description", "lift coefficients");
        bestCruiseCL.set_value(0.5);
        bestCruiseCL.update(rtIO_->acxml.at(subPath));
    } else {
        bestCruiseCL.read(rtIO_->acxml);
    }
    subPath = "analysis/masses_cg_inertia/design_mass/mass_properties/center_of_gravity/x";
    if (!(rtIO_->acxml.find(subPath))) {
        rtIO_->acxml[subPath];
        rtIO_->acxml.at(subPath).addAttrib("description", "x component");
        designCoGX.set_value(0.0);
        designCoGX.update(rtIO_->acxml.at(subPath));
    } else {
        designCoGX.read(rtIO_->acxml);
    }
    subPath = "analysis/masses_cg_inertia/design_mass/mass_properties/center_of_gravity/y";
    if (!(rtIO_->acxml.find(subPath))) {
        rtIO_->acxml[subPath];
        rtIO_->acxml.at(subPath).addAttrib("description", "y component");
        designCoGY.set_value(0.0);
        designCoGY.update(rtIO_->acxml.at(subPath));
    } else {
        designCoGY.read(rtIO_->acxml);
    }
    subPath = "analysis/masses_cg_inertia/design_mass/mass_properties/center_of_gravity/z";
    if (!(rtIO_->acxml.find(subPath))) {
        rtIO_->acxml[subPath];
        rtIO_->acxml.at(subPath).addAttrib("description", "z component");
        designCoGZ.set_value(0.0);
        designCoGZ.update(rtIO_->acxml.at(subPath));
    } else {
        designCoGZ.read(rtIO_->acxml);
    }
    subPath = "analysis/masses_cg_inertia/most_forward_mass/mass_properties/center_of_gravity/x";
    if (!(rtIO_->acxml.find(subPath))) {
        rtIO_->acxml[subPath];
        rtIO_->acxml.at(subPath).addAttrib("description", "x component");
        forwardCoGX.set_value(0.0);
        forwardCoGX.update(rtIO_->acxml.at(subPath));
    } else {
        forwardCoGX.read(rtIO_->acxml);
    }
    subPath = "analysis/masses_cg_inertia/most_forward_mass/mass_properties/center_of_gravity/y";
    if (!(rtIO_->acxml.find(subPath))) {
        rtIO_->acxml[subPath];
        rtIO_->acxml.at(subPath).addAttrib("description", "y component");
        forwardCoGY.set_value(0.0);
        forwardCoGY.update(rtIO_->acxml.at(subPath));
    } else {
        forwardCoGY.read(rtIO_->acxml);
    }
    subPath = "analysis/masses_cg_inertia/most_forward_mass/mass_properties/center_of_gravity/z";
    if (!(rtIO_->acxml.find(subPath))) {
        rtIO_->acxml[subPath];
        rtIO_->acxml.at(subPath).addAttrib("description", "z component");
        forwardCoGZ.set_value(0.0);
        forwardCoGZ.update(rtIO_->acxml.at(subPath));
    } else {
        forwardCoGZ.read(rtIO_->acxml);
    }
    subPath = "analysis/masses_cg_inertia/most_afterward_mass/mass_properties/center_of_gravity/x";
    if (!(rtIO_->acxml.find(subPath))) {
        rtIO_->acxml[subPath];
        rtIO_->acxml.at(subPath).addAttrib("description", "x component");
        aftCoGX.set_value(0.0);
        aftCoGX.update(rtIO_->acxml.at(subPath));
    } else {
        aftCoGX.read(rtIO_->acxml);
    }
    subPath = "analysis/masses_cg_inertia/most_afterward_mass/mass_properties/center_of_gravity/y";
    if (!(rtIO_->acxml.find(subPath))) {
        rtIO_->acxml[subPath];
        rtIO_->acxml.at(subPath).addAttrib("description", "y component");
        aftCoGY.set_value(0.0);
        aftCoGY.update(rtIO_->acxml.at(subPath));
    } else {
        aftCoGY.read(rtIO_->acxml);
    }
    subPath = "analysis/masses_cg_inertia/most_afterward_mass/mass_properties/center_of_gravity/z";
    if (!(rtIO_->acxml.find(subPath))) {
        rtIO_->acxml[subPath];
        rtIO_->acxml.at(subPath).addAttrib("description", "z component");
        aftCoGZ.set_value(0.0);
        aftCoGZ.update(rtIO_->acxml.at(subPath));
    } else {
        aftCoGZ.read(rtIO_->acxml);
    }
    designCoG.xCoordinate = designCoGX.value();
    designCoG.yCoordinate = designCoGY.value();
    designCoG.zCoordinate = designCoGZ.value();
    forwardCoG.xCoordinate = forwardCoGX.value();
    forwardCoG.yCoordinate = forwardCoGY.value();
    forwardCoG.zCoordinate = forwardCoGZ.value();
    aftCoG.xCoordinate = aftCoGX.value();
    aftCoG.yCoordinate = aftCoGY.value();
    aftCoG.zCoordinate = aftCoGZ.value();
    CoGPositions["design"] = designCoG;
    CoGPositions["forward"] = forwardCoG;
    CoGPositions["aft"] = aftCoG;
    numberDevices = rtIO_->acxml.getVector(("wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices/control_device"), 3).size();
    // read in device names to later distinguish between control and high lift device
    for (size_t deviceID(0); deviceID < numberDevices; ++deviceID) {
        wingDevice aWingDevice;
        aWingDevice.type =
                    std::string(rtIO_->acxml.at("/aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices/control_device@"
                    + num2Str(deviceID) + "/type/value"));
        aWingDevice.maxNegativeDeflection =
                    rtIO_->acxml.at("/aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices/control_device@"
                            + num2Str(deviceID) + "/deflection/full_negative_deflection/value");
        aWingDevice.maxPositiveDeflection =
                    rtIO_->acxml.at("/aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices/control_device@"
                            + num2Str(deviceID) + "/deflection/full_positive_deflection/value");
        aWingDevice.innerRelSpanPosition =
                    rtIO_->acxml.at("/aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices/control_device@"
                            + num2Str(deviceID) + "/position/inner_position/spanwise/value");
        aWingDevice.innerRelFwdChordPosition =
                    rtIO_->acxml.at("/aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices/control_device@"
                            + num2Str(deviceID) + "/position/inner_position/chord/from/value");
        aWingDevice.innerRelAftChordPosition =
                    rtIO_->acxml.at("/aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices/control_device@"
                            + num2Str(deviceID) + "/position/inner_position/chord/to/value");
        aWingDevice.outerRelSpanPosition =
                    rtIO_->acxml.at("/aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices/control_device@"
                            + num2Str(deviceID) + "/position/outer_position/spanwise/value");
        aWingDevice.outerRelFwdChordPosition =
                    rtIO_->acxml.at("/aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices/control_device@"
                            + num2Str(deviceID) + "/position/outer_position/chord/from/value");
        aWingDevice.outerRelAftChordPosition =
                    rtIO_->acxml.at("/aircraft_exchange_file/component_design/wing/specific/geometry/aerodynamic_surface@0/parameters/control_devices/control_device@"
                            + num2Str(deviceID) + "/position/outer_position/chord/to/value");
        wingDevices.push_back(aWingDevice);
    }
}
