/*  Copyright (C) 2010-2024 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "bwbDefaultStrategy.h"

#include <aircraftGeometry2/hull_surface.h>
#include <aircraftGeometry2/processing/measure.h>
#include <runtimeInfo/runtimeInfo.h>
#include <moduleBasics/runtimeIO.h>
#include <unitConversion/unitConversion.h>
#include <CGAL/Surface_mesh/IO/PLY.h>
#include <aircraftGeometry2/processing/transform.h>

#include <algorithm>
#include <exception>
#include <iomanip>
#include <string>
#include <memory>
#include <vector>
#include <filesystem>
#include <utility>  // make_pair

#include "../polar.h"
#include "../methods/generalMethods/auxFunctions.h"
#include "../methods/liftingLinePolar.h"
#include "../methods/viscDragRaymer.h"
#include "../methods/waveDragMason.h"
#include "../methods/semiEmpiricalHighLiftAdaption.h"

using std::endl;
using std::exception;
using std::fixed;
using std::ofstream;
using std::setprecision;
using std::string;
using std::stringstream;
using std::pair;
using std::vector;

bwbDefaultStrategy::bwbDefaultStrategy(const std::shared_ptr<RuntimeIO>& rtIO, const std::shared_ptr<bwbCalculatePolarConfig>& config)
    :
    HLconfigs{"takeoff", "takeoff_landing_gear_retracted", "climb", "approach", "approach_landing_gear_out", "landing"},
    rtIO_(rtIO),
    data_(std::make_shared<bwbDefaultData>(rtIO_)),
    config_(config) {
}

void bwbDefaultStrategy::run() {
    this->createAircraftGeometry();
    this->runAeroCalculations(data_->CoGPositions.at(config_->CGPositions.front()));
    this->copyCleanPolars();
    this->copyHighLiftPolars();
}

void bwbDefaultStrategy::createAircraftGeometry() {
    std::filesystem::path file{rtIO_->acxml.getName()};
    std::shared_ptr<node> AcXML = aixml::openDocument(file);
    // create wing
    geom2::WingFactory wingFactory{AcXML, rtIO_->getAirfoilDataDir()};
    geom2::MultisectionSurface<geom2::AirfoilSection> mainWing = wingFactory.create("wing/specific/geometry/aerodynamic_surface@0");
    liftingSurfaces.push_back(mainWing);
    // create VTP
    geom2::WingFactory VTPFactory{AcXML, rtIO_->getAirfoilDataDir()};
    VTP = VTPFactory.create("empennage/specific/geometry/aerodynamic_surface@1");
    geom2::Mesh wing_mesh = geom2::transform::to_mesh(mainWing);
    geom2::Mesh BWB_mesh = geom2::transform::to_mesh(VTP);
    BWB_mesh += wing_mesh;
    CGAL::IO::write_STL("BWB_mesh.stl", BWB_mesh);
}

void bwbDefaultStrategy::runAeroCalculations(const Point CoGPosition) {
    // lifting line for TAW
    LiLiPolarInitialization myLiLiPolarInit;
    this->initializeLiLiPolar(&myLiLiPolarInit);
    liftingLinePolar* myLiftingLinePolar = new liftingLinePolar(rtIO_, config_->liftingLineFolderPath, myLiLiPolarInit);
    myLiftingLinePolar->run(CoGPosition);
    this->getPolarDataFromLiftingLine(*myLiftingLinePolar);
    // viscous drag by raymer
    viscDragRaymerInitialization myViscDragInit;
    this->initializeViscDragRaymer(&myViscDragInit);
    viscDragRaymer* myViscDragRamyer = new viscDragRaymer(rtIO_, myViscDragInit);
    this->calculateViscDragRaymer(myViscDragRamyer);
    // wave drag by mason
    waveDragInitialization myWaveDragInit;
    this->initializeWaveDragMason(&myWaveDragInit);
    waveDragMason* myWaveDragMason = new waveDragMason(rtIO_, myWaveDragInit);
    this->calculateWaveDragMason(myWaveDragMason);
    this->sumUpDragComponents();
    // high lift calculation
    semiEmpiricalHighLiftAdaption* mySemiEmpHLAdaption = new semiEmpiricalHighLiftAdaption(rtIO_, liftingSurfaces.front());
    this->copyWingDevicesAndStabilizer(mySemiEmpHLAdaption);
    this->copyCleanPolarForHLCalculation(mySemiEmpHLAdaption, myLiftingLinePolar);
    mySemiEmpHLAdaption->runHighLiftAdaption();
    this->calculateHighLiftPolars(mySemiEmpHLAdaption, CoGPosition.xCoordinate);
}

void bwbDefaultStrategy::initializeLiLiPolar(LiLiPolarInitialization *myLiLiPolarInit) {
    myLiLiPolarInit->the_delta_isa = data_->delta_isa;
    myLiLiPolarInit->CLModeAirfoilSelection = config_->CLModeAirfoilSelection.value();
    myLiLiPolarInit->stepWidthCL = config_->stepWidthCL.value();
    myLiLiPolarInit->delta_CM0 = config_->delta_CM0.value();
    myLiLiPolarInit->delta_dCMdCL = config_->delta_dCMdCL.value();
    myLiLiPolarInit->indDragCtCorrForCalibration = config_->indDragCtCorrForCalibration.value();
    myLiLiPolarInit->factorDragCleanPolar = config_->factorIndDragCleanPolar.value();
    myLiLiPolarInit->bestCruiseCL = data_->bestCruiseCL.value();
    myLiLiPolarInit->initialMachCruise = data_->initialCruiseMachNumber;
    for (size_t flightConditionID(0); flightConditionID < config_->flightConditions.size(); ++flightConditionID) {
        flightCondition* aLiLiFlightCondition = new flightCondition;
        aLiLiFlightCondition->Mach = config_->flightConditions.at(flightConditionID).Mach;
        aLiLiFlightCondition->Altitude = config_->flightConditions.at(flightConditionID).Altitude;
        aLiLiFlightCondition->extrapolationMargin = config_->flightConditions.at(flightConditionID).extrapolationMargin;
        aLiLiFlightCondition->allowGridChange = config_->flightConditions.at(flightConditionID).allowGridChange;
        myLiLiPolarInit->flightConditions.push_back(*aLiLiFlightCondition);
        if (accuracyCheck(aLiLiFlightCondition->Mach, data_->initialCruiseMachNumber, ACCURACY_HIGH)) {
            myLiLiPolarInit->cruiseMachID = flightConditionID;
        }
        delete aLiLiFlightCondition;
    }
    myLiLiPolarInit->liftingSurfaces = liftingSurfaces;
}

void bwbDefaultStrategy::initializeViscDragRaymer(viscDragRaymerInitialization *myViscDragInit) {
    // copy geometry
    myViscDragInit->theWing = liftingSurfaces.at(0);
    myViscDragInit->calculateWing = true;
    myViscDragInit->theVTP = VTP;
    myViscDragInit->calculateVTP = true;
    // initialize transition parameters
    myViscDragInit->manualTransitionLocation = config_->manualTransition.value();
    myViscDragInit->transitionLocationWing = config_->manualTransitionWing.value();
    myViscDragInit->transitionLocationVTP = config_->manualTransitionFin.value();
    // initialize calibration parameters
    myViscDragInit->doCalibration = config_->doViscDragCalibration.value();
    myViscDragInit->CDvisc_CDsum = config_->CDvisc_CDsum.value();
    myViscDragInit->CDvisc_CLexp = config_->CDvisc_CLexp.value();
    myViscDragInit->CDvisc_CLfact = config_->CDvisc_CLfact.value();
    myViscDragInit->CDvisc_CDsum_lowMa = config_->CDvisc_CDsum_lowMa.value();
    myViscDragInit->CDvisc_CLexp_lowMa = config_->CDvisc_CLexp_lowMa.value();
    myViscDragInit->CDvisc_CLfact_lowMa = config_->CDvisc_CLfact_lowMa.value();
    // initialize drag count corrections
    myViscDragInit->doViscDragCtCorrForCalibration = config_->doViscousDragRaymerDragCountCorrection.value();
    myViscDragInit->viscDragCtCorrForCalibrationWing = config_->deltaViscousDragCountWing.value();
    myViscDragInit->viscDragCtCorrForCalibrationVTP = config_->deltaViscousDragCountVTP.value();
    myViscDragInit->viscDragCtCorrForCalibrationNacelles = config_->deltaViscousDragCountNacelles.value();
    myViscDragInit->viscDragCtCorrForCalibrationMisc = config_->deltaViscousDragCountMisc.value();
}

void bwbDefaultStrategy::initializeWaveDragMason(waveDragInitialization *myWaveDragInit) {
    // fill configuration parameters and values in the wave drag method
    myWaveDragInit->CLExp = config_->CDwave_CLExp.value();
    myWaveDragInit->CLFact = config_->CDwave_CLFact.value();
    myWaveDragInit->factorKa = config_->Ka.value();
    myWaveDragInit->initialMachCruise = data_->initialCruiseMachNumber;
    myWaveDragInit->numberOfSegments = config_->numberWingStrips.value();
    myWaveDragInit->waveDragCtCorrForCalibrationWing = config_->deltaWaveDragCountWing.value();
    myWaveDragInit->maxAllowedSweep = convertUnit(RADIAN, DEGREE, config_->maximumSegmentSweep.value());
    myWaveDragInit->liftingSurfaces = liftingSurfaces;
}

void bwbDefaultStrategy::getPolarDataFromLiftingLine(const liftingLinePolar& myLiftingLinePolar) {
    for (size_t machID(0); machID < config_->flightConditions.size(); ++machID) {
        tawDefaultPolar aCleanPolar;
        aCleanPolar.extrapolationMargin = myLiftingLinePolar.theCleanPolars.at(machID).extrapolationMargin;
        aCleanPolar.allowGridChange = myLiftingLinePolar.theCleanPolars.at(machID).allowGridChange;
        aCleanPolar.altitude = myLiftingLinePolar.theCleanPolars.at(machID).altitude;
        aCleanPolar.MachNumber = myLiftingLinePolar.theCleanPolars.at(machID).MachNumber;
        aCleanPolar.ReynoldsNumber = myLiftingLinePolar.theCleanPolars.at(machID).ReynoldsNumber;
        for (size_t pointID(0); pointID < myLiftingLinePolar.theCleanPolars.at(machID).Points.size(); ++pointID) {
            aCleanPolar.Points.push_back(defaultPolarPoint());
            aCleanPolar.Points.back().CL = myLiftingLinePolar.theCleanPolars.at(machID).Points.at(pointID).CL;
            aCleanPolar.Points.back().AoA = myLiftingLinePolar.theCleanPolars.at(machID).Points.at(pointID).AoA;
            aCleanPolar.Points.back().CD_ind = myLiftingLinePolar.theCleanPolars.at(machID).Points.at(pointID).CDind;
            aCleanPolar.Points.back().CM = myLiftingLinePolar.theCleanPolars.at(machID).Points.at(pointID).CM;
            aCleanPolar.Points.back().iStabPolar = myLiftingLinePolar.theCleanPolars.at(machID).Points.at(pointID).iStabPolar;
            aCleanPolar.PointsWing.push_back(defaultPolarPointLiftSurface());
            aCleanPolar.PointsWing.back().yStations = myLiftingLinePolar.theCleanPolars.at(machID).liftingSurfacePolars.at(0).at(pointID).yStations;
            aCleanPolar.PointsWing.back().Cl_distr = myLiftingLinePolar.theCleanPolars.at(machID).liftingSurfacePolars.at(0).at(pointID).Cl_distr;
            aCleanPolar.PointsWing.back().Lift_distr = myLiftingLinePolar.theCleanPolars.at(machID).liftingSurfacePolars.at(0).at(pointID).Lift_distr;
            aCleanPolar.PointsWing.back().CL = myLiftingLinePolar.theCleanPolars.at(machID).liftingSurfacePolars.at(0).at(pointID).CL;
            aCleanPolar.PointsWing.back().CD_ind = myLiftingLinePolar.theCleanPolars.at(machID).liftingSurfacePolars.at(0).at(pointID).CDind;
            aCleanPolar.PointsFin.push_back(defaultPolarPoint());
        }
        currentPolarSet.cleanPolars.insert(std::make_pair(machID, aCleanPolar));
    }
    LILI_Version = myLiftingLinePolar.LILI_version;
}

void bwbDefaultStrategy::calculateViscDragRaymer(viscDragRaymer *myViscDragRaymer) {
    myViscDragRaymer->preprocessGeometry();
    // run viscous drag calculation by Raymer over the current polar set
    for (std::pair<size_t, tawDefaultPolar> polarIterator : currentPolarSet.cleanPolars) {
        size_t polarID = polarIterator.first;
        vector<double> theComponentDrags = myViscDragRaymer->calculateViscDrag(currentPolarSet.cleanPolars.at(polarID), 0.0, 0.0, 0.0);
        // loop over all points to calculate CL dependent visc drag correction
        for (size_t pointID(0); pointID < currentPolarSet.cleanPolars.at(polarID).Points.size(); ++pointID) {
            vector <double> theCorrectedComponentDrags = myViscDragRaymer->calculateCorrectedViscDrag(theComponentDrags, currentPolarSet.cleanPolars.at(polarID).MachNumber,
                                                                                                        currentPolarSet.cleanPolars.at(polarID).Points.at(pointID).CL);
            // assign total and componentwise viscous drag to the polar map
            currentPolarSet.cleanPolars.at(polarID).Points.at(pointID).CD_visc = std::accumulate(theCorrectedComponentDrags.begin(), theCorrectedComponentDrags.end(), 0.0);
            currentPolarSet.cleanPolars.at(polarID).PointsWing.at(pointID).CD_visc = theCorrectedComponentDrags.at(0);
            currentPolarSet.cleanPolars.at(polarID).PointsFin.at(pointID).CD_visc = theCorrectedComponentDrags.at(2);
        }
    }
    myRuntimeInfo->out<< "... Viscous drag has been calculated sucessfully!" << endl;
}

void bwbDefaultStrategy::calculateWaveDragMason(waveDragMason *myWaveDragMason) {
    // run wave drag calculation by Mason over the current polar set
    for (std::pair<size_t, tawDefaultPolar> polarIterator : currentPolarSet.cleanPolars) {
        size_t polarID = polarIterator.first;
        for (size_t pointID(0); pointID < currentPolarSet.cleanPolars.at(polarID).Points.size(); ++pointID) {
            currentPolarSet.cleanPolars.at(polarID).PointsWing.at(pointID).CD_wave
                                            = myWaveDragMason->calculateWaveDrag(currentPolarSet.cleanPolars.at(polarID).Points.at(pointID),
                                                                                currentPolarSet.cleanPolars.at(polarID).MachNumber,
                                                                                0, 0, currentPolarSet.cleanPolars.at(polarID).PointsWing.at(pointID).yStations,
                                                                                currentPolarSet.cleanPolars.at(polarID).PointsWing.at(pointID).Cl_distr, true);
            currentPolarSet.cleanPolars.at(polarID).Points.at(pointID).CD_wave = currentPolarSet.cleanPolars.at(polarID).PointsWing.at(pointID).CD_wave;
        }
    }
    myRuntimeInfo->out << "... Wave drag has been calculated sucessfully!" << endl;
}

void bwbDefaultStrategy::sumUpDragComponents() {
    myRuntimeInfo->out << "Sum up drag components" << endl;
    for (size_t polarID(0); polarID < currentPolarSet.cleanPolars.size(); ++polarID) {
        for (size_t pointID(0); pointID < currentPolarSet.cleanPolars.at(polarID).Points.size(); ++pointID) {
            currentPolarSet.cleanPolars.at(polarID).Points.at(pointID).CD =  currentPolarSet.cleanPolars.at(polarID).Points.at(pointID).CD_ind
                                                                          +  currentPolarSet.cleanPolars.at(polarID).Points.at(pointID).CD_visc
                                                                          +  currentPolarSet.cleanPolars.at(polarID).Points.at(pointID).CD_wave;
        }
    }
}

void bwbDefaultStrategy::copyWingDevicesAndStabilizer(semiEmpiricalHighLiftAdaption *mySemiEmpHLAdaption) {
    myRuntimeInfo->out << "Copy wing devices" << endl;
    for (size_t deviceID(0); deviceID < data_->numberDevices; ++deviceID) {
        semiEmpiricalHighLiftAdaption::WingDevice* aWingDevice = new semiEmpiricalHighLiftAdaption::WingDevice;
        aWingDevice->name = data_->wingDevices.at(deviceID).type;
        aWingDevice->innerRelFwdChordPosition = data_->wingDevices.at(deviceID).innerRelFwdChordPosition;
        aWingDevice->innerRelAftChordPosition = data_->wingDevices.at(deviceID).innerRelAftChordPosition;
        aWingDevice->outerRelFwdChordPosition = data_->wingDevices.at(deviceID).outerRelFwdChordPosition;
        aWingDevice->outerRelAftChordPosition = data_->wingDevices.at(deviceID).outerRelAftChordPosition;
        aWingDevice->innerSpan_rel = data_->wingDevices.at(deviceID).innerRelSpanPosition;
        aWingDevice->outerSpan_rel = data_->wingDevices.at(deviceID).outerRelSpanPosition;
        mySemiEmpHLAdaption->wingDevices.push_back(*aWingDevice);
    }
}

void bwbDefaultStrategy::copyCleanPolarForHLCalculation(semiEmpiricalHighLiftAdaption *mySemiEmpHLAdaption, liftingLinePolar *myLiftingLinePolar) {
    // find polar with minimum mach number
    double minMachNumber(1.0);
    size_t minMachID;
    for (std::pair<size_t, tawDefaultPolar> polarIterator : currentPolarSet.cleanPolars) {
        if (polarIterator.second.MachNumber < minMachNumber) {
            minMachID = polarIterator.first;
            minMachNumber = polarIterator.second.MachNumber;
        }
    }
    // copy clean polar to HL method
    mySemiEmpHLAdaption->theCleanPolar.MachNumber = currentPolarSet.cleanPolars.at(minMachID).MachNumber;
    mySemiEmpHLAdaption->theCleanPolar.altitude = currentPolarSet.cleanPolars.at(minMachID).altitude;
    mySemiEmpHLAdaption->theCleanPolar.extrapolationMargin = currentPolarSet.cleanPolars.at(minMachID).extrapolationMargin;
    mySemiEmpHLAdaption->theCleanPolar.allowGridChange = currentPolarSet.cleanPolars.at(minMachID).allowGridChange;
    mySemiEmpHLAdaption->theCleanPolar.ReynoldsNumber = currentPolarSet.cleanPolars.at(minMachID).ReynoldsNumber;
    mySemiEmpHLAdaption->theCleanPolar.Config = currentPolarSet.cleanPolars.at(minMachID).Config;
    for (size_t pointID(0); pointID < currentPolarSet.cleanPolars.at(minMachID).Points.size(); ++pointID) {
        mySemiEmpHLAdaption->theCleanPolar.PolarPoints.push_back(PolarPoint());
        mySemiEmpHLAdaption->theCleanPolar.PolarPoints.back().AoA = currentPolarSet.cleanPolars.at(minMachID).Points.at(pointID).AoA;
        mySemiEmpHLAdaption->theCleanPolar.PolarPoints.back().CL = currentPolarSet.cleanPolars.at(minMachID).Points.at(pointID).CL;
        mySemiEmpHLAdaption->theCleanPolar.PolarPoints.back().CD = currentPolarSet.cleanPolars.at(minMachID).Points.at(pointID).CD;
        mySemiEmpHLAdaption->theCleanPolar.PolarPoints.back().CM = currentPolarSet.cleanPolars.at(minMachID).Points.at(pointID).CM;
        mySemiEmpHLAdaption->theCleanPolar.PolarPoints.back().iStabPolar = currentPolarSet.cleanPolars.at(minMachID).Points.at(pointID).iStabPolar;
    }
    // copy other neccesary values from lifting line calculation
    mySemiEmpHLAdaption->CLmaxTO_req = 1.0;
    mySemiEmpHLAdaption->CLmaxLDG_req = 1.0;
    mySemiEmpHLAdaption->stepWidthCL = config_->stepWidthCL.value();
    mySemiEmpHLAdaption->reductionDragCountsHighLift = config_->deltaTotalDragHL.value();
    mySemiEmpHLAdaption->factorDragHighLift = config_->factorDragHL.value();
    mySemiEmpHLAdaption->CDmin_inc = myLiftingLinePolar->theAnalyticalPolarData.CDmin_inc.at(0);
    mySemiEmpHLAdaption->kFactor_inc = myLiftingLinePolar->theAnalyticalPolarData.kFactor_inc.at(0);
    mySemiEmpHLAdaption->CLatCDmin_inc = myLiftingLinePolar->theAnalyticalPolarData.CLatCDmin_inc.at(0);
    mySemiEmpHLAdaption->dCMtodCL_inc = myLiftingLinePolar->theAnalyticalPolarData.dCMtodCL_inc.at(0);
    mySemiEmpHLAdaption->CMatCL0_inc = myLiftingLinePolar->theAnalyticalPolarData.CMatCL0_inc.at(0);
    mySemiEmpHLAdaption->dCMtodAoA = myLiftingLinePolar->theAnalyticalPolarData.dCMtodAoA.at(0);
    mySemiEmpHLAdaption->dCLtodAoA = myLiftingLinePolar->theAnalyticalPolarData.dCLtodAoA.at(0);
    mySemiEmpHLAdaption->CLatAoA0 = myLiftingLinePolar->theAnalyticalPolarData.CLatAoA0.at(0);
    mySemiEmpHLAdaption->CMatAoA0 = myLiftingLinePolar->theAnalyticalPolarData.CMatAoA0.at(0);
}

void bwbDefaultStrategy::calculateHighLiftPolars(semiEmpiricalHighLiftAdaption *mySemiEmpHLAdaption, const double CoG_X_Position) {
    for (size_t configID(0); configID < HLconfigs.size(); ++configID) {
        myRuntimeInfo->out << "Calculate high lift polar for " << HLconfigs.at(configID) << endl;
        Polar aPolar = mySemiEmpHLAdaption->calculateHighLiftPolar(HLconfigs.at(configID), CoG_X_Position, 0.0);
        currentPolarSet.highLiftPolars[HLconfigs.at(configID)] = tawDefaultPolar();
        // copy polar data
        currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).Config = HLconfigs.at(configID);
        currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).extrapolationMargin = aPolar.extrapolationMargin;
        currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).allowGridChange = aPolar.allowGridChange;
        currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).altitude = aPolar.altitude;
        currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).MachNumber = aPolar.MachNumber;
        currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).ReynoldsNumber = aPolar.ReynoldsNumber;
        for (size_t pointID(0); pointID < aPolar.PolarPoints.size(); ++pointID) {
            currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).Points.push_back(defaultPolarPoint());
            currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).Points.back().AoA = aPolar.PolarPoints.at(pointID).AoA;
            currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).Points.back().CL = aPolar.PolarPoints.at(pointID).CL;
            currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).Points.back().CD = aPolar.PolarPoints.at(pointID).CD;
            currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).Points.back().CM = aPolar.PolarPoints.at(pointID).CM;
            currentPolarSet.highLiftPolars.at(HLconfigs.at(configID)).Points.back().iStabPolar = aPolar.PolarPoints.at(pointID).iStabPolar;
        }
    }
}

void bwbDefaultStrategy::copyCleanPolars() {
    for (size_t polarID(0); polarID < currentPolarSet.cleanPolars.size(); ++polarID) {
        polarDataTrimMap[0].cleanPolars[polarID] = currentPolarSet.cleanPolars.at(polarID);
    }
    polarDataCGMap[0] = polarDataTrimMap;
}

void bwbDefaultStrategy::copyHighLiftPolars() {
    for (size_t configID(0); configID < HLconfigs.size(); ++configID) {
        polarDataTrimMap[0].highLiftPolars[HLconfigs.at(configID)] = currentPolarSet.highLiftPolars.at(HLconfigs.at(configID));
    }
}

void bwbDefaultStrategy::writePolarData() {
     ofstream plot;
    /* plot all untrimmed polars */
    for (size_t CGID(0); CGID < polarDataCGMap.size(); ++CGID) {
        for (const pair<const size_t, PolarData>& thePolarIterator : polarDataCGMap.at(CGID)) {
            /* plot all clean polars */
            for (const pair<const size_t, tawDefaultPolar>& theMachIDIterator : thePolarIterator.second.cleanPolars) {
                stringstream machNumber;
                stringstream iStabStream;
                machNumber << setprecision(2) << fixed << theMachIDIterator.second.MachNumber;
                std::string csvFiles = rtIO_->getCsvFilesDir() + "/bwbCalculatePolar_M"
                + replaceAll(machNumber.str(), ".", "") + "_" + "design_plot.csv";
                /*
                csvFiles.push_back(mySettingsPt->getCsvFilesDirTool() + TOOL_NAME + "_M" + replaceAll(machNumber.str(), ".", "") + "_"
                                       + CGCases.at(CGID) + "_" + replaceAll(iStabStream.str(), ".", "") +"_plot.csv");
                                       */
                plot.open(csvFiles.c_str());
                if (!plot) {
                    myRuntimeInfo->err << csvFiles + " could not be opened!" << endl;
                    exit(1);
                } else {
                    plot << "# " + csvFiles << endl;
                    plot << "# Clean Polar" << endl;
                    plot << "# Mach Number " << theMachIDIterator.second.MachNumber << endl;
                    plot << "AoA; C_L; C_Dtotal; C_Dind; C_Dvisc; C_Dwave; C_M; L/D;" << endl;
                    for (size_t i(0); i < theMachIDIterator.second.Points.size(); ++i) {
                        plot << setprecision(3) << fixed << theMachIDIterator.second.Points.at(i).AoA << ";";
                        plot << setprecision(4) << fixed << theMachIDIterator.second.Points.at(i).CL << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.Points.at(i).CD << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.Points.at(i).CD_ind << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.Points.at(i).CD_visc << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.Points.at(i).CD_wave << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.Points.at(i).CM << ";";
                        plot << setprecision(6) << fixed << theMachIDIterator.second.Points.at(i).LoverD << ";";
                        plot << endl;
                    }
                }
                plot.close();
            }
        }
    }
    size_t cruiseMachID(0);
    stringstream machNumber;
    machNumber << setprecision(2) << fixed << data_->initialCruiseMachNumber;
    for (size_t machID(0); machID < polarDataTrimMap.at(0).cleanPolars.size(); ++machID) {
        if (accuracyCheck(polarDataTrimMap.at(0).cleanPolars.at(machID).MachNumber, data_->initialCruiseMachNumber, ACCURACY_HIGH)) {
            cruiseMachID = machID;
            break;
        }
    }
    std::string csvFiles = rtIO_->getCsvFilesDir() + "/bwbCalculatePolar_liftDistribution_"
    + replaceAll(machNumber.str(), ".", "") +"_plot.csv";
    plot.open(csvFiles.c_str());
    if (!plot) {
        myRuntimeInfo->err << csvFiles + " could not be opened!" << endl;
        exit(1);
    } else {
        plot << "# " + csvFiles << endl;
        plot << "# (1)C_L; (2)C_Lw; (3)y, m; (4)C_l; (5)Cl*l, m" << endl;
        for (size_t pointID(0); pointID < polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).PointsWing.size(); ++pointID) {
            vector <double> clDistribution_tmp = polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).PointsWing.at(pointID).Cl_distr;
            vector <double> yStations_tmp      = polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).PointsWing.at(pointID).yStations;
            reverse(clDistribution_tmp.begin(), clDistribution_tmp.end());
            reverse(yStations_tmp.begin(), yStations_tmp.end());
            // for Cl distr analysis at single CL
            double CL_tmp(0.51);
            if (true || fabs(polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).CL - CL_tmp) < 1e-5) {  // uncomment "true || " for analysis
                string delimiter(";");
                for (size_t j(0); j < clDistribution_tmp.size(); ++j) {
                    plot << polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).CL << delimiter;
                    plot << polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).PointsWing.at(pointID).CL << delimiter;
                    plot << yStations_tmp.at(j) << delimiter;
                    plot << clDistribution_tmp.at(j) << delimiter;
                    plot << getLocalChordLength(liftingSurfaces.front(), yStations_tmp.at(j)) << endl;
                }
                plot << endl << endl;
            }
        }
        plot.close();
    }
}

void bwbDefaultStrategy::writePolarXML() {
    myRuntimeInfo->out << "Create XML-Polar file ..." << endl;
    for (size_t CGID(0); CGID < polarDataTrimMap.size(); ++CGID) {
        polarXMLFilePath = rtIO_->getAeroDataDir() + "/";
        polarXMLFilePath += rtIO_->aircraft_model();
        this->polarXMLFilePath += "_" + config_->CGPositions.at(CGID) + "_polar.xml";
        this->designPolarXMLFilePath = polarXMLFilePath;
        this->designPolarXMLFilePath = designPolarXMLFilePath.substr(designPolarXMLFilePath.find_last_of("/") + 1);
        ofstream XML;
        XML.open(polarXMLFilePath.c_str());
        if (!XML) {
            myRuntimeInfo->err << polarXMLFilePath + " could not be opened!" << endl;
            exit(1);
        } else {
            XML << "<?xml version=\"1.0\"?>" << endl;
            XML << "<polar-exchange-file version=\"1.0\">" << endl;
            XML << "    <comment>The 'PEF-Browser' application, a polar file converter is available in the software section at:";
            XML << " http://www.MH-AeroTools.de/airfoils/</comment>" << endl;
            XML << "    <liftingLineVersion>" << LILI_Version << "</liftingLineVersion>" << endl;
            XML << "    <dragComponentMethods>" << endl;
            XML << "        <inducedDrag>Lifting Line + Corrections for fuselage and nacelles</inducedDrag>" << endl;
            XML << "        <viscousDrag>" << "Raymer" << "</viscousDrag>" << endl;
            XML << "        <waveDrag>" << "Mason" << "</waveDrag>" << endl;
            XML << "    </dragComponentMethods>" << endl;
            XML << "    <configurations>" << endl;
            XML << "        <name>" << rtIO_->aircraft_model() << "</name>" << endl;
            XML << "        <comment>Polars for wing configuration</comment>" << endl;
            XML << "        <configuration ID=\"clean\">" << endl;
            XML << "            <name>Clean</name>" << endl;
            XML << "            <comment>Clean Configuration</comment>" << endl;
            for (size_t MachID(0); MachID < polarDataTrimMap.at(CGID).cleanPolars.size(); MachID++) {
                XML << "                <polar ID=\"" << MachID + 1 << "\" ExtrapolationMargin=\"" << polarDataTrimMap.at(CGID).cleanPolars.at(MachID).extrapolationMargin
                                    << "\" AllowGridChange=\"" << polarDataTrimMap.at(CGID).cleanPolars.at(MachID).allowGridChange <<"\">" << endl;
                XML << "                    <name>" << "---" << "</name>" << endl;
                XML << "                    <reynoldsnumber>" <<polarDataTrimMap.at(CGID).cleanPolars.at(MachID).ReynoldsNumber << "</reynoldsnumber>" << endl;
                XML << "                    <altitude>" <<polarDataTrimMap.at(CGID).cleanPolars.at(MachID).altitude << "</altitude>" << endl;
                XML << "                    <machnumber>" << polarDataTrimMap.at(CGID).cleanPolars.at(MachID).MachNumber << "</machnumber>" << endl;
                std::string Alpha_str(""), Cl_str(""), Cd_str(""), Cd_turb_str(""), Cm_str(""), origin_polar_str(""), i_stab_str("");
                for (size_t i(0); i < polarDataTrimMap.at(CGID).cleanPolars.at(MachID).Points.size(); ++i) {
                    Alpha_str += num2Str(polarDataTrimMap.at(CGID).cleanPolars.at(MachID).Points.at(i).AoA);
                    Cl_str += num2Str(polarDataTrimMap.at(CGID).cleanPolars.at(MachID).Points.at(i).CL);
                    Cd_str += num2Str(polarDataTrimMap.at(CGID).cleanPolars.at(MachID).Points.at(i).CD);
                    Cd_turb_str += num2Str(polarDataTrimMap.at(CGID).cleanPolars.at(MachID).Points.at(i).CD);
                    Cm_str += num2Str(polarDataTrimMap.at(CGID).cleanPolars.at(MachID).Points.at(i).CM);
                    origin_polar_str += "seg0-0.0";
                    i_stab_str += num2Str(polarDataTrimMap.at(CGID).cleanPolars.at(MachID).Points.at(i).iStabPolar);
                    if (i < polarDataTrimMap.at(CGID).cleanPolars.at(MachID).Points.size() - 1) {
                        Alpha_str += ";";
                        Cl_str += ";";
                        Cd_str += ";";
                        Cd_turb_str += ";";
                        Cm_str += ";";
                        origin_polar_str += ";";
                        i_stab_str += ";";
                    }
                }
                XML << "                    <Alpha>" << Alpha_str << "</Alpha>" << endl;
                XML << "                    <Cl>" << Cl_str << "</Cl>" << endl;
                XML << "                    <Cd>" <<Cd_str << "</Cd>" << endl;
                XML << "                    <Cd_turb>" << Cd_turb_str << "</Cd_turb>" << endl;
                XML << "                    <Cm>" << Cm_str << "</Cm>" << endl;
                XML << "                    <origin_polar>"<< origin_polar_str << "</origin_polar>" << endl;
                XML << "                    <iStab_polar>"<< i_stab_str << "</iStab_polar>" << endl;
                XML << "                </polar>" << endl;
            }
            XML << "            </polars>" << endl;
            XML << "        </configuration>" << endl;
            for (const pair<const string, tawDefaultPolar>& thePolarIterator : polarDataTrimMap.at(CGID).highLiftPolars) {
                XML << "        <configuration ID=\"" << thePolarIterator.first << "\">" << endl;
                XML << "            <name>" << thePolarIterator.first << "</name>" << endl;
                XML << "            <comment>" << thePolarIterator.first << " Configuration</comment>" << endl;
                // Inner config loop: Since the myAcftPt->highLiftPolar elements are ordered by config name, configs can be gathered here until the next one appears.
                XML << "                <polar ID=\"" << num2Str(1) << "\" ExtrapolationMargin=\"" << thePolarIterator.second.extrapolationMargin
                                        << "\" AllowGridChange=\"" << thePolarIterator.second.allowGridChange << "\">" << endl;
                XML << "                    <name>" << "---" << "</name>" << endl;
                XML << "                    <reynoldsnumber>" << thePolarIterator.second.ReynoldsNumber << "</reynoldsnumber>" << endl;
                XML << "                    <altitude>" << thePolarIterator.second.altitude << "</altitude>" << endl;
                XML << "                    <machnumber>" <<thePolarIterator.second.MachNumber << "</machnumber>" << endl;
                std::string Alpha_str(""), Cl_str(""), Cd_str(""), Cm_str(""), origin_polar_str(""), i_stab_str("");
                for (size_t j(0); j < thePolarIterator.second.Points.size(); ++j) {
                    Alpha_str += num2Str(thePolarIterator.second.Points.at(j).AoA);
                    Cl_str += num2Str(thePolarIterator.second.Points.at(j).CL);
                    Cd_str += num2Str(thePolarIterator.second.Points.at(j).CD);
                    Cm_str += num2Str(thePolarIterator.second.Points.at(j).CM);
                    origin_polar_str += "seg0-0.0";
                    i_stab_str += num2Str(thePolarIterator.second.Points.at(j).iStabPolar);
                    if (j < thePolarIterator.second.Points.size() - 1) {
                        Alpha_str += ";";
                        Cl_str += ";";
                        Cd_str += ";";
                        Cm_str += ";";
                        origin_polar_str += ";";
                        i_stab_str += ";";
                    }
                }
                XML << "                    <Alpha>" << Alpha_str << "</Alpha>" << endl;
                XML << "                    <Cl>" << Cl_str << "</Cl>" << endl;
                XML << "                    <Cd>" <<Cd_str << "</Cd>" << endl;
                XML << "                    <Cm>" << Cm_str << "</Cm>" << endl;
                XML << "                    <origin_polar>"<< origin_polar_str << "</origin_polar>" << endl;
                XML << "                    <iStab_polar>"<< i_stab_str << "</iStab_polar>" << endl;
                XML << "                </polar>" << endl;
                XML << "            </polars>" << endl;
                XML << "        </configuration>" << endl;
                }
            XML << "    </configurations>" << endl;
            XML << "</polar-exchange-file>" << endl;
        }
        XML.close();
    }
}

void bwbDefaultStrategy::writeXMLOutputData() {
    myRuntimeInfo->out << "write aircraft xml output values..." << endl;
    /* Writing characteristic aerodynamic properties to AC-XML */
    string subPath = "aircraft_exchange_file/analysis/aerodynamics/";
    // reference values
    rtIO_->acxml.at(subPath + "reference_values/S_ref/value") = geom2::measure::reference_area(liftingSurfaces.front());
    rtIO_->acxml.at(subPath + "reference_values/b/value") = getHalfSpan(liftingSurfaces.front());
    rtIO_->acxml.at(subPath + "reference_values/MAC/value") = geom2::measure::mean_aerodynamic_chord(liftingSurfaces.front());
    // lift coefficients
    rtIO_->acxml.at(subPath + "lift_coefficients/C_LoptimumCruise/value") = getCLoptCruise();
    // polar file
    rtIO_->acxml.at(subPath + "polar_file/value") = designPolarXMLFilePath;
}

double bwbDefaultStrategy::getCLoptCruise() {
    double LoverDmax(0.0);
    double CLopt(0.0);
    size_t cruiseMachID(0);
    for (size_t polarID(0); polarID < polarDataTrimMap.at(0).cleanPolars.size(); ++polarID) {
        if (accuracyCheck(polarDataTrimMap.at(0).cleanPolars.at(polarID).MachNumber, data_->initialCruiseMachNumber, ACCURACY_HIGH)) {
            cruiseMachID = polarID;
        }
    }
    for (size_t pointID(0); pointID < polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.size(); ++pointID) {
        double LoverD = polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).CL
                        / polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).CD;
            if (LoverD > LoverDmax) {
                LoverDmax = LoverD;
                CLopt = polarDataTrimMap.at(0).cleanPolars.at(cruiseMachID).Points.at(pointID).CL;
            }
    }
    return CLopt;
}
