/*  Copyright (C) 2010-2024 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "liftingLineForTAW.h"

#include "generalMethods/auxFunctions.h"

#include <aircraftGeometry2/airfoil_surface.h>
#include <aircraftGeometry2/processing/measure.h>
#include <aixml/node.h>
#include <liftingLineInterface2/liftingLineInput.h>
#include <liftingLineInterface2/liftingLineOutput.h>
#include <liftingLineInterface2/liftingLineInterface.h>
#include <runtimeInfo/runtimeInfo.h>
#include <moduleBasics/runtimeIO.h>
#include <CGAL/Exact_predicates_exact_constructions_kernel.h>
#include <CGAL/Polygon_2.h>
#include <unitConversion/unitConversion.h>

#include <memory>
#include <vector>
#include <algorithm>
#include <exception>
#include <iomanip>
#include <string>
#include <utility>  // make_pair

using std::endl;
using std::exception;
using std::fixed;
using std::setprecision;
using std::string;
using std::stringstream;
using std::vector;

/* TODO:    - find way to integrate calibration
            - initialize all the config parameters externally
            - moment coefficients are at least one order of magnitude too high
*/

LiLiforTAWInitialization::LiLiforTAWInitialization()
    :
    liftCorrectionFuselage(true),
    liftCorrectionNacelle(true) {
    // ctor
}

liftingLineForTAW::liftingLineForTAW(const std::shared_ptr<RuntimeIO>& rtIO, const std::string& liftingLinePath,
                                    const LiLiforTAWInitialization& LiLiInit)
    :
    delta_isa(LiLiInit.the_delta_isa),
    rtIO_(rtIO),
    designCL(0.5),
    CLModeAirfoilSelection(LiLiInit.CLModeAirfoilSelection),
    liftingSurfaces(LiLiInit.liftingSurfaces),
    theFuselage(LiLiInit.theFuselage),
    theNacelles(LiLiInit.theNacelles),
    myLiftingLine(liftingLineInterface(liftingLinePath, rtIO->acxml.getName())),
    myLiftingLine_compr(liftingLineInterface(liftingLinePath, rtIO->acxml.getName())),
    bestCruiseCL(LiLiInit.bestCruiseCL),
    initialMachCruise(LiLiInit.initialMachCruise),
    engineType(LiLiInit.engineType),
    stepWidthCL(LiLiInit.stepWidthCL),
    CM_corr_fuselage_mode(LiLiInit.CM_corr_fuselage_mode),
    CM_corr_nacelle_mode(LiLiInit.CM_corr_nacelle_mode),
    delta_CM0(LiLiInit.delta_CM0),
    delta_dCMdCL(LiLiInit.delta_dCMdCL),
    indDragCtCorrForCalibration(LiLiInit.indDragCtCorrForCalibration),
    factorDragCleanPolar(LiLiInit.factorDragCleanPolar),
    flightConditions(LiLiInit.flightConditions),
    cruiseMachID(LiLiInit.cruiseMachID),
    liftCorrectionFuselage(LiLiInit.liftCorrectionFuselage),
    liftCorrectionNacelle(LiLiInit.liftCorrectionNacelle) {
    wingDesignPoint = new LiftDistributionPolarPoint();
    wingPointBestCruiseCL = new LiftDistributionPolarPoint();
    pathToLiftingLine = liftingLinePath;
    IOFilePath = rtIO_->acxml.getName();
    config = rtIO_->aircraft_configuration_type();
    airfoilPath = rtIO_->getAirfoilDataDir();
    currentStabAngle = convertUnit(RADIAN, DEGREE, liftingSurfaces.back().sections.front().get_twist_angle());
}

void liftingLineForTAW::run(const Point& aCOG) {
    myRuntimeInfo->out << "Run lifting line for TAW..." << endl;
    liftCoefficients.resize(flightConditions.size());
    myLiftingLine = this->runLiftingLine(aCOG, 0., 0.);  // hard coded for now
    myLiftingLine_compr = this->runLiftingLine(aCOG, flightConditions.at(cruiseMachID).Mach, flightConditions.at(cruiseMachID).Altitude);
    this->interpolate_LL_results();
    this->setLiftCoefficientVector();
    this->setWingDesignPoints();
    this->setAeroSections();
    this->getCleanPolar();
}

liftingLineInterface liftingLineForTAW::runLiftingLine(const Point& aCOG, const double& machNumber, const double& altitude) {
    myRuntimeInfo->debug << "Run LiftingLine" << endl;
    liftingLineInterface *myLL = new liftingLineInterface(pathToLiftingLine, rtIO_->getIODir() + "/");
    // Create LIFTING_LINE Input File
    // get reynoldsNumber
    double reynoldsNumber = getReynoldsNumber(geom2::measure::mean_aerodynamic_chord(liftingSurfaces.front()), altitude, machNumber, delta_isa);
    myLL->initializeLiftingLine(liftingSurfaces, theFuselage, machNumber, reynoldsNumber, geom2::measure::reference_area(liftingSurfaces.front()), aCOG);
    // Execute LIFTING_LINE
    myLL->executeLiftingLine();
    // Read aerodynamic coefficients from LIFTING_LINE XML output file
    myLL->theLLOutputPt->readLiftingLineResults(liftingSurfaces.size());
    return *myLL;
}

void liftingLineForTAW::interpolate_LL_results() {
    liftingLineInterface *myLL = &myLiftingLine;
    liftingLineInterface *myLL_compr = &myLiftingLine_compr;
    myRuntimeInfo->info << "Interpolate LIFTING_LINE results according to Prandtl-Glauert-rule ..." << endl;
    for (size_t flightConditionID(0); flightConditionID < flightConditions.size(); ++flightConditionID) {
        myLL_vec.push_back(liftingLineInterface(pathToLiftingLine, IOFilePath));
        if (flightConditions.at(flightConditionID).Mach < 0.3) {  // Use incompressible LiLi-calculation for Mach < 0.3 according to [Sch18] p. 317
            myLL_vec.back() = *myLL;
        } else if (flightConditionID == cruiseMachID) {
            myLL_vec.back() = *myLL_compr;
        } else {  // Use Prandtl-Glauert correction for compressibility effects
            myLL_vec.back().theLiftingLineInputPt->machNumber = flightConditions.at(flightConditionID).Mach;
            // coefficients of the overall configuration
            myLL_vec.back().theLLOutputPt->llPolar.resize(myLL->theLLOutputPt->llPolar.size());
            for (size_t LLPolarID(0); LLPolarID < myLL->theLLOutputPt->llPolar.size(); ++LLPolarID) {
                myLL_vec.back().theLLOutputPt->llPolar.at(LLPolarID).AoA    = myLL->theLLOutputPt->llPolar[LLPolarID].AoA;
                myLL_vec.back().theLLOutputPt->llPolar.at(LLPolarID).CL     = PrandtlGlauert_interpol(myLL_vec.back().theLiftingLineInputPt->machNumber,
                        myLL->theLLOutputPt->llPolar.at(LLPolarID).CL, myLL_compr->theLiftingLineInputPt->machNumber, myLL_compr->theLLOutputPt->llPolar.at(LLPolarID).CL);
                myLL_vec.back().theLLOutputPt->llPolar.at(LLPolarID).CDind  = PrandtlGlauert_interpol(myLL_vec.back().theLiftingLineInputPt->machNumber,
                        myLL->theLLOutputPt->llPolar.at(LLPolarID).CDind, myLL_compr->theLiftingLineInputPt->machNumber, myLL_compr->theLLOutputPt->llPolar.at(LLPolarID).CDind);
                myLL_vec.back().theLLOutputPt->llPolar.at(LLPolarID).CM     = PrandtlGlauert_interpol(myLL_vec.back().theLiftingLineInputPt->machNumber,
                        myLL->theLLOutputPt->llPolar.at(LLPolarID).CM, myLL_compr->theLiftingLineInputPt->machNumber, myLL_compr->theLLOutputPt->llPolar.at(LLPolarID).CM);
            }
            // coefficients of wing components
            myLL_vec.back().theLLOutputPt->llPolarWings.resize(myLL->theLLOutputPt->llPolarWings.size());
            for (size_t liftingSurfaceID(0); liftingSurfaceID < myLL->theLLOutputPt->llPolarWings.size(); ++liftingSurfaceID) {
                liftingLineOutput::liftingLinePolarWings* currentAircraftPolar = &(myLL_vec.back().theLLOutputPt->llPolarWings.at(liftingSurfaceID));
                const liftingLineOutput::liftingLinePolarWings& currentLiftingLinePolarOutputIncomp = myLL->theLLOutputPt->llPolarWings.at(liftingSurfaceID);
                const liftingLineOutput::liftingLinePolarWings& currentLiftingLinePolarOutputComp = myLL_compr->theLLOutputPt->llPolarWings.at(liftingSurfaceID);
                currentAircraftPolar->llPolarW.resize(currentLiftingLinePolarOutputIncomp.llPolarW.size());
                for (size_t outpPolarID(0); outpPolarID < currentLiftingLinePolarOutputIncomp.llPolarW.size(); ++outpPolarID) {  // loop over angles of attack
                    currentAircraftPolar->numberOfSegments  = currentLiftingLinePolarOutputIncomp.numberOfSegments;
                    currentAircraftPolar->yPositions        = currentLiftingLinePolarOutputIncomp.yPositions;
                    currentAircraftPolar->llPolarW.resize(currentLiftingLinePolarOutputIncomp.llPolarW.size());
                    currentAircraftPolar->llPolarW.at(outpPolarID).AoA    = currentLiftingLinePolarOutputIncomp.llPolarW.at(outpPolarID).AoA;
                    currentAircraftPolar->llPolarW.at(outpPolarID).CL     =
                            PrandtlGlauert_interpol(myLL_vec.back().theLiftingLineInputPt->machNumber, currentLiftingLinePolarOutputIncomp.llPolarW.at(outpPolarID).CL,
                                                    myLL_compr->theLiftingLineInputPt->machNumber, currentLiftingLinePolarOutputComp.llPolarW.at(outpPolarID).CL);
                    currentAircraftPolar->llPolarW.at(outpPolarID).CDind  =
                            PrandtlGlauert_interpol(myLL_vec.back().theLiftingLineInputPt->machNumber, currentLiftingLinePolarOutputIncomp.llPolarW.at(outpPolarID).CDind,
                                                    myLL_compr->theLiftingLineInputPt->machNumber, currentLiftingLinePolarOutputComp.llPolarW.at(outpPolarID).CDind);
                    currentAircraftPolar->llPolarW.at(outpPolarID).CM     =
                            PrandtlGlauert_interpol(myLL_vec.back().theLiftingLineInputPt->machNumber, currentLiftingLinePolarOutputIncomp.llPolarW.at(outpPolarID).CM,
                                                    myLL_compr->theLiftingLineInputPt->machNumber, currentLiftingLinePolarOutputComp.llPolarW.at(outpPolarID).CM);
                    for (int segmentID(0); segmentID < currentLiftingLinePolarOutputIncomp.numberOfSegments; segmentID++) {  // loop over Lifting Line segments
                        const liftingLinePolars::liftingLineSegment& currentLiftingLineSegmentIncomp
                                                                        (currentLiftingLinePolarOutputIncomp.llPolarW.at(outpPolarID).wingSegments.at(segmentID));
                        const liftingLinePolars::liftingLineSegment& currentLiftingLineSegmentComp
                                                                        (currentLiftingLinePolarOutputComp.llPolarW.at(outpPolarID).wingSegments.at(segmentID));
                        currentAircraftPolar->llPolarW.at(outpPolarID).wingSegments.resize(currentLiftingLinePolarOutputIncomp.llPolarW.at(outpPolarID).wingSegments.size());
                        currentAircraftPolar->llPolarW.at(outpPolarID).wingSegments.at(segmentID).yPos            = currentLiftingLinePolarOutputIncomp.yPositions.at(segmentID);
                        currentAircraftPolar->llPolarW.at(outpPolarID).wingSegments.at(segmentID).llPolarSeg      = new liftingLinePolars::liftingLinePolar();
                        currentAircraftPolar->llPolarW.at(outpPolarID).wingSegments.at(segmentID).llPolarSeg->AoA = currentLiftingLinePolarOutputIncomp.llPolarW.at(
                                    outpPolarID).wingSegments.at(segmentID).llPolarSeg->AoA;
                        currentAircraftPolar->llPolarW.at(outpPolarID).wingSegments.at(segmentID).llPolarSeg->CL  =
                            PrandtlGlauert_interpol(myLL_vec.back().theLiftingLineInputPt->machNumber, currentLiftingLineSegmentIncomp.llPolarSeg->CL,
                                                    myLL_compr->theLiftingLineInputPt->machNumber, currentLiftingLineSegmentComp.llPolarSeg->CL);
                        currentAircraftPolar->llPolarW.at(outpPolarID).wingSegments.at(segmentID).llPolarSeg->CDind =
                            PrandtlGlauert_interpol(myLL_vec.back().theLiftingLineInputPt->machNumber,    currentLiftingLineSegmentIncomp.llPolarSeg->CDind,
                                                    myLL_compr->theLiftingLineInputPt->machNumber, currentLiftingLineSegmentComp.llPolarSeg->CDind);
                        currentAircraftPolar->llPolarW.at(outpPolarID).wingSegments.at(segmentID).llPolarSeg->CM    =
                            PrandtlGlauert_interpol(myLL_vec.back().theLiftingLineInputPt->machNumber,    currentLiftingLineSegmentIncomp.llPolarSeg->CM,
                                                    myLL_compr->theLiftingLineInputPt->machNumber, currentLiftingLineSegmentComp.llPolarSeg->CM);
                    }
                }
            }
        }
        // Fit data to get analytical polar
        myLL_vec.back().theLLOutputPt->getAnalyticalPolar();
        // Get analytical polar for each wing separately
        size_t numbWings_tmp(liftingSurfaces.size());
        for (size_t j(0); j < numbWings_tmp; ++j) {
            myLL_vec.back().theLLOutputPt->getAnalyticalPolarLiftingSurface(myLL_vec.back().theLLOutputPt->llPolarWings.at(j),
                    myLL_vec.back().theLLOutputPt->llPolar.size());
        }
        theAnalyticalPolarData.dCMtodAoA.push_back(myLL_vec.back().theLLOutputPt->dCMtodAoA);
        theAnalyticalPolarData.dCLtodAoA.push_back(myLL_vec.back().theLLOutputPt->dCLtodAoA);
        theAnalyticalPolarData.CLatAoA0.push_back(myLL_vec.back().theLLOutputPt->CLatAoA0);
        theAnalyticalPolarData.CMatAoA0.push_back(myLL_vec.back().theLLOutputPt->CMatAoA0);
    }
    myLL_vec.back().theLLOutputPt->CDmin = 0.0;
    LILI_version = myLL->myLLsettings.LL_version;
}

double liftingLineForTAW::PrandtlGlauert_interpol(double targetMach, double y_incompr, double Ma_compr, double y_compr) {
    double PG_compr = sqrt(1. - pow(Ma_compr, 2.));
    return y_incompr + (y_compr - y_incompr) / (PG_compr - 1.) * (sqrt(1. - pow(targetMach, 2.)) - 1);
}

void liftingLineForTAW::setLiftCoefficientVector() {
    // set clean CL_min und CL_max
    double CL_min(this->getCL_startFromLiftingLineResults());
    for (size_t i(0); i< flightConditions.size(); ++i) {
        // set clean CL_max
        double CL_max = this->getCL_endFromLiftingLineResults(flightConditions.at(i).Mach);
        // loop over all CLs
        for (double j = CL_min; j <= CL_max; j += stepWidthCL) {  // loop over CLs
            liftCoefficients.at(i).push_back(Rounding(j, 4));
        }
    }
}

double liftingLineForTAW::getCL_startFromLiftingLineResults() {
    /* set CL_min and CL_max for loop of current Mach-number */
    double CL_start(0.);
    double CL_min_LL = myLiftingLine.theLLOutputPt->llPolar.front().CL;

    if (CL_min_LL <= -1. || CL_min_LL >= 1.) {
        myRuntimeInfo->err << "Smallest CL-value calculated by LIFTING-LINE (" << CL_min_LL << " is outside the valid CL-interval (-1.;1.). Abort program!" << endl;
        exit(1);
    } else {
        CL_start = ceil(CL_min_LL*10.)/10.;
    }
    return CL_start;
}

double liftingLineForTAW::getCL_endFromLiftingLineResults(const double& aMachNumber) {
    double CL_end = this->getCLmax_clean(aMachNumber);
    double CL_max_LL = myLiftingLine.theLLOutputPt->llPolar.back().CL;

    if (CL_end > CL_max_LL) {
        myRuntimeInfo->warn << setprecision(2) << fixed << "For Ma = " << aMachNumber << ": CL_max=" << CL_end << " > CL_max (LI_LI)=" << CL_max_LL << ". Using Extrapolation!"
                            << endl;
    }
    return CL_end;
}

double liftingLineForTAW::getCLmax_clean(double Mach) {
    /* Sources: [Ray18] p. 279 ff. (Original: DATCOM, Kap. 4.1.3.4); [Obe09] chapter 28 -> for Buffet onset */
    /* Assumptions and given values */
    double CLmax                = 0.0;
    double Clmax_sec            = 1.8;  // section Cl_max (general assumption) geom2::measure.phi_LE_right * PI / 180.;
    double phi_LE_rad           = geom2::measure::sweep(liftingSurfaces.front(), 0.0, 0.0) * PI / 180.;
    double Ma_dive              = std::min(initialMachCruise + 0.1, 1.);
    // ratio CLmax/Clmax according to [Ray18] Fig. 12.8
    // (small) influence of sweep -> assumption delta y ~ 2.1
    double CLmax2Clmax_phiLE0   = 0.9;  // [Ray18] Fig. 12.8
    double CLmax2Clmax_phiLE60  = 0.8;  // [Ray18] Fig. 12.8
    double CLmax2Clmax_phiLE    = CLmax2Clmax_phiLE0 + (CLmax2Clmax_phiLE60 - CLmax2Clmax_phiLE0) / (cos(60.*PI / 180.) - 1.) * (cos(phi_LE_rad) - 1.);  // [Ray18] Fig. 12.8
    double CLmax_inc            = CLmax2Clmax_phiLE * Clmax_sec;
    // Decrease due to Mach number
    double delta_CLmax_Ma06 = -0.2 * cos(phi_LE_rad);  // according to [Ray18] Fig. 12.9 (not exact, but influence neglectable --> not too big)
    double CLmax_Ma06 = CLmax_inc + delta_CLmax_Ma06;
    double delta_CLmax_Ma = delta_CLmax_Ma06 * (Mach - 0.2) / (0.6 - 0.2);
    /* Determination clean CLmax (wing) via distinction by mach-number intervals */
    if (Mach <= 0.2) {
        CLmax = CLmax_inc;
    } else if (Mach > 0.2 && Mach <= 0.6) {  // [Ray18] Figure 12.9
        CLmax = CLmax_inc + delta_CLmax_Ma;
    } else if (Mach > 0.6 && Mach <= Ma_dive) {  // Assump.: Buffet-Limit, acc.: [Obe09] Fig. 28.10-28.13
        double CL_buffet_Ma_dive = 0.2;
        // acc.: PG -> set Prandtl-Glauert factors for interpolation
        double beta_Ma06    = sqrt(1. - pow(0.6, 2.));
        double beta_Ma_dive = sqrt(1. - pow(Ma_dive, 2.));
        double beta_Ma      = sqrt(1. - pow(Mach, 2.));
        CLmax = CLmax_Ma06 + (CL_buffet_Ma_dive - CLmax_Ma06) / (beta_Ma_dive - beta_Ma06) * (beta_Ma - beta_Ma06);
    } else {
        myRuntimeInfo->err << "Ma=" << Mach << " for determination CLmax_clean higher than allowed limit Ma_lim=" << Ma_dive << endl;
        myRuntimeInfo->err << "Abort program!" << endl;
        exit(1);
    }
    return CLmax;
}

void liftingLineForTAW::setWingDesignPoints() {
    // Store Lifting Line stations and resort them from inner to outer station
    yPositionsLL = myLiftingLine_compr.theLLOutputPt->llPolarWings.at(0).yPositions;
    reverse(yPositionsLL.begin(), yPositionsLL.end());
    // Choose design lift distribution
    if (CLModeAirfoilSelection == "mode_0" || CLModeAirfoilSelection == "mode_1") {
        interpolateLiftDistribution(designCL, myLL_vec.at(cruiseMachID).theLLOutputPt->llPolarWings.at(0), cruiseMachID, 0,
                                          wingDesignPoint);
        liftDistr_design = wingDesignPoint->Cl_distr;
    } else if (CLModeAirfoilSelection == "mode_2") {
        interpolateLiftDistribution(bestCruiseCL, myLL_vec.at(cruiseMachID).theLLOutputPt->llPolarWings.at(0), cruiseMachID,
                                          0, wingPointBestCruiseCL);
        liftDistr_design = wingPointBestCruiseCL->Cl_distr;
    } else {
        myRuntimeInfo->err << "Unknown CL mode for airfoil selection mode: " << CLModeAirfoilSelection << " (allowed: 0, 1, 2). Abort program!" << endl;
        exit(1);
    }
    reverse(liftDistr_design.begin(), liftDistr_design.end());
}

/* Function creates aero sections for Q3D method */
void liftingLineForTAW::setAeroSections() {
    myRuntimeInfo->out << "Create Aero-Sections ..." << endl;
    // Create geometrical aero sections
    std::vector<aeroSection> theGeometricalAeroSections(this->setGeometricalAeroSections());
    // Create lifting line aero sections */
    std::vector<aeroSection> theLiftingLineAeroSections(this->setLiftingLineAeroSections());
    // Combine geometrical and lifting line sections
    aeroSections_all = this->combineGeometricalAndLiftingLineSections(theGeometricalAeroSections, theLiftingLineAeroSections);
    // Assignment of the corresponding area segments for later weighted summation of the Cds over wings
    this->setSectionAreas(&(aeroSections_all));
}

std::vector<liftingLineForTAW::aeroSection> liftingLineForTAW::setGeometricalAeroSections() {
    std::vector<aeroSection> geometricalAeroSections;
    enum creationMethod {MANUAL, AUTOMATED};
    uint8_t theCreationMethod(AUTOMATED);
    #ifdef DATABASE_ENABLED
        if (mySettingsPt->q3DMethod && this->theAeroDatabasePt->myAeroDatabaseSettings.selectAeroSections) theCreationMethod = MANUAL;
    #endif  // DATABASE_ENABLED
    switch (theCreationMethod) {
        case MANUAL:
            #ifdef DATABASE_ENABLED
                this->setAeroSectionsManually();
            #endif  // DATABASE_ENABLED
            break;
        default: this->setAeroSectionsAutomatically();
    }
    geometricalAeroSections = aeroSections_airfoilSel;
    return geometricalAeroSections;
}

std::vector<liftingLineForTAW::aeroSection> liftingLineForTAW::setLiftingLineAeroSections() {
    std::vector<aeroSection> theLiftingLineAeroSections;
    size_t secID_tmp(0);
    for (size_t i(0); i < yPositionsLL.size(); ++i) {
        double y_geoSec_tmp = getTrueCoordinates(liftingSurfaces.front(), secID_tmp).yCoordinate;
        if (yPositionsLL.at(i) > y_geoSec_tmp) {
            secID_tmp++;
        }
        double y_i_tmp = yPositionsLL.at(i);
        double l_i_tmp = geom2::measure::chord(liftingSurfaces.front(), y_i_tmp);
        theLiftingLineAeroSections.push_back(liftingLineForTAW::aeroSection(liftingSurfaces.front(), y_i_tmp, l_i_tmp, 0.,
                                                         "tempAirfoilNameHolder", theLiftingLineAeroSections.size() + 1, flightConditions.size()));
    }
    return theLiftingLineAeroSections;
}

std::vector<liftingLineForTAW::aeroSection> liftingLineForTAW::combineGeometricalAndLiftingLineSections(const std::vector<aeroSection>& theGeometricalAeroSections,
                                                                             const std::vector<aeroSection>& theLiftingLineAeroSections) {
    std::vector <aeroSection> theCombinedAeroSections;
    int secID_tmp(0);
    for (const aeroSection& aLiftingLineSection : theLiftingLineAeroSections) {
        double y_geoSec_tmp = theGeometricalAeroSections.at(secID_tmp).y_sec;
        if (aLiftingLineSection.y_sec > y_geoSec_tmp) {
            theCombinedAeroSections.push_back(liftingLineForTAW::aeroSection(liftingSurfaces.front(), y_geoSec_tmp, theGeometricalAeroSections.at(secID_tmp).l_sec, 0.,
                                                           theGeometricalAeroSections.at(secID_tmp).airfoilFilename, theCombinedAeroSections.size() + 1, flightConditions.size()));
            secID_tmp++;
        }
        double l_i_tmp = getLocalChordLength(liftingSurfaces.front(), aLiftingLineSection.y_sec);
        theCombinedAeroSections.push_back(liftingLineForTAW::aeroSection(liftingSurfaces.front(), aLiftingLineSection.y_sec, l_i_tmp, 0.,
                                                      theGeometricalAeroSections.at(secID_tmp).airfoilFilename, theCombinedAeroSections.size() + 1, flightConditions.size()));
    }
    // Insert last geo section
    double y_last_geoSeg = getTrueCoordinates(liftingSurfaces.front(), liftingSurfaces.front().sections.size() -1).yCoordinate;
    if (y_last_geoSeg > theLiftingLineAeroSections.back().y_sec) {
        size_t lastGeoSegID = liftingSurfaces.front().sections.size() - 1;
        theCombinedAeroSections.push_back(liftingLineForTAW::aeroSection(liftingSurfaces.front(), y_last_geoSeg,
                                            liftingSurfaces.front().sections.at(lastGeoSegID).get_chord_length(), 0.,
                                            "tmp_airfoilNameHolder", theCombinedAeroSections.size() + 1, flightConditions.size()));
    }
    return theCombinedAeroSections;
}

void liftingLineForTAW::setSectionAreas(std::vector<liftingLineForTAW::aeroSection>* theAeroSections) {
    for (size_t i(0); i < theAeroSections->size(); ++i) {
        aeroSection* currentAeroSection(&(theAeroSections->at(i)));
        if (i < theAeroSections->size() - 1) {
            const aeroSection& nextAeroSection(theAeroSections->at(i+1));
            currentAeroSection->s_sec = nextAeroSection.y_sec - currentAeroSection->y_sec;
            currentAeroSection->area = 0.5 * (nextAeroSection.l_sec + currentAeroSection->l_sec) * currentAeroSection->s_sec;
        } else {
            currentAeroSection->s_sec    = 0.;
            currentAeroSection->area     = 0.;
        }
    }
}

void liftingLineForTAW::setAeroSectionsAutomatically() {
    std::vector <int> fusSeg_IDs;
    int firstSurfaceSeg_ID(-1);
    int lastSurfaceSeg_ID(-1);
    // Loop wing sections
    for (size_t i(0); i < liftingSurfaces.front().sections.size(); ++i) {
        // Create aeroSection object
        if (i < liftingSurfaces.front().sections.size() - 1) {
            aeroSections_airfoilSel.push_back(liftingLineForTAW::aeroSection(liftingSurfaces.front(),
            getTrueCoordinates(liftingSurfaces.front(), i).yCoordinate,  // y_i
            liftingSurfaces.front().sections.at(i).get_chord_length(),  // l_i
            getTrueCoordinates(liftingSurfaces.front(), i + 1).yCoordinate - getTrueCoordinates(liftingSurfaces.front(), i).yCoordinate,  // s_i
            "tmpAirfoilNameHolder", i + 1,  // airfoil name
            flightConditions.size()));
            // create the last, virtual aerosection
        } else {
            aeroSections_airfoilSel.push_back(liftingLineForTAW::aeroSection(liftingSurfaces.front(),
            getTrueCoordinates(liftingSurfaces.front(), i).yCoordinate,  // y_i
            liftingSurfaces.front().sections.at(i).get_chord_length(),  // l_i
            ACCURACY_HIGH,  // s_i
            "tmpAirfoilNameHolder", i + 1,  // airfoil name
            flightConditions.size()));
        }
        // set flag for special case Root-Airfoil
        if (compareStrings(aeroSections_airfoilSel.back().airfoilFilename, "root", false)) {
            aeroSections_airfoilSel.back().rootAirfoil = true;
        }
        // Assignment of aerodynamic data by segment type and interpolation/direct access
        aeroSections_airfoilSel.back().theSurfaceType = 1;  // set surface  type manually for now
        switch (aeroSections_airfoilSel.back().theSurfaceType) {
            case FUSELAGE:
            {
                // ------if (myAcftPt->theFuselage.empty()) {
                if (true) {  // no Idea for now how to check for fuselage segments
                    myRuntimeInfo->err << "Error while creating the aero-sections (Sec-ID " << i + 1 << "): " <<
                                          "No fuselage segment allowed when no fuselage is available. Abort program!" << endl;
                    exit(1);
                }
            }
            case SURFACE:
            {
                myRuntimeInfo->out << "* Section " << aeroSections_airfoilSel.size() << " (Surface Segment";
                if (firstSurfaceSeg_ID < 0) {  // set ID for first Surface-Segment
                    myRuntimeInfo->out << " (1st)";
                    firstSurfaceSeg_ID = i;
                } else {  // once set, there must be no more fuselage segment
                    // --------if (myAcftPt->theWing.at(0).RightWing.at(i).segmentType == "Fuselage") {
                    if ("surface" == "Fuselage") {
                        myRuntimeInfo->out << ")" << endl;
                        myRuntimeInfo->err << "Error while creating the aero-sections (Sec-ID " << i + 1 << "): "
                                           << "No fuselage segment allowed outboard of surface-segments. Abort program!" << endl;
                        exit(1);
                    }
                }
                if (lastSurfaceSeg_ID < 0) {
                    // -------if (i == (myAcftPt->theWing.at(0).RightWing.size() - 2) || myAcftPt->theWing.at(0).RightWing.at(i + 1).segmentType == "Tip"
                    if (i == liftingSurfaces.front().sections.size() - 2) {
                        myRuntimeInfo->out << " (last)";
                        lastSurfaceSeg_ID = i;
                    }
                    // ------if (i == (myAcftPt->theWing.at(0).RightWing.size() - 1)) {
                    if (i == (liftingSurfaces.front().sections.size() - 1)) {
                        myRuntimeInfo->out << ")" << endl;
                        myRuntimeInfo->err << "Error while creating the aero-sections (Sec-ID " << i + 1 << "): "
                                           << "Last Surface Segment ID no set yet in last segment. BUG! Abort program!" << endl;
                        exit(1);
                    }
                } else {
                    if (i != (liftingSurfaces.front().sections.size() - 1)) {
                        myRuntimeInfo->out << ")" << endl;
                        myRuntimeInfo->err << "Error while creating the aero-sections (Sec-ID " << i + 1 << "): "
                                           << "Surface Segment after lastSegID has to be last segment. BUG! Abort program!" << endl;
                        exit(1);
                    } else {  // last wing segment is handled as TIP -- hflc does not work yet, block is obsolete!
                        if ("HLFC" == "hlfc") {  // ------add laminar outboard segment
                            myRuntimeInfo->out << " (last, outboard))" << endl;
                            aeroSections_airfoilSel.insert(aeroSections_airfoilSel.end()-1, liftingLineForTAW::aeroSection(liftingSurfaces.front(),
                                getTrueCoordinates(liftingSurfaces.front(), i).yCoordinate,
                                liftingSurfaces.front().sections.at(i).get_chord_length(),
                                getTrueCoordinates(liftingSurfaces.front(), i + 1).yCoordinate - getTrueCoordinates(liftingSurfaces.front(), i).yCoordinate,
                                "tmpAirfoilNamePlaceholder", i + 1, flightConditions.size()));
                            aeroSections_airfoilSel.at(i).theSurfaceType = SURFACE;
                            myRuntimeInfo->out << "* Section " << aeroSections_airfoilSel.size() << " (Surface Segment";
                        }
                        myRuntimeInfo->out << " (virtual tip)";
                        aeroSections_airfoilSel.back().theSurfaceType = TIP;
                    }
                }
                myRuntimeInfo->out << ")" << endl;
                break;
            }
            case TIP:
            {
                myRuntimeInfo->out << "case: TIP" << endl;
                /*
                if ((myAcftPt->flowType == "hlfc") && (i == static_cast<size_t>(lastSurfaceSeg_ID + 1))) {  // add laminar outboard segment
                    myRuntimeInfo->out << "* Section " << myAcftPt->aeroSections_airfoilSel.size() << " (Surface segment (last, outboard))" << endl;
                    myAcftPt->aeroSections_airfoilSel.insert(myAcftPt->aeroSections_airfoilSel.end()-1,
                                                             aeroSection(*mySettingsPt, *myAcftPt,
                        myAcftPt->theWing.at(0).RightWing.at(i).segReference[1] - myAcftPt->theWing.at(0).referencePty - ACCURACY_HIGH,  // y_i
                        myAcftPt->theWing.at(0).RightWing.at(i).l_i_segment, ACCURACY_HIGH,  // l_i, s_i
                        myAcftPt->theWing.at(0).RightWing.at(i).segInbSide.sectionProfileName, i + 1));  // airfoil name
                    myAcftPt->aeroSections_airfoilSel.at(i).theSurfaceType = SURFACE;
                }
                if (i != (myAcftPt->theWing.at(0).RightWing.size() - 1)) {
                    myRuntimeInfo->out << "* Section " << myAcftPt->aeroSections_airfoilSel.size() << " (Tip Segment)" << endl;
                } else {
                    myRuntimeInfo->out << "* Section " << myAcftPt->aeroSections_airfoilSel.size() << " (virtual tip)" << endl;
                }
                // Handle not allowed cases 
                if (i == 0) {  // tip in 1. seg. impossible
                    myRuntimeInfo->err << "Error while creating the aero-sections (Sec-ID " << myAcftPt->aeroSections_airfoilSel.size() << "): "
                                       << "Tip in first segment not possible! Abort program!" << endl;
                    exit(1);
                }
                if (lastSurfaceSeg_ID < 0) {  // last surface segID must be set already
                    myRuntimeInfo->err << "Error while creating the aero-sections (Sec-ID " << myAcftPt->aeroSections_airfoilSel.size() << "): "
                                       << "No tip-segment allowed without previous surface-segment. Abort program!" << endl;
                    exit(1);
                }
                if (myAcftPt->aeroSections_airfoilSel.back().rootAirfoil) {
                    myRuntimeInfo->err << "Invalid airfoil filename '" << myAcftPt->aeroSections_airfoilSel.back().airfoilFilename << "' : "
                                       << "No root-airfoil allowed at wing tip! Abort program!" << endl;
                    exit(1);
                }
                myAcftPt->y_wing_tip.push_back(myAcftPt->aeroSections_airfoilSel.back().y_sec);
                break;
                */
            }
            default: myRuntimeInfo->err << "setAeroSections: Unknown sectionType. Bug in program. Abort program." << endl; exit(1);
        }
    }
}

void liftingLineForTAW::getCleanPolar() {
    // Loop over all off-design Mach numbers
    for (size_t m(0); m < flightConditions.size(); ++m) {
        this->calculateCleanPolarForMach(flightConditions.at(m).Mach, m);
    }
    // If there is no LILI-error and handling is switched ON, delete backup directory
    if (myLiftingLine.myLLsettings.error_handling == 1 && myLiftingLine.LILI_bkp_dir_created && fileExists(myLiftingLine.LILI_bkp_dir)) {
        // liftingLine settings are the same for incompressible and compressible. No need to check again for myLiftingLine_compr object.
        deleteDirectory(getFullPathString(myLiftingLine.LILI_bkp_dir), false);
        deleteDirectory(getFullPathString(myLiftingLine_compr.LILI_bkp_dir), false);
    }
    // Incompressible analytical Polars
    for (size_t machID(0); machID < flightConditions.size(); machID++) {
        vector<double> Lift;
        vector<double> Drag;
        vector<double> Moment;
        bool allNAN(true);
        for (size_t i(0); i < theCleanPolars.at(machID).Points.size(); ++i) {
            if (!std::isnan(theCleanPolars.at(machID).Points.at(i).CDind)) {
                Lift.push_back(theCleanPolars.at(machID).Points.at(i).CL);
                Drag.push_back(theCleanPolars.at(machID).Points.at(i).CDind);
                Moment.push_back(theCleanPolars.at(machID).Points.at(i).CM);
                allNAN = false;
            }
        }
        if (allNAN) {
            myRuntimeInfo->err << "Could not determine incompressible polars because all values are NAN. Abort program!" << endl;
            exit(1);
        } else {
            vector<double> MomCoeff;
            vector<double> DragCoeff;
            // CM=f(CL) -> linear fit
            MomCoeff = calcRegressionCoefficientsUsingQRdecomp(Lift, Moment, 1);
            theAnalyticalPolarData.dCMtodCL_inc.push_back(MomCoeff.at(1));
            theAnalyticalPolarData.CMatCL0_inc.push_back(MomCoeff.at(0));
            MomCoeff.clear();
            // CD=f(CL) -> quadratic fit (for incompr. drag polar no wave drag, therefore CD_total (s.a.))
            DragCoeff = calcRegressionCoefficientsUsingQRdecomp(Lift, Drag, 2);
            theAnalyticalPolarData.CDmin_inc.push_back(DragCoeff.at(0) - 0.25 * pow(DragCoeff.at(1), 2) / DragCoeff.at(2));
            theAnalyticalPolarData.kFactor_inc.push_back(DragCoeff.at(2));
            theAnalyticalPolarData.CLatCDmin_inc.push_back(- 0.5 * DragCoeff.at(1) / DragCoeff.at(2));
            DragCoeff.clear();
            Lift.clear();
            Drag.clear();
            Moment.clear();
        }
    }
    myRuntimeInfo->out << "... moments and induced drag have been calculated successfully!" << endl;
}

void liftingLineForTAW::calculateCleanPolarForMach(const double& Mach, const uint16_t Mach_id) {
    myRuntimeInfo->out << setprecision(2) << fixed << "Calculate polar for Ma = " << flightConditions.at(Mach_id).Mach;
    // create "Polar Object"
    theCleanPolars.push_back(LiLiPolar());  // create "cleanPolar"
    theCleanPolars.back().Config = "Clean";
    theCleanPolars.back().extrapolationMargin = flightConditions.at(Mach_id).extrapolationMargin;
    theCleanPolars.back().allowGridChange = flightConditions.at(Mach_id).allowGridChange;
    theCleanPolars.back().MachNumber = flightConditions.at(Mach_id).Mach;
    theCleanPolars.back().altitude = flightConditions.at(Mach_id).Altitude;
    theCleanPolars.back().ReynoldsNumber = getReynoldsNumber(geom2::measure::mean_aerodynamic_chord(liftingSurfaces.front()),
                                                            theCleanPolars.back().altitude, theCleanPolars.back().MachNumber, delta_isa);
    for (size_t i(0); i < liftingSurfaces.size(); ++i) {
        theCleanPolars.back().liftingSurfacePolars.push_back(aLiftingSurfacePolar);
    }
    // Check if polar exists
    try {
        if (liftCoefficients.at(Mach_id).empty()) {
            stringstream errorMessage;
            errorMessage << "The polar for Ma = " << flightConditions.at(Mach_id).Mach << " is empty.";
            throwError(__FILE__, __func__, __LINE__, errorMessage.str());
        }
    } catch (const exception& e) {
        stringstream errorMessage;
        errorMessage << "The polar for Ma = " << flightConditions.at(Mach_id).Mach << " does not exist. (" << e.what() << ").";
        throwError(__FILE__, __func__, __LINE__, errorMessage.str());
    }
    // set clean CL_max
    theCleanPolars.back().CLmax = liftCoefficients.at(Mach_id).back();
    for (const aeroSection& anAeroSection : aeroSections_airfoilSel) {
        theCleanPolars.back().CLmax = std::min(anAeroSection.Cl_max.at(Mach_id), theCleanPolars.back().CLmax);
    }
    // set CL_min und CL_max for loop of Mach number
    double CL_start(liftCoefficients.at(Mach_id).front());
    double CL_end(liftCoefficients.at(Mach_id).back());
    myRuntimeInfo->out  << " | CL-Range [" << CL_start << ";" << CL_end << "]" << endl;
    // Loop over all CLs
    for (const double& aLiftCoefficient : liftCoefficients.at(Mach_id)) {  // Loop over CLs
        // create object for polar point
        myRuntimeInfo->debug << "  - Create a polar point ..." << endl;
        theCleanPolars.back().Points.push_back(LiLiPolarPoint());
        // Check if liftingLine object exists
        try {
            if (!myLL_vec.at(Mach_id).theLLOutputPt) {
                stringstream errorMessage;
                errorMessage << "The liftingLine output object for Ma = " << flightConditions.at(Mach_id).Mach << " does not exist.";
                throwError(__FILE__, __func__, __LINE__, errorMessage.str());
            }
        } catch (const exception& e) {
            stringstream errorMessage;
            errorMessage << "The liftingLine object for Ma = " << flightConditions.at(Mach_id).Mach << " does not exist. (" << e.what() << ").";
            throwError(__FILE__, __func__, __LINE__, errorMessage.str());
        }
        /* CLalpha-Correction by fuselage */
        myRuntimeInfo->debug << "  - Apply CLalpha correction for fuselage ..." << endl;
        double CLalpha_fus_corr(1.);
        // CLalpha Correction regarding fuselage influence according to [Tor13] , eq. (E-30) + (E-33)
        // only lift curve slope (without angular influences, K_II etc), since it is assumed that this is covered by LI_LI
        //-> check this eventually!
        if (theFuselage.sections.size() > 0 && (CM_corr_fuselage_mode == "mode_2")) {
            double S_net = geom2::measure::reference_area(liftingSurfaces.front()) - geom2::measure::chord(liftingSurfaces.front(), 0.) * geom2::measure::width_max(theFuselage);
            CLalpha_fus_corr = (1. + 2.15 * geom2::measure::width_max(theFuselage) / getHalfSpan(liftingSurfaces.front())) * 2 *
                               S_net / geom2::measure::reference_area(liftingSurfaces.front()) + PI /
                               (2. * convertUnit(RADIAN, DEGREE, myLL_vec.at(Mach_id).theLLOutputPt->liftingSurface_dCLtodAoA.at(0))) *
                               pow(geom2::measure::width_max(theFuselage), 2.) / geom2::measure::reference_area(liftingSurfaces.front());
        }
        myLL_vec.at(Mach_id).theLLOutputPt->dCLtodAoA_wf = CLalpha_fus_corr * myLL_vec.at(Mach_id).theLLOutputPt->liftingSurface_dCLtodAoA.at(0);
        // determine lift polar CL=f(AOA)
        myRuntimeInfo->debug << "  - Apply CLalpha correction for wing/htp ..." << endl;
        double dCLtodAoA_corr = myLL_vec.at(Mach_id).theLLOutputPt->dCLtodAoA_wf;
        for (size_t i(1); i < liftingSurfaces.size(); ++i) {
            dCLtodAoA_corr += myLL_vec.at(Mach_id).theLLOutputPt->liftingSurface_dCLtodAoA.at(i);
        }
        theCleanPolars.back().Points.back().AoA = (aLiftCoefficient - myLL_vec.at(Mach_id).theLLOutputPt->CLatAoA0) / dCLtodAoA_corr;
        theCleanPolars.back().Points.back().CL  = aLiftCoefficient;
        theCleanPolars.back().Points.back().iStabPolar = currentStabAngle;
        // Calculate induced drag from LiftingLine
        myRuntimeInfo->debug << "  - Calculate drag ..." << endl;
        this->calculateInducedDrag(Mach_id);
        /* Calculate moment coefficients */
        myRuntimeInfo->debug << "  - Calculate moment coefficient ..." << endl;
        this->getMomentCoefficient(Mach_id);
    }  // End CL-Loop
}

void liftingLineForTAW::calculateInducedDrag(const uint16_t& machID) {
    myRuntimeInfo->debug << "    * Calculate induced drag ..." << endl;
    /* CDmin + k*(CL - CLatCDmin)^2 leads again to the starting point CDind = a + b*CL + c*CL^2 (kFactor = c)
    which was solved with QR decomposition in liftingLine */
    theCleanPolars.back().Points.back().CDind = myLL_vec.at(machID).theLLOutputPt->CDmin + myLL_vec.at(machID).theLLOutputPt->kFactor
                                                  * pow(theCleanPolars.back().Points.back().CL - myLL_vec.at(machID).theLLOutputPt->CLatCDmin, 2.);
    // induced total drag correction
    theCleanPolars.back().Points.back().CDind = (theCleanPolars.back().Points.back().CDind + 0.0001 * indDragCtCorrForCalibration)
                                                    * factorDragCleanPolar;
    // loop over all lifting surfaces
    for (size_t surfaceID(0); surfaceID < liftingSurfaces.size(); ++surfaceID) {
        theCleanPolars.back().liftingSurfacePolars.at(surfaceID).push_back(LiftDistributionPolarPoint());
        // Read lift distributions from lifting line results and set induced drag for current liftingsurface */
        // Interpolate Cl(y) distribution for CL
        this->interpolateLiftDistribution(theCleanPolars.back().Points.back().CL, myLL_vec.at(machID).theLLOutputPt->llPolarWings.at(surfaceID), machID, surfaceID,
                                              &(theCleanPolars.back().liftingSurfacePolars.at(surfaceID).back()));
        // Set CDind for lift surface
        theCleanPolars.back().liftingSurfacePolars.at(surfaceID).back().CDind = myLL_vec.at(machID).theLLOutputPt->liftingSurface_CDmin.at(surfaceID)
                                                              + myLL_vec.at(machID).theLLOutputPt->liftingSurface_kFactor.at(surfaceID)
                                                              * pow(theCleanPolars.back().liftingSurfacePolars.at(surfaceID).back().CL
                                                              - myLL_vec.at(machID).theLLOutputPt->liftingSurface_CLatCDmin.at(surfaceID), 2.);
        // induced drag correction
        theCleanPolars.back().liftingSurfacePolars.at(surfaceID).back().CDind += indDragCtCorrForCalibration * 0.0001;
    }
    // calculate induced drag for fuselage and nacelles
    theCleanPolars.back().PointsFuselage.push_back(LiLiPolarPoint());
    theCleanPolars.back().PointsNacelles.push_back(LiLiPolarPoint());
    theCleanPolars.back().PointsFuselage.back().CDind = getInducedDragFuselage(theCleanPolars.back().Points.back().AoA, flightConditions.at(machID).Mach);
    theCleanPolars.back().PointsNacelles.back().CDind = getInducedDragNacelle(theCleanPolars.back().Points.back().AoA, flightConditions.at(machID).Mach);
    theCleanPolars.back().Points.back().CDind += theCleanPolars.back().PointsFuselage.back().CDind + theCleanPolars.back().PointsNacelles.back().CDind;
}

double liftingLineForTAW::getInducedDragFuselage(double AoA, double Mach) {
    // [Ros18]
    double indDrag(0.0);
    if (theFuselage.sections.size() > 0) {
        double l(geom2::measure::length(theFuselage));
        double baseArea(0.0);
        double planArea(0.0);
        double neta(0.0);
        double Cdc(0.0);
        baseArea = 0.25 * 3.1416 * 0.5 * (geom2::measure::width(theFuselage, 0.9 * l) + geom2::measure::height(theFuselage, 0.9 * l));
        for (int i = 0; i <= 20; i++) {
            planArea = planArea + l / 20. * geom2::measure::width(theFuselage, i * l / 20.);
        }
        neta = 0.4936 * pow((l / geom2::measure::width_max(theFuselage)), 0.1433);
        Cdc = -5.6772 * pow(Mach, 4.) + 8.6876 * pow(Mach, 3.)
              - 2.693 * pow(Mach, 2.) + 0.2186 * pow(Mach, 1.) + 1.2;
        AoA = AoA * 3.1416 / 180;
        indDrag = 2. * pow(AoA, 2.) * baseArea / geom2::measure::reference_area(liftingSurfaces.front())
                  + neta * Cdc * pow(AoA, 3.) * planArea / geom2::measure::reference_area(liftingSurfaces.front());
        /* At the moment, multi-fuselage is not possible; should be reactivated in future */
        // indDrag = myAcftPt->theFuselage.size() * indDrag;
    }
    if (isnan(indDrag) || isinf(indDrag)) {
        myRuntimeInfo->err << "Value for induced drag (fuselage) is NAN or INF: Cd_ind = " << indDrag << " at CL = " << theCleanPolars.back().Points.back().CL
                           << ". Abort program!" << endl;
        exit(1);
    }
    if (indDrag < 0.) {
        indDrag = 0.;
    }
    return indDrag;
}

double liftingLineForTAW::getInducedDragNacelle(double AoA, double Mach) {
    // [Ros18]
    double indDrag(0.0);
    if (!theNacelles.empty()) {
        double baseArea(0.0);
        double planArea(0.0);
        double neta(0.0);
        double Cdc(0.0);
        // baseArea = myAcftPt->theNacelle.at(0).R_exit; // initially is R_section defined as (innerWidth + innerHeight) / 4.
        double exit_segment_inner_width(geom2::measure::width(theNacelles.front(), theNacelles.front().sections.back().origin.z()));
        double exit_segment_inner_height(geom2::measure::height(theNacelles.front(), theNacelles.front().sections.back().origin.z()));
        baseArea = (exit_segment_inner_width + exit_segment_inner_height) / 4.;
        planArea = 0.85 * geom2::measure::length(theNacelles.front()) * geom2::measure::width_max(theNacelles.front());
        neta = 0.4936 * pow((geom2::measure::length(theNacelles.front()) / geom2::measure::width_max(theNacelles.front())), 0.1433);
        Cdc = -5.6772 * pow(Mach, 4.) + 8.6876 * pow(Mach, 3.)
              - 2.693 * pow(Mach, 2.) + 0.2186 * pow(Mach, 1.) + 1.2;
        AoA = AoA * 3.1416 / 180.;
        indDrag = 2. * pow(AoA, 2.) * baseArea / geom2::measure::reference_area(liftingSurfaces.front())
                  + neta * Cdc * pow(AoA, 3.) * planArea / geom2::measure::reference_area(liftingSurfaces.front());
        indDrag = theNacelles.size() * indDrag;
    }
    if (isnan(indDrag) || isinf(indDrag)) {
        myRuntimeInfo->err << "Value for induced drag (nacelle) is NAN or INF: Cd_ind = " << indDrag << " at CL = " << theCleanPolars.back().Points.back().CL
                           << ". Abort program!" << endl;
        exit(1);
    }
    if (indDrag < 0.) {
        indDrag = 0.;
    }
    return indDrag;
}

void liftingLineForTAW::getMomentCoefficient(const uint16_t& machID) {
    myRuntimeInfo->debug << "    * Calculate moment coefficient ..." << endl;
    // Calculate moment coefficients
    // Fuselage and nacelle correction for dCM/dCL
    double dCMdCL_corr = myLL_vec.at(machID).theLLOutputPt->liftingSurface_dCMtodCL.at(0);
    if (!theFuselage.sections.empty() && (CM_corr_fuselage_mode == "mode_2")) {
        dCMdCL_corr -= getFuselageCM_AeroCenterCorrection(machID);
    }
    if (!theNacelles.empty() && (CM_corr_nacelle_mode == "mode_2")) {
        dCMdCL_corr -= getNacelleCM_AeroCenterCorrection(machID);
    }
    // Fuselage correction for CM0
    double CMatAOA0_corr = myLL_vec.at(machID).theLLOutputPt->CMatAoA0;
    if (!theFuselage.sections.empty() && CM_corr_fuselage_mode == "mode_2") {
        CMatAOA0_corr += (getFuselageCM0(machID) + dCMdCL_corr * myLL_vec.at(machID).theLLOutputPt->liftingSurface_CLatAoA0.at(0));
    }
    CMatAOA0_corr += delta_CM0;
    // Determination of total moment coefficient according to:
    // CM  = CM0(CL=0) + dCM/dCL * CL = (CM(AOA=0)+dCM/dCL*CL0) + dCM/dCL * CLalpha * alpha
    //     = (CM(AOA=0)+dCM/dCL*CL0) + (dCM/dCL_wfn*CLalpha_wfn + dCM/dCL_H*CLalpha_H)  * alpha
    theCleanPolars.back().Points.back().CM = CMatAOA0_corr + (dCMdCL_corr * myLL_vec.at(machID).theLLOutputPt->dCLtodAoA_wf
            + myLL_vec.at(machID).theLLOutputPt->liftingSurface_dCMtodCL.at(1) * myLL_vec.at(machID).theLLOutputPt->liftingSurface_dCLtodAoA.at(1)
            + delta_dCMdCL * myLL_vec.at(machID).theLLOutputPt->dCLtodAoA) * theCleanPolars.back().Points.back().AoA;
    // myRuntimeInfo->out << theCleanPolars.back().Points.back().CM << endl;
    /*
    theCleanPolars.back().Points.back().CM = CMatAOA0_corr + (dCMdCL_corr * myLL_vec.at(machID).theLLOutputPt->dCLtodAoA_wf +
                                            delta_dCMdCL * myLL_vec.at(machID).theLLOutputPt->dCLtodAoA) * theCleanPolars.back().Points.back().AoA;
    for (size_t i(0); i < liftingSurfaces.size(); ++i) {
        theCleanPolars.back().Points.back().CM += myLL_vec.at(machID).theLLOutputPt->liftingSurface_dCMtodCL.at(i)
                                                * myLL_vec.at(machID).theLLOutputPt->liftingSurface_dCLtodAoA.at(i)
                                                * theCleanPolars.back().Points.back().AoA;
    }
    */
    // if necessary, fuselage and nacelle corrections according to Raymer
    if (!theFuselage.sections.empty() && (CM_corr_fuselage_mode == "mode_1")) {
        theCleanPolars.back().Points.back().CM += getCMalpha_correction_Raymer(true) * theCleanPolars.back().Points.back().AoA;
    }
    if (!theNacelles.empty() && (CM_corr_nacelle_mode == "mode_1")) {
        theCleanPolars.back().Points.back().CM +=  static_cast<double>(theNacelles.size()) * getCMalpha_correction_Raymer(false) *
                                                         theCleanPolars.back().Points.back().AoA;
    }
}

double liftingLineForTAW::getFuselageCM0(const uint16_t& machID) {
    // [Tor13] p. 480, equation(E-40): delta_f1 -> C_M0 fuselage contribution
    /*
    double deltaFuselage = -1.8 * (1. - 2.5 * myAcftPt->theFuselage.at(0).maxWidth / myAcftPt->theFuselage.at(0).l)
                           * (PI * myAcftPt->theFuselage.at(0).maxWidth * myAcftPt->theFuselage.at(0).maxHeight * myAcftPt->theFuselage.at(0).l)
                           / (4. * myAcftPt->Sref * myAcftPt->theWing.at(0).l_mu_right)
                           * (aLiftingLineOutput->CLatAoA0 / (aLiftingLineOutput->dCLtodAoA_wf * 180. / PI));
    */
    double deltaFuselage = -1.8 * (1. - 2.5 * geom2::measure::width_max(theFuselage) / geom2::measure::length(theFuselage))
                           * (PI * geom2::measure::width_max(theFuselage) * geom2::measure::height_max(theFuselage) * geom2::measure::length(theFuselage))
                           / (4. * geom2::measure::reference_area(liftingSurfaces.front()) * geom2::measure::mean_aerodynamic_chord(liftingSurfaces.front()))
                           * (myLL_vec.at(machID).theLLOutputPt->CLatAoA0 / (myLL_vec.at(machID).theLLOutputPt->dCLtodAoA_wf * 180. / PI));
    return deltaFuselage;
}

double liftingLineForTAW::getFuselageCM_AeroCenterCorrection(const uint16_t& machID) {
    // [Tor13] p. 480, equations: (E-37), (E-38)
    // equation (E-37): delta_f1 -> represents forward shift due to fuselage sections fwd (main contribution!) and aft of the wing
    // double l_fn = myAcftPt->theWing.at(0).referencePtx - myAcftPt->theFuselage.at(0).referencePtx;  // s. Fig. E-13
    double l_fn = liftingSurfaces.front().origin.x() - theFuselage.origin.x();  // s. Fig. E-13
    /*
    double delta_fus_1 = - 1.8 / (myLL_vec.at(machID).theLLOutputPt->dCLtodAoA_wf * 180. / PI)
                         * (myAcftPt->theFuselage.at(0).maxWidth * myAcftPt->theFuselage.at(0).maxHeight * l_fn)
                         / (myAcftPt->Sref * myAcftPt->theWing.at(0).l_mu_right);
                         */
    double delta_fus_1 = - 1.8 / (myLL_vec.at(machID).theLLOutputPt->dCLtodAoA_wf * 180. / PI)
                         * (geom2::measure::width_max(theFuselage) * geom2::measure::height_max(theFuselage) * l_fn)
                         / (geom2::measure::reference_area(liftingSurfaces.front()) * geom2::measure::mean_aerodynamic_chord(liftingSurfaces.front()));
    // [Tor13] (E-38): delta_f2 -> accounts for lift loss in the region where wing/fuselage lift carry-over is concentrated
    /** \todo [T789] (https://unicado.ilr.rwth-aachen.de/T789) Check delta_fus_2 for high positive values
     * Commented out, because high positive values are returned (ca. delta_fus_2 ~ -delta_fus_1)
     * Might be redundant, because lift influence on fuselage is already taken account of with twist reduction in Lifting line
    */ 
    double delta_fus_2 = 0.;
    // v this block was commented out before I modularized it
    //    double delta_fus_2 = 0.273/(1.+myAcftPt->theWing.at(0).TR_right)
    //                            * ( myAcftPt->theFuselage.at(0).maxWidth * (myAcftPt->Sref/myAcftPt->theWing.at(0).b) * (myAcftPt->theWing.at(0).b -
    //                                                                                                                  myAcftPt->theFuselage.at(0).maxWidth) )
    //                            / ( pow(myAcftPt->theWing.at(0).l_mu_right, 2.) * (myAcftPt->theWing.at(0).b + 2.15 * myAcftPt->theFuselage.at(0).maxWidth) )
    //                            * tan( myAcftPt->theWing.at(0).phi_25_right * PI/180.);
    return (delta_fus_1 + delta_fus_2);
}

double liftingLineForTAW::getNacelleCM_AeroCenterCorrection(const uint16_t& machID) {
    // [Tor13] p. 481, equation(E-41): delta_n -> correction of aerodynamic center location
    double delta_nac(0.0);
    double k_nac(0.0);
    if (theNacelles.size() > 0) {
        k_nac = -4.;
    } else {
        k_nac = -2.5;
    }
    // loop over all gondolas
    for (size_t i(0); i < theNacelles.size(); ++i) {
        // double l_nac = myAcftPt->theWing.at(0).l_mu_right_refPt[0] - myAcftPt->theNacelle.at(i).referencePtx;
        double l_mu_right_refPt_x = get_l_mu_ref_Point(liftingSurfaces.front()).xCoordinate;
        double l_nac = l_mu_right_refPt_x - theNacelles.at(i).origin.x();
        if (engineType == "jet" || engineType == "gtf" || engineType == "fan") {
            delta_nac += (k_nac * pow(geom2::measure::width_max(theNacelles.front()), 2.) * l_nac)
                         / (geom2::measure::reference_area(liftingSurfaces.front()) * geom2::measure::mean_aerodynamic_chord(liftingSurfaces.front())
                         * myLL_vec.at(machID).theLLOutputPt->dCLtodAoA_wf * 180. / PI);
        } else {
            myRuntimeInfo->err << "Nacelle correction not yet implemented for Props (cf. TB Eq. E-42). Abort program!" << endl;
            exit(1);
        }
    }
    return delta_nac;
}

double liftingLineForTAW::getCMalpha_correction_Raymer(bool fuselage) {
    // [Ray18] page 427, equation (16.22)
    double width(0.);
    double length(0.);
    if (fuselage) {
        width = geom2::measure::width_max(theFuselage);
        length = geom2::measure::length(theFuselage);
    } else {  // nacelle
        width = geom2::measure::width_max(theNacelles.front());
        length = geom2::measure::length(theNacelles.front());
    }
    double l_mu_right_refPt_x = get_l_mu_ref_Point(liftingSurfaces.front()).xCoordinate;
    double x_rel = l_mu_right_refPt_x / geom2::measure::length(theFuselage);
    if (x_rel < 0.1 || x_rel > 0.6) {
        myRuntimeInfo->err << "Fehler in Cmalpha-Korrektur ";
        if (fuselage)
            myRuntimeInfo->err << "Fuselage";
        else
            myRuntimeInfo->err << "Nacelle";
        myRuntimeInfo->err << " Raymer. x_rel=" << x_rel << " ausserhalb [0.1,0.6]. Abort program!" << endl;
        exit(1);
    }
    // K_f according to Fig. 16.14 (digitized and fitted with quadratic polynomial, R^2~1)
    double K_f = 0.2039656553 * pow(x_rel, 2.) - 0.0623153062 * x_rel + 0.0101385526;
    if (K_f < 0.004 || K_f > 0.055) {
        myRuntimeInfo->err << "Fehler in Cmalpha-Korrektur ";
        if (fuselage)
            myRuntimeInfo->err << "Fuselage";
        else
            myRuntimeInfo->err << "Nacelle";
        myRuntimeInfo->err << " Raymer. K_f=" << K_f << " ausserhalb [0.004,0.055]. Abort program!" << endl;
        exit(1);
    }
    return (K_f * pow(width, 2.) * length / (geom2::measure::reference_area(liftingSurfaces.front()) * geom2::measure::mean_aerodynamic_chord(liftingSurfaces.front())));
}

void liftingLineForTAW::interpolateLiftDistribution(double aLiftCoefficient, const liftingLinePolars::liftingLinePolarWings& aLLwingPolarObject, int machID, int liftSurfaceID,
                                           LiftDistributionPolarPoint* aPolarPoint) const {
    int lastPos(aLLwingPolarObject.llPolarW.size() - 1);
    aPolarPoint->yStations = aLLwingPolarObject.yPositions;
    if ((aLiftCoefficient - myLL_vec.at(machID).theLLOutputPt->llPolar.front().CL) < -ACCURACY_HIGH) {  // Check, if CL under calculated alpha-polar -> extrapolate
        double x((aLiftCoefficient - myLL_vec.at(machID).theLLOutputPt->llPolar.at(0).CL)
                 / (myLL_vec.at(machID).theLLOutputPt->llPolar.at(1).CL - myLL_vec.at(machID).theLLOutputPt->llPolar.at(0).CL));
        aPolarPoint->AoA = (1. - x) * aLLwingPolarObject.llPolarW.at(0).AoA + x * aLLwingPolarObject.llPolarW.at(1).AoA;
        aPolarPoint->CL  = (1. - x) * aLLwingPolarObject.llPolarW.at(0).CL  + x * aLLwingPolarObject.llPolarW.at(1).CL;
        aPolarPoint->CM  = (1. - x) * aLLwingPolarObject.llPolarW.at(0).CM  + x * aLLwingPolarObject.llPolarW.at(1).CM;
        for (int j(0); j < aLLwingPolarObject.numberOfSegments; j++) {
            aPolarPoint->Cl_distr.push_back((1. - x) * aLLwingPolarObject.llPolarW.at(0).wingSegments.at(j).llPolarSeg->CL
                                           + x * aLLwingPolarObject.llPolarW.at(1).wingSegments.at(j).llPolarSeg->CL);
            aPolarPoint->Lift_distr.push_back(aPolarPoint->Cl_distr.back() *
                                            getLocalChordLength(liftingSurfaces.at(liftSurfaceID),
                                            (aPolarPoint->yStations.at(j) - getTrueCoordinates(liftingSurfaces.at(liftSurfaceID), 0).yCoordinate)));
        }
    } else if ((aLiftCoefficient - myLL_vec.at(machID).theLLOutputPt->llPolar.back().CL) >= ACCURACY_HIGH) {  // Check, if CL over calculated alpha-polar -> extrapolate
        double x((aLiftCoefficient - myLL_vec.at(machID).theLLOutputPt->llPolar.at(lastPos - 1).CL)
                 / (myLL_vec.at(machID).theLLOutputPt->llPolar.at(lastPos).CL - myLL_vec.at(machID).theLLOutputPt->llPolar.at(lastPos - 1).CL));
        aPolarPoint->AoA = (1. - x) * aLLwingPolarObject.llPolarW.at(lastPos - 1).AoA + x * aLLwingPolarObject.llPolarW.at(lastPos).AoA;
        aPolarPoint->CL  = (1. - x) * aLLwingPolarObject.llPolarW.at(lastPos - 1).CL  + x * aLLwingPolarObject.llPolarW.at(lastPos).CL;
        aPolarPoint->CM  = (1. - x) * aLLwingPolarObject.llPolarW.at(lastPos - 1).CM  + x * aLLwingPolarObject.llPolarW.at(lastPos).CM;
        for (int j(0); j < aLLwingPolarObject.numberOfSegments; j++) {
            aPolarPoint->Cl_distr.push_back((1. - x) * aLLwingPolarObject.llPolarW.at(lastPos - 1).wingSegments.at(j).llPolarSeg->CL
                                           + x * aLLwingPolarObject.llPolarW.at(lastPos).wingSegments.at(j).llPolarSeg->CL);
            aPolarPoint->Lift_distr.push_back(aPolarPoint->Cl_distr.back() *
                                             getLocalChordLength(liftingSurfaces.at(liftSurfaceID),
                                            (aPolarPoint->yStations.at(j) - getTrueCoordinates(liftingSurfaces.at(liftSurfaceID), 0).yCoordinate)));
        }
    } else {  // Otherwise: check between which points of the polar the CL value is located and interpolate
        for (size_t i(0); i < myLL_vec.at(machID).theLLOutputPt->llPolar.size() - 1; ++i) {
            if ((aLiftCoefficient - myLL_vec.at(machID).theLLOutputPt->llPolar.at(i).CL) >= -ACCURACY_HIGH &&
                (aLiftCoefficient - myLL_vec.at(machID).theLLOutputPt->llPolar.at(i + 1).CL) < ACCURACY_HIGH) {
                double x((aLiftCoefficient - myLL_vec.at(machID).theLLOutputPt->llPolar.at(i).CL)
                         / (myLL_vec.at(machID).theLLOutputPt->llPolar.at(i + 1).CL - myLL_vec.at(machID).theLLOutputPt->llPolar.at(i).CL));
                aPolarPoint->AoA   = (1. - x) * aLLwingPolarObject.llPolarW.at(i).AoA   + x * aLLwingPolarObject.llPolarW.at(i + 1).AoA;
                aPolarPoint->CL    = (1. - x) * aLLwingPolarObject.llPolarW.at(i).CL    + x * aLLwingPolarObject.llPolarW.at(i + 1).CL;
                aPolarPoint->CM    = (1. - x) * aLLwingPolarObject.llPolarW.at(i).CM    + x * aLLwingPolarObject.llPolarW.at(i + 1).CM;
                for (int j(0); j < aLLwingPolarObject.numberOfSegments; j++) {
                    aPolarPoint->Cl_distr.push_back((1. - x) * aLLwingPolarObject.llPolarW.at(i).wingSegments.at(j).llPolarSeg->CL
                                                   + x * aLLwingPolarObject.llPolarW.at(i + 1).wingSegments.at(j).llPolarSeg->CL);
                    try {
                        aPolarPoint->Lift_distr.push_back(aPolarPoint->Cl_distr.back() *
                                                         getLocalChordLength(liftingSurfaces.at(liftSurfaceID),
                                            (aPolarPoint->yStations.at(j) - getTrueCoordinates(liftingSurfaces.at(liftSurfaceID), 0).yCoordinate)));
                    } catch (const string& text) {
                        myRuntimeInfo->err << std::setprecision(10) << "y-positions [j = " << j + 1 << "] of LiftingLine [y_LL: "
                                           << aPolarPoint->yStations.at(j) - getTrueCoordinates(liftingSurfaces.at(liftSurfaceID), 0).yCoordinate << "] out of wing span [s: "
                                           << getHalfSpan(liftingSurfaces.at(liftSurfaceID)) << ". Redistribute panels. Abort program." << std::setprecision(4) << endl;
                        exit(1);
                    }
                }
                break;
            }
        }
    }
}

liftingLineForTAW::aeroSection::aeroSection(const geom2::MultisectionSurface<geom2::AirfoilSection> liftingSurface, const double& y_sec_i, const double& l_sec_i,
                                            const double& s_sec_i, const std::string& filename_i, const uint16_t& anID, const size_t numberOfFlightConditions)
    :
    airfoilFilename(filename_i.substr(filename_i.rfind('\\') + 1)),
    secID(anID),
    writePlotData(false),
    rootAirfoil(false),
    Cl_max(numberOfFlightConditions, NAN),
    // geometry
    y_sec(y_sec_i),
    l_sec(l_sec_i),
    s_sec(s_sec_i),
    area(0.) {
    // ctor
    if (y_sec < 0 || y_sec > getHalfSpan(liftingSurface)) {
        myRuntimeInfo->err << "Default y-position for aero section (constructor) outside of the wing geometry. BUG! Abort program!" << endl;
        exit(1);
    }
}

// Destructor Aero-Section
liftingLineForTAW::aeroSection::~aeroSection() {
    // dtor
}

liftingLineForTAW::LiLiPolar::LiLiPolar() {
    // ctor
}

liftingLineForTAW::LiLiPolarPoint::LiLiPolarPoint()
    :
    AoA(0.),
    CL(0.),
    CDind(0.),
    CM(0.),
    LoverD(0.),
    iStabPolar(0.) {
}

liftingLineForTAW::LiftDistributionPolarPoint::LiftDistributionPolarPoint() {
    // ctor
}
