/*  Copyright (C) 2010-2024 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
 * Raymer, D. P. - Aircraft Design: A Conceptual Approach, American Institute of Aeronautics and Astronautics, 2018, ISBN: 978-1-62410-574-6
 * p. 280 ff. (pdf p. 294 ff.) 
 */

#include "viscDragRaymer.h"

#include <aixml/node.h>
#include <aircraftGeometry2/fuselage.h>
#include <aircraftGeometry2/airfoil_surface.h>
#include <aircraftGeometry2/processing/measure.h>
#include <moduleBasics/module.h>
#include <moduleBasics/runtimeIO.h>
#include <runtimeInfo/runtimeInfo.h>
#include <moduleBasics/strategySelector.h>

#include <filesystem>
#include <vector>
#include <memory>
#include <algorithm>
#include <exception>
#include <iomanip>
#include <string>
#include <utility>  // make_pair

#include "../polar.h"
#include "generalMethods/auxFunctions.h"

using std::endl;
using std::exception;
using std::fixed;
using std::ofstream;
using std::setprecision;
using std::string;
using std::stringstream;
using std::isinf;
using std::isnan;
using std::vector;
/* TODO:    -find way to substract fuselage from wetted surface wing
            -find way to calculate fuselage upsweep angle
            -check if wing drag is for entire wing or just right side
*/
viscDragRaymerInitialization::viscDragRaymerInitialization()
    :
    calculateWing(false),
    calculateFuselage(false),
    calculateVTP(false),
    calculateHTP(false),
    calculateNacelles(false),
    calculateMisc(false),
    manualTransitionLocation(false),
    transitionLocationWing(NAN),
    transitionLocationHTP(NAN),
    transitionLocationVTP(NAN),
    doCalibration(false),
    CDvisc_CDsum(NAN),
    CDvisc_CLfact(NAN),
    CDvisc_CLexp(NAN),
    CDvisc_CDsum_lowMa(NAN),
    CDvisc_CLfact_lowMa(NAN),
    CDvisc_CLexp_lowMa(NAN),
    jetEngine(false),
    the_delta_isa(0.) {
}

viscDragRaymer::viscDragRaymer(const std::shared_ptr<RuntimeIO>& rtIO, const viscDragRaymerInitialization& viscDragInit)
    :
    // the following variables need to be initialized externally
    calculateWing(viscDragInit.calculateWing),
    calculateFuselage(viscDragInit.calculateFuselage),
    calculateVTP(viscDragInit.calculateVTP),
    calculateHTP(viscDragInit.calculateHTP),
    calculateNacelles(viscDragInit.calculateNacelles),
    calculateMisc(viscDragInit.calculateMisc),
    manualTransitionLocation(viscDragInit.manualTransitionLocation),
    transitionLocationWing(viscDragInit.transitionLocationWing),
    transitionLocationHTP(viscDragInit.transitionLocationHTP),
    transitionLocationVTP(viscDragInit.transitionLocationVTP),
    doCalibration(viscDragInit.doCalibration),
    CDvisc_CDsum(viscDragInit.CDvisc_CDsum),
    CDvisc_CLfact(viscDragInit.CDvisc_CLfact),
    CDvisc_CLexp(viscDragInit.CDvisc_CLexp),
    CDvisc_CDsum_lowMa(viscDragInit.CDvisc_CDsum_lowMa),
    CDvisc_CLfact_lowMa(viscDragInit.CDvisc_CLfact_lowMa),
    CDvisc_CLexp_lowMa(viscDragInit.CDvisc_CLexp_lowMa),
    doViscDragCtCorrForCalibration(viscDragInit.doViscDragCtCorrForCalibration),
    viscDragCtCorrForCalibrationWing(viscDragInit.viscDragCtCorrForCalibrationWing),
    viscDragCtCorrForCalibrationFuselage(viscDragInit.viscDragCtCorrForCalibrationFuselage),
    viscDragCtCorrForCalibrationHTP(viscDragInit.viscDragCtCorrForCalibrationHTP),
    viscDragCtCorrForCalibrationVTP(viscDragInit.viscDragCtCorrForCalibrationVTP),
    viscDragCtCorrForCalibrationNacelles(viscDragInit.viscDragCtCorrForCalibrationNacelles),
    viscDragCtCorrForCalibrationMisc(viscDragInit.viscDragCtCorrForCalibrationMisc),
    jetEngine(viscDragInit.jetEngine),
    delta_isa(viscDragInit.the_delta_isa),
    Sref_Wing(NAN),
    Swet_Wing(NAN),
    MAC_Wing(NAN),
    avgTtoC_Wing(NAN),
    mXtoC_Wing(NAN),
    eta_ref_Wing(NAN),
    phiAtmXtoC_Wing(NAN),
    QC_Wing(NAN),
    Swet_HTP(NAN),
    MAC_HTP(NAN),
    avgTtoC_HTP(NAN),
    mXtoC_HTP(NAN),
    eta_ref_HTP(NAN),
    phiAtmXtoC_HTP(NAN),
    QC_HTP(NAN),
    Swet_VTP(NAN),
    MAC_VTP(NAN),
    avgTtoC_VTP(NAN),
    mXtoC_VTP(NAN),
    eta_ref_VTP(NAN),
    phiAtmXtoC_VTP(NAN),
    QC_VTP(NAN),
    Swet_Fuselage(NAN),
    l_Fuselage(NAN),
    maxWidthFuselage(NAN),
    maxHeightFuselage(NAN),
    QC_Fuselage(NAN),
    fuselageUpsweepAngle(NAN),
    Swet_Nacelles(NAN),
    l_Nacelles(NAN),
    maxWidthNacelles(NAN),
    maxHeightNacelles(NAN),
    QC_Nacelles(NAN),
    factorProtuberances(NAN) {
    if (calculateWing) {
        theAircraftComponents.myWing = viscDragInit.theWing;
    }
    if (calculateFuselage) {
        theAircraftComponents.myFuselage = viscDragInit.theFuselage;
    }
    if (calculateNacelles) {
        theAircraftComponents.myNacelles = viscDragInit.theNacelles;
    }
    if (calculateHTP) {
        theAircraftComponents.myHTP = viscDragInit.theHTP;
    }
    if (calculateVTP) {
        theAircraftComponents.myVTP = viscDragInit.theVTP;
    }
    if (!manualTransitionLocation) {
        transitionLocationWing = 0.0;
        transitionLocationHTP = 0.0;
        transitionLocationVTP = 0.0;
    }
}

void viscDragRaymer::preprocessGeometry() {
    myRuntimeInfo->debug << "viscDragRaymer: preprocess the aircraft geometry" << endl;
    if (calculateWing) {
        Sref_Wing = geom2::measure::reference_area(theAircraftComponents.myWing);
        Swet_Wing = geom2::measure::area(theAircraftComponents.myWing);
        MAC_Wing = geom2::measure::mean_aerodynamic_chord(theAircraftComponents.myWing);
        avgTtoC_Wing = getAverageTtoC(theAircraftComponents.myWing);
        mXtoC_Wing = 0.5;  // Assumption by Raymer for chordwise location of airfoil maximum thickness point for high-speed airfoils
        eta_ref_Wing = 0.65;  // Reference position used for local sweep <- cant  find this value in the source
        phiAtmXtoC_Wing = getLocalSweep(theAircraftComponents.myWing, eta_ref_Wing * getHalfSpan(theAircraftComponents.myWing), mXtoC_Wing, "RADIAN");
        QC_Wing = 1.;  // Assumtion for high, mid, and well fitted low wing
        // Correct wetted surface if wing is within fuselage
        double wFuselage(0.), zFuselageTop(0.), zFuselageBottom(0.);
        if (!theAircraftComponents.myFuselage.sections.empty()) {
            wFuselage = geom2::measure::width(theAircraftComponents.myFuselage,
                                              theAircraftComponents.myWing.origin.x() + theAircraftComponents.myWing.sections.front().origin.x());
            zFuselageTop = theAircraftComponents.myFuselage.origin.z() +
                           geom2::measure::top(theAircraftComponents.myFuselage,
                                               theAircraftComponents.myWing.origin.x() + theAircraftComponents.myWing.sections.front().origin.x()).y();
            zFuselageBottom = theAircraftComponents.myFuselage.origin.z() +
                              geom2::measure::bottom(theAircraftComponents.myFuselage,
                                                     theAircraftComponents.myWing.origin.x() + theAircraftComponents.myWing.sections.front().origin.x()).y();
        }
        // Check if wing object is located in fuselage
        double origin_z_wing(theAircraftComponents.myWing.origin.z() + theAircraftComponents.myWing.sections.front().origin.y());
        if (origin_z_wing <= zFuselageTop && origin_z_wing >= zFuselageBottom) {  // If tailplane Z-coordinate inside fuselage
            Swet_Wing -= 2.* (wFuselage *  theAircraftComponents.myWing.sections.front().get_chord_length());
        }
    }
    if (calculateHTP) {
        Swet_HTP = geom2::measure::area(theAircraftComponents.myHTP);
        MAC_HTP = geom2::measure::mean_aerodynamic_chord(theAircraftComponents.myHTP);
        avgTtoC_HTP = getAverageTtoC(theAircraftComponents.myHTP);
        mXtoC_HTP = 0.4;  // Assumption by Raymer for chordwise location of airfoil maximum thickness point for HTP
        eta_ref_HTP = 0.65;  // Reference position used for local sweep
        phiAtmXtoC_HTP = getLocalSweep(theAircraftComponents.myHTP, eta_ref_HTP * getHalfSpan(theAircraftComponents.myHTP), mXtoC_HTP, "RADIAN");
        QC_HTP = 1.04;  // Form factor HTP
        // Correct wetted surface if wing is within fuselage
        double wFuselage(0.), zFuselageTop(0.), zFuselageBottom(0.);
        if (!theAircraftComponents.myFuselage.sections.empty()) {
            wFuselage = geom2::measure::width(theAircraftComponents.myFuselage,
                                              theAircraftComponents.myHTP.origin.x() + theAircraftComponents.myHTP.sections.front().origin.x());
            zFuselageTop = theAircraftComponents.myFuselage.origin.z() +
                           geom2::measure::top(theAircraftComponents.myFuselage,
                                               theAircraftComponents.myHTP.origin.x() + theAircraftComponents.myHTP.sections.front().origin.x()).y();
            zFuselageBottom = theAircraftComponents.myFuselage.origin.z() +
                              geom2::measure::bottom(theAircraftComponents.myFuselage,
                                                     theAircraftComponents.myHTP.origin.x() + theAircraftComponents.myHTP.sections.front().origin.x()).y();
        }
        // Check if wing object is located in fuselage
        double origin_z_htp(theAircraftComponents.myHTP.origin.z() + theAircraftComponents.myHTP.sections.front().origin.y());
        if (origin_z_htp <= zFuselageTop && origin_z_htp >= zFuselageBottom) {  // If tailplane Z-coordinate inside fuselage
            Swet_HTP -= 2.* (wFuselage *  theAircraftComponents.myHTP.sections.front().get_chord_length());
        }
    }
    if (calculateVTP) {
        Swet_VTP = geom2::measure::area(theAircraftComponents.myVTP);
        MAC_VTP = geom2::measure::mean_aerodynamic_chord(theAircraftComponents.myVTP);
        avgTtoC_VTP = getAverageTtoC(theAircraftComponents.myVTP);
        mXtoC_VTP = 0.4;  // Assumption by Raymer for chordwise location of airfoil maximum thickness point for high-speed airfoils
        eta_ref_VTP = 0.65;
        phiAtmXtoC_VTP = getLocalSweep(theAircraftComponents.myVTP, eta_ref_VTP * getHalfSpan(theAircraftComponents.myVTP), mXtoC_VTP, "RADIAN");
        QC_VTP = 1.04;  // Form factor VTP
    }
    if (calculateFuselage) {
        Swet_Fuselage = geom2::measure::area(theAircraftComponents.myFuselage);
        l_Fuselage = geom2::measure::length(theAircraftComponents.myFuselage);
        maxWidthFuselage = geom2::measure::width_max(theAircraftComponents.myFuselage);
        maxHeightFuselage = geom2::measure::height_max(theAircraftComponents.myFuselage);
        QC_Fuselage = 1.0;
        fuselageUpsweepAngle = getFuselageUpsweepAngle(theAircraftComponents.myFuselage);
    }
    if (calculateNacelles) {
        Swet_Nacelles = geom2::measure::area(theAircraftComponents.myNacelles.front());
        l_Nacelles = geom2::measure::length(theAircraftComponents.myNacelles.front());
        maxWidthNacelles = geom2::measure::width_max(theAircraftComponents.myNacelles.front());
        maxHeightNacelles = geom2::measure::height_max(theAircraftComponents.myNacelles.front());
        QC_Nacelles = 1.4;
    }
    // according to [Ray18] (chap. Aerodynamics - Leakage and Protuberances): 2-5% for jet, 5-10% for prop
    if (jetEngine) {
        factorProtuberances = 0.03;
    } else {
        factorProtuberances = 0.07;
    }
    myRuntimeInfo->out << "Viscous drag calculation by Raymer successfully initialized for:" << endl;
    if (calculateWing)
        myRuntimeInfo->out << " -wing" << endl;
    if (calculateHTP)
        myRuntimeInfo->out << " -HTP" << endl;
    if (calculateVTP)
        myRuntimeInfo->out << " -VTP" << endl;
    if (calculateFuselage)
        myRuntimeInfo->out << " -fuselage" << endl;
    if (calculateNacelles)
        myRuntimeInfo->out << " -nacelles" << endl;
    if (calculateMisc)
        myRuntimeInfo->out << " -miscellaneous" << endl;
}

std::vector<double> viscDragRaymer::calculateViscDrag(const Polar& aPolar, const double transitionLocationWing,
                                                    const double transitionLocationHTP, const double transitionLocationVTP) {
    // [Ray18] p. 280 ff. (pdf p. 294 ff.)
    if (calculateWing == false) {
        std::stringstream errorMsg;
        errorMsg << "No wing in viscous drag raymer. Unable to calculate drag";
        throwError(__FILE__, __func__, __LINE__, errorMsg.str());
    }
    // vector with component order: 0 = Wing, 1 = HTP, 2 = VTP, 3 = Fuselage, 4 = Nacelles, 5 = Miscellaneous
    vector <double> theComponentDrags;
    theComponentDrags.push_back(calculateViscDragWing(aPolar, transitionLocationWing));
    // calculate HTP drag if switch is on
    if (calculateHTP) {
        theComponentDrags.push_back(calculateViscDragHTP(aPolar, transitionLocationHTP));
    } else {
        theComponentDrags.push_back(0.);
    }
    // calculate VTP drag if switch is on
    if (calculateVTP) {
        theComponentDrags.push_back(calculateViscDragVTP(aPolar, transitionLocationVTP));
    } else {
        theComponentDrags.push_back(0.);
    }
    // calculate fuselage drag if switch is on
    if (calculateFuselage) {
        theComponentDrags.push_back(calculateViscDragFuselage(aPolar));
    } else {
        theComponentDrags.push_back(0.);
    }
    // calculate nacelle drag if switch is on
    if (calculateNacelles) {
        theComponentDrags.push_back(calculateViscDragNacelles(aPolar));
    } else {
        theComponentDrags.push_back(0.);
    }
    // calculate miscellaneous and protuberance drag if switch is on
    if (calculateMisc) {
        theComponentDrags.push_back(calculateViscDragMisc(aPolar));
        // according to [Ray18] (chap. Aerodynamics - Leakage and Protuberances): 2-5% for jet, 5-10% for prop (Page 289)
        double CD_protub = std::accumulate(theComponentDrags.begin(), theComponentDrags.end() - 1, 0.0) * factorProtuberances;
        theComponentDrags.back() += CD_protub;
    } else {
        theComponentDrags.push_back(0.);
    }
    // maybe do a plausibility check here..
    return theComponentDrags;
}

std::vector<double> viscDragRaymer::calculateCorrectedViscDrag(const std::vector<double>& componentDrags, const double aMachNumber, const double CL) {
    /* Drag calibration. First introduced in commit [4eab519c2] based on experimental data in [Lam2010], 23.06.2010. After several changes, 
       the current definition and values in config are from Risse in commit [2ec91580] (02.08.2015) */
    std::vector <double> correctedComponentDrags = componentDrags;
    if (doCalibration) {
        // loop over all points to calculate CL dependent visc drag correction for wing (only wing drag will be corrected to get the correct behavor of form drag due to separation)
        if (aMachNumber > 0.5) {
            correctedComponentDrags.front() += CDvisc_CDsum + CDvisc_CLfact * pow(CL, CDvisc_CLexp);
        } else if (aMachNumber <= 0.5) {
            correctedComponentDrags.front() += CDvisc_CDsum_lowMa + CDvisc_CLfact_lowMa * pow(CL, CDvisc_CLexp_lowMa);
        }
    }
    return correctedComponentDrags;
}

double viscDragRaymer::calculateViscDragWing(const Polar& aPolar, const double transitionLocationWing) {
    // [Ray18] p. 280 ff. (pdf p. 294 ff.)
    myRuntimeInfo->debug << "      - Get viscous drag wing ..." << endl;
    double CDvisc_Wing(0.);
    double reynoldNumberWing = getReynoldsNumber(MAC_Wing, aPolar.altitude, aPolar.MachNumber, delta_isa);
    double Cf;
    double Cflaminar;
    double FF;
    // Friction Coefficient (turbulent) for the wing
    Cf = (getFrictionCoefficient(aPolar.MachNumber, reynoldNumberWing));
    // Friction Coefficient (laminar) for the wing
    Cflaminar = (1.328 / sqrt(reynoldNumberWing));
    // Form factor for the components (eq. 12.30)
    FF = (1. + 0.6 / mXtoC_Wing * avgTtoC_Wing + 100.*pow(avgTtoC_Wing, 4.))
                 * (1.34 * pow(aPolar.MachNumber, 0.18) * pow(cos(phiAtmXtoC_Wing), 0.28));
    // fully turbulent for transitionLocation = 0.0
    CDvisc_Wing = (Cflaminar * transitionLocationWing + Cf * (1. - transitionLocationWing))
                                                            * FF * QC_Wing * Swet_Wing / Sref_Wing;
    return CDvisc_Wing;
}

double viscDragRaymer::calculateViscDragHTP(const Polar& aPolar, const double transitionLocationHTP) {
    // [Ray18] p. 280 ff. (pdf p. 294 ff.)
    myRuntimeInfo->debug << "      - Get viscous drag HTP ..." << endl;
    double CDvisc_HTP(0.0);
    double reynoldNumberHTP = getReynoldsNumber(MAC_HTP, aPolar.altitude, aPolar.MachNumber, delta_isa);
    double Cf;
    double Cflaminar;
    double FF;
    // Friction Coefficient (turbulent) for the HTP
    Cf = getFrictionCoefficient(aPolar.MachNumber, reynoldNumberHTP);
    // Friction Coefficient (laminar) for the HTP
    Cflaminar = 1.328 / sqrt(reynoldNumberHTP);
    // Form factor for the HTP (eq. 12.30)
    FF = ((1. + 0.6 / mXtoC_HTP * avgTtoC_HTP + 100.*pow(avgTtoC_HTP, 4.))
                 * (1.34 * pow(aPolar.MachNumber, 0.18) * pow(cos(phiAtmXtoC_HTP), 0.28)));
    CDvisc_HTP = (Cflaminar * transitionLocationHTP + Cf * (1. - transitionLocationHTP))
                                                                * FF * QC_HTP * Swet_HTP / Sref_Wing;
    return CDvisc_HTP;
}

double viscDragRaymer::calculateViscDragVTP(const Polar& aPolar, const double transitionLocationVTP) {
    // [Ray18] p. 280 ff. (pdf p. 294 ff.)
    myRuntimeInfo->debug << "      - Get viscous drag VTP ..." << endl;
    double CDvisc_VTP(0.0);
    double reynoldNumberVTP = getReynoldsNumber(MAC_VTP, aPolar.altitude, aPolar.MachNumber, delta_isa);
    double Cf;
    double Cflaminar;
    double FF;
    // Friction Coefficient (turbulent) for component
    Cf = getFrictionCoefficient(aPolar.MachNumber, reynoldNumberVTP);
    Cflaminar = 1.328 / sqrt(reynoldNumberVTP);
    // Form factor for components
    FF = (1. + 0.6 / mXtoC_VTP * avgTtoC_VTP + 100.*pow(avgTtoC_VTP, 4.))
                 * (1.34 * pow(aPolar.MachNumber, 0.18) * pow(cos(phiAtmXtoC_VTP), 0.28));
    // Store components
    CDvisc_VTP = (Cflaminar * transitionLocationVTP + Cf * (1. - transitionLocationVTP))
                                                            * FF * QC_VTP * Swet_VTP / Sref_Wing;
    return CDvisc_VTP;
}

double viscDragRaymer::calculateViscDragFuselage(const Polar& aPolar) {
    // [Ray18] p. 280 ff. (pdf p. 294 ff.)
    myRuntimeInfo->debug << "      - Get viscous drag Fuselage ..." << endl;
    double CDvisc_Fuselage(0.0);
    double reynoldsNumberFuselage = getReynoldsNumber(l_Fuselage, aPolar.altitude, aPolar.MachNumber, delta_isa);
    double Cf;
    double FF;
    // Friction Coefficient (turbulent) for component
    Cf = getFrictionCoefficient(aPolar.MachNumber, reynoldsNumberFuselage);
    // Form factor for fuselage
    double f_Fuselage = l_Fuselage / sqrt(maxHeightFuselage * maxWidthFuselage);
    FF = 1. + (60. / pow(f_Fuselage, 3.)) + (f_Fuselage / 400.);
    CDvisc_Fuselage = Cf * FF * QC_Fuselage * Swet_Fuselage / Sref_Wing;
    return CDvisc_Fuselage;
}

double viscDragRaymer::calculateViscDragNacelles(const Polar& aPolar) {
    // [Ray18] p. 280 ff. (pdf p. 294 ff.)
    myRuntimeInfo->debug << "      - Get viscous drag Nacelles ..." << endl;
    double CDvisc_Nacelles(0.0);
    double reynoldsNumberNacelles = getReynoldsNumber(l_Nacelles, aPolar.altitude, aPolar.MachNumber, delta_isa);
    double Cf;
    double FF;
    // Reynolds number of component
    reynoldsNumberNacelles = getReynoldsNumber(l_Nacelles, aPolar.altitude, aPolar.MachNumber, delta_isa);
    // Friction Coefficient (turbulent) for nacelles
    Cf = getFrictionCoefficient(aPolar.MachNumber, reynoldsNumberNacelles);
    // Form factor for components
    double f_Nacelles = l_Nacelles / sqrt(maxHeightNacelles * maxWidthNacelles);
    FF = 1 + (0.35 / f_Nacelles);
    // Store components
    CDvisc_Nacelles = theAircraftComponents.myNacelles.size() * Cf * FF * QC_Nacelles * Swet_Nacelles / Sref_Wing;
    return CDvisc_Nacelles;
}

double viscDragRaymer::calculateViscDragMisc(const Polar& aPolar) {
    // [Ray18] p. 280 ff. (pdf p. 294 ff.)
    myRuntimeInfo->debug << "      - Get viscous drag Miscellaneous ..." << endl;
    double CDvisc_misc = 0.0;
    if (calculateFuselage) {
        CDvisc_misc = 3.83 * pow(fuselageUpsweepAngle, 2.5) * (PI * 0.5 * maxWidthFuselage * 0.5 * maxHeightFuselage) / Sref_Wing;
    }
    return CDvisc_misc;
}

double viscDragRaymer::getFrictionCoefficient(double flightMach, double ReynoldsNumber) {
    // [Ray18] p. 280 ff. (pdf p. 294 ff.)
    // Friction Coefficient (turbulent) for dcomponent
    return 0.455 / (pow(log10(ReynoldsNumber), 2.58) * pow(1 + 0.144 * pow(flightMach, 2.), 0.65));
}
