/*  Copyright (C) 2010-2024 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/

/** References
 *  [Gur2010] Gur, O. - Full-Configuration Drag Estimation, DOI: 10.2514/1.47557
 *  [Lam2010] Anton, E.; Lammering, T 2010 - A comparative analysis of operations towards fuel efficiency in civil aviation -
 *  I am not shure if this is the right source, but ist the only paper I can find (LN 2024)
 *  [Mas2011] Mason, W. H. - Lesson 26: Transonic Airfoils, 2011, [Lecture] Virginia Polytechnic Institute and State University,
 *                              url: https://www.dept.aoe.vt.edu/~mason/Mason_f/TransonicAirfoils.pdf [Accessed: 12/12/2021]
 *  [Ris2016] Risse, K. 2016 - Preliminary Overall Aircraft Design with Hybrid Laminar Flow Control, ISBN 978-3-8440-4950-3
 */

#include "waveDragMason.h"
#include "generalMethods/auxFunctions.h"
#include "../polar.h"

#include <runtimeInfo/runtimeInfo.h>
#include <moduleBasics/runtimeIO.h>
#include <unitConversion/unitConversion.h>
#include <aircraftGeometry2/processing/measure.h>

#include <numeric>
#include <algorithm>
#include <vector>
#include <exception>
#include <iomanip>
#include <memory>
#include <string>

using std::endl;
using std::exception;
using std::fixed;
using std::ofstream;
using std::setprecision;
using std::string;
using std::stringstream;
using std::vector;

/*  TODO - find a consisten way to handle calibration and drag count correction 
         - approximate lift distribution with a spline
         */

waveDragInitialization::waveDragInitialization()
    :
    factorKa(NAN),
    maxAllowedSweep(NAN),
    numberOfSegments(NAN),
    initialMachCruise(NAN),
    doWaveDragMasonCalibration(false),
    CLFact(NAN),
    CLExp(NAN),
    doWaveDragMasonDragCountCorrection(false),
    waveDragCtCorrForCalibrationWing(NAN),
    waveDragCtCorrForCalibrationStabilizer(NAN) {
}

waveDragMason::waveDragMason(const std::shared_ptr<RuntimeIO>& rtIO, const waveDragInitialization& waveDragInit)
    :
    factorKa(waveDragInit.factorKa),
    maxAllowedSweep(waveDragInit.maxAllowedSweep),
    numberOfSegments(waveDragInit.numberOfSegments),
    initialMachCruise(waveDragInit.initialMachCruise),
    useCostomReferenceAngleForSweep(waveDragInit.useCostomReferenceAngleForSweep),
    customSweepPosition(waveDragInit.customSweepPosition),
    doWaveDragMasonCalibration(waveDragInit.doWaveDragMasonCalibration),
    CLFact(waveDragInit.CLFact),
    CLExp(waveDragInit.CLExp),
    doWaveDragMasonDragCountCorrection(waveDragInit.doWaveDragMasonDragCountCorrection),
    waveDragCtCorrForCalibrationWing(waveDragInit.waveDragCtCorrForCalibrationWing),
    waveDragCtCorrForCalibrationStabilizer(waveDragInit.waveDragCtCorrForCalibrationStabilizer),
    liftingSurfaces(waveDragInit.liftingSurfaces),
    theFuselage(waveDragInit.theFuselage),
    tooLargeSweep(false) {
    surfaceSectionThicknesses.resize(liftingSurfaces.size());
    for (size_t surfaceID(0); surfaceID < liftingSurfaces.size(); ++surfaceID) {
        for (size_t sectionID(0); sectionID < liftingSurfaces.at(surfaceID).sections.size(); ++sectionID) {
            spanStationWing aSpanStationWing;
            aSpanStationWing.span = getTrueCoordinates(liftingSurfaces.at(surfaceID), sectionID).yCoordinate;
            aSpanStationWing.thicknessRatio = getThicknessToChordRatioMax(liftingSurfaces.at(surfaceID).sections.at(sectionID));
            surfaceSectionThicknesses.at(surfaceID).push_back(aSpanStationWing);
        }
    }
}

double waveDragMason::calculateWaveDrag(const PolarPoint& aPolarPoint, const double aMachNumber, size_t liftingSurfaceIndex, const double chordWiseSweepPosition,
                                        const std::vector<double>& spanStations, const std::vector<double>& Cl_distribution, const bool doCorrection) {
    // consistency check
    if (spanStations.size() != Cl_distribution.size()) {
            stringstream errorMessage;
            errorMessage << "calculateWaveDrag: Input vector sizes dont match! SpanStations size: " << spanStations.size() << ", CL_distribution size: " << Cl_distribution.size();
            throwError(__FILE__, __func__, __LINE__, errorMessage.str());
    }
    double C_DWaveTotal(0.0);
    vector<double> C_DWaveSegment;
    double liftingSurfaceRefArea = geom2::measure::reference_area(liftingSurfaces.at(liftingSurfaceIndex));
    double segmentSpan = getHalfSpan(liftingSurfaces.at(liftingSurfaceIndex)) / numberOfSegments;
    for (size_t segmentID(0); segmentID < numberOfSegments; ++segmentID) {
        bool tooLargeSweepInSegment(false);
        double ySegmentCenter = segmentSpan * (static_cast<double>(segmentID) + 0.5);
        double ySegmentInboard = segmentSpan * static_cast<double>(segmentID);
        double ySegmentOutboard = segmentSpan  * (static_cast<double>(segmentID) + 1.0);
        // use of trapezoidal area measurement for the wing strips
        double segmentArea = (getLocalChordLength(liftingSurfaces.at(liftingSurfaceIndex), ySegmentInboard)
                            + getLocalChordLength(liftingSurfaces.at(liftingSurfaceIndex), ySegmentOutboard))
                            * 0.5 * fabs(ySegmentInboard - ySegmentOutboard);
        double fuselage_width_at_liftingSurface_intersection(geom2::measure::width(theFuselage,
            (liftingSurfaces.at(liftingSurfaceIndex).origin.x() + geom2::measure::offset_LE(liftingSurfaces.at(liftingSurfaceIndex), -ySegmentCenter).x()
             + 0.5 * geom2::measure::chord(liftingSurfaces.at(liftingSurfaceIndex), -ySegmentCenter))));
        
        if (ySegmentInboard > 0.5 * fuselage_width_at_liftingSurface_intersection) {
            double t2c = interpolateThicknessRatio(liftingSurfaceIndex, ySegmentCenter);
            /* Distinguish between subsonic and transonic Mach number for the used sweep angle position.*/
            double refLength(1);
            /* Check whether sign changes between LE and TE. If so, take the position where phi=0 as new reference length*/
            if (signum(getLocalSweep(liftingSurfaces.at(liftingSurfaceIndex), ySegmentCenter, 0., "RADIAN"))
                != signum(getLocalSweep(liftingSurfaces.at(liftingSurfaceIndex), ySegmentCenter, 1.0, "RADIAN"))) {
                    myRuntimeInfo->debug << "LE sweep and TE sweep have different signs. "
                                    << "LE sweep: " << getLocalSweep(liftingSurfaces.at(liftingSurfaceIndex), ySegmentCenter, 0., "DEGREE")
                                    << " TE sweep: " << getLocalSweep(liftingSurfaces.at(liftingSurfaceIndex), ySegmentCenter, 1.0, "DEGREE")
                                    << ". New reference length is calculated from LE to "
                                    << "chord position where sweep = 0 degree ." << std::endl;
                    double gradient = (getLocalSweep(liftingSurfaces.at(liftingSurfaceIndex), ySegmentCenter, 1.0, "RADIAN")
                                - getLocalSweep(liftingSurfaces.at(liftingSurfaceIndex), ySegmentCenter, 0.0, "RADIAN"));
                    double intersection = getLocalSweep(liftingSurfaces.at(liftingSurfaceIndex), ySegmentCenter, 0.0, "RADIAN");
                    refLength = -intersection/gradient; // Solution of linear equation. x(at phi=0)= -phi(0)/(phi(1)-phi(0))
                    myRuntimeInfo->debug << "New reference length: " << refLength << "." << std::endl;
            }
            double sweep = getLocalSweep(liftingSurfaces.at(liftingSurfaceIndex), ySegmentCenter, chordWiseSweepPosition * refLength, "RADIAN");
            // check if maximum sweep is violated
            if (fabs(convertUnit(RADIAN, DEGREE, sweep)) > maxAllowedSweep) {
                // just one warning, otherwise the console gets clogged
                if (!tooLargeSweep) {
                    myRuntimeInfo->warn << "Some wing segments have too large sweep" << endl;
                    myRuntimeInfo->warn << "Maximum allowed sweep is " << maxAllowedSweep << " degree" << endl;
                    myRuntimeInfo->warn << "Wing segments with to large sweep are excluded from wave drag calculation!" << endl;
                }
                tooLargeSweepInSegment = true;
                tooLargeSweep = true;
            }
            // Local Cl from lift distribution (getClFromLiftDistribution needs an ascending order of span- and lift-distribution)
            vector<double> reverseClDistribution(Cl_distribution);
            vector<double> reverseSpanDistribution(spanStations);
            reverse(reverseClDistribution.begin(), reverseClDistribution.end());
            reverse(reverseSpanDistribution.begin(), reverseSpanDistribution.end());
            double Cl_loc(getClFromLiftDistribution(ySegmentCenter, reverseSpanDistribution, reverseClDistribution));
            // Determination of Mdd, Mcrit and Cd_wave according to [Gur2010] Eq 15 - 18 page 1360 ff and [Mas2011]
            double Mdd = factorKa / cos(sweep) - t2c / pow(cos(sweep), 2.) - 0.1 * Cl_loc / pow(cos(sweep), 3.);
            double Mcrit = Mdd - pow(0.1 / 80.0, 1. / 3.);
            if (tooLargeSweepInSegment) {
                liftingSurfaceRefArea -=  2. * segmentArea;
                C_DWaveSegment.push_back(0.);
            } else if (aMachNumber >= Mcrit) {
                C_DWaveSegment.push_back(20. * pow((aMachNumber - Mcrit), 4.) * segmentArea);
            } else {
                C_DWaveSegment.push_back(0.);
            }
            // Transformation from 2D to 3D according to [Ris2016] Eq. 3.30 page 87
            C_DWaveSegment.back() *= pow(cos(sweep), 3);
        }
    }
    // double the drag to account for the whole wing
    C_DWaveTotal = 2. * std::accumulate(C_DWaveSegment.begin(), C_DWaveSegment.end(), 0.0) / liftingSurfaceRefArea;
    if (doCorrection) {
        C_DWaveTotal += calculateWaveDragCorrectionByCalibration(C_DWaveTotal, aPolarPoint, aMachNumber);
    }
    return C_DWaveTotal;
}

double waveDragMason::calculateWaveDragCorrectionByCalibration(const double& waveDragBeforeCorrection, const PolarPoint& aPolarPoint, const double& aMachNumber) {
    double totalWaveDragCorrection(0.);
    /* Drag calibration. First introduced in commit [4eab519c2] based on experimental data in [Lam2010], 23.06.2010. After several changes, 
       the current definition and values in config are from Risse, commit [2ec91580] on 02.08.2015. Especially for wave drag calibration, the values should be used 
       carefully, as they usually only fit to one specific airfoil family and, thus, k-factor. */    
    totalWaveDragCorrection = CLFact * pow(aMachNumber * aPolarPoint.CL, CLExp);
    if (aMachNumber >= 0.5) {
        totalWaveDragCorrection = CLFact * pow(aMachNumber * aPolarPoint.CL, CLExp);
    }
    return totalWaveDragCorrection;
}

double waveDragMason::calculateWaveDragCorrectionByDragCounts(const double& waveDragCtCorrForCalibration) {
    double totalWaveDragCorrection(0.);
    // Drag correction based on user input in config
    if (doWaveDragMasonDragCountCorrection) {
        totalWaveDragCorrection += waveDragCtCorrForCalibration * 0.0001;
    }
    return totalWaveDragCorrection;
}

double waveDragMason::interpolateThicknessRatio(const size_t surfaceID, const double& yCoordinate) {
    size_t innerSectionID(0), outerSectionID(0);
    for (size_t sectionID(0); sectionID < surfaceSectionThicknesses.at(surfaceID).size(); ++sectionID) {
        if (accuracyCheck(yCoordinate, surfaceSectionThicknesses.at(surfaceID).at(sectionID).span, ACCURACY_LOW)) {
            return surfaceSectionThicknesses.at(surfaceID).at(sectionID).thicknessRatio;
        } else if (yCoordinate > surfaceSectionThicknesses.at(surfaceID).at(sectionID).span && yCoordinate < surfaceSectionThicknesses.at(surfaceID).at(sectionID + 1).span) {
            innerSectionID = sectionID;
            outerSectionID = sectionID + 1;
            break;
        }
    }
    double localthickness = linearInterpolation(yCoordinate,
                                                surfaceSectionThicknesses.at(surfaceID).at(innerSectionID).span,
                                                surfaceSectionThicknesses.at(surfaceID).at(innerSectionID).thicknessRatio,
                                                surfaceSectionThicknesses.at(surfaceID).at(outerSectionID).span,
                                                surfaceSectionThicknesses.at(surfaceID).at(outerSectionID).thicknessRatio);
    return localthickness;
}
