/*  Copyright (C) 2010-2024 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef CALCULATEPOLAR_SRC_POLAR_H_
#define CALCULATEPOLAR_SRC_POLAR_H_

#include <string>
#include <unordered_map>
#include <vector>
#include <map>

// ------------------------ Basic Polar Point to inherit from ---------------------------- //
class PolarPoint {
 public:
        double AoA; /**< Angle of attack [deg]*/
        double CL;  /**< Lift coefficient [-] */
        double CD; /**< Total drag coefficient [-] */
        double CM; /**< Moment coefficient [-]*/
        double LoverD; /**< Lift-to-drag ratio [-] */
        double iStabPolar; /**< current stabilzer incidence angle [deg] */

        /** \brief Function sums all drag coefficients to CD and calculates drag-to-lift ratio for non-zero CDtotal
         */
        /** \brief Constructor of polarPoint class
         */
        PolarPoint();
        /** \brief Virtual destructor of polarPoint class
         */
        virtual ~PolarPoint() {}
};

// ---------------------------- implement new types of Polar Points here -------------------------------------//



class defaultPolarPoint : public PolarPoint {
 public:
    double CD_ind; /**< Total induced drag coefficient [-] */
    double CD_visc; /**< Viscous drag coefficient [-]*/
    double CD_wave; /**< Wave drag coefficient [-] */

    defaultPolarPoint();
    virtual ~defaultPolarPoint() {}



};

class tawDBPolarPoint : public PolarPoint {
 public:
    double CD_ind; /**< Total induced drag coefficient [-] */
    double CD_visc_turb;
    double CD_visc_lam;
    double CD_wave; /**< Wave drag coefficient [-] */

    tawDBPolarPoint();
    virtual ~tawDBPolarPoint() {}
};

    /** \brief Class for lifting surfaces, also contains lift distribution and coefficients of respective areas
     */
class defaultPolarPointLiftSurface: public defaultPolarPoint {
 public:
        std::vector <double> yStations;     /**< Vector of y stations of lifting surface [m] */
        std::vector <double> Cl_distr;      /**< Vector of local lift coefficients [-] */
        std::vector <double> Lift_distr;    /**< Vector of lift distribution calculated by multiplying local lift coefficient with the corresponding cord length [-] */
        /** \brief Constructor of polarPointLiftSurface class
         */
        defaultPolarPointLiftSurface();
        virtual ~defaultPolarPointLiftSurface() {}
};

// -------------------------------- Basic Polar to inherit from ---------------------------- //

    /** \brief Class for total polar
    */
class Polar {
 public:
    std::string Config;             /**< Vector of calculated mission section configurations [-] */
    double MachNumber;              /**< Mach number [-]*/
    double altitude;                /**< Altitude at the polar point [m] */
    double ReynoldsNumber;          /**< Reynold's number [-] */
    //double CLmax;                   /**< Maximum lift coefficient [-] */
    //double LoverDmax;               /**< Maximum lift-to-drag ratio [-] */
    double extrapolationMargin; /**< Extrapolation margin */
    bool allowGridChange; /**< Switch whether to allow Mach number grid change when processing final polar data */
    std::vector<PolarPoint> PolarPoints;
    Polar();
    virtual ~Polar() {}
};

// ------------------------------------- implement new polar types here -------------------------------------------- //

class tawDefaultPolar : public Polar {
 public:
    std::vector<defaultPolarPoint> Points;
    std::vector<defaultPolarPoint> PointsFuselage;
    std::vector<defaultPolarPoint> PointsNacelles;
    std::vector<defaultPolarPoint> PointsFin;
    std::vector<defaultPolarPoint> PointsMisc;
    std::vector<defaultPolarPointLiftSurface> PointsWing;
    std::vector<defaultPolarPointLiftSurface> PointsStab;

    tawDefaultPolar();
    virtual ~tawDefaultPolar() {}
};

#endif  // CALCULATEPOLAR_SRC_POLAR_H_
