/*  Copyright (C) 2010-2024 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef CALCULATEPOLAR_SRC_TAW_TAWCALCULATEPOLARCONFIG_H_
#define CALCULATEPOLAR_SRC_TAW_TAWCALCULATEPOLARCONFIG_H_

#include <aixml/endnode.h>

#include <map>
#include <set>
#include <string>
#include <vector>

#include "../flightCondition.h"

/** \brief Settings class for calculatePolar. Initializes tool with settings set in the respective config file.
 *  Inherits from the settings class (located in libraries and in unicadoRuntimeLibs)
 */

class tawCalculatePolarConfig{
 public:
    /* !!! WARNING: All EndnodeReadOnly Values are converted automatically to SI-Units/RADIAN when read in */
    EndnodeReadOnly<std::string>execution_mode;
    EndnodeReadOnly<std::string>pathToLiftingLine;
    // general lifting line settings
    EndnodeReadOnly<double> stepWidthCL;
    EndnodeReadOnly<std::string> CLModeAirfoilSelection;
    // fligt condition parameters
    EndnodeReadOnly<std::string> setFlightConditionsMode;
    EndnodeReadOnly<double> extrapolationMargin;
    EndnodeReadOnly<bool> allowGrideChange;
    // lifting line pitching moment parameters
    EndnodeReadOnly<std::string> CM_corr_fuselage_mode;
    EndnodeReadOnly<std::string> CM_corr_nacelle_mode;
    EndnodeReadOnly<double> delta_CM0;
    EndnodeReadOnly<double> delta_dCMdCL;
    // lifting line for TAW induced drag calibraitons
    EndnodeReadOnly<double> indDragCtCorrForCalibration;
    EndnodeReadOnly<double> factorIndDragCleanPolar;
    // visc drag raymer
    EndnodeReadOnly<bool> manualTransition;
    EndnodeReadOnly<double> manualTransitionWing;
    EndnodeReadOnly<double> manualTransitionStabilizer;
    EndnodeReadOnly<double> manualTransitionFin;
    EndnodeReadOnly<bool> doViscDragCalibration;
    EndnodeReadOnly<double> CDvisc_CDsum;
    EndnodeReadOnly<double> CDvisc_CLfact;
    EndnodeReadOnly<double> CDvisc_CLexp;
    EndnodeReadOnly<double> CDvisc_CDsum_lowMa;
    EndnodeReadOnly<double> CDvisc_CLfact_lowMa;
    EndnodeReadOnly<double> CDvisc_CLexp_lowMa;
    EndnodeReadOnly<bool> doViscousDragRaymerDragCountCorrection;
    EndnodeReadOnly<double> deltaViscousDragCountWing;
    EndnodeReadOnly<double> deltaViscousDragCountFuselage;
    EndnodeReadOnly<double> deltaViscousDragCountHTP;
    EndnodeReadOnly<double> deltaViscousDragCountVTP;
    EndnodeReadOnly<double> deltaViscousDragCountNacelles;
    EndnodeReadOnly<double> deltaViscousDragCountMisc;
    // wave drag mason
    EndnodeReadOnly<int> numberWingStrips;
    EndnodeReadOnly<bool> useCostomReferenceAngleForSweep;
    EndnodeReadOnly<double> Ka;
    EndnodeReadOnly<double> maximumSegmentSweep;
    EndnodeReadOnly<double> customSweepPosition;
    EndnodeReadOnly<bool> doWaveDragMasonCalibration;
    EndnodeReadOnly<double> CDwave_CLFact;
    EndnodeReadOnly<double> CDwave_CLExp;
    EndnodeReadOnly<bool> doWaveDragMasonDragCountCorrection;
    EndnodeReadOnly<double> deltaWaveDragCountWing;
    EndnodeReadOnly<double> deltaWaveDragCountStabilizer;
    // trimInterpolationSettings
    EndnodeReadOnly<bool> doTrim;                           /**< Switch to do trim calculations */
    EndnodeReadOnly<bool> trimAdditionalCGPositions;        /**< Switch to calculate trimmed polars for additional center of gravity positions */
    EndnodeReadOnly<bool> trimHighLiftPolars;               /**< Switch to trimm the high lift polars */
    EndnodeReadOnly<std::string> stabAngleGridMode;         /**< Mode selector to define how many stabilizer angles are calculated */
    EndnodeReadOnly<std::string> customStabAngleGrid;       /**< Custom sequence of stabilizer angles read in from the config file */
    double design_trim_MTOM_fraction;                       /**< gives the fraction of MTOM  in regqards to design center of gravity calculation, unit [-] */
    // semi empirical high lift adaptions
    EndnodeReadOnly<double> deltaTotalDragHL;
    EndnodeReadOnly<double> factorDragHL;
    bool do_individual_hl_calibration;
    bool apply_manual_CLmax_hl_values;

    //Methods
    void readflightConditions(const node& config, size_t ID);
    void initializeTrimSettings(const node& config);
    void read_hl_calibration_factors(const node& config);
    explicit tawCalculatePolarConfig(const node& config);
    virtual ~tawCalculatePolarConfig() {}
    // explicit calculatePolarSettings(const node& configXML);
    /** \brief  Virtual destructor of calculatePolarSettings
     */
    // virtual ~calculatePolarSettings();
    /* Member Variables */
    std::string liftingLineFolderPath; /**< Path to LiftingLine folder */

    // trim settings
    std::vector<double> stabAngles; /**< vector of stabilizer angles in [deg] */
    std::vector<std::string> CGPositions;  /**< vector of CG positions, order: design, fwd, aft*/
    size_t neutralStabAngleID;

    /* Flight Conditions */
    std::vector<flightCondition> flightConditions; /**< Vector of flightCondition objects representing the mission data */
    std::vector<double> machNumbers;  /**< Vector with mach numbers corrsponding to their mach ID */

    /* High lift calibration data */
    struct hl_polar_calibration_values {
        double CL_max;          /**< maximum CL value which should be reached */
        double delta_CD;        /**< Constant delta for drag calibration */
        double factor_CD;       /**< Factor for drag calibration */
    };
    std::unordered_map<std::string, hl_polar_calibration_values> hl_polar_calibration_data; /**< Container for storing all high lift polar calibration and correction values */

 private:
    /* Member functions */
    /** \brief Function reads polar settings from configuration XML and stores them in polarConfig object
     *  \param &polarNode aixml node that contains polar configuration information from corresponding XML file
     */
    void readPolarSettings(const node& polarNode);
    /** \brief Function reads flight condition settings from configuration XML and stores them in class variables
     *  \param &flightConditionNode aixml node that contains flight condition information from corresponding XML file
     */
    void readFlightConditions(const node& flightConditionNode);
    /** \brief Function reads viscous drag settings from configuration XML and stores them in class variables
     *  \param &viscousDragNode aixml node that contains flight condition information from corresponding XML file
     */
    void readViscousDragSettings(const node& viscousDragNode);
    /** \brief Function reads wave drag settings from configuration XML and stores them in class variables
     *  \param &viscousDragNode aixml node that contains flight condition information from corresponding XML file
     */
    void readWaveDragSettings(const node& viscousWaveNode);
};

#endif  // CALCULATEPOLAR_SRC_TAW_TAWCALCULATEPOLARCONFIG_H_
