/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include <cmath>
#include <vector>
#include <numbers>
#include <iostream>
#include <string>
#include <algorithm>
#include <format>
#include <variant>
#include <moduleBasics/module.h>
#include <engine/engine.h>
#include "io/aircraft_xml.h"
#include "constraint_analysis.h"
#include "energy_based/energy_based.h"

void DefaultAnalysis::initialize() {
    /* Load XML resources */
    this->aircraft_xml = aixml::openDocument(this->get_RuntimeIO()->acxmlAccess);
    this->configuration_xml = aixml::openDocument(this->get_RuntimeIO()->moduleConfAccess);
    std::string engine_model = this->aircraft_xml->at("aircraft_exchange_file/component_design/propulsion/specific/propulsion/engine/model/value");
    std::filesystem::path file = this->get_RuntimeIO()->acxmlAccess;
    this->engine_directory = file.parent_path() / "engine_data" / engine_model;
    this->mission_directory = file.parent_path() / "mission_data";
    std::string polar_file = this->aircraft_xml->at("aircraft_exchange_file/analysis/aerodynamics/polar/polar_file/value");
    std::filesystem::path polarFile = file.parent_path() / "aero_data" / polar_file;
    auto polar_xml = aixml::openDocument(polarFile);
    std::filesystem::path plot_dir = this->get_RuntimeIO()->getPlotDir();

    /* Select the constraint analysis strategy*/
    const std::string method(this->configuration_xml->at("module_configuration_file/program_settings/method/value"));
    if (method == "Energy_Based")
        this->constraint_analyzer = std::make_unique<EnergyBased>(
            this->configuration_xml,
            polar_xml,
            this->aircraft_xml,
            this->engine_directory,
            this->mission_directory,
            plot_dir);
    else
        throw std::runtime_error("[Constraint_Analysis] The constraint analysis method '" + method + "' is not supported.");

    this->constraint_analyzer->initialize();
}

void DefaultAnalysis::run() {
    (*this->constraint_analyzer)();;
}

void DefaultAnalysis::update() {
}

void DefaultAnalysis::report() {
}

void DefaultAnalysis::save() {
}
