/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "energy_based.h"
#include <algorithm>
#include <filesystem>
#include <format>
#include <list>
#include <ranges>
#include <cmath>
#include <vector>
#include <numbers>
#include <iostream>
#include <string>
#include <moduleBasics/module.h>
#include <engine/engine.h>
#include "../../include/constraint_analysis/ca_functions.h"
#include "../../include/constraint_analysis/ca_parser.h"
#include "../constraint_analysis.h"
#include "../../include/constraint_analysis/ca_plotting.h"
#include "../io/aircraft_xml.h"
#include "../../include/constraint_analysis/ca_minfinder.h"


void EnergyBased::initialize()
{
    if (!this->aircraft_xml()) {
        throw std::runtime_error("[EnergyBased::initialize()] aircraft_xml is null!");
    };
    if (!this->configuration()) {
        throw std::runtime_error("[EnergyBased::initialize()] config_xml is null!");
    };
    if (!this->polar()) {
        throw std::runtime_error("[EnergyBased::initialize()] polar_xml is null!");
    };
}

void EnergyBased::operator()()
{

    /*Define relevant Wing_loading range. Selected values should be valid for all possible configurations*/
    const int WS_start = 50;
    const int WS_end = 820;
    constexpr size_t WS_size = WS_end - WS_start + 1; // Range from 50 to 820, inclusive
    std::vector<int> numbers(WS_size);

    std::vector<double> W_over_S_data(WS_size);

    /*Transform from kg/m^2 into N/m^2*/
    for (int i = 0; i < WS_size; ++i) {
        W_over_S_data[i] = (WS_start + i) * 9.81;
    }

    Mattingly::constraint_analysis constraint_analysis_tool;

    atmosphere atm;

    // Set atmosphere according to the delta ISA requirement for the sizing process
    double temperature_ISA = ISA_TEMPERATURE + this->aircraft_xml()->at("aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission/delta_ISA/value");
    atm.setAtmosphere(0.0, temperature_ISA, ISA_PRESSURE);

    constraint_analysis_tool.atm = atm;

    double CD = 0.0;
    double CL = 0.0;
    std::vector<double> CD_vector = {};
    std::vector<double> CL_vector = {};

    Engine engine = Engine(this->engine_directory());
    auto calculate_thrust_lapse = [&atm, &engine](double M, double altitude) {
        const double thrust_lapse = engine.get_thrust_lapse("cruise", atm, M, altitude);
        return thrust_lapse;
        };

    auto get_num_engines = [&]() {
        double num_engines = 0;

        // Get the node vector for all defined propulsion systems
        auto propulsors = this->aircraft_xml()->getVector(
            "aircraft_exchange_file/requirements_and_specifications/design_specification/propulsion/propulsor");

        // Increment num_engines for each propulsor
        for (const auto& propulsor : propulsors) {
            num_engines += 1;
        }
        return num_engines;
    };
    std::filesystem::path missionCSV = this->mission_directory();
    for (const auto& entry : fs::directory_iterator(this->mission_directory())) {
        std::string filename = entry.path().filename().string();

        if (filename.find("design_mission") != std::string::npos && filename.find("_out.csv") != std::string::npos) {
            missionCSV /= filename;
            break;
        }
    }
    readMission mission_data(missionCSV);

    /* Take-off Segment */
    double CL_max_TO = this->aircraft_xml()->at("aircraft_exchange_file/analysis/aerodynamics/lift_coefficients/C_LmaxT-O/value");
    double s_G_TO = this->aircraft_xml()->at("aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/takeoff_distance/value");
    double takeoff_climb_angle = this->configuration()->at("module_configuration_file/program_settings/takeoff_climb_angle/value");
    double M_TO = this->configuration()->at("module_configuration_file/program_settings/Mach_TO/value"); 
    double thrust_lapse_TO = calculate_thrust_lapse(M_TO, 0.0);
    double weight_fraction_TO = mission_data.get_beta(" takeoff", 1.0); //from mission analysis
    const auto to_ground_roll = constraint_analysis_tool.takeoff_ground_roll(
        W_over_S_data,
        weight_fraction_TO,
        thrust_lapse_TO,
        M_TO,
        0.0,
        s_G_TO,
        CL_max_TO);

    readPolar takeoff_config(this->polar(), "takeoff", "linear", M_TO);
    std::vector<double> coeffs = takeoff_config.quadratic_coefficients();
    double K_1 = coeffs[0];
    double K_2 = coeffs[1];
    double C_D0 = coeffs[2];

    const auto to_climb_angle = constraint_analysis_tool.takeoff_climb_angle(
        W_over_S_data,
        K_1,
        K_2,
        C_D0,
        weight_fraction_TO,
        thrust_lapse_TO,
        M_TO,
        0.0,
        CL_max_TO,
        takeoff_climb_angle);

    /* OEI Climb */
    double climb_speed = this->aircraft_xml()->at(
        "aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission/climb_speed_schedule/climb_speed_below_FL100/value");
    double climb_gradient_OEI = this->configuration()->at("module_configuration_file/program_settings/climb_gradient_OEI/value"); 
    double M_climb = climb_speed / atm.getSpeedOfSound(0.0);

    double num_engines = get_num_engines();

    readPolar climb_config(this->polar(), "takeoff", "linear", M_TO);
    for (int i = 0; i < W_over_S_data.size(); ++i) {
        CD = climb_config.getCD(W_over_S_data[i], 1.0, weight_fraction_TO, (climb_speed / atm.getSpeedOfSound(0.0)), 0.0);
        CD_vector.push_back(CD);
    };

    double thrust_factor_OEI = thrust_lapse_TO *((num_engines - 1)/ num_engines);
    double climb_rate_OEI = climb_gradient_OEI * climb_speed;
    double weight_fraction_climb_OEI = mission_data.get_beta(" takeoff");
    /*minimal climb rate of 2.4% defined in CS25 for second segment climb*/
    const auto climb_OEI = constraint_analysis_tool.constant_speed_climb(
        W_over_S_data,
        CD_vector,
        weight_fraction_climb_OEI,
        thrust_factor_OEI,
        M_climb,
        0.0,
        climb_rate_OEI); /* DEFINE CLIMB RATE CONSTRAINT */

    /* Service Ceiling */
    double altitude_max = this->aircraft_xml()->at(
        "aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/flight_envelope/maximum_operating_altitude/value");
    double M_cruise = this->aircraft_xml()->at(
        "aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/study_mission/initial_cruise_mach_number/value");

    double M_polar = M_cruise;
    double weight_fraction_climb_service_ceiling = mission_data.get_beta(" climb_to_cruise");
    
    bool polar_read = false;
    readPolar clean_config_M_max(this->polar(), "clean", "linear", M_TO);
    while (!polar_read) {
        try {
            clean_config_M_max = readPolar(this->polar(), "clean", "linear", M_polar);
            if (clean_config_M_max.CD_polar.size() >= 1)
            {
                polar_read = true;
            }
            else {
                M_polar += 0.01;
            };
        }
        catch (...) {
            M_polar += 0.01;
        }
    }
    double thrust_lapse_segment = calculate_thrust_lapse(M_cruise, altitude_max);
    CD_vector = {};
    CL_vector = {};
    coeffs = clean_config_M_max.quadratic_coefficients();
    K_1 = coeffs[0];
    K_2 = coeffs[1];
    C_D0 = coeffs[2];

    const std::string method(this->configuration()->at("module_configuration_file/program_settings/aero_method/value"));
    if (method == "Read_Polar") {
        for (int i = 0; i < W_over_S_data.size(); ++i) {
            CD = clean_config_M_max.getCD(W_over_S_data[i], 1.0, weight_fraction_climb_service_ceiling, M_polar, altitude_max);
            CD_vector.push_back(CD);
        }
    } else if (method == "Calculate_Polar") {
        const double b = this->aircraft_xml()->at("aircraft_exchange_file/analysis/aerodynamics/reference_values/b/value");
        const double S_ref = this->aircraft_xml()->at("aircraft_exchange_file/analysis/aerodynamics/reference_values/S_ref/value");
        double AR = pow(b, 2) / S_ref;
        const double e = this->configuration()->at("module_configuration_file/program_settings/oswald_factor/value");
        for (int i = 0; i < W_over_S_data.size(); ++i) {
            CD = clean_config_M_max.calculateCD(W_over_S_data[i], 1.0, weight_fraction_climb_service_ceiling, M_polar, altitude_max, AR, e, C_D0);
            CD_vector.push_back(CD);
        }
    } else {
        throw std::runtime_error("Invalid method for calculating CD");
    }

    double minimum_climb_rate = this->configuration()->at("module_configuration_file/program_settings/minimum_climb_rate/value");

    /* Service ceiling with the minimum climb rate requirement */
    const auto service_ceiling = constraint_analysis_tool.constant_speed_climb(
        W_over_S_data,
        CD_vector,
        weight_fraction_climb_service_ceiling,
        thrust_lapse_segment,
        M_cruise,
        altitude_max,
        minimum_climb_rate); 

    /* Landing */
    double thrust_lapse_landing = calculate_thrust_lapse(M_TO, 0.0);
    double CL_max_L = this->aircraft_xml()->at("aircraft_exchange_file/analysis/aerodynamics/lift_coefficients/C_LmaxLanding/value");
    double my_B = this->aircraft_xml()->at("aircraft_exchange_file/requirements_and_specifications/requirements/additional_requirements/landing_gear/braking_coefficient/value");
    double s_G_L = this->aircraft_xml()->at("aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/takeoff_distance/value");
    readPolar landing_config(this->polar(), "landing", "linear", M_TO);
    double CD_max_L = landing_config.interpCD(CL_max_L);
    double max_landing_weight = this->aircraft_xml()->at("aircraft_exchange_file/analysis/masses_cg_inertia/maximum_landing_mass/mass_properties/mass/value");
    double MTOM = this->aircraft_xml()->at("aircraft_exchange_file/analysis/masses_cg_inertia/maximum_takeoff_mass/mass_properties/mass/value");
    double weight_fraction_landing = max_landing_weight / MTOM;
    const auto landing = constraint_analysis_tool.braking_roll(CD_max_L, CL_max_L, weight_fraction_landing, thrust_lapse_landing, M_TO, 0.0, my_B, s_G_L);

    /* Gust */
    double AoA_1 = landing_config.interpalpha(CL_max_L);
    double AoA_2 = landing_config.interpalpha(CL_max_L - CL_max_L * 0.9);
    double CL_alpha = (CL_max_L - (CL_max_L - CL_max_L * 0.9)) / ((AoA_1 - AoA_2) * PI / 180.0);
    double V_TO_L = M_TO * atm.getSpeedOfSound(0.0);
    double w_g = this->configuration()->at("module_configuration_file/program_settings/gust_speed/value");
    double dn_G = this->configuration()->at("module_configuration_file/program_settings/gust_load_factor/value");
    const auto gust = constraint_analysis_tool.gust(CL_alpha, 0.0, V_TO_L, w_g, dn_G, weight_fraction_TO);

    /*Define Constraint out of each non-vertical constraint case*/
    Constraint Case1(W_over_S_data, to_ground_roll, "Takeoff Ground Roll");
    Constraint Case2(W_over_S_data, to_climb_angle, "Takeoff Climb Angle");
    Constraint Case3(W_over_S_data, climb_OEI, "OEI Climb");
    Constraint Case4(W_over_S_data, service_ceiling, "Service Ceiling");

    std::vector<Constraint> constraint_list = { Case1, Case2, Case3, Case4 };

    /* Define boundaries out of vertical constraint cases*/
    std::vector<double> boundaries = { gust, landing };

    /* Define Simple_Analysis*/
    Simple_Analysis min_finder = Simple_Analysis(constraint_list, this->aircraft_xml(), boundaries);

    /* Find dominant curve of the design chart*/
    min_finder.find_dominant_curve();

    double safety_factor = this->configuration()->at("module_configuration_file/program_settings/safety_factor/value");

    /* Find design point with safety factor*/
    min_finder.find_design_point(safety_factor);

    /*Update design point in ACxml*/
    min_finder.update_design_point();

    std::string plot_dir_string=this->plot_dir().string();

    // Add output of design point in logfile

    myRuntimeInfo->out << "New wing loading: " << min_finder.design_wing_loading * 9.81  << " N/m^2. New thrust to weight ratio: " <<  min_finder.design_thrust_to_weight << std::endl;

    min_finder.write_design_point(plot_dir_string + "/csv_files/design_point.csv");

    /* Plot the complete design chart with constraints, boundaries and design point*/
    std::string plotting = this->configuration()->at("module_configuration_file/control_settings/plot_output/enable/value");
    if (plotting == "true") {
        ConstraintAnalysis_Plot ca_plot;

        ca_plot.fill_infeasible_area(W_over_S_data, min_finder.dominant_thrust_to_weight, boundaries);
        ca_plot.add_curve(W_over_S_data, to_ground_roll, "Takeoff Ground Roll");
        ca_plot.add_curve(W_over_S_data, to_climb_angle, "Takeoff Climb Angle");
        ca_plot.add_curve(W_over_S_data, climb_OEI, "OEI Climb");
        ca_plot.add_curve(W_over_S_data, service_ceiling, "Service Ceiling");
        ca_plot.add_curve(landing, "Landing");
        ca_plot.add_curve(gust, "Gust");
        ca_plot.add_curve({ min_finder.design_wing_loading * 9.81 }, { min_finder.design_thrust_to_weight }, "Design Point");

        ca_plot.save_figure(plot_dir_string + "/constraint_plot.svg");
    };
}
