/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

 /* === Includes === */
#include <gtest/gtest.h>
#include <aixml/node.h>

/* Unit Under Test */
#include "constraint_analysis/ca_minfinder.h"

TEST(MinFinder, initialization)
{
    std::vector<double> W_over_S = { 150,160,170,180,190,200,210,220,230,240,250 };
    std::vector<double> T_over_W = { 1.5,1.6,1.7,1.8,2.2,2.5,2.8,3.3,4.5,6.8,9.1 };
    std::string name = "trial";
    Constraint case1 = Constraint(W_over_S, T_over_W, name);
    T_over_W = { 1.5,1.7,1.9,2.1,2.3,2.5,2.7,2.9,3.1,3.3,3.5 };
    name = "trial2";
    Constraint case2 = Constraint(W_over_S, T_over_W, name);
    std::vector<Constraint> constraint_list = {case1, case2};
    std::vector<double> boundaries = { 200, 230 };
    Simple_Analysis simple_analysis(constraint_list, boundaries);
};

TEST(MinFinder, initialization2)
{
    std::filesystem::path file{ CMAKE_TEST_STUBS_DIR };
    std::filesystem::path acXMLFile = file / "csmr-2020_startCSD.xml";

    std::vector<double> W_over_S = { 150,160,170,180,190,200,210,220,230,240,250 };
    std::vector<double> T_over_W = { 1.5,1.6,1.7,1.8,2.2,2.5,2.8,3.3,4.5,6.8,9.1 };
    std::string name = "trial";
    Constraint case1 = Constraint(W_over_S, T_over_W, name);
    T_over_W = { 1.5,1.7,1.9,2.1,2.3,2.5,2.7,2.9,3.1,3.3,3.5 };
    name = "trial2";
    Constraint case2 = Constraint(W_over_S, T_over_W, name);
    T_over_W = { 9.8,7.6,5.5,4.3,2.1,1.0,1.0,1.2,1.3,1.4,1.5 };
    name = "trial3";
    Constraint case3 = Constraint(W_over_S, T_over_W, name);
    std::vector<Constraint> constraint_list = { case1, case2, case3 };
    std::vector<double> boundaries = { 170, 260 };
    Simple_Analysis simple_analysis(constraint_list, aixml::openDocument(acXMLFile),  boundaries);
    simple_analysis.find_dominant_curve();
    simple_analysis.find_design_point(0.0);
    simple_analysis.update_design_point();
};

TEST(MinFinder, simpleAnalysis)
{
    std::vector<double> W_over_S = { 150,160,170,180,190,200,210,220,230,240,250 };
    std::vector<double> T_over_W = { 1.5,1.6,1.7,1.8,2.2,2.5,2.8,3.3,4.5,6.8,9.1 };
    std::string name = "trial";
    Constraint case1 = Constraint(W_over_S, T_over_W, name);
    T_over_W = { 1.5,1.7,1.9,2.1,2.3,2.5,2.7,2.9,3.1,3.3,3.5 };
    name = "trial2";
    Constraint case2 = Constraint(W_over_S, T_over_W, name);
    T_over_W = { 9.8,7.6,5.5,4.3,2.1,1.0,1.0,1.2,1.3,1.4,1.5 };
    name = "trial3";
    Constraint case3 = Constraint(W_over_S, T_over_W, name);
    std::vector<Constraint> constraint_list = { case1, case2, case3 };
    std::vector<double> boundaries = { 200, 230 };
    Simple_Analysis simple_analysis(constraint_list, boundaries);
    simple_analysis.find_dominant_curve();
    EXPECT_NEAR(simple_analysis.dominant_thrust_to_weight[0], 9.8, 1e-5);
    EXPECT_NEAR(simple_analysis.dominant_thrust_to_weight[2], 5.5, 1e-5);
    EXPECT_NEAR(simple_analysis.dominant_thrust_to_weight[4], 2.3, 1e-5);
    EXPECT_NEAR(simple_analysis.dominant_thrust_to_weight[6], 2.8, 1e-5);
    EXPECT_NEAR(simple_analysis.dominant_thrust_to_weight[8], 4.5, 1e-5);
    EXPECT_NEAR(simple_analysis.dominant_thrust_to_weight[10], 9.1, 1e-5);
};

TEST(MinFinder, minFind)
{
    std::vector<double> W_over_S = { 150,160,170,180,190,200,210,220,230,240,250 };
    std::vector<double> T_over_W = { 1.5,1.6,1.7,1.8,2.2,2.5,2.8,3.3,4.5,6.8,9.1 };
    std::string name = "trial";
    Constraint case1 = Constraint(W_over_S, T_over_W, name);
    T_over_W = { 1.5,1.7,1.9,2.1,2.3,2.5,2.7,2.9,3.1,3.3,3.5 };
    name = "trial2";
    Constraint case2 = Constraint(W_over_S, T_over_W, name);
    T_over_W = { 9.8,7.6,5.5,4.3,2.1,1.0,1.0,1.2,1.3,1.4,1.5 };
    name = "trial3";
    Constraint case3 = Constraint(W_over_S, T_over_W, name);
    std::vector<Constraint> constraint_list = { case1, case2, case3 };
    std::vector<double> boundaries = {170, 260};
    Simple_Analysis simple_analysis(constraint_list, boundaries);
    simple_analysis.find_dominant_curve();
    simple_analysis.find_design_point(0.0);
    EXPECT_NEAR(simple_analysis.design_thrust_to_weight, 2.3, 1e-5);
    EXPECT_NEAR(simple_analysis.design_wing_loading, 190, 1e-5);
};