# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2024 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

# Import standard libraries.
from numpy import sqrt


def calculate_air_traffic_control_costs(dict_ac_data, dict_operating_costs, runtime_output):
    """Calculate air traffic control (ATC) costs.

    The calculation of the ATC costs is based on the EUROCONTROL route charge formula, more precisely the aircraft
    weight factor. "The weight factor (expressed to two decimals) is determined by dividing, by fifty (50), the
    certificated Maximum Take-Off Weight (MTOW) of the aircraft (in metric tonnes, to one decimal) and subsequently
    taking the square root of the result rounded to the second decimal" [1]. The ATC price factor considers the fact
    that the price scenarios are varying strongly for each continent (or even region).

    This function adds the 'air_traffic_control_costs' dict to the 'dict_operating_costs' that contains the following
    parameters:
        - array air_traffic_control_costs: ATC costs (range dependent) in EUR
        - float air_traffic_control_costs_design_point: ATC costs at design point of current aircraft in EUR

    [1] EUROCONTROL Central Route Charge Office (CRCO): "Customer Guide to Charges", November 2022.
    URL: https://www.eurocontrol.int/sites/default/files/2022-11/eurocontrol-customer-guide-to-charges.pdf.
    (Last access: 26.09.2023).

    :param dict dict_ac_data: Dict containing parameters and values from aircraft exchange and module config file
    :param dict dict_operating_costs: Dict containing parameters from cost estimation calculation
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict dict_operating_costs: Dict containing parameters from cost estimation calculation
    """

    # Extract necessary variables from 'dict_ac_data'.
    air_traffic_control_price_factor = dict_ac_data['air_traffic_control_price_factor']
    m_takeoff_max = dict_ac_data['m_takeoff_max']
    # Extract necessary values from 'dict_operating_costs'.
    n_annual_flights = dict_operating_costs['dependent_parameter']['flights_per_year']
    idx_design_point = dict_operating_costs['payload_range_diagram_data']['idx_design_point']
    ranges_for_prd = dict_operating_costs['payload_range_diagram_data']['ranges_for_prd']/1000

    # Calculate ATC costs.
    air_traffic_control_costs = (ranges_for_prd * air_traffic_control_price_factor * sqrt((m_takeoff_max / 1000) / 50)
                                 * n_annual_flights)

    # Print.
    runtime_output.print('Air traffic control costs (per year): '
                         + f"{int(air_traffic_control_costs[idx_design_point]):,}" + ' €')

    # Prepare output data.
    dict_operating_costs['air_traffic_control_costs'] = {}
    dict_operating_costs['air_traffic_control_costs']['air_traffic_control_costs'] = air_traffic_control_costs
    dict_operating_costs['air_traffic_control_costs']['air_traffic_control_costs_design_point'] = \
        air_traffic_control_costs[idx_design_point]

    return dict_operating_costs
