/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 * 
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include <aixml/endnode.h>
#include <string>
#include <moduleBasics/module.h>
#include <unordered_map>
#include "standard_mission/standard_mission.h"

void StandardMission::set_departure_profile() {
    /* Create objects for take off */
    if (this->takeoff_procedure == 0) { // This is for a standard take-off procedure, others can be added later
        /* Transfer parameters: Mode, rating, derate, config., endAlt., rateOfclimb, endCAS*/
        this->departure_steps.push_back(DepartureStep("takeoff", "takeoff", 1., "takeoff", NAN, NAN, NAN));
        this->departure_steps.push_back(DepartureStep("climb", this->climb_thrust_setting, 1.,
                                        "takeoff_landing_gear_retracted", convertUnit(FOOT, METER, 1500.), this->maximum_rate_of_climb, NAN));
        this->departure_steps.push_back(DepartureStep("climb", this->climb_thrust_setting, 1.,
                                        "takeoff_landing_gear_retracted", convertUnit(FOOT, METER, 3000.), this->maximum_rate_of_climb, NAN));
        this->departure_steps.push_back(DepartureStep("accelerate", this->climb_thrust_setting, 1.,
                                        "climb", NAN, convertUnit(FOOTPERMINUTE, METERPERSECOND, 1000.),  convertUnit(KNOTS, METERPERSECOND, 210.)));
        this->departure_steps.push_back(DepartureStep("accelerate", this->climb_thrust_setting, 1.,
                                        "clean", NAN, convertUnit(FOOTPERMINUTE, METERPERSECOND, 1200.), this->CAS_ATC_limit_climb));
        this->departure_steps.push_back(DepartureStep("climb", this->climb_thrust_setting, 1., "clean",
                                        convertUnit(FOOT, METER, 10000.), this->maximum_rate_of_climb, NAN));
    } else if (this->takeoff_procedure == 1) { // The ICAO-A take-off procedure has not yet been implemented, error message
        throwError(__FILE__, __func__, __LINE__, "ICAO-A Take-Off Procedure not yet implemented.");
    } else if (this->takeoff_procedure == 2) { // The ICAO-B take-off procedure has not yet been implemented, error message
        throwError(__FILE__, __func__, __LINE__, "ICAO-B Take-Off Procedure not yet implemented.");
    } else if (this->takeoff_procedure == 3) { // Standard_19-Seat_Commuter take-off procedure
        /* Parameters:                                 #Step, Mode,         rating,     derate, config.,              endAlt., rateOfclimb, endCAS */
        this->departure_steps.push_back(DepartureStep("takeoff", "takeoff", 1., "takeoff", NAN, NAN, NAN));
        this->departure_steps.push_back(DepartureStep("climb", this->climb_thrust_setting, 1.,
                                        "takeoff_landing_gear_retracted", convertUnit(FOOT, METER, 200.), this->maximum_rate_of_climb, NAN));
        this->departure_steps.push_back(DepartureStep("accelerate", this->climb_thrust_setting, 1., "clean", NAN, 1200.,
                                        this->CAS_ATC_limit_climb));
        this->departure_steps.push_back(DepartureStep("climb",
                                        this->climb_thrust_setting, 1., "clean", convertUnit(FOOT, METER, 10000.), this->maximum_rate_of_climb,  NAN));
    } else { // Unknown take-off procedure, error message
        myRuntimeInfo->err << "Unkown take-Off procedure in config file. Abort!" << std::endl;
        exit(1);
    }
}

void StandardMission::set_cruise_profile() {
    /* Create objects for cruise */
    if (this->CAS_over_flight_level_100_climb >= this->CAS_ATC_limit_climb) {
        /* If 2nd CAS (e.g. 300KTAS) larger than 1st CAS (e.g. 250KTAS) -> create segment for acceleration to second CAS ("change_speed_to_CAS") */
        this->cruise_steps.push_back(CruiseStep("change_speed_to_CAS",
                                            this->climb_thrust_setting, 1., "clean", NAN, this->auto_select_flight_level,
                                            this->round_to_regular_flight_level, this->auto_climb_altitude_steps, false,
                                            convertUnit(FOOTPERMINUTE, METERPERSECOND, 300.), this->CAS_over_flight_level_100_climb, NAN, NAN));
    } else {
        /* If 2nd CAS (e.g. 300KTAS) smaller than 1st CAS -> create segment for deceleration to second CAS ("change_speed_to_CAS") */
        this->cruise_steps.push_back(CruiseStep("change_speed_to_CAS",
                                            this->climb_thrust_setting, 1., "clean", NAN, this->auto_select_flight_level,
                                            this->round_to_regular_flight_level, this->auto_climb_altitude_steps, false,
                                            this->maximum_rate_of_climb, this->CAS_over_flight_level_100_climb, NAN, NAN));
    }
    if (this->mission_type == "requirements_mission") { // Mission type: requirement mission
        /* For requirement mission only one additional cruise step --> climb to ceiling requirements! After that, the requirement mission can be aborted */
        this->cruise_steps.push_back(CruiseStep("climb_to_ceiling",
                                            this->climb_thrust_setting, 1., "clean",
                                            this->cruise_step_segments.front().altitude,
                                            this->auto_select_initial_cruise_altitude, this->round_to_regular_flight_level,
                                            this->auto_climb_altitude_steps, false, NAN,
                                            NAN, this->initial_cruise_mach.value(), NAN));
    } else { // Mission type: design/study mission
        /* For design and study mission, additional cruise segments have to be created. climb to ICA and accelerate to crossover Mach number */
        this->cruise_steps.push_back(CruiseStep("climb_to_cruise", this->climb_thrust_setting, 1., "clean",
                                            this->cruise_step_segments.front().altitude,
                                            this->auto_select_initial_cruise_altitude, this->round_to_regular_flight_level,
                                            this->auto_climb_altitude_steps, false, this->maximum_rate_of_climb, NAN,
                                            this->mach_climb, NAN));
        /* accelerate to initial cruise Mach number */
        this->cruise_steps.push_back(CruiseStep("change_speed_to_Mach", this->climb_thrust_setting, 1., "clean", NAN,
                                            this->auto_select_flight_level, this->round_to_regular_flight_level,
                                            this->auto_climb_altitude_steps, false, NAN, NAN, this->initial_cruise_mach.value(), NAN));
        std::string mode;
        std::string rating;
        double rate_of_climb_limit;
        double relative_segment_length;
        double end_altitude;
        /* Create additional steps (according to definition in XML file) */
        int CruiseStep_blocks(3 + (2 * this->cruise_step_segments.size() - 1) + 1);

        for (int i = 4; i <= CruiseStep_blocks - 1; i++) {
            if (i % 2 == 0) { // Mode: cruise
                mode = "cruise";
                rating = "cruise";
                if (this->auto_select_flight_level) {
                    rate_of_climb_limit = this->auto_rate_of_climb_steps;
                } else {
                    rate_of_climb_limit = NAN;
                }
                relative_segment_length = this->cruise_step_segments[i / 2 - 2].relative_segment_length;
                end_altitude = NAN;
            } else { // Mode: change flight level...
                if (this->auto_rate_of_climb_steps > 0.) { // ... at a constant rate of climb
                    rate_of_climb_limit = this->auto_rate_of_climb_steps;
                    mode = "change_flight_level_constant_ROC";
                } else { // ... at an undefined rate of climb
                    rate_of_climb_limit = NAN;
                    mode = "change_flight_level";
                }
                rating = this->climb_thrust_setting;
                relative_segment_length = NAN;
                end_altitude =  this->cruise_step_segments[(i - 1) / 2 - 1].altitude;
            }
            this->cruise_steps.push_back(CruiseStep(mode, rating, 1., "clean", end_altitude,
                                            this->auto_select_flight_level, this->round_to_regular_flight_level,
                                            this->auto_climb_altitude_steps, false, rate_of_climb_limit, NAN, NAN, relative_segment_length));
        }
        this->cruise_steps.push_back(CruiseStep("descend_to_approach", "idle", 1., "clean", convertUnit(FOOT, METER, 10000.),
                                            this->auto_select_flight_level, this->round_to_regular_flight_level,
                                            this->auto_climb_altitude_steps, false, NAN,
                                            this->CAS_over_flight_level_100_descent, NAN, NAN));
    }
}

void StandardMission::set_approach_profile() {
    /* Create objects for approach - by now, no rate of climb value used in those procedures, but are kept in for possible adaption */
    if (this->approach_procedure == 0) { // Standard approach procedure
        /* Get speeds */
        double v_1_temp(this->CAS_ATC_limit_descent);
        double v_2_temp(std::max(convertUnit(KNOTS, METERPERSECOND, 170.), this->approach_speed + convertUnit(KNOTS, METERPERSECOND, 5.)));
        double v_3_temp(this->approach_speed + convertUnit(KNOTS, METERPERSECOND, 5.));
        double v_4_temp(this->approach_speed);
        /* Creation of the approach steps: mode, rating, derate, config., endAlt, rate_of_climb_limit, endCAS, GD, glidePath */
        this->approach_steps.push_back(ApproachStep("change_speed", "idle", 1., "clean", NAN, NAN, v_1_temp, convertUnit(FOOT, METER, 5000.), 0.));
        this->approach_steps.push_back(ApproachStep("descend", "cruise", 1., "clean", convertUnit(FOOT, METER, 3000.), NAN, NAN, NAN, convertUnit(DEGREE, RADIAN, -3.)));
        this->approach_steps.push_back(ApproachStep("change_speed", "idle", 1., "approach", NAN, NAN, v_2_temp, convertUnit(FOOT, METER, 10000.), 0.));
        this->approach_steps.push_back(ApproachStep("level_glide_slope_interception", "cruise", 1., "approach_landing_gear_out", NAN, NAN, NAN, NAN, NAN));
        this->approach_steps.push_back(ApproachStep("change_speed", "idle", 1., "approach_landing_gear_out",
                                            NAN, NAN, v_3_temp,  convertUnit(FOOT, METER, 6000.), convertUnit(DEGREE, RADIAN, -3.)));
        this->approach_steps.push_back(ApproachStep("descend", "cruise", 1., "approach_landing_gear_out",
                                            convertUnit(FOOT, METER, 1500.), NAN, NAN, NAN, convertUnit(DEGREE, RADIAN, -3.)));
        this->approach_steps.push_back(ApproachStep("change_speed", "idle", 1., "landing",
                                            NAN, NAN, v_4_temp, convertUnit(FOOT, METER, 5000.), convertUnit(DEGREE, RADIAN, -3.)));
        this->approach_steps.push_back(ApproachStep("descend", "cruise", 1., "landing", convertUnit(FOOT, METER, 1000.), NAN, NAN, NAN, convertUnit(DEGREE, RADIAN, -3.)));
        this->approach_steps.push_back(ApproachStep("descend", "cruise", 1., "landing", convertUnit(FOOT, METER, 50.), NAN, NAN, NAN, convertUnit(DEGREE, RADIAN, -3.)));
        this->approach_steps.push_back(ApproachStep("landing", "takeoff", 1., "landing", 0., NAN, NAN, NAN, convertUnit(DEGREE, RADIAN, -3.)));
    } else if (this->approach_procedure == 1) { // The CDA approach procedure has not yet been implemented, error message
        myRuntimeInfo->err << "Continuous descent approach procedure not yet implemented. Abort!" << std::endl;
        exit(1);
    } else if (this->approach_procedure == 2) { // The CDA-Steep approach procedure has not yet been implemented, error message
        myRuntimeInfo->err << "Steep continuous descent approach procedure not yet implemented. Abort!" << std::endl;
        exit(1);
    } else if (this->approach_procedure == 3) { // Standard_19-Seat-Commuter approach procedure
        /* Get speeds */
        double v_1_temp(this->CAS_ATC_limit_descent);
        double v_2_temp(this->approach_speed);
        if (fabs(this->CAS_ATC_limit_descent - this->CAS_over_flight_level_100_descent) > ACCURACY_LOW) {
            /* Creation of the approach steps: mode , rating, derate, config., endAlt, rate_of_climb_limit, endCAS, GD, glidePath */
            this->approach_steps.push_back(ApproachStep("change_speed", "idle", 1., "clean", NAN, NAN, v_1_temp, convertUnit(FOOT, METER, 5000), 0.));
            this->approach_steps.push_back(ApproachStep("descend", "cruise", 1., "clean", convertUnit(FOOT, METER, 2000.),  NAN, NAN, NAN, convertUnit(DEGREE, RADIAN, -3.)));
            this->approach_steps.push_back(ApproachStep("change_speed", "idle", 1., "approach",
                                                            NAN, NAN, v_2_temp, convertUnit(FOOT, METER, 6200.), convertUnit(DEGREE, RADIAN, -3.)));
            this->approach_steps.push_back(ApproachStep("descend", "cruise", 1., "landing", convertUnit(FOOT, METER, 50.), NAN, NAN, NAN, convertUnit(DEGREE, RADIAN, -3.)));
            this->approach_steps.push_back(ApproachStep("landing", "takeoff", 1., "landing", 0., NAN, NAN, NAN, convertUnit(DEGREE, RADIAN, -3.)));
        } else {
            /* Creation of the approach steps: mode, rating, derate, config., endAlt, rate_of_climb_limit, endCAS, GD, glidePath */
            this->approach_steps.push_back(ApproachStep("descend", "cruise", 1., "clean", convertUnit(FOOT, METER, 2000.),  NAN, NAN, NAN, convertUnit(DEGREE, RADIAN, -3.)));
            this->approach_steps.push_back(ApproachStep("change_speed", "idle",   1., "approach",
                                                            NAN, NAN, v_2_temp, convertUnit(FOOT, METER, 6200.), convertUnit(DEGREE, RADIAN, -3.)));
            this->approach_steps.push_back(ApproachStep("descend", "cruise", 1., "landing", convertUnit(FOOT, METER, 50.), NAN, NAN, NAN, convertUnit(DEGREE, RADIAN, -3.)));
            this->approach_steps.push_back(ApproachStep("landing", "takeoff",     1., "landing", 0., NAN, NAN, NAN, convertUnit(DEGREE, RADIAN, -3.)));
        }
    } else { // Unknown approach procedure, error message
        myRuntimeInfo->err << "Unknown approach procedure in config file. Abort!" << std::endl;
        exit(1);
    }
}
