# mission_emissions {#mission-emissions}
The submodule _mission_emissions_ is the only part of _ecological\_assessment_ that will always be executed. It is handled via the class `Mission` and provides various options to calculate the emissions of kerosene or hydrogen-burning engines during a mission.

## General principles {#mission-emissions-generalprinciples}
The following steps are executed within the mission class:
- `get_mission_data`, `write_emissions_path_csv`, and `get_engine_thermodynamics_LTO`: Data from the `mission.csv` file is imported, engine data for every mission step is determined and saved in a CSV file, and the engine thermodynamics during the landing and takeoff phase according to ICAO definition are calculated.
- Emission calculation: Depending on the defined engine carrier, the emissions will be calculated for every mission step.
    - Kerosene:
        - The emissions of CO2, H2O, SO2, and SO4 are considered to be proportional to the fuel flow. Therefore, they are calculated via \f$ m_{emission} = EI * m_{fuel}\f$, with
            - \f$ m_{emission}\f$: emission mass \f$[kg]\f$
            - \f$ EI \f$: emission index \f$[\frac{kg_{emission}}{kg_{fuel}}]\f$
            - \f$ m_{fuel} \f$: fuel mass  \f$[kg]\f$
        - All other emissions are considered to be non-proportional. For NOx emissions, there are a P3T3 Method \cite Nor03, Boeing Fuel Flow Method 2 \cite Sch13, and the calculation based on data generated by GasTurb available. For HC as well as CO emissions, the DLR Omega method and Boeing Fuel Flow Method 2 \cite Sch13 are implemented. Additionally, there is the option to calculate the landing and takeoff cycle emissions based on constants provided by ICAO. Soot emissions can be determined via a DLR correlation based on ICAO smoke numbers or a correlation by R.B. Whyte \cite Kug05. Alternatively, it can be assumed to be proportional to the consumed fuel.
    - Hydrogen: Only H2O and NOx emissions are produced.
        - H2O is assumed to be proportional to the fuel flow.
        - For NOx emissions, the emissions of kerosene-burning engines are calculated via the P3T3 method, and a correction factor is used to derive the hydrogen-burning emissions \cite Koss22. You can choose between the calculation of the correction factor and the use of constant emission indices, which were determined in \cite Koss22 for one engine type. As the determination of NOx emissions during hydrogen burning is subject to great uncertainty, the low-fidelity method of using constants is both used as the default method and also implemented as a fallback method if the calculation of a correction factor fails.

## Input Data {#mission-emissions-input}
For the mission emission calculation, the following parameters are needed in the `aircraft_exchange_file`:
```xml
<aircraft_exchange_file>
    <requirements_and_specifications>
        <general>
            <type>
                <value>CeRAS</value>
            </type>
            <model>
                <value>SMR-2020</value>
            </model>
        </general>
        <design_specification>
            <configuration>
                <configuration_type>
                    <value>tube_and_wing</value>
                </configuration_type>
                <aerodynamic_technologies>
                    <value>false</value>
                </aerodynamic_technologies>
            </configuration>
            <energy_carriers>
                <energy_carrier ID="0">
                    <type>
                        <value>kerosene</value>
                    </type>
                </energy_carrier>
            </energy_carriers>
            <assessment_scenario>
                <flights_per_year>
                    <unit>1</unit>
                    <value>2227</value>
                    <lower_boundary>0</lower_boundary>
                    <upper_boundary>4000</upper_boundary>
                </flights_per_year>
            </assessment_scenario>
        </design_specification>
    <requirements_and_specifications>
<aircraft_exchange_file>
```

In the `ecological_assessment_conf.xml`, next to the control settings block, the following program settings are needed:
```xml
<program_settings>
    <strategy_selector>
        <value>STANDARD</value>
    </strategy_selector>
    <standard_strategy>
        <emission_calculation>
            <mission_emissions>
                <emission_methods>
                    <kerosene>
                        <HC_method_selector>
                            <value>mode_0</value>
                        </HC_method_selector>
                        <CO_method_selector>
                            <value>mode_0</value>
                        </CO_method_selector>
                        <NOx_method_selector>
                            <value>mode_0</value>
                        </NOx_method_selector>
                        <soot_method_selector>
                            <value>mode_0</value>
                        </soot_method_selector>
                    </kerosene>
                    <hydrogen_combustion>
                        <NOx_method_selector>
                            <value>mode_0</value>
                        </NOx_method_selector>
                    </hydrogen_combustion>
                </emission_methods>
                <relative_humidity>
                    <unit>1</unit>
                    <value>0.6</value>
                    <lower_boundary>0</lower_boundary>
                    <upper_boundary>1</upper_boundary>
                </relative_humidity>
            </mission_emissions>
        </emission_calculation>
        <duration_operation>
            <unit>a</unit>
            <value>25</value>
            <lower_boundary>0</lower_boundary>
            <upper_boundary>50</upper_boundary>
        </duration_operation>
    </standard_strategy>
</program_settings>
```

As the engine library will be used, you can check the documentation page of the library to get information about input files. From the `engine_data_file`, the following parameters are needed:
```xml
<EngineDataFile>
    <EngineDesignCondition>
        <flightAltitude>35000</flightAltitude>
        <flightMachNumber>0.8</flightMachNumber>
    </EngineDesignCondition>
    <Deck>
        <WFToP3min>45</WFToP3min>
        <WFToP3max_MTO>101.83</WFToP3max_MTO>
        <WFToP3max_MCT>98.74</WFToP3max_MCT>
        <TempMax_MTO>848.69</TempMax_MTO>
        <TempMax_MCT>772.35</TempMax_MCT>
        <N2max>1.034</N2max>
        <NfractionClimbRating>0.97</NfractionClimbRating>
        <NfractionCruiseRating>0.875</NfractionCruiseRating>
        <RelBleedMax>0.3</RelBleedMax>
        <ShaftPowerExtractionMax>150</ShaftPowerExtractionMax>
        <UnscaledSLST>110.3126953</UnscaledSLST>
    </Deck>
    <EmissionFactors>
        <NOxFactor>32</NOxFactor>
    </EmissionFactors>
    <ICAOEngineData>
        <LTOFuelFlow>
            <Taxi>0.128</Taxi>
            <Takeoff>1.053</Takeoff>
            <ClimbOut>0.88</ClimbOut>
            <Approach>0.319</Approach>
        </LTOFuelFlow>
        <LTOEmissions>
            <HCFactor>
                <Taxi>0.105e-3</Taxi>
                <Approach>0.061e-3</Approach>
                <ClimbOut>0.041e-3</ClimbOut>
                <Takeoff>0.041e-3</Takeoff>
            </HCFactor>
            <COFactor>
                <Taxi>12.43e-3</Taxi>
                <Approach>2.44e-3</Approach>
                <ClimbOut>0.62e-3</ClimbOut>
                <Takeoff>0.53e-3</Takeoff>
            </COFactor>
            <NOxFactor>
                <Taxi>4.7e-3</Taxi>
                <Approach>8.9e-3</Approach>
                <ClimbOut>22.3e-3</ClimbOut>
                <Takeoff>26.5e-3</Takeoff>
            </NOxFactor>
            <SNFactor>
                <Taxi>2.6</Taxi>
                <Takeoff>5.2</Takeoff>
                <ClimbOut>7.2</ClimbOut>
                <Approach>4.2</Approach>
            </SNFactor>
        </LTOEmissions>
    </ICAOEngineData>
</EngineDataFile>
```

Also, some data from `mission.xml` will be read:
```xml
<mission>
    <taxi_time_origin>
        <unit>s</unit>
        <value>540</value>
        <lower_boundary>0</lower_boundary>
        <upper_boundary>10000</upper_boundary>
    </taxi_time_origin>
    <taxi_time_destination>
        <unit>s</unit>
        <value>300</value>
        <lower_boundary>0</lower_boundary>
        <upper_boundary>10000</upper_boundary>
    </taxi_time_destination>
    <range>
        <unit>m</unit>
        <value>4630000</value>
        <lower_boundary>0</lower_boundary>
        <upper_boundary>100000000</upper_boundary>
    </range>
</mission>
```
And last but not least, you need to provide the `mission.csv` file written by the UNICADO mission module and located in _aircraft\_exchange\_file\_directory/missionData_.

## Output-Data {#mission-emissions-output}
The central output of the mission submodule is the `ecological_assessment_results_file.xml` which you will find in the *aircraft\_exchange\_file\_directory/reporting/report_xml*. It contains all calculated emission masses. Additionally, there is a CSV file in the folder *aircraft_exchange_file_directory/reporting/plots/csv_files/* with mission and engine data for every mission step. As described in [getting started](#getting-started), an HTML report including some plots will be generated.
