/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include <vector>
#include <algorithm>

#include "development.h"
#include <aixml/node.h>
#include <runtimeInfo/runtimeInfo.h>
#include <unitConversion/unitConversion.h>
#include <standardFiles/functions.h>

using std::pow;

Development::Development()
    :
    myEcoDatabase(ecoDatabase()),
    energy_total(0.),
    fuel_total(0.),
    taxi_fuel_test_factor(0.),
    test_aircraft_production_energy(0.),
    test_engine_production_energy(0.) {
        //ctor
    }

void Development::calculate_development_resources(const std::shared_ptr<SchaeferIOData>& data) {
    /* Tests at SL und ISA+delta_isa (assumed); max. Takeoff Thrust SL and ISA (see Ruijgrok, P. 298 and ICAO 2008 P. III-2-2*/
    // TODO(franz#1#): T142 check if with offtakes (see engineSizing output cpp, here if necessary with a part released out again) at which atm. conditions engine tests
    //max. Thrust
    /*BEGINN: from engineSizing output.cpp, to determine Max. Cont. Thrust for SL */
    /*Maximum Continuous */
    std::vector<double>  N1;
    /* For bucket curve without offtakes */
    std::vector<double>  Thrust;
    std::vector<double>  FuelFlow;
    std::vector<double>  SFC;
    int ID_maxThrust(0);
    int ID_bucketSFC(0);
    do {
        if (static_cast<int>(N1.size()) == 0) {
            N1.push_back(0.7);
        } else {
            N1.push_back(N1.back() + 0.001);
        }
        /* Set engine without offtakes */
        data->myEngine.calculate_N1_with_penalties(0, 0., myAtmosphere, Rounding(N1.back(), 4), "maximum_continuous", 0., 0.);
        Thrust.push_back(data->myEngine.get_thrust_aircraft() / data->number_of_engines);
        FuelFlow.push_back(data->myEngine.get_aircraft_fuelflow() / data->number_of_engines);
        SFC.push_back(FuelFlow.back() / Thrust.back());
    } while (N1.back() <= 1.);
    // Determination of the bucket point (run through from the back, as discontinuity is possible at low rotation speeds with props)
    double tmp_SFC = SFC.back();
    for (unsigned int i(SFC.size() - 2); i > 0; i--) {
        if (SFC.at(i) <= tmp_SFC) {
            tmp_SFC = SFC.at(i);
            ID_bucketSFC = i;
        } else {
            break;
        }
    }
    /* max thrust*/
    std::vector<double>::iterator it1 = max_element(Thrust.begin(), Thrust.end());
    double MaxContThrustSL(*it1);// Thrust.at(ID_maxThrust); Maximum continuous thrust at Sea Level for one scaled engine [kN]
    ID_maxThrust = static_cast<int>(it1 - Thrust.begin());
    double SFCMaxContThrustSL = SFC.at(ID_maxThrust); // Thrust specific fuel consumption at continuous thrust at Sea Level for one scaled engine [(kg/s)/kN]
    /* bucket thrust; assumption: = const. thrust*/
    // TODO(franz#1#): T143 check if you can get a rating for TW; BucketThrust assumption wrong (at FL = 0 and Ma = 0 there is no bucket, curve goes down constantly)
    double ContThrustSL = Thrust.at(ID_bucketSFC); //Continuous thrust at Sea Level for one scaled engine [kN]
    double SFCContThrustSL = SFC.at(ID_bucketSFC); //Thrust specific fuel consumption at continuous thrust at Sea Level for one scaled engine [(kg/s)/kN]
    /*END: from engineSizing output.cpp, to determine Max. Cont. Thrust for SL */

    /* calculation unit MJ! (1 kWh = 3.6 MJ) **/
    if (data->configuration.development_emission_setting != "mode_0") {
        /* Calculation of personnel expenses **/
        theEngeneering.engineering_hours = 1.309 * 5.18 * pow(data->ome, 0.777) * pow(convertUnit(METERPERSECOND, KILOMETERPERHOUR, data->max_operating_speed), 0.894)
                                * pow(data->configuration.structural_test_aircraft_number + data->configuration.flight_tests_aircraft_number, 0.163);
        theEngeneering.engineering_energy_electricity = energy_perSqmYear_dev[0] * F_netFloor_dev / spaceEfficiency_dev * theEngeneering.engineering_hours / t_labourYear_dev
                                    * (1 + data->configuration.engine_engineering_factor);
        theEngeneering.engineering_energy_heat = energy_perSqmYear_dev[1] * F_netFloor_dev / spaceEfficiency_dev * theEngeneering.engineering_hours / t_labourYear_dev
                                    * (1 + data->configuration.engine_engineering_factor);
        theEngeneering.engineering_energy_total = theEngeneering.engineering_energy_electricity + theEngeneering.engineering_energy_heat;
        myRuntimeInfo->info << "F_netFloor_dev " << F_netFloor_dev << std::endl;
        myRuntimeInfo->info << "spaceEfficiency_dev " << spaceEfficiency_dev << std::endl;
        myRuntimeInfo->info << "T_eng " << theEngeneering.engineering_hours << std::endl;
        myRuntimeInfo->info << "t_labourYear_dev " << t_labourYear_dev << std::endl;
        /* Calculation of Tests (according to SA by Kuno Jandaurek)**/
        /* Energy of the wind tunnel test **/
        testResources.energy_wind_tunnel_test = convertUnit(HOUR, SECOND, data->configuration.wind_tunnel_test_hours) * 1.782; //Power of subsonic wind tunnel: 1782kW
        if (data->configuration.development_emission_setting == "mode_1") {
            myRuntimeInfo->info << "-----Test Emissions-----" << std::endl;
            myRuntimeInfo->info << "CAUTION: Test emissions without production and end-of-life emissions!" << std::endl;
        }
        myRuntimeInfo->info << "energy_wind_tunnel_test: " << testResources.energy_wind_tunnel_test << std::endl;
        /* Energy of Structure tests**/
        testResources.energy_fatigue_test = data->configuration.structural_test_cycles * data->mtom / 1000. * 1.45;
        myRuntimeInfo->info << "energy_fatigue_test: " << testResources.energy_fatigue_test << std::endl;
        /* Energy of System tests**/
        testResources.energy_system_tests = convertUnit(KILO, WATTHOUR, MEGA, JOULE, (data->configuration.system_integration_test_hours + data->configuration.iron_bird_test_hours)
                                                      * data->propulsion_design_power);
        myRuntimeInfo->info << "energy_system_tests: " << testResources.energy_system_tests << std::endl;
        if (data->configuration.do_engine_tests) {
            /* Fuel consumption for engine tests on rig (assumption: Sea Level Conditions; LTO emissions according to
               ICAO Engine Emission Databank (also NOx!); DesignMission as reference) **/
            // TODO(franz#1#): T136 NOx from TW-Decks instead of from ICAO Databank (like mission-class!))
            double rigHours;
            if (data->configuration.test_ETOPS) {
                if (data->configuration.include_ETOPS) {
                    rigHours = data->configuration.rig_test_hours;
                } else {
                    rigHours = data->configuration.rig_test_hours + 3000.*0.5;
                }
            } else {
                rigHours = data->configuration.rig_test_hours;
            }
            testResources.fuel_engine_test_rig = data->configuration.number_new_engines * rigHours * 3600.
                                               *(SFCMaxContThrustSL * MaxContThrustSL * data->configuration.max_cont_thrust_percentage
                                                 + SFCContThrustSL * ContThrustSL * (1. - data->configuration.max_cont_thrust_percentage));
            myRuntimeInfo->debug << "SFCMaxContThrustSL: " << SFCMaxContThrustSL << std::endl;
            myRuntimeInfo->debug << "MaxContThrustSL: " << MaxContThrustSL << std::endl;
            myRuntimeInfo->debug << "ContThrustSL: " <<  ContThrustSL << std::endl;
            /* Fuel consumption in engine tests at the FTB (assumption: cruise conditions; emissions of the
               design mission as reference, only specifications of the new engine are integrated) **/
            testResources.fuel_engine_test_flight = data->configuration.number_new_engines * convertUnit(HOUR, SECOND, data->configuration.flying_test_bed_hours)
                * ((data->configuration.flying_test_bed_engine_number * data->configuration.flying_test_bed_fuel_consumption) + data->fuel_flow_cont_thrust);
            myRuntimeInfo->info << "fuel_engine_test_flight: " << testResources.fuel_engine_test_flight << std::endl;
        } else {
            myRuntimeInfo->info << "No engine tests are performed because do_engine_tests switch is off." << std::endl;
        }
        /* Energy of ground tests */
        testResources.energy_ground_test = convertUnit(KILO, WATTHOUR, MEGA, JOULE, 0.95 * data->configuration.flight_test_hours * data->propulsion_design_power);
        myRuntimeInfo->info << "energy_ground_test: " << testResources.energy_ground_test << std::endl;
        /* Fuel consumption for flight tests */
        if (data->configuration.include_engine_options)
            testResources.flight_hours_flight_test = data->configuration.flight_test_hours;
        else
            testResources.flight_hours_flight_test = (0.63 + 0.37 * data->configuration.engine_options_number) * data->configuration.flight_test_hours;
        myRuntimeInfo->info << "flightHours: " << testResources.flight_hours_flight_test << std::endl;
        testResources.fuel_flight_test = data->blockfuel_design / data->blocktime_design *
                                        testResources.flight_hours_flight_test; //strongly simplified: Assumption: Const. Fuel consumption, determined from design mission.
        /* Test total: fuel consumption (without taxi, added at calcEmission) and energy (without Fatigue Test Airframe)**/
        testResources.energy_test_total = testResources.energy_wind_tunnel_test + testResources.energy_fatigue_test + testResources.energy_system_tests
                                            + testResources.energy_ground_test;
        testResources.fuel_engine_test_total =  testResources.fuel_engine_test_rig + testResources.fuel_engine_test_flight;
        myRuntimeInfo->debug << "testResources.fuel_engine_test_rig: " << testResources.fuel_engine_test_rig << std::endl;
        myRuntimeInfo->debug << "testResources.fuel_engine_test_flight: " << testResources.fuel_engine_test_flight << std::endl;
        testResources.fuel_test_total = testResources.fuel_engine_test_total + testResources.fuel_flight_test;
        myRuntimeInfo->debug << "testResources.fuel_engine_test_total: " << testResources.fuel_engine_test_total << std::endl;
        myRuntimeInfo->debug << "testResources.fuel_flight_test: " << testResources.fuel_flight_test << std::endl;
        energy_total = theEngeneering.engineering_energy_total + testResources.energy_test_total;
        fuel_total = testResources.fuel_test_total;
        // TODO(franz#1#): T135 Possible improvement: Fuel consumption for various types of fuel. Determine missisonen (duration);
        myRuntimeInfo->info << "fuel_flight_test: " << testResources.fuel_flight_test << std::endl;
    }
}

void Development::calculate_development_emissions(const std::shared_ptr<SchaeferIOData>& data, const Production &theProduction, const EndOfLife &theEndOfLife) {
    if (data->configuration.development_emission_setting != "mode_0") {
    double emission_index_CO2 = 3.149; // EI for kersosene combustion
    double emission_index_SO2 = 0.84*pow(10, -3); // EI for kersosene combustion
        /*Test phase*/
        /* Wind tunnel Test*/
        myRuntimeInfo->info << "      Wind tunnel Test Energy: " << testResources.energy_wind_tunnel_test << " MJ" << std::endl;
        windTunnelTests.CO2 = testResources.energy_wind_tunnel_test * myEcoDatabase.elecEmData.CO2;
        windTunnelTests.CH4 = testResources.energy_wind_tunnel_test * myEcoDatabase.elecEmData.CH4;
        windTunnelTests.N2O = testResources.energy_wind_tunnel_test * myEcoDatabase.elecN2O;
        windTunnelTests.PFC = testResources.energy_wind_tunnel_test * myEcoDatabase.elecPFC;
        windTunnelTests.NOx = testResources.energy_wind_tunnel_test * myEcoDatabase.elecEmData.NOx;
        windTunnelTests.SO2 = testResources.energy_wind_tunnel_test * myEcoDatabase.elecEmData.SO2;
        windTunnelTests.CO = testResources.energy_wind_tunnel_test * myEcoDatabase.elecEmData.CO;
        windTunnelTests.NMVOC = testResources.energy_wind_tunnel_test * myEcoDatabase.elecNMVOC;
        windTunnelTests.CED = testResources.energy_wind_tunnel_test * myEcoDatabase.elecEmData.specificEnergy;
        /* Structure Test*/
        myRuntimeInfo->info << "      Structure Test Energy (Fatigue Test): " << testResources.energy_fatigue_test << " MJ" << std::endl;
        structureTests.CO2 = testResources.energy_fatigue_test * myEcoDatabase.elecEmData.CO2;
        structureTests.CH4 = testResources.energy_fatigue_test * myEcoDatabase.elecEmData.CH4;
        structureTests.N2O = testResources.energy_fatigue_test * myEcoDatabase.elecN2O;
        structureTests.PFC = testResources.energy_fatigue_test * myEcoDatabase.elecPFC;
        structureTests.NOx = testResources.energy_fatigue_test * myEcoDatabase.elecEmData.NOx;
        structureTests.SO2 = testResources.energy_fatigue_test * myEcoDatabase.elecEmData.SO2;
        structureTests.CO = testResources.energy_fatigue_test * myEcoDatabase.elecEmData.CO;
        structureTests.NMVOC = testResources.energy_fatigue_test * myEcoDatabase.elecNMVOC;
        structureTests.CED = testResources.energy_fatigue_test * myEcoDatabase.elecEmData.specificEnergy;
        /* System Test*/
        myRuntimeInfo->info << "      System Test Energy: " << testResources.energy_system_tests << " MJ" << std::endl;
        systemTests.CO2 = testResources.energy_system_tests * myEcoDatabase.elecEmData.CO2;
        systemTests.CH4 = testResources.energy_system_tests * myEcoDatabase.elecEmData.CH4;
        systemTests.N2O = testResources.energy_system_tests * myEcoDatabase.elecN2O;
        systemTests.PFC = testResources.energy_system_tests * myEcoDatabase.elecPFC;
        systemTests.NOx = testResources.energy_system_tests * myEcoDatabase.elecEmData.NOx;
        systemTests.SO2 = testResources.energy_system_tests * myEcoDatabase.elecEmData.SO2;
        systemTests.CO = testResources.energy_system_tests * myEcoDatabase.elecEmData.CO;
        systemTests.NMVOC = testResources.energy_system_tests * myEcoDatabase.elecNMVOC;
        systemTests.CED = testResources.energy_system_tests * myEcoDatabase.elecEmData.specificEnergy;
        /* Ground Test incl. fuel prod.*/
        taxi_fuel_test_factor = 0.05 * data->configuration.flight_test_hours / (data->taxi_time / (3600.));
        myRuntimeInfo->info << "      Ground Test (System) Energy: " << testResources.energy_ground_test << " MJ" << std::endl;
        myRuntimeInfo->info << "      Taxi Test Fuel: " << data->taxi_fuel_design *taxi_fuel_test_factor << " kg" << std::endl;
        testResources.fuel_test_total += data->taxi_fuel_design * taxi_fuel_test_factor;
        fuel_total += data->taxi_fuel_design * taxi_fuel_test_factor;
        groundTests.CO2 = testResources.energy_ground_test * myEcoDatabase.elecEmData.CO2
                         + data->taxi_fuel_design * taxi_fuel_test_factor * (emission_index_CO2 + myEcoDatabase.fuelProdEmData.at(0).CO2);
        groundTests.CH4 = testResources.energy_ground_test * myEcoDatabase.elecEmData.CH4
                         + data->taxi_fuel_design * taxi_fuel_test_factor * myEcoDatabase.fuelProdEmData.at(0).CH4;
        groundTests.N2O = testResources.energy_ground_test * myEcoDatabase.elecN2O;
        groundTests.PFC = testResources.energy_ground_test * myEcoDatabase.elecPFC;
        groundTests.NOx = testResources.energy_ground_test * myEcoDatabase.elecEmData.NOx
                         + data->theTestMission.LTO.NOx * taxi_fuel_test_factor
                         + data->taxi_fuel_design * taxi_fuel_test_factor * myEcoDatabase.fuelProdEmData.at(0).NOx;;
        groundTests.SO2 = testResources.energy_ground_test * myEcoDatabase.elecEmData.SO2
                         + data->taxi_fuel_design * taxi_fuel_test_factor * emission_index_SO2;
        groundTests.CO = testResources.energy_ground_test * myEcoDatabase.elecEmData.CO
                        + data->theTestMission.LTO.CO * taxi_fuel_test_factor
                        + data->taxi_fuel_design * taxi_fuel_test_factor * myEcoDatabase.fuelProdEmData.at(0).CO;
        groundTests.NMVOC = testResources.energy_ground_test * myEcoDatabase.elecNMVOC;
        groundTests.HC = data->theTestMission.LTO.HC * taxi_fuel_test_factor;
        groundTests.H2O = data->theTestMission.LTO.H2O * taxi_fuel_test_factor
                         + data->taxi_fuel_design * taxi_fuel_test_factor * myEcoDatabase.fuelProdEmData.at(0).H2O;;
        groundTests.PM10 = data->theTestMission.LTO.soot * taxi_fuel_test_factor
                          + data->taxi_fuel_design * taxi_fuel_test_factor * myEcoDatabase.fuelProdEmData.at(0).PM10;
        groundTests.CED = testResources.energy_ground_test * myEcoDatabase.elecEmData.specificEnergy
                         + data->taxi_fuel_design * taxi_fuel_test_factor * myEcoDatabase.fuelSpecificEnergy.at(0);
        /* Engine Test incl. Fuel Production*/
        if (data->configuration.do_engine_tests) {
            double engine_testsFuelTotal = testResources.fuel_engine_test_total;
            engineTests.CO2 = engine_testsFuelTotal * (emission_index_CO2 + myEcoDatabase.fuelProdEmData.at(0).CO2);
            engineTests.CH4 = engine_testsFuelTotal * myEcoDatabase.fuelProdEmData.at(0).CH4;
            engineTests.N2O = 0.;
            engineTests.PFC = 0.;
            engineTests.NOx = testResources.fuel_engine_test_rig * (data->theTestMission.LTO.NOx / data->LTO_fuel_design)
                             + (data->theTestMission.cruise.NOx + data->theTestMission.LTO.NOx) / data->blocktime_design * data->configuration.flying_test_bed_hours
                             + engine_testsFuelTotal * myEcoDatabase.fuelProdEmData.at(0).NOx;
            engineTests.SO2 = (testResources.fuel_engine_test_rig + testResources.fuel_engine_test_flight) * emission_index_SO2;
            engineTests.CO = testResources.fuel_engine_test_rig * (data->theTestMission.LTO.CO / data->LTO_fuel_design)
                            + data->theTestMission.LTO.CO / data->blocktime_design * data->configuration.flying_test_bed_hours
                            + engine_testsFuelTotal * myEcoDatabase.fuelProdEmData.at(0).CO;; //no values for CO_Cruise, because from LTO data!
            engineTests.NMVOC = 0.;
            engineTests.HC = testResources.fuel_engine_test_rig * (data->theTestMission.LTO.HC / data->LTO_fuel_design)
                            + (data->theTestMission.cruise.HC + data->theTestMission.LTO.HC) / data->blocktime_design * data->configuration.flying_test_bed_hours;
            engineTests.H2O = testResources.fuel_engine_test_rig * (data->theTestMission.LTO.H2O / data->LTO_fuel_design)
                             + (data->theTestMission.cruise.H2O + data->theTestMission.LTO.H2O) / data->blocktime_design * data->configuration.flying_test_bed_hours
                             + engine_testsFuelTotal * myEcoDatabase.fuelProdEmData.at(0).H2O;; //but actually no matter, only important for cirrus formation
            engineTests.PM10 = testResources.fuel_engine_test_rig * (data->theTestMission.LTO.soot / data->LTO_fuel_design)
                              + (data->theTestMission.LTO.soot) / data->blocktime_design * data->configuration.flying_test_bed_hours
                              + engine_testsFuelTotal * myEcoDatabase.fuelProdEmData.at(0).PM10;; //PM10 interesting only during LTO, no greenhouse gas and not ozone-forming
            engineTests.CED = engine_testsFuelTotal * myEcoDatabase.fuelSpecificEnergy.at(0);
        }
        /* Flight Test incl. fuel prod. */
        myRuntimeInfo->info << "      Flight Test Fuel: " << testResources.fuel_flight_test << " kg" << std::endl;
        flightTests.CO2 = testResources.fuel_flight_test * (emission_index_CO2 + myEcoDatabase.fuelProdEmData.at(0).CO2);
        flightTests.CH4 = testResources.fuel_flight_test * myEcoDatabase.fuelProdEmData.at(0).CH4;
        flightTests.N2O = 0.;
        flightTests.PFC = 0.;
        flightTests.NOx = (data->theTestMission.cruise.NOx + data->theTestMission.LTO.NOx) / data->blocktime_design * testResources.flight_hours_flight_test
                         + testResources.fuel_flight_test * myEcoDatabase.fuelProdEmData.at(0).NOx;
        flightTests.SO2 = testResources.fuel_flight_test * emission_index_SO2;
        flightTests.CO = data->theTestMission.LTO.CO / data->blocktime_design * testResources.flight_hours_flight_test
                        + testResources.fuel_flight_test * myEcoDatabase.fuelProdEmData.at(0).CO;//no values for CO_Cruise, because from LTO data!
        flightTests.NMVOC = 0.;
        flightTests.HC = (data->theTestMission.cruise.HC + data->theTestMission.LTO.HC) / data->blocktime_design * testResources.flight_hours_flight_test; //ozone-forming
        flightTests.H2O = (data->theTestMission.cruise.H2O + data->theTestMission.LTO.H2O) / data->blocktime_design * testResources.flight_hours_flight_test
                         + testResources.fuel_flight_test * myEcoDatabase.fuelProdEmData.at(0).H2O;
        flightTests.PM10 = (data->theTestMission.LTO.soot) / data->blocktime_design * testResources.flight_hours_flight_test
                          + testResources.fuel_flight_test
                          * myEcoDatabase.fuelProdEmData.at(0).PM10;//PM10 only interesting during LTO, no greenhouse gas and no ozone-forming
        flightTests.CED = testResources.fuel_flight_test * myEcoDatabase.fuelSpecificEnergy.at(0);
    }
    if (data->configuration.development_emission_setting == "mode_2") {
        test_aircraft_production_energy = theProduction.energy_total * data->configuration.flight_tests_aircraft_number
                       + (theProduction.energy_total - theProduction.energy_engine)
                       * data->configuration.structural_test_aircraft_number; //because Fuel from Testflight and FuelProd. does not contribute to Energy!
        testResources.energy_test_total += test_aircraft_production_energy;
        energy_total += test_aircraft_production_energy;

        /*Production and EoL of test aircrafts*/
        myRuntimeInfo->info << "      Test AC Production Energy: " << test_aircraft_production_energy << " MJ" << std::endl;
        productionTestAircraft.CO2 = (theProduction.productionTotal.CO2 - theProduction.finalFlightTests.CO2 + theEndOfLife.endOfLifeTotal.CO2)
                                        * data->configuration.flight_tests_aircraft_number
                                        + (theProduction.productionTotal.CO2 - theProduction.productionEngine.CO2 - theProduction.finalFlightTests.CO2
                                             + theEndOfLife.endOfLifeTotal.CO2 - theEndOfLife.endOfLifeMaterialEngine.CO2 - theEndOfLife.endOfLifeTransport.CO2)
                                        * data->configuration.structural_test_aircraft_number;
        productionTestAircraft.H2O = (theProduction.productionTotal.H2O - theProduction.finalFlightTests.H2O + theEndOfLife.endOfLifeTotal.H2O)
                                        * data->configuration.flight_tests_aircraft_number
                                        + (theProduction.productionTotal.H2O - theProduction.productionEngine.H2O - theProduction.finalFlightTests.H2O
                                            + theEndOfLife.endOfLifeTotal.H2O - theEndOfLife.endOfLifeMaterialEngine.H2O - theEndOfLife.endOfLifeTransport.H2O)
                                        * data->configuration.structural_test_aircraft_number;
        productionTestAircraft.NOx = (theProduction.productionTotal.NOx - theProduction.finalFlightTests.NOx + theEndOfLife.endOfLifeTotal.NOx)
                                        * data->configuration.flight_tests_aircraft_number
                                        + (theProduction.productionTotal.NOx - theProduction.productionEngine.NOx - theProduction.finalFlightTests.NOx
                                            + theEndOfLife.endOfLifeTotal.NOx - theEndOfLife.endOfLifeMaterialEngine.NOx - theEndOfLife.endOfLifeTransport.NOx)
                                        * data->configuration.structural_test_aircraft_number;
        productionTestAircraft.SO2 = (theProduction.productionTotal.SO2 - theProduction.finalFlightTests.SO2 + theEndOfLife.endOfLifeTotal.SO2)
                                        * data->configuration.flight_tests_aircraft_number
                                        + (theProduction.productionTotal.SO2 - theProduction.productionEngine.SO2 - theProduction.finalFlightTests.SO2
                                            + theEndOfLife.endOfLifeTotal.SO2 - theEndOfLife.endOfLifeMaterialEngine.SO2 - theEndOfLife.endOfLifeTransport.SO2)
                                        * data->configuration.structural_test_aircraft_number;
        productionTestAircraft.CO = (theProduction.productionTotal.CO - theProduction.finalFlightTests.CO + theEndOfLife.endOfLifeTotal.CO)
                                        * data->configuration.flight_tests_aircraft_number
                                        + (theProduction.productionTotal.CO - theProduction.productionEngine.CO - theProduction.finalFlightTests.CO
                                             + theEndOfLife.endOfLifeTotal.CO - theEndOfLife.endOfLifeMaterialEngine.CO - theEndOfLife.endOfLifeTransport.CO)
                                        * data->configuration.structural_test_aircraft_number;
        productionTestAircraft.CH4 = (theProduction.productionTotal.CH4 - theProduction.finalFlightTests.CH4 + theEndOfLife.endOfLifeTotal.CH4)
                                        * data->configuration.flight_tests_aircraft_number
                                        + (theProduction.productionTotal.CH4 - theProduction.productionEngine.CH4 - theProduction.finalFlightTests.CH4
                                            + theEndOfLife.endOfLifeTotal.CH4 - theEndOfLife.endOfLifeMaterialEngine.CH4 - theEndOfLife.endOfLifeTransport.CH4)
                                        * data->configuration.structural_test_aircraft_number;
        productionTestAircraft.PM10 = (theProduction.productionTotal.PM10 - theProduction.finalFlightTests.PM10 + theEndOfLife.endOfLifeTotal.PM10)
                                        * data->configuration.flight_tests_aircraft_number
                                        + (theProduction.productionTotal.PM10 - theProduction.productionEngine.PM10 - theProduction.finalFlightTests.PM10
                                            + theEndOfLife.endOfLifeTotal.PM10 - theEndOfLife.endOfLifeMaterialEngine.PM10 - theEndOfLife.endOfLifeTransport.PM10)
                                        * data->configuration.structural_test_aircraft_number;
        productionTestAircraft.N2O = (theProduction.productionTotal.N2O - theProduction.finalFlightTests.N2O + theEndOfLife.endOfLifeTotal.N2O)
                                        * data->configuration.flight_tests_aircraft_number
                                        + (theProduction.productionTotal.N2O - theProduction.productionEngine.N2O - theProduction.finalFlightTests.N2O
                                            + theEndOfLife.endOfLifeTotal.N2O - theEndOfLife.endOfLifeMaterialEngine.N2O - theEndOfLife.endOfLifeTransport.N2O)
                                        * data->configuration.structural_test_aircraft_number;
        productionTestAircraft.PFC = (theProduction.productionTotal.PFC - theProduction.finalFlightTests.PFC + theEndOfLife.endOfLifeTotal.PFC)
                                        * data->configuration.flight_tests_aircraft_number
                                        + (theProduction.productionTotal.PFC - theProduction.productionEngine.PFC - theProduction.finalFlightTests.PFC
                                             + theEndOfLife.endOfLifeTotal.PFC - theEndOfLife.endOfLifeMaterialEngine.PFC - theEndOfLife.endOfLifeTransport.PFC)
                                        * data->configuration.structural_test_aircraft_number;
        productionTestAircraft.NMVOC = (theProduction.productionTotal.NMVOC - theProduction.finalFlightTests.NMVOC + theEndOfLife.endOfLifeTotal.NMVOC)
                                        * data->configuration.flight_tests_aircraft_number
                                        + (theProduction.productionTotal.NMVOC - theProduction.productionEngine.NMVOC - theProduction.finalFlightTests.NMVOC
                                            + theEndOfLife.endOfLifeTotal.NMVOC - theEndOfLife.endOfLifeMaterialEngine.NMVOC - theEndOfLife.endOfLifeTransport.NMVOC)
                                        * data->configuration.structural_test_aircraft_number;
        productionTestAircraft.HC = (theProduction.productionTotal.HC - theProduction.finalFlightTests.HC + theEndOfLife.endOfLifeTotal.HC)
                                        * data->configuration.flight_tests_aircraft_number
                                        + (theProduction.productionTotal.HC - theProduction.productionEngine.HC - theProduction.finalFlightTests.HC
                                            + theEndOfLife.endOfLifeTotal.HC - theEndOfLife.endOfLifeMaterialEngine.HC - theEndOfLife.endOfLifeTransport.HC)
                                        * data->configuration.structural_test_aircraft_number;
        productionTestAircraft.CED = (theProduction.productionTotal.CED - theProduction.finalFlightTests.CED + theEndOfLife.endOfLifeTotal.CED)
                                        * data->configuration.flight_tests_aircraft_number
                                        + (theProduction.productionTotal.CED - theProduction.productionEngine.CED - theProduction.finalFlightTests.CED
                                            + theEndOfLife.endOfLifeTotal.CED - theEndOfLife.endOfLifeMaterialEngine.CED - theEndOfLife.endOfLifeTransport.CED)
                                        * data->configuration.structural_test_aircraft_number;

        /* Production and EoL (EoL Scen only, without Disassembly+Dismant.) of the Test Engines */
        if (data->configuration.do_engine_tests) {
            test_engine_production_energy = theProduction.energy_engine * data->configuration.number_test_engines;
            testResources.energy_test_total += test_engine_production_energy;
            energy_total += test_engine_production_energy;
            myRuntimeInfo->info << "      Test Engine Production Energy: " << test_engine_production_energy << " MJ" << std::endl;
            productionTestEngine.CO2 = (theProduction.productionEngine.CO2 + theEndOfLife.endOfLifeMaterialEngine.CO2) * data->configuration.number_test_engines;
            productionTestEngine.H2O = (theProduction.productionEngine.H2O + theEndOfLife.endOfLifeMaterialEngine.H2O) * data->configuration.number_test_engines;
            productionTestEngine.NOx = (theProduction.productionEngine.NOx + theEndOfLife.endOfLifeMaterialEngine.NOx) * data->configuration.number_test_engines;
            productionTestEngine.SO2 = (theProduction.productionEngine.SO2 + theEndOfLife.endOfLifeMaterialEngine.SO2) * data->configuration.number_test_engines;
            productionTestEngine.CO = (theProduction.productionEngine.CO + theEndOfLife.endOfLifeMaterialEngine.CO) * data->configuration.number_test_engines;
            productionTestEngine.CH4 = (theProduction.productionEngine.CH4 + theEndOfLife.endOfLifeMaterialEngine.CH4) * data->configuration.number_test_engines;
            productionTestEngine.PM10 = (theProduction.productionEngine.PM10 + theEndOfLife.endOfLifeMaterialEngine.PM10) * data->configuration.number_test_engines;
            productionTestEngine.N2O = (theProduction.productionEngine.N2O + theEndOfLife.endOfLifeMaterialEngine.N2O) * data->configuration.number_test_engines;
            productionTestEngine.PFC = (theProduction.productionEngine.PFC + theEndOfLife.endOfLifeMaterialEngine.PFC) * data->configuration.number_test_engines;
            productionTestAircraft.NMVOC = (theProduction.productionEngine.NMVOC + theEndOfLife.endOfLifeMaterialEngine.NMVOC) * data->configuration.number_test_engines;
            productionTestAircraft.HC = (theProduction.productionEngine.HC + theEndOfLife.endOfLifeMaterialEngine.HC) * data->configuration.number_test_engines;
            productionTestEngine.CED = (theProduction.productionEngine.CED + theEndOfLife.endOfLifeMaterialEngine.CED) * data->configuration.number_test_engines;
        }
    }

    /* Addition all Test-Emissions */
    myRuntimeInfo->info << "    Total Test Energy: " << testResources.energy_test_total << " MJ" << std::endl;
    myRuntimeInfo->info << "    Total Test Fuel: " << testResources.fuel_test_total << " kg" << std::endl;
    testTotal.CO2  = windTunnelTests.CO2 + structureTests.CO2 + systemTests.CO2 + groundTests.CO2
                    + engineTests.CO2 + flightTests.CO2 + productionTestAircraft.CO2;
    testTotal.H2O  = windTunnelTests.H2O + structureTests.H2O + systemTests.H2O + groundTests.H2O
                    + engineTests.H2O + flightTests.H2O + productionTestAircraft.H2O;
    testTotal.NOx  = windTunnelTests.NOx + structureTests.NOx + systemTests.NOx + groundTests.NOx
                    + engineTests.NOx + flightTests.NOx + productionTestAircraft.NOx;
    testTotal.SO2  = windTunnelTests.SO2 + structureTests.SO2 + systemTests.SO2 + groundTests.SO2
                    + engineTests.SO2 + flightTests.SO2 + productionTestAircraft.SO2;
    testTotal.CO   = windTunnelTests.CO  + structureTests.CO + systemTests.CO + groundTests.CO
                    + engineTests.CO + flightTests.CO + productionTestAircraft.CO;
    testTotal.CH4  = windTunnelTests.CH4 + structureTests.CH4 + systemTests.CH4 + groundTests.CH4
                    + engineTests.CH4 + flightTests.CH4 + productionTestAircraft.CH4;
    testTotal.PM10 = windTunnelTests.PM10 + structureTests.PM10 + systemTests.PM10 + groundTests.PM10
                    + engineTests.PM10 + flightTests.PM10 + productionTestAircraft.PM10;
    testTotal.N2O  = windTunnelTests.N2O + structureTests.N2O + systemTests.N2O + groundTests.N2O
                    + engineTests.N2O + flightTests.N2O + productionTestAircraft.N2O;
    testTotal.PFC  = windTunnelTests.PFC + structureTests.PFC + systemTests.PFC + groundTests.PFC
                    + engineTests.PFC + flightTests.PFC + productionTestAircraft.PFC;
    testTotal.NMVOC = windTunnelTests.NMVOC + structureTests.NMVOC + systemTests.NMVOC + groundTests.NMVOC
                    + engineTests.NMVOC + flightTests.NMVOC + productionTestAircraft.NMVOC;
    testTotal.HC   = windTunnelTests.HC + structureTests.HC + systemTests.HC + groundTests.HC
                    + engineTests.HC + flightTests.HC + productionTestAircraft.HC;
    testTotal.CED   = windTunnelTests.CED + structureTests.CED + systemTests.CED + groundTests.CED
                    + engineTests.CED + flightTests.CED + productionTestAircraft.CED;

    /* Calculation of emissions from personnel expenses in the development phase*/
    //electricity
    myRuntimeInfo->info << "      Test Engineering Electric Energy: " << theEngeneering.engineering_energy_electricity << " MJ" << std::endl;
    engineeringElectricity.CO2 = theEngeneering.engineering_energy_electricity * myEcoDatabase.elecEmData.CO2;
    engineeringElectricity.H2O = theEngeneering.engineering_energy_electricity * myEcoDatabase.elecEmData.H2O;
    engineeringElectricity.NOx = theEngeneering.engineering_energy_electricity * myEcoDatabase.elecEmData.NOx;
    engineeringElectricity.SO2 = theEngeneering.engineering_energy_electricity * myEcoDatabase.elecEmData.SO2;
    engineeringElectricity.CO = theEngeneering.engineering_energy_electricity * myEcoDatabase.elecEmData.CO;
    engineeringElectricity.CH4 = theEngeneering.engineering_energy_electricity * myEcoDatabase.elecEmData.CH4;
    engineeringElectricity.PM10 = theEngeneering.engineering_energy_electricity * myEcoDatabase.elecEmData.PM10;
    engineeringElectricity.CED = theEngeneering.engineering_energy_electricity * myEcoDatabase.elecEmData.specificEnergy;
    //heating
    myRuntimeInfo->info << "      Test Engineering Heat Energy: " << theEngeneering.engineering_energy_heat << " MJ" << std::endl;
    engineeringHeat.CO2 = theEngeneering.engineering_energy_heat * myEcoDatabase.heatEmData.CO2;
    engineeringHeat.H2O = theEngeneering.engineering_energy_heat * myEcoDatabase.heatEmData.H2O;
    engineeringHeat.NOx = theEngeneering.engineering_energy_heat * myEcoDatabase.heatEmData.NOx;
    engineeringHeat.SO2 = theEngeneering.engineering_energy_heat * myEcoDatabase.heatEmData.SO2;
    engineeringHeat.CO = theEngeneering.engineering_energy_heat * myEcoDatabase.heatEmData.CO;
    engineeringHeat.CH4 = theEngeneering.engineering_energy_heat * myEcoDatabase.heatEmData.CH4;
    engineeringHeat.PM10 = theEngeneering.engineering_energy_heat * myEcoDatabase.heatEmData.PM10;
    engineeringHeat.CED = theEngeneering.engineering_energy_heat * myEcoDatabase.heatEmData.specificEnergy;
    //Total personnel expenses
    myRuntimeInfo->info << "    Test Engineering Energy Total: " << theEngeneering.engineering_energy_total << " MJ" << std::endl;
    engineeringTotal.CO2 = engineeringElectricity.CO2 + engineeringHeat.CO2;
    engineeringTotal.H2O = engineeringElectricity.H2O + engineeringHeat.H2O;
    engineeringTotal.NOx = engineeringElectricity.NOx + engineeringHeat.NOx;
    engineeringTotal.SO2 = engineeringElectricity.SO2 + engineeringHeat.SO2;
    engineeringTotal.CO = engineeringElectricity.CO + engineeringHeat.CO;
    engineeringTotal.CH4 = engineeringElectricity.CH4 + engineeringHeat.CH4;
    engineeringTotal.PM10 = engineeringElectricity.PM10 + engineeringHeat.PM10;
    engineeringTotal.N2O = engineeringElectricity.N2O + engineeringHeat.N2O;
    engineeringTotal.PFC = engineeringElectricity.PFC + engineeringHeat.PFC;
    engineeringTotal.NMVOC = engineeringElectricity.NMVOC + engineeringHeat.NMVOC;
    engineeringTotal.HC = engineeringElectricity.HC + engineeringHeat.HC;
    engineeringTotal.CED = engineeringElectricity.CED + engineeringHeat.CED;

    /*Entire development phase*/
    myRuntimeInfo->info << "  Development Power Total: " << energy_total << " MJ" << std::endl;
    myRuntimeInfo->info << "  Development Fuel Total: " << fuel_total << " kg" << std::endl;
    developmentTotal.CO2 = engineeringTotal.CO2 + testTotal.CO2;
    developmentTotal.H2O = engineeringTotal.H2O + testTotal.H2O;
    developmentTotal.NOx = engineeringTotal.NOx + testTotal.NOx;
    developmentTotal.SO2 = engineeringTotal.SO2 + testTotal.SO2;
    developmentTotal.CO  = engineeringTotal.CO + testTotal.CO;
    developmentTotal.CH4 = engineeringTotal.CH4 + testTotal.CH4;
    developmentTotal.PM10 = engineeringTotal.PM10 + testTotal.PM10;
    developmentTotal.N2O = engineeringTotal.N2O + testTotal.N2O;
    developmentTotal.PFC = engineeringTotal.PFC + testTotal.PFC;
    developmentTotal.NMVOC = engineeringTotal.NMVOC + testTotal.NMVOC;
    developmentTotal.HC = engineeringTotal.HC + testTotal.HC;
    developmentTotal.CED = engineeringTotal.CED + testTotal.CED;
    myRuntimeInfo->info << "  Development Primary Power Total: " << developmentTotal.CED << " MJ" << std::endl;
}


Development::~Development() {
    //dtor
}

// TODO(franz#1#): Mantis ID: 0000325 Class structure even clearer?! by class Energy, class Fuel, and class:  Calculate taxi fuel already here

Development::test::test()
    :
    energy_wind_tunnel_test(0.),
    energy_fatigue_test(0.),
    energy_system_tests(0.),
    fuel_engine_test_rig(0.),
    fuel_engine_test_flight(0.),
    fuel_engine_test_total(0.),
    energy_ground_test(0.),
    fuel_flight_test(0.),
    flight_hours_flight_test(0.),
    energy_test_total(0.),
    fuel_test_total(0.) {
}
Development::test::~test() {
    //dtor
}

Development::Engineering::Engineering()
    :
    engineering_hours(0.),
    engineering_energy_electricity(0.),
    engineering_energy_heat(0.),
    engineering_energy_total(0.) {
}
Development::Engineering::~Engineering() {
}
