/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_EMISSION_CALCULATION_LIFE_CYCLE_EMISSIONS_LCA_SCHAEFER_DEVELOPMENT_H_
#define ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_EMISSION_CALCULATION_LIFE_CYCLE_EMISSIONS_LCA_SCHAEFER_DEVELOPMENT_H_

#include "../../../emissionsClass.h"
#include "../../ecoDatabase.h"
#include "production.h"
#include "endOfLife.h"
#include "schaeferIOData.h"
#include <memory>

class SchaeferIOData;
class Production;
class EndOfLife;

/** \class Development
 * \brief Contains all properties of the development phase
 */
class Development{
 protected:
    const ecoDatabase myEcoDatabase; /**< Object contains data of ecoDatabase*/
    atmosphere myAtmosphere;/**< Atmosphere object */

 public:
    /** \brief Constructor of class development
     */
    Development();
    /** \brief Destructor of class development
     */
    virtual ~Development();

    /** \class test
     * \brief Class contains the properties of the expenses for testing
     */
    class test {
     public:
        /* Necessary resources for the entire certification process (i.e. for the entire aircraft program) */
        double energy_wind_tunnel_test;/**< Required energy for wind tunnel testing [MJ] */
        double energy_fatigue_test;/**< Required energy for fatigue tests [MJ] */
        double energy_system_tests;/**< Required energy for system tests [MJ] */
        double fuel_engine_test_rig;/**< Required fuel for engine test rig [kg] */
        double fuel_engine_test_flight;/**< Required fuel for engine flight tests [kg]*/
        double fuel_engine_test_total;/**< Total required fuel for engine testing [kg]*/
        double energy_ground_test;/**< Required energy for ground tests [MJ]*/
        double fuel_flight_test;/**< Total required fuel for flight tests [kg]*/
        double flight_hours_flight_test;/**< Flight hours during flight test phase [h]*/
        double energy_test_total;/**< Total required energy for test phase [MJ]*/
        double fuel_test_total;/**< Total required fuel for test phase [kg]*/

        /** \brief Constructor of class test
         */
        test();

        /** \brief Destructor of class test
         */
        virtual ~test();
    };
    test testResources; /**< Object of class test contains data of test resources*/

    /** \class Engineering
     * \brief Contains the properties of personnel expenses for engineering
     */
    class Engineering {
     public:
        double engineering_hours;/**< Airframe Engineering Hours [h] */
        double engineering_energy_electricity;/**< Test Engineering Electric Energy [MJ] */
        double engineering_energy_heat;/**< Test Engineering Heat Energy [MJ] */
        double engineering_energy_total;/** Test Engineering total Energy [MJ] */

        /** \brief Constructor of class Engineering
         */
        Engineering();

        /** \brief Destructor of class Engineering
         */
        virtual ~Engineering();
    };
    Engineering theEngeneering; /**< Object of class engineering contains personnel expenses for engineering */

    double energy_total;/**< Total energy demand for the development phase [MJ] */
    double fuel_total;/**< total fuel demand for the development phase [kg] */

    Emissions engineeringElectricity;/**< Emissions due to electricity generated by development personnel */
    Emissions engineeringHeat;/**< Emissions due to heating of development personnel */
    Emissions engineeringTotal;/**< Total emissions from development personnel */
    Emissions windTunnelTests;/**< Emissions due to wind tunnel tests */
    Emissions structureTests;/**< Emissions due to structure tests */
    Emissions systemTests;/**< Emissions due to system tests */
    Emissions engineTests;/**< Emissions due to engine tests */
    Emissions groundTests;/**< Emissions due to ground test incl. fuel production */
    double taxi_fuel_test_factor;/**< Factor to calculate taxi fuel for test mission [-] */
    Emissions flightTests;/**< Emissions due to flight test incl. fuel production */
    Emissions testTotal;/**< All test emissions (excluding fuel and material production) */
    Emissions productionTestAircraft;/**< Emissions of production of the test aircraft */
    Emissions productionTestEngine;/**< Emissions of production of the test aircraft engines */
    double test_aircraft_production_energy; /**< Needed energy for production of test aircraft [MJ]*/
    double test_engine_production_energy;/**< Needed energy for production of test aircraft engine [MJ]*/
    /** \todo [T137] (https://unicado.ilr.rwth-aachen.de/T137): Put resources into respective class if possible
     */
    Emissions developmentTotal;/**< Total emissions during the development phase */

    /** \brief Function to calculate the material and energy demand of the development phase
     * \param data IO Data pointer
     */
    void calculate_development_resources(const std::shared_ptr<SchaeferIOData>& data);

    /** \brief Function to calculate the emissions of the development phase
     * \param data IO Data pointer
     * \param theProduction production object containing the production data
     * \param theEndOfLife production object containing the end of life data
     */
    void calculate_development_emissions(const std::shared_ptr<SchaeferIOData>& data, const Production&theProduction, const EndOfLife &theEndOfLife);
};


#endif // ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_EMISSION_CALCULATION_LIFE_CYCLE_EMISSIONS_LCA_SCHAEFER_DEVELOPMENT_H_
