/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "endOfLife.h"
#include <aixml/node.h>
#include <runtimeInfo/runtimeInfo.h>
#include <standardFiles/functions.h>

EndOfLife::EndOfLife()
    :
    myEcoDatabase(ecoDatabase()),
    cost_DD(0.),
    engineering_energy_electricity_DD(0.),
    energy_EoL_total(0.),
    energy_total(0.),
    transport_fuel(0.) {
    }

void EndOfLife::calculate_EoL_resources(const std::shared_ptr<SchaeferIOData>& data) {
    /*Energy for Disassembly and Dismantling*/
    cost_DD = (218800 * data->ome) / (data->ome + 31600);
    engineering_energy_electricity_DD = energy_PerWorker_pr[0] * cost_DD / (C_labourHour_pr * t_labourYear_pr);
    for (int i(0); i < data->configuration.parts_number; i++) {
        theAcComp.push_back(mainComponent(data->main_parts_name[i]));
        theAcComp.back().mass = data->main_parts_mass[i];
        myRuntimeInfo->debug << std::endl;
        myRuntimeInfo->debug << std::endl;
        myRuntimeInfo->debug << "Component: " << theAcComp.back().component_name << std::endl;
        /* Energy calculation: distribution according to aircraft components (Main Parts Modus) */
        //Material distribution. For numeration see production.h: const double matDistr
        double endOfLifeEnergy = 0.;
        for (int j(0); j < numberOfMaterials; j++) {
            theAcComp.back().rate_material.push_back(matDistr[i][j]);
            theAcComp.back().rate_recycling.push_back(recyclingScen[i][j]);
            theAcComp.back().rate_incineration.push_back(incinerationScen[i][j]);
            theAcComp.back().rate_landfill.push_back(landfillScen[i][j]);
            double massMaterialComp = theAcComp.back().rate_material.back() * theAcComp.back().mass;
            theAcComp.back().recovered_material_mass.push_back(massMaterialComp * theAcComp.back().rate_recycling.back()*recoveryRate[j]);
            if (j != 1)
                theAcComp.back().energy_recycling = 1 / recyclingEnergyRatio[j] * myEcoDatabase.matSpecificEnergy.at(j)
                                               / myEcoDatabase.elecEmData.specificEnergy * massMaterialComp * theAcComp.back().rate_recycling.back();
            else
                theAcComp.back().energy_recycling = 0.; //Composites will not be recycled, recyclingEnergyRatio[j] = 0
            theAcComp.back().energy_incineration = myEcoDatabase.incinerationSpecificEnergy.at(j)
                                              / myEcoDatabase.elecEmData.specificEnergy * massMaterialComp * theAcComp.back().rate_incineration.back();
            theAcComp.back().energy_landfill = myEcoDatabase.disposalSpecificEnergy.at(j)
                                          / myEcoDatabase.elecEmData.specificEnergy * massMaterialComp * theAcComp.back().rate_landfill.back();
            endOfLifeEnergy += theAcComp.back().energy_recycling + theAcComp.back().energy_incineration + theAcComp.back().energy_landfill;
            myRuntimeInfo->debug << std::endl;
            myRuntimeInfo->debug << "Material " << j + 1 << "..." << std::endl;
            myRuntimeInfo->debug << "ER: " << recyclingEnergyRatio[j] << std::endl;
            myRuntimeInfo->debug << "Mass fraction: " << theAcComp.back().rate_material.back() << std::endl;
            myRuntimeInfo->debug << "Recycling fraction: " << theAcComp.back().rate_recycling.back() << std::endl;
            myRuntimeInfo->debug << "Incinaration fraction: " << theAcComp.back().rate_incineration.back() << std::endl;
            myRuntimeInfo->debug << "Landfill fraction: " << theAcComp.back().rate_landfill.back() << std::endl;
            myRuntimeInfo->debug << "Mass material: " << massMaterialComp << std::endl;
            myRuntimeInfo->debug << "energyRecycling: " << theAcComp.back().energy_recycling << std::endl;
            myRuntimeInfo->debug << "energyIncinenartion: " << theAcComp.back().energy_incineration << std::endl;
            myRuntimeInfo->debug << "energyLandfill: " << theAcComp.back().energy_landfill << std::endl;
            myRuntimeInfo->debug << "endOfLifeEnergy: " << endOfLifeEnergy << std::endl;
        }
        theAcComp.back().energy_endOfLife = endOfLifeEnergy;
        energy_EoL_total += theAcComp.back().energy_endOfLife;
        myRuntimeInfo->debug << "Mass:   " << theAcComp.back().mass << std::endl;
        myRuntimeInfo->debug << "Energy: " << theAcComp.back().energy_endOfLife << std::endl;
    }
    myRuntimeInfo->debug << "E_EoL_total: " << energy_EoL_total << std::endl;
//    getchar();
    /* Energy Calculation: Distribution by Materials (Material Mode) */
    //TO DO: delete?
    //E_EoL_total = 0.;
    for (int i(0); i < numberOfMaterials; i++) {
        theMaterial.push_back(EndOfLife::material());
        theMaterial.back().material_name = data->material_name[i];
    }
    myRuntimeInfo->debug << "E_EoL_total: " << energy_EoL_total << std::endl;
//    getchar();
    energy_total = energy_EoL_total + engineering_energy_electricity_DD;
}

void EndOfLife::calculate_EoL_emissions(const std::shared_ptr<SchaeferIOData>& data, const Production&theProduction) {
    /*Transport to End-of-Life Site */
    /* emission calculation */
    double transportFactor = data->configuration.distance_to_EoL_site / data->range_design;
    transport_fuel = data->blockfuel_design * transportFactor; //strongly simplified: Assumption: Const. Fuel consumption, determined from design mission.
    myRuntimeInfo->info << "  EoL Fuel: " << transport_fuel << " kg" << std::endl;
    endOfLifeTransport.CO2 = data->theTestMission.LTO.CO2 + data->theTestMission.cruise.CO2 * transportFactor + transport_fuel * myEcoDatabase.fuelProdEmData.at(0).CO2;
    endOfLifeTransport.H2O = data->theTestMission.LTO.H2O + data->theTestMission.cruise.H2O * transportFactor + transport_fuel * myEcoDatabase.fuelProdEmData.at(0).H2O;
    endOfLifeTransport.NOx = data->theTestMission.LTO.NOx + data->theTestMission.cruise.NOx * transportFactor + transport_fuel * myEcoDatabase.fuelProdEmData.at(0).NOx;
    endOfLifeTransport.SO2 = data->theTestMission.LTO.SO2 + data->theTestMission.cruise.SO2 * transportFactor + transport_fuel * myEcoDatabase.fuelProdEmData.at(0).SO2;
    endOfLifeTransport.CO = data->theTestMission.LTO.CO + data->theTestMission.cruise.CO * transportFactor + transport_fuel * myEcoDatabase.fuelProdEmData.at(0).CO;
    endOfLifeTransport.CH4 = transport_fuel * myEcoDatabase.fuelProdEmData.at(0).CH4;
    endOfLifeTransport.PM10 = data->theTestMission.LTO.soot + data->theTestMission.cruise.soot * transportFactor + transport_fuel * myEcoDatabase.fuelProdEmData.at(0).PM10;
    endOfLifeTransport.HC = data->theTestMission.LTO.HC + data->theTestMission.cruise.HC * transportFactor;
    endOfLifeTransport.CED = transport_fuel * myEcoDatabase.fuelSpecificEnergy.at(0);
//    // TODO (student#1#): Noch weitere Emissionen wichtig?
    /*Disassembly and Dismantling */
    //Elektrische Emissionen
    myRuntimeInfo->info << "    Disassembly and Dismantling Energy: " << engineering_energy_electricity_DD << " MJ" << std::endl;
    endOfLifeDD.CO2 = engineering_energy_electricity_DD * myEcoDatabase.elecEmData.CO2;
    endOfLifeDD.H2O = engineering_energy_electricity_DD * myEcoDatabase.elecEmData.H2O;
    endOfLifeDD.NOx = engineering_energy_electricity_DD * myEcoDatabase.elecEmData.NOx;
    endOfLifeDD.SO2 = engineering_energy_electricity_DD * myEcoDatabase.elecEmData.SO2;
    endOfLifeDD.CO =  engineering_energy_electricity_DD * myEcoDatabase.elecEmData.CO;
    endOfLifeDD.CH4 = engineering_energy_electricity_DD * myEcoDatabase.elecEmData.CH4;
    endOfLifeDD.PM10 = engineering_energy_electricity_DD * myEcoDatabase.elecEmData.PM10;
    endOfLifeDD.CED = engineering_energy_electricity_DD * myEcoDatabase.elecEmData.specificEnergy;
    /*End-of-Life Scenario (Energy consumption through processes and energy gain through material extraction*/

    if (data->configuration.production_mode == "mode_0") {
        //Material modus
        this->emissions_material_mode();
    } else if (data->configuration.production_mode == "mode_1") {
        //Main Parts Modus
        this->emissions_main_parts_mode(data, theProduction);
    } else {
        myRuntimeInfo->err << "Unknown production mode " << data->configuration.production_mode << ". (config-entry \"production_mode\") " << std::endl;
        myRuntimeInfo->err << "Abort program!" << std::endl;
        exit(1);
    }
}

void EndOfLife::emissions_material_mode() {
//    double material_energy=0.;
    for (int i(0); i < numberOfMaterials; i++) {
        std::string materialEntry(theMaterial.at(i).material_name);
        std::vector<std::string> material_name_tmp;
        for (unsigned int k(0.); k < myEcoDatabase.matEmData.size(); k++) {
            material_name_tmp.push_back(myEcoDatabase.matEmData.at(k).specification);
        }
        std::vector<std::string>::iterator entry = find(material_name_tmp.begin(), material_name_tmp.end(), materialEntry); //Iterator to find the configuration entry
        if (entry == material_name_tmp.end()) { //Entry not available!
            myRuntimeInfo->err << "Material " << materialEntry << " not found. Abort!" << std::endl;
            exit(1);
        } else {
            unsigned int pos(0);
            pos = static_cast<int>(entry - material_name_tmp.begin());
            /* emission calculation */
            endOfLifeMaterial.push_back(Emissions());
            endOfLifeMaterial.back().CO2 = - theMaterial.at(i).mass_recovered * myEcoDatabase.matEmData.at(pos).CO2 //Row material
                                     + theMaterial.at(i).energy_endOfLife * myEcoDatabase.elecEmData.CO2; //manufacturing
            endOfLifeMaterial.back().H2O = - theMaterial.at(i).mass_recovered * myEcoDatabase.matEmData.at(pos).H2O
                                     + theMaterial.at(i).energy_endOfLife * myEcoDatabase.elecEmData.H2O;
            endOfLifeMaterial.back().NOx = - theMaterial.at(i).mass_recovered * myEcoDatabase.matEmData.at(pos).NOx
                                     + theMaterial.at(i).energy_endOfLife * myEcoDatabase.elecEmData.NOx;
            endOfLifeMaterial.back().SO2 = - theMaterial.at(i).mass_recovered * myEcoDatabase.matEmData.at(pos).SO2
                                     + theMaterial.at(i).energy_endOfLife * myEcoDatabase.elecEmData.SO2;
            endOfLifeMaterial.back().CO = - theMaterial.at(i).mass_recovered * myEcoDatabase.matEmData.at(pos).CO
                                    + theMaterial.at(i).energy_endOfLife * myEcoDatabase.elecEmData.CO;
            endOfLifeMaterial.back().CH4 = - theMaterial.at(i).mass_recovered * myEcoDatabase.matEmData.at(pos).CH4
                                     + theMaterial.at(i).energy_endOfLife * myEcoDatabase.elecEmData.CH4;
            endOfLifeMaterial.back().PM10 = - theMaterial.at(i).mass_recovered * myEcoDatabase.matEmData.at(pos).PM10
                                      + theMaterial.at(i).energy_endOfLife * myEcoDatabase.elecEmData.PM10;
            endOfLifeMaterial.back().CED = - theMaterial.at(i).mass_recovered * myEcoDatabase.matSpecificEnergy.at(pos)
                                     + theMaterial.at(i).energy_endOfLife * myEcoDatabase.elecEmData.specificEnergy;;
        }
        endOfLifeMaterialTotal.CO2  += endOfLifeMaterial.back().CO2;
        endOfLifeMaterialTotal.H2O  += endOfLifeMaterial.back().H2O;
        endOfLifeMaterialTotal.NOx  += endOfLifeMaterial.back().NOx;
        endOfLifeMaterialTotal.SO2  += endOfLifeMaterial.back().SO2;
        endOfLifeMaterialTotal.CO   += endOfLifeMaterial.back().CO;
        endOfLifeMaterialTotal.CH4  += endOfLifeMaterial.back().CH4;
        endOfLifeMaterialTotal.PM10 += endOfLifeMaterial.back().PM10;
        endOfLifeMaterialTotal.N2O += endOfLifeMaterial.back().N2O;
        endOfLifeMaterialTotal.PFC += endOfLifeMaterial.back().PFC;
        endOfLifeMaterialTotal.NMVOC += endOfLifeMaterial.back().NMVOC;
        endOfLifeMaterialTotal.HC += endOfLifeMaterial.back().HC;
        endOfLifeMaterialTotal.CED += endOfLifeMaterial.back().CED;
    }
    myRuntimeInfo->info << "    EoL Scenario Primary Energy Total: " << endOfLifeMaterialTotal.CED << " MJ" << std::endl;
}

void EndOfLife::emissions_main_parts_mode(const std::shared_ptr<SchaeferIOData>& data, const Production&theProduction) {
    for (int i(0); i < data->configuration.parts_number; i++) {
        /* emission calculation */
        endOfLifeMaterial.push_back(Emissions());
        endOfLifeMaterial.back().CO2 = theAcComp.at(i).energy_endOfLife * myEcoDatabase.elecEmData.CO2;
        endOfLifeMaterial.back().H2O = theAcComp.at(i).energy_endOfLife * myEcoDatabase.elecEmData.H2O;
        endOfLifeMaterial.back().NOx = theAcComp.at(i).energy_endOfLife * myEcoDatabase.elecEmData.NOx;
        endOfLifeMaterial.back().SO2 = theAcComp.at(i).energy_endOfLife * myEcoDatabase.elecEmData.SO2;
        endOfLifeMaterial.back().CO = theAcComp.at(i).energy_endOfLife * myEcoDatabase.elecEmData.CO;
        endOfLifeMaterial.back().CH4 = theAcComp.at(i).energy_endOfLife * myEcoDatabase.elecEmData.CH4;
        endOfLifeMaterial.back().PM10 = theAcComp.at(i).energy_endOfLife * myEcoDatabase.elecEmData.PM10;
        endOfLifeMaterial.back().CED = theAcComp.at(i).energy_endOfLife * myEcoDatabase.elecEmData.specificEnergy;
        for (int j(0); j < numberOfMaterials; j++) {
            std::string materialEntry(theProduction.theMaterial.at(j).material_name);
            std::vector<std::string> material_name_tmp;
            for (unsigned int k(0.); k < myEcoDatabase.matEmData.size(); k++) {
                material_name_tmp.push_back(myEcoDatabase.matEmData.at(k).specification);
            }
            std::vector<std::string>::iterator entry = find(material_name_tmp.begin(), material_name_tmp.end(), materialEntry); //Iterator to find the configuration entry
            if (entry == material_name_tmp.end()) { //Entry not available!
                myRuntimeInfo->err << "Material " << materialEntry << " in EcoDatabase not found. Abort!" << std::endl;
                exit(1);
            } else {
                unsigned int pos(0);
                pos = static_cast<int>(entry - material_name_tmp.begin());
                /* emission calculation */
                endOfLifeMaterial.back().CO2 -= theAcComp.back().recovered_material_mass.at(j) * myEcoDatabase.matEmData.at(pos).CO2;
                endOfLifeMaterial.back().H2O -= theAcComp.back().recovered_material_mass.at(j) * myEcoDatabase.matEmData.at(pos).H2O;
                endOfLifeMaterial.back().NOx -= theAcComp.back().recovered_material_mass.at(j) * myEcoDatabase.matEmData.at(pos).NOx;
                endOfLifeMaterial.back().SO2 -= theAcComp.back().recovered_material_mass.at(j) * myEcoDatabase.matEmData.at(pos).SO2;
                endOfLifeMaterial.back().CO -= theAcComp.back().recovered_material_mass.at(j) * myEcoDatabase.matEmData.at(pos).CO;
                endOfLifeMaterial.back().CH4 -= theAcComp.back().recovered_material_mass.at(j) * myEcoDatabase.matEmData.at(pos).CH4;
                endOfLifeMaterial.back().PM10 -= theAcComp.back().recovered_material_mass.at(j) * myEcoDatabase.matEmData.at(pos).PM10;
                endOfLifeMaterial.back().CED -= theAcComp.back().recovered_material_mass.at(j) * myEcoDatabase.matSpecificEnergy.at(pos);
            }
        }
        endOfLifeMaterialTotal.CO2 += endOfLifeMaterial.back().CO2;
        endOfLifeMaterialTotal.H2O += endOfLifeMaterial.back().H2O;
        endOfLifeMaterialTotal.NOx += endOfLifeMaterial.back().NOx;
        endOfLifeMaterialTotal.SO2 += endOfLifeMaterial.back().SO2;
        endOfLifeMaterialTotal.CO += endOfLifeMaterial.back().CO;
        endOfLifeMaterialTotal.CH4 += endOfLifeMaterial.back().CH4;
        endOfLifeMaterialTotal.PM10 += endOfLifeMaterial.back().PM10;
        endOfLifeMaterialTotal.N2O += endOfLifeMaterial.back().N2O;
        endOfLifeMaterialTotal.PFC += endOfLifeMaterial.back().PFC;
        endOfLifeMaterialTotal.NMVOC += endOfLifeMaterial.back().NMVOC;
        endOfLifeMaterialTotal.HC += endOfLifeMaterial.back().HC;
        endOfLifeMaterialTotal.CED += endOfLifeMaterial.back().CED;
    }
    if (data->configuration.engine_mode) {
        endOfLifeMaterialEngine.CO2 = endOfLifeMaterial.back().CO2;
        endOfLifeMaterialEngine.H2O = endOfLifeMaterial.back().H2O;
        endOfLifeMaterialEngine.NOx = endOfLifeMaterial.back().NOx;
        endOfLifeMaterialEngine.SO2 = endOfLifeMaterial.back().SO2;
        endOfLifeMaterialEngine.CO = endOfLifeMaterial.back().CO;
        endOfLifeMaterialEngine.CH4 = endOfLifeMaterial.back().CH4;
        endOfLifeMaterialEngine.PM10 = endOfLifeMaterial.back().PM10;
        endOfLifeMaterialEngine.N2O = endOfLifeMaterial.back().N2O;
        endOfLifeMaterialEngine.PFC = endOfLifeMaterial.back().PFC;
        endOfLifeMaterialEngine.NMVOC = endOfLifeMaterial.back().NMVOC;
        endOfLifeMaterialEngine.HC = endOfLifeMaterial.back().HC;
        endOfLifeMaterialEngine.CED = endOfLifeMaterial.back().CED;
    }
    myRuntimeInfo->info << "    EoL Scenario Primary Energy Total: " << endOfLifeMaterialTotal.CED << " MJ" << std::endl;

    /*EoL Phase Total*/
    myRuntimeInfo->info << "  EoL Energy Total: " << energy_total << " MJ" << std::endl;
    endOfLifeTotal.CO2 = endOfLifeTransport.CO2 + endOfLifeDD.CO2 + endOfLifeMaterialTotal.CO2;
    endOfLifeTotal.H2O = endOfLifeTransport.H2O + endOfLifeDD.H2O + endOfLifeMaterialTotal.H2O;
    endOfLifeTotal.NOx = endOfLifeTransport.NOx + endOfLifeDD.NOx + endOfLifeMaterialTotal.NOx;
    endOfLifeTotal.SO2 = endOfLifeTransport.SO2 + endOfLifeDD.SO2 + endOfLifeMaterialTotal.SO2;
    endOfLifeTotal.CO = endOfLifeTransport.CO + endOfLifeDD.CO + endOfLifeMaterialTotal.CO;
    endOfLifeTotal.CH4 = endOfLifeTransport.CH4 + endOfLifeDD.CH4 + endOfLifeMaterialTotal.CH4;
    endOfLifeTotal.PM10 = endOfLifeTransport.PM10 + endOfLifeDD.PM10 + endOfLifeMaterialTotal.PM10;
    endOfLifeTotal.N2O = endOfLifeTransport.N2O + endOfLifeDD.N2O + endOfLifeMaterialTotal.N2O;
    endOfLifeTotal.PFC = endOfLifeTransport.PFC + endOfLifeDD.PFC + endOfLifeMaterialTotal.PFC;
    endOfLifeTotal.NMVOC = endOfLifeTransport.NMVOC + endOfLifeDD.NMVOC + endOfLifeMaterialTotal.NMVOC;
    endOfLifeTotal.HC = endOfLifeTransport.HC + endOfLifeDD.HC + endOfLifeMaterialTotal.HC;
    endOfLifeTotal.CED = endOfLifeTransport.CED + endOfLifeDD.CED + endOfLifeMaterialTotal.CED;
    myRuntimeInfo->info << "  EoL Primary Energy Total: " << endOfLifeTotal.CED << " MJ" << std::endl;
}

EndOfLife::~EndOfLife() {
}

EndOfLife::mainComponent::mainComponent(const std::string& name)
    :
    component_name(name),
    mass(0.),
    energy_recycling(0.),
    energy_incineration(0.),
    energy_landfill(0.),
    energy_endOfLife(0.) {
}
EndOfLife::mainComponent::~mainComponent() {
}

EndOfLife::material::material()
    :
    material_name(""),
    mass(0.),
    mass_recovered(0.),
    energy_recycling(0.),
    energy_incineration(0.),
    energy_landfill(0.),
    energy_endOfLife(0.) {
}
EndOfLife::material::~material() {
}
