/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include <algorithm>

#include "operation.h"
#include "production.h"

#include <aixml/node.h>
#include <standardFiles/functions.h>
#include <runtimeInfo/runtimeInfo.h>
#include <moduleBasics/runtimeIO.h>

using std::vector;
using std::string;

// TODO(franz#1#): T129 Emission calculation at Derate correct? Check! (Problem: Base curves with EI_ICAO (without Derate))
Operation::Operation()
    :
    energy_spare_parts_total(0.),
    energy_labour_total(0.),
    energy_total(0.),
    missions_fuel(0.) {
    }

void Operation::calculate_operation_resources(const std::shared_ptr<SchaeferIOData>& data, const std::shared_ptr<RuntimeIO>& rtIO,
                                                const Production &theProduction) {//Due to maintenance
    /* Maintenance Spare Parts */
    double engineCostRatioMaterial(0.75);//Assumption: Material Engine Maintenance 3 times as high as Labour (see PhD Schaefer)
    double conversionFactor = convertUnit(SECOND, HOUR, data->flight_time_study) * data->flights_per_year * data->duration_operation; // Used for [$/FH] -> [$/LC]
    for (uint8_t i(0); i < data->configuration.spare_parts_number; i++) {
        /*Creating the object for the corresponding components*/
        theSpareParts.push_back(spareParts(data->spare_parts_ATA[i], data->spare_parts_name[i]));
        /*RC components*/
        theSpareParts.back().RC_component = theProduction.theAcComp.at(i).RC;
        /*Material maintenance costs: Calculation according to Americal Airlines model*/
        double CPI_factor = 3.833; // Converts 1976-USD to 2010-USD [-]
        if (data->spare_parts_name[i] == "Wing") { //ATA 57
            double timeFactor = (convertUnit(SECOND, HOUR, data->flight_time_study) * 0.49 + 0.51) / (2.5 * 0.49 + 0.51);
            theSpareParts.back().maintenance_material_costs = CPI_factor * (0.126 + 0.00506 * data->wing_reference_area) * timeFactor * conversionFactor;
        } else if (data->spare_parts_name[i] == "Fuselage") { //Airframe
            for (uint8_t k(0); k < 4; k++) { // spare_parts_ATA_airframe: ATA51, ATA52, ATA53, ATA56
                /*Order of coefficients: ATA51, ATA52, ATA53, ATA56*/
                vector<double> c{0., 0.387, 0.5833, 0.};// Method coefficient
                vector<double> d{0., 0.00785, 0., 0.0362};// Method coefficient
                vector<double> X{data->airframe_mass/1000, data->available_seats, data->airframe_mass/1000, data->available_seats};// Method coefficient
                vector<double> Y{1., 1., 1., 1.}; // Method coefficient
                vector<double> HR{1., 0.51, 0.50, 0.80};// Flight hour related cost ratio (ATA51, ATA52, ATA53, ATA56)
                vector<double> CR{0., 0.49, 0.50, 0.20};// Flight cycle related cost ratio (ATA51, ATA52, ATA53, ATA56)
                double timeFactor = (convertUnit(SECOND, HOUR, data->flight_time_study) * HR[k] + CR[k]) / (2.5 * HR[k] + CR[k]);
                theSpareParts.back().maintenance_material_costs += CPI_factor * (c[k] + d[k]*X[k]*Y[k]) * timeFactor * conversionFactor;
            }
        } else if (data->spare_parts_name[i] == "Tailplane") {//ATA 55
            double timeFactor = (convertUnit(SECOND, HOUR, data->flight_time_study) * 0.49 + 0.51) / (2.5 *  0.49 + 0.51);
            theSpareParts.back().maintenance_material_costs = data->stabilizer_mass_ratio * CPI_factor * 0.3737 * timeFactor * conversionFactor;
        } else if (data->spare_parts_name[i] == "Fin") { //ATA 55
            double timeFactor = (convertUnit(SECOND, HOUR, data->flight_time_study) * 0.49 + 0.51) / (2.5 *  0.49 + 0.51);
            theSpareParts.back().maintenance_material_costs = data->fin_mass_ratio * CPI_factor * 0.3737 * timeFactor * conversionFactor;
        } else if (data->spare_parts_name[i] == "LandingGear") { //ATA 32
            double timeFactor = (convertUnit(SECOND, HOUR, data->flight_time_study) * 0.18 + 0.82) / (2.5 *  0.18 + 0.82);
            theSpareParts.back().maintenance_material_costs = CPI_factor * (4.961 + .1810 * data->mtom / 1000) * timeFactor *conversionFactor;
        } else if (data->spare_parts_name[i] == "Pylons") { //ATA 54
            double timeFactor = (convertUnit(SECOND, HOUR, data->flight_time_study) * 0.8 + 0.2) / (2.5 *  0.8 + 0.2);
            theSpareParts.back().maintenance_material_costs = data->pylon_mass_ratio * CPI_factor * 0.1391 * data->number_of_engines * timeFactor * conversionFactor;
        } else if (data->spare_parts_name[i] == "PowerUnit") { //spare_parts_ATA_energy_unit: ATA28, ATA36, ATA54
            for (uint8_t k(0); k < 3; k++) {
                double airflow(0.454 * data->study_mission_PAX); // Air conditioning mass flow [kg/min]
                /*Order of coefficients: ATA28, ATA36, ATA54*/
                vector<double> c{0.595, 0., 0.};// Method coefficient
                vector<double> d{0.0123, 0.0019, 0.1391};// Method coefficient
                vector<double> X{data->blockfuel_study/1000, airflow, static_cast<double>(data->number_of_engines)};// Method coefficient
                vector<double> Y{1., data->myEngine.get_scaled_SLST()/10000, 1.}; // Method coefficient
                vector<double> HR{0.94, 0.26, 0.80};// Flight hour related cost ratio (ATA28, ATA36, ATA54)
                vector<double> CR{0.06, 0.74, 0.20};// Flight cycle related cost ratio (ATA28, ATA36, ATA54)
                double timeFactor = (convertUnit(SECOND, HOUR, data->flight_time_study) * HR[k] + CR[k]) / (2.5 * HR[k] + CR[k]);
                theSpareParts.back().maintenance_material_costs += CPI_factor * (c[k] + d[k]*X[k]*Y[k]) * timeFactor * conversionFactor;
            }
        } else {
            myRuntimeInfo->err << "Invalid name of spare part." << std::endl;
        }
        theSpareParts.back().energy_consumption_ratio = theSpareParts.back().maintenance_material_costs / theSpareParts.back().RC_component;
        myRuntimeInfo->info << "spare_parts_name: " << data->spare_parts_name[i] << std::endl;
        myRuntimeInfo->info << "RC_component [$]: " << theSpareParts.back().RC_component << std::endl;
        myRuntimeInfo->info << "maintenance_material_costs [$]: " << theSpareParts.back().maintenance_material_costs << std::endl;
        myRuntimeInfo->info << "available_seats [-]: " << data->available_seats << std::endl;
        myRuntimeInfo->info << "range_study [m]: " << data->range_study << std::endl;
        myRuntimeInfo->info << "energy_consumption_ratio [-]: " << theSpareParts.back().energy_consumption_ratio << std::endl;
        string compEntry(data->spare_parts_name[i]);
        vector<string> component_name_tmp;
        for (unsigned int k(0.); k < theProduction.theAcComp.size(); k++) {
            component_name_tmp.push_back(theProduction.theAcComp.at(k).component_name);
        }
        vector<string>::iterator entry = std::find(component_name_tmp.begin(), component_name_tmp.end(), compEntry); //Iterator to find the configuration input
        if (entry == component_name_tmp.end()) { //Input not available!
            myRuntimeInfo->err << "components " << compEntry << " not found. Abort!" << std::endl;
            exit(1);
        } else {
            unsigned int pos(0);
            pos = static_cast<int>(entry - component_name_tmp.begin());
            theSpareParts.back().energy_spare_parts_manufacturing = theSpareParts.back().energy_consumption_ratio * theProduction.theAcComp.at(pos).energy_manufacturing;
        }
        energy_spare_parts_total += theSpareParts.back().energy_spare_parts_manufacturing;
    }
    /* Maintenance Labour (Cost calculation: DOC TUB)*/
    myRuntimeInfo->info << "  Maintenance labour starts" << std::endl;
    double convertEuroToDollar(1/0.72); // Conversion factor EUR -> USD
    double wageCostLabourTUB(50); // [€/h]
    double burden = 2.0; //Cost burden
    double SLST = data->myEngine.get_scaled_SLST()/1000/G_FORCE; // [t] converted thrust from N
    double engineDMC = convertEuroToDollar * (data->number_of_engines * (1.5 * SLST + //[$/yr]
                30.5 * convertUnit(SECOND, HOUR, data->flight_time_study) + 10.6)) * data->flights_per_year;
    double engineLabourCosts = (1 - engineCostRatioMaterial) * engineDMC; //[$/yr]
    double airframeLabourCosts = convertEuroToDollar * wageCostLabourTUB * (1 + burden) * ((0.655 + 0.01 * data->ome / 1000.) //[USD/yr]
                                     *convertUnit(SECOND, HOUR, data->flight_time_study) + 0.254 + 0.01 * data->ome / 1000.) * data->flights_per_year;
    theMaintLabour.DMC_labor_per_year = (airframeLabourCosts + engineLabourCosts);
    myRuntimeInfo->debug << "Available Seats " << data->available_seats << std::endl;
    myRuntimeInfo->debug << "flight time [s] " << data->flight_time_study << std::endl;
    myRuntimeInfo->debug << "range [km] " << convertUnit(NOPREFIX, METER, KILO, METER, data->range_study) << std::endl;
    myRuntimeInfo->debug << "Flights per year " << data->flights_per_year << std::endl;
    myRuntimeInfo->debug << "DMC_labor_per_year [$] " << theMaintLabour.DMC_labor_per_year << std::endl;
    theMaintLabour.engineering_energy_electricity = energy_PerWorker_op[0] * theMaintLabour.DMC_labor_per_year / (C_labourHour_op * t_labourYear_op) * data->duration_operation;
    theMaintLabour.engineering_energy_heat = energy_PerWorker_op[1] * theMaintLabour.DMC_labor_per_year / (C_labourHour_op * t_labourYear_op) * data->duration_operation;
    energy_labour_total += theMaintLabour.engineering_energy_electricity + theMaintLabour.engineering_energy_heat;
    energy_total = energy_spare_parts_total + energy_labour_total;
    myRuntimeInfo->debug << "Check Labour Hours Maintenance: " << theMaintLabour.DMC_labor_per_year/(C_labourHour_op) << std::endl;
}

void Operation::calculate_operation_emissions(const std::shared_ptr<SchaeferIOData>& data, const Production &theProduction) {
    /* Missions in the operational phase */
    missions_fuel = (data->blockfuel_study) * data->flights_per_year * data->duration_operation;
    myRuntimeInfo->info << "  Operation Fuel: " << missions_fuel << " kg" << std::endl;
    missions.CO2 = (data->theStudyMission.LTO.CO2 + data->theStudyMission.cruise.CO2) * data->flights_per_year * data->duration_operation;
    missions.H2O = (data->theStudyMission.LTO.H2O + data->theStudyMission.cruise.H2O) * data->flights_per_year * data->duration_operation;
    missions.NOx = (data->theStudyMission.LTO.NOx + data->theStudyMission.cruise.NOx) * data->flights_per_year * data->duration_operation;
    missions.SO2 = (data->theStudyMission.LTO.SO2 + data->theStudyMission.cruise.SO2) * data->flights_per_year * data->duration_operation;
    missions.HC = (data->theStudyMission.LTO.HC + data->theStudyMission.cruise.HC) * data->flights_per_year * data->duration_operation;
    missions.CO = (data->theStudyMission.LTO.CO + data->theStudyMission.cruise.CO) * data->flights_per_year * data->duration_operation;
    missions.CH4 = (data->theStudyMission.LTO.CH4 + data->theStudyMission.cruise.CH4) * data->flights_per_year * data->duration_operation;
    missions.PM10 = (data->theStudyMission.LTO.soot + data->theStudyMission.cruise.soot) * data->flights_per_year * data->duration_operation;

    /*Fuel production in the operational phase*/
    operationFuelProduction.CO2 = missions_fuel * myEcoDatabase.fuelProdEmData.at(0).CO2;
    operationFuelProduction.H2O = missions_fuel * myEcoDatabase.fuelProdEmData.at(0).H2O;
    operationFuelProduction.NOx = missions_fuel * myEcoDatabase.fuelProdEmData.at(0).NOx;
    operationFuelProduction.SO2 = missions_fuel * myEcoDatabase.fuelProdEmData.at(0).SO2;
    operationFuelProduction.CO = missions_fuel * myEcoDatabase.fuelProdEmData.at(0).CO;
    operationFuelProduction.CH4 = missions_fuel * myEcoDatabase.fuelProdEmData.at(0).CH4;
    operationFuelProduction.PM10 = missions_fuel * myEcoDatabase.fuelProdEmData.at(0).PM10;
    operationFuelProduction.CED = missions_fuel * myEcoDatabase.fuelSpecificEnergy.at(0);

    /*Maintenance Labour*/
    //Electrical emissions
    myRuntimeInfo->info << "      Maintenance (Labour) Energy electrical: " << theMaintLabour.engineering_energy_electricity << " MJ" << std::endl;
    maintenanceLabourElectricity.CO2 = theMaintLabour.engineering_energy_electricity * myEcoDatabase.elecEmData.CO2;
    maintenanceLabourElectricity.H2O = theMaintLabour.engineering_energy_electricity * myEcoDatabase.elecEmData.H2O;
    maintenanceLabourElectricity.NOx = theMaintLabour.engineering_energy_electricity * myEcoDatabase.elecEmData.NOx;
    maintenanceLabourElectricity.SO2 = theMaintLabour.engineering_energy_electricity * myEcoDatabase.elecEmData.SO2;
    maintenanceLabourElectricity.CO =  theMaintLabour.engineering_energy_electricity * myEcoDatabase.elecEmData.CO;
    maintenanceLabourElectricity.CH4 = theMaintLabour.engineering_energy_electricity * myEcoDatabase.elecEmData.CH4;
    maintenanceLabourElectricity.PM10 = theMaintLabour.engineering_energy_electricity * myEcoDatabase.elecEmData.PM10;
    maintenanceLabourElectricity.CED = theMaintLabour.engineering_energy_electricity * myEcoDatabase.elecEmData.specificEnergy;
    //Heating emissions
    myRuntimeInfo->info << "      Maintenance (Labour) Energy heat: " << theMaintLabour.engineering_energy_heat << " MJ" << std::endl;
    maintenanceLabourHeat.CO2 = theMaintLabour.engineering_energy_heat * myEcoDatabase.heatEmData.CO2;
    maintenanceLabourHeat.H2O = theMaintLabour.engineering_energy_heat * myEcoDatabase.heatEmData.H2O;
    maintenanceLabourHeat.NOx = theMaintLabour.engineering_energy_heat * myEcoDatabase.heatEmData.NOx;
    maintenanceLabourHeat.SO2 = theMaintLabour.engineering_energy_heat * myEcoDatabase.heatEmData.SO2;
    maintenanceLabourHeat.CO = theMaintLabour.engineering_energy_heat * myEcoDatabase.heatEmData.CO;
    maintenanceLabourHeat.CH4 = theMaintLabour.engineering_energy_heat * myEcoDatabase.heatEmData.CH4;
    maintenanceLabourHeat.PM10 = theMaintLabour.engineering_energy_heat * myEcoDatabase.heatEmData.PM10;
    maintenanceLabourHeat.CED = theMaintLabour.engineering_energy_heat * myEcoDatabase.heatEmData.specificEnergy;
    //Total maintenance Labour Emissions
    myRuntimeInfo->info << "    Maintenance (Labour) Energy: " << energy_labour_total << " MJ" << std::endl;
    maintenanceLabourTotal.CO2 = maintenanceLabourElectricity.CO2 + maintenanceLabourHeat.CO2;
    maintenanceLabourTotal.H2O = maintenanceLabourElectricity.H2O + maintenanceLabourHeat.H2O;
    maintenanceLabourTotal.NOx = maintenanceLabourElectricity.NOx + maintenanceLabourHeat.NOx;
    maintenanceLabourTotal.SO2 = maintenanceLabourElectricity.SO2 + maintenanceLabourHeat.SO2;
    maintenanceLabourTotal.CO = maintenanceLabourElectricity.CO + maintenanceLabourHeat.CO;
    maintenanceLabourTotal.CH4 = maintenanceLabourElectricity.CH4 + maintenanceLabourHeat.CH4;
    maintenanceLabourTotal.PM10 = maintenanceLabourElectricity.PM10 + maintenanceLabourHeat.PM10;
    maintenanceLabourTotal.CED = maintenanceLabourElectricity.CED + maintenanceLabourHeat.CED;
    //are zero, nevertheless perform:
    maintenanceLabourTotal.N2O = maintenanceLabourElectricity.N2O + maintenanceLabourHeat.N2O;
    maintenanceLabourTotal.PFC = maintenanceLabourElectricity.PFC + maintenanceLabourHeat.PFC;
    maintenanceLabourTotal.NMVOC = maintenanceLabourElectricity.NMVOC + maintenanceLabourHeat.NMVOC;
    maintenanceLabourTotal.HC = maintenanceLabourElectricity.HC + maintenanceLabourHeat.HC;
    /*Maintenance Spare Parts*/
    for (int i(0); i < data->configuration.spare_parts_number; i++) {
        /* emission calculation */
        //Electrical energy of production
        SpareParts.push_back(Emissions());
        SpareParts.back().CO2 = theSpareParts.at(i).energy_spare_parts_manufacturing * myEcoDatabase.elecEmData.CO2;
        SpareParts.back().H2O = theSpareParts.at(i).energy_spare_parts_manufacturing * myEcoDatabase.elecEmData.H2O;
        SpareParts.back().NOx = theSpareParts.at(i).energy_spare_parts_manufacturing * myEcoDatabase.elecEmData.NOx;
        SpareParts.back().SO2 = theSpareParts.at(i).energy_spare_parts_manufacturing * myEcoDatabase.elecEmData.SO2;
        SpareParts.back().CO =  theSpareParts.at(i).energy_spare_parts_manufacturing * myEcoDatabase.elecEmData.CO;
        SpareParts.back().CH4 = theSpareParts.at(i).energy_spare_parts_manufacturing * myEcoDatabase.elecEmData.CH4;
        SpareParts.back().PM10 = theSpareParts.at(i).energy_spare_parts_manufacturing * myEcoDatabase.elecEmData.PM10;
        SpareParts.back().CED = theSpareParts.at(i).energy_spare_parts_manufacturing * myEcoDatabase.elecEmData.specificEnergy;
        //Find components from production with the same name
        string compEntry(data->spare_parts_name[i]);
        vector<string> component_name_tmp;
        for (unsigned int j(0.); j < theProduction.theAcComp.size(); j++) {
            component_name_tmp.push_back(theProduction.theAcComp.at(j).component_name);
        }
        vector<string>::iterator entry = find(component_name_tmp.begin(), component_name_tmp.end(), compEntry); //Iterator to find the configuration entry
        if (entry == component_name_tmp.end()) { //Entry not available!
            myRuntimeInfo->err << "component " << compEntry << " not found. Abort!" << std::endl;
            exit(1);
        } else {
            unsigned int compPos(0);
            compPos = static_cast<int>(entry - component_name_tmp.begin());
            //raw material extraction
            for (int k(0); k < numberOfMaterials; k++) {
                string materialEntry(theProduction.theMaterial.at(k).material_name);
                vector<string> material_name_tmp;
                for (unsigned int m(0.); m < myEcoDatabase.matEmData.size(); m++) {
                    material_name_tmp.push_back(myEcoDatabase.matEmData.at(m).specification);
                }
                entry = find(material_name_tmp.begin(), material_name_tmp.end(), materialEntry); //Iterator to find the configuration entry
                if (entry == material_name_tmp.end()) { //Entry not available!
                    myRuntimeInfo->err << "Material " << materialEntry << " in EcoDatabase not found!" << std::endl;
                    exit(1);
                } else {
                    unsigned int matPos(0);
                    matPos = static_cast<int>(entry - material_name_tmp.begin());
                    /* emission calculation */
                    //The factor energy_consumption_ratio also defines here as the relationship between production and maintenance.
                    //The mass of the total component is therefore taken here with this factor
                    SpareParts.back().CO2 += theSpareParts.at(i).energy_consumption_ratio * (1 + clippings[k])
                    * theProduction.theAcComp.at(compPos).mass * theProduction.theAcComp.at(compPos).rate_material.at(k) * myEcoDatabase.matEmData.at(matPos).CO2;
                    SpareParts.back().H2O += theSpareParts.at(i).energy_consumption_ratio * (1 + clippings[k])
                    * theProduction.theAcComp.at(compPos).mass * theProduction.theAcComp.at(compPos).rate_material.at(k) * myEcoDatabase.matEmData.at(matPos).H2O;
                    SpareParts.back().NOx += theSpareParts.at(i).energy_consumption_ratio * (1 + clippings[k])
                    * theProduction.theAcComp.at(compPos).mass * theProduction.theAcComp.at(compPos).rate_material.at(k) * myEcoDatabase.matEmData.at(matPos).NOx;
                    SpareParts.back().SO2 += theSpareParts.at(i).energy_consumption_ratio * (1 + clippings[k])
                    * theProduction.theAcComp.at(compPos).mass * theProduction.theAcComp.at(compPos).rate_material.at(k) * myEcoDatabase.matEmData.at(matPos).SO2;
                    SpareParts.back().CO += theSpareParts.at(i).energy_consumption_ratio * (1 + clippings[k])
                    * theProduction.theAcComp.at(compPos).mass * theProduction.theAcComp.at(compPos).rate_material.at(k) * myEcoDatabase.matEmData.at(matPos).CO;
                    SpareParts.back().CH4 += theSpareParts.at(i).energy_consumption_ratio * (1 + clippings[k])
                    * theProduction.theAcComp.at(compPos).mass * theProduction.theAcComp.at(compPos).rate_material.at(k) * myEcoDatabase.matEmData.at(matPos).CH4;
                    SpareParts.back().PM10 += theSpareParts.at(i).energy_consumption_ratio * (1 + clippings[k])
                    * theProduction.theAcComp.at(compPos).mass * theProduction.theAcComp.at(compPos).rate_material.at(k) * myEcoDatabase.matEmData.at(matPos).PM10;
                    SpareParts.back().CED += theSpareParts.at(i).energy_consumption_ratio * (1 + clippings[k])
                    * theProduction.theAcComp.at(compPos).mass * theProduction.theAcComp.at(compPos).rate_material.at(k) * myEcoDatabase.matSpecificEnergy.at(matPos);
                }
            }
        }
        //Spare part emissions total
        sparePartsTotal.CO2 += SpareParts.back().CO2;
        sparePartsTotal.H2O += SpareParts.back().H2O;
        sparePartsTotal.NOx += SpareParts.back().NOx;
        sparePartsTotal.SO2 += SpareParts.back().SO2;
        sparePartsTotal.CO += SpareParts.back().CO;
        sparePartsTotal.CH4 += SpareParts.back().CH4;
        sparePartsTotal.PM10 += SpareParts.back().PM10;
        sparePartsTotal.N2O += SpareParts.back().N2O;
        sparePartsTotal.PFC += SpareParts.back().PFC;
        sparePartsTotal.NMVOC += SpareParts.back().NMVOC;
        sparePartsTotal.HC += SpareParts.back().HC;
        sparePartsTotal.CED += SpareParts.back().CED;
    }
    myRuntimeInfo->info << "    Maintenance (Spare Parts) Energy: " << energy_spare_parts_total << " MJ" << std::endl;

    /*Operation phase total*/
    myRuntimeInfo->info << "  Operating Energy Total: " << energy_total << " MJ" << std::endl;
    operationTotal.CO2 = missions.CO2 + operationFuelProduction.CO2 + maintenanceLabourElectricity.CO2 + maintenanceLabourHeat.CO2 + sparePartsTotal.CO2;
    operationTotal.H2O = missions.H2O + operationFuelProduction.H2O + maintenanceLabourElectricity.H2O + maintenanceLabourHeat.H2O + sparePartsTotal.H2O;
    operationTotal.NOx = missions.NOx + operationFuelProduction.NOx + maintenanceLabourElectricity.NOx + maintenanceLabourHeat.NOx + sparePartsTotal.NOx;
    operationTotal.SO2 = missions.SO2 + operationFuelProduction.SO2 + maintenanceLabourElectricity.SO2 + maintenanceLabourHeat.SO2 + sparePartsTotal.SO2;
    operationTotal.CO = missions.CO + operationFuelProduction.CO + maintenanceLabourElectricity.CO + maintenanceLabourHeat.CO + sparePartsTotal.CO;
    operationTotal.CH4 = missions.CH4 + operationFuelProduction.CH4 + maintenanceLabourElectricity.CH4 + maintenanceLabourHeat.CH4 + sparePartsTotal.CH4;
    operationTotal.PM10 = missions.PM10 + operationFuelProduction.PM10 + maintenanceLabourElectricity.PM10 + maintenanceLabourHeat.PM10 + sparePartsTotal.PM10;
    operationTotal.N2O = missions.N2O + operationFuelProduction.N2O + maintenanceLabourElectricity.N2O + maintenanceLabourHeat.N2O + sparePartsTotal.N2O;
    operationTotal.PFC = missions.PFC + operationFuelProduction.PFC + maintenanceLabourElectricity.PFC + maintenanceLabourHeat.PFC + sparePartsTotal.PFC;
    operationTotal.NMVOC = missions.NMVOC + operationFuelProduction.NMVOC + maintenanceLabourElectricity.NMVOC + maintenanceLabourHeat.NMVOC + sparePartsTotal.NMVOC;
    operationTotal.HC = missions.HC + operationFuelProduction.HC + maintenanceLabourElectricity.HC + maintenanceLabourHeat.HC + sparePartsTotal.HC;
    operationTotal.CED = operationFuelProduction.CED + maintenanceLabourElectricity.CED + maintenanceLabourHeat.CED + sparePartsTotal.CED;
    myRuntimeInfo->info << "     missions.HC: " << missions.HC << " kg" << std::endl;
    myRuntimeInfo->info << "     studyMission.LTO.HC: " << data->theStudyMission.LTO.HC << " kg" << std::endl;
    myRuntimeInfo->info << "     studyMission.cruise.HC: " << data->theStudyMission.cruise.HC << " kg" << std::endl;
    myRuntimeInfo->info << "     maintenanceLabourHeat.CO2: " << maintenanceLabourHeat.CO2 << " kg" << std::endl;
    myRuntimeInfo->info << "     missions.CH4: " << missions.CH4 << " kg" << std::endl;
    myRuntimeInfo->info << "     operationFuelProduction.CH4: " << operationFuelProduction.CH4 << " kg" << std::endl;
    myRuntimeInfo->info << "     maintenanceLabourElectricity.CH4: " << maintenanceLabourElectricity.CH4 << " kg" << std::endl;
    myRuntimeInfo->info << "     maintenanceLabourHeat.CH4: " << maintenanceLabourHeat.CH4 << " kg" << std::endl;
    myRuntimeInfo->info << "  Operating Primary Energy Total: " << operationTotal.CED << " MJ" << std::endl;
}


Operation::~Operation() {
    //dtor
}

Operation::Labour::Labour()
    :
    DMC_labor_per_year(0.),
    engineering_energy_electricity(0.),
    engineering_energy_heat(0.) {
}
Operation::Labour::~Labour() {
}

Operation::spareParts::spareParts(const string& anATA_ID, const string& name)
    :
    ATA_ID(anATA_ID),
    component_name(name),
    RC_component(0.),
    maintenance_material_costs(0.),
    energy_consumption_ratio(0.),
    energy_spare_parts_manufacturing(0.) {
}
Operation::spareParts::~spareParts() {
}

