/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_EMISSION_CALCULATION_LIFE_CYCLE_EMISSIONS_LCA_SCHAEFER_PRODUCTION_H_
#define ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_EMISSION_CALCULATION_LIFE_CYCLE_EMISSIONS_LCA_SCHAEFER_PRODUCTION_H_

#include <string>
#include <vector>
#include <memory>
#include <moduleBasics/runtimeIO.h>
#include "../../ecoDatabase.h"
#include "../../../emissionsClass.h"
#include "schaeferIOData.h"

class SchaeferIOData;

/** \class Production
 * \brief Contains the properties of the production phase
 */
class Production{
 private:
    const ecoDatabase myEcoDatabase; /**< EcoDatase object */

 public:
    std::vector<Emissions> materialProduction;/**< Emissions due to material production distributed according to the individual components */
    std::vector<Emissions> materialProcessing;/**< Emissions from material processing to components, distributed according to the individual components */
    Emissions materialProductionTotal;/**< Total emissions from material production */
    Emissions materialProcessingTotal;/**< Total emissions from material processing of components */
    Emissions materialProductionEngine;/**< Total emissions from material production of engines */
    Emissions materialProcessingEngine;/**< Total emissions from material processing of engines */
    double material_energy;/**< Energy needed for material [MJ] */
    Emissions labourProductionElectricity;/**< Emissions due to electricity from manufacturing personnel (not process! light/heating, etc.) */
    Emissions labourProductionHeat;/**< Emissions due to heat from manufacturing personnel (not process! light/heating, etc.) */
    Emissions labourProductionTotal;/**< Total emissions due to heat from manufacturing personnel (not process! light/heating, etc.) */
    Emissions labourProductionElectricityEngine;/**< Emissions due to electricity from manufacturing personnel for engines (not process! light/heating, etc.) */
    Emissions labourProductionHeatEngine;/**< Emissions due to heat from manufacturing personnel for engines (not process! light/heating, etc.) */
    Emissions labourProductionEngine;/**< Total emissions from manufacturing personnel for engines (not process! light/heating, etc.) */
    Emissions finalAssemblyElectricity;/**< Emissions due to electricity generated by assembly personnel */
    Emissions finalAssemblyHeat;/**< Emissions due to heating of assembly personnel */
    Emissions finalAssemblyTotal;/**< Total emissions of assembly personnel */
    Emissions finalFlightTests;/**< Emissions due to test flights after final assembly (incl. fuel production) */
    double test_flight_fuel;/**< Fuel consumption for one test flight [kg]*/
    Emissions finalTestFlightFuelProduction;/**< Emissions due to fuel production for final test flight */
    Emissions transport;/**< Emissions from transport means */
    Emissions productionTotal;/**< Total emissions during the production phase */
    Emissions productionEngine;/**< Emissions during the engine production phase */
    double total_RC;/**< Total recurring costs [$]*/
    double energy_manufacturing_engine;/**< Total energy demand for engine manufacturing [MJ] */
    double energy_manufacturing_total;/**< Total energy demand for manufacturing [MJ] */
    double energy_engine;/**< Total energy demand for the engine in the production phase [MJ] */
    double energy_total;/**< Total energy demand for the entire production phase [MJ] */

    std::vector<double> recovery;/**< Recovery rate of the materials [-]*/

    /** \class mainComponent
     * \brief Distribution of emissions by main components
     */
    class mainComponent {
     public:
        std::string component_name;/**< Name of the main component ("Wing", "Fuselage", "Tailplane", "Fin", "LandingGear", "Pylons", "PowerUnit")*/

        double mass;/**< Mass of the component [kg]  */
        double energy_manufacturing;/**< Energy demand for manufacturing the component [MJ] */
        double RC;/**< Recurring cost [USD] */

        std::vector<double> rate_material;/**<Percentage of different materials. s. const double matDistr [-] */
        std::vector<double> raw_material;/**< Necessary raw material [kg] */

        /** \brief Constructor of class mainComponent
         * \param name Name of the component
         */
        explicit mainComponent(const std::string& name);
        /** \brief Destructor of class mainComponent */
        virtual ~mainComponent();
    };
    std::vector<mainComponent> theAcComp; /**< Vector contains aircraft components */

    /** \class material
     * \brief Distribution of emissions by materials
     */
    class material {
     public:
        std::string material_name;/**< Name of the material ("Al","CFRP","St","Ti","Ni", "Cr","Cu","Ni","Pl") */

        double mass;/**< Mass of the material [kg]*/
        double raw_material_mass;/**< Mass of the raw material [kg]*/
        double energy_manufacturing;/**< Manufacturing energy demand of the material [MJ] */

        /** \brief Constructor of class material */
        material();
        /** \brief Destructor of class material */
        virtual ~material();
    };
    std::vector<material> theMaterial; /**< Contains materials */

    /** \class  manufacturingWork
     * \brief Class contains personnel expenditure for manufacturing
     */
    class manufacturingWork {
     public:
        double labour_costs;/**< Cost for labour [USD]  */
        double engineering_energy_electricity;/**< Total electric energy demand [MJ] */
        double engineering_energy_electricity_eng;/**< Electric energy demand for engineering [MJ] */
        double engineering_energy_heat;/**< Total heat energy demand [MJ] */
        double engineering_energy_heat_eng;/**< Heat energy demand for engineering [MJ] */

        /** \brief Constructor of class manufacturingWork */
        manufacturingWork();
        /** \brief Destructor of class manufacturingWork */
        virtual ~manufacturingWork();
    };
    manufacturingWork theManufacturingWork; /**< Object contains personnel expenditure for manufacturing */

    /** \class Assembly
     * \brief Class contains personnel expenditure for assembly
     */
    class Assembly {
     public:
        double costs;/**< Assembly costs [USD] */
        double engineering_energy_electricity;/**< Electric energy demand for assembly [MJ] */
        double engineering_energy_heat;/**< Heat energy demand for assembly [MJ] */

        /** \brief Constructor of class Assembly */
        Assembly();
        /** \brief Destructor of class Assembly */
        virtual ~Assembly();
    };
    Assembly theFinalAssembly; /**< Object contains personnel expenditure for final assembly */

    /** \class TransportLogistics
     * \brief Transport of the components
     */
    class TransportLogistics {
     public:
        std::string transport_type;/**< Type of transport ("Aircraft", "Truck", "Ship", "Train") */
        std::vector<std::string> name_component;/**< Name of the component ("Wing", "Fuselage", "Tailplane", "Fin", "LandingGear", "Pylons", "PowerUnit") */
        std::vector<double> mass_component;/**< Mass of the component [kg] */
        std::vector<double> distance_component;/**< Transport distance of the component [km] */

        double mass_total;/**< Total mass of the transported components [kg] */
        double distance_total;/**< Total distance of the transported components [km]*/
        double fuel;/**< Total fuel demand for transport [kg] */

      	/** \brief Constructor of class Assembly
         *  \param name Name of the mean of transportation
         */
        explicit TransportLogistics(const std::string& name);
        /** \brief Destructor of class TransportLogistics */
        virtual ~TransportLogistics();
    };
    std::vector<TransportLogistics> theTransportation; /**< Object contains transport data*/

    /** \brief Constructor of class production
     */
    Production();

    /** \brief Destructor of class production
     */
    virtual ~Production();

    /** \brief Function to calculate the energy demand of the production phase
     * \param data Access to IO data
     * \param rtIO Access to rtIO
     */
    void calculate_production_resources(const std::shared_ptr<SchaeferIOData>& data, const std::shared_ptr<RuntimeIO>& rtIO);

    /** \brief Function to calculate the emissions of the production phase
     * \param data Access to IO data
     */
    void calculate_production_emissions(const std::shared_ptr<SchaeferIOData>& data);

    /** \brief NASA cost method. can be applied for all aircraft components
     *
     * \param Mass double: mass of the component
     * \param Quantity double: production quantity
     * \param A0 double: Factor A0
     * \param A1 double: Factor A1
     * \param A2 double: Factor A2
     * \return double: [Mio USD] cost value
     *
     */
    double get_NASA_costs(double Mass, double Quantity, double A0, double A1, double A2);
};
#endif // ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_EMISSION_CALCULATION_LIFE_CYCLE_EMISSIONS_LCA_SCHAEFER_PRODUCTION_H_
