/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "schaeferIOData.h"
#include "schaeferConfig.h"
#include "development.h"
#include "production.h"
#include "operation.h"
#include "endOfLife.h"
#include <unordered_map>
#include <string>
#include <aixml/node.h>
#include <runtimeInfo/runtimeInfo.h>
#include <moduleBasics/runtimeIO.h>
#include <standardFiles/functions.h>
#include <unitConversion/unitConversion.h>
#include <energyCarriers/energyCarriers.h>

using std::endl;
using std::ostream;
using std::string;
using std::vector;

SchaeferIOData::SchaeferIOData(std::shared_ptr<RuntimeIO> rtIO, std::shared_ptr<node> ecoAssXML)
    :
    configuration(SchaeferConfig()),
    myEngine(rtIO->getEngineDataDir() + "/" +
            EndnodeReadOnly<std::string>("component_design/propulsion/specific/propulsion/engine/model").read(rtIO->acxml).value(),
            EndnodeReadOnly<double>("component_design/propulsion/specific/propulsion/engine/scale_factor").read(rtIO->acxml).value()),
    theTestMission(mission()),
    /*Requirements and specifications*/
    max_operating_speed(EndnodeReadOnly<double>(string("requirements_and_specifications/requirements/top_level_aircraft_requirements/flight_envelope/")
                                    + string("maximum_operating_velocity")).read(rtIO->acxml).value()),
    flights_per_year(EndnodeReadOnly<int>("assessment/cost_estimation/operating_cost/direct_operating_cost/flights_per_year_study_mission").read(rtIO->acxml).value()),
    duration_operation(EndnodeReadOnly<int>("requirements_and_specifications/assessment_scenario/duration_operation").read(rtIO->acxml).value_xml()),
    available_seats(EndnodeReadOnly<double>(string("requirements_and_specifications/design_specification/transport_task/passenger_definition/")
                                            + string("total_number_passengers")).read(rtIO->acxml).value()),
    /*Aerodynamics*/
    wing_reference_area(EndnodeReadOnly<double>("analysis/aerodynamics/reference_values/S_ref").read(rtIO->acxml).value()),
    /*Masses*/
    mtom(EndnodeReadOnly<double>("analysis/masses_cg_inertia/maximum_takeoff_mass/mass_properties/mass").read(rtIO->acxml).value()),
    ome(EndnodeReadOnly<double>("analysis/masses_cg_inertia/operating_mass_empty/mass_properties/mass").read(rtIO->acxml).value()),
    spare_parts_mass(6),
    main_parts_mass(7),
    transported_component_mass(6),
    airframe_mass(0.),
    fin_mass_ratio(0.),
    stabilizer_mass_ratio(0.),
    pylon_mass_ratio(0.),
    nacelle_mass_ratio(0.),
    engine_mass(0.),
    nacelle_mass(0.),
    pylon_mass(0.),
    bleed_air_system_mass(0.),
    fuel_system_mass(0.),
    other_engine_systems_mass(0.),
    /*Propulsion*/
    number_of_engines(rtIO->acxml.getVector("aircraft_exchange_file/component_design/propulsion/specific/propulsion/").size()),
    fuel_flow_cont_thrust(0.),
    /*Systems*/
    propulsion_design_power(0.),
    /*Design Mission*/
    range_design(EndnodeReadOnly<double>("analysis/mission/design_mission/range").read(rtIO->acxml).value()),
    blocktime_design(convertUnit(SECOND, HOUR, EndnodeReadOnly<double>("analysis/mission/design_mission/block_time").read(rtIO->acxml).value())),
    flight_time_design(convertUnit(SECOND, HOUR, EndnodeReadOnly<double>("analysis/mission/design_mission/flight_time").read(rtIO->acxml).value())),
    taxi_fuel_design(0.),
    tripfuel_design(0.),
    blockfuel_design(0.),
    LTO_fuel_design(0.),
    /*Study Mission*/
    range_study(EndnodeReadOnly<double>("requirements_and_specifications/requirements/top_level_aircraft_requirements/study_mission/range").read(rtIO->acxml).value()),
    flight_time_study(EndnodeReadOnly<double>("analysis/mission/study_mission/flight_time").read(rtIO->acxml).value()),
    tripfuel_study(0.),
    study_mission_PAX(0.),
    blockfuel_study(0.) {
        configuration.read_all_config_data(rtIO->moduleConfig);
        read_aircraft(rtIO);
        get_mission_file_data(rtIO);
        read_ecological_assessment_xml(ecoAssXML);
    }

/* Input data methods*/
void SchaeferIOData::read_aircraft(std::shared_ptr<RuntimeIO> rtIO) {
    double propDesignPowerElecHyd(EndnodeReadOnly<double>("component_design/systems/specific/maximium_power_demand").read(rtIO->acxml).value());
    propulsion_design_power = (fabs(propDesignPowerElecHyd)) * 1 / 0.27 /1000; //kW; assumption: 27% electrical+hydr. power, rest: pneumatically:
    //...s. _Wheeler_MoreElectricAircraft (under AircraftDesign/Systems/...)
    number_of_engines = rtIO->acxml.getVector("aircraft_exchange_file/component_design/propulsion/specific/propulsion/").size();
    myEngine.set_same_engine_type_quantity(number_of_engines);
    double buckt_point_thrust(EndnodeReadOnly<double>("component_design/propulsion/specific/propulsion/engine/bucket_point/thrust").read(rtIO->acxml).value());
    double buckt_point_tsfc(EndnodeReadOnly<double>("component_design/propulsion/specific/propulsion/engine/bucket_point/tsfc").read(rtIO->acxml).value());
    fuel_flow_cont_thrust = buckt_point_thrust * buckt_point_tsfc;
    study_mission_PAX = std::round(EndnodeReadOnly<double>(std::string("requirements_and_specifications/requirements/top_level_aircraft_requirements/study_mission/")
                                                + std::string("payload_fractions/passenger_mass_fraction")).read(rtIO->acxml).value()
                                    * EndnodeReadOnly<double>(std::string("requirements_and_specifications/design_specification/transport_task/")
                                                + std::string("passenger_definition/total_number_passengers")).read(rtIO->acxml).value());
    get_masses(rtIO);
}

void SchaeferIOData::get_mission_file_data(std::shared_ptr<RuntimeIO> rtIO) {
    std::string mission_xml_name = std::string(rtIO->acxml.at("/aircraft_exchange_file/requirements_and_specifications/mission_files/design_mission_file/value"));
    const node& mission_xml(aixml::openDocument(rtIO->getMissionDataDir() + "/" + mission_xml_name));
    taxi_time = EndnodeReadOnly<double>("mission/taxi_time_origin").read(mission_xml).value() + EndnodeReadOnly<double>("mission/taxi_time_destination").read(mission_xml).value();
}

void SchaeferIOData::get_masses(std::shared_ptr<RuntimeIO> rtIO) {
    /* Aircraft masses */
    double MME(EndnodeReadOnly<double>("analysis/masses_cg_inertia/manufacturer_mass_empty/mass_properties/mass").read(rtIO->acxml).value());
    double systemsMass(EndnodeReadOnly<double>("component_design/systems/mass_properties/mass").read(rtIO->acxml).value());
    double wingMass(EndnodeReadOnly<double>("component_design/wing/mass_properties/mass").read(rtIO->acxml).value());
    double fuselageMass(EndnodeReadOnly<double>("component_design/fuselage/mass_properties/mass").read(rtIO->acxml).value());
    double landingGearMass(EndnodeReadOnly<double>("component_design/landing_gear/mass_properties/mass").read(rtIO->acxml).value());

    /* Propulsion masses and ratio */
    double powerUnitMass(0.);
    int numberOfPropulsors(rtIO->acxml.getVector("aircraft_exchange_file/component_design/propulsion/specific/propulsion/").size());
    for (int i(0); i < numberOfPropulsors; i++) {
       nacelle_mass += EndnodeReadOnly<double>("component_design/propulsion/specific/nacelle/mass_properties/mass").read(rtIO->acxml).value();
       pylon_mass += EndnodeReadOnly<double>("component_design/propulsion/specific/pylon/mass_properties/mass").read(rtIO->acxml).value();
       engine_mass += EndnodeReadOnly<double>("component_design/propulsion/specific/engine/mass_properties/mass").read(rtIO->acxml).value();
    }
    bleed_air_system_mass = EndnodeReadOnly<double>("component_design/systems/specific/geometry/mass_properties/bleed_air_system").read(rtIO->acxml).value();
    fuel_system_mass = EndnodeReadOnly<double>("component_design/systems/specific/geometry/mass_properties/fuel_system").read(rtIO->acxml).value();
    //other_engine_systems_mass = todo;
    /*TODO: check if engine systems are part of propulsion mass for powerUnitMass, if not add it from systems block*/
    powerUnitMass = EndnodeReadOnly<double>("component_design/propulsion/mass_properties/mass").read(rtIO->acxml).value() - pylon_mass;
    pylon_mass_ratio = pylon_mass / (pylon_mass + nacelle_mass);
    nacelle_mass_ratio = nacelle_mass / (pylon_mass + nacelle_mass);

    /* Empennage masses and ratios */
    double finMass(0.);
    double stabiliserMass(0.);
    node empennageGeometry(rtIO->acxml.at("aircraft_exchange_file/component_design/empennage/specific/geometry/"));
    int numberOfSurfaces(empennageGeometry.getVector("aerodynamic_surface").size());
    for (int i = 0; i < numberOfSurfaces; i++) {
        std::string name(EndnodeReadOnly<std::string>("aerodynamic_surface@" + num2Str(i) +"/name").read(empennageGeometry).value());
        double mass(EndnodeReadOnly<double>("aerodynamic_surface@" + num2Str(i) +"/mass_properties/mass").read(empennageGeometry).value());
        if (name == "vertical_stabiliser")
            finMass += mass;
        else if (name == "horizontal_stabiliser")
            stabiliserMass += mass;
    }
    empennage_mass = finMass + stabiliserMass;
    fin_mass_ratio = finMass / (empennage_mass);
    stabilizer_mass_ratio = stabiliserMass / (empennage_mass);

    /* Airframe mass */
    airframe_mass =  MME - systemsMass - engine_mass - nacelle_mass;
    /* Spare parts masses*/
    spare_parts_mass = {wingMass, fuselageMass, stabiliserMass, finMass, landingGearMass, powerUnitMass};
    /* Main parts masses*/
    main_parts_mass = {wingMass, fuselageMass, stabiliserMass, finMass, landingGearMass, pylon_mass, powerUnitMass};
    /* Transported components masses*/
    transported_component_mass = {wingMass, fuselageMass, stabiliserMass, finMass, landingGearMass, powerUnitMass};

    /* Fuel masses */
    double gravimetric_energy_density = EnergyCarrier("kerosene", rtIO->get_fuel_density(0)).gravimetric_energy_density;
    tripfuel_design = EndnodeReadOnly<double>("analysis/mission/design_mission/trip_energy").read(rtIO->acxml).value()/gravimetric_energy_density;;
    LTO_fuel_design = (EndnodeReadOnly<double>("analysis/mission/design_mission/takeoff_energy").read(rtIO->acxml).value()
                    + EndnodeReadOnly<double>("analysis/mission/design_mission/landing_energy").read(rtIO->acxml).value())/gravimetric_energy_density;
    taxi_fuel_design = EndnodeReadOnly<double>("analysis/mission/design_mission/taxi_energy").read(rtIO->acxml).value()/gravimetric_energy_density;;
    blockfuel_design = taxi_fuel_design + tripfuel_design;
    taxi_fuel_study = EndnodeReadOnly<double>("analysis/mission/study_mission/taxi_energy").read(rtIO->acxml).value()/gravimetric_energy_density;;
    tripfuel_study = EndnodeReadOnly<double>("analysis/mission/study_mission/trip_energy").read(rtIO->acxml).value()/gravimetric_energy_density;;
    blockfuel_study = tripfuel_study + taxi_fuel_study;
}

void SchaeferIOData::read_ecological_assessment_xml(std::shared_ptr<node> ecologicalAssessmentXML) {
    const node mission_emissions(ecologicalAssessmentXML->at("ecological_data_file/mission_emissions"));
    theTestMission.LTO.CO2 = EndnodeReadOnly<double>("design_mission/emissions/LTO_cycle/CO2").read(mission_emissions).value();
    theTestMission.LTO.H2O = EndnodeReadOnly<double>("design_mission/emissions/LTO_cycle/H2O").read(mission_emissions).value();
    theTestMission.LTO.SO2 = EndnodeReadOnly<double>("design_mission/emissions/LTO_cycle/SO2").read(mission_emissions).value();
    theTestMission.LTO.NOx = EndnodeReadOnly<double>("design_mission/emissions/LTO_cycle/NOx").read(mission_emissions).value();
    theTestMission.LTO.HC = EndnodeReadOnly<double>("design_mission/emissions/LTO_cycle/HC").read(mission_emissions).value();
    theTestMission.LTO.CH4 = EndnodeReadOnly<double>("design_mission/emissions/LTO_cycle/CH4").read(mission_emissions).value();
    theTestMission.LTO.CO = EndnodeReadOnly<double>("design_mission/emissions/LTO_cycle/CO").read(mission_emissions).value();
    theTestMission.LTO.soot = EndnodeReadOnly<double>("design_mission/emissions/LTO_cycle/soot").read(mission_emissions).value();
    theTestMission.cruise.CO2 = EndnodeReadOnly<double>("design_mission/emissions/cruise/CO2").read(mission_emissions).value();
    theTestMission.cruise.H2O = EndnodeReadOnly<double>("design_mission/emissions/cruise/H2O").read(mission_emissions).value();
    theTestMission.cruise.SO2 = EndnodeReadOnly<double>("design_mission/emissions/cruise/SO2").read(mission_emissions).value();
    theTestMission.cruise.NOx = EndnodeReadOnly<double>("design_mission/emissions/cruise/NOx").read(mission_emissions).value();
    theTestMission.cruise.HC = EndnodeReadOnly<double>("design_mission/emissions/cruise/HC").read(mission_emissions).value();
    theTestMission.cruise.CH4 = EndnodeReadOnly<double>("design_mission/emissions/cruise/CH4").read(mission_emissions).value();
    theTestMission.cruise.CO = EndnodeReadOnly<double>("design_mission/emissions/cruise/CO").read(mission_emissions).value();
    theTestMission.cruise.soot = EndnodeReadOnly<double>("design_mission/emissions/cruise/soot").read(mission_emissions).value();
    theStudyMission.LTO.CO2 = EndnodeReadOnly<double>("study_mission/emissions/LTO_cycle/CO2").read(mission_emissions).value();
    theStudyMission.LTO.H2O = EndnodeReadOnly<double>("study_mission/emissions/LTO_cycle/H2O").read(mission_emissions).value();
    theStudyMission.LTO.SO2 = EndnodeReadOnly<double>("study_mission/emissions/LTO_cycle/SO2").read(mission_emissions).value();
    theStudyMission.LTO.NOx = EndnodeReadOnly<double>("study_mission/emissions/LTO_cycle/NOx").read(mission_emissions).value();
    theStudyMission.LTO.HC = EndnodeReadOnly<double>("study_mission/emissions/LTO_cycle/HC").read(mission_emissions).value();
    theStudyMission.LTO.CH4 = EndnodeReadOnly<double>("study_mission/emissions/LTO_cycle/CH4").read(mission_emissions).value();
    theStudyMission.LTO.CO = EndnodeReadOnly<double>("study_mission/emissions/LTO_cycle/CO").read(mission_emissions).value();
    theStudyMission.LTO.soot = EndnodeReadOnly<double>("study_mission/emissions/LTO_cycle/soot").read(mission_emissions).value();
    theStudyMission.cruise.CO2 = EndnodeReadOnly<double>("study_mission/emissions/cruise/CO2").read(mission_emissions).value();
    theStudyMission.cruise.H2O = EndnodeReadOnly<double>("study_mission/emissions/cruise/H2O").read(mission_emissions).value();
    theStudyMission.cruise.SO2 = EndnodeReadOnly<double>("study_mission/emissions/cruise/SO2").read(mission_emissions).value();
    theStudyMission.cruise.NOx = EndnodeReadOnly<double>("study_mission/emissions/cruise/NOx").read(mission_emissions).value();
    theStudyMission.cruise.HC = EndnodeReadOnly<double>("study_mission/emissions/cruise/HC").read(mission_emissions).value();
    theStudyMission.cruise.CH4 = EndnodeReadOnly<double>("study_mission/emissions/cruise/CH4").read(mission_emissions).value();
    theStudyMission.cruise.CO = EndnodeReadOnly<double>("study_mission/emissions/cruise/CO").read(mission_emissions).value();
    theStudyMission.cruise.soot = EndnodeReadOnly<double>("study_mission/emissions/cruise/soot").read(mission_emissions).value();
}

/* Output data methods*/
void SchaeferIOData::update_ecological_assessment_xml(std::shared_ptr<node> ecologicalAssessmentXML, const Development& theDevelopment,
                const Production& theProduction, const Operation& theOperation, const EndOfLife& theEndOfLife) {
    myRuntimeInfo->out << "  Update ecological_assessment_results.xml" << std::endl;
    std::string lca_path("ecological_data_file/life_cycle_assessment/");
    std::string sub_path(lca_path + "emissions/development");
    /*Development Block*/
    if (configuration.development_emission_setting != "mode_0") {
        Endnode<double>("CO2", "Total CO2 emissions of development phase",
                        Rounding(theDevelopment.developmentTotal.CO2, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
        Endnode<double>("H2O", "Total H2O emissions of development phase",
                        Rounding(theDevelopment.developmentTotal.H2O, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
        Endnode<double>("NOx", "Total NOx emissions of development phase",
                         Rounding(theDevelopment.developmentTotal.NOx, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
        Endnode<double>("CO", "Total CO emissions of development phase",
                        Rounding(theDevelopment.developmentTotal.CO, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
        Endnode<double>("SO2", "Total SO2 emissions of development phase",
                        Rounding(theDevelopment.developmentTotal.SO2, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
        Endnode<double>("CH4", "Total CH4 emissions of development phase",
                        Rounding(theDevelopment.developmentTotal.CH4, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
        Endnode<double>("PM10", "Total PM10 emissions of development phase",
                        Rounding(theDevelopment.developmentTotal.PM10, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
        Endnode<double>("N2O", "Total N2O emissions of development phase",
                        Rounding(theDevelopment.developmentTotal.N2O, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
        Endnode<double>("PFC", "Total PFC emissions of development phase",
                        Rounding(theDevelopment.developmentTotal.PFC, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
        Endnode<double>("NMVOC", "Total NMVOC emissions of development phase",
                        Rounding(theDevelopment.developmentTotal.NMVOC, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
        Endnode<double>("HC", "Total HC emissions of development phase",
                Rounding(theDevelopment.developmentTotal.HC, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
        ecologicalAssessmentXML->at(sub_path).addAttrib("description", "Development emissions");
    }

    /*Production Block*/
    sub_path = lca_path + "emissions/production";
    Endnode<double>("CO2", "Total CO2 emissions of production phase",
                    Rounding(theProduction.productionTotal.CO2, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("H2O", "Total H2O emissions of production phase",
                    Rounding(theProduction.productionTotal.H2O, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("NOx", "Total NOx emissions of production phase",
                    Rounding(theProduction.productionTotal.NOx, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("CO", "Total CO emissions of production phase",
                    Rounding(theProduction.productionTotal.CO, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("SO2", "Total SO2 emissions of production phase",
                    Rounding(theProduction.productionTotal.SO2, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("CH4", "Total CH4 emissions of production phase",
                    Rounding(theProduction.productionTotal.CH4, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("PM10", "Total PM10 emissions of production phase",
                    Rounding(theProduction.productionTotal.PM10, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("N2O", "Total N2O emissions of production phase",
                    Rounding(theProduction.productionTotal.N2O, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("PFC", "Total PFC emissions of production phase",
                    Rounding(theProduction.productionTotal.PFC, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("NMVOC", "Total NMVOC emissions of production phase",
                    Rounding(theProduction.productionTotal.NMVOC, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("HC", "Total HC emissions of production phase",
                    Rounding(theProduction.productionTotal.HC, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    ecologicalAssessmentXML->at(sub_path).addAttrib("description", "Production emissions");

    /* Operation Block */
    sub_path = lca_path + "emissions/operation";
    Endnode<double>("CO2", "Total CO2 emissions of operation phase",
                    Rounding(theOperation.operationTotal.CO2, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("H2O", "Total H2O emissions of operation phase",
                    Rounding(theOperation.operationTotal.H2O, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("NOx", "Total NOx emissions of operation phase",
                    Rounding(theOperation.operationTotal.NOx, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("CO", "Total CO emissions of operation phase",
                    Rounding(theOperation.operationTotal.CO, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("SO2", "Total SO2 emissions of operation phase",
                    Rounding(theOperation.operationTotal.SO2, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("CH4", "Total CH4 emissions of operation phase",
                    Rounding(theOperation.operationTotal.CH4, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("PM10", "Total PM10 emissions of operation phase",
                    Rounding(theOperation.operationTotal.PM10, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("N2O", "Total N2O emissions of operation phase",
                    Rounding(theOperation.operationTotal.N2O, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("PFC", "Total PFC emissions of operation phase",
                    Rounding(theOperation.operationTotal.PFC, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("NMVOC", "Total NMVOC emissions of operation phase",
                    Rounding(theOperation.operationTotal.NMVOC, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("HC", "Total HC emissions of operation phase",
                    Rounding(theOperation.operationTotal.HC, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    ecologicalAssessmentXML->at(sub_path).addAttrib("description", "Operation emissions");

    /*End Of Life Block*/
    sub_path = lca_path + "emissions/end_of_life";
    Endnode<double>("CO2", "Total CO2 emissions of end-of-life phase",
                    Rounding(theEndOfLife.endOfLifeTotal.CO2, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("H2O", "Total H2O emissions of end-of-life phase",
                    Rounding(theEndOfLife.endOfLifeTotal.H2O, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("NOx", "Total NOx emissions of end-of-life phase",
                    Rounding(theEndOfLife.endOfLifeTotal.NOx, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("CO", "Total CO emissions of end-of-life phase",
                    Rounding(theEndOfLife.endOfLifeTotal.CO, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("SO2", "Total SO2 emissions of end-of-life phase",
                    Rounding(theEndOfLife.endOfLifeTotal.SO2, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("CH4", "Total CH4 emissions of end-of-life phase",
                    Rounding(theEndOfLife.endOfLifeTotal.CH4, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("PM10", "Total PM10 emissions of end-of-life phase",
                    Rounding(theEndOfLife.endOfLifeTotal.PM10, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("N2O", "Total N2O emissions of end-of-life phase",
                    Rounding(theEndOfLife.endOfLifeTotal.N2O, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("PFC", "Total PFC emissions of end-of-life phase",
                    Rounding(theEndOfLife.endOfLifeTotal.PFC, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("NMVOC", "Total NMVOC emissions of end-of-life phase",
                    Rounding(theEndOfLife.endOfLifeTotal.NMVOC, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("HC", "Total HC emissions of end-of-life phase",
                    Rounding(theEndOfLife.endOfLifeTotal.HC, 3), "kg").update((*ecologicalAssessmentXML)[sub_path]);
    ecologicalAssessmentXML->at(sub_path).addAttrib("description", "End of life emissions");

    /*CED*/
    sub_path = lca_path + "cumulative_energy_demand/";
    double CEDtotal = theDevelopment.developmentTotal.CED + theProduction.productionTotal.CED +
                    theOperation.operationTotal.CED + theEndOfLife.endOfLifeTotal.CED;
    Endnode<double>("total", "Total cumulative energy demand", Rounding(CEDtotal*pow(10, 6), 3), "J").update((*ecologicalAssessmentXML)[sub_path]);
    if (configuration.development_emission_setting != "mode_0") {
        Endnode<double>("development", "Total cumulative energy demand of development phase",
                        Rounding(theDevelopment.developmentTotal.CED*pow(10, 6), 3), "J").update((*ecologicalAssessmentXML)[sub_path]);
    }
    Endnode<double>("production", "Total cumulative energy demand of production phase",
                        Rounding(theProduction.productionTotal.CED*pow(10, 6), 3), "J").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("operation", "Total cumulative energy demand of operation phase",
                        Rounding(theOperation.operationTotal.CED*pow(10, 6), 3), "J").update((*ecologicalAssessmentXML)[sub_path]);
    Endnode<double>("end_of_life", "Total cumulative energy demand of end-of-life phase",
                        Rounding(theEndOfLife.endOfLifeTotal.CED*pow(10, 6), 3), "J").update((*ecologicalAssessmentXML)[sub_path]);
    ecologicalAssessmentXML->at(sub_path).addAttrib("description", "Cumulative enrrgy demand of life cycle");

    ecologicalAssessmentXML->at(lca_path).addAttrib("description", "Results of life cycle assessment");
}

