/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_EMISSION_CALCULATION_LIFE_CYCLE_EMISSIONS_LCA_SCHAEFER_SCHAEFERIODATA_H_
#define ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_EMISSION_CALCULATION_LIFE_CYCLE_EMISSIONS_LCA_SCHAEFER_SCHAEFERIODATA_H_

#include <string>
#include <unordered_map>
#include <utility>
#include <vector>
#include <array>
#include <memory>

#include <moduleBasics/runtimeIO.h>
#include <aixml/node.h>
#include <engine/engine.h>

#include "../../../emissionsClass.h"
#include "schaeferConfig.h"
#include "development.h"
#include "production.h"
#include "operation.h"
#include "endOfLife.h"

class Development;
class Production;
class Operation;
class EndOfLife;

// TODO(franz#1#): T126 check if sum of components = 100%
/** \class SchaeferIOData
 * \brief Class for the settings that are needed in the schaefer emissions strategy.
 */
class SchaeferIOData {
 public:
    const std::string spare_parts_ATA[7] = { /**< ATA chapters which should be exchanged during the operation phase */
        "ATA57", "Fuselage", "ATA55", "ATA55", "ATA32", "ATA54", "PowerUnit"};
    const std::string spare_parts_name[7] = { /**< Name of the spare part (needs to be the same as main parts right now)*/
        "Wing", "Fuselage", "Tailplane", "Fin", "LandingGear", "Pylons", "PowerUnit"};
    const std::string spare_parts_ATA_energy_unit[3] = { /**< ATA chapter of spare parts for power unit*/
        "ATA28", "ATA36", "ATA54"};//Belong to Engine Maintenance Cost because in PowerUnit RC with in it, attention: ATA54 Pylon+Nacelles -> weight ratio
    const std::string spare_parts_ATA_airframe[4] = { /**< ATA chapter of spare parts for airframe*/
        "ATA51", "ATA52", "ATA53", "ATA56"};//Fuselage Maintenance Cost because in Fuselage RC implicitly included (only ATA not separately mentioned in RC Systems etc.)
    const std::string main_parts_name[7] = {"Wing", "Fuselage", "Tailplane", "Fin", "LandingGear", "Pylons", "PowerUnit"};/**< Names of the aircraft's main parts */
    const std::string material_name[5] = {"Al", "CFRP", "St", "Ti", "Ni"};/**< Name of used materials*/
    const std::string available_mean_of_transportation[4] = {"Aircraft", "Truck", "Ship", "Train"};/**< Name of the means of transportation*/
    const std::string transported_component_name[6] = {/**< Name of the transported component (if component is also contained in main_parts_name, the name must be the same here)*/
        "Wing", "Fuselage", "Tailplane", "Fin", "LandingGear", "PowerUnit"};
    const std::string chosen_mean_of_transportation[6] = {"Aircraft", "Aircraft", "Aircraft", "Aircraft", "Truck", "Aircraft"};/**< Chosen mean of transportation*/

    SchaeferConfig configuration; /**< Object of class schaeferConfig which contains data from configuration file*/
    Engine myEngine; /**< Object of class engine that contains engine data */

    /** \class mission
     * \brief Class for the emissions of one mission
     */
    class mission {
     public:
        Emissions LTO; /**< LTO emissions of mission [kg] */
        Emissions cruise; /**< Cruise emissions of mission [kg] */
        /** \brief Constructor of class mission */
        mission() {}
        /** \brief Destructor of class mission */
        virtual ~mission() {}
    };
    mission theTestMission; /**< Object of class mission which contains the test mission emissions */
    mission theStudyMission; /**< Object of class mission which contains the study mission emissions */

    /*Requirements and specifications*/
    double max_operating_speed;/**<Maximum operating speed (maximum dynamic pressure)[m/s]*/
    int flights_per_year;/**< Number of flights per year [-]*/
    int duration_operation; /**<Duration of the operating phase [yr]*/
    double available_seats; /**< Total number of available seats [-] */
    /*Aerodynamics*/
    double wing_reference_area;/**< Reference area of main wing [m^2]*/
    /* Masses */
    double mtom; /**< Maximum take-off mass [kg] */
    double ome; /**< Operating empty mass [kg] */
    std::vector<double> spare_parts_mass;/**< Vector contains masses of spare parts wing, fuselage, stabiliser, fin, landing gear and power unit [kg]*/
    std::vector<double> main_parts_mass;/**< Vector contains masses of main parts wing, fuselage, stabiliser, fin, landing gear, pylon and power unit [kg]*/
    std::vector<double> transported_component_mass;/**< Vector contains masses of transported component wing, fuselage, stabiliser, fin, landing gear and power unit [kg]*/
    double airframe_mass; /**< Mass of the airframe components [kg]*/
    double empennage_mass; /**< Mass of empennage [kg] */
    double fin_mass_ratio;/**< Ratio of fin mass to empennage mass [-] */
    double stabilizer_mass_ratio;/**< Ratio stabilizer mass to empennage mass [-] */
    double pylon_mass_ratio;/**< Ratio pylon mass to pylon+nacelle mass [-]*/
    double nacelle_mass_ratio;/**< Ratio pylon mass to pylon+nacelle mass [-]*/
    double engine_mass; /** Mass of all engines [kg] */
    double nacelle_mass;/**< Mass of all nacelles [kg]*/
    double pylon_mass;/**< Mass of all pylons [kg]*/
    double bleed_air_system_mass;/**< Mass of overall bleed air system [kg]*/
    double fuel_system_mass;/**< Mass of overall fuel system [kg]*/
    double other_engine_systems_mass;/**< Mass of other engine system [kg]*/

    /*Propulsion*/
    int number_of_engines; /**< Number of engines [-] */
    double fuel_flow_cont_thrust; /**< Fuel flow at continuous thrust at cruise for one scaled engine [kg/s]*/

    /*Systems*/
    double propulsion_design_power;/**< Propulsion System: Total energy demand [kW]*/

    /*Performance Design Mission*/
    double range_design;/**< Range Design Mission [m] */
    double blocktime_design;/**< Block time of design mission [h]*/
    double flight_time_design;/**< Flight time of design mission [h]*/
    double taxi_fuel_design;/**< Total taxi fuel for design mission [kg]*/
    double tripfuel_design;/**< Trip fuel of design mission [kg]*/
    double blockfuel_design;/**< Block fuel of design mission [kg]*/
    double LTO_fuel_design;/**< Fuel mass of design mission for landing and take off [kg]*/
    double taxi_time;/** Total taxi time of design mission [s]*/

    /*Performance Study Mission*/
    double range_study; /**< Range Study Mission [m] */
    double flight_time_study;/**< Flight time of study mission [s]*/
    double taxi_fuel_study;/**< Taxi fuel of study mission before takeoff [kg]*/
    double tripfuel_study;/**< Trip fuel of study mission [kg]*/
    double study_mission_PAX;/**< Number of PAX for study mission [-]*/
    double blockfuel_study;/**< Block fuel of study mission [kg]*/


    /** \brief Construct a new schaefer IO data object
     * \param rtIO Access to rtIO
     * \param ecoAssXML Access to ecological assessment results xml
     */
    explicit SchaeferIOData(std::shared_ptr<RuntimeIO> rtIO, std::shared_ptr<node> ecoAssXML);

    /** \brief Destructor of class SchaeferIOData
     */
    virtual ~SchaeferIOData() {}

    /** \brief Function reads all input data from aircraft exchange file
     * \param rtIO Access to rtIO
     */
    void read_aircraft(std::shared_ptr<RuntimeIO> rtIO);

    /** \brief Function reads masses, fills mass vectors and calculate mass ratios
     * \param rtIO Access to rtIO
     */
    void get_masses(std::shared_ptr<RuntimeIO> rtIO);

    /** \brief Function reads data from mision xml file
     * \param rtIO Access to rtIO
     */
    void get_mission_file_data(std::shared_ptr<RuntimeIO> rtIO);

    /** \brief Function reads emissions from ecological_assessment_results.xml
     * \param ecologicalAssessmentXML Access to ecological_assessment_results.xml data
     */
    void read_ecological_assessment_xml(std::shared_ptr<node> ecologicalAssessmentXML);

    /** \brief Function writes results to ecological_assessment_results.xml
     * \param ecologicalAssessmentXML Access to ecological_assessment_results.xml data
     * \param theDevelopment Access to development results
     * \param theProduction Access to production results
     * \param theOperation Accesss to operation results
     * \param theEndOfLife Access to end of life results
     */
    void update_ecological_assessment_xml(std::shared_ptr<node> ecologicalAssessmentXML, const Development& theDevelopment,
                    const Production& theProduction, const Operation& theOperation, const EndOfLife& theEndOfLife);
};

#endif // ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_EMISSION_CALCULATION_LIFE_CYCLE_EMISSIONS_LCA_SCHAEFER_SCHAEFERIODATA_H_
