/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include <string>
#include "../../../standard_strategy.h"
#include "../../../../toolinfo.h"
#include <moduleBasics/plot.h>
#include <moduleBasics/colors.h>

void StandardStrategy::generate_plots_schaefer() {
     std::string pltName = std::string(TOOL_NAME) + "_life_cycle_emissions";
     this->generate_life_cycle_emissions_plot_schaefer(pltName);
}

void StandardStrategy::generate_life_cycle_emissions_plot_schaefer(const std::string& pltName) {
     std::vector<double> development_emissions = {
          theLifeCycleEmissionsSchaefer.theDevelopment.developmentTotal.CO2,
          theLifeCycleEmissionsSchaefer.theDevelopment.developmentTotal.H2O,
          theLifeCycleEmissionsSchaefer.theDevelopment.developmentTotal.NOx,
          theLifeCycleEmissionsSchaefer.theDevelopment.developmentTotal.CO,
          theLifeCycleEmissionsSchaefer.theDevelopment.developmentTotal.SO2,
          theLifeCycleEmissionsSchaefer.theDevelopment.developmentTotal.CH4,
          theLifeCycleEmissionsSchaefer.theDevelopment.developmentTotal.PM10,
          theLifeCycleEmissionsSchaefer.theDevelopment.developmentTotal.N2O,
          theLifeCycleEmissionsSchaefer.theDevelopment.developmentTotal.PFC,
          theLifeCycleEmissionsSchaefer.theDevelopment.developmentTotal.NMVOC,
          theLifeCycleEmissionsSchaefer.theDevelopment.developmentTotal.HC
     };
     if (theLifeCycleEmissionsSchaefer.schaeferIOData->configuration.number_produced_aircraft != 0) {
          for (uint8_t i = 0; i < development_emissions.size(); i++) {
            development_emissions[i] /= theLifeCycleEmissionsSchaefer.schaeferIOData->configuration.number_produced_aircraft; //scale to one aircraft
          }
    } else {
          myRuntimeInfo->warn << "Number of produced aircraft is zero. Full development emissions will be shown in LCA plot" << std::endl;
    }
     std::vector<double> production_emissions = {
          theLifeCycleEmissionsSchaefer.theProduction.productionTotal.CO2,
          theLifeCycleEmissionsSchaefer.theProduction.productionTotal.H2O,
          theLifeCycleEmissionsSchaefer.theProduction.productionTotal.NOx,
          theLifeCycleEmissionsSchaefer.theProduction.productionTotal.CO,
          theLifeCycleEmissionsSchaefer.theProduction.productionTotal.SO2,
          theLifeCycleEmissionsSchaefer.theProduction.productionTotal.CH4,
          theLifeCycleEmissionsSchaefer.theProduction.productionTotal.PM10,
          theLifeCycleEmissionsSchaefer.theProduction.productionTotal.N2O,
          theLifeCycleEmissionsSchaefer.theProduction.productionTotal.PFC,
          theLifeCycleEmissionsSchaefer.theProduction.productionTotal.NMVOC,
          theLifeCycleEmissionsSchaefer.theProduction.productionTotal.HC
     };
     std::vector<double> operation_emissions = {
          theLifeCycleEmissionsSchaefer.theOperation.operationTotal.CO2,
          theLifeCycleEmissionsSchaefer.theOperation.operationTotal.H2O,
          theLifeCycleEmissionsSchaefer.theOperation.operationTotal.NOx,
          theLifeCycleEmissionsSchaefer.theOperation.operationTotal.CO,
          theLifeCycleEmissionsSchaefer.theOperation.operationTotal.SO2,
          theLifeCycleEmissionsSchaefer.theOperation.operationTotal.CH4,
          theLifeCycleEmissionsSchaefer.theOperation.operationTotal.PM10,
          theLifeCycleEmissionsSchaefer.theOperation.operationTotal.N2O,
          theLifeCycleEmissionsSchaefer.theOperation.operationTotal.PFC,
          theLifeCycleEmissionsSchaefer.theOperation.operationTotal.NMVOC,
          theLifeCycleEmissionsSchaefer.theOperation.operationTotal.HC
     };
     std::vector<double> eol_emissions = {
          theLifeCycleEmissionsSchaefer.theEndOfLife.endOfLifeTotal.CO2,
          theLifeCycleEmissionsSchaefer.theEndOfLife.endOfLifeTotal.H2O,
          theLifeCycleEmissionsSchaefer.theEndOfLife.endOfLifeTotal.NOx,
          theLifeCycleEmissionsSchaefer.theEndOfLife.endOfLifeTotal.CO,
          theLifeCycleEmissionsSchaefer.theEndOfLife.endOfLifeTotal.SO2,
          theLifeCycleEmissionsSchaefer.theEndOfLife.endOfLifeTotal.CH4,
          theLifeCycleEmissionsSchaefer.theEndOfLife.endOfLifeTotal.PM10,
          theLifeCycleEmissionsSchaefer.theEndOfLife.endOfLifeTotal.N2O,
          theLifeCycleEmissionsSchaefer.theEndOfLife.endOfLifeTotal.PFC,
          theLifeCycleEmissionsSchaefer.theEndOfLife.endOfLifeTotal.NMVOC,
          theLifeCycleEmissionsSchaefer.theEndOfLife.endOfLifeTotal.HC
     };
     std::vector<std::string> emissions = {"CO2", "H2O", "NOx", "CO", "SO2", "CH4", "PM10", "N2O", "PFC", "NMVOC", "HC"};
     matplot::figure_handle fig = matplot::figure(true);
     matplot::axes_handle ax = fig->current_axes();
     ax->colororder(colors::unicado_to_matplot_colororder());
     std::vector<std::vector<double>> emissions_data = {development_emissions, production_emissions, operation_emissions, eol_emissions};
     ax->bar(emissions_data);
     ax->title("Life Cycle Emissions (Correct order of bars: Development, Production, Operation, End of Life)");
     ax->ylabel("Emissions [kg]");
     ax->x_axis().ticklabels(emissions);
     ax->y_axis().scale(matplot::axis_type::axis_scale::log);
     auto legend = matplot::legend(ax, {"Development", "Production", "Operation", "End of Life"});// \todo Correct legend
     legend->location(matplot::legend::general_alignment::topright);
     std::string plotFilePath = rtIO->getPlotDir() + "/" + pltName + ".svg";
     matplot::save(plotFilePath);
     matplot::cla();
}
