/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "mission.h"
#include <algorithm>
#include <numeric>
#include <fstream>
#include <iomanip>
#include <sstream>
#include <memory>
#include <unordered_map>
#include <aixml/node.h>
#include <runtimeInfo/runtimeInfo.h>
#include <standardFiles/functions.h>
#include <unitConversion/unitConversion.h>
#include <moduleBasics/runtimeIO.h>


// TODO(franz#1#): emission calculation at derate correct? Check! (Problem: Base curves with EI_ICAO (without Derate))
Mission::Mission()
    :
    mission_csv_name(""),
    delta_isa(0.),
    myEcoDatabase(ecoDatabase()),
    mission_range(0.),
    cruise_number_of_steps(0),
    fuel_mass_taxi(0.),
    fuel_mass_takeoff(0.),
    fuel_mass_climbout(0.),
    fuel_mass_approach(0.),
    fuel_mass_taxi_ICAO(0.),
    fuel_mass_takeoff_ICAO(0.),
    fuel_mass_climbout_ICAO(0.),
    fuel_mass_approach_ICAO(0.),
    fuel_mass_LTO(0.),
    fuel_mass_cruise(0.),
    taxi_time(0.),
    takeoff_time(0.),
    climbout_time(0.),
    approach_time(0.),
    LTO_time(0.),
    thrust_taxi(0.),
    thrust_takeoff(0.),
    thrust_climbout(0.),
    thrust_approach(0.),
    T3_ID(0.),
    T3_TO(0.),
    T3_CL(0.),
    T3_AP(0.),
    P3_ID(0.),
    P3_TO(0.),
    P3_CL(0.),
    P3_AP(0.),
    P3t_ID(0.),
    P3t_TO(0.),
    P3t_CL(0.),
    P3t_AP(0.),
    St3Mach_ID(0.),
    St3Mach_TO(0.),
    St3Mach_CL(0.),
    St3Mach_AP(0.),
    St3MassFlow_ID(0.),
    St3MassFlow_TO(0.),
    St3MassFlow_CL(0.),
    St3MassFlow_AP(0.),
    fuelFlow_ID(0.),
    fuelFlow_TO(0.),
    fuelFlow_CL(0.),
    fuelFlow_AP(0.),
    number_of_steps(0.),
    time_step(0.),
    flights_per_lifecycle(0.),
    c_soot_LTO_max(0.) {
}

void Mission::run(const std::string& missionType, const std::shared_ptr<RuntimeIO>& rtIO, std::shared_ptr<node> ecoAssXML) {
    missionIOData = std::make_shared<MissionIOData>(rtIO);
    myRuntimeInfo->out << "  Calculation of " << missionType << " mission emissions" << std::endl;
    calculate_mission(missionType, rtIO);
    missionIOData->update_ecological_assessment_xml(ecoAssXML, *this);
}

void Mission::calculate_mission(const std::string& missionType, const std::shared_ptr<RuntimeIO>& rtIO) {
    rtIOPt = rtIO;
    mission_type = missionType;
    /*Read mission data*/
    get_mission_data();
    myFlightPath = flightPath();
    for (unsigned int i(0); i < missionInfo.size(); i++) {
        myFlightPath.pathPoints.push_back(flightPathPoint(missionInfo.at(i).time,
                                          missionInfo.at(i).range,
                                          missionInfo.at(i).altitude,
                                          missionInfo.at(i).TAS_ms,
                                          missionInfo.at(i).Mach,
                                          missionInfo.at(i).Tamb,
                                          missionInfo.at(i).Pamb,
                                          missionInfo.at(i).fuelConsumed,
                                          missionInfo.at(i).fuelFlow,
                                          missionInfo.at(i).NOxFlow,
                                          missionInfo.at(i).N1,
                                          missionInfo.at(i).St3MassFlow,
                                          missionInfo.at(i).St3TotalTemperature,
                                          missionInfo.at(i).St3StaticTemperature,
                                          missionInfo.at(i).St3TotalPressure,
                                          missionInfo.at(i).St3Mach,
                                          missionInfo.at(i).St4FuelAir,
                                          missionInfo.at(i).BPR));
}
    /* Create emissionspath.csv*/
    write_emissions_path_csv();
    /* Determination of mission segments data*/
    get_mission_segment_data();
    /* Values which are used for ICAO database measurements */
    get_engine_thermodynamics_LTO();
    /* Set mission/aircraft parameter */
    number_of_steps =  this->myFlightPath.pathPoints.size();
    LTO_time = (taxi_time + takeoff_time + climbout_time + approach_time) / 60.;
    fuel_mass_taxi_ICAO = missionIOData->number_of_engines * (missionIOData->myEngine.get_LTO_fuelflow(LTOPhases::taxi) * myEcoDatabase.taxi_time_ICAO);
    fuel_mass_takeoff_ICAO = missionIOData->number_of_engines * (missionIOData->myEngine.get_LTO_fuelflow(LTOPhases::takeoff) * myEcoDatabase.takeoff_time_ICAO);
    fuel_mass_climbout_ICAO = missionIOData->number_of_engines * (missionIOData->myEngine.get_LTO_fuelflow(LTOPhases::climb) * myEcoDatabase.climbout_time_ICAO);
    fuel_mass_approach_ICAO = missionIOData->number_of_engines * (missionIOData->myEngine.get_LTO_fuelflow(LTOPhases::approach) * myEcoDatabase.appTime_ICAO);
    flights_per_lifecycle = missionIOData->duration_operation * missionIOData->flights_per_year;
    /* Calculate emissions */
    if (rtIO->aircraft_energy_carrier_type() == "kerosene") {
        calculate_kerosene_emissions();
    } else if (rtIO->aircraft_energy_carrier_type() == "liquid_hydrogen") {
        calculate_hydrogen_combustion_emissions();
    } else {
        myRuntimeInfo->err << "No calculation method implemented for fuel type " << rtIO->aircraft_energy_carrier_type() << std::endl;
        exit(1);
    }
    /*Compare calculated LTO values with ICAO data*/
    LTO_output_comparison();
}

void Mission::LTO_output_comparison() {
    /* LTO Output Comparison ICAO vs. MissionAnalysis*/
    myRuntimeInfo->info << "Comparison ICAO vs MissionAnalysis:" << std::endl;
    myRuntimeInfo->info << " Taxi" << std::endl;
    myRuntimeInfo->info << "   Fuel (total) Mission [kg]:" << fuel_mass_taxi << std::endl;
    myRuntimeInfo->info << "   Fuel Flow (total) Mission [kg/s]:" << fuel_mass_taxi / taxi_time << std::endl;
    myRuntimeInfo->info << "   Fuel Flow (total) ICAO  [kg/s]:" << fuel_mass_taxi_ICAO / myEcoDatabase.taxi_time_ICAO << std::endl;
    myRuntimeInfo->info << "   time Mission  [min]:" << taxi_time / 60. << std::endl;
    myRuntimeInfo->info << "   time ICAO  [min]:" << myEcoDatabase.taxi_time_ICAO / 60. << std::endl;
    myRuntimeInfo->info << "   P3  [kPa]:" << P3_ID << std::endl;
    myRuntimeInfo->info << "   T3  [K]:" << T3_ID << std::endl;
    myRuntimeInfo->info << " T/O" << std::endl;
    myRuntimeInfo->info << "   Fuel (total) Mission [kg]:" << fuel_mass_takeoff << std::endl;
    myRuntimeInfo->info << "   Fuel Flow (total) Mission [kg/s]:" << fuel_mass_takeoff / takeoff_time << std::endl;
    myRuntimeInfo->info << "   Fuel Flow (total) ICAO  [kg/s]:" << fuel_mass_takeoff_ICAO / myEcoDatabase.takeoff_time_ICAO << std::endl;
    myRuntimeInfo->info << "   time Mission  [min]:" << takeoff_time / 60. << std::endl;
    myRuntimeInfo->info << "   time ICAO  [min]:" << myEcoDatabase.takeoff_time_ICAO / 60. << std::endl;
    myRuntimeInfo->info << "   P3  [kPa]:" << P3_TO << std::endl;
    myRuntimeInfo->info << "   T3  [K]:" << T3_TO << std::endl;
    myRuntimeInfo->info << " Climbinfo" << std::endl;
    myRuntimeInfo->info << "   Fuel (total) Mission [kg]:" << fuel_mass_climbout << std::endl;
    myRuntimeInfo->info << "   Fuel Flow (total) Mission [kg/s]:" << fuel_mass_climbout / climbout_time << std::endl;
    myRuntimeInfo->info << "   Fuel Flow (total) ICAO  [kg/s]:" << fuel_mass_climbout_ICAO / myEcoDatabase.climbout_time_ICAO << std::endl;
    myRuntimeInfo->info << "   time Mission  [min]:" << climbout_time / 60. << std::endl;
    myRuntimeInfo->info << "   time ICAO  [min]:" << myEcoDatabase.climbout_time_ICAO / 60. << std::endl;
    myRuntimeInfo->info << "   P3  [kPa]:" << P3_CL << std::endl;
    myRuntimeInfo->info << "   T3  [K]:" << T3_CL << std::endl;
    myRuntimeInfo->info << " Approach" << std::endl;
    myRuntimeInfo->info << "   Fuel (total) Mission [kg]:" << fuel_mass_approach << std::endl;
    myRuntimeInfo->info << "   Fuel Flow (total) Mission [kg/s]:" << fuel_mass_approach / approach_time << std::endl;
    myRuntimeInfo->info << "   Fuel Flow (total) ICAO  [kg/s]:" << fuel_mass_approach_ICAO / myEcoDatabase.appTime_ICAO << std::endl;
    myRuntimeInfo->info << "   time Mission  [min]:" << approach_time / 60. << std::endl;
    myRuntimeInfo->info << "   time ICAO  [min]:" << myEcoDatabase.appTime_ICAO / 60. << std::endl;
    myRuntimeInfo->info << "   P3  [kPa]:" << P3_AP << std::endl;
    myRuntimeInfo->info << "   T3  [K]:" << T3_AP << std::endl;
}

void Mission::get_engine_thermodynamics_LTO() {
    missionIOData->myEngine.calculate_N1_with_penalties(0., 0., myAtmosphere, 1., "takeoff", 0., 0.);
    //for all engines together; max. takeoff Thrust SL and ISA (not ISA+15!)(see Ruijgrok, p. 298 and ICAO 2008 p. III-2-2
    double maximum_takeoff_thrust_sealevel = missionIOData->myEngine.get_thrust_aircraft();
    //Taxi
    thrust_taxi = 0.07 * maximum_takeoff_thrust_sealevel;//for all engines together!
    //Takeoff
    thrust_takeoff = 1. * maximum_takeoff_thrust_sealevel;//for all engines together!
    //ClimbOut
    thrust_climbout = 0.85 * maximum_takeoff_thrust_sealevel;//for all engines together!
    //Approach
    thrust_approach = 0.3 * maximum_takeoff_thrust_sealevel;//for all engines!

    /* Set engine parameter for ICAO database measurements*/
    // Taxi (Idle)
    enum thrustLimit{DERATED = 2, LOWERBOUNDLIMITED, UPPERBOUNDLIMITED};
    try {
        missionIOData->myEngine.calculate_N1_with_thrustlimit(0., 0., myAtmosphere, 1., "takeoff", 0., 0., thrust_taxi); // SI unit
    } catch (const int& engineStatus) {
        switch (engineStatus) {
        case DERATED:
            myRuntimeInfo->warn << "required \"thrust_taxi\" violates thrustLimit from engine deck!" << std::endl;
            myRuntimeInfo->warn << "Engine is derated." << std::endl;
            myRuntimeInfo->warn << "thrust_taxi        = " << thrust_taxi << " N" << std::endl;
            myRuntimeInfo->warn << "engineThrustLimit = " << missionIOData->myEngine.get_thrust_aircraft() << " N" << std::endl;
            myRuntimeInfo->warn << "Engine rating set to idle." << std::endl;
            missionIOData->myEngine.calculate_N1_with_penalties(0., 0., myAtmosphere, 1., "idle", 0., 0.);
        case LOWERBOUNDLIMITED:
            myRuntimeInfo->warn << "required \"thrust_taxi\" violates thrustLimit from engine deck!" << std::endl;
            myRuntimeInfo->warn << "Reached lower speed limit." << std::endl;
            myRuntimeInfo->warn << "thrust_taxi        = " << thrust_taxi << " N" << std::endl;
            myRuntimeInfo->warn << "engineThrustLimit = " << missionIOData->myEngine.get_thrust_aircraft() << " N" << std::endl;
            myRuntimeInfo->warn << "Engine rating set to idle." << std::endl;
            missionIOData->myEngine.calculate_N1_with_penalties(0., 0., myAtmosphere, 1., "idle", 0., 0.);
        case UPPERBOUNDLIMITED:
            myRuntimeInfo->warn << "required \"thrust_taxi\" violates thrustLimit from engine deck!" << std::endl;
            myRuntimeInfo->warn << "Reached upper speed limit." << std::endl;
            myRuntimeInfo->warn << "thrust_taxi        = " << thrust_taxi << " N" << std::endl;
            myRuntimeInfo->warn << "engineThrustLimit = " << missionIOData->myEngine.get_thrust_aircraft() << " N" << std::endl;
            myRuntimeInfo->warn << "Engine rating set to idle." << std::endl;
            missionIOData->myEngine.calculate_N1_with_penalties(0., 0., myAtmosphere, 1., "idle", 0., 0.);
        default:
            break;
        }
    } catch (...) {
        missionIOData->myEngine.calculate_N1_with_penalties(0., 0., myAtmosphere, 1., "idle", 0., 0.);
        myRuntimeInfo->out << "An unknown error has occured during N1 calculation." << std::endl;
    }
    T3_ID = missionIOData->myEngine.get_physical_properties_stage(StageProperties::StaticTemperature, EngineStage::St3);
    P3t_ID = missionIOData->myEngine.get_physical_properties_stage(StageProperties::TotalPressure, EngineStage::St3);
    St3Mach_ID = missionIOData->myEngine.get_physical_properties_stage(StageProperties::MachNumber, EngineStage::St3);
    P3_ID = P3t_ID / (1. + 1.4 / 2.*pow(St3Mach_ID, 2));
    St3MassFlow_ID = missionIOData->myEngine.get_physical_properties_stage(StageProperties::MassFlow, EngineStage::St3);
    fuelFlow_ID = missionIOData->myEngine.get_aircraft_fuelflow();
    // Take off
    missionIOData->myEngine.calculate_N1_with_thrustlimit(0., 0., myAtmosphere, 1., "takeoff", 0., 0., thrust_takeoff);
    T3_TO = missionIOData->myEngine.get_physical_properties_stage(StageProperties::StaticTemperature, EngineStage::St3);
    P3t_TO = missionIOData->myEngine.get_physical_properties_stage(StageProperties::TotalPressure, EngineStage::St3);
    St3Mach_TO = missionIOData->myEngine.get_physical_properties_stage(StageProperties::MachNumber, EngineStage::St3);
    P3_TO = P3t_TO / (1. + 1.4 / 2.*pow(St3Mach_TO, 2));
    St3MassFlow_TO = missionIOData->myEngine.get_physical_properties_stage(StageProperties::MassFlow, EngineStage::St3);
    fuelFlow_TO = missionIOData->myEngine.get_aircraft_fuelflow();
    // Climb
    missionIOData->myEngine.calculate_N1_with_thrustlimit(0., 0., myAtmosphere, 1., "takeoff", 0., 0., thrust_climbout);
    T3_CL = missionIOData->myEngine.get_physical_properties_stage(StageProperties::StaticTemperature, EngineStage::St3);
    P3t_CL = missionIOData->myEngine.get_physical_properties_stage(StageProperties::TotalPressure, EngineStage::St3);
    St3Mach_CL = missionIOData->myEngine.get_physical_properties_stage(StageProperties::MachNumber, EngineStage::St3);
    P3_CL = P3t_CL / (1. + 1.4 / 2.*pow(St3Mach_CL, 2));
    St3MassFlow_CL = missionIOData->myEngine.get_physical_properties_stage(StageProperties::MassFlow, EngineStage::St3);
    fuelFlow_CL = missionIOData->myEngine.get_aircraft_fuelflow();
    // Approach
    missionIOData->myEngine.calculate_N1_with_thrustlimit(0., 0., myAtmosphere, 1., "takeoff", 0., 0., thrust_approach);
    T3_AP = missionIOData->myEngine.get_physical_properties_stage(StageProperties::StaticTemperature, EngineStage::St3);
    P3t_AP = missionIOData->myEngine.get_physical_properties_stage(StageProperties::TotalPressure, EngineStage::St3);
    St3Mach_AP = missionIOData->myEngine.get_physical_properties_stage(StageProperties::MachNumber, EngineStage::St3);
    P3_AP = P3t_AP / (1. + 1.4 / 2.*pow(St3Mach_AP, 2));
    St3MassFlow_AP = missionIOData->myEngine.get_physical_properties_stage(StageProperties::MassFlow, EngineStage::St3);
    fuelFlow_AP = missionIOData->myEngine.get_aircraft_fuelflow();
}


/* Functions for mission data*/

void Mission::get_mission_data() {
    /* Read from acxml */
    std::string mission_file = EndnodeReadOnly<std::string>("requirements_and_specifications/mission_files/" + mission_type + "_mission_file").read(rtIOPt->acxml).value();
    this->delta_isa = EndnodeReadOnly<double>("requirements_and_specifications/requirements/top_level_aircraft_requirements/" + mission_type
                        + "_mission/delta_ISA").read(rtIOPt->acxml).value();;
    myAtmosphere.setAtmosphere(0., ISA_TEMPERATURE + this->delta_isa, ISA_PRESSURE); //sets atmosphere to delta_isa conditions
    double cargo_fraction(1.);
    double pax_mass_fraction(1.);
    if (mission_type == "study") {
        cargo_fraction = EndnodeReadOnly<double>(std::string("requirements_and_specifications/requirements/top_level_aircraft_requirements/")
                            + std::string("study_mission/payload_fractions/cargo_fraction")).read(rtIOPt->acxml).value();
        pax_mass_fraction = EndnodeReadOnly<double>(std::string("requirements_and_specifications/requirements/top_level_aircraft_requirements/")
                            + std::string("study_mission/payload_fractions/passenger_mass_fraction")).read(rtIOPt->acxml).value();
    }
    double cargo_mass = cargo_fraction * EndnodeReadOnly<double>(std::string("requirements_and_specifications/design_specification/")
                                + std::string("transport_task/cargo_definition/additional_cargo_mass")).read(rtIOPt->acxml).value();

    double pax_mass = pax_mass_fraction
                        * EndnodeReadOnly<double>(std::string("requirements_and_specifications/design_specification/")
                                + std::string("transport_task/passenger_definition/total_number_passengers")).read(rtIOPt->acxml).value()
                        * (EndnodeReadOnly<double>(std::string("requirements_and_specifications/design_specification/")
                                + std::string("transport_task/passenger_definition/mass_per_passenger")).read(rtIOPt->acxml).value()
                            + EndnodeReadOnly<double>(std::string("requirements_and_specifications/design_specification/")
                                + std::string("transport_task/passenger_definition/luggage_mass_per_passenger")).read(rtIOPt->acxml).value());
    double  payload = cargo_mass + pax_mass;
    cruise_number_of_steps = rtIOPt->acxml.getVector("/aircraft_exchange_file/analysis/mission/" + mission_type + "_mission/cruise/cruise_steps/cruise_step").size();
    for (int i = 0; i < cruise_number_of_steps; i++) {
        relative_end_of_step.push_back(EndnodeReadOnly<double>("analysis/mission/" + mission_type +
                        "_mission/cruise/cruise_steps/cruise_step@" + num2Str(i) + "/relative_end_of_cruise_step").read(rtIOPt->acxml).value());
        flight_level.push_back(convertUnit(METER, FOOT, (EndnodeReadOnly<double>("analysis/mission/" + mission_type +
                        "_mission/cruise/cruise_steps/cruise_step@" + num2Str(i) + "/altitude").read(rtIOPt->acxml).value())/100));
    }
    /* Read from mission.xml */
    const node& mission_xml(aixml::openDocument(rtIOPt->getMissionDataDir() + "/" + mission_file));
    taxi_time = EndnodeReadOnly<double>("mission/taxi_time_origin").read(mission_xml).value() + EndnodeReadOnly<double>("mission/taxi_time_destination").read(mission_xml).value();
    mission_range = convertUnit(METER, NAUTICALMILE, EndnodeReadOnly<double>("mission/range").read(mission_xml).value());

    /* Read from mission.csv */
    std::string outputFilename = rtIOPt->acxmlAccess.filename().string();
    outputFilename.resize(outputFilename.rfind('.')); // cut off any extension, if it exists
    std::stringstream output_name;
    output_name << outputFilename << "_" << mission_file.substr(mission_file.rfind(FILESEPERATOR) + 1, mission_file.rfind('.') - mission_file.rfind(FILESEPERATOR) - 1)
               << "_R" << int(ceil(mission_range))
               << "_PL" << int(ceil(payload)) << "_out.csv";
    this->mission_csv_name = rtIOPt->getMissionDataDir() + "/" + output_name.str();
    /* Reading the mission file into a std::vector*/
    std::ifstream mission_csv;
    mission_csv.open(mission_csv_name.c_str(), std::ios::in);
    if (!mission_csv) {
        myRuntimeInfo->err << mission_csv_name << " can not be opened." << std::endl;
        exit(1);
    } else {
        std::vector<std::string> header_vector;
        std::unordered_map<std::string, std::function<void(const std::string&, MissionData&)>> fieldMapping = { // Map for header and functions
            {"Time [s]", [](const std::string& value, MissionData& data) { data.time = std::atof(value.c_str()); }},
            {"Range [m]", [](const std::string& value, MissionData& data) { data.range = std::atof(value.c_str()); }},
            {"Altitude [m]", [](const std::string& value, MissionData& data) { data.altitude = std::atof(value.c_str()); }},
            {"TAS [m/s]", [](const std::string& value, MissionData& data) { data.TAS_ms = std::atof(value.c_str()); }},
            {"CAS [m/s]", [](const std::string& value, MissionData& data) { data.CAS_ms = std::atof(value.c_str()); }},
            {"Total mass [kg]", [](const std::string& value, MissionData& data) { data.mass = std::atof(value.c_str()); }},
            {"Fuel consumed", [](const std::string& value, MissionData& data) { data.fuelConsumed = std::atof(value.c_str()); }},
            {"Thrust [N]", [](const std::string& value, MissionData& data) { data.thrust = std::fabs(std::atof(value.c_str())); }},
            {"Fuelflow [kg/s]", [](const std::string& value, MissionData& data) { data.fuelFlow = std::atof(value.c_str()); }},
            {"TAS [kts]", [](const std::string& value, MissionData& data) { data.TAS_kts = std::atof(value.c_str()); }},
            {"Mach [-]", [](const std::string& value, MissionData& data) { data.Mach = std::atof(value.c_str()); }},
            {"FL [100 ft]", [](const std::string& value, MissionData& data) { data.flight_level = std::atof(value.c_str()); }},
            {"ROC [fpm]", [](const std::string& value, MissionData& data) { data.ROC = std::atof(value.c_str()); }},
            {"C_L [-]", [](const std::string& value, MissionData& data) { data.C_L = std::atof(value.c_str()); }},
            {"SAR [m/kg]", [](const std::string& value, MissionData& data) { data.SAR = std::atof(value.c_str()); }},
            {"Aero Config [-]", [](const std::string& value, MissionData& data) {
                std::string trimmed = value;
                trimmed.erase(std::remove_if(trimmed.begin(), trimmed.end(), ::isspace), trimmed.end());
                data.aeroConfig = trimmed;
            }},
            {"Bleed [kg/s]", [](const std::string& value, MissionData& data) { data.bleedAirOfftake = std::atof(value.c_str()); }},
            {"Shaft power offtake [W]", [](const std::string& value, MissionData& data) { data.shaftPowerOfftake = std::atof(value.c_str()); }},
            {"Engine Rating [-]", [](const std::string& value, MissionData& data) {
                std::string trimmed = value;
                trimmed.erase(std::remove_if(trimmed.begin(), trimmed.end(), ::isspace), trimmed.end());
                data.engineRating = trimmed.empty() || trimmed == "0" ? "takeoff" : trimmed;
            }},
            {"Engine N1", [](const std::string& value, MissionData& data) { data.N1 = std::atof(value.c_str()); }},
            {"Angle of attack [deg]", [](const std::string& value, MissionData& data) { data.AOA = std::atof(value.c_str()); }},
            {"Spoiler Factor [-]", [](const std::string& value, MissionData& data) { data.spoilerFactor = std::atof(value.c_str()); }},
            {"Mode name [-]", [](const std::string& value, MissionData& data) {
                std::string trimmed = value;
                trimmed.erase(std::remove_if(trimmed.begin(), trimmed.end(), ::isspace), trimmed.end());
                data.modeName = trimmed;
            }}
        };

        std::string line;
        int line_number = 0;
        while (getline(mission_csv, line)) {
            std::vector<std::string> line_data;
            split(line, ';', &line_data);
            if (line_number == 0) {
                header_vector = line_data; // Save header of columns
                for (const auto& [key, _] : fieldMapping) {
                    bool found = std::any_of(header_vector.begin(), header_vector.end(), [&](const std::string& header) {
                        return header.find(key) != std::string::npos;
                    });
                    if (!found) {
                        throw(std::format("Corrupted mission output file {}. Entry '{}' not found: Abort program!", mission_csv_name, key));
                    }
                }
            } else if (line_number > 1) { // Without header and first data line
                MissionData tmpMissionData;
                for (size_t i = 0; i < line_data.size(); ++i) {
                    for (const auto& [key, func] : fieldMapping) {
                        if (header_vector[i].find(key) != std::string::npos) {
                            func(line_data[i], tmpMissionData); // Save csv value in temporary mission data
                            break;
                        }
                    }
                }
                missionIOData->myEngine.set_operating_point({tmpMissionData.N1, tmpMissionData.altitude, tmpMissionData.Mach});
                tmpMissionData.Tamb = myAtmosphere.getTemperature(tmpMissionData.altitude);
                tmpMissionData.Pamb = myAtmosphere.getPressure(tmpMissionData.altitude);
                tmpMissionData.NOxFlow = missionIOData->myEngine.get_engine_NOx_emission_index() * missionIOData->myEngine.get_aircraft_fuelflow();
                tmpMissionData.St3MassFlow = missionIOData->myEngine.get_physical_properties_stage(StageProperties::MassFlow, EngineStage::St3);
                tmpMissionData.St3TotalTemperature = missionIOData->myEngine.get_physical_properties_stage(StageProperties::TotalTemperature, EngineStage::St3);
                tmpMissionData.St3StaticTemperature = missionIOData->myEngine.get_physical_properties_stage(StageProperties::StaticTemperature, EngineStage::St3);
                tmpMissionData.St3TotalPressure = missionIOData->myEngine.get_physical_properties_stage(StageProperties::TotalPressure, EngineStage::St3);
                tmpMissionData.St3Mach = missionIOData->myEngine.get_physical_properties_stage(StageProperties::MachNumber, EngineStage::St3);
                tmpMissionData.St4FuelAir = missionIOData->myEngine.get_physical_properties_stage(StageProperties::FuelToAirRatio, EngineStage::St4);
                tmpMissionData.BPR = (missionIOData->myEngine.get_physical_properties_stage(StageProperties::MassFlow, EngineStage::St22) > 0.) ?
                                     missionIOData->myEngine.get_physical_properties_stage(StageProperties::MassFlow, EngineStage::St13)
                                     / missionIOData->myEngine.get_physical_properties_stage(StageProperties::MassFlow, EngineStage::St22) : 0.;
                missionInfo.push_back(MissionData());
                missionInfo.back() = tmpMissionData;
            }
            ++line_number;
        }
        mission_csv.close();
    }
}

void Mission::get_mission_segment_data() {
    /* Calculate Fuel and distance lengths for different phases */
    /* Taxi (T/O and Landing)*/
    if (mission_type == "design") {
        fuel_mass_taxi = EndnodeReadOnly<double>("analysis/mission/design_mission/taxi_energy").read(rtIOPt->acxml).value()
                        /EnergyCarrier(rtIOPt->get_fuel_type(0), rtIOPt->get_fuel_density(0)).gravimetric_energy_density;
    } else {
        fuel_mass_taxi = EndnodeReadOnly<double>("analysis/mission/study_mission/taxi_energy").read(rtIOPt->acxml).value()
                        /EnergyCarrier(rtIOPt->get_fuel_type(0), rtIOPt->get_fuel_density(0)).gravimetric_energy_density;
    }

    unsigned int tmp_line = 1;//Line 0 would be first line with description of columns, therefore start at 1
    double tmp_altitude = this->missionInfo.at(tmp_line).altitude; //Height in meter as 3rd entry in csv file

    /* Takeoff */
    double takeoff_altitude = convertUnit(FOOT, METER, 35.); // T/O until 35 ft, afterwards ClimbOut, vgl. "Elements of Aircraft Pollution", von Ruijgrok, Fig. 11.4-1
    double LTO_altitude = Rounding(convertUnit(FOOT, METER, 3000.), 1); // ICAO: LTO Cycle until 3000 ft.

    /* Start until TO-Height */
    while (tmp_altitude <= takeoff_altitude) {
        tmp_line = tmp_line + 1;
        tmp_altitude = this->missionInfo.at(tmp_line).altitude; //Height in meter as 3rd entry in csv file
    }
    tmp_line = tmp_line - 1;//while Schleife einmal zuviel durchgelaufen
    fuel_mass_takeoff = -missionInfo.at(tmp_line).fuelConsumed;//Fuel consumed als 7. in csv-Datei
    takeoff_time = missionInfo.at(tmp_line).time - 0.;

    /* ClimbOut */
    while (tmp_altitude <= LTO_altitude) {
        tmp_line = tmp_line + 1;
        tmp_altitude = missionInfo.at(tmp_line).altitude; //Height in meter as 3rd entry in csv file
    }
    tmp_line = tmp_line - 1;
    tmp_altitude = missionInfo.at(tmp_line).altitude;//necessary for query if (tmp_altitude == missionInfo.at(tmp_line).altitude) below in while loop
    fuel_mass_climbout = -missionInfo.at(tmp_line).fuelConsumed - fuel_mass_takeoff; //Fuel consumed als 7. in csv-Datei
    climbout_time = missionInfo.at(tmp_line).time - takeoff_time;

    /* LTO to 1st Cruise Step (already counts as "Cruise" for NOx calculation) */
    //Attention, gives rounding error and initialCruiseHeight greater than the height in the csv -> below query when 2 times same height comes in a row -> 1st Step
    double inital_cruise_altitude = this->flight_level.front() * convertUnit(FOOT, METER, 100.);
    while (tmp_altitude < inital_cruise_altitude) {
        tmp_line = tmp_line + 1;
        if (accuracyCheck(tmp_altitude, missionInfo.at(tmp_line).altitude, ACCURACY_HIGH)) { //To catch the rounding error (see double inital_cruise_altitude
            inital_cruise_altitude = tmp_altitude;
        } else {
            if (tmp_line == missionInfo.size()-1) { //last precautionary measure, but should never come into play due to the previous query
                myRuntimeInfo->err << "Cruise Altitude < beforeInitalCruiseAlt; check if Acft.xml and mission.xml belong together" <<
                                   std::endl; //in case of repeated occurrence of controlled termination
                exit(1);
            }
            tmp_altitude = missionInfo.at(tmp_line).altitude; //Height in meter as 3. in csv file
        }
    }
    double last_cruise_altitude = this->flight_level.back() * convertUnit(FOOT, METER, 100.);
    double before_final_cruise_altitude = floor(last_cruise_altitude);// instead of asking if height occurs 2 times in the same direction (see InitialCruise Height)
    if (this->flight_level.size() > 1) {
        /* flown range from Initial Cruise to beginning of last Cruise Step*/
        while (tmp_altitude < before_final_cruise_altitude) { //
            tmp_line = tmp_line + 1;
            tmp_altitude = missionInfo.at(tmp_line).altitude; //Height in meter 3rd entry in csv file
       }
    }
    //here last Cruise Step reached, which still must be flown before height is reduced again
    while (tmp_altitude >= before_final_cruise_altitude) { //falls flight_level.size()=1 finalCruiseAlt=initalCruiseAlt
        tmp_line = tmp_line + 1;
        tmp_altitude = missionInfo.at(tmp_line).altitude; //Height in meter as 3rd entry in csv file
    }
    tmp_line = tmp_line - 1;

    /*Descend until LTO Height*/
    while (tmp_altitude >= LTO_altitude) {
        tmp_line = tmp_line + 1;
        tmp_altitude = missionInfo.at(tmp_line).altitude; //Height in meter as 3. entry in csv file
    }
    tmp_line = tmp_line - 1;//for NOx the FuelFlow up to 1st Cruise Step interesting, not for 1st Cruise Step
    fuel_mass_cruise = -missionInfo.at(tmp_line).fuelConsumed - fuel_mass_takeoff - fuel_mass_climbout;
    double time_until_approach = missionInfo.at(tmp_line).time;

    /* Approach */
    while (tmp_line < (missionInfo.size() - 2)) { //tmp_line: 1.Linie=1, missionInfo.size: Vektor! 1.Line = 0; and then it goes through again!
        tmp_line = tmp_line + 1;
    }
    fuel_mass_approach = (-missionInfo.at(tmp_line).fuelConsumed + missionInfo.at(1).fuelConsumed) - fuel_mass_cruise - fuel_mass_takeoff - fuel_mass_climbout;
    fuel_mass_LTO = fuel_mass_taxi + fuel_mass_takeoff + fuel_mass_climbout + fuel_mass_approach;
    approach_time = missionInfo.at(tmp_line).time - time_until_approach;
}

void Mission::write_emissions_path_csv() {
    std::ofstream emissions_csv_stream;
    std::string emission_csv_name;
    emission_csv_name = mission_csv_name.substr(0, mission_csv_name.rfind("_out")) + "_emissionspath.csv";
    emissions_csv_stream.open(emission_csv_name.c_str());
    if (!emissions_csv_stream) {
        myRuntimeInfo->err << "could not open " << emission_csv_name << std::endl;
        exit(1);
    }
    emissions_csv_stream << "(1) Time [s];(2) Range [m];(3) Altitude [m];(4) TAS m/s];(5) Mach [-];(6) T_amb [K];(7) P_amb [N/m^2];(8) Consumed Fuel [kg];"
           "(9) Fuel Flow [kg/s];(10) NOx [kg/s];(11) N_1 [-];(12) ST3MassFlow [kg/s];(13) St3TotalTemperature [K];(14) St3StaticTemperature [K];"
           "(15) St3TotalPressure [kPa];(16) St3Mach [-];(17) St4FuelAir [-];(18) BPR [-];(19) modeName;";
    for (unsigned int i = 0; i < missionInfo.size(); i++) {
        emissions_csv_stream << std::endl;
        emissions_csv_stream << std::setprecision(10) << missionInfo.at(i).time << ";";
        emissions_csv_stream << std::setprecision(10) << missionInfo.at(i).range << ";";
        emissions_csv_stream << std::setprecision(10) << missionInfo.at(i).altitude << ";";
        emissions_csv_stream << std::setprecision(10) << missionInfo.at(i).TAS_ms <<  ";";
        emissions_csv_stream << std::setprecision(10) << missionInfo.at(i).Mach <<  ";";
        emissions_csv_stream << std::setprecision(10) << missionInfo.at(i).Tamb <<  ";";
        emissions_csv_stream << std::setprecision(10) << missionInfo.at(i).Pamb <<  ";";
        emissions_csv_stream << std::setprecision(10) << missionInfo.at(i).fuelConsumed <<  ";";
        emissions_csv_stream << std::setprecision(10) << missionInfo.at(i).fuelFlow <<  ";";
        emissions_csv_stream << std::setprecision(10) << missionInfo.at(i).NOxFlow <<  ";";
        emissions_csv_stream << std::setprecision(10) << missionInfo.at(i).N1 << ";";
        emissions_csv_stream << std::setprecision(10) << missionInfo.at(i).St3MassFlow << ";";
        emissions_csv_stream << std::setprecision(10) << missionInfo.at(i).St3TotalTemperature << ";";
        emissions_csv_stream << std::setprecision(10) << missionInfo.at(i).St3StaticTemperature << ";";
        emissions_csv_stream << std::setprecision(10) << missionInfo.at(i).St3TotalPressure << ";";
        emissions_csv_stream << std::setprecision(10) << missionInfo.at(i).St3Mach << ";";
        emissions_csv_stream << std::setprecision(10) << missionInfo.at(i).St4FuelAir << ";";
        emissions_csv_stream << std::setprecision(10) << missionInfo.at(i).BPR << ";";
        emissions_csv_stream << missionInfo.at(i).modeName << ";";
    }
    emissions_csv_stream.close();
}

/* Functions for interpolation */

// cppcheck-suppress unusedFunction
double Mission::point_to_point_fit(double REI_TO, double REI_CL, double REI_AP, double REI_ID,
                                int mission_point) { //Linear interpolation of the ICAO data from NOx, where REI is the intersection of Wff with the interpolated curve.
    //Wf values from LTO ICAO cycle corrected with Boeing's correction factor r;
    //r = 1.100 for ID; r = 1.010 for takeoff; r = 1.013 for CL; r = 1.020 for AP;
    double Wf_ID = missionIOData->myEngine.get_LTO_fuelflow(LTOPhases::taxi) * 7936 * 1.100; // [lbm/h] (1 lbm/h = 1 kg/s * 7936);
    double Wf_TO = missionIOData->myEngine.get_LTO_fuelflow(LTOPhases::takeoff) * 7936 * 1.010; // [lbm/h] (1 lbm/h = 1 kg/s * 7936);
    double Wf_CL = missionIOData->myEngine.get_LTO_fuelflow(LTOPhases::climb) * 7936 * 1.013; // [lbm/h] (1 lbm/h = 1 kg/s * 7936);
    double Wf_AP = missionIOData->myEngine.get_LTO_fuelflow(LTOPhases::approach) * 7936 * 1.020; // [lbm/h] (1 lbm/h = 1 kg/s * 7936);
    double tempWf = (this->myFlightPath.pathPoints.at(mission_point).fuelFlow / this->missionIOData->number_of_engines) * 7936; // [lbm/h] (1 lbm/h = 1 kg/s * 7936);
    double REI = 0;
    if (tempWf <= Wf_AP) {
        REI = pow(10., (log10(REI_AP) - log10(REI_ID)) / (log10(Wf_AP) - log10(Wf_ID)) * (log10(tempWf) - log10(Wf_ID)) + log10(REI_ID));
    } else if ((tempWf > Wf_AP) && (tempWf <= Wf_CL)) {
        REI = pow(10., (log10(REI_CL) - log10(REI_AP)) / (log10(Wf_CL) - log10(Wf_AP)) * (log10(tempWf) - log10(Wf_AP)) + log10(REI_AP));
    } else {
        REI = pow(10., (log10(REI_TO) - log10(REI_CL)) / (log10(Wf_TO) - log10(Wf_CL)) * (log10(tempWf) - log10(Wf_CL)) + log10(REI_CL));
    }
    return REI;
}

double Mission::linear_least_square_fit(double REI_TO, double REI_CL, double REI_AP, double REI_ID, int mission_point) {
    double Wf_ID = missionIOData->myEngine.get_LTO_fuelflow(LTOPhases::taxi) * 7936 * 1.100; // [lbm/h] (1 lbm/h = 1 kg/s * 7936);
    double Wf_TO = missionIOData->myEngine.get_LTO_fuelflow(LTOPhases::takeoff) * 7936 * 1.010; // [lbm/h] (1 lbm/h = 1 kg/s * 7936);
    double Wf_CL = missionIOData->myEngine.get_LTO_fuelflow(LTOPhases::climb) * 7936 * 1.013; // [lbm/h] (1 lbm/h = 1 kg/s * 7936);
    double Wf_AP = missionIOData->myEngine.get_LTO_fuelflow(LTOPhases::approach) * 7936 * 1.020; // [lbm/h] (1 lbm/h = 1 kg/s * 7936);
    double tempWf = (this->myFlightPath.pathPoints.at(mission_point).fuelFlow / this->missionIOData->number_of_engines) * 7936; // [lbm/h] (1 lbm/h = 1 kg/s * 7936);
    double p = log10(Wf_TO) + log10(Wf_CL) + log10(Wf_AP) + log10(Wf_ID);
    double q = log10(REI_TO) + log10(REI_CL) + log10(REI_AP) + log10(REI_ID);
    double r = pow(log10(Wf_TO), 2.) + pow(log10(Wf_CL), 2.) + pow(log10(Wf_AP), 2.) + pow(log10(Wf_ID), 2.);
    double s = log10(Wf_TO) * log10(REI_TO) + log10(Wf_CL) * log10(REI_CL) + log10(Wf_AP) * log10(REI_AP) + log10(Wf_ID) * log10(REI_ID);
    double REI = 0;
    int n = 4;
    REI = pow(10., (n * s - p * q) / (n * r - pow(p, 2.)) * log10(tempWf) + (q * r - p * s) / (n * r - pow(p, 2.)));
    if (REI > REI_TO) {
        REI = REI_TO;
        return REI;
    } else {
        return REI;
    }
}

// Calculation of the EI at reference conditions (REI), linear least square fit

double Mission::bi_linear_least_square_fit(double REI_TO, double REI_CL, double REI_AP, double REI_ID, int mission_point) {
    //Wf values from LTO ICAO cycle corrected with Boeing's correction factor r;
    //r = 1.100 for ID; r = 1.010 for takeoff; r = 1.013 for CL; r = 1.020 for AP;
    double Wf_ID = missionIOData->myEngine.get_LTO_fuelflow(LTOPhases::taxi) * 7936 * 1.100; // [lbm/h] (1 lbm/h = 1 kg/s * 7936);
    double Wf_AP = missionIOData->myEngine.get_LTO_fuelflow(LTOPhases::approach) * 7936 * 1.020; // [lbm/h] (1 lbm/h = 1 kg/s * 7936);
    double tempWf = (this->myFlightPath.pathPoints.at(mission_point).fuelFlow / this->missionIOData->number_of_engines) * 7936; // [lbm/h] (1 lbm/h = 1 kg/s * 7936);
    double p = log10(Wf_AP) + log10(Wf_ID);
    double q = log10(REI_AP) + log10(REI_ID);
    double r = pow(log10(Wf_AP), 2.) + pow(log10(Wf_ID), 2.);
    double s = log10(Wf_AP) * log10(REI_AP) + log10(Wf_ID) * log10(REI_ID);
    int n = 2;
    double Wf_intersect = 0; // [lbm/h]
    double REI = 0;
    Wf_intersect = pow(10., ((log10(REI_TO) + log10(REI_CL)) / 2. - (q * r - p * s) / (n * r - pow(p, 2.))) * (n * r - pow(p, 2.)) / (n * s - p * q));
    if (fabs(tempWf) < ACCURACY_HIGH) {
        REI = 0;
        return REI;
    } else {
        if (tempWf <= Wf_intersect) {
            REI = pow(10., (n * s - p * q) / (n * r - pow(p, 2.)) * log10(tempWf) + (q * r - p * s) / (n * r - pow(p, 2.)));
            if (REI > REI_ID) {
                REI = REI_ID;
                return REI;
            } else {
                return REI;
            }
        } else {
            REI = pow(10, (log10(REI_TO) + log10(REI_CL)) / 2.);
            return REI;
        }
    }
}

//potential_fit creates a graphic in the form y=ax^b
//when Y==false "input" should be a value in "Y-axis" and the value returned will be a value from "X-axis"
double Mission::potential_fit(double input, double x_ID, double x_TO, double x_CL, double x_AP, double y_ID, double y_TO, double y_CL, double y_AP, bool Y) {
    double x = log10(x_ID) + log10(x_TO) + log10(x_CL) + log10(x_AP);
    double x_2 = pow(log10(x_ID), 2) + pow(log10(x_TO), 2) + pow(log10(x_CL), 2) + pow(log10(x_AP), 2);
    double y = log10(y_ID) + log10(y_TO) + log10(y_CL) + log10(y_AP);
    double y_2 = log10(x_ID) * log10(y_ID) + log10(x_TO) * log10(y_TO) + log10(x_CL) * log10(y_CL) + log10(x_AP) * log10(y_AP);
    double a = pow(10, (y_2 - y / x * x_2) / (x - 4 / x * x_2));
    double b = (y - 4 * log10(a)) / x;
    if (Y) {
        double x_mom = input;
        double y_mom = a * pow(x_mom, b);
        return y_mom;
    } else {
        double y_mom = input;
        double x_mom = pow(y_mom / a, 1 / b);
        return x_mom;
    }
}

// cppcheck-suppress unusedFunction
double Mission::potential_fit_2(double input, double x_ID, double x_TO, double x_CL, double x_AP, double y_ID, double y_TO, double y_CL, double y_AP, bool Y) {
    double x = log10(x_ID) + log10(x_TO) + log10(x_CL) + log10(x_AP);
    double x_2 = pow(log10(x_ID), 2) + pow(log10(x_TO), 2) + pow(log10(x_CL), 2) + pow(log10(x_AP), 2);
    double y = log10(y_ID) + log10(y_TO) + log10(y_CL) + log10(y_AP);
    double y_2 = log10(x_ID) * log10(y_ID) + log10(x_TO) * log10(y_TO) + log10(x_CL) * log10(y_CL) + log10(x_AP) * log10(y_AP);
    double a = pow(10, (y_2 - y / x * x_2) / (x - 4 / x * x_2));
    double b = (y - 4 * log10(a)) / x;
    if (Y) {
        double x_mom = input;
        double y_mom = a * pow(x_mom, b);
        return y_mom;
    } else {
        double y_mom = input;
        double x_mom = pow(y_mom / a, 1 / b);
        return x_mom;
    }
}

// cppcheck-suppress unusedFunction
double Mission::potential_fit_3(double input, double x_ID, double x_TO, double x_CL, double x_AP, double y_ID, double y_TO, double y_CL, double y_AP, bool Y) {
    double x = log10(x_ID) + log10(x_TO) + log10(x_CL) + log10(x_AP);
    double x_2 = pow(log10(x_ID), 2) + pow(log10(x_TO), 2) + pow(log10(x_CL), 2) + pow(log10(x_AP), 2);
    double y = log10(y_ID) + log10(y_TO) + log10(y_CL) + log10(y_AP);
    double y_2 = log10(x_ID) * log10(y_ID) + log10(x_TO) * log10(y_TO) + log10(x_CL) * log10(y_CL) + log10(x_AP) * log10(y_AP);
    double a = pow(10, (y_2 - y / x * x_2) / (x - 4 / x * x_2));
    double b = (y - 4 * log10(a)) / x;
    if (Y) {
        double x_mom = input;
        double y_mom = a * pow(x_mom, b);
        return y_mom;
    } else {
        double y_mom = input;
        double x_mom = pow(y_mom / a, 1 / b);
        return x_mom;
    }
}

double Mission::quadratic_fit(double x_mom, double x_ID, double x_TO, double x_CL, double x_AP, double y_ID, double y_TO, double y_CL, double y_AP, bool Y) {
    double x = x_ID + x_TO + x_CL + x_AP;
    double x_2 = pow(x_ID, 2) + pow(x_TO, 2) + pow(x_CL, 2) + pow(x_AP, 2);
    double x_3 = pow(x_ID, 3) + pow(x_TO, 3) + pow(x_CL, 3) + pow(x_AP, 3);
    double x_4 = pow(x_ID, 4) + pow(x_TO, 4) + pow(x_CL, 4) + pow(x_AP, 4);
    double y = y_ID + y_TO + y_CL + y_AP;
    double y_2 = x_ID * y_ID + x_TO * y_TO + x_CL * y_CL + x_AP * y_AP;
    double y_3 = pow(x_ID, 2) * y_ID + pow(x_TO, 2) * y_TO + pow(x_CL, 2) * y_CL + pow(x_AP, 2) * y_AP;
    double eigenval = 4 * x_2 * x_4 + 2 * x * x_3 * x_2 - pow(x_2, 3) - 4 * pow(x_3, 2) - pow(x, 2) * x_4;
    double a = (y * x_2 * x_4 + y_2 * x_3 * x_2 + y_3 * x * x_3 - pow(x_2, 2) * y_3 - y_2 * x * x_4 - y * pow(x_3, 2)) / eigenval;
    double b = (4 * y_2 * x_4 + x * y_3 * x_2 + y * x_3 * x_2 - pow(x_2, 2) * y_2 - 4 * x_3 * y_3 - x * x_4 * y) / eigenval;
    double c = (4 * x_2 * y_3 + x * x_3 * y + x * y_2 * x_2 - pow(x_2, 2) * y - 4 * x_3 * y_2 - pow(x, 2) * y_3) / eigenval;
    double y_mom = a + b * x_mom + c * pow(x_mom, 2);
    if (Y) {
        if (y_mom > std::max(y_ID, y_TO)) {
            y_mom = std::max(y_ID, y_TO);
            return y_mom;
        } else {
            return y_mom;
        }
    } else {
        return y_mom;
    }
}

double Mission::linear_interpolation(double x, double x_ID, double x_TO, double x_CL, double x_AP, double y_ID, double y_TO, double y_CL, double y_AP) {
    if (x < x_ID) {
        double y = y_ID * x / y_ID;
        return y;
    } else if ((x_ID <= x) && (x < x_AP)) {
        double y = y_ID + (y_AP - y_ID) * (x - x_ID) / (y_AP - y_ID);
        return y;
    } else if ((x_AP <= x) && (x < x_CL)) {
        double y = y_AP + (y_CL - y_AP) * (x - x_AP) / (y_CL - y_AP);
        return y;
    } else if ((x_CL <= x) && (x < x_TO)) {
        double y = y_CL + (y_TO - y_CL) * (x - x_CL) / (y_TO - y_CL);
        return y;
    } else {
        double y = y_TO;
        return y;
    }
}

/* Constuctors and Destructor*/
Mission::~Mission() {
    //dtor
}

Mission::MissionData::MissionData()
    :
    time(0.),
    range(0.),
    altitude(0.),
    TAS_ms(0.),
    CAS_ms(0.),
    mass(0.),
    fuelConsumed(0.),
    thrust(0.),
    fuelFlow(0.),
    TAS_kts(0.),
    Mach(0.),
    flight_level(0.),
    ROC(0.),
    C_L(0.),
    SAR(0.),
    aeroConfig("TakeOff"),
    bleedAirOfftake(0.),
    shaftPowerOfftake(0.),
    engineRating("takeoff"),
    AOA(0.),
    spoilerFactor(1.),
    Tamb(0.),
    Pamb(0.),
    NOxFlow(0.),
    N1(0.),
    St3MassFlow(0.),
    St3TotalTemperature(0.),
    St3StaticTemperature(0.),
    St3TotalPressure(0.),
    St3Mach(0.),
    St4FuelAir(0.),
    BPR(0.) {
        //ctor
    };

