/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_IMPACT_CALCULATION_CLIMATE_MODEL_DALLARA_DALLARA_H_
#define ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_IMPACT_CALCULATION_CLIMATE_MODEL_DALLARA_DALLARA_H_

#include <memory>
#include <vector>
#include <string>
#include <moduleBasics/runtimeIO.h>
#include <aixml/node.h>
#include "dallaraIOData.h"

/** \class ClimateModelDallara
 * \brief Class for environmental impact calculation according to Dallara 2011
 */
class ClimateModelDallara {
 public:
    /** \brief Constructor of class ClimateModelDallara
     */
    ClimateModelDallara() {}
    /** \brief Destructor of class ClimateModelDallara */
    ~ClimateModelDallara() = default;

    /** \brief Function to run Dallara method
     * \param rtIO Access to rtIO
     * \param ecoAssXML Pointer to node of ecological assessment results
     */
    void run(const std::shared_ptr<RuntimeIO> &rtIO, std::shared_ptr<node> ecoAssXML);

    /** \brief Function to calculate the climate impact (AGWP, AGTP, ATR)
     * \param fuelType Type of energy carrier
     */
    void calculate_climate_impact_dallara(const std::string &fuelType);

    /** \brief Function to calculate the avarage forcing factors for of the mission
     * \param emissionNo Index of emission species
     */
    void get_average_forcing_factors(int emissionNo);

    /** \brief Function to calculate the (normalized) radiative forcing of all emission species
     * \param emissionNo Index of emission species
     * \param fuelType Type of energy carrier
     */
    void calculate_radiative_forcing(int emissionNo, const std::string &fuelType);

    /** \brief Function to calculate the temperature change of emission species for every year
     * \param emissionNo Index of emission species
     */
    void calculate_temperature_change(int emissionNo);

    /** \brief Function to calculate the absolute global temperature potential (AGTP) and absolute global warming potential (AGWP)
     *  of emission species for every year
     * \param emissionNo Index of emission species
     */
    void calculate_agtp_and_agwp(int emissionNo);

    /** \brief Function to calculate the average temperature response of all emission species
     * \param emissionNo Index of emission species
     */
    void calculate_average_temperature_response(int emissionNo);

    /** \brief Function to sum up results of all emission species ("total")
     */
    void calculate_total_impact();

    /** \brief Function to calculate the altitude dependent forcing factors
    * \param emissionNo Index of emission species
    * \param flightHeight Flight altitude [ft]
    * \return Forcing factor [-]
    */
    double get_forcing_factor(int emissionNo, double flightHeight);

    /** \brief Function to calculate the response function
    * \param emissionNo Index of emission species [-]
    * \param time Time between current year in outer and inner loop [yrs]
    * \param fuelType Energy carrier [-]
    * \return Response function result [(W/m^2)/kg]
    */
    double get_response_function(int emissionNo, double time, const std::string &fuelType);

    /** \brief Function calculates a weighting factor for temperature change (s. eq. (2)). It assigns unity weighting to temperature change during the operating lifetime
    * of the aircraft and an exponential devaluation rate after the aircraft has ceased operation
    * \param time Time [yrs]
    * \return Weighting factor [-]
    */
    double weighting_function(double time);

    /** \brief Function calculates factor to scale values per emission unit to annual values
    * \param emissionNo Index of emission species
    * \return Scale factor [-]
    */
    double get_scale_factor(int emissionNo);

    std::shared_ptr<DallaraIOData> dallaraIOData; /**< IOData object for climate impact calculation according to dallara */
    double forcing_factor_average[8]{0., 0., 0., 0., 0., 0., 0., 0.};/**< Mission average forcing factors [-] \cite Dal11 page 1607 equation(18) or modified version \cite Ste13 is used.*/
    /* Climate metrics per emission species */
    std::vector<std::vector<double>> radiative_forcing;/**< Radiative Forcing for a unit emission (cf. \cite Dal11 eq. (5),(8),(10),(11)) [W/m^2]*/
    std::vector<std::vector<double>> normalized_radiative_forcing;/**< Normalized radiative_forcing [-]*/
    std::vector<std::vector<double>> annual_radiative_forcing; /**< Absolute radiative forcing [W/m^2]*/
    std::vector<std::vector<double>> annual_normalized_radiative_forcing; /**< Absolute normalized radiative forcing [-] */
    std::vector<std::vector<double>> temperature_change;/**<Temperature Change for a unit of the different emission types at different times (yrs) \cite Dal11 eq.(14) [kg/year,NM/year]*/
    std::vector<std::vector<double>> temperature_change_weighted;/**<Weighted Temperature Change for a unit of the different emission types at different times (yrs) \cite Dal11 eq.(14) [kg/year,NM/year]*/
    std::vector<std::vector<double>> annual_temperature_change; /**< Absolute Temperature Change [K]*/
    std::vector<std::vector<double>> annual_temperature_change_weighted; /**< Absolute Temperature Change weighted [K]*/
    std::vector<std::vector<double>> agtp; /**< Absolute Global Temperature Potential for a unit Emission [K]*/
    std::vector<std::vector<double>> agwp; /**< Absolute Global Warming Potential for a unit Emission [(W/m2)*year]*/
    std::vector<std::vector<double>> annual_agwp;/**< Absolute Global Warming Potential [(W/m2)*year]*/
    std::vector<std::vector<double>> annual_agtp;/**< Absolute Global Temperature Potential after final year of operation [K] */

    /* Adding up the values of the individual components*/
    std::vector<double> total_annual_temperature_change; /**< Calculate Total Temperature Changes resulting from all emissions [K]*/
    std::vector<double> total_annual_temperature_change_weighted; /**< Calculate Total Temperature Changes weighted resulting from all emissions [K]*/
    std::vector<double> total_annual_agtp; /**< Calculate AGTP, integrated Temp change resulting from all emissions [K]*/
    std::vector<double> annual_cooling_NOx; /**< Calculate NOX Cooling Temperature Changes [K]*/
    std::vector<double> annual_warming_H2O_SO4_Soot;/**< Calculate Temperature Changes resulting from H2O, SO4 and soot [K]*/
    std::vector<double> total_annual_radiative_forcing;/**< Absolute radiative_forcing resulting from all emissions [W/m^2]*/
    std::vector<double> total_annual_normalized_radiative_forcing;/**< Absolute normalized radiative_forcing resulting from all emissions [-]*/
    std::vector<double> total_annual_agwp; /**< Calculate AGWP resulting from all emissions [(W/m2)*year]*/
    std::vector<double> ATR_per_year;/**< Average Temperature Response for one year per emission [K]*/
    std::vector<double> emissions_per_year; /**< Emission per emission for one year [kg] */

    /* Constants for climate model*/

    //For ATR calculation: \cite Dall11
    //For 13 height steps (column 1) the forcing factors (s) for CH4&O3L (column 2), O3S (column 3),
    //AIC (column 4) (see \cite Dal11 Fig.3) and ExcelTab: std::vectorsForcingFactors */
    const double forcing_factor_height_Dall[13][5] = { /**< Table of forcing factors s_i [-] for AIC, CH4/O3L, O3S, H2O (column 2-5) depending on flight height [ft] (column 1) \cite Dal11 */
        {17500., 0.0278, 0.8814, 0.4885, 1.},//height: 17500 ft, s_AIC: 0.0278, s_CH4O3L: 0.8814, s_O3S: 0.4885, s_H20: 1
        {19500., 0.0007, 0.9424, 0.5766, 1.},//height: 19500 ft,...
        {21500., 0.0075, 0.9695, 0.6443, 1.},
        {23500., 0.1972, 0.9763, 0.7188, 1.},
        {25500., 0.4207, 0.9560, 0.7392, 1.},
        {27500., 0.8340, 0.9424, 0.8340, 1.},
        {29500., 1.2879, 0.9356, 0.9424, 1.},
        {31500., 1.7351, 0.9492, 1.0305, 1.},
        {33500., 2.1213, 0.9898, 1.1592, 1.},
        {35500., 1.8300, 1.1524, 1.4573, 1.},
        {37500., 1.5522, 1.2202, 1.6538, 1.},
        {39500., 1.0034, 1.2134, 1.8232, 1.},
        {41500., 0.8001, 1.2134, 1.9519, 1.}
    };

    const double forcing_factor_height_Dahl[68][5] = {/**< Table of forcing factors s_i [-] for AIC, CH4/O3L, O3S, H2O (column 2-5) depending on flight height [ft] (column 1) \cite Koss22 */
        {0., 0.250952189, 1.343876525, 0.657057175, 0.},//height: 0 ft, s_AIC: 0.250952189, s_CH4O3L: 1.343876525, s_O3S: 0.657057175, s_H20: 0
        {16500., 0.250952189, 1.343876525, 0.657057175, 0.},//height: 16500 ft,...
        {17000., 0.295101185, 1.327572997, 0.672735470, 0.},
        {17500., 0.339250182, 1.311269469, 0.688413765, 0.},
        {18000., 0.383399178, 1.294965941, 0.704092060, 0.},
        {18500., 0.427548174, 1.278662413, 0.719770355, 0.},
        {19000., 0.453108119, 1.262358885, 0.731870616, 0.},
        {19500., 0.478668064, 1.246055357, 0.743970876, 0.},
        {20000., 0.504228010, 1.229751829, 0.756071137, 0.},
        {20500., 0.529787955, 1.213448301, 0.768171398, 0.},
        {21000., 0.557671531, 1.197144773, 0.779035611, 0.},
        {21500., 0.585555108, 1.180841245, 0.789899824, 0.},
        {22000., 0.613438684, 1.164537717, 0.800764036, 0.},
        {22500., 0.641322261, 1.148234189, 0.811628249, 0.},
        {23000., 0.692442151, 1.131639526, 0.826135551, 0.},
        {23500., 0.743562042, 1.115044864, 0.840642853, 0.},
        {24000., 0.794681932, 1.098450201, 0.855150155, 0.},
        {24500., 0.845801823, 1.081855539, 0.869657457, 0.},
        {25000., 0.880656293, 1.076323985, 0.878114628, 0.},
        {25500., 0.915510764, 1.070792430, 0.886571800, 0.},
        {26000., 0.950365235, 1.065260876, 0.895028972, 0.},
        {26500., 0.985219705, 1.059729322, 0.903486143, 0.},
        {27000., 1.022397807, 1.037894240, 0.910707267, 0.},
        {27500., 1.059575910, 1.016059158, 0.917928390, 0.},
        {28000., 1.096754012, 0.994224076, 0.925149513, 0.024009536},
        {28500., 1.133932114, 0.972388993, 0.932370637, 0.048019072},
        {29000., 1.171110216, 0.961617019, 0.938485822, 0.072028608},
        {29500., 1.208288318, 0.950845046, 0.944601008, 0.096038145},
        {30000., 1.245466420, 0.940073072, 0.950716194, 0.096038145},
        {30500., 1.282644522, 0.929301098, 0.956831379, 0.096038145},
        {31000., 1.322146256, 0.907466016, 0.973680667, 0.096038145},
        {31500., 1.361647989, 0.885630933, 0.990529955, 0.096038145},
        {32000., 1.401149723, 0.863795851, 1.007379243, 0.096038145},
        {32500., 1.440651456, 0.841960769, 1.024228531, 0.096038145},
        {33000., 1.470858664, 0.830897661, 1.044785963, 0.096038145},
        {33500., 1.501065872, 0.819834552, 1.065343396, 0.096038145},
        {34000., 1.531273080, 0.808771444, 1.085900828, 0.120047681},
        {34500., 1.561480288, 0.797708336, 1.106458261, 0.144057217},
        {35000., 1.554509394, 0.781404808, 1.162991200, 0.168066753},
        {35500., 1.547538500, 0.765101280, 1.219524139, 0.192076289},
        {36000., 1.540567605, 0.748797752, 1.276057078, 0.192076289},
        {36500., 1.533596711, 0.732494224, 1.332590017, 0.192076289},
        {37000., 1.508036766, 0.715899561, 1.403630258, 0.192076289},
        {37500., 1.482476821, 0.699304899, 1.474670499, 0.192076289},
        {38000., 1.456916876, 0.682710236, 1.545710740, 0.216085825},
        {38500., 1.431356930, 0.666115574, 1.616750981, 0.240095362},
        {39000., 1.384884303, 0.649812046, 1.733459949, 0.264104898},
        {39500., 1.338411675, 0.633508518, 1.850168916, 0.288114434},
        {40000., 1.291939048, 0.617204990, 1.966877884, 0.336133506},
        {40500., 1.245466420, 0.600901462, 2.083586851, 0.384152579},
        {41000., 1.178081110, 0.568294406, 2.201531867, 0.432171651},
        {41500., 1.110695800, 0.535687350, 2.319476883, 0.480190723},
        {42000., 1.043310490, 0.503080294, 2.437421898, 0.576228868},
        {42500., 0.975925180, 0.470473237, 2.555366914, 0.672267013},
        {43000., 0.915510764, 0.437866181, 2.693739252, 0.768305157},
        {43500., 0.855096348, 0.405259125, 2.832111589, 0.864343302},
        {44000., 0.794681932, 0.372652069, 2.970483927, 0.960381447},
        {44500., 0.734267516, 0.340045013, 3.108856265, 1.056419591},
        {45000., 0.687794889, 0.307729092, 3.242414520, 1.152457736},
        {45500., 0.641322261, 0.275413170, 3.375972776, 1.248495881},
        {46000., 0.594849633, 0.243097248, 3.509531031, 1.512600778},
        {46500., 0.548377006, 0.210781327, 3.643089287, 1.776705676},
        {47000., 0.501904378, 0.178465405, 3.776647542, 2.040810574},
        {47500., 0.455431751, 0.146149483, 3.910205798, 2.304915472},
        {48000., 0.408959123, 0.113833562, 4.043764053, 2.641048978},
        {48500., 0.362486495, 0.081517640, 4.177322309, 2.977182484},
        {49000., 0.316013868, 0.049201719, 4.310880564, 3.313315991},
        {49500., 0.269541240, 0.016885797, 4.444438820, 3.649449497}
    };

    const double efficacy[8] = {1., 1.18, 1.37, 1.37, 1.14, 0.7, 0.9, 0.59}; /**< Efficacy f_i of CO2, CH4, O3L, O3S, H2O, soot, SO4, AIC \cite Dal11 Table A2 and A3 [-]*/

    // Parameter necessary to calculate the transfer functions G or, if no transfer function is required, the RFref/Eref values (see calculateEmissions.cpp)
    /* CO2 parameter values and distribution, s. \cite Dal11 Table A1, for equation (6)*/
    const double A_CO2 = 1.8E-15; /**< Parameter to calculate the transfer function of CO2 [W/m2/kgCO2] */
    const double alphaC1 = 0.259;/**< Parameter to calculate the transfer function of CO2 according to \cite Dal11 Table A1 [-] */
    const double alphaC2 = 0.338;/**< Parameter to calculate the transfer function of CO2 according to \cite Dal11 Table A1 [-] */
    const double alphaC3 = 0.186;/**< Parameter to calculate the transfer function of CO2 according to \cite Dal11 Table A1[-] */
    const double tauC1 = 172.9; /**< Parameter to calculate the transfer function of CO2 according to \cite Dal11 Table A1 [yr] */
    const double tauC2 = 18.51; /**< Parameter to calculate the transfer function of CO2 according to \cite Dal11 Table A1 [yr] */
    const double tauC3 = 1.186; /**< Parameter to calculate the transfer function of CO2 according to \cite Dal11 Table A1 [yr] */
    /*NOx parameter values and distribution, s. \cite Dal11 Table A2, for equation (7) or equation (10)*/
    const double A_CH4 = -5.16E-13; /**< Parameter to calculate the transfer function of CH4 according to \cite Dal11 Table A2 [W/m2/kgNOx] */
    const double A_O3L = -1.21E-13; /**< Parameter to calculate the transfer functions of O3L according to \cite Dal11 Table A2[W/m2/kgNOx] */
    const double tauN = 12.; /**< Parameter to calculate the transfer functions of CH4 and O3L according to \cite Dal11 Table A2 [yr] */
    const double RFrefEref_O3S = 1.01E-11; /**< Radiative forcing per emission for O3S according to \cite Dal11 Table A2 [W/m^2/kgNOx] */
    /* H2O, So4, soot and AIC parameter values and distributions, s. \cite Dal11 Table A3, for equation (10)*/
    const double RFrefEref_H2O = 7.63E-15; /**< Radiative forcing per emission for H2O according to \cite Dal11 Table A3 [W/m^2/kgH2] */
    const double RFrefEref_SO4 = -2.01E-11; /**< Radiative forcing per emission for SO4 according to \cite Dal11 Table A3 [W/m^2/kgSO4] */
    const double RFrefEref_soot = 4.48E-10; /**< Radiative forcing per emission for soot according to \cite Dal11 Table A3 [W/m^2/kgSoot] */
    const double RFrefLref_AIC = 2.21E-12; /**< Radiative forcing per flight length for AIC according to \cite Dal11 Table A3 [W/m^2/NM] */
    /*Temperature changes model parameters values, s. \cite Dal11 Table A4*/
    const double RF_2CO2 = 3.7; /**< Parameter for temperature change model according to \cite Dal11 Table A4 [W/m^2] */
    const double climate_sensitivity_parameter = 3.; /**< Climate sensitivity parameter S according to \cite Dal11 Table A4 [K]*/
    const double alphaT = 0.595; /**< Parameter for temperature change model according to \cite Dal11 Table A4 [-] */
    const double tauT1 = 8.4;  /**< Parameter for temperature change model according to \cite Dal11 Table A4 [yr] */
    const double tauT2 = 409.5;  /**< Parameter for temperature change model according to \cite Dal11 Table A4 [yr] */
};

#endif //ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_IMPACT_CALCULATION_CLIMATE_MODEL_DALLARA_DALLARA_H_
