/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "dallaraIOData.h"
#include "dallaraConfig.h"
#include <aixml/node.h>
#include <aixml/endnode.h>
#include <standardFiles/functions.h>
#include <unitConversion/unitConversion.h>
#include <vector>


DallaraIOData::DallaraIOData(const node& acXML, const node& config, std::shared_ptr<node> ecoAssXML)
    :
    configuration(DallaraConfig()),
    range_study(0.),
    top_of_climb_range(0.),
    top_of_descent_range(0.),
    flights_per_year(0),
    relative_end_of_cruise_step(),
    flight_level(),
    ATR(0.) {
        configuration.read_all_config_data(config);
        read_aircraft(acXML);
        read_ecological_assessment_xml(*ecoAssXML);
    }

/* Input data methods*/
void DallaraIOData::read_aircraft(const node& acXML) {
    range_study = EndnodeReadOnly<double>("requirements_and_specifications/requirements/top_level_aircraft_requirements/study_mission/range").read(acXML).value();
    top_of_climb_range = EndnodeReadOnly<double>("analysis/mission/study_mission/cruise/top_of_climb_range").read(acXML).value();
    top_of_descent_range = EndnodeReadOnly<double>("analysis/mission/study_mission/cruise/top_of_descent_range").read(acXML).value();
    flights_per_year = EndnodeReadOnly<int>("assessment/cost_estimation/operating_cost/direct_operating_cost/flights_per_year_study_mission").read(acXML).value();
    duration_operation = EndnodeReadOnly<int>("requirements_and_specifications/assessment_scenario/duration_operation").read(acXML).value_xml();
    emissionMass[7] = convertUnit(METER, NAUTICALMILE, range_study); //AIC; Climateimpact over flown length [nm] not over emitted emissions
    int number_of_cruise_steps = acXML.getVector("aircraft_exchange_file/analysis/mission/study_mission/cruise/cruise_steps/cruise_step").size();
    for (int i = 0; i < number_of_cruise_steps; i++) {
        EndnodeReadOnly<double> cruise_step_end("analysis/mission/study_mission/cruise/cruise_steps/cruise_step@" + num2Str(i) + "/relative_end_of_cruise_step");
        relative_end_of_cruise_step.push_back(cruise_step_end.read(acXML).value());
        EndnodeReadOnly<double> cruise_step_altitude("analysis/mission/study_mission/cruise/cruise_steps/cruise_step@" + num2Str(i) + "/altitude");
        flight_level.push_back(convertUnit(METER, FOOT, cruise_step_altitude.read(acXML).value())/100.);
    }
}

void DallaraIOData::read_ecological_assessment_xml(const node& ecoAssXML) {
    /* Fill std::vector emissionMass: emitted emissions per mission or for AIC only route length*/ //
    std::string emissions_path("ecological_data_file/mission_emissions/study_mission/emissions/");
    emissionMass[0] = EndnodeReadOnly<double>("LTO_cycle/CO2").read(ecoAssXML.at(emissions_path)).value() //CO2
                        + EndnodeReadOnly<double>("cruise/CO2").read(ecoAssXML.at(emissions_path)).value();
    emissionMass[1] = EndnodeReadOnly<double>("LTO_cycle/NOx").read(ecoAssXML.at(emissions_path)).value() //CH4 (via NOx)
                        + EndnodeReadOnly<double>("cruise/NOx").read(ecoAssXML.at(emissions_path)).value();
    emissionMass[2] = EndnodeReadOnly<double>("LTO_cycle/NOx").read(ecoAssXML.at(emissions_path)).value() //O3L (via NOx)
                        + EndnodeReadOnly<double>("cruise/NOx").read(ecoAssXML.at(emissions_path)).value();
    emissionMass[3] = EndnodeReadOnly<double>("LTO_cycle/NOx").read(ecoAssXML.at(emissions_path)).value() //O3S (via NOx)
                        + EndnodeReadOnly<double>("cruise/NOx").read(ecoAssXML.at(emissions_path)).value();
    emissionMass[4] = EndnodeReadOnly<double>("LTO_cycle/H2O").read(ecoAssXML.at(emissions_path)).value() //H2O
                        + EndnodeReadOnly<double>("cruise/H2O").read(ecoAssXML.at(emissions_path)).value();
    emissionMass[5] = EndnodeReadOnly<double>("LTO_cycle/soot").read(ecoAssXML.at(emissions_path)).value() //Soot
                        + EndnodeReadOnly<double>("cruise/soot").read(ecoAssXML.at(emissions_path)).value();
    emissionMass[6] = EndnodeReadOnly<double>("LTO_cycle/SO4").read(ecoAssXML.at(emissions_path)).value() //SO4
                        + EndnodeReadOnly<double>("cruise/SO4").read(ecoAssXML.at(emissions_path)).value();
}

/* Output data methods*/
void DallaraIOData::update_aircraft_xml(std::shared_ptr<RuntimeIO> rtIO) {
    Endnode<double> ATRnode("average_temperature_response",
                            "Integrated temperature change per year caused by aircraft operation divided by operating lifetime",
                            ATR, "K", 0, 1e-5);
    std::string subpath("assessment/ecological_assessment/");
    if (!(rtIO->acxml.find(subpath))) { // if ecological_assessment node doesn't exist, add one
        rtIO->acxml[subpath];
        rtIO->acxml.at(subpath).addAttrib("description", "Ecological assessment data");
    }
    //ATRnode.update(rtIO->acxml.at("aircraft_exchange_file/assessment/ecological_assessment/"));
    rtIO->saveAcXML();
}

