/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include <iostream>
#include <memory>
#include <algorithm>
#include <string>
#include <filesystem>
#include <unitConversion/unitConversion.h>
#include <aixml/node.h>
#include "standard_strategy.h"
#include "../toolinfo.h"


void StandardStrategy::initialize() {
    myRuntimeInfo->out << "Initialize standard strategy" << std::endl;
    /* Read calculation methods from config */
    life_cycle_emissions_method = EndnodeReadOnly<std::string>("method").read(rtIO->moduleConfig.at("program_settings/"
                                + std::string("standard_strategy/emission_calculation/life_cycle_emissions_methods/"))).value();
    myRuntimeInfo->out << "  Method " << life_cycle_emissions_method << " will be used for life cycle emission calculation" << std::endl;
    climate_model_method = EndnodeReadOnly<std::string>("method").read(rtIO->moduleConfig.at("program_settings/"
                                + std::string("standard_strategy/impact_calculation/climate_model_methods/"))).value();
    myRuntimeInfo->out << "  Method " << climate_model_method << " will be used for climate model" << std::endl;
    air_quality_method = EndnodeReadOnly<std::string>("method").read(rtIO->moduleConfig.at("program_settings/"
                                + std::string("standard_strategy/impact_calculation/air_quality_methods/"))).value();
    myRuntimeInfo->out << "  Method " << air_quality_method << " will be used for air quality calculation" << std::endl;

    /* Prepare ecological_assessment.xml */
    if (!std::filesystem::exists(rtIO->getXmlDir())) std::filesystem::create_directory(rtIO->getXmlDir());
    std::string filepath = rtIO->getXmlDir() + FILESEPERATOR + TOOL_NAME + "_results.xml";
    if (fileExists(filepath)) { // Delete old calculation results of this phase
        ecological_assessment_xml = aixml::openDocument(std::filesystem::path(filepath));
        node* node = ecological_assessment_xml->find("/ecological_data_file");
        if (node != nullptr) {
            node->deleteChildren();
        }
    } else {
        ecological_assessment_xml = std::make_shared<node>(filepath);
        ecological_assessment_xml->appendChild("ecological_data_file");
    }

    /* Initialize Report */
    report_ = std::shared_ptr<Report>(new Report(rtIO));
}

void StandardStrategy::run() {
    myRuntimeInfo->out << "Run standard strategy" << std::endl;

    /* Mission emission calculation */
    theDesignMission.run("design", rtIO, ecological_assessment_xml);
    theStudyMission.run("study", rtIO, ecological_assessment_xml);

    /* Life cycle emission calculation */
    if (life_cycle_emissions_method == "NO") {
    } else if (life_cycle_emissions_method == "SCHAEFER") {
        theLifeCycleEmissionsSchaefer.run(rtIO, ecological_assessment_xml);
    } else {
        myRuntimeInfo->warn << "Life cycle emission method" << life_cycle_emissions_method << " unknow. Please check settings in config file!" << std::endl;
    }

    /* Climate model */
    if (climate_model_method == "NO") {
    } else if (climate_model_method == "DALLARA") {
        theClimateModelDallara.run(rtIO, ecological_assessment_xml);
    } else {
        myRuntimeInfo->warn << "Climate model method" << climate_model_method << " unknown. Please check settings in config file!" << std::endl;
    }

    /* Air quality */
    if (air_quality_method == "NO") {
    } else if (air_quality_method == "SCHAEFER") {
        theAQISchaefer.run(rtIO, ecological_assessment_xml);
    } else {
        myRuntimeInfo->warn << "Air quality calculation method " << air_quality_method << " unknown. Please check settings in config file!" << std::endl;
    }
}

void StandardStrategy::update() {
    myRuntimeInfo->out << "Update " + rtIO->acxmlAccess.filename().string() << std::endl;
    theClimateModelDallara.dallaraIOData->update_aircraft_xml(rtIO);
}

void StandardStrategy::report() {
    /* Plots */
    if (rtIO->plotOn) {
        myRuntimeInfo->out << "Generate plots" << std::endl;
        myRuntimeInfo->out << "  Generate study mission plots" << std::endl;
        generate_mission_emissions_plot("study");
        myRuntimeInfo->out << "  Generate design mission plots" << std::endl;
        generate_mission_emissions_plot("design");
        if (life_cycle_emissions_method == "SCHAEFER" && rtIO->aircraft_energy_carrier_type() == "kerosene" && rtIO->aircraft_configuration_type() == "tube_and_wing") {
            myRuntimeInfo->out << "  Generate " << life_cycle_emissions_method << " emissions plots" << std::endl;
            generate_plots_schaefer();
        }
        if (climate_model_method == "DALLARA") {
            myRuntimeInfo->out << "  Generate " << climate_model_method << " plots " << std::endl;
            generate_plots_dallara();
        }
    }

    /* Reports */
    myRuntimeInfo->out << "Write reports" << std::endl;
    report_->htmlReportStream() << html::div_start("class=\"box\""); // start html box
    myRuntimeInfo->out << "  Write mission report body" << std::endl;
    set_html_body_mission(rtIO);
    set_tex_body_mission();
    if (life_cycle_emissions_method == "SCHAEFER" && rtIO->aircraft_energy_carrier_type() == "kerosene" && rtIO->aircraft_configuration_type() == "tube_and_wing") {
        myRuntimeInfo->out << "  Write " << life_cycle_emissions_method << " emissions report body" << std::endl;
        set_html_body_LCA_schaefer(rtIO);
        set_tex_body_LCA_schaefer(rtIO);

        generate_life_cycle_emissions_csv_schaefer();
    }
    if (climate_model_method == "DALLARA") {
        myRuntimeInfo->out << "  Write " << climate_model_method << " climate report body" << std::endl;
        set_html_body_dallara(rtIO);
        set_tex_body_dallara();
    }
    if (air_quality_method == "SCHAEFER") {
        myRuntimeInfo->out << "  Write " << air_quality_method << " air quality report body" << std::endl;
        set_html_body_AQI_schaefer();
        set_tex_body_AQI_schaefer(rtIO);
    }
    report_->htmlReportStream() << html::div_end; // end html box
    report_->generateReports();
}

void StandardStrategy::save() {
    aixml::saveDocument(*ecological_assessment_xml);
}

