/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include <moduleBasics/colors.h>
#include "plot_utility.h"

std::tuple<std::vector<double>, std::vector<double>, std::string> gen_polygon(
    const geom2::MultisectionSurface<geom2::AirfoilSection>& surface,
    const geom2::MultisectionSurface<geom2::PolygonSection>& object) {
  double half_span = surface.is_symmetric ? 0.5 * geom2::measure::span(surface) : geom2::measure::span(surface);
  /* Spar coordinates positive values normally*/
  double from = object.sections.at(0).origin.z() * half_span;
  double to = object.sections.at(1).origin.z() * half_span;

  /* Chordwise coordinates */
  double chord_from_fwd = object.sections.at(0).get_contour().vertex(0).x();
  double chord_from_aft = object.sections.at(0).get_contour().vertex(1).x();
  double chord_to_fwd = object.sections.at(1).get_contour().vertex(0).x();
  double chord_to_aft = object.sections.at(1).get_contour().vertex(1).x();

  auto f_chord_fwd = [chord_from_fwd, chord_to_fwd, half_span](double span_position) {
    return std::lerp(chord_from_fwd, chord_to_fwd, fabs(span_position) / half_span);
  };
  auto f_chord_aft = [chord_from_aft, chord_to_aft, half_span](double span_position) {
    return std::lerp(chord_from_aft, chord_to_aft, fabs(span_position) / half_span);
  };
  /* Add object from */
  std::vector<double> spanwise_coordinates{-fabs(from)};

  /* Add intermediate sections */
  for (auto& section : surface.sections) {
    if (fabs(section.origin.z()) > fabs(from) && fabs(section.origin.z()) < fabs(to)) {
      spanwise_coordinates.push_back(section.origin.z());
    }
  }
  /* Add object to */
  spanwise_coordinates.push_back(-fabs(to));

  /**/
  std::vector<double> x;
  std::vector<double> y;
  /* From in to out*/
  for (auto it = spanwise_coordinates.begin(); it != spanwise_coordinates.end(); ++it) {
    y.push_back(fabs(*it));
    x.push_back(geom2::measure::offset_LE(surface, *it).x() + f_chord_fwd(*it) * geom2::measure::chord(surface, *it));
  }
  /* From out to in*/
  for (auto rit = spanwise_coordinates.rbegin(); rit != spanwise_coordinates.rend(); ++rit) {
    y.push_back(fabs(*rit));
    x.push_back(geom2::measure::offset_LE(surface, *rit).x() +
                f_chord_aft(*rit) * geom2::measure::chord(surface, *rit));
  }

  /* Close loop */
  y.push_back(y.front());
  x.push_back(x.front());

  return {x, y, object.name};
}


std::array<float,4UL> get_spar_colors()
{
  return colors::to_matplot_color({99, 110, 114}, 0.0);
}

std::array<float, 4UL> get_control_device_colors(const std::string& control_device)
{
  std::map<std::string, std::array<float, 4UL>> control_device_colors;
  control_device_colors.insert({"elevator",colors::to_matplot_color({108, 92, 231}, 0.0)});
  control_device_colors.insert({"rudder",colors::to_matplot_color({232, 67, 147}, 0.0)});

  return control_device_colors.at(control_device);
}
