/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef LOW_CONVENTIONAL_TAW_H_
#define LOW_CONVENTIONAL_TAW_H_

#include <aircraftGeometry2/airfoil_surface.h>
#include <airfoils/airfoils.h>
#include <moduleBasics/module.h>
#include <moduleBasics/report.h>
#include <moduleBasics/strategySelector.h>

#include <functional>
#include <map>
#include <vector>

#include "conventionalEmpennageDesignConfig.h"
#include "conventionalEmpennageDesignData.h"

namespace taw {
namespace low {
class Conventional : public Strategy {
 public:
  Conventional(const std::shared_ptr<RuntimeIO>& rtIO);
  ~Conventional() = default;
  /**
   * \brief Initialize conventional method
   * 
   */
  void initialize();

  /**
   * \brief Run conventional method
   * 
   */
  void run();

  /**
   * \brief Update conventional method
   * 
   */
  void update();

  /**
   * \brief Report generation for conventional method
   * 
   */
  void report();

  /**
   * \brief Save results for conventional method
   * 
   */
  void save();

  /**
   * \brief Design method for conventional empennage
   * 
   */
  void design(void);

  /**
   * \brief Redesign method for conventional empennage
   * 
   */
  void redesign(void);

  /**
   * \brief Design of a vertical stabilizer
   * 
   * \param wing_sweep 
   * \param wing_reference_area 
   * \param wing_span 
   * \param wing_taper_ratio 
   * \param wing_aspect_ratio 
   * \param wing_neutral_position 
   * \param fuselage_length 
   */
  void design_vertical_stabilizer(const double wing_sweep, const double wing_reference_area, const double wing_span,
                                  const double wing_taper_ratio, const double wing_aspect_ratio,
                                  geom2::Point_3& wing_neutral_position, const double fuselage_length);

  /**
   * \brief Design of a horizontal stabilizer
   * 
   * \param wing_sweep 
   * \param wing_reference_area 
   * \param wing_mac 
   * \param wing_taper_ratio 
   * \param wing_aspect_ratio 
   * \param wing_neutral_position 
   * \param fuselage_length 
   */
  void design_horizontal_stabilizer(const double wing_sweep, const double wing_reference_area, const double wing_mac,
                                    const double wing_taper_ratio, const double wing_aspect_ratio,
                                    geom2::Point_3& wing_neutral_position, const double fuselage_length);

  /**
   * \brief Redesign of a vertical stabilizer
   * 
   * \param wing_sweep 
   * \param wing_reference_area 
   * \param wing_span 
   * \param wing_neutral_position 
   * \param fuselage_length 
   */
  void redesign_vertical_stabilizer(const double wing_sweep, const double wing_reference_area, const double wing_span,
                                    geom2::Point_3& wing_neutral_position, const double fuselage_length);

  /**
   * \brief Redesign of a horizontal stabilizer
   * 
   * \param wing_sweep 
   * \param wing_reference_area 
   * \param wing_span 
   * \param wing_neutral_position 
   * \param fuselage_length 
   */
  void redesign_horizontal_stabilizer(const double wing_sweep, const double wing_reference_area, const double wing_span,
                                      geom2::Point_3& wing_neutral_position, const double fuselage_length);

  /**
   * \brief Design of a stabilizer
   * 
   * \param wing_sweep_rad 
   * \param wing_reference_area 
   * \param reference_length 
   * \param wing_neutral_position 
   * \param fuselage_length 
   * \param orientation 
   * \param volume_coefficient 
   * \param delta_sweep_rad 
   * \param stabilizer_taper_ratio 
   * \param stabilizer_aspect_ratio 
   * \param offset_position 
   * \param inner_profile_name 
   * \param outer_profile_name 
   * \param symmetric 
   * \param number_of_stabilizers 
   * \return geom2::MultisectionSurface<geom2::AirfoilSection> 
   */
  geom2::MultisectionSurface<geom2::AirfoilSection> design_stabilizer(
      const double wing_sweep_rad, const double wing_reference_area, const double reference_length,
      geom2::Point_3& wing_neutral_position, const double fuselage_length, geom2::Direction_3& orientation,
      const double volume_coefficient, const double delta_sweep_rad, const double stabilizer_taper_ratio,
      const double stabilizer_aspect_ratio, geom2::Point_3& offset_position, const std::string& inner_profile_name,
      const std::string& outer_profile_name, bool symmetric, int number_of_stabilizers);

  /**
   * \brief Design of a stabilizer
   * 
   * \param wing_sweep_rad 
   * \param wing_reference_area 
   * \param reference_length 
   * \param wing_taper_ratio 
   * \param wing_aspect_ratio 
   * \param wing_neutral_position 
   * \param fuselage_length 
   * \param tail 
   * \param align 
   * \param symmetric 
   * \return geom2::MultisectionSurface<geom2::AirfoilSection> 
   */
  geom2::MultisectionSurface<geom2::AirfoilSection> design_stabilizer(
      const double wing_sweep_rad, const double wing_reference_area, const double reference_length,
      const double wing_taper_ratio, const double wing_aspect_ratio, geom2::Point_3& wing_neutral_position,
      const double fuselage_length, TailElement& tail, geom2::Direction_3& align, bool symmetric);
  
  /**
   * \brief Set the empennage position 
   * 
   */
  void set_empennage_position(void);

  /**
   * \brief Method to create shape of a control surface
   * 
   * \param name name of the control surface
   * \param inner_chord_rel_pos_from 
   * \param inner_chord_rel_pos_to 
   * \param inner_spanwise_rel_pos 
   * \param outer_chord_rel_pos_from 
   * \param outer_chord_rel_pos_to 
   * \param outer_spanwise_rel_pos 
   * \return geom2::MultisectionSurface<geom2::PolygonSection> 
   */
  geom2::MultisectionSurface<geom2::PolygonSection> control_device(
      const std::string& name, double inner_chord_rel_pos_from, double inner_chord_rel_pos_to,
      double inner_spanwise_rel_pos, double outer_chord_rel_pos_from, double outer_chord_rel_pos_to,
      double outer_spanwise_rel_pos);
  
  /**
   * \brief Method to create shape of a predefined spar 
   * 
   * \param name Name of the spar
   * \param inner_chord_rel_pos_from  
   * \param inner_chord_rel_pos_to 
   * \param inner_spanwise_rel_pos 
   * \param outer_chord_rel_pos_from 
   * \param outer_chord_rel_pos_to 
   * \param outer_spanwise_rel_pos 
   * \return geom2::MultisectionSurface<geom2::PolygonSection> spar element as a polygonsection 
   */
  geom2::MultisectionSurface<geom2::PolygonSection> spar(const std::string& name, double inner_chord_rel_pos_from,
                                                             double inner_chord_rel_pos_to,
                                                             double inner_spanwise_rel_pos,
                                                             double outer_chord_rel_pos_from,
                                                             double outer_chord_rel_pos_to,
                                                             double outer_spanwise_rel_pos);
  /**
   * \brief Add spars method 
   * 
   */
  void add_spars();

  /**
   * \brief Calculate empennage mass (vertical stabilizer / horizontal stabilizer) and associated center of gravity by
   *        Flops method
   * 
   */
  void flops_mass();

  /**
   * \brief Calculate empennage mass by raymer method - not implemented yet
   * 
   */
  void raymer_mass();

  /**
   * \brief Set the html body report
   * 
   */
  void set_html_body();

  /**
   * \brief Generates plots for stabilizer planform and returns relative path to plot (svg)
   * 
   * \param surface stabilizer 
   * \param spars spars vector
   * \param devices control devices vector
   * \return fs::path relative path between report and plot
   */
  fs::path plot_stabilizer_planform(const geom2::MultisectionSurface<geom2::AirfoilSection>& surface,
                                    const std::vector<geom2::MultisectionSurface<geom2::PolygonSection>>& spars,
                                    const std::vector<geom2::MultisectionSurface<geom2::PolygonSection>>& devices);

  /**
   * \brief Generate unique plots of airfoils 
   * 
   * \param surface stabilizer
   * \return std::vector<fs::path> vector of relative paths between report and plots
   */
  std::vector<fs::path> plot_airfoils(const geom2::MultisectionSurface<geom2::AirfoilSection>& surface);

  /**
   * \brief Generate plot of thickness and twist distribution 
   * 
   * \param surface stabilizer
   * \return fs::path relative path between report and plot
   */
  fs::path plot_thickness_and_twist_distribution(const geom2::MultisectionSurface<geom2::AirfoilSection>& surface);

public:
  std::string selected_design_mode;
  std::string selected_mass_mode;
  std::map<std::string, std::function<void(void)>> design_mode_runner;
  std::map<std::string, std::function<void(void)>> mass_mode_runner;
  std::shared_ptr<Airfoils> airfoils_library;


  const std::shared_ptr<RuntimeIO>& rtIO;
  const std::shared_ptr<taw::low::ConventionalConfig> config;
  const std::shared_ptr<taw::low::ConventionalData> data;
  Report reporter;
};

}  // namespace low
}  // namespace taw

#endif  // LOW_CONVENTIONAL_TAW_H_