/**
 * \file T_TAILEmpennageDesignData.cpp
 * \author Christopher Ruwisch (christopher.ruwisch@tu-berlin.de)
 * \brief Tube and Wing T_TAIL Data source file
 * \version 0.1
 * \date 2023-11-24
 *
 * @copyright Copyright (c) 2023
 *
 */

#include "t_tail_empennage_design_data.h"

#include <aircraftGeometry2/airfoil_surface.h>
#include <aircraftGeometry2/fuselage.h>
#include <aircraftGeometry2/geometry/surface.h>

namespace taw {
namespace low {

T_Tail_Data::T_Tail_Data()
    : maximum_takeoff_mass(
          EndnodeReadOnly<double>("analysis/masses_cg_inertia/maximum_takeoff_mass/mass_properties/mass")),
      empennage_mass(MassPropertiesIO("component_design/empennage/", "T_TAIL tail overall mass")),
      vertical_stabilizer_mass(MassPropertiesIO("component_design/empennage/specific/geometry/aerodynamic_surface@0/",
                                                "vertical stabiliser mass")),
      horizontal_stabilizer_mass(MassPropertiesIO("component_design/empennage/specific/geometry/aerodynamic_surface@1/",
                                                  "horizontal stabiliser mass")),
      wing_position_x(EndnodeReadOnly<double>("component_design/wing/position/x")),
      empennage_position(PositionIO("component_design/empennage/position", "empennage")) {}

void T_Tail_Data::read(const node &xml) {
  wing_position_x.read(xml);
  maximum_takeoff_mass.read(xml);
}

void T_Tail_Data::read_fuselage(std::filesystem::path path, std::filesystem::path geometry_data_dir) {
  std::shared_ptr<node> acxml = aixml::openDocument(path);
  geom2::FuselageFactory fuselage_factory{acxml, geometry_data_dir};
  fuselage = fuselage_factory.create("fuselage/specific/geometry/fuselage@0");
}

void T_Tail_Data::read_wing(std::filesystem::path path, std::filesystem::path airfoil_data_dir) {
  std::shared_ptr<node> acxml = aixml::openDocument(path);
  geom2::WingFactory wing_factory{acxml, airfoil_data_dir};
  wing = wing_factory.create("wing/specific/geometry/aerodynamic_surface@0");
}

void T_Tail_Data::update_vertical_stabilizer(node &xml) {
  geom2::io::SurfaceType io = geom2::io::Wing(vertical_stabilizer);
  const std::string path_to_geometry = "component_design/empennage/specific/geometry";
  std::visit(geom2::io::AixmlConverter(xml[path_to_geometry], {"aerodynamic_surface", "0", vertical_stabilizer.name}),
             io);
}

void T_Tail_Data::update_vertical_stabilizer_spars(node &xml) {
  int i = 0;
  for (auto &spar : vertical_stabilizer_spars) {
    geom2::io::SurfaceType io_spar = geom2::io::Spar(spar);
    const std::string path_to_spars =
        "component_design/empennage/specific/geometry/aerodynamic_surface@0/parameters/spars";
    std::visit(geom2::io::AixmlConverter(xml[path_to_spars], {"spar", std::to_string(i++), spar.name}), io_spar);
  }
}

void T_Tail_Data::update_vertical_stabilizer_control_devices(node &xml) {
  int i = 0;
  for (auto control_device : vertical_stabilizer_control_devices) {
    geom2::io::SurfaceType io_control_device = geom2::io::ControlDevice(control_device);
    std::string path_to_device =
        "component_design/empennage/specific/geometry/aerodynamic_surface@0/parameters/control_devices";
    std::visit(
        geom2::io::AixmlConverter(xml[path_to_device], {"control_device", std::to_string(i), control_device.name}),
        io_control_device);
    Endnode<double> min_deflection(
        path_to_device + "/control_device@" + std::to_string(i) + "/deflection/full_negative_deflection",
        "full negative deflection");
    Endnode<double> max_deflection(
        path_to_device + "/control_device@" + std::to_string(i) + "/deflection/full_positive_deflection",
        "full positive deflection");
    const auto [min, max] = vertical_stabilizer_control_devices_deflections[i];
    min_deflection.set_unit("rad");
    min_deflection.set_value(min);
    min_deflection.update(xml);
    max_deflection.set_unit("rad");
    max_deflection.set_value(max);
    max_deflection.update(xml);
    i++;
  }
}

void T_Tail_Data::update_horizontal_stabilizer(node &xml) {
  geom2::io::SurfaceType io = geom2::io::Wing(horizontal_stabilizer);
  const std::string path_to_geometry = "component_design/empennage/specific/geometry";
  std::visit(geom2::io::AixmlConverter(xml[path_to_geometry], {"aerodynamic_surface", "1", horizontal_stabilizer.name}),
             io);
}

void T_Tail_Data::update_horizontal_stabilizer_spars(node &xml) {
  int i = 0;
  for (auto &spar : horizontal_stabilizer_spars) {
    geom2::io::SurfaceType io_spar = geom2::io::Spar(spar);
    const std::string path_to_spars =
        "component_design/empennage/specific/geometry/aerodynamic_surface@1/parameters/spars";
    std::visit(geom2::io::AixmlConverter(xml[path_to_spars], {"spar", std::to_string(i++), spar.name}), io_spar);
  }
}

void T_Tail_Data::update_horizontal_stabilizer_control_devices(node &xml) {
  int i = 0;
  for (auto &control_device : horizontal_stabilizer_control_devices) {
    geom2::io::SurfaceType io_control_device = geom2::io::ControlDevice(control_device);
    const std::string path_to_device =
        "component_design/empennage/specific/geometry/aerodynamic_surface@1/parameters/control_devices";
    std::visit(
        geom2::io::AixmlConverter(xml[path_to_device], {"control_device", std::to_string(i), control_device.name}),
        io_control_device);
    Endnode<double> min_deflection(
        path_to_device + "/control_device@" + std::to_string(i) + "/deflection/full_negative_deflection",
        "full negative deflection");
    Endnode<double> max_deflection(
        path_to_device + "/control_device@" + std::to_string(i) + "/deflection/full_positive_deflection",
        "full positive deflection");
    const auto [min, max] = horizontal_stabilizer_control_devices_deflections[i];
    min_deflection.set_unit("rad");
    min_deflection.set_value(min);
    min_deflection.update(xml);
    max_deflection.set_unit("rad");
    max_deflection.set_value(max);
    max_deflection.update(xml);
    i++;
  }
}

void T_Tail_Data::update(node &xml) {
  auto tmp = xml.find("component_design/empennage");
  if (tmp != nullptr) {
    tmp->deleteChildren();
  }
  empennage_position.update(xml);
  empennage_mass.update(xml);
  update_vertical_stabilizer(xml);
  update_vertical_stabilizer_spars(xml);
  update_vertical_stabilizer_control_devices(xml);
  vertical_stabilizer_mass.update(xml);
  update_horizontal_stabilizer(xml);
  update_horizontal_stabilizer_spars(xml);
  update_horizontal_stabilizer_control_devices(xml);
  horizontal_stabilizer_mass.update(xml);
}

}  // namespace low
}  // namespace taw
