# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2024 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

# Import standard modules.
import sys
import xml.etree.ElementTree as ET

# Import own modules.
from src.tube_and_wing.empirical.fuselage_design_tu_berlin.general.fuselagedesign import fuselage_design  # noPep8 e501
#from src.wing_mounted.empirical.landing_gear_design_tu_berlin.general.call_functions._07_estimatecog import estimate_cog
#from src.wing_mounted.empirical.landing_gear_design_tu_berlin.general.plot_functions.generateacnplots import generate_acn_plots  # noPep8 e501
#from src.wing_mounted.empirical.landing_gear_design_tu_berlin.general.call_functions._06_estimatemasses import estimate_masses  # noPep8 e501
#from src.wing_mounted.empirical.landing_gear_design_tu_berlin.general.call_functions._08_setexistinggeometry import set_existing_geometry  # noPep8 e501
#from src.wing_mounted.empirical.landing_gear_design_tu_berlin.general.plot_functions.generatelandinggearplots import generate_landing_gear_plots  # noPep8 e501
#from src.wing_mounted.empirical.landing_gear_design_tu_berlin.general.aicraft_classification_number.estiamteaircraftclassificationnumber import estimate_aircraft_classification_number  # noPep8 e501


def use_existing_geometry(paths_and_names, routing_dict, dict_ac_exchange, dict_mod_config, runtime_output):
    """ Use existing geometry prepares the fuselage design module to use a fuselage geometry from an external
         file.

        :param dict paths_and_names: Dictionary containing system paths and ElementTrees
        :param dict routing_dict: Dictionary containing routing parameters
        :param dict dict_ac_exchange: Dict containing parameter and according values from aircraft exchange file
        :param dict dict_mod_config: Dict containing parameter and according values from module configuration file
        :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
        :return dict kerosene_output_dict: Dict containing output parameter and values of landing gear estimation
    """
    ''' initialize local parameter '''
    existing_x_position = []

    ''' initial operation for use existing geometry mode '''
    # Try to read the existing geometry xml file of landing gear design.
    try:
        # Attempt to create an ElementTree and get the root element from the XML file.
        existing_geometry_tree = ET.ElementTree(file=paths_and_names['working_directory'] + '/'
                                                + dict_mod_config['name_of_existing_geometry_file'])
        root_of_existing_geometry_tree = existing_geometry_tree.getroot()

    # Exception handling for operating system (OS) error.
    except OSError:
        runtime_output.critical('Error: The file: "' + dict_mod_config['name_of_existing_geometry_file']
                                + '" was found but could not be read. \n'
                                '                                            Program aborted!')
        sys.exit('Exit information: Existing fuselage geometry file could not be read!')

    # Get landing gear assembly x-positions from existing geometry
    existing_elements = root_of_existing_geometry_tree.findall(
        './fuselage/specific/geometry/fuselage')
    # Check if the required x-positions are in the existing landing gear geometry
    #  -> if true: -> set to aircraft exchange dictionary and start further estimations.
    if existing_elements is not None:
        for element in existing_elements:
            existing_x_position.append(round(float(element.text), 4))

        # Check if at least three landing gear assembly are defined in the given existing geometry
        #  -> otherwise raise an error end abort program.
        if len(existing_elements) < 3:
            runtime_output.critical('Error: Not enough landing gear assembly has been defined in the existing '
                                    'landing gear geometry. \n'
                                    '                                            Program aborted!')
            sys.exit('Exit information: Missing required lansing gear assembly in the existing landing gear '
                     'geometry!')

        # Check if the current aircraft exchange file contains no landing gear parameters.
        #   -> if true: -> initialize landing gear position dictionary
        if len(dict_ac_exchange['landing_gear_positions_in_x_direction']) == 1:
            i = 0
            while i < len(existing_elements):
                name_of_parameter = 'landing_gear_positions_in_x_direction_ID' + str(i)
                dict_ac_exchange['landing_gear_positions_in_x_direction'][name_of_parameter] \
                    = existing_x_position[i]
                i += 1
        # Else condition: Landing gear parameter are already existing in the aircraft exchange file
        #  -> overwrite with data from existing geometry
        else:
            # Check if the given existing geometry has a lower number of landing gear assemblies as the number of
            # entries in the aircraft exchange file -> if true: -> cut off the aircraft exchange dictionary
            if len(dict_ac_exchange['landing_gear_positions_in_x_direction']) > len(existing_elements):
                truncated_dict = {k: v for i, (k, v) in
                                  enumerate(dict_ac_exchange['landing_gear_positions_in_x_direction'].items())
                                  if i < len(existing_elements)}
                dict_ac_exchange['landing_gear_positions_in_x_direction'] = truncated_dict

            count = 0
            for key, _ in dict_ac_exchange['landing_gear_positions_in_x_direction'].items():
                dict_ac_exchange['landing_gear_positions_in_x_direction'][key] = existing_x_position[count]
                count += 1

    # Else conditions: The required x-positions are not in the existing landing gear geometry
    #  -> raise an error and abort program.
    else:
        runtime_output.critical(
            'Error: The required x-positions for each landing gear strut are not found in the'
            'existing landing gear geometry. \n'
            '                                            Program aborted!')
        sys.exit('Exit information: Missing required parameter in the existing landing gear geometry!')

    ''' Run code for use existing geometry mode '''
    # Check if the existing geometry should be used for further estimation
    #  -> if true: -> call landing gear design with loaded data from geometry file
    if dict_mod_config['use_as_starting_point']:
        # Call function to run the landing gear design process with an existing geometry.
        kerosene_output_dict = fuselage_design(paths_and_names, routing_dict, dict_ac_exchange, dict_mod_config,
                                               runtime_output)

    # Else condition: Existing geometry should not be used for further estimation -> skip landing gear estimation steps
    else:
        runtime_output.warning('Attention! The Use existing geometry mode is selected for the landing gear design. '
                               'In this mode, clearances and limitations are not checked.')

        # Call function to set all necessary landing gear parameter from existing geometry file.
        kerosene_output_dict = set_existing_geometry(dict_ac_exchange, dict_mod_config, root_of_existing_geometry_tree,
                                                     runtime_output)

        # Call function to estimate existing geometry dependent landing gear masses.
        kerosene_output_dict = estimate_masses(dict_ac_exchange, dict_mod_config, kerosene_output_dict, runtime_output)

        # Call function to estimate existing geometry dependent landing gear CoGs.
        kerosene_output_dict = estimate_cog(dict_ac_exchange, dict_mod_config, kerosene_output_dict, runtime_output)

        ''' estimate aircraft classification number (acn) '''
        # initialize parameter to estimate ACN for flex pavement
        dict_acn_flex_parameter, dict_constants = \
            estimate_aircraft_classification_number(paths_and_names, kerosene_output_dict, dict_ac_exchange,
                                                    dict_mod_config, runtime_output)

        # visualisation of results
        dict_landing_gear_output = {}
        root_of_module_config_tree = paths_and_names['root_of_module_config_tree']
        plot_switch = (eval(root_of_module_config_tree.find('.//plot_output/enable/value').text.capitalize()))
        if plot_switch:
            dict_landing_gear_output = generate_landing_gear_plots(dict_ac_exchange, kerosene_output_dict,
                                                                   runtime_output)
            dict_acn_flex_parameter = generate_acn_plots(dict_acn_flex_parameter, dict_constants, runtime_output)

        kerosene_output_dict['acn_estimation'] = dict_acn_flex_parameter
        kerosene_output_dict['dict_landing_gear_output'] = dict_landing_gear_output

        runtime_output.debug('landing_gear_design with existing geometry successfully executed!')

    return kerosene_output_dict
