# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2024 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

# Import own modules.
from src.tube_and_wing.empirical.fuselage_design_tu_berlin.general.call_functions.fuselage_geometry.estimatecabinwidth import estimate_cabin_width  # noPep8 e501
from src.tube_and_wing.empirical.fuselage_design_tu_berlin.general.call_functions.fuselage_geometry.cabinslendernessratiocalculation import cabin_slenderness_ratio_calculation  # noPep8 e501
from src.tube_and_wing.empirical.fuselage_design_tu_berlin.general.call_functions.fuselage_geometry.cabinlengthcalculation import cabin_length_calculation  # noPep8 e501
from src.tube_and_wing.empirical.fuselage_design_tu_berlin.general.call_functions.fuselage_geometry.cabinwallthicknesscalculation import cabin_wall_thickness_calculation  # noPep8 e501
from src.tube_and_wing.empirical.fuselage_design_tu_berlin.general.call_functions.fuselage_geometry.cabinfloorthicknesscalculation import cabin_floor_thickness_calculation  # noPep8 e501
from src.tube_and_wing.empirical.fuselage_design_tu_berlin.general.call_functions.fuselage_geometry.fuselagedimensioncalculation import fuselage_dimension_calculation  # noPep8 e501


def generate_geometry(dict_ac_exchange, dict_mod_config, fuselage_description, runtime_output):
    """ Generates main geometry parameter of the aircraft fuselage.

        :param dict dict_ac_exchange: Dict containing parameter and according values from aircraft exchange file
        :param dict dict_mod_config: Dict containing parameter and according values from module configuration file
        :param dict fuselage_description: Dict containing parameter and values of current fuselage design
        :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
        :return dict fuselage_description: Dict containing parameter and values of current fuselage design
    """

    ''' initialize local parameter and dictionaries '''
    geometry_description = {'number_of_fuselages': 1,
                            'number_of_payload_tubes': 1,
                            'fuselage_dimension': {},
                            'total_cabin_width': float(),
                            'total_cabin_length': float(),
                            'cabin_slenderness_ratio': float(),
                            'cabin_wall_thickness': float(),
                            'cabin_floor_thickness': float(),
                            'cabin_lower_z': float(),
                            'cabin_inner_upper_z': float(),
                            'cabin_outer_upper_z': float(),
                            'geometry_flag': False
                            }

    ''' estimate cabin width '''
    # Function call for cabin width calculation.
    fuselage_description['geometry_description'] = estimate_cabin_width(dict_ac_exchange, dict_mod_config,
                                                                        geometry_description, runtime_output)

    ''' estimate cabin slenderness ratio '''
    # Function call for cabin slenderness ratio calculation.
    fuselage_description['geometry_description'], fuselage_description['interior_description'] \
        = cabin_slenderness_ratio_calculation(dict_ac_exchange, dict_mod_config, geometry_description, runtime_output)

    ''' estimate cabin length '''
    # Function call for cabin length calculation.
    fuselage_description['geometry_description'] = \
        cabin_length_calculation(fuselage_description['geometry_description'], runtime_output)

    ''' estimate cabin wall thickness '''
    # Function call for cabin wall thickness calculation.
    fuselage_description['geometry_description'] = \
        cabin_wall_thickness_calculation(fuselage_description['geometry_description'], runtime_output)

    ''' estimate cabin floor thickness '''
    # Function call for cabin floor thickness calculation.
    fuselage_description['geometry_description'] = \
        cabin_floor_thickness_calculation(fuselage_description['geometry_description'], runtime_output)

    ''' estimate fuselage diameter in y and z direction '''
    # Function call for fuselage diameter calculation
    fuselage_description['geometry_description'] = fuselage_dimension_calculation(dict_ac_exchange, dict_mod_config,
                                                                                  geometry_description,
                                                                                  runtime_output)

    ''' store geometry flag to description dictionary '''
    geometry_description['geometry_flag'] = True
    fuselage_description['geometry_description'] = geometry_description

    ''' debug print '''
    runtime_output.debug('Debug: The generate_geometry function was successfully executed!')

    return fuselage_description
