# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2024 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

# Import standard modules.
import sys
import math


def fuselage_dimension_calculation(dict_ac_exchange, dict_mod_config, geometry_description, runtime_output):
    """ Calculates of fuselage dimensions in x, y and z direction.

        References and calculation method:
            [1] Prof. Dr.-Ing. Juergen Thorbeck TU-Berlin (2006), Script "Flugzeugentwurf I und II" page: XYZ
                fuselage_diameter = cabin_width + 2 * wall_thickness [inch]

            [2] M.Sc. Andreas Gobbin (2015), Master Thesis "Numerische Modellierung des Auslegungsprozesses für
                                                            Passagierkabinen von Verkehrsflugzeugen unter
                                                            Berücksichtigung der wichtigsten Auslegungsforderungen und
                                                            Implementierung in MatLab"

    :param dict dict_ac_exchange: Dict containing parameter and according values from aircraft exchange file
    :param dict dict_mod_config: Dict containing parameter and according values from module configuration file
    :param dict geometry_description: Dict containing geometry parameter and values of current fuselage design
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict geometry_description: Dict containing geometry parameter and values of current fuselage design
    """
    ''' initialize local parameter '''
    # ToDo: ersetzen mit constants from lib
    convert_inch_to_meter = 0.0254
    convert_meter_to_inch = 39.370078740161

    max_height = 0.0
    max_width_at_base = 0.0
    fuselage_dimension = {'fuselage_length': float(),
                          'fuselage_width': float(),
                          'fuselage_upper_height': float(),
                          'fuselage_lower_height': float(),
                          'fuselage_total_height': float()
                          }

    ''' calculation of fuselage length in x-direction according to [2]'''
    cabin_length_in_meter = geometry_description["total_cabin_length"] * convert_inch_to_meter

    # Check the type of fuselage and use type associated regression for fuselage length estimation.
    if dict_ac_exchange['fuselage_type'] == "single_aisle":
        fuselage_dimension["fuselage_length"] = \
            round((cabin_length_in_meter /
                   (0.23482756 * math.log(float(cabin_length_in_meter)) - 0.05106017)
                   * convert_meter_to_inch), 2)

    elif dict_ac_exchange['fuselage_type'] == "wide_body":
        fuselage_dimension["fuselage_length"] = \
            round((cabin_length_in_meter / (0.1735 * math.log(float(cabin_length_in_meter)) + 0.0966)
                   * convert_meter_to_inch), 2)

    else:
        runtime_output.critical('Error: The selected fuselage type is not valid for current fuselage design method.\n'
                                '                                     Program aborted!')
        sys.exit('Fuselage type does not match the calculation method.')

    if dict_ac_exchange['additional_fuselage_length'] is not None:
        fuselage_dimension["fuselage_length"] = \
            round(fuselage_dimension["fuselage_length"]
                  + (dict_ac_exchange['additional_fuselage_length'] * convert_meter_to_inch), 2)

    # Calculation of fuselage diameter in y-direction according to [1].
    fuselage_dimension["fuselage_width"] = \
        round(geometry_description["total_cabin_width"] + 2 * geometry_description["cabin_wall_thickness"], 2)

    # Calculation of fuselage diameter in z-direction according to [1].
    uld_type = dict_mod_config['uld_type']
    # check which type of cargo accommodation is selected and extract dimensions depending on
    if uld_type == 'mixed':
        uld_types = dict_ac_exchange['relevant_cargo_accommodation']['uld_type']
        for key, value in uld_types.items():
            max_height = max(max_height, round(float(uld_types[key]['height']['value']), 2))
            max_width_at_base = max(max_width_at_base, round(float(uld_types[key]['width_at_base']['value']), 2))
    else:
        uld_definition = dict_ac_exchange['relevant_cargo_accommodation']['uld_type'].items()
        container_name = next(iter(uld_definition))[0]
        max_height = round(float(dict_ac_exchange['relevant_cargo_accommodation']['uld_type']
                                 [container_name]['height']['value']), 2)
        max_width_at_base = round(float(dict_ac_exchange['relevant_cargo_accommodation']['uld_type']
                                        [container_name]['width_at_base']['value']), 2)

    # Calculation of fuselage height in negative z-direction according to [1].
    distance_to_cargo_bottom = \
        round(max_height + geometry_description['cabin_floor_thickness']
              + float(dict_ac_exchange['dict_cs_requirements']['fuselage_design_cs_requirements']['container_distances']
                      ['container_to_ceiling_distance']['value']) + dict_mod_config['cabin_floor_offset'], 2)
    distance_to_lower_compartment_edge = \
        round(float(dict_ac_exchange['dict_cs_requirements']['fuselage_design_cs_requirements']['container_distances']
                    ['container_to_wall_distance']['value']) + 1 / 2 * max_width_at_base, 2)
    geometry_description["cabin_lower_z"] = round(pow(distance_to_lower_compartment_edge ** 2
                                                      + distance_to_cargo_bottom ** 2, 0.5), 2)

    fuselage_dimension["fuselage_lower_height"] = round(geometry_description["cabin_lower_z"]
                                                        + geometry_description["cabin_wall_thickness"], 2)

    # Calculation of fuselage height in positiv z-direction according to [1].
    geometry_description["cabin_inner_upper_z"] \
        = dict_mod_config['passenger_aisle_standing_height'] - dict_mod_config['cabin_floor_offset']
    geometry_description["cabin_outer_upper_z"] = \
        geometry_description["cabin_inner_upper_z"] + dict_mod_config['system_bay_height_above_cabin']
    fuselage_dimension["fuselage_upper_height"] = round(geometry_description["cabin_outer_upper_z"]
                                                        + geometry_description["cabin_wall_thickness"], 2)

    # Check if the fuselage upper height lower than the half of fuselage width
    #  -> if true: -> set fuselage upper height to half of fuselage width
    if dict_ac_exchange['fuselage_type'] == 'wide_body' \
            and fuselage_dimension["fuselage_upper_height"] < fuselage_dimension["fuselage_width"] / 2:
        fuselage_dimension["fuselage_upper_height"] = fuselage_dimension["fuselage_width"] / 2
        geometry_description["cabin_outer_upper_z"] = round(fuselage_dimension["fuselage_upper_height"]
                                                            - geometry_description["cabin_wall_thickness"], 2)

    # Calculation of fuselage total height
    fuselage_dimension["fuselage_total_height"] = round(fuselage_dimension["fuselage_upper_height"]
                                                        + fuselage_dimension["fuselage_lower_height"], 2)

    # Check for forced circular cross section -> if true: -> set fuselage width and height to max of both
    if dict_mod_config['force_circular_cross_section']:
        max_diameter = max(fuselage_dimension["fuselage_width"], fuselage_dimension["fuselage_total_height"])
        fuselage_dimension["fuselage_width"] = max_diameter
        fuselage_dimension["fuselage_upper_height"] = round(max_diameter / 2, 2)
        fuselage_dimension["fuselage_lower_height"] = round(max_diameter / 2, 2)
        fuselage_dimension["fuselage_total_height"] = max_diameter
        runtime_output.warning('Attention! The force circle cross-section mode is selected. \n'
                               '                                    Fuselage height and width are set to the maximum '
                               'of both.')

    ''' store total cabin width to geometry description dictionary '''
    geometry_description['fuselage_dimension'] = fuselage_dimension

    ''' debug print '''
    runtime_output.debug('Debug: The fuselage_dimension_calculation function was successfully executed!')

    return geometry_description
