# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2024 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

# Import standard modules.
import csv
import os
import sys


def read_reference_ellipses(paths_and_names, runtime_output):
    """ Function reads the fuselage reference ellipsis for scaling.

    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict reference_ellipses: Dict containing fuselage reference ellipses
    """
    ''' initialize local parameter '''
    reference_ellipses = {'cockpit_section': {},
                          'constant_section': {},
                          'tail_section': {},
                          'scaling_factors': {},
                          'reference_diameter': {}
                          }

    ''' read fuselage reference ellipses from csv-files '''
    # set path to csv-files in subdirectory
    path_to_sub_folder = paths_and_names['working_directory'] + '/fuselage_design_lib/aircraft_ellipses'
    path_to_sub_folder = path_to_sub_folder.replace(os.sep, '/')

    # check for existing file of cockpit section, if not exist -> program will abort with exit code 1
    if os.path.isfile(path_to_sub_folder + '/reference_cockpit_section.csv'):
        # read ellipses for cockpit section
        with open(path_to_sub_folder + '/reference_cockpit_section.csv', 'r') as csv_file:
            ellipses = csv.reader(csv_file, delimiter=',')
            cockpit_section = {}
            for row in ellipses:
                cockpit_section[row[0]] = row[1:]
            reference_ellipses['cockpit_section'] = cockpit_section
            csv_file.close()
    else:
        runtime_output.critical('Error: The reference_cockpit_section.csv could not be opened. '
                                'Please update the workflow. \n                                     Program aborted!')
        sys.exit('Could not read fuselage reference ellipses.')

    # check for existing file of constant section, if not exist -> program will abort with exit code 1
    if os.path.isfile(path_to_sub_folder + '/reference_constant_section.csv'):
        # read ellipses for constant section
        with open(path_to_sub_folder + '/reference_constant_section.csv', 'r') as csv_file:
            ellipses = csv.reader(csv_file, delimiter=',')
            constant_section = {}
            for row in ellipses:
                constant_section[row[0]] = row[1:]
            reference_ellipses['constant_section'] = constant_section
            csv_file.close()
    else:
        runtime_output.critical('Error: The reference_constant_section.csv could not be opened. '
                                'Please update the workflow. \n                                     Program aborted!')
        sys.exit('Could not read fuselage reference ellipses.')

    # check for existing file of tail section, if not exist -> program will abort with exit code 1
    if os.path.isfile(path_to_sub_folder + '/reference_tail_section.csv'):
        # read ellipses for tail section
        with open(path_to_sub_folder + '/reference_tail_section.csv', 'r') as csv_file:
            ellipses = csv.reader(csv_file, delimiter=',')
            tail_section = {}
            for row in ellipses:
                tail_section[row[0]] = row[1:]
            reference_ellipses['tail_section'] = tail_section
            csv_file.close()
    else:
        runtime_output.critical('Error: The reference_tail_section.csv could not be opened. '
                                'Please update the workflow. \n                                     Program aborted!')
        sys.exit('Could not read fuselage reference ellipses.')

    # check for existing file of scaling factors, if not exist -> program will abort with exit code 1
    if os.path.isfile(path_to_sub_folder + '/reference_scaling_factors.csv'):
        # read scaling factors for reference aircraft
        with open(path_to_sub_folder + '/reference_scaling_factors.csv', 'r') as csv_file:
            ellipses = csv.reader(csv_file, delimiter=',')
            scaling_factors = {}
            for row in ellipses:
                scaling_factors[row[0]] = row[1:]
            reference_ellipses['scaling_factors'] = scaling_factors
            csv_file.close()
    else:
        runtime_output.critical('Error: The reference_scaling_factors.csv could not be opened. '
                                'Please update the workflow. \n                                     Program aborted!')
        sys.exit('Could not read fuselage reference ellipses.')

    # check for existing file of reference diameter, if not exist -> program will abort with exit code 1
    if os.path.isfile(path_to_sub_folder + '/reference_diameter.csv'):
        # read reference diameter for reference aircraft
        with open(path_to_sub_folder + '/reference_diameter.csv', 'r') as csv_file:
            ellipses = csv.reader(csv_file, delimiter=',')
            reference_diameter = {}
            for row in ellipses:
                reference_diameter[row[0]] = row[1:]
            reference_ellipses['reference_diameter'] = reference_diameter
            csv_file.close()
    else:
        runtime_output.critical('Error: The reference_diameter.csv could not be opened. '
                                'Please update the workflow. \n                                     Program aborted!')
        sys.exit('Could not read fuselage reference ellipses.')

    ''' debug print '''
    runtime_output.debug('Debug: The read_reference_ellipses function was successfully executed!')

    return reference_ellipses
