# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2024 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

# Import standard modules.
import os
import sys
import json
import importlib

# Import own modules.
from src.tube_and_wing.empirical.fuselage_design_tu_berlin.general.call_functions.estimatemasses import estimate_masses
from src.tube_and_wing.empirical.fuselage_design_tu_berlin.general.call_functions.prepareexternaldata import prepare_external_data  # noPep8 e501
from src.tube_and_wing.empirical.fuselage_design_tu_berlin.general.plot_functions.tailsectioncalculation import tail_section_calculation  # noPep8 e501
from src.tube_and_wing.empirical.fuselage_design_tu_berlin.general.call_functions.estimatepositionsandcog import estimate_positions_and_cog  # noPep8 e501
from src.tube_and_wing.empirical.fuselage_design_tu_berlin.general.plot_functions.cockpitsectioncalculation import cockpit_section_calculation  # noPep8 e501
from src.tube_and_wing.empirical.fuselage_design_tu_berlin.general.plot_functions.constantsectioncalculation import constant_section_calculation  # noPep8 e501
from src.tube_and_wing.empirical.fuselage_design_tu_berlin.general.call_functions.fuselage_geometry.readreferenceellipses import read_reference_ellipses  # noPep8 e501


def fuselage_design(paths_and_names, routing_dict, dict_ac_exchange, dict_mod_config, runtime_output):
    """  Module providing calculation functions for fuselage and cabin design in accordance CS 25.801ff.

    :param dict paths_and_names: Dictionary containing system paths and ElementTrees
    :param dict routing_dict: Dictionary containing routing parameter
    :param dict dict_ac_exchange: Dict containing parameter and according values from aircraft exchange file
    :param dict dict_mod_config: Dict containing parameter and according values from module configuration file
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict fuselage_description: Dict containing parameter and values of current fuselage design
    """
    ''' Initialize local parameter '''
    convert_inch_to_meter = 0.0254
    convert_meter_to_inch = 39.370078740161
    convergence_flag = False
    fuselage_description = {}
    word_to_insert_after = 'generate'
    files_to_remove = ['prepareexternaldata.py', 'estimatemasses.py', 'estimatepositionsandcog.py']

    ''' Read and prepare external data for fuselage and cabin design '''
    # Call function to prepare external data.
    dict_ac_exchange = prepare_external_data(paths_and_names, routing_dict, dict_ac_exchange, dict_mod_config,
                                             runtime_output)

    # Call function to read fuselage reference ellipses.
    dict_ac_exchange['reference_ellipses'] = read_reference_ellipses(paths_and_names, runtime_output)

    ''' Run fuselage design '''
    # If the script is compiled, sys._MEIPASS gives the temp directory for PyInstaller.
    if getattr(sys, 'frozen', False):    
        # List all entries in the directory.     
        all_entries = os.listdir(sys._MEIPASS + '/' +
                                 routing_dict['module_import_name'].replace('.', '/') + '/general/call_functions')
        # Filter out only files (ignore directories)
        call_function_list = [entry for entry in all_entries
                            if os.path.isfile(sys._MEIPASS + '/' + routing_dict['module_import_name'].replace('.', '/')
                                              + '/general/call_functions' + '/' + entry)]
    
    # Check if the system prefix is different from the base prefix -> code is runnig in a vitual environment
    elif hasattr(sys, 'base_prefix') and sys.base_prefix != sys.prefix:
        # List all entries in the directory.
        all_entries = os.listdir(paths_and_names['working_directory'] + '/' +
                                routing_dict['module_import_name'].replace('.', '/') + '/general/call_functions')
        # Filter out only files (ignore directories).
        call_function_list = [entry for entry in all_entries
                            if os.path.isfile(paths_and_names['working_directory'] + '/' +
                                                routing_dict['module_import_name'].replace('.', '/')
                                                + '/general/call_functions' + '/' + entry)] 
        
    # If running as a regular Python script, use the current directory.
    else:
        # List all entries in the directory.
        all_entries = os.listdir(os.getcwd() + '/' +
                                routing_dict['module_import_name'].replace('.', '/') + '/general/call_functions')
        # Filter out only files (ignore directories).
        call_function_list = [entry for entry in all_entries
                            if os.path.isfile(os.getcwd() + '/' +
                                                routing_dict['module_import_name'].replace('.', '/')
                                                + '/general/call_functions' + '/' + entry)]

    # Remove unnecessary files from call function list.
    for file in files_to_remove:
        if file in call_function_list:
            call_function_list.remove(file)

    call_function_list = sorted(call_function_list, key=lambda x: (int(x.split('_')[1]), x))

    i = 1
    while not convergence_flag and i < 100:
        runtime_output.print('Current iteration of fuselage design is: ' + str(i))
        # Loop across all python files from call_function directory to execute them one by one.
        for python_module in call_function_list:
            try:
                # Import own modules from call function directory.
                import_call_function = (routing_dict['module_import_name']
                                        + '.general.call_functions.' + os.path.splitext(python_module)[0])

                if 'generate' in import_call_function:
                    position_to_insert = python_module.find(word_to_insert_after) + len(word_to_insert_after)
                    python_module = word_to_insert_after + '_' + python_module[position_to_insert:]

                    function_name = os.path.splitext(python_module)[0]
                    # Import calculation module.
                    module = importlib.import_module(import_call_function)

                    # Call function depending on routing parameter.
                    fuselage_description = getattr(module, function_name)(dict_ac_exchange, dict_mod_config,
                                                                          fuselage_description, runtime_output)

            # Exception handling for module import error.
            except ModuleNotFoundError as module_import_error:
                runtime_output.critical('Error: ' + str(module_import_error) + ' found in '
                                        + routing_dict['module_name'] + '. Program aborted.')
                sys.exit('Exit information: The own python module to be imported could not be found!')
        
        ''' output prints of current loop '''
        runtime_output.print('Total length of fuselage is: ' 
                             + str(round(fuselage_description['geometry_description']['fuselage_dimension']['fuselage_length'] 
                                   * convert_inch_to_meter, 4)) + ' meter.')
        runtime_output.print('Total width of fuselage is: ' 
                             + str(round(fuselage_description['geometry_description']['fuselage_dimension']['fuselage_width'] 
                                   * convert_inch_to_meter, 4)) + ' meter.')
        runtime_output.print('Total height of fuselage is: ' 
                             + str(round(fuselage_description['geometry_description']['fuselage_dimension']['fuselage_total_height'] 
                                   * convert_inch_to_meter, 4)) + ' meter.')
        runtime_output.print('Total length of cabin is: ' 
                             + str(round(fuselage_description['geometry_description']['total_cabin_length'] 
                                   * convert_inch_to_meter, 4)) + ' meter.')
        runtime_output.print('Total width of cabin is: ' 
                             + str(round(fuselage_description['geometry_description']['total_cabin_width'] 
                                   * convert_inch_to_meter, 4)) + ' meter.')
        runtime_output.print('Wall thickness of cabin is: ' 
                             + str(round(fuselage_description['geometry_description']['cabin_wall_thickness'] 
                                   * convert_inch_to_meter, 4)) + ' meter.')
        runtime_output.print('Floor thickness of cabin is: ' 
                             + str(round(fuselage_description['geometry_description']['cabin_floor_thickness'] 
                                   * convert_inch_to_meter, 4)) + ' meter.')

        ''' convergence check '''
        # check if the fuselage geometry flag is true
        #  -> if true: -> check convergence -> otherwise restart landing gear estimation
        if fuselage_description['geometry_description']['geometry_flag']:
            convergence_flag = True
            runtime_output.print('Fuselage design successful after ' + str(i) + ' iterations!')

        i += 1

    # Call function to estimate fuselage and cabin masses.
    fuselage_description = estimate_masses(dict_ac_exchange, dict_mod_config, fuselage_description, runtime_output)

    # Call function to estimate positions and center of gravity.
    fuselage_description = estimate_positions_and_cog(dict_ac_exchange, dict_mod_config, fuselage_description,
                                                      runtime_output)

    ''' generate fuselage shape of current aircraft project '''
    dict_fuselage_deign_output = {
        'cockpit_section': cockpit_section_calculation(dict_ac_exchange, fuselage_description['geometry_description'],
                                                       runtime_output),
        'tail_section': tail_section_calculation(dict_ac_exchange, fuselage_description['geometry_description'],
                                                 runtime_output),
        'total_sections': {'origin': [], 'upper_height': [], 'lower_height': [], 'width': []}}

    dict_fuselage_deign_output['constant_section'] = \
        constant_section_calculation(dict_ac_exchange, fuselage_description['geometry_description'],
                                     dict_fuselage_deign_output, runtime_output)

    # Check if an additional fuselage length exist -> if true: -> add delta in x direction to x-cords of tail section
    if dict_ac_exchange['additional_fuselage_length'] is not None:
        for i in range(1, len(dict_fuselage_deign_output['tail_section']['x_direction_coordinates']) + 1):
            new_x_values_for_tail_section = \
                [round(x + (dict_ac_exchange['additional_fuselage_length'] * convert_meter_to_inch), 2)
                 for x in dict_fuselage_deign_output['tail_section']['x_direction_coordinates'][
                     'ellipse_number_' + str(i)]]
            dict_fuselage_deign_output['tail_section']['x_direction_coordinates']['ellipse_number_' + str(i)] \
                = new_x_values_for_tail_section

    ''' Set total section parameter '''
    # cockpit sections
    for key, value in dict_fuselage_deign_output['cockpit_section']['output_data']['origin'].items():
        dict_fuselage_deign_output['total_sections']['origin'].append(value)
    for key, value in dict_fuselage_deign_output['cockpit_section']['output_data']['upper_height'].items():
        dict_fuselage_deign_output['total_sections']['upper_height'].append(value)
    for key, value in dict_fuselage_deign_output['cockpit_section']['output_data']['lower_height'].items():
        dict_fuselage_deign_output['total_sections']['lower_height'].append(value)
    for key, value in dict_fuselage_deign_output['cockpit_section']['output_data']['width'].items():
        dict_fuselage_deign_output['total_sections']['width'].append(value)

    # constant sections
    for key, value in dict_fuselage_deign_output['constant_section']['output_data']['origin'].items():
        dict_fuselage_deign_output['total_sections']['origin'].append(value)
    for key, value in dict_fuselage_deign_output['constant_section']['output_data']['upper_height'].items():
        dict_fuselage_deign_output['total_sections']['upper_height'].append(value)
    for key, value in dict_fuselage_deign_output['constant_section']['output_data']['lower_height'].items():
        dict_fuselage_deign_output['total_sections']['lower_height'].append(value)
    for key, value in dict_fuselage_deign_output['constant_section']['output_data']['width'].items():
        dict_fuselage_deign_output['total_sections']['width'].append(value)

    # tail sections
    for key, value in dict_fuselage_deign_output['tail_section']['output_data']['origin'].items():
        dict_fuselage_deign_output['total_sections']['origin'].append(value)
    for key, value in dict_fuselage_deign_output['tail_section']['output_data']['upper_height'].items():
        dict_fuselage_deign_output['total_sections']['upper_height'].append(value)
    for key, value in dict_fuselage_deign_output['tail_section']['output_data']['lower_height'].items():
        dict_fuselage_deign_output['total_sections']['lower_height'].append(value)
    for key, value in dict_fuselage_deign_output['tail_section']['output_data']['width'].items():
        dict_fuselage_deign_output['total_sections']['width'].append(value)

    fuselage_description['dict_fuselage_deign_output'] = dict_fuselage_deign_output

    ''' write dictionary with fuselage design ellipses to json file (for use as existing geometry) '''
    # set path to csv-files in subdirectory
    path_to_sub_folder = paths_and_names['working_directory'] + '/fuselage_design_lib/aircraft_ellipses'
    path_to_sub_folder = path_to_sub_folder.replace(os.sep, '/')
    json.dump(fuselage_description['dict_fuselage_deign_output'],
              open(path_to_sub_folder + '/' + 'fuselage_design_ellipses.json', 'w'), indent=4)

    ''' debug print '''
    runtime_output.debug('Debug: The fuselage_design function was successfully executed!')

    return fuselage_description
