# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2024 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

"""Module providing report functionalities for current calculation method."""
from yattag import Doc
from bs4 import BeautifulSoup
import os


def method_html_report(paths_and_names, routing_dict, data_dict, method_specific_output_dict, runtime_output):
    """HTML report function.

    This function is responsible for creating reports.
    Note: Please ensure to format the data in the preparation part of the code (e.g., round floats)!

    :param dict paths_and_names: Dictionary containing system paths and ElementTrees
    :param dict routing_dict: Dictionary containing routing parameter
    :param dict data_dict: Dictionary containing results of module execution
    :return: None
    """
    runtime_output.print(f'HTML report is generated and saved...')

    """ Preparation and data extraction."""
    # Extract tool and project name.
    tool_name = paths_and_names['tool_name']
    report_title = (tool_name.replace('_', ' ')).capitalize()
    project_name = paths_and_names['name_of_project']
    root_of_module_config_tree = paths_and_names['root_of_module_config_tree']
    plot_switch = (eval(root_of_module_config_tree.find('.//plot_output/enable/value').text.capitalize()))
    conv_inch_to_m = 0.0254
    
    # Fuselage general data.
    general = [
        ("Number of fuselages", data_dict['geometry_description']['number_of_fuselages']),
        ("Number of payload tubes", data_dict['geometry_description']['number_of_payload_tubes']),
        ("Required number of flight crew members",
         data_dict['geometry_description']['required_number_of_flight_crew_member']),
        ("Required number of cabin crew members",
         data_dict['geometry_description']['required_number_of_flight_attendants'])
    ]
    general_title = "General"
    general_column_titles = ["Parameter", "Value"]
    general_column_alignment = ["l", "r"]

    # Fuselage geometry description.
    fuselage_geometry = []
    for key, value in data_dict['geometry_description']['fuselage_dimension'].items():
        formatted_key = key.replace('_', ' ').capitalize()
        # Format values.
        if isinstance(value, (int, float)):
            if 'volume' in key:
                formatted_value = f"{value:,.2f} L"
            elif 'area' in key:
                formatted_value = f"{value:,.2f} sqm"
            else:
                formatted_value = f"{value*conv_inch_to_m:,.2f} m"
        else:
            formatted_value = value
        fuselage_geometry.append((formatted_key, formatted_value))
    fuselage_geometry_title = "Fuselage geometry"
    fuselage_geometry_column_titles = ["Parameter", "Value"]
    fuselage_geometry_column_alignment = ["l", "r"]
    # Cabin geometry description.
    cabin_geometry = []
    exclude_list = ['number_of_fuselages', 'number_of_payload_tubes', 'fuselage_dimension', 'geometry_flag',
                    'required_number_of_flight_attendants', 'required_number_of_flight_crew_member']
    for key, value in data_dict['geometry_description'].items():
        if key not in exclude_list:
            formatted_key = key.replace('_', ' ').capitalize()
            # Format values.
            if isinstance(value, (int, float)):
                if 'ratio' in key:
                    formatted_value = f"{value:,.2f}"
                else:
                    formatted_value = f"{value*conv_inch_to_m:,.2f} m"
            else:
                formatted_value = value
            cabin_geometry.append((formatted_key, formatted_value))
    cabin_geometry_title = "Fuselage geometry"
    cabin_geometry_column_titles = ["Parameter", "Value"]
    cabin_geometry_column_alignment = ["l", "r"]

    # Fuselage overall masses.
    structural_mass = sum(data_dict['fuselage_design_masses']['fuselage_structural_masses'].values())
    furnishing_mass = data_dict['fuselage_design_masses']['fuselage_furnishing_masses']['total_furnishing_group_mass']
    operator_item_mass = data_dict['fuselage_design_masses']['fuselage_operator_item_masses']['total_group_mass']
    overall_mass = [
        ("Structural mass", f"{structural_mass:,.2f} kg"),
        ("Furnishing mass", f"{furnishing_mass:,.2f} kg"),
        ("Operator item mass", f"{operator_item_mass:,.2f} kg"),
        ("Overall fuselage mass", f"{data_dict['fuselage_design_masses']['total_fuselage_mass']:,.2f} kg")
    ]
    overall_mass_title = "Fuselage group masses"
    overall_mass_column_titles = ["Group", "Mass"]
    overall_mass_column_alignment = ["l", "r"]
    # Fuselage structural masses breakdown.
    structural_mass_breakdown = [
        (key.replace("_", " ").capitalize(), f"{value:,.2f} kg")
        for key, value in data_dict['fuselage_design_masses']['fuselage_structural_masses'].items()
    ]
    structural_mass_breakdown.append(("Fuselage structural mass total", f"{structural_mass:,.2f} kg"))
    structural_mass_breakdown_title = "Fuselage structural masses"
    structural_mass_breakdown_column_titles = ["Component", "Mass"]
    structural_mass_breakdown_alignment = ["l", "r"]
    # Fuselage furnishing masses breakdown.
    furnishing_mass_breakdown = [
        (key.replace("_", " ").capitalize(), f"{value:,.2f} kg")
        for key, value in data_dict['fuselage_design_masses']['fuselage_furnishing_masses'].items()
        if key != 'total_furnishing_group_mass'
    ]
    furnishing_mass_breakdown.append(("Fuselage furnishing mass total", f"{furnishing_mass:,.2f} kg"))
    furnishing_mass_breakdown_title = "Fuselage furnishing masses"
    furnishing_mass_breakdown_column_titles = ["Component", "Mass"]
    furnishing_mass_breakdown_alignment = ["l", "r"]
    # Fuselage operator item group masses.
    operator_item_mass_breakdown = [
        (key.replace("_", " ").capitalize(), f"{value:,.2f} kg")
        for key, value in data_dict['fuselage_design_masses']['fuselage_operator_item_masses']['group_masses'].items()
        if key != 'total_furnishing_group_mass'
    ]
    operator_item_mass_breakdown.append(("Fuselage operator item group total", f"{operator_item_mass:,.2f} kg"))
    operator_item_mass_breakdown_title = "Fuselage operator item masses"
    operator_item_mass_breakdown_column_titles = ["Component", "Mass"]
    operator_item_mass_breakdown_alignment = ["l", "r"]
    # Fuselage positions.
    positions = []
    for key, value in data_dict['positions'].items():
        if isinstance(value, dict):
            if any(isinstance(sub_value, dict) for sub_value in value.values()):
                for sub_key, sub_value in value.items():
                    positions.append((sub_key.replace("_", " ").capitalize(), 
                                *[f"{v:,.2f} m" for v in sub_value.values()]))
            else:
                positions.append((key.replace("_", " ").capitalize(), 
                            *[f"{v:,.2f} m" for v in value.values()]))
        else:
            positions.append((key.replace("_", " ").capitalize(), 
                        f"{value:,.2f} m"))
    positions_title = "Positions"
    positions_column_titles = ["Component", "x", "y", "z"]
    positions_alignments = ["l", "m", "m", "m"]

    # Fuselage interior.
    interior = []
    class_keys = list(data_dict['interior_description'].keys())
    seat_abreast = ["Seat abreast"]
    width_of_seat_rows_per_class = ["Width of seat rows per class in inch"]
    cabin_length_per_class = ["Cabin length per class in m"]
    for classes in class_keys:
        seat_abreast.append('-'.join(data_dict['interior_description'][classes]['seat_abreast']))
        width_of_seat_rows_per_class.append(
            '-'.join(
                map(str, data_dict['interior_description'][classes]['width_of_seat_rows_per_class'])))
        cabin_length_per_class.append(
            f"{data_dict['interior_description'][classes]['cabin_length_per_class'][0]*conv_inch_to_m:,.2f}")
    exclude_list = ['seat_abreast', 'cabin_length_per_class', 'width_of_seat_rows_per_class']

    for key in data_dict['interior_description'][class_keys[0]].keys():
        if key not in exclude_list:
            # values = [data_dict['interior_description'][class_key][key] for class_key in class_keys]
            values = [
                ('-'.join(map(str, value)) if isinstance(value, list) and len(value) > 1 else f"{value[0]:,.2f}" if isinstance(value, list) else value)
                for class_key in class_keys
                for value in [data_dict['interior_description'][class_key][key]]
            ]
            interior.append((key.replace("_", " ").capitalize(), *values))
    
    interior.append(tuple(seat_abreast))
    interior.append(tuple(cabin_length_per_class))
    interior.append(tuple(width_of_seat_rows_per_class))
    interior_title = "Interior data"
    interior_column_titles = ["Parameter"]
    interior_column_titles.extend(class_keys)
    additional_alignments = len(class_keys)
    interior_alignments = ["l"]
    interior_alignments.extend(["r"] * additional_alignments)
    
    """Generate report."""
    # Start Yattag document.
    doc, tag, text = Doc().tagtext()

    # Generate header (without creating a new <html>).
    with tag('html', lang="en"):
        # Generate report header (adds <head>).
        generate_report_header(doc, report_title)

        # Generate body.
        with tag('body'):
            with tag('div', klass="logo"):
                text("")
            with tag('div', klass="content"):
                with tag('h1'):
                    text(f"Report - {report_title} of {project_name}")
                with tag('div', klass="container"):

                    # Data on the left side.
                    with tag('div', klass="box data"):
                        
                        # General fuselage data.
                        with tag('h2'):
                            text("General")
                        # Add table for general data.
                        generate_table(doc, tag, text, general, general_title,
                                       general_column_titles, general_column_alignment)

                        # General fuselage data.
                        with tag('h2'):
                            text("Geometry")
                        # Add table for general data.
                        generate_table(doc, tag, text, fuselage_geometry, fuselage_geometry_title,
                                       fuselage_geometry_column_titles, fuselage_geometry_column_alignment)
                        # Add table for general data.
                        generate_table(doc, tag, text, cabin_geometry, cabin_geometry_title,
                                       cabin_geometry_column_titles, cabin_geometry_column_alignment)

                        # Mass data.
                        with tag('h2'):
                            text("Masses")
                        # Add table for fuselage group masses.
                        generate_table(doc, tag, text, overall_mass, overall_mass_title,
                                       overall_mass_column_titles, overall_mass_column_alignment, last_line_bold=True)
                        # Add table for structural mass breakdown.
                        generate_table(doc, tag, text, structural_mass_breakdown, structural_mass_breakdown_title,
                                       structural_mass_breakdown_column_titles, structural_mass_breakdown_alignment,
                                       last_line_bold=True)
                        # Add table for furnishing mass breakdown.
                        generate_table(doc, tag, text, furnishing_mass_breakdown, furnishing_mass_breakdown_title,
                                       furnishing_mass_breakdown_column_titles, furnishing_mass_breakdown_alignment,
                                       last_line_bold=True)
                        # Add table for operator item mass breakdown.
                        generate_table(doc, tag, text, operator_item_mass_breakdown,
                                       operator_item_mass_breakdown_title, operator_item_mass_breakdown_column_titles,
                                       operator_item_mass_breakdown_alignment, last_line_bold=True)

                        # Position data.
                        with tag('h2'):
                            text("Positions")
                        # Add table for fuselage positions.
                        generate_table(doc, tag, text, positions, positions_title, positions_column_titles,
                                       positions_alignments)

                        # Interior data.
                        with tag('h2'):
                            text("Interior")
                        # Add table for fuselage positions.
                        generate_table(doc, tag, text, interior, interior_title, interior_column_titles,
                                       interior_alignments)
                    # Plots on the right side.
                    # Ensure that plots are enabled before displaying them on report.
                    if plot_switch:
                        with tag('div', klass="box plot"):
                            with tag('h2'):
                                text("Plots")
                                # 3D shape.
                                with tag('h3'):
                                    text('Fuselage 3D shape')
                                    src_name = "../plots/fuselage_design_3D_shape.svg"
                                with tag('img',src=src_name, klass="image-plot", style="margin-bottom: 20px;"):
                                    pass
                                # 3D shape.
                                with tag('h3'):
                                    text('Fuselage isometric view')
                                    src_name = "../plots/fuselage_design_Isometric view of fuselage shape.svg"
                                with tag('img',src=src_name, klass="image-plot", style="margin-bottom: 20px;"):
                                    pass

    # Convert the document to HTML string.
    html_output = doc.getvalue()

    # Pretty HTML output with BeautifulSoup.
    soup = BeautifulSoup(html_output, 'html.parser')
    pretty_html = soup.prettify()

    """ Save report to current aircraft project directory. """
    # Save the report.
    project_directory = paths_and_names["project_directory"]
    report_directory = os.path.join(project_directory, 'reporting', 'report_html')
    os.makedirs(report_directory, exist_ok=True)
    output_path = os.path.join(report_directory, f"{tool_name}_report.html")

    # Write the HTML to the file.
    with open(output_path, "w", encoding="utf-8") as report:
        report.write(pretty_html)
    
    return


def generate_table(doc, tag, text, data, caption, column_titles, alignment=None, last_line_bold=False,
                   table_without_head=False):
    """Generate an HTML table.

    This function creates an HTML table from the provided data.
    Note: Options for the alignment types are "l" for left, "r" for right, and "m" for center. If not provided, columns
    are left-aligned by default.

    :param doc: Yattag Doc instance
    :param tag: Yattag tag function
    :param text: Yattag text function
    :param str caption: Table caption
    :param list column_titles: Titles of table columns
    :param list alignment: List of alignment types for ech column (defaults to None)
    :param bool last_line_bold: Flag if last table line should be printed bold
    :param bool table_without_head: Flag if table has no table head
    :param list data: List of tuples containing table data
    """
    with tag('table', klass="content-table"):
        with tag('caption'):
            text(caption)
        if table_without_head:
            pass
        else:
            with tag('thead'):
                with tag('tr'):
                    for col_num, col in enumerate(column_titles):
                        if alignment:
                            if alignment[col_num] == "m":
                                with tag('th', style="text-align: center;"):
                                    text(f"{col}")
                            elif alignment[col_num] == "r":
                                with tag('th', style="text-align: right;"):
                                    text(f"{col}")
                            elif alignment[col_num] == "l":
                                with tag('th', style="text-align: left;"):
                                    text(f"{col}")
                        else:                       
                            with tag('th'):
                                text(f"{col}")
        with tag('tbody'):
            row_number = 0
            for row in data:
                if last_line_bold:
                    is_last_row = (row_number == len(data) - 1)
                else:
                    is_last_row = False
                with tag('tr', style="font-weight: normal;" if is_last_row else ""):
                    column_number = 0
                    for col in row:
                        if col is None:
                            with tag('td'):
                                text('N.A.')
                        else:
                            if alignment is not None:
                                if alignment[column_number] == "m":
                                    with tag('td', style="text-align: center;"):
                                        text(col)
                                elif alignment[column_number] == "r":
                                    with tag('td', style="text-align: right;"):
                                        text(col)
                                elif alignment[column_number] == "l":
                                    with tag('td', style="text-align: left;"):
                                        text(col)
                                else:
                                    with tag('td'):
                                        text(col)
                            else:
                                with tag('td'):
                                    text(col)
                        column_number += 1
                row_number += 1
    
    return


def generate_report_header(doc, tool_name):
    """Generate HTML header.
    
    This function generates an HTML header with meta information, style sheets, and title.

    :param doc: Yattag Doc instance
    :param str tool_name: Name of the tool
    """
    with doc.tag('head'):
        doc.stag('meta', name="viewport", content="width=device-width, initial-scale=1.0")
        doc.stag('meta', charset="UTF-8")
        with doc.tag('title'):
            doc.text(tool_name)
        doc.stag('link', rel="stylesheet", href="style.css")
    
    return
