/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef INITIALSIZING_SRC_TAW_ANALYTICAL_TAWANALYTICAL_H_  
#define INITIALSIZING_SRC_TAW_ANALYTICAL_TAWANALYTICAL_H_  

#include <memory>
#include <vector>
#include <moduleBasics/strategySelector.h>
#include <moduleBasics/report.h>
#include <moduleBasics/plot.h>
#include "../../initialSizing.h"
#include "../TawInitialSizingIOData.h"
#include "../TawInitialSizingData.h"


class TawAnalytical : public Strategy {
 public:
    explicit TawAnalytical(std::shared_ptr<RuntimeIO> rtIO);
    ~TawAnalytical() = default;
    void initialize();
    void run();
    void update();
    void report();
    void save();
    void size();
    void study();
    void calculatePayloadAndCrew();
    void calculateFuelFraction();
    void calculateMassEstimation();
    bool checkConvergence(double mo_iteration, double mo_iteration_previous);
    double final_mass_estimation_;  // Member variable to store the final mass estimation
    double mo_iteration_;
    void calculateCDO();
    void calculateTakeoff();
    void calculateClimb();
    void calculateCruise();
    void calculateLanding();
    void findIntersectionPoints_LTO();
    void findIntersectionPoints_CL();
    void findIntersectionPoints_CR();
    void findIntersectionPoints_CT();
    void findIntersectionPoints_TC();
    void getoptiCL();
    void findIntersectionPoints_optiCL_Cruise();
    void findIntersectionPoints_optiCL_Climb();
    void findIntersectionPoints_optiCL_Takeoff();
    void findDesignPoint();

    std::vector<double> takeoffDataWS;
    std::vector<double> climbDataWS;
    std::vector<double> cruiseDataWS;
    std::vector<double> landingDataWS;
    std::vector<double> takeoffDataTW;
    std::vector<double> climbDataTW;
    std::vector<double> cruiseDataTW;
    std::vector<double> landingDataTW;
    std::vector<double> optiCLDataTW;
    std::vector<double> optiCLDataWS;

    std::vector<std::pair<double, double>> intersectionPoints_CL;
    std::vector<std::pair<double, double>> intersectionPoints_CR;
    std::vector<std::pair<double, double>> intersectionPoints_LTO;
    std::vector<std::pair<double, double>> intersectionPoints_CT;
    std::vector<std::pair<double, double>> intersectionPoints_TC;
    std::vector<std::pair<double, double>> intersectionPoints_optiCL_Climb;
    std::vector<std::pair<double, double>> intersectionPoints_optiCL_Cruise;
    std::vector<std::pair<double, double>> intersectionPoints_optiCL_Takeoff;


    double designWS_;
    double designTW_;
    

    /* Reporting elements */
    /**
    * \brief Defines the content of the HTML Report body. The content depends on wether the sizing or study mode
    * is used. Since that masses or only calculated in the sizing mode, they are also only written in this mode.
    * \return void
    */
    void setHtmlBody();
    void setTexBody();
    void generatePlotData();

 private:  // NOLINT
    std::shared_ptr<RuntimeIO> rtIO_;
    std::shared_ptr<TawInitialSizingIOData> data_;
    Report report_;

    // Member variables to store calculated values
    // double timeLoiter = data_->data.MissionInfo.timeLoiter;
    // double LDInitialLoiter = data_->configuration.LDInitialLoiter;
    double ff_warmup = 0.;
    double ff_taxi = 0.;
    double ff_takeoff = 0.;
    double ff_climb = 0.;
    double ff_descent = 0.;
    double ff_land = 0.;
    double ff_missedandclimb = 0.;
    double altitude = 0.;
    double c_kerosine = 0.;
    double c_LH2 = 0.;
    double speedOfSound = 0.;
    double passengerWeight = 0.;
    double cargoWeight = 0.;
    int numCrew = 0.;
    int numEconomyCrew = 0.;
    int numPremiumEconomyCrew = 0.;
    int numBusinessClassCrew = 0.;
    int numFirstClassCrew = 0.;
    int numPresidentCrew = 0.;
    double mo_initial = 0.;
    double mo_iteration = 0.;
    double mo_iteration_previous = 0.;
    double FuelFraction = 0.;
    double crewWeight = 0.;
    double Payload = 0.;
    double final_mass_estimation = 0.;
    double CD0_clean = 0.;
    double STOFL = 0.;
    double CLmaxTO = 0.;
    double mo_S = 0.;
    double numberOfEngines = 0.;
    double vCruise = 0.;
    double AR = 0.;
    double e = 0.;
    double me_mo = 0.;
    double machNumber = 0.;
    double densityCruise = 0.;
    std::string energyCarrier = "";
    double SFCKerosene  = 0.;
    double LDInitialCruise = 0.;
    double LDInitialLoiter = 0.;
    double T_W_TO = 0.;
    double T_W_C = 0.;
    double T_W_R = 0.;
    double moS_landing = 0.;
    double designWS = 0.;
    double designTW = 0.;
    double mcr_mo = 0.;
    double range = 0.;
    double contingency = 0.;
    double CruiseFuel  = 0.;
    double distAlter = 0.;
	double AltFuel = 0.;
    double timeLoiter = 0.;
	double LoiterFuel = 0.;
    double c_fuel = 0.;
    
};

#endif // INITIALSIZING_SRC_TAW_ANALYTICAL_TAWANALYTICAL_H_
