# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2024 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

# Import standard modules.
import sys

# Import own modules.
from src.wing_mounted.empirical.landing_gear_design_tu_berlin.general.call_functions._01_estimatedistances import estimate_distances  # noPep8 e501


def estimate_loads(dict_ac_exchange, dict_mod_config, landing_gear_description, runtime_output):
    """ Landing gear load estimation for main nose gear and main gear struts
        in accordance to Norman S. Currey. [ref N. S. Currey, 1988]

    :param dict dict_ac_exchange: Dict containing parameter and according values from aircraft exchange file
    :param dict dict_mod_config: Dict containing parameter and according values from module configuration file
    :param dict landing_gear_description: Dict containing parameter and values of current landing gear estimation
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict landing_gear_description: Dict containing parameter and values of current landing gear estimation
    """
    # initialize local parameter
    load_flag = False
    landing_gear_loads = {}
    landing_gear_limitations = {}
    minimum_nose_gear_position_reached = False
    maximum_nose_gear_position_reached = False
    nose_gear_minimum_static_load = float()
    nose_gear_maximum_static_load = float()
    nose_gear_maximum_dynamic_load = float()
    percentage_of_nose_gear_minimum_load = float()
    percentage_of_nose_gear_maximum_load = float()
    percentage_of_nose_gear_dynamic_load = float()
    # ToDo: gravitational_acceleration should be in a separate python standard package
    gravitational_acceleration = 9.80665

    ''' Initialize landing gear limitations '''
    # set maximum aft position of nose landing gear
    if not ('landing_gear_limitations' in landing_gear_description) \
            and not ('maximum_aft_position_of_nose_gear' in landing_gear_limitations):
        landing_gear_limitations['maximum_aft_position_of_nose_gear'] = 6.5
    else:
        landing_gear_limitations['maximum_aft_position_of_nose_gear'] = \
            landing_gear_description['landing_gear_limitations']['maximum_aft_position_of_nose_gear']

    # set minimum foremost position of nose landing gear
    if not ('landing_gear_limitations' in landing_gear_description) \
            and not ('minimum_forward_position_of_nose_gear' in landing_gear_limitations):
        landing_gear_limitations['minimum_forward_position_of_nose_gear'] = 4.5
    else:
        landing_gear_limitations['minimum_forward_position_of_nose_gear'] = \
            landing_gear_description['landing_gear_limitations']['minimum_forward_position_of_nose_gear']

    # set minimum foremost position of main landing gear
    if not ('landing_gear_limitations' in landing_gear_description) \
            and not ('minimum_forward_position_of_main_gear' in landing_gear_limitations):
        landing_gear_limitations['minimum_forward_position_of_main_gear'] \
            = (landing_gear_description['lever_arms']['distance_between_nose_gear_and_rearmost_center_of_gravity']
               + landing_gear_description['landing_gear_positions']['nose_gear_x_position']
               + landing_gear_description['lever_arms']['distance_between_main_gear_and_rearmost_center_of_gravity'])
    else:
        landing_gear_limitations['minimum_forward_position_of_main_gear'] = \
            landing_gear_description['landing_gear_limitations']['minimum_forward_position_of_main_gear']

    ''' Maximum ramp weight estimation. '''
    # check if the key 'maximum_ramp_weight' not exists in the landing gear description dictionary
    #  -> if true: -> Estimate maximum ramp weight starting value
    if not ('landing_gear_loads' in landing_gear_description) and not ('maximum_ramp_weight' in landing_gear_loads):
        # Check if the maximum take of mass and the taxi fuel mass for take-off are zero.
        #  -> if true: -> Check which kind of fuselage type is selected and set specific starting values.
        if dict_ac_exchange['maximum_take_off_mass'] is None and dict_ac_exchange['taxi_fuel_take_off'] is None:
            # Check if the fuselage type of current project is single aisle
            #  -> if true: -> Set starting value for maximum ramp weight for single aisle aircraft to 70.000 * g0.
            if dict_ac_exchange['fuselage_type'] == 'single_aisle':
                landing_gear_loads['maximum_ramp_weight'] = gravitational_acceleration * 70000
            # Else condition: The fuselage type of current project is weight body.
            #  -> Set starting value for maximum ramp weight for weight body aircraft to 250.000 * g0.
            else:
                landing_gear_loads['maximum_ramp_weight'] = gravitational_acceleration * 250000
        # Else if condition: Check if the taxi fuel take of mass is zero
        #  -> if true: -> Add 1000 kg to the maximum take of mass.
        elif dict_ac_exchange['taxi_fuel_take_off'] is None:
            landing_gear_loads['maximum_ramp_weight'] = gravitational_acceleration \
                                                        * (dict_ac_exchange['maximum_take_off_mass'] + 1000)
        # Else condition: Maximum take-off mass and taxi fuel mass for take-off exists -> estimate maximum ramp weight
        else:
            landing_gear_loads['maximum_ramp_weight'] = gravitational_acceleration \
                                                        * (dict_ac_exchange['maximum_take_off_mass']
                                                           + dict_ac_exchange['taxi_fuel_take_off'])
    # Else condition: The key 'maximum_ramp_weight' exists in the landing gear description dictionary
    #  -> Use the value for further estimation.
    else:
        landing_gear_loads['maximum_ramp_weight'] = \
            landing_gear_description['landing_gear_loads']['maximum_ramp_weight']

    ''' Nose gear load estimation '''
    # check if the main landing gear is not on the minimum most forward position
    #  -> if true: -> set main gear position to minimum most forward position.
    if not (landing_gear_description['landing_gear_positions']['main_gear_x_position']
            == landing_gear_limitations['minimum_forward_position_of_main_gear']):
        landing_gear_description['landing_gear_positions']['main_gear_x_position'] \
            = landing_gear_limitations['minimum_forward_position_of_main_gear']

    i = 0
    while i < 1000 and not load_flag:
        # minimum static nose gear load
        nose_gear_minimum_static_load = \
            ((landing_gear_loads['maximum_ramp_weight']
              * (landing_gear_description['lever_arms']['distance_between_nose_and_main_gear']
                 - landing_gear_description['lever_arms']
                 ['distance_between_nose_gear_and_rearmost_center_of_gravity']))
             / landing_gear_description['lever_arms']['distance_between_nose_and_main_gear'])

        # maximum static nose gear load
        nose_gear_maximum_static_load = \
            ((landing_gear_loads['maximum_ramp_weight']
              * (landing_gear_description['lever_arms']['distance_between_nose_and_main_gear']
                 - landing_gear_description['lever_arms']
                 ['distance_between_nose_gear_and_foremost_center_of_gravity']))
             / landing_gear_description['lever_arms']['distance_between_nose_and_main_gear'])

        # maximum static nose gear load
        nose_gear_maximum_dynamic_load = \
            (nose_gear_maximum_static_load
             + (10
                * landing_gear_description['lever_arms'][
                    'vertical_distance_between_ground_and_aft_center_of_gravity']
                * landing_gear_loads['maximum_ramp_weight'])
             / (32.2 * landing_gear_description['lever_arms']['distance_between_nose_and_main_gear']))

        percentage_of_nose_gear_minimum_load = (nose_gear_minimum_static_load
                                                / landing_gear_loads['maximum_ramp_weight']) * 100
        percentage_of_nose_gear_maximum_load = (nose_gear_maximum_static_load
                                                / landing_gear_loads['maximum_ramp_weight']) * 100
        percentage_of_nose_gear_dynamic_load = (nose_gear_maximum_dynamic_load
                                                / landing_gear_loads['maximum_ramp_weight']) * 100

        ''' Check if the minimum and maximum allowed nose gear load is in accordance to Norman S. Currey '''
        # Check if the minimum and maximum permissible percentage of the nose gear load is violated at the same time
        #  -> if true: -> Raise an error and abort landing gear design
        if percentage_of_nose_gear_maximum_load > 20 and percentage_of_nose_gear_minimum_load < 6:
            runtime_output.critical('Error: The minimum and maximum permissible percentage of the nose landing gear'
                                    ' load is violated at the same time. \n '
                                    '                                    Load estimation for landing gear '
                                    'design not possible. Program aborted!')
            sys.exit('Landing gear load estimation failed for landing gear design!')

        # check if the maximum allowed percentage of the nose gear load is greater than 20
        #  -> if true: -> Try move nose gear forward.
        elif percentage_of_nose_gear_maximum_load > 20:
            # Check if the foremost nose gear position (8 percent nose gear load) is not reached
            #  -> if true: -> Move the x-position of the nose landing gear forward by 0.05 meter.
            if percentage_of_nose_gear_minimum_load > 6 \
                    and landing_gear_description['landing_gear_positions']['nose_gear_x_position'] \
                    > landing_gear_limitations['minimum_forward_position_of_nose_gear']:
                landing_gear_description['landing_gear_positions']['nose_gear_x_position'] -= 0.05
            # Else condition: Most forward x-position of nose gear is reached. -> No further movement possible.
            else:
                minimum_nose_gear_position_reached = True
            # Check if both x-position have reached the limit. -> if true: -> Raise an error and abort program.
            if minimum_nose_gear_position_reached:
                
                runtime_output.critical('Error: The minimum nose gear position is reached but the maximum'
                                        ' allowed percentage of the nose gear load is greater than 20. \n '
                                        '                                    Load estimation for landing gear '
                                        'design not possible. Program aborted!')
                sys.exit('Landing gear load estimation failed for landing gear design!')

            landing_gear_description['landing_gear_positions'] = {
                key: round(value, 4) for key, value in landing_gear_description['landing_gear_positions'].items()}
            landing_gear_description = estimate_distances(dict_ac_exchange, dict_mod_config,
                                                          landing_gear_description, runtime_output)

        # Else if condition: Check if the minimum allowed percentage of the nose gear load is lower than 6
        elif percentage_of_nose_gear_minimum_load < 6:
            # Check if the aft most nose gear position (20 percent nose gear load) is not reached
            #  -> if true: -> Move the x-position of the nose landing gear backward by 0.05 meter.
            if percentage_of_nose_gear_minimum_load < 20 \
                    and landing_gear_description['landing_gear_positions']['nose_gear_x_position'] \
                    < landing_gear_limitations['maximum_aft_position_of_nose_gear']:
                landing_gear_description['landing_gear_positions']['nose_gear_x_position'] += 0.05
            # Else condition: Most backward x-position of nose gear is reached. -> No further movement possible.
            else:
                maximum_nose_gear_position_reached = True

            # Check if the maximum aft position of nose landing gear is reached
            #  -> if true: -> Raise an error and abort program.
            if maximum_nose_gear_position_reached:
                runtime_output.critical(
                    'Error: The maximum nose gear position is reached but the minimum allowed percentage of '
                    'the nose gear load is greater lower than 6 \n '
                    '                                    Load estimation for landing gear '
                    'design not possible. Program aborted!')
                sys.exit('Landing gear load estimation failed for landing gear design!')

            landing_gear_description['landing_gear_positions'] = {
                key: round(value, 4) for key, value in landing_gear_description['landing_gear_positions'].items()}
            landing_gear_description = estimate_distances(dict_ac_exchange, dict_mod_config,
                                                          landing_gear_description, runtime_output)

        # Else condition: Min and max load of the nose gear is not between 6 and 20 percent of maximum ramp weight.
        #  -> Break the loop and continue landing gear estimation.
        else:
            load_flag = True

        i += 1

    if not load_flag:
        runtime_output.critical(
            'Error: The maximum number of iteration (i=1000) for nose gear load estimation is reached. \n '
            '                                    Load estimation for landing gear design not possible. '
            'Program aborted!')
        sys.exit('Landing gear load estimation failed for landing gear design!')

    ''' Estimate dynamic nose gear loads for take-off and landing condition according to CS 25.733 (b)(2) and (b)(3) '''
    # maximum static nose gear landing load
    # check if the given maximum landing mass is equal to zero
    #  -> if true: -> use 90 percent of maximum ramp weight for initial estimation
    if dict_ac_exchange['maximum_landing_mass'] is None:
        nose_gear_maximum_static_landing_load = (0.9 * landing_gear_loads['maximum_ramp_weight']
                                                 * percentage_of_nose_gear_maximum_load/100)
    # Else condition: The given maximum landing mass is not equal to zero -> use for dynamic load estimation
    else:
        nose_gear_maximum_static_landing_load = \
            (((dict_ac_exchange['maximum_landing_mass'] * gravitational_acceleration)
              * (landing_gear_description['lever_arms']['distance_between_nose_and_main_gear']
                 - landing_gear_description['lever_arms']
                 ['distance_between_nose_gear_and_foremost_center_of_gravity']))
             / landing_gear_description['lever_arms']['distance_between_nose_and_main_gear'])

    # nose gear dynamic landing load (CS 25.733 (b)(2))
    nose_gear_maximum_dynamic_landing_load = (nose_gear_maximum_static_landing_load
                                              + 0.31 * nose_gear_maximum_static_landing_load)

    # nose gear dynamic take-off load (CS 25.733 (b)(3))
    nose_gear_maximum_dynamic_take_off_load = nose_gear_maximum_static_load + 0.2 * nose_gear_maximum_static_load

    landing_gear_loads['nose_gear_minimum_static_load'] = nose_gear_minimum_static_load
    landing_gear_loads['nose_gear_maximum_static_load'] = nose_gear_maximum_static_load
    landing_gear_loads['nose_gear_maximum_dynamic_load'] = nose_gear_maximum_dynamic_load
    landing_gear_loads['nose_gear_maximum_dynamic_take_off_load'] = nose_gear_maximum_dynamic_take_off_load
    landing_gear_loads['nose_gear_maximum_static_landing_load'] = nose_gear_maximum_static_landing_load
    landing_gear_loads['nose_gear_maximum_dynamic_landing_load'] = nose_gear_maximum_dynamic_landing_load
    landing_gear_loads['percentage_of_nose_gear_minimum_load'] = percentage_of_nose_gear_minimum_load
    landing_gear_loads['percentage_of_nose_gear_maximum_load'] = percentage_of_nose_gear_maximum_load
    landing_gear_loads['percentage_of_nose_gear_dynamic_load'] = percentage_of_nose_gear_dynamic_load

    ''' Main gear load estimation '''
    # Estimation of total main gear load (100 percent - minimum_nose_gear_load_in_percent)/100 * maximum_ramp_weight
    landing_gear_loads['main_gear_maximum_total_load'] = ((100 -
                                                          landing_gear_loads['percentage_of_nose_gear_minimum_load'])
                                                          / 100 * landing_gear_loads['maximum_ramp_weight'])

    ''' Most possible formost and aft position of nose landing gear depending on maximum load '''
    # maximum of 20 percent of maximum ramp weight allowed as maximum static nose gear load
    nose_gear_maximum_possible_static_load = 0.2 * landing_gear_loads['maximum_ramp_weight']

    minimum_required_distance_between_nose_gear_and_formmost_center_of_gravity = \
        round((landing_gear_loads['maximum_ramp_weight']
               * landing_gear_description['lever_arms']['distance_between_nose_and_main_gear']
               - nose_gear_maximum_possible_static_load
               * landing_gear_description['lever_arms']['distance_between_nose_and_main_gear'])
              / landing_gear_loads['maximum_ramp_weight'], 4)

    if dict_ac_exchange['most_forward_center_of_gravity_in_x_direction'] is not None:
        landing_gear_limitations['maximum_aft_position_of_nose_gear'] = \
            (dict_ac_exchange['most_forward_center_of_gravity_in_x_direction']
             - minimum_required_distance_between_nose_gear_and_formmost_center_of_gravity)
    else:
        landing_gear_limitations['maximum_aft_position_of_nose_gear'] = \
            (landing_gear_description['landing_gear_positions']['temporary_center_of_gravity']
             - minimum_required_distance_between_nose_gear_and_formmost_center_of_gravity)

    # minimum of 6 percent of maximum ramp weight allowed as minimum static nose gear load
    nose_gear_minimum_possible_static_load = 0.06 * landing_gear_loads['maximum_ramp_weight']

    maximum_required_distance_between_nose_gear_and_aft_most_center_of_gravity = \
        round((landing_gear_loads['maximum_ramp_weight']
               * landing_gear_description['lever_arms']['distance_between_nose_and_main_gear']
               - nose_gear_minimum_possible_static_load
               * landing_gear_description['lever_arms']['distance_between_nose_and_main_gear'])
              / landing_gear_loads['maximum_ramp_weight'], 4)

    if dict_ac_exchange['most_aft_center_of_gravity_in_x_direction'] is not None:
        landing_gear_limitations['minimum_forward_position_of_nose_gear'] = \
            (dict_ac_exchange['most_aft_center_of_gravity_in_x_direction']
             - maximum_required_distance_between_nose_gear_and_aft_most_center_of_gravity)
    else:
        landing_gear_limitations['minimum_forward_position_of_nose_gear'] = \
            (landing_gear_description['landing_gear_positions']['temporary_center_of_gravity']
             - maximum_required_distance_between_nose_gear_and_aft_most_center_of_gravity)

    # prepare output dictionary
    landing_gear_loads = {key: round(value, 4) for key, value in landing_gear_loads.items()}
    landing_gear_description['landing_gear_loads'] = landing_gear_loads
    landing_gear_limitations = {key: round(value, 4) for key, value in landing_gear_limitations.items()}
    landing_gear_description['landing_gear_limitations'] = landing_gear_limitations

    # update landing gear distances
    landing_gear_description['landing_gear_positions'] = {
        key: round(value, 4) for key, value in landing_gear_description['landing_gear_positions'].items()}
    landing_gear_description = estimate_distances(dict_ac_exchange, dict_mod_config,
                                                  landing_gear_description, runtime_output)

    runtime_output.debug('estimate_loads successfully executed!')

    return landing_gear_description
