# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2024 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

# Import standard modules.
import math

# Import own modules.
from src.wing_mounted.empirical.landing_gear_design_tu_berlin.general.aicraft_classification_number.flexible_pavement.acnflexcomp import acn_flex_comp  # noPep8 e501


def generate_acn_plots(dict_acn_flex_parameter, dict_constants, runtime_output):
    """ Function to generate the acn vs gross weight charts for flexible and rigid pavement.

    :param dict dict_acn_flex_parameter: Dictionary contains all results of acn flex estimation
    :param dict dict_constants: Dictionary contains all global constants and conversion factors
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return: none
    """

    ''' initialize local parameter '''
    eswl_output_list = []
    alpha_factor_output_list = []
    pavement_thickness_in_mm_output_list = []
    california_bearing_ratio_output_list = []
    pavement_thickness_in_inch_output_list = []
    aircraft_classification_number_output_list = []

    if dict_acn_flex_parameter['InputAlpha'] == 0.0:
        dict_acn_flex_parameter['InputAlpha'] = round(dict_acn_flex_parameter['B'][0], 4)

    ''' initialize iteration masses for plot chart '''
    # round up the maximum ramp mass to the next tens of thousands
    if dict_acn_flex_parameter['AMASS'] > 100000:
        max_ramp_mass = math.ceil(dict_acn_flex_parameter['AMASS'] / 10000) * 10000 + 10000
    elif dict_acn_flex_parameter['AMASS'] > 10000:
        max_ramp_mass = math.ceil(dict_acn_flex_parameter['AMASS'] / 1000) * 1000 + 2500
    else:
        max_ramp_mass = math.ceil(dict_acn_flex_parameter['AMASS'] / 100) * 100 + 500

    # round up the maximum manufacturer empty mass to the next tens of thousands
    if dict_acn_flex_parameter['MEM'] > 100000:
        max_manufacturer_empty_mass = math.floor(dict_acn_flex_parameter['MEM'] / 10000) * 10000
        step_size = 10000
    elif dict_acn_flex_parameter['MEM'] > 10000:
        max_manufacturer_empty_mass = math.floor(dict_acn_flex_parameter['MEM'] / 1000) * 1000
        step_size = 2500
    else:
        max_manufacturer_empty_mass = math.floor(dict_acn_flex_parameter['MEM'] / 100) * 100
        step_size = 500

    # generate list of gross masses for acn estimation
    gross_mass_list = [i for i in range(max_manufacturer_empty_mass, max_ramp_mass, step_size)]

    # save initial values to output lists
    eswl_output_list.append(dict_acn_flex_parameter['PESWL'])
    pavement_thickness_in_mm_output_list.append(dict_acn_flex_parameter['PZ'])
    california_bearing_ratio_output_list.append(dict_acn_flex_parameter['PCBR'])
    aircraft_classification_number_output_list.append(dict_acn_flex_parameter['PACN'])
    alpha_factor_output_list.append(dict_acn_flex_parameter['AlphaFactor'])
    pavement_thickness_in_inch_output_list.append(dict_acn_flex_parameter['CBRThickness'])

    # loop across all gross mass list entries to estimate the aircraft classification number depending on
    for i in range(0, len(gross_mass_list)):
        # reset the output lists for each loop of estimation
        dict_acn_flex_parameter['PZ'] = []
        dict_acn_flex_parameter['PESWL'] = []
        dict_acn_flex_parameter['PCBR'] = []
        dict_acn_flex_parameter['PACN'] = []
        dict_acn_flex_parameter['AlphaFactor'] = []
        dict_acn_flex_parameter['CBRThickness'] = []

        # set the current aircraft gross mass to dictionary mass
        dict_acn_flex_parameter['AMASS'] = gross_mass_list[i]

        # call function to estimate the acn of current aircraft gross mass
        dict_acn_flex_parameter = acn_flex_comp(dict_acn_flex_parameter, dict_constants, runtime_output)

        eswl_output_list.append(dict_acn_flex_parameter['PESWL'])
        pavement_thickness_in_mm_output_list.append(dict_acn_flex_parameter['PZ'])
        aircraft_classification_number_output_list.append(dict_acn_flex_parameter['PACN'])
        pavement_thickness_in_inch_output_list.append(dict_acn_flex_parameter['CBRThickness'])

    dict_acn_flex_parameter['aircraft_classification_number_output_list'] = aircraft_classification_number_output_list
    dict_acn_flex_parameter['gross_mass_list'] = gross_mass_list
    dict_acn_flex_parameter['step_size'] = step_size

    runtime_output.debug('generate_acn_plots successfully executed!')

    return dict_acn_flex_parameter
