/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 * 
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include <memory>
#include <map>
#include <string>
#include <vector>

#include "mission_analysis.h"
#include "standard_mission/standard_mission/standard_mission.h"

/**
 * @brief Construct a new mission_analysis:: mission_analysis object
 * 
 * @param argc argument count
 * @param argv vector of C++ style argument strings
 * @param tool_name entry from toolinfo.h
 * @param tool_version entry from toolinfo.h
 */
MissionAnalysis::MissionAnalysis(const int argc, char *argv[], const std::string &tool_name, const std::string &tool_version)
        : Module(argc, argv, tool_name, tool_version) {
    std::string method = (EndnodeReadOnly<std::string>("program_settings/mode/mission_methods/fidelity_level").read(rtIO_->moduleConfig).value());
    std::string mission_type(EndnodeReadOnly<std::string>("program_settings/mode/mission_methods/mission_type").read(rtIO_->moduleConfig).value());

    if (mission_type == "design_mission" && (!rtIO_->acxml.find("analysis/mission/design_mission"))) { // Catch if we are in the first loop...
        myRuntimeInfo->out << "First iteration for this design mission: Low-fidelity Breguet method was set automatically!" << std::endl;
        myRuntimeInfo->out << "Mid-fidelity methods will be applied in the next iteration..." << std::endl;
        method = "low";
    }
    /* Register low fidelity method (by now, no use-case for other missions than design mission)*/
    missionStrategy_.registerStrategy<breguet::StandardMission>             ({"low", "design_mission"});

    /* Register mid fidelity method */
    missionStrategy_.registerStrategy<design_mission::StandardMission>      ({"mid", "design_mission"      });
    missionStrategy_.registerStrategy<study_mission::StandardMission>       ({"mid", "study_mission"       });
    missionStrategy_.registerStrategy<requirements_mission::StandardMission>({"mid", "requirements_mission"});

    /* Register mid fidelity method */
    // 4D once ready

    /* Start method */
    missionStrategy_.setStrategy({method, mission_type}, rtIO_);
}
