/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 * 
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "../standard_mission/standard_mission.h"

void StandardMission::tanking_and_load_control() {
    this->zero_fuel_mass = this->OME + this->payload_mass.value(); // Initialization of the zero fuel mass
    this->loaded_fuel_mass = this->previous_loaded_fuel_mass; // Get loaded fuel
    double aircraft_mass(this->zero_fuel_mass + this->extra_fuel_mass.second); // Set start value for aircraft mass
    this->set_taxi_fuel(); //Taxi fuel
    this->set_extra_fuel(&aircraft_mass); // Extra fuel on pilot's discretion
    if (fabs(this->get_fuel_sum(this->loaded_fuel_mass)) < ACCURACY_HIGH || low_fidelity) { //no fuel loaded or low fidelity switch on
        if (this->mission_file->air_to_air_fuel_mass < 0.) { // The air to air fuel mass is smaller than 0
            aircraft_mass -= // AirToAir Fuel is only subtracted as a conservative estimation since hereby more fuel consumption will be calculated
                this->mission_file->air_to_air_fuel_mass;
        }
        this->set_trip_fuel(&aircraft_mass);
        this->calculate_reserve_fuel();
        this->update_loaded_fuel();
    }
    /* Check loading condition */
    this->load_control();
}

void StandardMission::set_taxi_fuel() {
    int energy_carrier_id = EndnodeReadOnly<int>("energy_carrier_ID").read(
                                rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion/energy_provider/taxi")).value();
    /* Taxi fuel */
    this->taxi_fuel_mass_origin.first      = energy_carrier_id;
    this->taxi_fuel_mass_destination.first = energy_carrier_id;
    this->engine.calculate_N1_with_penalties(energy_carrier_id, 0., 0., this->atm, 1., "idle",
                                                this->mission_file->mission_data.front().background_bleed,
                                                this->mission_file->mission_data.front().background_shaft_power);
    /* Calculation of the taxi fuel mass for engine warm up after electric taxiing... */
    if (this->mission_file->taxiing_procedure == true) {
        /* ... at origin airport */
        this->taxi_fuel_mass_origin.second = this->mission_file->engine_warmup_time * this->engine.get_aircraft_fuelflow(energy_carrier_id) *
                                EndnodeReadOnly<double>("program_settings/general/fuel_planning/fuel_flow_factor_taxiing").read(this->rtIO->moduleConfig).value();
        /* ... at destination airport */
        this->taxi_fuel_mass_destination.second = 0.;
        /* Output of the total taxi fuel mass after electric taxiing */
        myRuntimeInfo->out << "m_taxiFuel for Engine Warmup after Electric Taxiing: "
                            << this->taxi_fuel_mass_origin.second + this->taxi_fuel_mass_destination.second << std::endl;
    } else { // Calculation of the taxi fuel mass without electric taxiing...
        /* ... at origin airport */
        this->taxi_fuel_mass_origin.second = this->mission_file->taxi_time_origin * this->engine.get_aircraft_fuelflow(energy_carrier_id) *
                                 EndnodeReadOnly<double>("program_settings/general/fuel_planning/fuel_flow_factor_taxiing").read(this->rtIO->moduleConfig).value();
        /* ... at destination airport */
        this->taxi_fuel_mass_destination.second = this->mission_file->taxi_time_destination * this->engine.get_aircraft_fuelflow(energy_carrier_id) *
                                 EndnodeReadOnly<double>("program_settings/general/fuel_planning/fuel_flow_factor_taxiing").read(this->rtIO->moduleConfig).value();
        /* Output of the total taxi fuel mass without electric taxiing */
        myRuntimeInfo->out << "Taxi fuel: " << this->taxi_fuel_mass_origin.second + this->taxi_fuel_mass_destination.second << " kg" << std::endl;
    }
}

void StandardMission::set_extra_fuel(double *aircraft_mass) {
    /* Set the extra fuel mass: Captain Discretion Fuel - "at the captain's discretion" - according to the pilot's "good feeling" - e.g. weather forecast, ... */
    switch (this->fuel_planning) {
        case this->JAR: {
            this->extra_fuel_mass.first = // TODO(Gerrit): Check if it makes sense: Most fuel needed for cruise -> extra fuel added to that
                EndnodeReadOnly<int>("energy_carrier_ID").read(rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion/energy_provider/cruise")).value();
            this->extra_fuel_mass.second = EndnodeReadOnly<double>(
                "program_settings/general/fuel_planning/fuel_estimation/joint_aviation_requirements_parameters/extra_fuel").read(rtIO->moduleConfig).value();
            break;
        }
        default: {
            this->extra_fuel_mass.first = 0; //not provided in FAR
            this->extra_fuel_mass.second = 0.; //not provided in FAR
        }
    }
    /* Output: extra fuel mass rounded to the first decimal place */
    myRuntimeInfo->out << "Extra Fuel: " << Rounding(this->extra_fuel_mass.second, 1) << " kg" << std::endl;
    *aircraft_mass += this->extra_fuel_mass.second;
}

void StandardMission::set_trip_fuel(double* aircraft_mass) {
    /* Setup takeoff and landing fuel mass maps */
    for (const auto& [energy_carrier_id, energy_carrier_map] : rtIO->aircraft_energy_carriers()) {
        this->takeoff_fuel_mass[energy_carrier_id] = 0.;
        this->landing_fuel_mass[energy_carrier_id] = 0.;
    }
    /* If possible, use the trip fuel from the previous run; if not, use a Breguet-based estimation */
    if (this->get_fuel_sum(this->previous_trip_fuel_mass) < 10. || // If no pre-calculated fuel is available, estimate trip fuel for the later used takeoff mass;
        low_fidelity || // if global low fidelity method is turned on, always proceed with estimate_trip_fuel;
        (EndnodeReadOnly<bool>( // if each iteration shall pre-calculate/estimate its own trip fuel, also use estimate_trip_fuel
            "program_settings/general/fuel_planning/fuel_estimation/fuel_estimation_switch").read(this->rtIO->moduleConfig).value())) {
        this->estimate_trip_fuel(*aircraft_mass);
    } else {
        this->trip_fuel_mass = this->previous_trip_fuel_mass;
    }
    /* Output: estimated trip fuel mass rounded to the first decimal place */
    myRuntimeInfo->out << "Trip fuel: " << Rounding(this->get_fuel_sum(this->trip_fuel_mass), 1) << " kg" << std::endl;
    *aircraft_mass += this->get_fuel_sum(this->trip_fuel_mass);
}

void StandardMission::set_contingency_fuel(double *aircraft_mass) {
    /* Contingency Fuel  is defined:
     * JAR:             maximum of the following quantities:
     *                      - 5 min holding at 1500 ft above destination airport at holding speed and ISA-Conditions
     *                      - 5 % of trip-fuel or 3% if en-route alternate is available
     *                      - 15 min holding at 1500 ft above destination airport at holding speed and ISA-Conditions (not used here)
     *                      - 20 min with trip-fuel consumption (not used here)
     * FAR_DOMESTIC:    no contingency fuel
     * FAR_FLAG:        10% of the total required time from brake release at the departure airport to landing to the destination airport (assuming trip-fuel consumption)
     */
    std::string subpath("requirements_and_specifications/design_specification/propulsion/energy_provider/");
    switch (this->fuel_planning) {
        case this->JAR: {
            /* 5 min holding 1500ft above destination */
            this->contingency_fuel_mass.first = // TODO(Gerrit): Check if it makes senese: Assuming descent energy carrier used since we are 'holding at 1500 ft'
                EndnodeReadOnly<int>("energy_carrier_ID").read(rtIO->acxml.at("/descent")).value();
            double holding_time(convertUnit(MINUTE, SECOND, 5.)); //in [s]
            double holding_fuel(this->calculate_holding_fuel(*aircraft_mass, holding_time));
            /* percentage of trip fuel */
            double en_route_allowence(EndnodeReadOnly<double>(
                "program_settings/general/fuel_planning/fuel_estimation/joint_aviation_requirements_parameters/contingency_fuel").read(this->rtIO->moduleConfig).value() *
                this->get_fuel_sum(this->trip_fuel_mass));

            this->contingency_fuel_mass.second = std::max(holding_fuel, en_route_allowence);
            break;
        }
        case this->FAR_FLAG: {
            if (!accuracyCheck(this->previous_block_time, 0., ACCURACY_HIGH) && !accuracyCheck(this->get_fuel_sum(this->loaded_fuel_mass), 0., ACCURACY_HIGH)) {
                this->contingency_fuel_mass.first = // TODO(Gerrit): Check if it makes senese: Assuming cruise energy carrier used since it is derived from cruise
                    EndnodeReadOnly<int>("energy_carrier_ID").read(rtIO->acxml.at("/cruise")).value();
                double contingency_time(0.1 * this->previous_block_time); // 10% of total required time from break release to landing
                this->contingency_fuel_mass.second = contingency_time * this->get_mean_cruise_fuel_flow();
                break;
            }
        }
        default: {
            this->contingency_fuel_mass.first = 0;
            this->contingency_fuel_mass.second = 0.; //not provided in FAR_DOMESTIC
        }
    }
    /* Output: contingency fuel */
    myRuntimeInfo->out << "  - Contingency fuel: " << Rounding(this->contingency_fuel_mass.second, 2) << " kg" << std::endl;
    *aircraft_mass += this->contingency_fuel_mass.second;
}

void StandardMission::set_additional_fuel(double *aircraft_mass) {
    /* Additional Fuel - Holding for 15 Min - 1500 ft */
    additional_fuel_mass.first  = // TODO(Gerrit): Check if it makes senese: Assuming descent energy carrier used since we are 'holding at 1500 ft'
            EndnodeReadOnly<int>("energy_carrier_ID").read(rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion/energy_provider/descent")).value();
    std::string subpath("program_settings/general/fuel_planning/fuel_estimation");
    switch (this->fuel_planning) {
        case this->JAR: {
            if (!EndnodeReadOnly<bool>(subpath +
                    "/joint_aviation_requirements_parameters/use_additional_fuel").read(this->rtIO->moduleConfig).value()) {
                this->additional_fuel_mass.second = 0.;
                break;
            }
        }
        case this->FAR_DOMESTIC:
            this->additional_fuel_mass.second = 0.;
            break;
        case this->FAR_FLAG: {
            if (!EndnodeReadOnly<bool>(subpath +
                    "/federal_aviation_regulations_parameter/use_additional_fuel").read(this->rtIO->moduleConfig).value()) {
                this->additional_fuel_mass.second = 0.;
                break;
            }
        }
        default: {
            double holding_time(convertUnit(MINUTE, SECOND, 15.)); //in [s]
            this->additional_fuel_mass.second = this->calculate_holding_fuel(*aircraft_mass, holding_time);
        }
    }
    /* Output: additional fuel mass rounded to the first decimal place */
    myRuntimeInfo->out << "  - Additional Fuel: " << Rounding(this->additional_fuel_mass.second, 2) << " kg" << std::endl;
    *aircraft_mass += this->additional_fuel_mass.second;
}

void StandardMission::set_final_reserve_fuel(double *aircraft_mass) {
    /* Final Reserve Fuel is defined:
     * JAR:             30 min holding at 1500 ft above destination airport at holding speed and ISA-Conditions
     * FAR_DOMESTIC:    45 min at normal cruise consumption
     * FAR_FLAG:        30 min holding at 1500 ft above destination airport at holding speed and ISA-Conditions
     * JAR and FAR_FLAG final reserve fuel calculation procedure is the same and defined as the default procedure
     */
    std::string subpath("requirements_and_specifications/design_specification/propulsion/energy_provider");
    switch (this->fuel_planning) {
        case this->FAR_DOMESTIC: {
            double final_reserve_time(convertUnit(MINUTE, SECOND, 45.)); // in [s]
            final_reserve_fuel_mass.first = // TODO(Gerrit): Check if it makes senese: Assuming cruise energy carrier used for 'normal cruise consumption'
                EndnodeReadOnly<int>("energy_carrier_ID").read(rtIO->acxml.at(subpath + "/cruise")).value();
            this->final_reserve_fuel_mass.second = final_reserve_time * this->get_mean_cruise_fuel_flow();
            break;
        }
        default: {
            double holding_time(convertUnit(MINUTE, SECOND, 30.));
            final_reserve_fuel_mass.first = // TODO(Gerrit): Check if it makes senese: Assuming descent energy carrier used since we are 'holding at 1500 ft'
            EndnodeReadOnly<int>("energy_carrier_ID").read(rtIO->acxml.at(subpath + "/descent")).value();
            this->final_reserve_fuel_mass.second = this->calculate_holding_fuel(*aircraft_mass, holding_time);
        }
    }
    /* Output: final reserve fuel mass rounded to the first decimal place */
    myRuntimeInfo->out << "  - Final Reserve: " << Rounding(this->final_reserve_fuel_mass.second, 2) << " kg" << std::endl;
    *aircraft_mass += this->final_reserve_fuel_mass.second;
}

void StandardMission::set_alternate_fuel(double* aircraft_mass) {
    int energy_carrier_id = // TODO(Gerrit): Check if it makes sense: Assuming cruise energy carrier used for alternate fuel reserves (derived from cruise conditions)
        EndnodeReadOnly<int>("energy_carrier_ID").read(rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion/energy_provider/cruise")).value();
    /* Alternate fuel: Calculation of the flight altitude from alternate. Climb gradient (2.5°) and decent (1.5°) estimated based on typical ROC and TAS */
    // TODO(Gerrit): Get angles e.g. from certifications or other values previously used
    double alternate_FL = this->mission_file->alternate_distance * 0.8 / (1. / tan(convertUnit(DEGREE, RADIAN, 2.5)) + 1. / tan(convertUnit(DEGREE, RADIAN, 1.5)));
    /* Maximum alternate at FL200 based on Schmitt and Gollnick, Air Transport System */
    if (alternate_FL > convertUnit(HECTO, FOOT, NOPREFIX, METER, 200.)) {
        alternate_FL = convertUnit(HECTO, FOOT, NOPREFIX, METER, 200.);
    }
    double climb_mach(convertUnit(CALIBRATEDAIRSPEED, MACH, alternate_FL, this->atm, this->climb_speed_below_FL100));
    /* Alternate climbing */
    this->engine.calculate_N1_with_penalties(energy_carrier_id, alternate_FL, climb_mach, this->atm, 1., "climb", 0., 0.);
    /* Calculate aerodynamics */
    double L_over_D(this->aero.getCruiseLoverD(climb_mach, alternate_FL, 0., *aircraft_mass, "clean", this->atm));
    if (sophistication_level != UNSOPHISTICATED) {
        double current_cg(get_current_cg());
        L_over_D = this->interpolateViaCoG(current_cg, L_over_D,
            getAlternativeAero(current_cg)->getCruiseLoverD(climb_mach, alternate_FL, 0., *aircraft_mass, "clean", this->atm));
    }
    /* Calculate the specific fuel consumption */
    double SFC = this->engine.get_aircraft_fuelflow(energy_carrier_id) / this->engine.get_thrust_aircraft(energy_carrier_id);
    /* Alternate climbing: Estimate the climbing duration with 5 deg climb angle */
    double climb_time_altn = alternate_FL / (tan(convertUnit(DEGREE, RADIAN, 5.)) * convertUnit(MACH, CALIBRATEDAIRSPEED, alternate_FL, this->atm, climb_mach));
    /* Calculate the alternate climbing fuel mass */
    double m_climb_altn = *aircraft_mass * expm1(climb_time_altn * SFC * G_FORCE / L_over_D);
    /* Calculate the alternate fuel mass */
    L_over_D = this->aero.getCruiseLoverD(this->mission_file->desired_cruise_speed, alternate_FL, 0., *aircraft_mass + m_climb_altn, "clean", this->atm);
    if (sophistication_level != UNSOPHISTICATED) {
        double current_cg(get_current_cg());
        L_over_D = this->interpolateViaCoG(current_cg, L_over_D,
            getAlternativeAero(current_cg)->getCruiseLoverD(this->mission_file->desired_cruise_speed, alternate_FL, 0., *aircraft_mass + m_climb_altn, "clean", this->atm));
    }
    double cruise_drag(this->aero.getCruiseDrag(this->mission_file->desired_cruise_speed, alternate_FL, 0., *aircraft_mass + m_climb_altn, "clean", this->atm));
    if (sophistication_level != UNSOPHISTICATED) {
        double current_cg(get_current_cg());
        cruise_drag = this->interpolateViaCoG(current_cg, cruise_drag,
            getAlternativeAero(current_cg)->getCruiseLoverD(this->mission_file->desired_cruise_speed, alternate_FL, 0., *aircraft_mass + m_climb_altn, "clean", this->atm));
    }
    try {
        this->engine.calculate_N1_with_thrustlimit(energy_carrier_id, alternate_FL, this->mission_file->desired_cruise_speed,
            this->atm, 1., "maximum_continuous", this->average_cruise_bleed, this->average_shaft_bleed, cruise_drag);
    } catch (const int& engine_status) {
        switch (engine_status) {
            case 2: {
                std::stringstream error_message;
                error_message << "Required thrust (" << convertUnit(NOPREFIX, NEWTON, KILO, NEWTON, cruise_drag)
                                << " kN) cannot be delivered by engines during cruise to alternate.";
                error_message << "Reduce drag.";
                throwError(__FILE__, __func__, __LINE__, error_message.str());
            }
            default:
                throwError(__FILE__, __func__, __LINE__, "Engine could not be set to requested status.");
        }
    }
    SFC = this->engine.get_aircraft_fuelflow(energy_carrier_id) / this->engine.get_thrust_aircraft(energy_carrier_id);
    this->alternate_fuel_mass.first = energy_carrier_id;
    this->alternate_fuel_mass.second =  m_climb_altn + (*aircraft_mass + m_climb_altn) *
                        expm1(this->mission_file->alternate_distance / (this->mission_file->desired_cruise_speed *
                        this->atm.getSpeedOfSound(alternate_FL)) * SFC * G_FORCE / L_over_D);
    /* Output: alternate fuel rounded to the first decimal place */
    myRuntimeInfo->out << "  - Alternate Fuel: " << Rounding(this->alternate_fuel_mass.second, 2) << " kg" << std::endl;
    *aircraft_mass += this->alternate_fuel_mass.second;
}

void StandardMission::calculate_reserve_fuel() {
    /* reserve_fuel (EASA) : JAR             := Alternate Fuel + Final Reserve Fuel + Additional Fuel + Contingency Fuel
     * reserve_fuel (FAA)  : FAR_DOMESTIC    := Alternate Fuel + Final Reserve Fuel
     *                      FAR_FLAG        := Alternate Fuel + Final Reserve Fuel + Additional Fuel + Contingency Fuel
     * JAR and FAR_FLAG tanking procedure is the same in terms of fuel quantities and are defined as default procedure
     * the reserve fuel is calculated reverse from end of mission (at zero fuel mass - no extra fuel considered)
    */
    myRuntimeInfo->out << "---------- Reserve Fuel ----------" << std::endl;
    double aircraft_mass(this->zero_fuel_mass);
    switch (this->fuel_planning) {
        case this->FAR_DOMESTIC: {
            this->set_final_reserve_fuel(&aircraft_mass);
            this->set_alternate_fuel(&aircraft_mass);
            break;
        }
        default: {
            this->set_contingency_fuel(&aircraft_mass);
            this->set_additional_fuel(&aircraft_mass);
            this->set_final_reserve_fuel(&aircraft_mass);
            this->set_alternate_fuel(&aircraft_mass);
        }
    }
    myRuntimeInfo->out << "----------------------------------" << std::endl;
}

double StandardMission::calculate_holding_fuel(const double& aircraft_mass, const double& holding_time) {
    int energy_carrier_id = // TODO(Gerrit): Check if it makes sense: Assuming descent energy carrier used for reserves
        EndnodeReadOnly<int>("energy_carrier_ID").read(rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion/energy_provider/descent")).value();
    atmosphere holding_atm; //Airbus defines to calculate holding pattern at ISA-Conditions
    double holding_mach(EndnodeReadOnly<double>("/program_settings/general/fuel_planning/holding/holding_mach_number").read(this->rtIO->moduleConfig).value());
    double holding_altitude(EndnodeReadOnly<double>("/program_settings/general/fuel_planning/holding/holding_altitude").read(this->rtIO->moduleConfig).value());
    if (EndnodeReadOnly<bool>("program_settings/general/fuel_planning/holding/use_economical_speed").read(this->rtIO->moduleConfig).value()) {
        this->calculate_econ_cruise_speed(&holding_mach, holding_altitude, 0., aircraft_mass, this->aero.cruiseConfiguration, "cruise");
    }
    if (convertUnit(METERPERSECOND, KNOTS, convertUnit(MACH, CALIBRATEDAIRSPEED, holding_altitude, holding_atm, holding_mach)) > 250.) {
        holding_mach = convertUnit(CALIBRATEDAIRSPEED, MACH, holding_altitude, holding_atm, convertUnit(KNOTS, METERPERSECOND, 250.));
    }
    /* Calculate the aerodynamics */
    double holding_L_over_D(this->aero.getCruiseLoverD(holding_mach, holding_altitude, 0., aircraft_mass, this->aero.cruiseConfiguration, holding_atm));
    double holding_drag(this->aero.getCruiseDrag(holding_mach, holding_altitude, 0., aircraft_mass, this->aero.cruiseConfiguration, holding_atm));
    if (sophistication_level != UNSOPHISTICATED) {
        double current_cg(get_current_cg());
        holding_L_over_D = this->interpolateViaCoG(current_cg, holding_L_over_D,
            getAlternativeAero(current_cg)->getCruiseLoverD(holding_mach, holding_altitude, 0., aircraft_mass, this->aero.cruiseConfiguration, holding_atm));
        holding_drag = this->interpolateViaCoG(current_cg, holding_drag,
            getAlternativeAero(current_cg)->getCruiseLoverD(holding_mach, holding_altitude, 0., aircraft_mass, this->aero.cruiseConfiguration, holding_atm));
    }
    try {
        this->engine.calculate_N1_with_thrustlimit(energy_carrier_id, holding_altitude, holding_mach, holding_atm, 1.,
                                                        "maximum_continuous", 0., 0., holding_drag);
    } catch (const int& engine_status) {
        switch (engine_status) {
            case 2: {
                std::stringstream error_message;
                error_message << "Required  (" << convertUnit(NOPREFIX, NEWTON, KILO, NEWTON, holding_drag) << " kN) cannot be delivered by engines during holding pattern.";
                error_message << "Reduce drag.";
                throwError(__FILE__, __func__, __LINE__, error_message.str());
            }
            default:
                throwError(__FILE__, __func__, __LINE__, "Engine could not be set to requested status.");
        }
    }
    /* Calculate the specific fuel consumption */
    double SFC = this->engine.get_aircraft_fuelflow(energy_carrier_id) / this->engine.get_thrust_aircraft(energy_carrier_id);
    /* Calculate the mass of the holding fuel */
    return aircraft_mass * expm1(holding_time * SFC * G_FORCE / holding_L_over_D);
}

void StandardMission::estimate_trip_fuel(const double& aircraft_mass) {
    myRuntimeInfo->out << "******************* Estimate initial trip fuel mass *************************" << std::endl;
    int energy_carrier_id = // Assuming cruise energy carrier has highest impact; TODO(Gerrit): Ggf. anhand energy anteilig berechnen
        EndnodeReadOnly<int>("energy_carrier_ID").read(rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion/energy_provider/cruise")).value();
    double current_aircraft_mass(aircraft_mass);
    double current_trip_fuel_mass(0.);
    /* Calculate reserve fuel */
    myRuntimeInfo->info << "    - Calculate Reserve Fuel" << std::endl;
    this->calculate_reserve_fuel(); // Calculate the different reserve fuels
    current_aircraft_mass += this->alternate_fuel_mass.second + this->final_reserve_fuel_mass.second + // Add those reserve fuels
                                    this->additional_fuel_mass.second + this->contingency_fuel_mass.second;
    /* Calculate the mean flight level for cruise */
    double mean_cruise_altitude(
            (this->mission_file->mission_data.at(this->top_of_climb_index).altitude +
                this->mission_file->mission_data.at(this->top_of_descent_index).altitude) * 0.5); // in m

    /* Calculate the specific fuel consumption SFC during cruise */
    // Calculate the cruise drag
    double cruise_drag(this->aero.getCruiseDrag(this->mission_file->desired_cruise_speed, mean_cruise_altitude,
                        0., current_aircraft_mass, this->aero.cruiseConfiguration, this->atm));
    if (sophistication_level != UNSOPHISTICATED) {
        double current_cg(get_current_cg());
        cruise_drag = this->interpolateViaCoG(current_cg, cruise_drag,
            getAlternativeAero(current_cg)->getCruiseDrag(this->mission_file->desired_cruise_speed, mean_cruise_altitude,
                                                            0., current_aircraft_mass, this->aero.cruiseConfiguration, this->atm));
    }
    /* Calculate the cruise thrust using maximum_continuous 
     * Could also be calculated with drag as thrust limit to assure T = D, but that can cause Errors in early design stages.
     * Assuming a high thrust rating in cruise for a rough trip fuel estimation to start the calculation is feasible and robuster
     */
    this->engine.calculate_N1_with_penalties(energy_carrier_id, mean_cruise_altitude, this->mission_file->desired_cruise_speed,
                                            this->atm, 1., "maximum_continuous", this->average_cruise_bleed, this->average_shaft_bleed);
    double cruise_thrust = this->engine.get_thrust_aircraft(energy_carrier_id);
    if (cruise_thrust <= cruise_drag) {
                std::string warning("Required thrust (" + num2Str(convertUnit(NOPREFIX, NEWTON, KILO, NEWTON, cruise_drag)));
                warning += " kN) cannot be delivered by engines even in rating 'maximum_continuous' (";
                warning += num2Str(convertUnit(NOPREFIX, NEWTON, KILO, NEWTON, cruise_thrust));
                warning += " kN). Reduce drag, adapt MTOM or upscale engine!";
                // Within the estimation, only print a warning. If it's the first loop, it might solve itself. If not, it will break soon anyway
                myRuntimeInfo->warn << std::string(warning) << std::endl;
    }

    double SFC_cruise(this->engine.get_aircraft_fuelflow(energy_carrier_id) / this->engine.get_thrust_aircraft(energy_carrier_id));

    /* Calculate the L/D during cruise */
    double L_over_D(0.);
    /* Check if the average cruise lift coefficient is lower than 0.1 */
    if (this->CL_average < 0.1) {
        /* Check if the average cruise lift coefficient is equal to 0. If so, output: CL_average has not yet been determined */
        if (this->CL_average == 0.) {
            myRuntimeInfo->out << "An avg. Cruise-CL has not yet been determined. Use CL_optCruise for L/D calculation." << std::endl;
        } else { // If the average cruise lift coefficient is lower than 0.1 but not equal to 0, warning
            myRuntimeInfo->warn << "Avg. cruise CL is less than 0.1. Use CL_optCruise for L/D calculation." << std::endl;
        }
        L_over_D = this->aero.getCLLoverD(
                    this->mission_file->desired_cruise_speed, this->aero.CLoptimumCruise, this->aero.cruiseConfiguration);
    } else {
        L_over_D = this->aero.getCLLoverD(this->mission_file->desired_cruise_speed, this->CL_average, this->aero.cruiseConfiguration);
        if (sophistication_level != UNSOPHISTICATED) {
            double current_cg(get_current_cg());
            L_over_D = this->interpolateViaCoG(current_cg, L_over_D,
                getAlternativeAero(current_cg)->getCLLoverD(this->mission_file->desired_cruise_speed, this->CL_average, this->aero.cruiseConfiguration));
        }
    }
    /* Tank estimation for trip fuel mass */
    myRuntimeInfo->info << "    - Calculate Cruise Fuel" << std::endl;
    current_trip_fuel_mass += current_aircraft_mass * expm1(this->mission_file->Range / (this->mission_file->desired_cruise_speed *
                            this->atm.getSpeedOfSound(mean_cruise_altitude)) * SFC_cruise * G_FORCE / L_over_D);
    current_aircraft_mass += current_trip_fuel_mass;
    /* Climb fuel mass */
    myRuntimeInfo->info << "    - Calculate Climb Fuel" << std::endl;
    double climb_fuel_mass(0.);
    double climb_mach(convertUnit(CALIBRATEDAIRSPEED, MACH, convertUnit(FOOT, METER, 10000.), this->atm, this->climb_speed_below_FL100));
    /* Get L_over_D */
    L_over_D = this->aero.getCruiseLoverD(climb_mach, convertUnit(FOOT, METER, 10000.), 0., current_aircraft_mass, "clean", this->atm);
    if (sophistication_level != UNSOPHISTICATED) {
        double current_cg(get_current_cg());
        L_over_D = this->interpolateViaCoG(current_cg, L_over_D,
            getAlternativeAero(current_cg)->getCruiseLoverD(climb_mach, convertUnit(FOOT, METER, 10000.), 0., current_aircraft_mass, "clean", this->atm));
    }
    /* Estimation with the values at 10.000 ft */
    this->engine.calculate_N1_with_penalties(energy_carrier_id, convertUnit(FOOT, METER, 10000.), climb_mach, this->atm, 1., "climb", 0., 0.);
    /* Calculate the specific fuel consumption */
    double SFC = this->engine.get_aircraft_fuelflow(energy_carrier_id) / this->engine.get_thrust_aircraft(energy_carrier_id);
    /* Estimate climb duration at 5 deg climb angle */
    double climb_time = this->mission_file->mission_data.at(this->top_of_climb_index).altitude/ (tan(convertUnit(DEGREE, RADIAN, 5.)) *
                     convertUnit(CALIBRATEDAIRSPEED, TRUEAIRSPEED, convertUnit(FOOT, METER, 10000.), this->atm, convertUnit(KNOTS, METERPERSECOND, 200.)));
    /* Calculate the climb fuel mass according to the fuel planning method */
    climb_fuel_mass += current_aircraft_mass * expm1(climb_time * SFC * G_FORCE / L_over_D);
    /* Calculate the trip fuel mass and update trip_fuel_mass accordingly */
    current_trip_fuel_mass += climb_fuel_mass;
    this->trip_fuel_mass[energy_carrier_id] = current_trip_fuel_mass;
}

void StandardMission::update_mass_values() {
    /* Loaded fuel is the summation of taxiFuelOrigin + trip fuel + reserve fuel + extra fuel
     * taxi fuel and extra fuel are independent of mission and do not have to be updated here
     */
    for (const auto& [energy_carrier_id, fuel_mass] : this->mission_profile.back().end_total_fuel_mass) { // Sum all used fuel masses
        this->trip_fuel_mass[energy_carrier_id] = -fuel_mass; // Since consumed fuel is negative, subtract those (negative) masses
    }
    this->calculate_reserve_fuel();
    this->update_loaded_fuel();
    this->takeoff_mass = this->zero_fuel_mass + this->get_fuel_sum(this->loaded_fuel_mass) - this->taxi_fuel_mass_origin.second;
}

void StandardMission::update_loaded_fuel() {
    // Reset loaded fuel masses and start filling them with trip fuel
    for (const auto& [energy_carrier_id, energy_carrier_map] : rtIO->aircraft_energy_carriers()) {
        this->loaded_fuel_mass[energy_carrier_id] = 0; // Reset loaded fuel mass of this energy carrier
        if (trip_fuel_mass.find(energy_carrier_id) != trip_fuel_mass.end()) {
            this->loaded_fuel_mass[energy_carrier_id] += trip_fuel_mass.at(energy_carrier_id);
        }
    }

    // add extra fuel (part of trip fuel)
    this->loaded_fuel_mass[this->extra_fuel_mass.first] += this->extra_fuel_mass.second;

    // Add reserve fuels
    this->loaded_fuel_mass[this->alternate_fuel_mass.first]     += this->alternate_fuel_mass.second;
    this->loaded_fuel_mass[this->final_reserve_fuel_mass.first] += this->final_reserve_fuel_mass.second;
    this->loaded_fuel_mass[this->additional_fuel_mass.first]    += this->additional_fuel_mass.second;
    this->loaded_fuel_mass[this->contingency_fuel_mass.first]   += this->contingency_fuel_mass.second;

    // Add taxi fuel at start only: taxi_fuel_mass_destination is already included in Reserve Fuel according to Airbus/DLR
    this->loaded_fuel_mass[taxi_fuel_mass_origin.first] += this->taxi_fuel_mass_origin.second;
}

void StandardMission::load_control() {
    /* Check if the air to air fuel mass is greater than the half of the mission fuel mass. If so, error message */
    if (this->get_fuel_sum(this->loaded_fuel_mass) / 2. < this->mission_file->air_to_air_fuel_mass) {
        myRuntimeInfo->err << "AirToAir fuel amount " <<  this->mission_file->air_to_air_fuel_mass <<
            " kg > half of mission fuel " << this->get_fuel_sum(this->loaded_fuel_mass) / 2. << " kg! Abort program!" << std::endl;
        exit(1);
    }
    /* Check if the loaded fuel is larger than the tank capacity of the aircraft. If so, error message */
    if (this->get_fuel_sum(this->loaded_fuel_mass) > this->max_fuel_mass) {
        myRuntimeInfo->warn << "Attention: Tanked up fuel mass (" << this->get_fuel_sum(this->loaded_fuel_mass) <<
            " kg) larger than tank capacity (" << this->max_fuel_mass << " kg)!" << std::endl;
        /* Check if the deviation is greater than the allowed fuel overshoot. If so, error message */
        if (EndnodeReadOnly<bool>("program_settings/program_specific/exit_if_fuel_limit_reached/enable").read(this->rtIO->moduleConfig).value() &&
            this->get_fuel_sum(this->loaded_fuel_mass) > (1 + this->rel_fuel_overshoot)*this->max_fuel_mass) {
            myRuntimeInfo->err << "Deviation greater than permissible overshoot from " << 100*this->rel_fuel_overshoot << " percent. Abort program!" << std::endl;
            exit(3);
        }
    }
    /* Check if the actual payload is heavier than the maximum allowed payload. If so, error message */
    if (this->payload_mass.value() > this->max_payload_mass) {
        myRuntimeInfo->err << "Payload too heavy!" << std::endl;
        exit(4);
    }
    /* Calculation of the take off mass */
    double takeoff_mass_tmp = this->zero_fuel_mass + this->get_fuel_sum(this->loaded_fuel_mass) - this->taxi_fuel_mass_origin.second;
    /* Check if the take off mass is exceeded. If so, warning message */
    if (!this->estimate_MTOM && takeoff_mass_tmp > this->MTOM) { // estimate_MTOM = false -> study mission -> don't update MTOM
        myRuntimeInfo->warn << "Maximum takeoff mass exceeded by " << takeoff_mass_tmp - this->MTOM << " kg!" << std::endl;
        /* Check if the deviation is greater than the allowed maximum take off mass overshoot. If so, error message */
        if (EndnodeReadOnly<bool>("program_settings/program_specific/exit_if_maximum_takeoff_mass_limit_reached/enable").read(this->rtIO->moduleConfig).value() &&
            takeoff_mass_tmp > (1 + this->rel_MTOM_overshoot)*this->MTOM) {
            myRuntimeInfo->err << "Deviation greater than permissible overshoot of " << 100*this->rel_MTOM_overshoot << " percent. Abort program!" << std::endl;
            exit(2);
        }
        myRuntimeInfo->warn << "Deviation less than permissible overshoot of " << 100*this->rel_MTOM_overshoot << " percent. Continue calculation..." << std::endl;
    }
}

double StandardMission::get_mean_cruise_fuel_flow() {
    int energy_carrier_id =
        EndnodeReadOnly<int>("energy_carrier_ID").read(rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion/energy_provider/cruise")).value();
    std::string subpath("analysis/mission/" + this->mission_type+ "/cruise");
    /* Calculate the mean aircraft mass during cruise */
    double aircraft_mass_mid_cruise(0.);
    if (rtIO->acxml.find(subpath)) { // if it was calculated previously, take the mean of given cruise mass values
        aircraft_mass_mid_cruise = (EndnodeReadOnly<double>("top_of_climb_mass").read(rtIO->acxml.at(subpath)).value() +
                                    EndnodeReadOnly<double>("top_of_descent_mass").read(rtIO->acxml.at(subpath)).value()) / 2.;
    } else { // if not, use zero fuel mass
        aircraft_mass_mid_cruise = this->zero_fuel_mass;
    }

    /* Calculate the mean flight level for cruise */
    double mean_cruise_FL((this->mission_file->mission_data.at(
            this->top_of_climb_index).altitude+ this->mission_file->mission_data.at(this->top_of_descent_index).altitude) * 0.5); // in m
    /* Calculate the cruise drag */
    double cruise_drag(this->aero.getCruiseDrag(this->mission_file->desired_cruise_speed, mean_cruise_FL, 0.,
                                                aircraft_mass_mid_cruise, this->aero.cruiseConfiguration, this->atm));
    if (sophistication_level != UNSOPHISTICATED) {
        double current_cg(get_current_cg());
        cruise_drag = this->interpolateViaCoG(current_cg, cruise_drag,
            getAlternativeAero(current_cg)->getCruiseDrag(this->mission_file->desired_cruise_speed, mean_cruise_FL, 0.,
                                                            aircraft_mass_mid_cruise, this->aero.cruiseConfiguration, this->atm));
    }
    /* Calculate the cruise thrust */
    try {
        this->engine.calculate_N1_with_thrustlimit(energy_carrier_id, mean_cruise_FL, this->mission_file->desired_cruise_speed,
            this->atm, 1., "maximum_continuous", this->average_cruise_bleed, this->average_shaft_bleed, cruise_drag);
    } catch (const int& engine_status) {
        switch (engine_status) {
            case 2: {
                std::stringstream error_message;
                error_message << "Required thrust (" << convertUnit(NOPREFIX, NEWTON, KILO, NEWTON, cruise_drag)
                                << " kN) cannot be delivered by engines even in rating 'maximum_continuous'. ";
                error_message << "Reduce drag.";
                throwError(__FILE__, __func__, __LINE__, error_message.str());
            }
            default:
                throwError(__FILE__, __func__, __LINE__, "Engine could not be set to requested status.");
        }
    }
    return this->engine.get_aircraft_fuelflow(energy_carrier_id);
}

// Methods to calculate the CoG of total fuel mass
double StandardMission::calculate_fuel_cg(const double& fuel_mass) {
    double cg_tmp(0.);
        myRuntimeInfo->info << "Calculate fuel CoG based on fuel distribution." << std::endl;
        double wing_tank_cg(0), center_tank_cg(0), trim_tank_cg(0);
        double rest_fuel_volume(0.), prior_fuel_volume(0.), wing_tank_fuel_volume(0.), center_tank_fuel_volume(0.), trim_tank_fuel_volume(0.);
        rest_fuel_volume = prior_fuel_volume = fuel_mass / 0.785; // TODO(GERRIT): fuels.at(0).at(0) oder so statt 0.785
        myRuntimeInfo->info << "Fuel distribution: Try to fit remaining fuel (" + num2Str(rest_fuel_volume) + " liters) in wing tanks." << std::endl;
        /* Wing is filled first */
        if (this->number_wing_tanks == 1 || this->number_wing_tanks > 2) {
            throwError(__FILE__, __func__, __LINE__, "Invalid wing tank count: 1 or more than 2 specified, but only 0 or 2 are allowed. Abort program.");
        } else if (this->number_wing_tanks > 0) {
            uint8_t mode(Modeselector(EndnodeReadOnly<std::string>
                ("program_settings/mode/mission_methods/center_of_gravity_method").read(rtIO->moduleConfig).value()).get_mode_numeric());
            if (mode == SEMISOPHISTICATED) {
                wing_tank_cg = this->calculate_lifting_surface_tank_cg(&rest_fuel_volume, &this->wing_surface_vector.at(0));
            } else if (mode == SOPHISTICATED) {
                wing_tank_cg = this->get_Howe_lifting_surface_tank_cg(&rest_fuel_volume);
            } else {
                throwError(__FILE__, __func__, __LINE__, "Center of gravity method unkown. Abort program.");
            }
        }
        wing_tank_fuel_volume = prior_fuel_volume - rest_fuel_volume;
        if (rest_fuel_volume > 0.) {
            /* If fuel remains, center tank is used next */
            myRuntimeInfo->info << "Fuel distribution: Try to fit remaining fuel (" + num2Str(rest_fuel_volume) + " liters) in center tank." << std::endl;
            prior_fuel_volume = rest_fuel_volume;
            center_tank_cg = this->calculate_center_tank_cg(&rest_fuel_volume);
            center_tank_fuel_volume = prior_fuel_volume - rest_fuel_volume;
            if (rest_fuel_volume > 0.) {
                /* If fuel remains, trim tank is finally used */
                myRuntimeInfo->info << "Fuel distribution: Try to fit remaining fuel (" + num2Str(rest_fuel_volume) + " liters) in trim tank." << std::endl;
                prior_fuel_volume = rest_fuel_volume;
                if (number_trim_tanks == 1) {
                    wing_tank_cg = this->calculate_lifting_surface_tank_cg(&rest_fuel_volume, &horizontal_tail_surface_vector.at(0));
                    trim_tank_fuel_volume = prior_fuel_volume - rest_fuel_volume;
                }
            }
        }
        /* Final check if fuel has successfully been distributed */
        if (rest_fuel_volume > 0.) {
            myRuntimeInfo->warn << "Available tank volume not sufficient to store required fuel volume. This should not happen during"
                                << "the last iterations of convergenceLoop." << std::endl;
        } else {
            myRuntimeInfo->info << "Fuel distribution finished." << std::endl;
        }
        /* Set CoG of total fuel mass to CoG of total wing group */
        cg_tmp = (wing_tank_cg * wing_tank_fuel_volume + center_tank_cg * center_tank_fuel_volume +
                    trim_tank_cg * trim_tank_fuel_volume) / (fuel_mass / 0.785); // TODO(GErrit): Wieder echt fueldensity!
    return cg_tmp;
}

double StandardMission::calculate_center_tank_cg(double* rest_fuel_volume) { // TODO(Gerrit): Update to geom2
    /* Assumption: fuel is always distributed equally in the center tank, hence, the CoG is in the center */
    // Vec3 local_ref_point = this->wing_surface_vector.at(0).getLocalRefPoint(this->wing_surface_vector.at(0).referencePty, "right");
    Vec3 center_tank_fuel_cg(0, 0, 0);
    // Vec3 center_tank_fuel_cg(local_ref_point);
    /* Only x needs to be updated, using rel. x center of gravity position of wing tank referenced to quarter chord */
    // double deltaXLocal25toCoG = 0.5 * (this->wing_surface_vector.at(0).getLocalFrontSparPosition(local_ref_point[1], "right")
    //                                    + this->wing_surface_vector.at(0).getLocalRearSparPosition(local_ref_point[1], "right")) - 0.25;
    // deltaXLocal25toCoG *= this->wing_surface_vector.at(0).getLocalChord(local_ref_point[1], "right");
    // center_tank_fuel_cg[0] = local_ref_point[0] + deltaXLocal25toCoG;
    /* Check if center tank volume is enough to fit remaining fuel */
    this->calculate_rest_fuel_volume(rest_fuel_volume, "center");
    return center_tank_fuel_cg[0];
}

double StandardMission::calculate_lifting_surface_tank_cg(double* rest_fuel_volume, geom2::MultisectionSurface<geom2::AirfoilSection>* a_lifting_surface) {
    double fuel_cg(0.), right_side_fuel_cg(0.), left_side_fuel_cg(0.), nominator_cg_right_wing(0.); // TODO(Gerrit): Update to geom2
    // double right_side_tank_volume(0.), left_side_tank_volume(0.), step_length(0.), max_half_span_width(0.), fuselage_radius(0.), wing_tank_fuel_volume(0.);
    // uint8_t number_of_steps(25);
    // std::vector<uint16_t> wing_tank_IDs = this->find_elements_ID(rtIO->acxml, "aircraft_exchange_file/component_design/tank/specific/tank",
    //     wingTankTypes, "/name/value");
    // for (uint16_t i : wing_tank_IDs) {
    //     wing_tank_fuel_volume += EndnodeReadOnly<double>("component_design/tank/specific/tank@" + num2Str(i) + "/volume").read(rtIO->acxml).value();
    // }
    // max_half_span_width = a_lifting_surface->RightWing.back().segReference[1] - a_lifting_surface->referencePty;
    // fuselage_radius = fuselage_vector.at(0).getLocalWidth(a_lifting_surface->referencePtx) / 2.0;
    // step_length = (max_half_span_width - fuselage_radius) / number_of_steps;
    // /* Calculate tank volume for incremental parts of the tank. Stops if required fuel is reached.
    //     Simplification: Fuel stays in the incremental parts and does not float from gravity */
    // for (uint8_t steps(0); steps < number_of_steps - 1; ++steps) {
    //     /* Right lifting surface tank, starting at fuselage lifting surface junction */
    //     double left_y_position = fuselage_radius + steps * step_length;
    //     double base_area = a_lifting_surface->getAreaBetweenSpars(left_y_position, skin_thickness, "right");
    //     double area_at_middle_height = a_lifting_surface->getAreaBetweenSpars(left_y_position + 0.5 * step_length, skin_thickness, "right");
    //     double top_area = a_lifting_surface->getAreaBetweenSpars(left_y_position + step_length, skin_thickness, "right");
    //     /* Determine incremental tank volume using prismatoid equation */
    //     double incremental_tank_volume(step_length * (base_area + 4. * area_at_middle_height + top_area) / 6.);
    //     right_side_tank_volume += incremental_tank_volume;
    //     /* Determine local ref position position (quarter chord point) at half of incremental part */
    //     Vec3 local_ref_point(a_lifting_surface->getLocalRefPoint(left_y_position + 0.5 * step_length, "right"));
    //     Vec3 local_cg_point(local_ref_point);
    //     /* Determine x coordinate in more detail, using rel. x CoG position of wing tank referenced to quarter chord */
    //     double deltaXLocal25toCoG = 0.5 * (a_lifting_surface->getLocalFrontSparPosition(local_ref_point[1], "right")
    //                                         + a_lifting_surface->getLocalRearSparPosition(local_ref_point[1], "right")) - 0.25;
    //     deltaXLocal25toCoG *= a_lifting_surface->getLocalChord(local_ref_point[1], "right");
    //     local_cg_point[0] = local_ref_point[0] + deltaXLocal25toCoG;
    //     /* Calculate nominator for center of gravity */
    //     nominator_cg_right_wing += incremental_tank_volume * local_cg_point[0];
    //     /* Symmetry is assumed */
    //     left_side_tank_volume = right_side_tank_volume;
    //     /* Stop if required fuel fits into current wing tank space */
    //     if (convertUnit(VOLUME, NOPREFIX, METER, NOPREFIX, LITER, right_side_tank_volume)
    //         + convertUnit(VOLUME, NOPREFIX, METER, NOPREFIX, LITER, left_side_tank_volume) >= (*rest_fuel_volume)) {
    //         (*rest_fuel_volume) = 0.;
    //         break;
    //     }
    //     /* Stop if Tankvolume gets bigger than the values in Ac.xml */
    //     if (convertUnit(VOLUME, NOPREFIX, METER, NOPREFIX, LITER, right_side_tank_volume) +
    //         convertUnit(VOLUME, NOPREFIX, METER, NOPREFIX, LITER, left_side_tank_volume) >= wing_tank_fuel_volume) {
    //         break;
    //     }
    // }
    // /* Determine CoGs of both wing sides */
    // right_side_fuel_cg = nominator_cg_right_wing / right_side_tank_volume;
    // left_side_fuel_cg = right_side_fuel_cg;
    // /* Determine combined CoG from both wing sides */
    // fuel_cg = (right_side_fuel_cg * right_side_tank_volume + left_side_fuel_cg * left_side_tank_volume) / (right_side_tank_volume + left_side_tank_volume);
    // /* Determine remaining fuel volume */
    // if (*rest_fuel_volume != 0.) {
    //     right_side_tank_volume = convertUnit(VOLUME, NOPREFIX, METER, NOPREFIX, LITER, right_side_tank_volume);
    //     left_side_tank_volume = convertUnit(VOLUME, NOPREFIX, METER, NOPREFIX, LITER, left_side_tank_volume);
    //     (*rest_fuel_volume) -= (right_side_tank_volume + left_side_tank_volume);
    // }
    // /* Warning, if wingTank is completely filled up, but not all of the volume set in Ac.xml is used */
    // if ((*rest_fuel_volume) > 0 && right_side_tank_volume + left_side_tank_volume < wing_tank_fuel_volume) {
    //     myRuntimeInfo->warn << "Lifting Surface Tank volume for Fuel CoG is smaller than the tank volume value set in Ac.xml" << std::endl;
    // }
    return fuel_cg;
}

double StandardMission::get_Howe_lifting_surface_tank_cg(double* rest_fuel_volume) {
    this->calculate_rest_fuel_volume(rest_fuel_volume, "wing");
    return this->cg_wing;
}

void StandardMission::calculate_rest_fuel_volume(double* rest_fuel_volume, const std::string& tank_type) {
    double tank_volume(0.);
    if (tank_type == "center") {
        std::vector<uint16_t> centerTankIDs = this->find_elements_ID(rtIO->acxml, "aircraft_exchange_file/component_design/tank/specific/tank",
        this->centerTankTypes, "/name/value");
        for (const uint16_t& i : centerTankIDs) {
            tank_volume += EndnodeReadOnly<double>("component_design/tank/specific/tank@" + num2Str(i) + "/volume").read(rtIO->acxml).value();
        }
    } else if (tank_type == "wing") {
        std::vector<uint16_t> wing_tank_IDs = this->find_elements_ID(rtIO->acxml,  "aircraft_exchange_file/component_design/tank/specific/tank",
        this->wingTankTypes, "/name/value");
        for (const uint16_t& i : wing_tank_IDs) {
            tank_volume += EndnodeReadOnly<double>("component_design/tank/specific/tank@" + num2Str(i) + "/volume").read(rtIO->acxml).value();
        }
    } else {
        myRuntimeInfo->err << "Error calculating rest fuel volume with unknown tank type during CG calculation." << std::endl;
    }
    if ((*rest_fuel_volume) <= tank_volume) {
        (*rest_fuel_volume) = 0.;
    } else {
        (*rest_fuel_volume) -= tank_volume;
    }
}

double StandardMission::get_fuel_sum(std::map<int, double> fuel_map) {
    return (std::accumulate(fuel_map.begin(), fuel_map.end(), 0.,
            [](double total, const std::pair<int, double>& fuel_mass) { return total + fuel_mass.second; }));
}

std::map<int, double> StandardMission::fuel_map_setup(const std::string& subpath, const std::string& energy_nodes) {
    std::map<int, double> energy_map;
    if (rtIO->acxml.find(subpath + "/" + energy_nodes)) { // if energy from previous run is available, loop through all energy carriers
        for (node* energy_node : rtIO->acxml.at(subpath).getVector(energy_nodes)) {
            int energy_carrier_id = EndnodeReadOnly<int>("energy_carrier_ID").read(*(energy_node)).value();
            energy_map[energy_carrier_id] = EndnodeReadOnly<double>("consumed_energy").read(*(energy_node)).value() /
                EnergyCarrier(rtIO->get_fuel_type(energy_carrier_id), rtIO->get_fuel_density(energy_carrier_id)).gravimetric_energy_density;
        }
    }
    return energy_map;
}

void StandardMission::update_energy_nodes(const std::map<int, double>& fuel_map, const std::string& subpath,
                                            const std::string& description, Endnode<int>* energy_carrier_id,
                                            Endnode<double>* consumed_energy) {
    uint16_t xml_id(0);
    for (const auto& [key, fuel_mass] : fuel_map) { // loop through all used energy carriers
        consumed_energy->set_value(fuel_mass * EnergyCarrier(rtIO->get_fuel_type(key), rtIO->get_fuel_density(key)).gravimetric_energy_density);
        energy_carrier_id->set_value(key);
        this->check_and_add_nodes(subpath, description, xml_id); // if no nodes where added yet, do so
        consumed_energy->update(rtIO->acxml.at(subpath + "@" + num2Str(xml_id)));
        energy_carrier_id->update(rtIO->acxml.at(subpath + "@" + num2Str(xml_id)));
        xml_id++;
    }
}

std::vector<uint16_t> StandardMission::find_elements_ID(const node& xml, const std::string& path_to_node,
                                                                const std::vector<std::string>& keywords, const std::string& subpath) {
  std::vector<uint16_t> IDs;
  for (int i = 0; i < xml.getVector(path_to_node).size(); i++) {
    for (int j = 0; j < keywords.size(); j++) {
      if (std::string(xml.at(path_to_node + "@" + num2Str(i) + subpath)) ==  keywords[j]) {
        IDs.push_back(i);
        continue;
      }
    }
  }
  if (IDs.size() == 0) {
    std::string entries = "Specified tank components don't match list of known types. For this category, these are: ";
    for (int i = 0; i < keywords.size(); i++) {
      if (i != 0) {
        entries += ", " + keywords[i];
      } else {
        entries += keywords[0];
      }
    }
    myRuntimeInfo->warn << entries << std::endl;
  }
  return IDs;
}
