/*  Copyright (C) 2012-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "VnDiagram.h"

namespace low {

VnDiagram::VnDiagram(std::shared_ptr<aircraft> theAircraft, const std::shared_ptr<performance_assessment_settings> theSettings)
    :
    CleanCoefficientPos(0.),
    CleanCoefficientNeg(0.),
    FlapsCoefficientTO(0.),
    FlapsCoefficientLDG(0.),
    upperBoundClean(0.),
    lowerBoundClean(0.),
    upperBoundFlaps(0.),
    lowerBoundFlaps(0.),
    theAcftPt(theAircraft),
    mySettingsPt(theSettings),
    step(1.) {
}
void VnDiagram::doVnDiagram() {
    /* Calculation of the decisive coefficients, which define the shape of the first part of the V-n diagram (quadratic shape) Coeff*pow(v,2) */
    CleanCoefficientPos = (theAcftPt->aero.getCLmax(0., "clean") * 0.5 * theAcftPt->atm.getDensity(0.) * theAcftPt->S_Wing) / (theAcftPt->MTOM * G_FORCE);
    CleanCoefficientNeg = -CleanCoefficientPos;  // Assumption: minimum lift coefficient is as high as positive one
    FlapsCoefficientTO = (theAcftPt->aero.CLmaxTakeoff * 0.5 * theAcftPt->atm.getDensity(0.) * theAcftPt->S_Wing) / (theAcftPt->MTOM * G_FORCE);
    // Note here the use of MLM instead of MTOM
    FlapsCoefficientLDG = (theAcftPt->aero.CLmaxLanding * 0.5 * theAcftPt->atm.getDensity(0.) * theAcftPt->S_Wing) / (theAcftPt->MLM * G_FORCE);
    /* Diagram for die Clean-configuration */
    for (int i = 0 ; i <= theAcftPt->VDive + convertUnit(KNOTS, METERPERSECOND, 5); i += step) {
        tempValues.speed = i;
        tempValues.n_upperBound = calc_upperBoundClean(CleanCoefficientPos, tempValues.speed);
        tempValues.n_lowerBound = calc_lowerBoundClean(CleanCoefficientNeg, tempValues.speed);
        myCleanDiagram.push_back(tempValues);
    }
    /* Diagram for the Takeoff configuration */
    for (int i = 0 ; i <= theAcftPt->VDive + 5; i += step) {  // TODO(Ufermann): Determine maximum speed for TakeOff configuration
        tempValues.speed = i;
        tempValues.n_upperBound = calc_upperBoundFlaps("takeoff", FlapsCoefficientTO, tempValues.speed);
        tempValues.n_lowerBound = 0;
        myTakeoffDiagram.push_back(tempValues);
    }
    /* Diagram for the Landing configuration */
    for (int i = 0 ; i <= theAcftPt->VDive + convertUnit(KNOTS, METERPERSECOND, 5); i += step) {  // TODO(Ufermann): Determine maximum speed for landing configuration
        tempValues.speed = i;
        tempValues.n_upperBound = calc_upperBoundFlaps("landing", FlapsCoefficientLDG, tempValues.speed);
        tempValues.n_lowerBound = 0;
        myLandingDiagram.push_back(tempValues);
    }
}


double VnDiagram::calc_upperBoundClean(double Coefficient, double v) {  // v in [m/s]
    double n;
    n = Coefficient * pow(v, 2);
    // Lower v Limit:
    if (v < theAcftPt->v1gS_Clean(theAcftPt->MTOM, 0)) {
        return 0;
    }
    // Upper v Limit
    if (v >= theAcftPt->VDive) {
        return 0;
    }
    // maximum n from config
    if (n >= mySettingsPt->myVnDiagramConstants.maxNClean) {
        return mySettingsPt->myVnDiagramConstants.maxNClean;
    } else {
        return n;
    }
}

double VnDiagram::calc_lowerBoundClean(double Coefficient, double v) {
    double n;
    n = Coefficient * pow(v, 2);
    // lower v limit
    if (v < theAcftPt->v1gS_Clean(theAcftPt->MTOM, 0)) {
        return 0;
    }
    // upper v limit
    if (v >= theAcftPt->VDive) {
        return 0;
    }
    // If v is interpolated linearly between VMO and VDive
    if (v >= theAcftPt->VMO && v <= theAcftPt->VDive) {
        return mySettingsPt->myVnDiagramConstants.minNClean + ((0 - mySettingsPt->myVnDiagramConstants.minNClean) /
                                                               (theAcftPt->VDive - theAcftPt->VMO)) *
                                                               (v - theAcftPt->VMO);
    }
    // n limit
    if (n <= mySettingsPt->myVnDiagramConstants.minNClean) {
        return mySettingsPt->myVnDiagramConstants.minNClean;
    } else {
        return n;
    }
}

double VnDiagram::calc_upperBoundFlaps(const string &configuration, double Coefficient, double v) {
    double n(Coefficient * pow(v, 2));
    if (configuration == "takeoff") {
        // upper limit v
        if (v >= theAcftPt->VDive) {
            return 0;
        }
        // lower limit v
        if (v <= theAcftPt->v1gS_TO(theAcftPt->MTOM, 0)) {
            return 0;
        }
        // upper limit n from config
        if (n >= mySettingsPt->myVnDiagramConstants.maxNFlaps) {
            return mySettingsPt->myVnDiagramConstants.maxNFlaps;
        } else {
            return n;
        }
    }
    if (configuration == "landing") {
        // upper limit v
        if (v >= theAcftPt->VDive) {
            return 0;
        }
        // lower limit v
        if (v <= theAcftPt->getReferenceStallSpeed(theAcftPt->MLM, 0., theAcftPt->aero.CLmaxLanding)) {
            return 0;
        }
        // upper limit n from config
        if (n >= mySettingsPt->myVnDiagramConstants.maxNFlaps) {
            return mySettingsPt->myVnDiagramConstants.maxNFlaps;
        } else {
            return n;
        }
    }
    return n;
}
}  // namespace low
