/*  Copyright (C) 2012-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "landingPerformance.h"

using std::endl;

namespace low {

landingPerformance::landingPerformance(std::shared_ptr<aircraft> theAircraft)
    :
    theAcftPt(theAircraft),
    runwaySlopeAngle(0.0),
    brakingFriction(0.0),
    windSpeed(0.0) {
    // ctor
}
landingPerformance::~landingPerformance() {
    // dtor
}

void landingPerformance::doLandingPerformance(double theRunwaySlope, double theGroundFrictionBraking, double theHeadWind) {
    runwaySlopeAngle = theRunwaySlope;
    brakingFriction = theGroundFrictionBraking;
    windSpeed = theHeadWind;
    /* Calculation of landing distance as a function of mass and height */
    for (double elevation = 0.0; elevation <= convertUnit(FOOT, METER, 12000.); convertUnit(FOOT, METER, elevation += 2000.)) {  // calculation for several runway altitudes [ft]
        myLDNField.push_back(dataElevationLDN());
        myLDNField.back().elevation = elevation;
        for (int i = 0; i <= 10; i++) {  // calculation for several masses between OME and MLM
            myLDNField.back().myLDN.push_back(dataLDN());
            myLDNField.back().myLDN.back().grossMass = theAcftPt->OME + static_cast<double>(i) / 10.*(theAcftPt->MLM - theAcftPt->OME);
            calcALD(&myLDNField.back().myLDN.back().ALD_AEO, myLDNField.back().myLDN.back().grossMass, myLDNField.back().elevation);
            // Safety margin for dry runway (FAR121.195: requiredDistance = 60 percent of availableDistance)
            myLDNField.back().myLDN.back().ALD_AEO /= 0.6;  // maximum allowable distance including safety margin
        }
    }
    theAcftPt->myChecks.LDN                         = checkForLDN();
    theAcftPt->myChecks.vApproach                   = checkForApproachSpeed();
    checkForApproachClimb();
}

void landingPerformance::calcALD(double* landingDistanceRequired, const double& acMass, const double& Altitude) {
    // landingDistanceRequired: Available landing distance (m): dry runway
    // Approach 50ft to 0ft Runway
    double glideSlopeAngle(3.);  // [deg]
    double loadFactorLanding(1.1);  // Load factor during flare in accordance to Bruening_1986_Flugleistungen
    // Landing distance x_L1 in accordance to Bruening_1986_Flugleistungen and Lecture Chap.10: Takeoff- and Landing in Flugzeugbau 1
    *landingDistanceRequired = convertUnit(FOOT, METER, 50.) / tan(convertUnit(DEGREE, RADIAN, glideSlopeAngle)) +
                               convertUnit(DEGREE, RADIAN, glideSlopeAngle) * pow(theAcftPt->vAPPR(acMass, Altitude), 2.) / (2. * G_FORCE * (loadFactorLanding - 1.));
    // Ground Run until Stop
    int stepSize(50);
    double deltaV(- (theAcftPt->vAPPR(acMass, Altitude) - windSpeed) / stepSize);
    for (int i = 0; i <= stepSize; i++) {
        double vGround(theAcftPt->vAPPR(acMass, Altitude) - windSpeed + i * deltaV);
        double vAir(0.0);
        if (vGround >= windSpeed) {
            vAir = windSpeed + vGround;
        }
        double deAcceleration = calcDeAccelerationGroundRoll(acMass, Altitude, vAir);
        // trapezoidal rule
        if (i == 0 || i == stepSize) {
            *landingDistanceRequired += 0.5 * vGround / deAcceleration * deltaV;
        } else {
            *landingDistanceRequired += vGround / deAcceleration * deltaV;
        }
    }
}

double landingPerformance::calcDeAccelerationGroundRoll(double mass, double alt, double Velocity) {
    double Mach = Velocity / theAcftPt->atm.getSpeedOfSound(alt);
    // theAcftPt->myEngine.setEngineRating( alt, Mach, theAcftPt->atm, 1.0, "TO", theAcftPt->myMission.bleedOffTO, theAcftPt->myMission.shaftOffTO );
    double tempThrust(0.0);
    double tempLift(0.0);
    double tempDrag(0.0);
    tempDrag = theAcftPt->aero.getCLDrag(Mach, alt, theAcftPt->aero.CLgroundRoll, "landing", theAcftPt->atm);
    return 1.0 / mass * (tempThrust - mass * G_FORCE * (brakingFriction + sin(runwaySlopeAngle)) - (tempDrag - brakingFriction * tempLift));
}

bool landingPerformance::checkForLDN() {
    bool Check(false);
    if (myLDNField.at(0).myLDN.back().ALD_AEO >= theAcftPt->myReqs.LDN) {
        myRuntimeInfo->out << "The required landing distance cannot be maintained!" << endl;
        myRuntimeInfo->out << "LDN_reg  = " << theAcftPt->myReqs.LDN << " m" << endl;
        myRuntimeInfo->out << "LDN_calc = " << myLDNField.at(0).myLDN.back().ALD_AEO << " m" << endl;
        Check = false;
    } else {
        Check = true;
    }
    return Check;
}

bool landingPerformance::checkForApproachSpeed() {
    bool Check(false);
    if (theAcftPt->vAPPR(theAcftPt->MLM, 0.0) > theAcftPt->myReqs.approachSpeed) {
        myRuntimeInfo->out << "The required approach speed cannot be maintained.!" << endl;
        myRuntimeInfo->out << "Req.  V_approach: " << theAcftPt->myReqs.approachSpeed << " m/s" << endl;
        myRuntimeInfo->out << "Calc. V_approach: " << theAcftPt->vAPPR(theAcftPt->MLM, 0.0) << " m/s" << endl;
        Check = false;
    } else {
        Check = true;
    }
    return Check;
}

void landingPerformance::checkForApproachClimb() {
    double tempGradientOEI(9999.);
    double tempGradientAEO(9999.);

    for (int i = 0; i <= 1; i++) {
        double tempAlt(0.);
        if (i == 0) {
            tempAlt = 50.*0.3048;
        }
        double tempTAS(theAcftPt->vAPPR(theAcftPt->MLM, tempAlt));
        double tempMach(tempTAS / theAcftPt->atm.getSpeedOfSound(tempAlt));
        double tempROC(0.0);
        tempROC = theAcftPt->ROC(theAcftPt->MLM, tempAlt, tempMach, 0.0,
                                 "landing", "takeoff", theAcftPt->myMission.bleedOffTO,  theAcftPt->myMission.shaftOffTO, true, theAcftPt->aero);
        if (tempGradientOEI > asin(tempROC / tempTAS)) {
            tempGradientOEI = 100. * asin(tempROC / tempTAS);
        }
        tempROC = theAcftPt->ROC(theAcftPt->MLM, tempAlt, tempMach, 0.0,
                                 "landing", "takeoff", theAcftPt->myMission.bleedOffTO,  theAcftPt->myMission.shaftOffTO, false, theAcftPt->aero);
        if (tempGradientAEO > asin(tempROC / tempTAS)) {
            tempGradientAEO = 100. * asin(tempROC / tempTAS);
        }
    }
    if (tempGradientOEI < theAcftPt->myReqs.climbGradientApproachOEI) {
        myRuntimeInfo->out << "The required gradient of climb in Missed Approach with OEI cannot be maintained.!" << endl;
        myRuntimeInfo->out << "Req. : " << theAcftPt->myReqs.climbGradientApproachOEI << " %" << endl;
        myRuntimeInfo->out << "Calc : " << tempGradientOEI << "%" << endl;
        theAcftPt->myChecks.climbGradientApproachOEI = false;
    } else {
        theAcftPt->myChecks.climbGradientApproachOEI = true;
    }
    if (tempGradientAEO < theAcftPt->myReqs.climbGradientApproachAEO) {
        myRuntimeInfo->out << "The required gradient of climb in Missed Approach with AEO cannot be maintained.!" << endl;
        myRuntimeInfo->out << "Req. : " << theAcftPt->myReqs.climbGradientApproachAEO  << " %" << endl;
        myRuntimeInfo->out << "Calc : " << tempGradientAEO  << " %" << endl;
        theAcftPt->myChecks.climbGradientApproachAEO = false;
    } else {
        theAcftPt->myChecks.climbGradientApproachAEO = true;
    }
}

}  // namespace low
