/*  Copyright (C) 2012-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "performance_assessment_factory.h"
#include "performance_assessment_output.h"  // Resolving circular dependency

namespace taw {

performance_assessment_factory::performance_assessment_factory(const std::shared_ptr<RuntimeIO>& rtIOin)
    :
    rtIO(rtIOin),
    mySettings(rtIOin),
    mySettingsPt(std::make_shared<performance_assessment_settings>(mySettings)),
    myAcft(rtIO, mySettingsPt),
    myAcftPt(std::make_shared<aircraft>(myAcft))
    // missionXML(aixml::openDocument(static_cast<std::string>(rtIO->acxml.at(
        // "/aircraft_exchange_file/requirements_and_specifications/mission_files/study_mission_file/value")))) 
        {
}

void performance_assessment_factory::initialize() {
    /* Get desired fidelity level of methods*/
    fidelityLevelFuelPlanning = static_cast<std::string>(rtIO->moduleConfig.at("/module_configuration_file/program_settings/fuel_planning/fidelity_level"));
    fidelityLevelOptimizeMission = static_cast<std::string>(rtIO->moduleConfig.at("/module_configuration_file/program_settings/mission/fidelity_level"));
    fidelityLevelPayloadRangeDiagram = static_cast<std::string>(rtIO->moduleConfig
        .at("/module_configuration_file/program_settings/performance_checks/payload_range_diagram/fidelity_level"));
    fidelityLevelEnginePerformance = static_cast<std::string>(rtIO->moduleConfig
        .at("/module_configuration_file/program_settings/performance_checks/engine_performance/fidelity_level"));
    fidelityLevelFlightEnvelopePerformance = static_cast<std::string>(rtIO->moduleConfig
        .at("/module_configuration_file/program_settings/performance_checks/flight_envelope_performance/fidelity_level"));
    fidelityLevelClimbPerformance = static_cast<std::string>(rtIO->moduleConfig
        .at("/module_configuration_file/program_settings/performance_checks/climb_performance/fidelity_level"));
    fidelityLevelTakeOffPerformance = static_cast<std::string>(rtIO->moduleConfig
        .at("/module_configuration_file/program_settings/performance_checks/takeoff_performance/fidelity_level"));
    fidelityLevelLandingPerformance = static_cast<std::string>(rtIO->moduleConfig
        .at("/module_configuration_file/program_settings/performance_checks/landing_performance/fidelity_level"));
    fidelityLevelVnDiagram = static_cast<std::string>(rtIO->moduleConfig.at("/module_configuration_file/program_settings/performance_checks/vn_diagram/fidelity_level"));
    fidelityLevelSARPerformance = static_cast<std::string>(rtIO->moduleConfig.at("/module_configuration_file/program_settings/performance_checks/sar_performance/fidelity_level"));

    /* Initialization of performance_assessment objects */
    /* Fuel Planning */
    if (fidelityLevelFuelPlanning == "high") {  // high fidelity fuel Planning
        // Custom implementation here, e.g.:
        // abstractFuelPlanningPt = std::make_shared<high::fuelPlanning>(someArgs);  // Create Object and puts it in interface class pointer
    } else if (fidelityLevelFuelPlanning == "low") {  // low fidelity fuelPlanning
        abstractFuelPlanningPt = std::make_shared<low::fuelPlanning>(myAcftPt, mySettingsPt, rtIO);  // Create Object and puts it in interface class pointer
        // abstractFuelPlanningPt->setFactory(shared_from_this());  // Enables Factory-Products to access the factory to check what fidelity levels other products are
    } else {
        myRuntimeInfo->warn << "Unavailable fidelity level for 'FuelPlanning' selected! Exiting"<< std::endl;
        exit(1);
    }
    /* Mission Optimization */
        if (fidelityLevelOptimizeMission == "high") {  // high fidelity optimizeMission
        // ...
    } else if (fidelityLevelFuelPlanning == "low") {  // low fidelity optimizeMission
        abstractOptimizeMissionProfilePt = std::make_shared<low::optimizeMissionProfile>(&myAcft);  // Create Object and puts it in interface class pointer
    } else {  // default case: low
        myRuntimeInfo->warn << "Unavailable fidelity level for 'MissionOptimization' selected! Exiting"<< std::endl;
        exit(1);
    }
    /* PayloadRange */
    if (fidelityLevelPayloadRangeDiagram == "high") {
        // ...
    } else if (fidelityLevelPayloadRangeDiagram == "low") {
        abstractPayloadRangePt = std::make_shared<low::payloadRange>(&myAcft, mySettings);
    } else {
        myRuntimeInfo->warn << "Unavailable fidelity level for 'PayloadRange' selected! Exiting"<< std::endl;
        exit(1);
    }
    /* EnginePerformance */
    if (fidelityLevelEnginePerformance == "high") {
        // ...
    } else if (fidelityLevelEnginePerformance == "low") {
        abstractEnginePerformancePt = std::make_shared<low::enginePerformance>(myAcftPt);
    } else {
        myRuntimeInfo->warn << "Unavailable fidelity level for 'EnginePerformance' selected! Exiting"<< std::endl;
        exit(1);
    }
    /* ClimbPerformance */
    if (fidelityLevelClimbPerformance == "high") {
        // ...
    } else if (fidelityLevelClimbPerformance == "low") {
        abstractCeilingPerformancePt = std::make_shared<low::ceilingPerformance>(myAcftPt);
    } else {
        myRuntimeInfo->warn << "Unavailable fidelity level for 'ClimbPerformance' selected! Exiting"<< std::endl;
        exit(1);
    }
    /* Flight Envelope */
    if (fidelityLevelFlightEnvelopePerformance == "high") {
        // ...
    } else if (fidelityLevelFlightEnvelopePerformance == "low") {
        abstractFlightEnvelopesPt = std::make_shared<low::flightEnvelopes>(myAcftPt);
    } else {
        myRuntimeInfo->warn << "Unavailable fidelity level for 'FlightEnvelope' selected! Exiting"<< std::endl;
        exit(1);
    }

    /* SARPerformance */
    if (fidelityLevelSARPerformance == "high") {
        // ...
    } else if (fidelityLevelSARPerformance == "low") {
        abstractSARPerformancePt = std::make_shared<low::SARPerformance>(myAcftPt);
    } else {
        myRuntimeInfo->warn << "Unavailable fidelity level for 'SARPerformance' selected! Exiting"<< std::endl;
        exit(1);
    }
    /* Take-Off performance */
    if (fidelityLevelTakeOffPerformance == "high") {
        // ...
    } else if (fidelityLevelTakeOffPerformance == "low") {
        abstractTakeOffPerformancePt = std::make_shared<low::takeOffPerformance>(myAcftPt);
    } else {
        myRuntimeInfo->warn << "Unavailable fidelity level for 'takeOffPerformance' selected! Exiting"<< std::endl;
        exit(1);
    }
    /* Landing Performance */
    if (fidelityLevelLandingPerformance == "high") {
        // ...
    } else if (fidelityLevelLandingPerformance == "low") {
        abstractLandingPerformancePt = std::make_shared<low::landingPerformance>(myAcftPt);
    } else {
        myRuntimeInfo->warn << "Unavailable fidelity level for 'landingPerformance' selected! Exiting"<< std::endl;
        exit(1);
    }
    /* Vn-Diagram */
    if (fidelityLevelVnDiagram == "high") {
        // ...
    } else if (fidelityLevelVnDiagram == "low") {
        abstractVnDiagramPt = std::make_shared<low::VnDiagram>(myAcftPt, mySettingsPt);
    } else {
        myRuntimeInfo->warn << "Unavailable fidelity level for 'VnDiagram' selected! Exiting"<< std::endl;
        exit(1);
    }
}  // initialize()


void performance_assessment_factory::run() {
/* Run the selected modules */
if (mySettingsPt->mySwitches.doFuelPlanning == true) {
    try {
        myRuntimeInfo->out << "  -> Fuel Planning" << std::endl;
        abstractFuelPlanningPt->doFuelPlanning();
        mySettingsPt->check_sucessfully_executed.doFuelPlanning = true;
    } catch (const int& errorCode) {
        mySettingsPt->check_sucessfully_executed.doFuelPlanning = false;
        myRuntimeInfo->out << "Error in fuel planning. See logLevel=2 (Info) for more details." << std::endl;
    }
}
/* Optimize Mission */
if (mySettingsPt->optimizeMission == true) {
    try {
        abstractOptimizeMissionProfilePt->doOptimizeMission(abstractFuelPlanningPt, rtIO, mySettingsPt->designForMTOM);
    } catch (const int& errorCode) {
        myRuntimeInfo->err << "Error in mission optimization. See logLevel=2 (Info) for more details." << std::endl;
    }
}

/* PayloadRange */
if (mySettingsPt->mySwitches.doPayloadRange == true) {
    try {
        if (mySettingsPt->mySwitches.doFuelPlanning == false || mySettingsPt->check_sucessfully_executed.doFuelPlanning == false) {
            throwError(__FILE__, __func__, __LINE__,
                "Error in payload range diagram calculation: Fuel planning needed for payload range calculation. Please turn on fuel planning.");
        }
        if (!mySettingsPt->designForMTOM && mySettingsPt->useStudyMissionForAnalysis) {
            myRuntimeInfo->out << "Attention: Payload range diagram incorrect when MTOM Design off or StudyMission is used!" << std::endl;
            abstractFuelPlanningPt->doPayloadRangeOLD();
        } else {
                myRuntimeInfo->out << "      ... Payload Range using OLD method" << std::endl;
                abstractFuelPlanningPt->doPayloadRangeOLD();
                abstractFuelPlanningPt->doFuelPayloadRangeOLD();
        }
        mySettingsPt->check_sucessfully_executed.doPayloadRange = true;
    } catch (const int& errorCode) {
        myRuntimeInfo->err << "Error in payload range calculation. See logLevel=2 (Info) for more details." << std::endl;
        mySettingsPt->check_sucessfully_executed.doPayloadRange = false;
    }
} else {
    abstractFuelPlanningPt->payloadRangeFromAcftXML();
}
if (mySettingsPt->mySwitches.doPayloadRange == true && mySettingsPt->mySwitches.doFuelPlanning == false) {
    myRuntimeInfo->out << "      ... Payload Range" << std::endl;  // TODO(hansmann): Cannot be reached (?)
    abstractPayloadRangePt->calcPayloadRange();
}

/* EnginePerformance */
if (mySettingsPt->mySwitches.doEnginePerformance == true) {
    try {
        myRuntimeInfo->out << "  -> Engine Performance" << std::endl;
        abstractEnginePerformancePt->doEngineCalculation(mySettingsPt->enginePerformance.Rating);
        mySettingsPt->check_sucessfully_executed.doEnginePerformance = true;
    } catch (const int& errorCode) {
        mySettingsPt->check_sucessfully_executed.doEnginePerformance = false;
        myRuntimeInfo->out << "Error in engine performance evaluation. See logLevel=2 (Info) for more details." << std::endl;
    }
}

/* Climb Performance */
if (mySettingsPt->mySwitches.doClimbPerformance == true) {
    try {
        myRuntimeInfo->out << "  -> Ceiling Performance" << std::endl;
        abstractCeilingPerformancePt->doCeilingCalculation();
        if (mySettingsPt->mySwitches.doFlightEnvelopePerformance == true) {
            try {
            myRuntimeInfo->out << "  -> Flight Envelopes" << std::endl;
            if (fabs(myAcftPt->MDive - 0) < ACCURACY_HIGH || fabs(myAcftPt->VDive - 0.) < ACCURACY_HIGH || mySettingsPt->mySwitches.overwriteInitialValues) {
                abstractFlightEnvelopesPt->doDiveSpeedLimits();
            }
                abstractFlightEnvelopesPt->doFlightEnvelope(abstractCeilingPerformancePt);
                myRuntimeInfo->out << "  -> SAR Performance" << std::endl;
                abstractSARPerformancePt->doSARCalculation(abstractFlightEnvelopesPt);
                mySettingsPt->check_sucessfully_executed.doFlightEnvelopePerformance = true;
            } catch (const int& errorCode) {
                mySettingsPt->check_sucessfully_executed.doFlightEnvelopePerformance = false;
                myRuntimeInfo->err << "Error in flight envelope calculation. See logLevel=2 (Info) for more details." << std::endl;
            }
        }
        mySettingsPt->check_sucessfully_executed.doClimbPerformance = true;
    } catch (const int& errorCode) {
        mySettingsPt->check_sucessfully_executed.doClimbPerformance = false;  // Prevents generation of plots with nonexisting data
        myRuntimeInfo->out << "Error in climb performance calculation. See logLevel=2 (Info) for more details." << std::endl;
    }
}

/* Take-Off Performance*/
if (mySettingsPt->mySwitches.doTOPerformance == true) {
    try {
        myRuntimeInfo->out << "  -> Takeoff Performance" << std::endl;
        abstractTakeOffPerformancePt->doTakeOffPerformance(mySettingsPt->myField.runwaySlope,
                myAcftPt->frictionCoefficient,
                myAcftPt->brakingCoefficient,
                mySettingsPt->myField.headWind);
                mySettingsPt->check_sucessfully_executed.doTOPerformance = true;
        } catch (const int& errorCode) {
            mySettingsPt->check_sucessfully_executed.doTOPerformance = false;  // Prevents generation of plots with nonexisting data
            myRuntimeInfo->out << "Error in takeoff performance calculation. See logLevel=2 (Info) for more details." << std::endl;
        }
    if (mySettingsPt->mySwitches.doBLFLPerformance == true) {  // BLFL performance cannot be calculated without TOperformance
        try {
            // Calculation of Balance Field Length
            myRuntimeInfo->out << "  -> Calculation of BLFL" << std::endl;
            abstractTakeOffPerformancePt->calcBLFL(mySettingsPt->myField.runwaySlope,
                                            myAcftPt->frictionCoefficient,
                                            myAcftPt->brakingCoefficient,
                                            mySettingsPt->myField.headWind);
            mySettingsPt->check_sucessfully_executed.doBLFLPerformance = true;
        } catch (const int& errorCode) {
            mySettingsPt->check_sucessfully_executed.doBLFLPerformance = false;
            myRuntimeInfo->out << "Error in balanced field length calculation. See logLevel=2 (Info) for more details." << std::endl;
        }
    }
}

/* Landing Performance */
if (mySettingsPt->mySwitches.doLandingPerformance == true) {
    try {
        myRuntimeInfo->out << "  -> Landing Performance" << std::endl;
        abstractLandingPerformancePt->doLandingPerformance(mySettingsPt->myField.runwaySlope,
        myAcftPt->brakingCoefficient,
        mySettingsPt->myField.headWind);
        mySettingsPt->check_sucessfully_executed.doLandingPerformance = true;
    } catch (const int& errorCode) {
        mySettingsPt->check_sucessfully_executed.doLandingPerformance = false;  // Prevents generation of plots with nonexisting data
        myRuntimeInfo->out << "Error in landing performance calculation. See logLevel=2 (Info) for more details." << std::endl;
    }
}
/* Vn-Diagram */
if (mySettingsPt->mySwitches.doVnDiagram == true) {
    try {
        myRuntimeInfo->out << "  -> V-n Diagram" << std::endl;
        abstractVnDiagramPt->doVnDiagram();
        mySettingsPt->check_sucessfully_executed.doVnDiagram = true;
    } catch (const int& errorCode) {
        mySettingsPt->check_sucessfully_executed.doVnDiagram = false;  // Prevents generation of plots with nonexisting data
        myRuntimeInfo->out << "Error in Vn diagram calculation. See logLevel=2 (Info) for more details." << std::endl;
    }
}
}  // end of run()

void performance_assessment_factory::update() {
  // update
}

void performance_assessment_factory::report() {
    performance_assessment_output theOutput(
        abstractFuelPlanningPt,
        abstractOptimizeMissionProfilePt,
        abstractPayloadRangePt,
        abstractEnginePerformancePt,
        abstractCeilingPerformancePt,
        abstractFlightEnvelopesPt,
        abstractSARPerformancePt,
        abstractTakeOffPerformancePt,
        abstractLandingPerformancePt,
        abstractVnDiagramPt,
        myAcftPt);

    theOutput.generateHtmlBody();
    theOutput.writeXmlOutput();
    theOutput.generatePlotData();
    theOutput.generatePlots();
}

void performance_assessment_factory::save() {
  // save data if necessary
}

}  // namespace taw
