/*  Copyright (C) 2012-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "performance_assessment_factory.h"
#include "performance_assessment_output.h"

performance_assessment_output::performance_assessment_output(
        std::shared_ptr<abstractFuelPlanning> abstractFuelPlanningPt,
        std::shared_ptr<abstractOptimizeMissionProfile> abstractOptimizeMissionProfilePt,
        std::shared_ptr<abstractPayloadRange> abstractPayloadRangePt,
        std::shared_ptr<abstractEnginePerformance> abstractEnginePerformancePt,
        std::shared_ptr<abstractCeilingPerformance> abstractCeilingPerformancePt,
        std::shared_ptr<abstractFlightEnvelopes> abstractFlightEnvelopesPt,
        std::shared_ptr<abstractSARPerformance> abstractSARPerformancePt,
        std::shared_ptr<abstractTakeOffPerformance> abstractTakeOffPerformancePt,
        std::shared_ptr<abstractLandingPerformance> abstractLandingPerformancePt,
        std::shared_ptr<abstractVnDiagram> abstractVnDiagramPt,
        const std::shared_ptr<aircraft> anAcftPt)
    :
    myAcftPt(anAcftPt),
    mySettingsPt(myAcftPt->mySettingsPt),
    rtIO(myAcftPt->rtIO),
    abstractFuelPlanningPt(abstractFuelPlanningPt),
    abstractOptimizeMissionProfilePt(abstractOptimizeMissionProfilePt),
    abstractPayloadRangePt(abstractPayloadRangePt),
    abstractEnginePerformancePt(abstractEnginePerformancePt),
    abstractCeilingPerformancePt(abstractCeilingPerformancePt),
    abstractFlightEnvelopesPt(abstractFlightEnvelopesPt),
    abstractSARPerformancePt(abstractSARPerformancePt),
    abstractTakeOffPerformancePt(abstractTakeOffPerformancePt),
    abstractLandingPerformancePt(abstractLandingPerformancePt),
    abstractVnDiagramPt(abstractVnDiagramPt),
    theReport(myAcftPt->rtIO) {
    // ctor
    if (mySettingsPt->mySwitches.doFuelPlanning && mySettingsPt->useStudyMissionForAnalysis) {
        theProgramName = rtIO->programname + "_study";
    } else {
        theProgramName = rtIO->programname;
    }
}

performance_assessment_output::~performance_assessment_output() {
    // dtor
}

void performance_assessment_output::generateHtmlBody() {
    theReport.htmlReportStream() << "<h1>Report " << myAcftPt->acftName << "</h1>" << std::endl;
    if (mySettingsPt->mySwitches.doFuelPlanning == true) {
        theReport.htmlReportStream() << "<h2>Fuel Planning</h2>" << std::endl
        << "<table class=\"ceras\">" << std::endl
        << "<tr>" << std::endl
        << "<td>Taxi Fuel: </td><td>" << abstractFuelPlanningPt->myFuel.taxiFuelTakeoff + abstractFuelPlanningPt->myFuel.taxiFuelLanding << " kg</td>" << std::endl
        << "</tr>" << std::endl
        << "<tr>" << std::endl
        << "<td>Mission Fuel: </td><td>" << abstractFuelPlanningPt->myFuel.missionFuel << " kg</td>" << std::endl
        << "</tr>" << std::endl
        << "<tr>" << std::endl
        << "<td>Trip Fuel: </td><td>" << abstractFuelPlanningPt->myFuel.tripFuel << " kg</td>" << std::endl
        << "</tr>" << std::endl
        << "<tr>" << std::endl
        << "<td>-> TOM: </td><td>" << Rounding(myAcftPt->OME
                + myAcftPt->myMission.missionPayload
                + abstractFuelPlanningPt->myFuel.missionFuel
                - abstractFuelPlanningPt->myFuel.taxiFuelTakeoff, 1) << " kg</td>" << std::endl
        << "</tr>" << std::endl
        << "<tr>" << std::endl
        << "<td>-> MTOM: </td><td>" << myAcftPt->MTOM << " kg</td>" << std::endl
        << "</tr>" << std::endl
        << "</table>" << std::endl
        << "<img class=\"ceras\" src=\"" + relativePath(rtIO->getPlotDir(), rtIO->getReportDir())
                + theProgramName + "_Mission_plot.svg\">" << std::endl;
    }
    theReport.htmlReportStream() << "<h2>Payload Range</h2>" << std::endl;
    if (mySettingsPt->mySwitches.doPayloadRange == true) {
        if (mySettingsPt->mySwitches.doFuelPlanning == true) {
            theReport.htmlReportStream() << "<img src=\"" + relativePath(rtIO->getPlotDir(), rtIO->getReportDir()) +
                   theProgramName + "_payloadRange.svg\">" << std::endl;
        } else {
            if (abstractPayloadRangePt->Range.at(2) > myAcftPt->designRange) {
                theReport.htmlReportStream() << "<img src=\"" + relativePath(rtIO->getPlotDir(), rtIO->getReportDir()) +
                       theProgramName + "_payloadRange.svg\">" << std::endl;
            } else {
                theReport.htmlReportStream() << "<table>" << std::endl
                << "<tr>" << std::endl
                << "<td>... not calculated due to error (Range_maxFuelatMTOM smaller DesignRange)!</td>" << std::endl
                << "</tr>" << std::endl
                << "</table>" << std::endl;
            }
        }
    } else {
        theReport.htmlReportStream() << "<table>" << std::endl
        << "<tr>" << std::endl
        << "<td>... not calculated!</td>" << std::endl
        << "</tr>" << std::endl
        << "</table>" << std::endl;
    }
    theReport.htmlReportStream() << "<h2>Flight Envelope</h2>" << std::endl;
    if (mySettingsPt->mySwitches.doFlightEnvelopePerformance == true) {
        theReport.htmlReportStream() << "<img src=\"" + relativePath(rtIO->getPlotDir(), rtIO->getReportDir()) +
               theProgramName + "_FlightEnvelope.svg\">" << std::endl;
    } else {
        theReport.htmlReportStream() << "<table>" << std::endl
        << "<tr>" << std::endl
        << "<td>... not calculated!</td>" << std::endl
        << "</tr>" << std::endl
        << "</table>" << std::endl;
    }
    theReport.htmlReportStream() << "<h2>Ceiling</h2>" << std::endl;
    if (mySettingsPt->mySwitches.doFlightEnvelopePerformance == true) {
        theReport.htmlReportStream() << "<img src=\"" + relativePath(rtIO->getPlotDir(), rtIO->getReportDir()) +
               theProgramName + "_Ceiling.svg\">" << std::endl;
    } else {
        theReport.htmlReportStream() << "<table>" << std::endl
        << "<tr>" << std::endl
        << "<td>... not calculated!</td>" << std::endl
        << "</tr>" << std::endl
        << "</table>" << std::endl;
    }
    theReport.htmlReportStream() << "<h2>Specific Air Range (SAR)</h2>" << std::endl;
    if (mySettingsPt->mySwitches.doFlightEnvelopePerformance == true) {
        theReport.htmlReportStream() << "<img src=\"" + relativePath(rtIO->getPlotDir(), rtIO->getReportDir()) +
               theProgramName + "_SAR.svg\">" << std::endl
        << "<div id=\"wrapper\">" << std::endl
        << "<div id=\"leftCol\">" << std::endl
        << "<h2>Best Specific Air Range (SAR)</h2>" << std::endl
        << "<img class=\"ceras\" src=\"" + relativePath(rtIO->getPlotDir(), rtIO->getReportDir()) +
               theProgramName + "_bestSAR.svg\">" << std::endl
        << "</div>" << std::endl
        << "</div>" << std::endl;
    } else {
        theReport.htmlReportStream() << "<table>" << std::endl
        << "<tr>" << std::endl
        << "<td>... not calculated!</td>" << std::endl
        << "</tr>" << std::endl
        << "</table>" << std::endl;
    }
    theReport.htmlReportStream() << "<h2>Takeoff Performance</h2>" << std::endl;
    if (mySettingsPt->mySwitches.doTOPerformance == true) {
        theReport.htmlReportStream() << "<img src=\"" + relativePath(rtIO->getPlotDir(), rtIO->getReportDir()) +
               theProgramName + "_TOFL.svg\">" << std::endl;
    } else {
        theReport.htmlReportStream() << "<table>" << std::endl
        << "<tr>" << std::endl
        << "<td>... not calculated!</td>" << std::endl
        << "</tr>" << std::endl
        << "</table>" << std::endl;
    }
    theReport.htmlReportStream() << "<h2>BFL Performance</h2>" << std::endl;
    if (mySettingsPt->mySwitches.doBLFLPerformance == true) {
        theReport.htmlReportStream() << "<img src=\"" + relativePath(rtIO->getPlotDir(), rtIO->getReportDir()) +
               theProgramName + "_BLFL.svg\">" << std::endl;
    } else {
        theReport.htmlReportStream() << "<table>" << std::endl
        << "<tr>" << std::endl
        << "<td>... not calculated!</td>" << std::endl
        << "</tr>" << std::endl
        << "</table>" << std::endl;
    }
    theReport.htmlReportStream() << "<h2>Landing Performance</h2>" << std::endl;
    if (mySettingsPt->mySwitches.doLandingPerformance == true) {
        theReport.htmlReportStream() << "<img src=\"" + relativePath(rtIO->getPlotDir(), rtIO->getReportDir()) +
               theProgramName + "_LDN.svg\">" << std::endl;
    } else {
        theReport.htmlReportStream() << "<table>" << std::endl
        << "<tr>" << std::endl
        << "<td>... not calculated!</td>" << std::endl
        << "</tr>" << std::endl
        << "</table>" << std::endl;
    }
    theReport.htmlReportStream() << "<h2>Engine Performance</h2>" << std::endl;
    if (mySettingsPt->mySwitches.doEnginePerformance == true) {
        theReport.htmlReportStream() << "<img src=\"" + relativePath(rtIO->getPlotDir(), rtIO->getReportDir()) +
               theProgramName + "_EnginePerformance.svg\">" << std::endl;
    } else {
        theReport.htmlReportStream() << "<table>" << std::endl
        << "<tr>" << std::endl
        << "<td>... not calculated!</td>" << std::endl
        << "</tr>" << std::endl
        << "</table>" << std::endl;
    }
    theReport.htmlReportStream() << "<h2>V-n-Diagram</h2>" << std::endl;
    if (mySettingsPt->mySwitches.doVnDiagram == true) {
        theReport.htmlReportStream() << "<img src=\"" + relativePath(rtIO->getPlotDir(), rtIO->getReportDir()) +
               theProgramName + "_VnDiagram.svg\">" << std::endl;
    } else {
        theReport.htmlReportStream() << "<table>" << std::endl
        << "<tr>" << std::endl
        << "<td>... not calculated!</td>" << std::endl
        << "</tr>" << std::endl
        << "</table>" << std::endl;
    }
}

void performance_assessment_output::writeXmlOutput() {
    string subPath;
    /* MTOM */
    if (mySettingsPt->designForMTOM == true && mySettingsPt->mySwitches.doFuelPlanning == true) {
        rtIO->acxml.at("analysis/masses_cg_inertia/maximum_takeoff_mass/mass_properties/mass/value") = myAcftPt->MTOM;
    }
    if (mySettingsPt->mySwitches.doFlightEnvelopePerformance == true) {
        /* Speeds */
        subPath = "assessment/performance/speed";
        Endnode<double> MMO(subPath + "/maximum_operating_mach_number", "Maximum operating mach number");
        MMO.set_value(myAcftPt->MMO);
        MMO.set_boundaries(0., 1.);
        MMO.set_unit("1");
        MMO.update(rtIO->acxml);
        Endnode<double> VMO(subPath + "/maximum_operating_velocity", "Maximum operating velocity");
        VMO.set_value(myAcftPt->VMO);
        VMO.set_boundaries(0., 400.);
        VMO.set_unit("m/s");
        VMO.update(rtIO->acxml);
        Endnode<double> MDive(subPath + "/dive_mach_number", "Maximum dive mach number");
        MDive.set_value(myAcftPt->MDive);
        MDive.set_boundaries(0., 1.);
        MDive.set_unit("1");
        MDive.update(rtIO->acxml);
        Endnode<double> VDive(subPath + "/dive_velocity", "Dive velocity");
        VDive.set_value(myAcftPt->VDive);
        VDive.set_boundaries(0., 400.);
        VDive.set_unit("m/s");
        VDive.update(rtIO->acxml);
        Endnode<double> v1gsClean(subPath + "/one_g_stall_speed_velocity", "One g stall speed");
        v1gsClean.set_value(myAcftPt->v1gS_Clean(myAcftPt->MTOM, 0.0));
        v1gsClean.set_boundaries(0., 400.);
        v1gsClean.set_unit("m/s");
        v1gsClean.update(rtIO->acxml);
    }
    /* Start */
    if (mySettingsPt->mySwitches.doTOPerformance == true && mySettingsPt->check_sucessfully_executed.doTOPerformance == true) {
        subPath = "assessment/performance/takeoff";
        // Endnode<double> todns(subPath + "/takeoff_distance_normal_safety", "Takeoff distance normal safety");
        // todns.set_value(abstractTakeOffPerformancePt->myTOFLField.at(0).myTOFL.back().TODR_AEO);
        // todns.set_lower_boundary(0);
        // todns.set_upper_boundary(20000);
        // todns.set_unit("m");
        // todns.update(rtIO->acxml);
        Endnode<double> vlof(subPath + "/lift_off_speed_velocity", "Lift off velocity");
        vlof.set_value(myAcftPt->vLOF(myAcftPt->MTOM, 0.0));
        vlof.set_boundaries(0, 400);
        vlof.set_unit("m/s");
        vlof.update(rtIO->acxml);
        Endnode<double> decision_speed(subPath + "/decision_speed", "Take off decision speed (V1)");
        decision_speed.set_value(abstractTakeOffPerformancePt->myBLFL.v1);
        decision_speed.set_lower_boundary(0);
        decision_speed.set_upper_boundary(200);
        decision_speed.set_unit("m/s");
        decision_speed.update(rtIO->acxml);
        Endnode<double> v2(subPath + "/takeoff_safety_speed", "Take off safety speed (V2)");
        v2.set_value(myAcftPt->v2(myAcftPt->MTOM, 0.0));
        v2.set_boundaries(0, 200);
        v2.set_unit("m/s");
        v2.update(rtIO->acxml);
        Endnode<double> vfto(subPath + "/final_takeoff_speed", "Final take off speed");
        vfto.set_value(myAcftPt->vFTO(myAcftPt->MTOM, 0.0));
        vfto.set_boundaries(0, 200);
        vfto.set_unit("m/s");
        vfto.update(rtIO->acxml);
        Endnode<double> ttsh(subPath + "/time_to_screen_height", "Time to screen height");
        ttsh.set_value(abstractTakeOffPerformancePt->myTOFLField.at(0).myTOFL.back().Time);
        ttsh.set_boundaries(0, 3600);
        ttsh.set_unit("s");
        ttsh.update(rtIO->acxml);
        Endnode<double> climbDescendecondSegment(subPath + "/climb_or_descend_segment_climb_gradient", "Climb or descend in second segment");
        climbDescendecondSegment.set_value(abstractTakeOffPerformancePt->scdSegmentGradient);
        climbDescendecondSegment.set_boundaries(0, 100);
        climbDescendecondSegment.set_unit("m/s");
        climbDescendecondSegment.update(rtIO->acxml);
        Endnode<double> finalSegmentGradient(subPath + "/final_segment_climb_gradient", "Climb gradient in final segment");
        finalSegmentGradient.set_value(abstractTakeOffPerformancePt->finalSegmentGradient);
        finalSegmentGradient.set_boundaries(0, 100);
        finalSegmentGradient.set_unit("m/s");
        finalSegmentGradient.update(rtIO->acxml);
    }
    /* BLFL */
    if (mySettingsPt->mySwitches.doBLFLPerformance == true && mySettingsPt->check_sucessfully_executed.doBLFLPerformance == true) {
        subPath = "assessment/performance/takeoff";
        Endnode<double> BLFL(subPath + "/balanced_field_length", "Balanced field length");
        BLFL.set_value(abstractTakeOffPerformancePt->myBLFL.BLFL);
        BLFL.set_boundaries(0, 100000);
        BLFL.set_unit("m");
        BLFL.update(rtIO->acxml);
    }
    /* Landing */
    if (mySettingsPt->mySwitches.doLandingPerformance == true && mySettingsPt->check_sucessfully_executed.doLandingPerformance == true) {
        subPath = "assessment/performance/landing";
        Endnode<double> needed_runway_length(subPath + "/needed_runway_length", "Needed runway length");
        needed_runway_length.set_value(abstractLandingPerformancePt->myLDNField.at(0).myLDN.back().ALD_AEO);
        needed_runway_length.set_boundaries(0, 100000);
        needed_runway_length.set_unit("m");
        needed_runway_length.update(rtIO->acxml);
        Endnode<double> vappr(subPath + "/approach_speed", "Approach speed");
        vappr.set_value(myAcftPt->vAPPR(myAcftPt->MLM, 0.0));
        vappr.set_boundaries(0, 1000);
        vappr.set_unit("m/s");
        vappr.update(rtIO->acxml);
    }
    if (mySettingsPt->mySwitches.doPayloadRange == true && mySettingsPt->check_sucessfully_executed.doPayloadRange == true) {
        /* Ranges */
        subPath = "assessment/performance/range";
        Endnode<double> range_max_payload_at_maximum_takeoff_mass(subPath + "/range_max_payload_at_maximum_takeoff_mass", "Range at max payload at maximum take off mass");
        Endnode<double> range_max_fuel_at_maximum_takeoff_mass(subPath + "/range_max_fuel_at_maximum_takeoff_mass", "Range at max fuel at maximum take off mass");
        Endnode<double> payload_maximum_fuel_at_maximum_takeoff_mass(subPath + "/payload_maximum_fuel_at_maximum_takeoff_mass", "Payload at maximum fuel at maximum takeoff mass");
        Endnode<double> range_maximum_fuel_empty(subPath + "/range_maximum_fuel_empty", "Range at maximum fuel empty");
        if (mySettingsPt->mySwitches.doFuelPlanning == true) {
            range_max_payload_at_maximum_takeoff_mass.set_value(abstractFuelPlanningPt->myPayloadRange.at(1).Range);
            range_max_fuel_at_maximum_takeoff_mass.set_value(abstractFuelPlanningPt->myPayloadRange.at(2).Range);
            payload_maximum_fuel_at_maximum_takeoff_mass.set_value(abstractFuelPlanningPt->myPayloadRange.at(2).Payload);
            range_maximum_fuel_empty.set_value(abstractFuelPlanningPt->myPayloadRange.at(3).Range);
        } else {
            if (abstractPayloadRangePt->Range.at(2) > myAcftPt->designRange) {
                range_max_payload_at_maximum_takeoff_mass.set_value(abstractPayloadRangePt->Range.at(1));
                range_max_fuel_at_maximum_takeoff_mass.set_value(abstractPayloadRangePt->Range.at(2));
                payload_maximum_fuel_at_maximum_takeoff_mass.set_value(abstractPayloadRangePt->Payload.at(2));
                range_maximum_fuel_empty.set_value(abstractPayloadRangePt->Range.at(3));
            }
        }
        range_max_payload_at_maximum_takeoff_mass.set_boundaries(0, INFINITY);
        range_max_payload_at_maximum_takeoff_mass.set_unit("m");
        range_max_payload_at_maximum_takeoff_mass.update(rtIO->acxml);
        range_max_fuel_at_maximum_takeoff_mass.set_boundaries(0, INFINITY);
        range_max_fuel_at_maximum_takeoff_mass.set_unit("m");
        range_max_fuel_at_maximum_takeoff_mass.update(rtIO->acxml);
        payload_maximum_fuel_at_maximum_takeoff_mass.set_boundaries(0, INFINITY);
        payload_maximum_fuel_at_maximum_takeoff_mass.set_unit("kg");
        payload_maximum_fuel_at_maximum_takeoff_mass.update(rtIO->acxml);
        range_maximum_fuel_empty.set_boundaries(0, INFINITY);
        range_maximum_fuel_empty.set_unit("m");
        range_maximum_fuel_empty.update(rtIO->acxml);
    }
    if (mySettingsPt->mySwitches.doFuelPlanning == true && mySettingsPt->check_sucessfully_executed.doFuelPlanning == true) {
        /* Mission Fuels */
        if (!mySettingsPt->designForMTOM && mySettingsPt->useStudyMissionForAnalysis) {
            subPath = "analysis/mission/study_mission";
        } else {
            subPath = "analysis/mission/design_mission";
        }
        Endnode<double> missionRange(subPath + "/range", "Traveled range from break release to end of taxi at destination");
        missionRange.set_value(myAcftPt->myMission.missionRange);
        missionRange.update(rtIO->acxml);
        // rtIO->acxml.at(subPath + "/range") = myAcftPt->myMission.missionRange;
        // Fuel
        // rtIO->acxml.at(subPath + "/mission_energy") = std::to_string(abstractFuelPlanningPt->myFuel.missionFuel) + "<-Fuel[kg]";  // Total Tanked Fuel
        // rtIO->acxml.at(subPath + "/trip_energy") = abstractFuelPlanningPt->myFuel.tripFuel;  // Tanked fuel - Reserve Fuel - Taxi Fuel
        // rtIO->acxml.at(subPath + "/taxi_energy") = abstractFuelPlanningPt->myFuel.taxiFuelTakeoff;  // TODO(hansmann): xpath&fuel2energyconversion
        // rtIO->acxml.at(subPath + "/TaxifuelLdg") = abstractFuelPlanningPt->myFuel.taxiFuelLanding;  // TODO(hansmann): xpath&fuel2energyconversion
        // Rest
        Endnode<double> tom(subPath + "/takeoff_mass", "Take off mass");
        tom.set_value(myAcftPt->OME + myAcftPt->myMission.missionPayload + abstractFuelPlanningPt->myFuel.missionFuel - abstractFuelPlanningPt->myFuel.taxiFuelTakeoff);
        tom.set_unit("kg");
        tom.set_boundaries(0, INFINITY);
        tom.update(rtIO->acxml);
        // rtIO->acxml.at(subPath + "/TOM") = myAcftPt->OME
        //                                        + myAcftPt->myMission.missionPayload
        //                                        + abstractFuelPlanningPt->myFuel.missionFuel
        //                                        - abstractFuelPlanningPt->myFuel.taxiFuelTakeoff;
        Endnode<double> payload_mass(subPath + "/payload_mass", "Mission payload mass");
        payload_mass.set_value(myAcftPt->myMission.missionPayload);
        payload_mass.set_boundaries(0, INFINITY);
        payload_mass.set_unit("kg");
        payload_mass.update(rtIO->acxml);
        Endnode<int> missionPax(subPath + "/number_of_pax", "Number of passengers considered for the mission");
        missionPax.set_value(myAcftPt->myMission.missionPAX);
        missionPax.set_boundaries(0, std::numeric_limits<int>::max());
        missionPax.set_unit("1");
        missionPax.update(rtIO->acxml);
        Endnode<double> cargo_mass(subPath + "/cargo_mass", "Cargo mass considered for the mission");
        cargo_mass.set_value(myAcftPt->myMission.missionCargo);
        cargo_mass.set_boundaries(0, INFINITY);
        cargo_mass.set_unit("kg");
        cargo_mass.update(rtIO->acxml);
        Endnode<double> takeoff_engine_derate(subPath + "/takeoff_engine_derate", "Engine power demand");
        takeoff_engine_derate.set_value(myAcftPt->myMission.Steps.at(0).derating);
        takeoff_engine_derate.set_unit("1");
        takeoff_engine_derate.update(rtIO->acxml);
        Endnode<double> block_time(subPath + "/block_time", "Block time for the whole mission: Time from break release to end of taxiing after landing");
        block_time.set_value(abstractFuelPlanningPt->blockTime);
        block_time.set_unit("s");
        block_time.update(rtIO->acxml);
        Endnode<double> flight_time(subPath + "/flight_time", "Flight time for the whole mission");
        flight_time.set_value(abstractFuelPlanningPt->tripTime);
        flight_time.set_unit("s");
        flight_time.update(rtIO->acxml);
    }
    /* Requirements Checks */
    // TLARS
    subPath = "aircraft_exchange_file/requirement_compliance/top_level_aircraft_requirements";
    Endnode<bool> design_takeoff_distance_checked(subPath + "/design_takeoff_distance/checked", "Indicates if the value has been checked against the requirement.");
    design_takeoff_distance_checked.set_value(false);
    if (mySettingsPt->mySwitches.doTOPerformance == true && mySettingsPt->check_sucessfully_executed.doTOPerformance == true) {
        Endnode<bool> design_takeoff_distance_maintainable(subPath + "/design_takeoff_distance/maintainable",
            "Switch indicating if design takeoff distance can be maintained.");
        design_takeoff_distance_maintainable.set_value(myAcftPt->myChecks.TOFL);
        design_takeoff_distance_maintainable.update(rtIO->acxml);
        design_takeoff_distance_checked.set_value(true);
    }
    design_takeoff_distance_checked.update(rtIO->acxml);

    Endnode<bool> design_landing_field_length_checked(subPath + "/design_landing_field_length/checked", "Indicates if the value has been checked against the requirement.");
    design_landing_field_length_checked.set_value(false);
    Endnode<bool> design_approach_speed_maintainable_checked(subPath + "/design_approach_speed/checked", "Indicates if the value has been checked against the requirement.");
    design_approach_speed_maintainable_checked.set_value(false);
    if (mySettingsPt->mySwitches.doLandingPerformance == true && mySettingsPt->check_sucessfully_executed.doLandingPerformance == true) {
        Endnode<bool> design_landing_field_length_maintainable(subPath + "/design_landing_field_length/maintainable",
            "Switch indicating if landing field length can be maintained.");
        design_landing_field_length_maintainable.set_value(myAcftPt->myChecks.LDN);
        design_landing_field_length_maintainable.update(rtIO->acxml);
        design_landing_field_length_checked.set_value(true);
        Endnode<bool> design_approach_speed_maintainable(subPath + "/design_approach_speed/maintainable", "Switch indicating if approach speed can be maintained.");
        design_approach_speed_maintainable.set_value(myAcftPt->myChecks.vApproach);
        design_approach_speed_maintainable.update(rtIO->acxml);
        design_approach_speed_maintainable_checked.set_value(true);
    }
    design_landing_field_length_checked.update(rtIO->acxml);
    design_approach_speed_maintainable_checked.update(rtIO->acxml);

    // Span limit
    Endnode<bool> span_limit(subPath + "/span_limit/maintained", "Switch indicating if the span limit according to ICAO aerodrome ref code is maintained.");
    span_limit.set_value(myAcftPt->myChecks.span_limit);
    span_limit.update(rtIO->acxml);
    Endnode<bool> span_limit_checked(subPath + "/span_limit/checked", "Indicates if the value has been checked against the requirement.");
    span_limit_checked.set_value(true);

    // Certification
    subPath = "aircraft_exchange_file/requirement_compliance/certification";
    Endnode<bool> climb_gradient_of_second_takeoff_segment_checked(subPath + "/climb_gradient_of_second_takeoff_segment/checked",
        "Indicates if the value has been checked against the requirement.");
    climb_gradient_of_second_takeoff_segment_checked.set_value(false);
    Endnode<bool> climb_gradient_of_final_takeoff_segment_checked(subPath + "/climb_gradient_of_final_takeoff_segment/checked",
        "Indicates if the value has been checked against the requirement.");
    climb_gradient_of_final_takeoff_segment_checked.set_value(false);

    if (mySettingsPt->mySwitches.doTOPerformance == true && mySettingsPt->check_sucessfully_executed.doTOPerformance == true) {
        Endnode<bool> climb_gradient_of_second_takeoff_segment(subPath + "/climb_gradient_of_second_takeoff_segment/maintainable",
            "Switch indicating if climb gradient of second takeoff segment can be maintained.");
        climb_gradient_of_second_takeoff_segment.set_value(myAcftPt->myChecks.climbGradientSecondSegment);
        climb_gradient_of_second_takeoff_segment.update(rtIO->acxml);
        climb_gradient_of_second_takeoff_segment_checked.set_value(true);
        Endnode<bool> climb_gradient_of_final_takeoff_segment(subPath + "/climb_gradient_of_final_takeoff_segment/maintainable",
            "Switch indicating if climb gradient of final takeoff segment can be maintained.");
        climb_gradient_of_final_takeoff_segment.set_value(myAcftPt->myChecks.climbGradientFinalSegment);
        climb_gradient_of_final_takeoff_segment.update(rtIO->acxml);
        climb_gradient_of_final_takeoff_segment_checked.set_value(true);
    }
    climb_gradient_of_second_takeoff_segment_checked.update(rtIO->acxml);
    climb_gradient_of_final_takeoff_segment_checked.update(rtIO->acxml);

    Endnode<bool> climb_gradient_approach_one_engine_inoperative_checked(subPath + "/climb_gradient_approach_one_engine_inoperative/checked",
        "Indicates if the value has been checked against the requirement.");
    climb_gradient_approach_one_engine_inoperative_checked.set_value(false);
    Endnode<bool> climb_gradient_all_engines_operative_checked(subPath + "/climb_gradient_all_engines_operative/checked",
        "Indicates if the value has been checked against the requirement.");
    climb_gradient_all_engines_operative_checked.set_value(false);
    if (mySettingsPt->mySwitches.doLandingPerformance == true && mySettingsPt->check_sucessfully_executed.doLandingPerformance) {
        Endnode<bool> climb_gradient_approach_one_engine_inoperative(subPath + "/climb_gradient_approach_one_engine_inoperative/maintainable",
            "Switch indicating if climb gradient approach one engine inoperative can be maintained.");
        climb_gradient_approach_one_engine_inoperative.set_value(myAcftPt->myChecks.climbGradientApproachOEI);
        climb_gradient_approach_one_engine_inoperative.update(rtIO->acxml);
        climb_gradient_approach_one_engine_inoperative_checked.set_value(true);
        Endnode<bool> climb_gradient_all_engines_operative(subPath + "/climb_gradient_all_engines_operative/maintainable",
            "Switch indicating if climb gradient with all engines operative can be maintained.");
        climb_gradient_all_engines_operative.set_value(myAcftPt->myChecks.climbGradientApproachAEO);
        climb_gradient_all_engines_operative.update(rtIO->acxml);
        climb_gradient_all_engines_operative_checked.set_value(true);
    }
    climb_gradient_approach_one_engine_inoperative_checked.update(rtIO->acxml);
    climb_gradient_all_engines_operative_checked.update(rtIO->acxml);
}


void performance_assessment_output::generatePlotData() {
    if (mySettingsPt->mySwitches.doFuelPlanning == true && mySettingsPt->check_sucessfully_executed.doFuelPlanning == true) {
        generatePlotDataMission();
    }
    if (mySettingsPt->mySwitches.doPayloadRange == true && mySettingsPt->check_sucessfully_executed.doPayloadRange == true) {
        generatePlotDataPayloadRange();
    }
    if (mySettingsPt->mySwitches.doEnginePerformance == true && mySettingsPt->check_sucessfully_executed.doEnginePerformance == true) {
        generateEnginePerformanceData();
    }
    if (mySettingsPt->mySwitches.doClimbPerformance == true && mySettingsPt->check_sucessfully_executed.doClimbPerformance == true) {
        generateCeilingPlotData();
    }
    if (mySettingsPt->mySwitches.doFlightEnvelopePerformance == true && mySettingsPt->check_sucessfully_executed.doFlightEnvelopePerformance == true) {
        generateFlightEnvelopeData();
        generateSARPlotData();
        generateBestSARAltPlotData();
    }
    if (mySettingsPt->mySwitches.doTOPerformance == true && mySettingsPt->check_sucessfully_executed.doTOPerformance == true) {
        generateTOFLPlotData();
    }
    if (mySettingsPt->mySwitches.doBLFLPerformance == true && mySettingsPt->check_sucessfully_executed.doBLFLPerformance == true) {
        generateBLFLPlotData();
    }
    if (mySettingsPt->mySwitches.doLandingPerformance == true && mySettingsPt->check_sucessfully_executed.doLandingPerformance == true) {
        generateLDNPlotData();
    }
    if (mySettingsPt->mySwitches.doVnDiagram == true && mySettingsPt->check_sucessfully_executed.doVnDiagram == true) {
        generateVnPlotData();
    }
}

void performance_assessment_output::generatePlotDataMission() {
std::ofstream csv;
    string csvDataFile(mySettingsPt->rtIO->getCsvFilesDirTool() + "/" + theProgramName + "_Mission.csv");
    csv.open(csvDataFile.c_str());
    if (!csv) {
        myRuntimeInfo->err << string(theProgramName) + "_Mission.csv could not be opened!" << std::endl;
        exit(1);
    } else {
        csv << "# " + mySettingsPt->rtIO->getCsvFilesDirTool() + theProgramName + "_Mission.csv" << std::endl;
        // Adjust data structure:
        csv << "# (1) ID; (2) Time [s]; (3) Range [m]; (4) Altitude [m]; (5) TAS [m/s]; (6) CAS [m/s]; (7) Mach; ";
        csv << "(8) A/C Mass [kg]; (9) consumed Fuel [kg]; (10) total consumed Fuel [kg]; (11) Thrust [N]; ";
        csv << "(12) Fuel Flow [kg/s]; (13) ROC [m/s]; (14) C_L" << std::endl;
        /*(1)*/ csv << "0; ";
        /*(1)*/ csv << "0; ";
        /*(3)*/ csv << "0; ";
        /*(4)*/ csv << "0; ";
        /*(5)*/ csv << "0; ";
        /*(6)*/ csv << "0; ";
        /*(7)*/ csv << "0; ";
        /*(8)*/ csv << abstractFuelPlanningPt->theMissionFuel.at(0).Mass << "; ";
        /*(9)*/ csv << "0; ";
        /*(10)*/ csv << "0; ";
        /*(11)*/ csv << "0; ";
        /*(12)*/ csv << "0; ";
        /*(13)*/ csv << "0; ";
        /*(14)*/ csv << "0" << std::endl;
        double time(0.);
        double range(0.);
        double fuel(0.);
        for (unsigned int i = 0; i <= myAcftPt->myMission.Steps.size() - 1; i++) {
            time += abstractFuelPlanningPt->theMissionFuel.at(i + 1).Time;
            range += abstractFuelPlanningPt->theMissionFuel.at(i + 1).Range;
            fuel += abstractFuelPlanningPt->theMissionFuel.at(i + 1).consumedFuel;
            double C_L = (2. * G_FORCE * abstractFuelPlanningPt->theMissionFuel.at(i + 1).Mass)
                  / (myAcftPt->atm.getDensity(myAcftPt->myMission.Steps.at(i).endAltitude)
                     *  pow(myAcftPt->myMission.Steps.at(i).endTAS, 2.)
                     * myAcftPt->S_Wing);
            /*(1)*/ csv << i + 1 << "; ";
            /*(1)*/ csv << time << "; ";
            /*(3)*/ csv << range << "; ";
            /*(4)*/ csv << myAcftPt->myMission.Steps.at(i).endAltitude << "; ";
            /*(5)*/ csv << myAcftPt->myMission.Steps.at(i).endTAS  << "; ";
            /*(6)*/ csv << myAcftPt->myMission.Steps.at(i).endCAS  << "; ";
            /*(7)*/ csv << myAcftPt->myMission.Steps.at(i).endMach  << "; ";
            /*(8)*/ csv << abstractFuelPlanningPt->theMissionFuel.at(i + 1).Mass << "; ";
            /*(9)*/ csv << abstractFuelPlanningPt->theMissionFuel.at(i + 1).consumedFuel << "; ";
            /*(10)*/ csv << fuel << "; ";
            /*(11)*/ csv << abstractFuelPlanningPt->theMissionFuel.at(i + 1).Thrust << "; ";
            /*(12)*/ csv << abstractFuelPlanningPt->theMissionFuel.at(i + 1).fuelFlow << "; ";
            /*(13)*/ csv << abstractFuelPlanningPt->theMissionFuel.at(i + 1).ROC << "; ";
            /*(14)*/ csv << C_L << std::endl;
        }
        csv.close();
    }
}

void performance_assessment_output::generatePlotDataPayloadRange() {
std::ofstream csv;
    string csvDataFile(mySettingsPt->rtIO->getCsvFilesDirTool() + "/" + theProgramName + "_payloadRange.csv");
    csv.open(csvDataFile.c_str());
    if (!csv) {
        myRuntimeInfo->err << string(theProgramName) + "_payloadRange.csv could not be opened!" << std::endl;
        exit(1);
    } else {
        csv << "# " + mySettingsPt->rtIO->getCsvFilesDirTool() + theProgramName + "_missionFuel.csv" << std::endl;
        // Adjust data structure:
        csv << "# (1) Range [m]; (2) Payload [kg]; (3) Fuel [kg]; (4) TOM [kg]; (5) OME [kg]" << std::endl;
        if (mySettingsPt->mySwitches.doFuelPlanning == true) {
            for (unsigned int i(0); i <= abstractFuelPlanningPt->myPayloadRange.size() - 1; i++) {
                csv << abstractFuelPlanningPt->myPayloadRange.at(i).Range << "; ";
                csv << abstractFuelPlanningPt->myPayloadRange.at(i).Payload << "; ";
                // csv << myPerfPt->abstractFuelPlanning.myPayloadRange.at(i).Fuel << "; ";
                // csv << myPerfPt->abstractFuelPlanning.myPayloadRange.at(i).TOM << "; ";
                // csv << myAcftPt->OME << "; ";
                csv << std::endl;
            }
        } else {
            for (unsigned int i(0); i <= abstractPayloadRangePt->Range.size() - 1; i++) {
                csv << abstractPayloadRangePt->Range.at(i) << "; ";
                csv << abstractPayloadRangePt->Payload.at(i) << "; ";
                csv << abstractPayloadRangePt->Fuel.at(i) << "; ";
                csv << abstractPayloadRangePt->TOM.at(i) << "; ";
                // csv << myPerfPt->abstractFuelPlanning.myPayloadRange.at(i).Fuel << "; ";
                // csv << myPerfPt->abstractFuelPlanning.myPayloadRange.at(i).TOM << "; ";
                // csv << myAcftPt->OME << "; ";
                csv << std::endl;
            }
        }
        csv << std::endl;
        csv << std::endl;
        csv << "# Design Point " << std::endl;
        csv << myAcftPt->designRange << "; ";
        csv << myAcftPt->designPayload << "; ";
        // csv << "N/A" << "; ";
        // csv << myAcftPt->MTOM << "; ";
        // csv << myAcftPt->OME << "; ";
        csv << std::endl;
        csv << std::endl;
        csv << std::endl;
        csv << "# Mission Point" << std::endl;
        csv << myAcftPt->myMission.missionRange << "; ";
        csv << myAcftPt->myMission.missionPayload << "; ";
        // csv << myPerfPt->abstractFuelPlanning.myFuel.tripFuel << "; ";
        // csv << myAcftPt->MTOM << "; ";
        // csv << myAcftPt->OME << "; ";
        csv.close();
    }
}

void performance_assessment_output::generateEnginePerformanceData() {
    for (unsigned int i(0); i <= abstractEnginePerformancePt->theRatingEngine.size() - 1; i++) {
        std::ofstream csv;
        string csvDataFile(mySettingsPt->rtIO->getCsvFilesDirTool() + "/" + theProgramName + "_EnginePerformance_" +
            abstractEnginePerformancePt->theRatingEngine.at(i).Rating + ".csv");
        csv.open(csvDataFile.c_str());
        if (!csv) {
            myRuntimeInfo->err << string(theProgramName) + "_EnginePerformance" + abstractEnginePerformancePt->theRatingEngine.at(i).Rating
                + ".csv could not be opened!" << std::endl;
            exit(1);
        } else {
            csv << "# " + mySettingsPt->rtIO->getCsvFilesDirTool() + theProgramName + "_EnginePerformance" +
                abstractEnginePerformancePt->theRatingEngine.at(i).Rating + ".csv" << std::endl;
            csv << "# Rating: " + abstractEnginePerformancePt->theRatingEngine.at(i).Rating << std::endl;
            /* Adjust data structure: */
            csv << "# (1) Mach; (2) Altitude [m]; (3) Thrust per Engine [N]; (4) SFC [g/Ns]" << std::endl;
            for (unsigned int j(0); j <= abstractEnginePerformancePt->theRatingEngine.at(i).theParameter.size() - 1; j++) {
                for (unsigned int n(0); n <= abstractEnginePerformancePt->theRatingEngine.at(i).theParameter.at(j).Mach.size() - 1; n++) {
                    csv << abstractEnginePerformancePt->theRatingEngine.at(i).theParameter.at(j).Mach.at(n) << "; ";
                    csv << abstractEnginePerformancePt->theRatingEngine.at(i).theParameter.at(j).Altitude << "; ";
                    csv << abstractEnginePerformancePt->theRatingEngine.at(i).theParameter.at(j).Thrust.at(n) << "; ";
                    csv << abstractEnginePerformancePt->theRatingEngine.at(i).theParameter.at(j).SFC.at(n);
                    csv << std::endl;
                }
                csv << std::endl;
                csv << std::endl;
            }
        }
        csv.close();
    }
}
void performance_assessment_output::generateCeilingPlotData() {
    std::ofstream csv;
    string csvDataFile(mySettingsPt->rtIO->getCsvFilesDirTool() + "/" + theProgramName + "_Ceiling.csv");
    csv.open(csvDataFile.c_str());
    if (!csv) {
        myRuntimeInfo->err << string(theProgramName) + "_Ceiling.csv could not be opened!" << std::endl;
        exit(1);
    } else {
        csv << "# " + mySettingsPt->rtIO->getCsvFilesDirTool() + theProgramName + "_Ceiling.csv" << std::endl;
        // Adjust data structure:
        csv << "# (1) Mach; (2) Gross Mass [kg]; (3) Service Ceiling [m]; (4) Max Ceiling [m] ; (5) OEI max Ceiling [m]; (6) Initial Cruise Altitude [m]" << std::endl;
        for (unsigned int i(0); i <= abstractCeilingPerformancePt->theCeiling.size() - 1; i++) {
            csv << abstractCeilingPerformancePt->theCeiling.at(i).Mach << "; ";
            csv << abstractCeilingPerformancePt->theCeiling.at(i).Mass << "; ";
            csv << abstractCeilingPerformancePt->theCeiling.at(i).serviceCeiling << "; ";
            csv << abstractCeilingPerformancePt->theCeiling.at(i).maxCeiling << "; ";
            csv << abstractCeilingPerformancePt->theCeiling.at(i).maxCeilingOEI << "; ";
            csv << myAcftPt->hInitialCruise << std::endl;
        }
        csv.close();
    }
}
void performance_assessment_output::generateFlightEnvelopeData() {
    std::ofstream csv;
    string csvDataFile(mySettingsPt->rtIO->getCsvFilesDirTool() + "/" + theProgramName + "_FlightEnvelope.csv");
    csv.open(csvDataFile.c_str());
    if (!csv) {
        myRuntimeInfo->err << string(theProgramName) + "_FlightEnvelope.csv could not be opened!" << std::endl;
        exit(1);
    } else {
        csv << "# " + string(theProgramName) + "_FlightEnvelope.csv" << std::endl;
        // Adjust data structure:
        csv << "# (1) Altitude [m]; (2) TAS [m/s]; (3) Gross Mass [kg]" <<std::endl;
        csv << "# Stall Limit" <<std::endl;
        for (unsigned int i(0); i <= abstractFlightEnvelopesPt->myFlightEnvelope.back().vStall.size() - 1; i++) {
            csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().vStall.at(i).height  << "; ";
            csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().vStall.at(i).speed << "; ";
            csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().grossMass << "; ";
            csv <<std::endl;
        }
        csv << std::endl;
        csv << std::endl;
        csv << "# VMO Limit" << std::endl;
        for (unsigned int i(0); i <= abstractFlightEnvelopesPt->myFlightEnvelope.back().vMO.size() - 1; i++) {
            csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().vMO.at(i).height << "; ";
            csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().vMO.at(i).speed << "; ";
            csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().grossMass << "; ";
            csv <<std::endl;
        }
        csv << std::endl;
        csv << std::endl;
        csv << "# MMO Limit" << std::endl;
        for (unsigned int i(0); i <= abstractFlightEnvelopesPt->myFlightEnvelope.back().MMO.size() - 1; i++) {
            csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().MMO.at(i).height << "; ";
            csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().MMO.at(i).speed << "; ";
            csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().grossMass << "; ";
            csv << std::endl;
        }
        csv <<std::endl;
        csv <<std::endl;
        csv << "# VDive Limit" <<std::endl;
        for (unsigned int i(0); i <= abstractFlightEnvelopesPt->myFlightEnvelope.back().vDive.size() - 1; i++) {
            csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().vDive.at(i).height  << "; ";
            csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().vDive.at(i).speed << "; ";
            csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().grossMass << "; ";
            csv <<std::endl;
        }
        csv <<std::endl;
        csv <<std::endl;
        csv << "# MDive Limit" <<std::endl;
        for (unsigned int i(0); i <= abstractFlightEnvelopesPt->myFlightEnvelope.back().MDive.size() - 1; i++) {
            csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().MDive.at(i).height << "; ";
            csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().MDive.at(i).speed << "; ";
            csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().grossMass << "; ";
            csv <<std::endl;
        }
        csv <<std::endl;
        csv <<std::endl;
        csv << "#Altitude Limit" <<std::endl;
        csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().vStall.back().height  << "; ";
        csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().vStall.back().speed << "; ";
        csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().grossMass << "; ";
        csv <<std::endl;
        csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().MDive.back().height << "; ";
        csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().MDive.back().speed << "; ";
        csv << abstractFlightEnvelopesPt->myFlightEnvelope.back().grossMass << "; ";
        csv.close();
    }
}
void performance_assessment_output::generateSARPlotData() {
    std::ofstream csv;
    for (unsigned int i(0); i < abstractSARPerformancePt->theSAR.size(); i++) {
        std::stringstream Number;
        Number << i;
        string csvDataFile(mySettingsPt->rtIO->getCsvFilesDirTool() + "/" + theProgramName + "_SAR_Mass" + Number.str() + ".csv");
        csv.open(csvDataFile.c_str());
        if (!csv) {
            myRuntimeInfo->err << csvDataFile + " could not be opened!" <<std::endl;
            exit(1);
        } else {
            csv << "# " + csvDataFile <<std::endl;
            // Adjust data structure:
            csv << "# (1) Mach; (2) Gross Mass [kg]; (3) Altitude [m]; (4) SAR [m/kg]" <<std::endl;
            for (unsigned int j(0); j < abstractSARPerformancePt->theSAR.at(i).tabSAR.size(); j++) {
                // csv << "# " << floor(myPerfPt->theSAR.at(i).Mass/1000)*1000 << " kg, " << floor(3.2808 * myPerfPt->theSAR.at(i).tabSAR.at(j).Altitude/100)*100<<std::endl;
                for (unsigned int n(0); n < abstractSARPerformancePt->theSAR.at(i).tabSAR.at(j).Mach.size(); n++) {
                    csv << abstractSARPerformancePt->theSAR.at(i).tabSAR.at(j).Mach.at(n) << "; ";
                    csv << abstractSARPerformancePt->theSAR.at(i).Mass << "; ";
                    csv << abstractSARPerformancePt->theSAR.at(i).tabSAR.at(j).Altitude << "; ";
                    csv << abstractSARPerformancePt->theSAR.at(i).tabSAR.at(j).SAR.at(n) <<std::endl;
                }
                csv <<std::endl;
                csv <<std::endl;
            }
        }
        csv.close();
    }
}
void performance_assessment_output::generateBestSARAltPlotData() {
// TODO(franz#1#): kathi (18.11.2013): Plot seltsam -> reparieren
    std::ofstream plot;
    string plotDataFile(mySettingsPt->rtIO->getCsvFilesDirTool() + "/" + theProgramName + "_bestSAR.csv");
    plot.open(plotDataFile.c_str());
    if (!plot) {
        myRuntimeInfo->err << string(theProgramName) + "_bestSAR.csv could not be opened!" << std::endl;
        exit(1);
    } else {
        plot << "# " + string(theProgramName) + "_bestSAR.csv" << std::endl;
        // Adjust data structure:
        plot << "# (1) Mach; (2) Gross Mass [kg]; (3) Opt. Altitude [m]; (4) Opt. SAR [m/kg]" << std::endl;
        for (unsigned int i(0); i <= abstractSARPerformancePt->theSAR.size() - 1; i++) {
            plot << abstractSARPerformancePt->theSAR.at(i).bestSARMach << "; ";
            plot << abstractSARPerformancePt->theSAR.at(i).Mass << "; ";
            plot << 3.2808 * abstractSARPerformancePt->theSAR.at(i).bestSARAlt << "; ";
            plot << abstractSARPerformancePt->theSAR.at(i).bestSAR << std::endl;
        }
        plot.close();
    }
}
void performance_assessment_output::generateTOFLPlotData() {
std::ofstream csv;
    string csvDataFile(mySettingsPt->rtIO->getCsvFilesDirTool() + "/" + theProgramName + "_TOFL.csv");
    csv.open(csvDataFile.c_str());
    if (!csv) {
        myRuntimeInfo->err << string(theProgramName) + "_TOFL.csv could not be opened!" << std::endl;
        exit(1);
    } else {
        csv << "# " + string(theProgramName) + "_TOFL.csv" << std::endl;
        csv << "# (1) Altitude [m]; (2) Gross Mass [kg]; (3) TODR_AEO [m]" << std::endl;
        for (unsigned int i(0); i <= abstractTakeOffPerformancePt->myTOFLField.size() - 1; i++) {
            for (unsigned int j(0); j <= abstractTakeOffPerformancePt->myTOFLField.at(i).myTOFL.size() - 1; j++) {
                csv << abstractTakeOffPerformancePt->myTOFLField.at(i).elevation * 1. / 0.3048 << "; ";
                csv << abstractTakeOffPerformancePt->myTOFLField.at(i).myTOFL.at(j).grossMass << "; ";
                csv << abstractTakeOffPerformancePt->myTOFLField.at(i).myTOFL.at(j).TODR_AEO << std::endl;
            }
            csv << std::endl;
            csv << std::endl;
        }
        // display TOFL
        csv << abstractTakeOffPerformancePt->myTOFLField.at(0).elevation << "; ";
        csv << abstractTakeOffPerformancePt->myTOFLField.at(0).myTOFL.back().grossMass << "; ";
        csv << abstractTakeOffPerformancePt->myTOFLField.at(0).myTOFL.back().TODR_AEO << std::endl;
        csv << std::endl;
        csv << std::endl;
        // Req TOFL
        csv << 0.0 << "; ";
        csv << myAcftPt->MTOM << "; ";
        csv << myAcftPt->myReqs.TOFL << std::endl;
        csv.close();
    }
}
void performance_assessment_output::generateBLFLPlotData() {
    std::ofstream csv;
    string csvDataFile(mySettingsPt->rtIO->getCsvFilesDirTool() + "/" + theProgramName + "_BLFL.csv");
    csv.open(csvDataFile.c_str());
    if (!csv) {
        myRuntimeInfo->err << string(theProgramName) + "_BLFL.csv could not be opened!" << std::endl;
        exit(1);
    } else {
        csv << "# " + string(theProgramName) + "_BLFL.csv" << std::endl;
        csv << "# (1) Altitude [m]; (2) Gross Mass [kg]; (3) TODR_AEO [m]; (4) TODR_OEI [m]; (5) ASDR_OEI [m]; (6) v1 [m/s]" << std::endl;
        for (unsigned int i(0); i <= abstractTakeOffPerformancePt->myBLFL.vEF.size() - 1; i++) {
            csv << "0.0" << "; ";
            csv << abstractTakeOffPerformancePt->myBLFL.grossMass << "; ";
            csv << abstractTakeOffPerformancePt->myBLFL.TODR_AEO << "; ";
            csv << abstractTakeOffPerformancePt->myBLFL.TODR_OEI.at(i) << "; ";
            csv << abstractTakeOffPerformancePt->myBLFL.ASDR_OEI.at(i) << "; ";
            csv << abstractTakeOffPerformancePt->myBLFL.vEF.at(i) << std::endl;
        }
        csv.close();
    }
}
void performance_assessment_output::generateLDNPlotData() {
    std::ofstream plot;
    string plotDataFile(mySettingsPt->rtIO->getCsvFilesDirTool() + "/" + theProgramName + "_LDN.csv");
    plot.open(plotDataFile.c_str());
    if (!plot) {
        myRuntimeInfo->err << string(theProgramName) + "_LDN.csv could not be opened!" << std::endl;
        exit(1);
    } else {
        plot << "# " + mySettingsPt->rtIO->getCsvFilesDirTool() + theProgramName + "_LDN.csv" << std::endl;
        plot << "# (1) Altitude [m]; (2) Gross Mass [kg]; (3) ALD_AEO [m]" << std::endl;
        for (unsigned int i(0); i <= abstractLandingPerformancePt->myLDNField.size() - 1; i++) {
            for (unsigned int j(0); j <= abstractLandingPerformancePt->myLDNField.at(i).myLDN.size() - 1; j++) {
                plot << abstractLandingPerformancePt->myLDNField.at(i).elevation << "; ";
                plot << abstractLandingPerformancePt->myLDNField.at(i).myLDN.at(j).grossMass << "; ";
                plot << abstractLandingPerformancePt->myLDNField.at(i).myLDN.at(j).ALD_AEO << std::endl;
            }
            plot << std::endl;
            plot << std::endl;
        }
        // display TOFL
        plot << abstractLandingPerformancePt->myLDNField.at(0).elevation << "; ";
        plot << abstractLandingPerformancePt->myLDNField.at(0).myLDN.back().grossMass << "; ";
        plot << abstractLandingPerformancePt->myLDNField.at(0).myLDN.back().ALD_AEO << std::endl;
        plot << std::endl;
        plot << std::endl;
        // Req TOFL
        plot << 0.0 << "; ";
        plot << myAcftPt->MLM << "; ";
        plot << myAcftPt->myReqs.LDN << std::endl;
        plot.close();
    }
}
void performance_assessment_output::generateVnPlotData() {
    std::ofstream plot;
    string plotDataFile(mySettingsPt->rtIO->getCsvFilesDirTool() + "/" + theProgramName + "_VnDiagram.csv");
    plot.open(plotDataFile.c_str());
    if (!plot) {
        myRuntimeInfo->err << string(theProgramName) + "_VnDiagram.csv could not be opened!" << std::endl;
        getchar();
        exit(1);
    } else {
        plot << "# " + mySettingsPt->rtIO->getCsvFilesDirTool() + theProgramName + "_VnDiagram.csv" << std::endl;
        plot << "# (1) v [m]; (2) max n (pos.) [-]; (3) min n (neg.) [-]" << std::endl;
        plot << "#Clean configuration" << std::endl;
        for (unsigned int i(0); i <= abstractVnDiagramPt->myCleanDiagram.size() - 1; i++) {
            plot << abstractVnDiagramPt->myCleanDiagram.at(i).speed << "; ";
            plot << abstractVnDiagramPt->myCleanDiagram.at(i).n_upperBound << "; ";
            plot << abstractVnDiagramPt->myCleanDiagram.at(i).n_lowerBound << std::endl;
        }
        plot << std::endl;
        plot << std::endl;
        plot << "#Takeoff configuration" << std::endl;
        for (unsigned int i(0); i <= abstractVnDiagramPt->myCleanDiagram.size() - 1; i++) {
            plot << abstractVnDiagramPt->myTakeoffDiagram.at(i).speed << "; ";
            plot << abstractVnDiagramPt->myTakeoffDiagram.at(i).n_upperBound << "; ";
            plot << abstractVnDiagramPt->myTakeoffDiagram.at(i).n_lowerBound << std::endl;
        }
        plot << std::endl;
        plot << std::endl;
        plot << "#Landing configuration" << std::endl;
        for (unsigned int i(0); i <= abstractVnDiagramPt->myCleanDiagram.size() - 1; i++) {
            plot << abstractVnDiagramPt->myLandingDiagram.at(i).speed << "; ";
            plot << abstractVnDiagramPt->myLandingDiagram.at(i).n_upperBound << "; ";
            plot << abstractVnDiagramPt->myLandingDiagram.at(i).n_lowerBound << std::endl;
        }
        plot.close();
    }
}


void performance_assessment_output::generatePlots() {

    // if (mySettingsPt->mySwitches.doFuelPlanning == true) {
    //     generate_plot_mission();
    // }
    if (mySettingsPt->mySwitches.doPayloadRange == true) {
        generate_plot_payload_range();
    }
    // if (mySettingsPt->mySwitches.doEnginePerformance == true) {
    //     generate_plot_engine_performance();
    // }
    if (mySettingsPt->mySwitches.doClimbPerformance == true && mySettingsPt->check_sucessfully_executed.doClimbPerformance == true) {
        generate_plot_ceiling_performance();
    }
    // if (mySettingsPt->mySwitches.doFlightEnvelopePerformance == true) {
    //     generate_plot_flight_envelope();
    //     generate_plot_SAR();
    //     generateBestSARAltPlotData();
    // }
    // if (mySettingsPt->mySwitches.doTOPerformance == true) {
    //     generate_plot_takeoff_field_length();
    // }
    // if (mySettingsPt->mySwitches.doBLFLPerformance == true) {
    //     generate_plot_balanced_field_length();
    // }
    // if (mySettingsPt->mySwitches.doLandingPerformance == true) {
    //     generate_plot_landing_performance();
    // }
    // if (mySettingsPt->mySwitches.doVnDiagram == true) {
    //     generate_plot_Vn();
    // }

}

void performance_assessment_output::generate_plot_mission() {
    // matplot::figure_handle fig = matplot::figure(true);
    // matplot::axes_handle ax = fig->current_axes();
    // (...)
}

void performance_assessment_output::generate_plot_payload_range() {
    matplot::figure_handle fig = matplot::figure(true);
    matplot::axes_handle ax = fig->current_axes();
    // Payload-range diagram
    matplot::plot(abstractPayloadRangePt->Range, abstractPayloadRangePt->Payload);
    matplot::title("Payload Range Diagram ");
    ax->xlabel("Range [m]");
    ax->ylabel("Payload [kg]");
    // Design point
    matplot::hold(matplot::on);
    // matplot::scatter(ax, myAcftPt->designRange, myAcftPt->designPayload, "+");
    // (...)
    std::string plot_file_path = rtIO->getPlotDir() + "/" + std::string(TOOL_NAME) + "_payload_range.svg";
    matplot::save(plot_file_path);
}

void performance_assessment_output::generate_plot_ceiling_performance() {
    auto fig = matplot::figure();
    auto ax = fig->current_axes();
    // Gather data
    vector<double> x;
    vector<double> service_ceiling_y;
    vector<double> max_ceiling_y;
    vector<double> max_oei_ceiling_y;
    for (int i(0); i < abstractCeilingPerformancePt->theCeiling.size(); i++) {
        service_ceiling_y.push_back(abstractCeilingPerformancePt->theCeiling.at(i).serviceCeiling);
        max_ceiling_y.push_back(abstractCeilingPerformancePt->theCeiling.at(i).maxCeiling);
        max_oei_ceiling_y.push_back(abstractCeilingPerformancePt->theCeiling.at(i).maxCeilingOEI);
        x.push_back(abstractCeilingPerformancePt->theCeiling.at(i).Mass);
    }
    // Plot
    ax->plot(x, service_ceiling_y)->line_style("-");
    matplot::hold(matplot::on);
    ax->plot(x, max_ceiling_y)->line_style("--");
    ax->plot(x, max_oei_ceiling_y);
    ax->grid(matplot::on);
    ax->legend({"Service Ceiling, Rate of climb 500 ft/min", "Max. ceiling, Rate of climb 0 ft/min", "OEI Ceiling, ROC 500 ft/min"});
    ax->legend()->location(matplot::legend::general_alignment::bottomright);
    ax->xlabel("Gross mass [kg]");
    ax->ylabel("Altitude [m]");
    std::string plot_file_path = rtIO->getPlotDir() + "/" + std::string(TOOL_NAME) + "_ceiling.svg";
    matplot::save(plot_file_path);
}
