/*  Copyright (C) 2024 Chair of Aircraft Design, Technical University Munich
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
/**
 * @file empirical_design.cpp
 * @author Sebastian Oberschwendtner (sebastian.oberschwendtner@tum.de)
 * @brief General empirical design methods.
 * @version 3.0.0
 * @date 2024-02-11
 */

/* === Includes === */
#include "empirical_design.h"
#include <algorithm>
#include <filesystem>
#include <format>
#include <list>
#include <ranges>
#include "../../io/engine_xml.h"
#include "../../utility.h"
#include "unitConversion/constants.h"

/* === Design implementations === */
namespace design
{
    void Empirical::initialize()
    {
        /* Initialize the EngineDesigner */
        EngineDesigner::initialize();
        this->BPR_ = static_cast<float>(this->configuration()->at("module_configuration_file/program_settings/propulsion/engine/Empirical/BPR/value"));
    }

    void Empirical::save()
    {
        /* Copy all engines which do not exist in the engine directory */
        utility::print("Saving the empirical engine results", Level::Info);
        auto is_in_project = [this](const std::string &engine)
        {
            return std::filesystem::exists(this->engine_directory() / engine);
        };
        std::list<std::string> engines_to_copy;
        std::ranges::copy_if(
            this->designed_engines(),
            std::back_inserter(engines_to_copy),
            std::not_fn(is_in_project));

        /* Copy the engines */
        for (const auto &engine : engines_to_copy)
        {
            utility::print("Copying engine: " + engine, Level::Debug);
            std::filesystem::copy(this->engine_database() / engine, this->engine_directory() / engine);
        }
    }

    template <EnergyCarrier carrier>
    auto Empirical::calculate_bucket_point(const Propulsion<carrier> &engine) const -> BucketPoint
    {
        /* Set the step size when looking for the minimum of the bucket curve */
        constexpr double step_size = 1e-2;
        constexpr auto maxiter = static_cast<std::size_t>(1.0 / step_size);

        /* Load the scaled engine */
        auto engine_scaled = io::load_engine_scaled(
            engine.model(),
            engine.scale(),
            {this->engine_directory(),
             this->engine_database()});

        /* Try to calculate the N1 */
        try
        {
            /* Get the maximum N1 */
            engine_scaled.calculate_N1_with_penalties(
                this->flight_condition().altitude,
                this->flight_condition().mach,
                this->flight_condition().ambiance,
                1.0,
                "idle",
                engine.offtakes().bleed_air,
                engine.offtakes().shaft_power / 1000);
        } catch (const std::exception &error) {
            utility::print("Calculating the N1 failed:" + std::string(error.what()) + "Bucket point will not be set!", Level::Warning);
            return {0.0, 0.0};
        }
        const double N1_min = engine_scaled.get_operating_point().N;

        /* Initialize the search point an lowest N1 */
        engine_scaled.calculate_N1_with_penalties(
            this->flight_condition().altitude,
            this->flight_condition().mach,
            this->flight_condition().ambiance,
            1.0,
            "takeoff",
            engine.offtakes().bleed_air,
            engine.offtakes().shaft_power / 1000);
        const double N1_max = engine_scaled.get_operating_point().N;

        /* Print the bucket curve range */
        utility::print(
            std::format(
                "Calculating bucket point for engine {} => N1_min = {:.2f}, N1_max = {:.2f}",
                engine.model(),
                N1_min,
                N1_max),
            Level::Debug);

        /* Set the engine operating point to the lower starting point of N1 */
        auto op = engine_scaled.get_operating_point();
        op.N = N1_min;
        engine_scaled.set_operating_point(op);

        /* Initialize the bucket N1 and TSFC */
        double N1_previous = N1_min;
        double tsfc_previous = engine_scaled.get_aircraft_fuelflow() / engine_scaled.get_thrust_aircraft();

        /* Loop until the fuel flow gradient becomes positive */
        for (std::size_t i = 0; i < maxiter; ++i)
        {
            /* Update the operating point of the engine */
            const double N1 = N1_previous + step_size;
            op.N = N1;
            engine_scaled.set_operating_point(op);

            /* Calculate the fuel flow at the current N1 */
            const double tsfc = engine_scaled.get_aircraft_fuelflow() / engine_scaled.get_thrust_aircraft();

            /* Check whether the gradient is positive */
            if (tsfc > tsfc_previous)
            {
                break;
            } else {
                /* Update the tsfc and N1 */
                tsfc_previous = tsfc;
                N1_previous = N1;
            }

            /* Check whether N1 is still valid */
            if (N1 > N1_max)
            {
                utility::print("Maximum N1 reached in bucket point calculation!", Level::Warning);
                break;
            }

            /* Warn if iter reached maxiter */
            if (i == maxiter - 1)
            {
                utility::print("Maximum number of iterations reached in bucket point calculation!", Level::Warning);
            }
        }

        /* Calculate the found bucket point */
        op.N = N1_previous;
        engine_scaled.set_operating_point(op);
        const double thrust_bucket = engine_scaled.get_thrust_aircraft();
        const double fuel_flow_bucket = engine_scaled.get_aircraft_fuelflow();
        utility::print(
            std::format(
                "Found bucket point for engine {} at N1 = {:.2f} => thrust = {:3.1f} kN, TSFC = {:1.3e} kg/Ns",
                engine.model(), N1_previous, thrust_bucket / 1000, fuel_flow_bucket / thrust_bucket),
            Level::Info);

        /* Return the bucket point */
        return {
            thrust_bucket,
            fuel_flow_bucket / thrust_bucket};
    };
    template auto Empirical::calculate_bucket_point(const Propulsion<EnergyCarrier::Kerosene> &) const -> BucketPoint;
    template auto Empirical::calculate_bucket_point(const Propulsion<EnergyCarrier::Liquid_Hydrogen> &) const -> BucketPoint;

    auto Empirical::calculate_thrust_table(double BPR, double required_thrust, std::vector<double> FL,  std::vector<double> Mach, std::vector<double> N) -> DeckData
    {
        DeckData data_deck = generate_empty_deck_data("thrust", FL,  Mach, N);
        
        // Define the shape of the multi_array
        boost::array<std::size_t, 3> shape = {data_deck.N.size(), data_deck.FL.size(), data_deck.Mach.size()};

        double c1;
        double c2;
        double c3;
        double c4;
        double c5;

        for (std::size_t i = 0; i < shape[0]; ++i) {
            for (std::size_t j = 0; j < shape[1]; ++j) {
                for (std::size_t k = 0; k < shape[2]; ++k) {
                    double test = data_deck.Mach[k];
                    if (BPR > 0 && BPR <= 1 && data_deck.Mach[k] < 0.4) {
                        c1 = 1;
                        c2 = 0;
                        c3 = -0.2;
                        c4 = 0.07;
                        c5 = 0.8;
                    }
                    else if (BPR > 0 && BPR <= 1 && data_deck.Mach[k] >= 0.4 && data_deck.Mach[k] <= 0.9) { 
                        c1 = 0.856;
                        c2 = 0.062;
                        c3 = 0.16;
                        c4 = -0.23;
                        c5 = 0.8;
                    }
                    //if 1 < BPR < 3, interpolate
                    if (BPR > 1 && BPR <= 3 && data_deck.Mach[k] < 0.4) {
                        c1 = 1; //same value in both cases
                        c2 = 0; //same value in both cases
                        c3 = -0.2 + ((-0.6+0.2)/(3-1))*BPR;
                        c4 = 0.07 + ((-0.04-0.07)/(3-1))*BPR;
                        c5 = 0.8 + ((0.7-0.8)/(3-1))*BPR;
                    }
                    else if (BPR > 1 && BPR <= 3 && data_deck.Mach[k] >= 0.4 && data_deck.Mach[k] <= 0.9) { 
                        c1 = 0.856 + ((0.88-0.856)/(3-1))*BPR;
                        c2 = 0.062 + ((-0.016-0.062)/(3-1))*BPR;
                        c3 = 0.16 + ((-0.3-0.16)/(3-1))*BPR;
                        c4 = -0.23 + ((0+0.23)/(3-1))*BPR;
                        c5 = 0.8 + ((0.7-0.8)/(3-1))*BPR;
                    }
                    else if (BPR >= 3 && BPR <= 6 && data_deck.Mach[k] >= 0.0 && data_deck.Mach[k] < 0.4) {
                        c1 = 1;
                        c2 = 0;
                        c3 = -0.6;
                        c4 = -0.04;
                        c5 = 0.7;
                    }
                    else if (BPR >= 3 && BPR <= 6 && data_deck.Mach[k] >= 0.4 && data_deck.Mach[k] <= 0.9) { 
                        c1 = 0.88;
                        c2 = -0.016;
                        c3 = -0.3;
                        c4 = 0;
                        c5 = 0.7;
                    }
                    // if 6 < BPR < 8, interpolate
                    else if (BPR > 6 && BPR < 8 && data_deck.Mach[k] >= 0.0 && data_deck.Mach[k] < 0.4) {
                        c1 = 1; //same value in both cases
                        c2 = 0; //same value in both cases
                        c3 = -0.6 + ((-0.595+0.6)/(8-6))*BPR; 
                        c4 = -0.04 + ((-0.03+0.04)/(8-6))*BPR; 
                        c5 = 0.7; //same value in both cases
                    }
                    else if (BPR > 6 && BPR < 8 && data_deck.Mach[k] >= 0.4 && data_deck.Mach[k] <= 0.9) { 
                        c1 = 0.88 + ((0.89-0.88)/(8-6))*BPR;
                        c2 = -0.016 + ((-0.014+0.016)/(8-6))*BPR;
                        c3 = -0.3; //same value in both cases
                        c4 = 0 + ((0.005-0)/(8-6))*BPR;
                        c5 = 0.7; //same value in both cases
                    }
                    // if BPR > 8, Use values for BPR = 8:
                    else if (BPR >= 8 && data_deck.Mach[k] >= 0.0 && data_deck.Mach[k] < 0.4) {
                        c1 = 1;
                        c2 = 0;
                        c3 = -0.595;
                        c4 = -0.03;
                        c5 = 0.7;
                    }
                    else if (BPR >= 8 && data_deck.Mach[k] >= 0.4 && data_deck.Mach[k] <= 0.9) { 
                        c1 = 0.89;
                        c2 = -0.014;
                        c3 = -0.3;
                        c4 = 0.005;
                        c5 = 0.7;
                    }
                    else{
                        // Else-Bedingung: Programmabbruch mit Fehlermeldung
                        std::cerr << "BPR or Mach out of borders. Selected BPR: " << BPR << " and should be 0 <= BPR. Selected Mach: "  << data_deck.Mach[k] <<  " and should be 0 <= Mach <= 0.9. Required_thrust: "<< required_thrust << std::endl;
                        // Abbruch des Programms mit exit
                        std::exit(EXIT_FAILURE);
                    }
                    if (data_deck.FL[j] > 110000){
                        c5 = 1.0;
                    }
                    double density = atmosphere().getDensity(data_deck.FL[j]);
                    double thrust_M_FL = required_thrust * (c1 + c2 * BPR + (c3 + c4 * BPR) * data_deck.Mach[k] * pow((density/1.225), c5));
                    double test_thrust = thrust_M_FL * pow(data_deck.N[i], 3.5);
                    data_deck.values[i][j][k] = thrust_M_FL * pow(data_deck.N[i], 3.5);
                }
            }
        }
        return data_deck;
    }

    auto Empirical::calculate_TSFC_table(double BPR, double required_thrust, DeckData thrust_deck) -> DeckData
    {
        DeckData data_deck = generate_empty_deck_data("TSFC", thrust_deck.FL,  thrust_deck.Mach, thrust_deck.N);
        std::vector<double> test3 = thrust_deck.FL;
        std::vector<double> test4 = thrust_deck.Mach;
        std::vector<double> test5 = thrust_deck.N;
        
        double c1 = 0.71 * 1/(9.81 * 3600) - 0.15 * 1/(9.81 * 3600) * std::sqrt(BPR);
        boost::array<std::size_t, 3> shape = {data_deck.N.size(), data_deck.FL.size(), data_deck.Mach.size()};

        for (std::size_t i = 0; i < shape[0]; ++i) {
            for (std::size_t j = 0; j < shape[1]; ++j) {
                for (std::size_t k = 0; k < shape[2]; ++k) {
                    double temp = atmosphere().getTemperature(data_deck.FL[j]);
                    double T = thrust_deck.values[i][j][k];
                    double T_share = T/required_thrust;
                    double TSFC_max_dry = (c1 + 0.0000126 * data_deck.Mach[k]) * std::sqrt(temp/288.15);
                    data_deck.values[i][j][k] = (c1 + 0.0000126 * data_deck.Mach[k]) * std::sqrt(temp/288.15) * (0.1/T_share + 0.24/pow(T_share, 0.8) + 0.66 * pow(T_share, 0.8) + 0.1 * data_deck.Mach[k] * (1/T_share - T_share));
                    if (data_deck.values[i][j][k] >= 1.5 * TSFC_max_dry){
                        data_deck.values[i][j][k] = 1.5 * TSFC_max_dry;
                        }
                    }
                }
            }
        return data_deck;
    }

    auto Empirical::calculate_fuelflow_table(DeckData thrust_deck, DeckData TSFC_deck) -> DeckData
    {
        DeckData data_deck = generate_empty_deck_data("fuelflow", thrust_deck.FL,  thrust_deck.Mach, thrust_deck.N);

        boost::array<std::size_t, 3> shape = {data_deck.N.size(), data_deck.FL.size(), data_deck.Mach.size()};

        for (std::size_t i = 0; i < shape[0]; ++i) {
            for (std::size_t j = 0; j < shape[1]; ++j) {
                for (std::size_t k = 0; k < shape[2]; ++k) {
                    data_deck.values[i][j][k] = TSFC_deck.values[i][j][k] * thrust_deck.values[i][j][k];
                    }
                }
            }
        return data_deck;
    }

    auto Empirical::calculate_energyflow_table(DeckData mf_deck,  double heating_value) -> DeckData
    {
        DeckData data_deck = generate_empty_deck_data("energyflow", mf_deck.FL,  mf_deck.Mach, mf_deck.N);

        boost::array<std::size_t, 3> shape = {data_deck.N.size(), data_deck.FL.size(), data_deck.Mach.size()};

        for (std::size_t i = 0; i < shape[0]; ++i) {
            for (std::size_t j = 0; j < shape[1]; ++j) {
                for (std::size_t k = 0; k < shape[2]; ++k) {
                    data_deck.values[i][j][k] = mf_deck.values[i][j][k] * heating_value;
                    }
                }
            }
        return data_deck;
    }

    auto Empirical::calculate_thermal_efficiency_table(DeckData thrust_deck,  DeckData energyflow_deck) -> DeckData
    {
        DeckData data_deck = generate_empty_deck_data("thermal_efficiency", thrust_deck.FL,  thrust_deck.Mach, thrust_deck.N);

        boost::array<std::size_t, 3> shape = {data_deck.N.size(), data_deck.FL.size(), data_deck.Mach.size()};

        for (std::size_t i = 0; i < shape[0]; ++i) {
            for (std::size_t j = 0; j < shape[1]; ++j) {
                for (std::size_t k = 0; k < shape[2]; ++k) {
                    double speed_of_sound = atmosphere().getSpeedOfSound(data_deck.FL[j]);
                    data_deck.values[i][j][k] = (data_deck.Mach[k] * speed_of_sound * thrust_deck.values[i][j][k]) / energyflow_deck.values[i][j][k];
                    }
                }
            }
        return data_deck;
    }
    
    auto Empirical::calculate_engine_dimensions(double required_thrust,  double BPR, double Mach_max) -> std::tuple<double, double>
    {
        // Raymer 10.3.1. Only valid for engines without afterburner!
        double L = 0.49 * pow(required_thrust, 0.4) * pow(Mach_max, 0.2);
        double D = 0.15 * pow(required_thrust, 0.5) * std::exp(0.04 * BPR);

        return {L, D};
    }
    
    auto Empirical::calculate_engine_mass(double required_thrust,  double BPR) -> double
    {
        // Raymer 10.3.1. Only valid for engines without afterburner!
        double W = 14.7 * pow(required_thrust, 1.1) * std::exp(-0.045 * BPR);

        return W;
    }

    auto Empirical::calculate_core_mass(double BPR, double D, DeckData thrust_deck) -> DeckData
    {
        
        DeckData core_mass = generate_empty_deck_data("core_mass", thrust_deck.FL,  thrust_deck.Mach, thrust_deck.N);
        boost::array<std::size_t, 3> shape = {core_mass.N.size(), core_mass.FL.size(), core_mass.Mach.size()};
        for (std::size_t i = 0; i < shape[0]; ++i) {
            for (std::size_t j = 0; j < shape[1]; ++j) {
                for (std::size_t k = 0; k < shape[2]; ++k) {
                    double speed_of_sound = atmosphere().getSpeedOfSound(thrust_deck.FL[j]);
                    double density = atmosphere().getDensity(thrust_deck.FL[j]);
                    core_mass.values[i][j][k] = 1/(1+BPR) * thrust_deck.Mach[k] * speed_of_sound * density * pow(D/2, 2) * PI;
                }
            }
        }
        return core_mass;
    }

} // namespace design