/*  Copyright (C) 2023 Chair of Aircraft Design, Technical University Munich
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
/**
 * @file empirical_design.h
 * @author Tobias Weckenmann (tobias.weckenmann@tum.de)
 * @brief The empirical engine designer
 * @version 3.0.0
 * @date 2023-12-15
 */
#ifndef SRC_ENGINE_DESIGN_EMPIRICAL_EMPIRICAL_DESIGN_H_
#define SRC_ENGINE_DESIGN_EMPIRICAL_EMPIRICAL_DESIGN_H_

/* === Includes === */
#include "../engine_design.h"
#include <filesystem>
#include <memory>
#include <optional>
#include <unordered_map>
#include <unordered_set>
#include <string>
#include <string_view>

namespace design
{
    /* === Classes === */
    /**
     * @class Empirical
     * @brief The empirical engine designer
     */
    class Empirical : public EngineDesigner
    {
      public:
        /* === Constructors === */
        /**
         * @brief Construct a new empirical engine designer.
         * @param configuration The module configuration provided by the user.
         * @param engine_directory The directory where the engine results are stored.
         * @param design_condition The design flight condition of the aircraft.
         */
        explicit Empirical(
            const std::shared_ptr<node> &configuration,
            const std::filesystem::path &engine_directory,
            const FlightCondition &design_condition)
            : EngineDesigner(configuration, engine_directory, design_condition) {}

        /* === Methods === */
        /**
         * @brief Finalize the initialization of the engine designer.
         */
        void initialize() final;

        /**
         * @brief Finalize the saving of the engine results.
         */
        void save() final;

        /**
         * @brief Design a turbofan engine with kerosene as fuel.
         * @note The engine WILL be modified!
         *
         * @param engine The engine to design.
         */
        void operator()(Turbofan<EnergyCarrier::Kerosene> &engine) final; // NOLINT runtime/references

    private:
        /* === Methods === */
        /**
         * @brief Calculate the bucket point of the engine.
         * The design flight condition and the current engine offtakes
         * are taken as the boundary conditions for the bucket point.
         *
         * @tparam carrier The energy carrier of the engine.
         * @param engine The engine to calculate the bucket point for.
         * @return BucketPoint [N, (kg/s)/N] The bucket point (thrust, tsfc) of the engine
         * at the current operating conditions.
         */
        template <EnergyCarrier carrier>
        [[nodiscard]] auto calculate_bucket_point(const Propulsion<carrier> &engine) const -> BucketPoint;

        /**
         * @brief Get the preselected engine for the given engine id.
         * When the specific id does not have a preselected engine,
         * the default preselected engine is returned. If there is no
         * default preselected engine, an empty optional is returned.
         *
         * @param engine_id The id of the engine.
         * @return std::optional<std::string_view> The preselected engine if available.
         */
        [[nodiscard]] auto preselected_engine(int engine_id) -> std::optional<std::string_view>;

        /**
         * @brief Selects the engine with the thrust closest to the required thrust.
         * It searches the database directory for the engine with the thrust closest
         * to the required thrust.
         *
         * @param thrust_required [N] The required thrust.
         * @return std::string The name of the selected engine.
         * @throws std::runtime_error if no engine is found in the database.
         */
        [[nodiscard]] auto select_engine(const double thrust_required) const -> std::string;

        /**
         * @brief Calculates the thrust deck based on empirical methods.
         *
         * @param BPR [-] The BPR ratio defined in the input file.
         * @param required_thrust [N] The required thrust.
         * @param FL vector of the flight level conditions.
         * @param Mach vector of the Mach conditions.
         * @param N vector of the spool speed conditions.
         * 
         * @return Calculated thrust data deck.
         */
        auto calculate_thrust_table(double BPR, double required_thrust, std::vector<double> FL,  std::vector<double> Mach, std::vector<double> N) -> DeckData;

        /**
         * @brief Calculates the TSFC deck based on empirical methods.
         *
         * @param BPR [-] The BPR ratio defined in the input file.
         * @param required_thrust [N] The required thrust.
         * @param thrust_deck The thrust DeckData.
         * 
         * @return Calculated thrust specific fuel consumption data deck.
         */
        auto calculate_TSFC_table(double BPR, double required_thrust, DeckData thrust_deck) -> DeckData;

        /**
         * @brief Calculates the fuelflow deck based on empirical methods.
         *
         * @param thrust_deck The thrust DeckData.
         * @param TSFC_deck The TSFC thrust DeckData.
         * 
         * @return Calculated fuel flow data deck.
         */
        auto calculate_fuelflow_table(DeckData thrust_deck, DeckData TSFC_deck) -> DeckData;

        /**
         * @brief Calculates the energyflow deck based on empirical methods.
         *
         * @param massflow_deck The massflow DeckData
         * @param heating_value Heating value of the fuel.
         * 
         * @return Calculated energy flow data deck.
         */
        auto calculate_energyflow_table(DeckData mf_deck,  double heating_value) -> DeckData;

        /**
         * @brief Calculates the thermal efficiency deck based on empirical methods.
         *
         * @param thrust_deck The thrust DeckData.
         * @param energyflow_deck The energy flow DeckData.
         * 
         * @return Calculated thermal efficiency data deck.
         */
        auto calculate_thermal_efficiency_table(DeckData thrust_deck,  DeckData energyflow_deck) -> DeckData;

        /**
         * @brief Calculates the engine dimensions.
         *
         * @param required_thrust [N] The required thrust.
         * @param BPR [-] The BPR ratio defined in the input file.
         * @param Mach_max [-] Maximal Mach number in data deck.
         * 
         * @return Engine length and diameter.
         */
        auto calculate_engine_dimensions(double required_thrust,  double BPR, double Mach_max) -> std::tuple<double, double>;

        /**
         * @brief Calculates the engine mass.
         *
         * @param required_thrust [N] The required thrust.
         * @param BPR [-] The BPR ratio defined in the input file.
         * 
         * @return Engine mass.
         */
        auto calculate_engine_mass(double required_thrust,  double BPR) -> double;

        /**
         * @brief Calculates the engine core mass flow.
         *
         * @param BPR [-] The BPR ratio defined in the input file.
         * @param D Engine Fan diameter.
         * @param thrust_deck Thrust data deck.
         * 
         * @return Engine core mass flow.
         */
        auto calculate_core_mass(double BPR, double D, DeckData thrust_deck) -> DeckData;


        /* === Properties === */
        //std::unordered_map<int, std::string> preselected_engines_; /**< The manual selected engines. */
        double BPR_{0.0}; 
    };
}; // namespace design

#endif // SRC_ENGINE_DESIGN_EMPIRICAL_EMPIRICAL_DESIGN_H_
