/*  Copyright (C) 2023 Chair of Aircraft Design, Technical University Munich
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
/**
 * @file turbofan.cpp
 * @author Tobias Weckenmann (tobias.wechenmann@tum.de)
 * @author Sebastian Oberschwendtner (sebastian.oberschwendtner@tum.de)
 * @brief Design a turbofan using the  engine designer.
 * @version 3.0.0
 * @date 2023-12-16
 */

/* === Includes === */
#include "empirical_design.h"
#include <cmath>
#include <format>
#include "../../utility.h"
#include "../../io/engine_xml.h"

/* === Design implementations === */
namespace design
{
    void Empirical::operator()(Turbofan<EnergyCarrier::Kerosene> &engine)
    {
        /*Get required thrust*/
        std::vector<double> FL = {0, 500, 1000, 1500, 2000, 2500, 3000, 3500, 4000, 4500, 5000, 5500, 6000, 6500, 7000, 7500, 8000, 8500, 9000, 9500, 10000, 10500, 11000, 11500, 12000, 12500, 13000, 13500, 14000}; // {Flugniveaus in Metern
        std::vector<double> Mach = {0, 0.05, 0.1, 0.15, 0.2, 0.25, 0.3, 0.35, 0.4, 0.45, 0.5, 0.55, 0.60, 0.65, 0.70, 0.75, 0.8, 0.85, 0.9};   // Mach-Zahlen
        std::vector<double> N = {1.1, 1.05, 1.0, 0.95, 0.9, 0.85, 0.8, 0.75, 0.7, 0.65, 0.6, 0.55, 0.5, 0.48, 0.46, 0.44, 0.42, 0.4, 0.39, 0.38, 0.37, 0.36, 0.35, 0.34, 0.33, 0.32, 0.31, 0.30, 0.29};

        engine.set_model("Empirical_engine");
        this->initialize();
        engine.set_bypass_ratio(this->BPR_);
        std::cout<<"required_thrust "<< engine.required_thrust()<<"\n";

        DeckData thrust_deck = this->calculate_thrust_table(engine.bypass_ratio(), engine.required_thrust(), FL, Mach, N);
        this->write_deck_value(thrust_deck, "Empirical_engine");
        DeckData TSFC_deck = calculate_TSFC_table(engine.bypass_ratio(), engine.required_thrust(), thrust_deck);
        this->write_deck_value(TSFC_deck, "Empirical_engine");
        DeckData ff_deck = calculate_fuelflow_table(thrust_deck, TSFC_deck);
        this->write_deck_value(ff_deck, "Empirical_engine");
        DeckData ef_deck = calculate_energyflow_table(ff_deck, 43100000);
        this->write_deck_value(ef_deck, "Empirical_engine");
        DeckData thermal_efficiency_deck = calculate_thermal_efficiency_table(thrust_deck, ef_deck);
        this->write_deck_value(thermal_efficiency_deck, "Empirical_engine");

        /* Set and sale the parameters */
        /** @todo Where does the scaling of the length come from ?*/
        double max_Mach = *std::max_element(Mach.begin(), Mach.end());
        auto dimensions = this->calculate_engine_dimensions(engine.required_thrust(), engine.bypass_ratio(), max_Mach);
        double d = std::get<1>(dimensions);
        double l = std::get<0>(dimensions);
        engine.set_dimension({d, d, l});
        engine.set_fan({d});

        DeckData core_mass = calculate_core_mass(engine.bypass_ratio(), d, thrust_deck);
        this->write_deck_value(core_mass, "Empirical_engine");

        /* Get the engine data from the project or database directory */
        const EngineData engine_data =
            io::load_engine_data(engine.model(),
                                {this->engine_directory(), this->engine_database()});

        /* Set the engine mass */
        PointMass engine_mass;
        engine_mass.mass = this->calculate_engine_mass(engine.required_thrust(), engine.bypass_ratio());
        engine.set_pointmass(Component::Engine, engine_mass);

        /* Set the bucket point of the engine */
        engine.set_bucket_point(
            this->calculate_bucket_point(engine));

        /* At this point the engine was successfully sized */
        this->add_designed_engine(engine.model());


        // /** @todo Should this selection always take place or should we stick to the selection
        //  * in the following iteration loops even if there is a better fitting engine now?
        //  */
        // engine.set_model(this->select_engine(engine.required_thrust()));

        // /*Calculate */
        // /* Get the engine data from the project or database directory */
        // const EngineData engine_data =
        //     io::load_engine_data(engine.model(),
        //                         {this->engine_directory(), this->engine_database()});

        // /* Scale the engine */
        // /** @todo How and which thrust do we need here? */
        // engine.set_scale(
        //     engine.required_thrust() > 0.0 ? engine.required_thrust() / engine_data.SLST() : 1.0);

        // /* Set and sale the parameters */
        // /** @todo Where does the scaling of the length come from ?*/
        // engine.set_dimension({engine_data.dimensions().height * std::pow(engine.scale(), 0.5),
        //                       engine_data.dimensions().width * std::pow(engine.scale(), 0.5),
        //                       engine_data.dimensions().length * std::pow(engine.scale(), 0.4)});
        // engine.set_fan({engine_data.dimensions().diameter * std::pow(engine.scale(), 0.5)});

        // /* Set the engine mass */
        // PointMass engine_mass;
        // engine_mass.mass = std::pow(engine.scale(), 1.1) * engine_data.dry_mass(); // Ray12 p. 285
        // engine_mass.mass *= this->technology_factor("engine_mass"); // Apply the technology factor
        // engine.set_pointmass(Component::Engine, engine_mass);

        // /* Set the bucket point of the engine */
        // engine.set_bucket_point(
        //     this->calculate_bucket_point(engine)
        // );

        // /* At this point the engine was successfully sized */
        // this->add_designed_engine(engine.model());

        // /* Print debug information */
        // utility::print(std::format(
        //     "Designed engine: {} with thrust = {:6.2f} kN, mass = {:5.1f} kg, scale = {:1.2f}",
        //     engine.model(), engine.required_thrust() / 1000, engine_mass.mass, engine.scale()), Level::Info);
    }

} // namespace design
