/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "aircraftSystem.h"

#include "systemsIOData.h"
#include <runtimeInfo/runtimeInfo.h>
#include <unitConversion/constants.h>
#include <memory>

/* constructor for all systems except for power sources*/
aircraftSystem::aircraftSystem(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO)
    :
    data_(data),
    rtIO_(rtIO),
    Name(systemName),
    ATA(""),
    Operating(systemIsOperating),
    Mass(0.0) {
    if (Name == "virtualSystem") {
        ATA = "ATA-virtual";
        myRuntimeInfo->out << "  Create: " << Name << "  ->  " << ATA << std::endl;
    } else if (Name == "dummySystemName") {
        ATA = "ATA-Dummy";
    }
    refPoint[0] = 0.0;
    refPoint[1] = 0.0;
    refPoint[2] = 0.0;
    bleedAir.designPower = (0.0);
    bleedAir.maxMissionPower = (0.0);
    /*Creating the hydraulic vector (hydraulic circuits)*/
    for (int n(0); n < data_->configuration.Hydraulic.numberOfhydraulicCircuits; n++) {
        Hydraulic.push_back(powerProfile());
        Hydraulic.back().designPower = (0.0);
        Hydraulic.back().maxMissionPower = (0.0);
    }
    /*Creating the electric vector (electric circuits)*/
    for (int n(0); n < data_->configuration.Electric.numberOfElectricCircuits; n++) {
        Electric.push_back(powerProfile());
        Electric.back().designPower = (0.0);
        Electric.back().maxMissionPower = (0.0);
    }
    heatLoad.designPower = 0.0;
    heatLoad.maxMissionPower = 0.0;
    /* Initializing the vectors for the mission-dependent power requirement **/
    initializePowerProfile();
    //ctor
}

aircraftSystem::~aircraftSystem() {
    //dtor
}

void aircraftSystem::initializePowerProfile() {
    //Hydraulics
    for (int n(0); n < data_->configuration.Hydraulic.numberOfhydraulicCircuits; n++) {
        Hydraulic.at(n).missionPower.clear();
        Hydraulic.at(n).designPower = (0.0);
        Hydraulic.at(n).maxMissionPower = (0.0);
        for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
            Hydraulic.at(n).missionPower.push_back(missionLoads());
        }
    }
    //Electric
    for (int n(0); n < data_->configuration.Electric.numberOfElectricCircuits; n++) {
        Electric.at(n).missionPower.clear();
        Electric.at(n).designPower = (0.0);
        Electric.at(n).maxMissionPower = (0.0);
        for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
            Electric.at(n).missionPower.push_back(missionLoads());
        }
    }
    //Bleed air and heat load
    bleedAir.missionPower.clear();
    bleedAir.designPower = (0.0);
    bleedAir.maxMissionPower = (0.0);
    heatLoad.missionPower.clear();
    heatLoad.designPower = 0.0;
    heatLoad.maxMissionPower = 0.0;
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        bleedAir.missionPower.push_back(missionLoads());
        heatLoad.missionPower.push_back(missionLoads());
    }
}

void aircraftSystem::calculateMaxMissionPower() { // cppcheck-suppress unusedFunction
    bleedAir.maxMissionPower = bleedAir.getMaxPower();
    for (int n(0); n < data_->configuration.Electric.numberOfElectricCircuits; n++) {
        Electric.at(n).maxMissionPower = Electric.at(n).getMaxPower();
    }
    heatLoad.maxMissionPower = heatLoad.getMaxPower();
    for (int n(0); n < data_->configuration.Hydraulic.numberOfhydraulicCircuits; n++) {
        Hydraulic.at(n).maxMissionPower = Hydraulic.at(n).getMaxPower();
    }
}

double aircraftSystem::getElectricPower() {
    powerProfile electricSum(getPowerProfileSum(Electric));
    return electricSum.designPower;
}

double aircraftSystem::getHydraulicPower() {
    powerProfile hydraulicSum(getPowerProfileSum(Hydraulic));
    return hydraulicSum.designPower;
}

void aircraftSystem::updateMass(std::string node_name, double value, std::string ATA) {
    std::string subPath = "aircraft_exchange_file/component_design/systems/specific/geometry/mass_properties";
    if (!(rtIO_.acxml.find(subPath))) {
        rtIO_.acxml[subPath];
    }
    Endnode<double> mass = data_->data.create_endnode(value, "kg", subPath + "/" + node_name, "mass of " + ATA);
    mass.set_boundaries(0, INFINITY);
    mass.update(rtIO_.acxml);
}
