/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "conventionalATA27.h"
#include "highLiftSystem/highLiftSystem.h"
#include <algorithm>

#include <runtimeInfo/runtimeInfo.h>
#include <unitConversion/unitConversion.h>
#include <aircraftGeometry2/processing/transform.h>

conventionalATA27::conventionalATA27(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO)
    :
    aircraftSystem(systemName, systemIsOperating, data, rtIO),
    rollFCMass(0.),
    yawFCMass(0.),
    pitchFCMass(0.),
    THSA_FCMass(0.),
    FlapsFCMass(0.),
    SpoilersFCMass(0.),
    SlatsFCMass(0.),
    commonInstFCMass(0.),
    primaryFCSMass(0.),
    secondaryFCSMass(0.),
    MaximumGrossMass(data_->data.Mass.MTOM) {
    ATA = "ATA-27";
    myRuntimeInfo->out << "  Create: " << Name << "  ->  " << ATA << std::endl;
    //ctor
}

conventionalATA27::~conventionalATA27() {
    //dtor
}

conventionalATA27::flightControlSystem::flightControlSystem()
    :
    flyMission(false),
    electricalFCS(false),
    roll_EFCSWeightFactor(0.),
    yaw_EFCSWeightFactor(0.),
    pitch_EFCSWeightFactor(0.),
    THSA_EFCSWeightFactor(0.),
    HL_TE_EFCSWeightFactor(0.),
    Spoiler_EFCSWeightFactor(0.),
    HL_LE_EFCSWeightFactor(0.),
    deploymentHighLiftTime(0.),
    commonInst_FCS_Factor(0.),
    numberOfSlatSystems(1),
    numberOfFlapSystems(1),
    numberOfAilerons(1),
    numberOfSpoilers(1),
    numberOfElevators(1),
    numberOfRudders(1),
    numberOfTHSAs(1),
    defaultAileronActuator(false),
    defaultSpoilerActuator(false),
    defaultElevatorActuator(false),
    defaultRudderActuator(false),
    defaultTHSAActuator(false) {
    //ctor
}

conventionalATA27::flightControlSystem::~flightControlSystem() {
    //dtor
}

/* Reading system specific parameters of configuration xml file */
void conventionalATA27::readParameters(const node& config) {
    scaling_factor = EndnodeReadOnly<double>("/module_configuration_file/program_settings/scaling_factors/flight_control_system").read(config).value();
    //Zeros of all control surfaces at the beginning of the calculation
    flightControl.ailerons.clear();
    flightControl.spoilers.clear();
    flightControl.elevators.clear();
    flightControl.rudders.clear();
    flightControl.THSAs.clear();
    flightControl.highLift.clear();
    // Flight Control
    std::string flightControlConfig = "/module_configuration_file/program_settings/systems_constants/flight_controls";
    flightControl.flyMission = EndnodeReadOnly<bool>(flightControlConfig + "/calculate_mission_loads").read(config).value();
    if (flightControl.flyMission) {
        myRuntimeInfo->out << "Energy consumption of the flight control is calculated via rudder deflections in the mission." << std::endl;
    } else {
        myRuntimeInfo->out << "Energy consumption of the flight control is calculated by average consumption (config file)." << std::endl;
    }
    //Number of THSAs is always read from config, because not defined in Acft-XML
    flightControl.numberOfTHSAs = config.at(flightControlConfig + "/trimmable_horizontal_stabilizer/number_of_trimmable_horizontal_stabilizers/value");
    flightControl.deploymentHighLiftTime =
        config.at(flightControlConfig + "/high_lift_system/duration_high_lift_deployment/value");
    data_->configuration.checkForValidity(flightControl.deploymentHighLiftTime, 5.0, 120.0, "flightControl.deploymentHighLiftTime");
    flightControl.electricalFCS = EndnodeReadOnly<bool>(flightControlConfig + "/electrical_fc_system").read(config).value();
    //Weight reduction factor when using a Fly-By-Wire system
    flightControl.roll_EFCSWeightFactor =
        config.at(flightControlConfig + "/ailerons/electric_flight_control_system_weight_factor/value");
    flightControl.yaw_EFCSWeightFactor =
        config.at(flightControlConfig + "/rudders/electric_flight_control_system_weight_factor/value");
    flightControl.pitch_EFCSWeightFactor =
        config.at(flightControlConfig + "/elevators/electric_flight_control_system_weight_factor/value");
    flightControl.THSA_EFCSWeightFactor =
        config.at(flightControlConfig + "/trimmable_horizontal_stabilizer/electric_flight_control_system_weight_factor/value");
    flightControl.HL_TE_EFCSWeightFactor =
        config.at(flightControlConfig + "/high_lift_system/trailing_edge/electric_flight_control_system_weight_factor/value");
    flightControl.Spoiler_EFCSWeightFactor =
        config.at(flightControlConfig + "/spoilers/electric_flight_control_system_weight_factor/value");
    flightControl.HL_LE_EFCSWeightFactor =
        config.at(flightControlConfig + "/high_lift_system/leading_edge/electric_flight_control_system_weight_factor/value");
    flightControl.commonInst_FCS_Factor =
        config.at(flightControlConfig + "/common_installation_weight_factor/value");
    //Creation of the default actuator layout, which is always used if no detailed information about the actuators of the respective control surface is known.
    flightControl.theDefaultActuator.powerSource.type =
        std::string(config.at(flightControlConfig + "/default_actuator/power_source/type/value"));
    flightControl.theDefaultActuator.powerSource.sourceID =
        config.at(flightControlConfig + "/default_actuator/power_source/source_ID/value");
    flightControl.theDefaultActuator.powerSource.operationFactor = 1.;//an actuator is always driven by only one energy source, therefore always "1".
    flightControl.theDefaultActuator.operationMode =
        std::string(config.at(flightControlConfig + "/default_actuator/operation_mode/value"));
    flightControl.theDefaultActuator.efficiency =
        config.at(flightControlConfig + "/default_actuator/efficiency/value");
    flightControl.theDefaultActuator.stdbyPower =
        config.at(flightControlConfig + "/default_actuator/standby_power/value");
    flightControl.theDefaultActuator.activePower =
        config.at(flightControlConfig + "/default_actuator/active_power/value");
    if (flightControl.flyMission == true && (!accuracyCheck(flightControl.theDefaultActuator.stdbyPower, 0., ACCURACY_HIGH)
            || !accuracyCheck(flightControl.theDefaultActuator.activePower, 0., ACCURACY_HIGH))) {
        myRuntimeInfo->err << "ActivePower or Standby Power of default actuator in config is greater than zero." << std::endl;
        myRuntimeInfo->err << "At the same time the CalculateMissionLoads switch is switched on."
                           << " Manual power inputs in config should only be available if CalculateMissionLoads = 0." << std::endl;
        myRuntimeInfo->err << "--> Abort!" << std::endl;
        exit(1);
    }
    flightControl.defaultAileronActuator = EndnodeReadOnly<bool>(flightControlConfig + "/ailerons/default_actuator_architecture").read(config).value();
    flightControl.defaultSpoilerActuator = EndnodeReadOnly<bool>(flightControlConfig + "/spoilers/default_actuator_architecture").read(config).value();
    flightControl.defaultElevatorActuator = EndnodeReadOnly<bool>(flightControlConfig + "/elevators/default_actuator_architecture").read(config).value();
    flightControl.defaultRudderActuator = EndnodeReadOnly<bool>(flightControlConfig + "/rudders/default_actuator_architecture").read(config).value();
    flightControl.defaultTHSAActuator = EndnodeReadOnly<bool>(flightControlConfig + "/trimmable_horizontal_stabilizer/default_actuator_architecture").read(config).value();

    this->getAilerons(config);
    this->getSpoilers(config);
    this->getElevators(config);
    this->getRudders(config);
    this->getTHSAs(config);
    this->getFlapSystems(config);
    this->getSlatSystems(config);

    //Searching for control surfaces
    /* Primary Controls */
    // Wing with "aileron"
    for (size_t i = 0; i < data_->theWing.size(); i++) { //loop over all wings
        for (size_t j = 0; j < data_->theWing.at(i).controlSurface.size(); j++) { //loop over all control surfaces of that wing
            int number_of_control_surfaces = 0; //for symmetric aerodynamic surfaces the control surfaces are only defined for one side
            if (data_->theWing.at(i).aerodynamicSurface.is_symmetric) {
                number_of_control_surfaces = 2;
            } else {
                number_of_control_surfaces = 1;
            }
            if (data_->theWing.at(i).controlSurface.at(j).name == "aileron" || data_->theWing.at(i).controlSurface.at(j).name == "droop_aileron" ||
                data_->theWing.at(i).controlSurface.at(j).name == "Aileron" || data_->theWing.at(i).controlSurface.at(j).name == "DroopAileron") {
                for (int k = 0; k < number_of_control_surfaces; k++) {
                    Aileron.push_back(horizontalControlSurface());
                    // TODO(aigner#1#): arbitrary assumption. Research would be helpful.
                    Aileron.back().relPosOfContrSurfHingeLine = 0.1;
                    Aileron.back().deltaC_H_0 = 0.14;
                    if (flightControl.defaultAileronActuator == true) {
                        setHorizontalContrSurfGeo(data_->theWing.at(i).controlSurface.at(j),
                                                data_->theWing.at(i).aerodynamicSurface,
                                                &Aileron.back(),
                                                flightControl.ailerons.back(),
                                                data_->theWing.at(i).controlDevicePosDeflection.at(j),
                                                data_->theWing.at(i).controlDeviceNegDeflection.at(j));
                    } else {
                        setHorizontalContrSurfGeo(data_->theWing.at(i).controlSurface.at(j),
                                                data_->theWing.at(i).aerodynamicSurface,
                                                &Aileron.back(),
                                                flightControl.ailerons.at(Aileron.size() - 1),
                                                data_->theWing.at(i).controlDevicePosDeflection.at(j),
                                                data_->theWing.at(i).controlDeviceNegDeflection.at(j));
                    }
                }
            }
            if (data_->theWing.at(i).controlSurface.at(j).name == "SpoilerAir" || data_->theWing.at(i).controlSurface.at(j).name == "SpoilerGround" ||
                data_->theWing.at(i).controlSurface.at(j).name == "spoiler_air" || data_->theWing.at(i).controlSurface.at(j).name == "spoiler_ground") {
                for (int k = 0; k < number_of_control_surfaces; k++) {
                    Spoiler.push_back(horizontalControlSurface());
                    //TODO(aigner#1#): arbitrary assumption. Research would be helpful.
                    Spoiler.back().relPosOfContrSurfHingeLine = 0.0;
                    Spoiler.back().deltaC_H_0 = 0.14;
                    if (flightControl.defaultSpoilerActuator == true) {
                        setHorizontalContrSurfGeo(data_->theWing.at(i).controlSurface.at(j),
                                                data_->theWing.at(i).aerodynamicSurface,
                                                &Spoiler.back(),
                                                flightControl.spoilers.back(),
                                                data_->theWing.at(i).controlDevicePosDeflection.at(j),
                                                data_->theWing.at(i).controlDeviceNegDeflection.at(j));
                    } else {
                        setHorizontalContrSurfGeo(data_->theWing.at(i).controlSurface.at(j),
                                                data_->theWing.at(i).aerodynamicSurface,
                                                &Spoiler.back(),
                                                flightControl.spoilers.at(Spoiler.size() - 1),
                                                data_->theWing.at(i).controlDevicePosDeflection.at(j),
                                                data_->theWing.at(i).controlDeviceNegDeflection.at(j));
                    }
                }
            }
        }
        if (Spoiler.size() == 0) {
            myRuntimeInfo->warn << "Design has no TED SpoilerAir! Design revision recommended!" << std::endl;
        }
        if (Aileron.size() == 0) {
            myRuntimeInfo->warn << "Design has no TED Aileron! Design revision recommended!" << std::endl;
        }
    }
    // Horizontal tail ("Stabiliser) with with "Elevator"
    if (data_->theEmpennage.empty()) {
        myRuntimeInfo->err << "There is no empennage mounted on the aircraft!" << std::endl;
        myRuntimeInfo->err << "HINT: An empennage sizing has to be executed first." << std::endl;
        exit(1);
    }
    for (size_t i = 0; i < data_->theEmpennage.size(); i++) {
        int number_of_control_surfaces = 0; //for symmetric aerodynamic surfaces the control surfaces are only defined for one side
        if (data_->theEmpennage.at(i).aerodynamicSurface.is_symmetric) {
            number_of_control_surfaces = 2;
        } else {
            number_of_control_surfaces = 1;
        }
        if (data_->theEmpennage.at(i).controlSurface.size() == 0) {
            myRuntimeInfo->warn << data_->theEmpennage.at(i).aerodynamicSurface.name << " has no control device! Design revision recommended!" << std::endl;
        } else {
            for (size_t j = 0; j < data_->theEmpennage.at(i).controlSurface.size(); j++) {
                if (data_->theEmpennage.at(i).controlSurface.at(j).name == "Elevator" || data_->theEmpennage.at(i).controlSurface.at(j).name == "elevator") {
                    for (int k = 0; k < number_of_control_surfaces; k++) {
                        Elevator.push_back(horizontalControlSurface());
                        // TODO(aigner#1#): arbitrary assumption. Research would be helpful.
                        Elevator.back().relPosOfContrSurfHingeLine = 0.1;
                        Elevator.back().deltaC_H_0 = 0.14;
                        if (flightControl.defaultElevatorActuator == true) {
                            setHorizontalContrSurfGeo(data_->theEmpennage.at(i).controlSurface.at(j),
                                                    data_->theEmpennage.at(i).aerodynamicSurface, &Elevator.back(),
                                                    flightControl.elevators.back(),
                                                    data_->theEmpennage.at(i).controlDevicePosDeflection.at(j),
                                                    data_->theEmpennage.at(i).controlDeviceNegDeflection.at(j));
                        } else {
                            setHorizontalContrSurfGeo(data_->theEmpennage.at(i).controlSurface.at(j),
                                                    data_->theEmpennage.at(i).aerodynamicSurface, &Elevator.back(),
                                                    flightControl.elevators.at(Elevator.size() - 1),
                                                    data_->theEmpennage.at(i).controlDevicePosDeflection.at(j),
                                                    data_->theEmpennage.at(i).controlDeviceNegDeflection.at(j));
                        }
                    }
                    // THSA
                    for (int i = 0; i < flightControl.numberOfTHSAs; i++) {
                        THSA.push_back(thsaControlSurface());
                        THSA.back().deltaC_H_0 = 0.16;
                        setTHSA_AsControlSurface(data_->theEmpennage.at(i).aerodynamicSurface, &THSA.back(),
                                                flightControl.THSAs.at(THSA.size() - 1));
                        for (int m = 0; m <= THSA.back().numberOfActuators - 1; m++) {
                            THSA.back().theActuators.push_back(actuatorPFCS());
                            //Reference point of the actuator (located at the leading edge)
                            THSA.back().theActuators.back().refPoint[0] = data_->theEmpennage.at(i).aerodynamicSurface.origin[0];
                            THSA.back().theActuators.back().refPoint[1] = data_->theEmpennage.at(i).aerodynamicSurface.origin[1];
                            THSA.back().theActuators.back().refPoint[2] = data_->theEmpennage.at(i).aerodynamicSurface.origin[2];
                            THSA.back().theActuators.back().powerSource =
                                flightControl.THSAs.at(THSA.size() - 1).actuators.at(THSA.back().theActuators.size() - 1).powerSource.type;
                            THSA.back().theActuators.back().powerSourceID =
                                flightControl.THSAs.at(THSA.size() - 1).actuators.at(THSA.back().theActuators.size() - 1).powerSource.sourceID;
                            THSA.back().theActuators.back().operationMode = flightControl.THSAs.at(THSA.size() - 1).actuators.at(THSA.back().theActuators.size() - 1).operationMode;
                            THSA.back().theActuators.back().efficiency = flightControl.THSAs.at(THSA.size() - 1).actuators.at(THSA.back().theActuators.size() - 1).efficiency;
                            THSA.back().theActuators.back().power += flightControl.THSAs.at(THSA.size() - 1).actuators.at(THSA.back().theActuators.size() - 1).stdbyPower;
                            if (THSA.back().theActuators.back().operationMode == "active") {
                                THSA.back().theActuators.back().power += flightControl.THSAs.at(THSA.size() - 1).actuators.at(THSA.back().theActuators.size() - 1).activePower;
                            }
                        }
                    }
                } else if (data_->theEmpennage.at(i).controlSurface.at(j).name == "Rudder" || data_->theEmpennage.at(i).controlSurface.at(j).name == "rudder") {
                    Rudder.push_back(verticalControlSurface());
                    // TODO(aigner#1#): arbitrary assumption. Research would be helpful.
                    Rudder.back().relPosOfContrSurfHingeLine = 0.1;
                    Rudder.back().deltaC_H_0 = 0.0;//Rudder is symmetrical
                    if (flightControl.defaultRudderActuator == true) {
                        setRudderTemp(data_->theEmpennage.at(i).controlSurface.at(j),
                                      data_->theEmpennage.at(i).aerodynamicSurface, "",
                                      &Rudder.back(),
                                      flightControl.rudders.back(),
                                      data_->theEmpennage.at(i).controlDevicePosDeflection.at(j),
                                      data_->theEmpennage.at(i).controlDeviceNegDeflection.at(j));
                    } else {
                        setRudderTemp(data_->theEmpennage.at(i).controlSurface.at(j),
                                      data_->theEmpennage.at(i).aerodynamicSurface, "",
                                      &Rudder.back(),
                                      flightControl.rudders.at(Rudder.size() - 1),
                                      data_->theEmpennage.at(i).controlDevicePosDeflection.at(j),
                                      data_->theEmpennage.at(i).controlDeviceNegDeflection.at(j));
                }
                } else { //unknown control surfac
                }
            }
        }
    }
    /* High Lift */
    for (size_t i = 0; i < data_->theWing.size(); i++) {
        theHighLiftSystem.push_back(highLiftSystem(data_->theWing.at(i), data_->configuration.controlDeviceWarning,
                                                   flightControl.highLift, 1.225, 200.*0.514444, data_->data.Mass.MTOM));
    }
}

void conventionalATA27::getAilerons(const node& config) {
    std::string unit;
    std::string flightControlConfig = "/module_configuration_file/program_settings/systems_constants/flight_controls";
    //Reading the Ailerons
    if (flightControl.defaultAileronActuator == false) {
        flightControl.numberOfAilerons = config.at(flightControlConfig + "/ailerons/number_of_control_surfaces/value");
    }
    bool reachedLeftSide(false);
    for (int i(0); i < flightControl.numberOfAilerons; i++) {
        flightControl.ailerons.push_back(flightControlSystem::ControlDevice());
        flightControl.ailerons.back().name = "aileron" + num2Str(i);
        //Request whether default actuator layout or manual input should be used
        if (flightControl.defaultAileronActuator == true) {
            // NOTE (aigner#1#): Applies to symmetrical wings only. In case of asymmetry config.XML "ManualActuatorArchitecture" must be selected.
            if (i <= flightControl.numberOfAilerons / 2.)
                flightControl.ailerons.back().Side =  "Right";
            else
                flightControl.ailerons.back().Side =  "Left";
            unit = std::string(config.at(flightControlConfig + "/ailerons/default_actuator_architecture/default_deflection_speed/unit"));
            flightControl.ailerons.back().maxDefSpeed =
                config.at(flightControlConfig + "/ailerons/default_actuator_architecture/default_deflection_speed/value");
            flightControl.ailerons.back().numberOfActuators =
                config.at(flightControlConfig + "/ailerons/default_actuator_architecture/number_of_actuators_per_control_surface/value");
            for (int j(0); j < flightControl.ailerons.back().numberOfActuators; j++) {
                flightControl.ailerons.back().actuators.push_back(flightControl.theDefaultActuator);
            }
        } else {
            std::string Side = EndnodeReadOnly<std::string>(flightControlConfig + "/ailerons/manual_actuator_architecture/aileron@"
                                            + num2Str(i) + "/side").read(config).value();
            std::string Side_upper = Side;
            std::transform(Side_upper.begin(), Side_upper.end(), Side_upper.begin(), ::toupper);
            if (Side_upper == "LEFT") reachedLeftSide = true;
            if (reachedLeftSide && Side_upper == "RIGHT") {
                myRuntimeInfo->err << "Ailerons are not sorted correctly in systemsDesign_conf.xml." << std::endl;
                myRuntimeInfo->err << "For \"ManualActuatorArchitecure\", the ailerons must be sorted exactly as given in the AiXML file." << std::endl;
                myRuntimeInfo->err << "Also please specify all right-side ailerons first, then all left-side ailerons. Please edit systemsDesign_conf.xml!" << std::endl;
                myRuntimeInfo->err << "Abort program!" << std::endl;
                exit(1);
            }
            flightControl.ailerons.back().Side = Side;
            unit = std::string(config.at(flightControlConfig + "/ailerons/manual_actuator_architecture/aileron@"
                                + num2Str(i) + "/deflection_speed/unit"));
            flightControl.ailerons.back().maxDefSpeed =
                config.at(flightControlConfig + "/ailerons/manual_actuator_architecture/aileron@" + num2Str(i) + "/deflection_speed/value");
            flightControl.ailerons.back().numberOfActuators =
                config.at(flightControlConfig + "/ailerons/manual_actuator_architecture/aileron@"
                             + num2Str(i) + "/actuator_layout/number_of_actuators/value");
            if (flightControl.ailerons.back().numberOfActuators < 1) {
                myRuntimeInfo->err << "No actuators for Aileron " << i << " defined in config. Abort!" << std::endl;
                exit(1);
            }
            for (int j(0); j < flightControl.ailerons.back().numberOfActuators; j++) {
                flightControl.ailerons.back().actuators.push_back(flightControlSystem::Actuator());
                flightControl.ailerons.back().actuators.back().powerSource.type =
                    std::string(config.at(flightControlConfig + "/ailerons/manual_actuator_architecture/aileron@" +
                                             num2Str(i) + "/actuator_layout/actuator@" + num2Str(j) + "/power_source/type/value"));
                flightControl.ailerons.back().actuators.back().powerSource.sourceID =
                    config.at(flightControlConfig + "/ailerons/manual_actuator_architecture/aileron@" + num2Str(
                                     i) + "/actuator_layout/actuator@" + num2Str(j) + "/power_source/source_ID/value");
                flightControl.ailerons.back().actuators.back().powerSource.operationFactor = 1.;
                flightControl.ailerons.back().actuators.back().operationMode =
                    std::string(config.at(flightControlConfig + "/ailerons/manual_actuator_architecture/aileron@"
                                             + num2Str(i) + "/actuator_layout/actuator@" + num2Str(j) + "/operation_mode/value"));
                flightControl.ailerons.back().actuators.back().efficiency =
                    config.at(flightControlConfig + "/ailerons/manual_actuator_architecture/aileron@"
                                 + num2Str(i) + "/actuator_layout/actuator@" + num2Str(j) + "/efficiency/value");
                flightControl.ailerons.back().actuators.back().stdbyPower =
                    config.at(flightControlConfig + "/ailerons/manual_actuator_architecture/aileron@"
                                 + num2Str(i) + "/actuator_layout/actuator@" + num2Str(j) + "/standby_power/value");
                flightControl.ailerons.back().actuators.back().activePower =
                    config.at(flightControlConfig + "/ailerons/manual_actuator_architecture/aileron@"
                                 + num2Str(i) + "/actuator_layout/actuator@" + num2Str(j) + "/active_power/value");
                if (flightControl.flyMission && (!accuracyCheck(flightControl.ailerons.back().actuators.back().activePower, 0., ACCURACY_HIGH)
                                                 || !accuracyCheck(flightControl.ailerons.back().actuators.back().stdbyPower, 0., ACCURACY_HIGH))) {
                    myRuntimeInfo->err << "ActivePower or StandbyPower of the actuator " << j << " of Aileron " << i << " is greater than zero." << std::endl;
                    myRuntimeInfo->err << "At the same time the CalculateMissionLoads switch is switched on. "
                                       << "Manual power inputs in config should only be available if CalculateMissionLoads = 0." << std::endl;
                    myRuntimeInfo->err << "--> Abort!" << std::endl;
                    exit(1);
                }
            }
        }
        //unit of DeflectionSpeed
        if (unit == "deg/s") {
            //conversion in rad
            flightControl.ailerons.back().maxDefSpeed = convertUnit(DEGREE, RADIAN, flightControl.ailerons.back().maxDefSpeed);
        } else if (unit == "rad/s") {
            //no conversion
        } else {
            myRuntimeInfo->err << "The unit of DeflectionSpeed is unknown: " << unit << std::endl;
            exit(1);
        }
        data_->configuration.checkForValidity(flightControl.ailerons.back().maxDefSpeed,
                         convertUnit(DEGREE, RADIAN, 10.0),
                         convertUnit(DEGREE, RADIAN, 120.0),
                         "flightControl.ailerons.at(" + num2Str(i) + ").maxDefSpeed");
    }
}

void conventionalATA27::getSpoilers(const node& config) {
    std::string unit;
    std::string flightControlConfig = "/module_configuration_file/program_settings/systems_constants/flight_controls";
    //Reading the spoilers
    if (flightControl.defaultSpoilerActuator == false) {
        flightControl.numberOfSpoilers = config.at(flightControlConfig + "/spoilers/number_of_control_surfaces/value");
    }
    bool reachedLeftSide(false);
    for (int i(0); i < flightControl.numberOfSpoilers; i++) {
        flightControl.spoilers.push_back(flightControlSystem::ControlDevice());
        flightControl.spoilers.back().name = "Spoiler" + num2Str(i);
        //Request whether default actuator layout or manual input should be used
        if (flightControl.defaultSpoilerActuator == true) {
            // NOTE (aigner#1#): Applies to symmetrical wings only. In case of asymmetry "ManualActuatorArchitecture" must be selected in config.XML.
            if (i <= flightControl.numberOfSpoilers / 2.)
                flightControl.spoilers.back().Side =  "Right";
            else
                flightControl.spoilers.back().Side =  "Left";
            unit = std::string(config.at(flightControlConfig + "/spoilers/default_actuator_architecture/default_deflection_speed/unit"));
            flightControl.spoilers.back().maxDefSpeed =
                config.at(flightControlConfig + "/spoilers/default_actuator_architecture/default_deflection_speed/value");
            flightControl.spoilers.back().numberOfActuators =
                config.at(flightControlConfig + "/spoilers/default_actuator_architecture/number_of_actuators_per_control_surface/value");
            for (int j(0); j < flightControl.spoilers.back().numberOfActuators; j++) {
                flightControl.spoilers.back().actuators.push_back(flightControl.theDefaultActuator);
            }
        } else {
            std::string Side = config.at(flightControlConfig + "/spoilers/manual_actuator_architecture/spoiler@"
                                            + num2Str(i) + "/side");
            std::string Side_upper = Side;
            std::transform(Side_upper.begin(), Side_upper.end(), Side_upper.begin(), ::toupper);
            if (Side_upper == "LEFT") reachedLeftSide = true;
            if (reachedLeftSide && Side_upper == "RIGHT") {
                myRuntimeInfo->err << "Spoilers are not sorted correctly in systemsDesign_conf.xml." << std::endl;
                myRuntimeInfo->err << "For \"ManualActuatorArchitecure\", the spoilers must be sorted exactly as given in the AiXML file." << std::endl;
                myRuntimeInfo->err << "Also please specify all right-side spoilers first, then all left-side spoilers. Please edit systemsDesign_conf.xml!" << std::endl;
                myRuntimeInfo->err << "Abort program!" << std::endl;
                exit(1);
            }
            flightControl.spoilers.back().Side = Side;
            unit = std::string(config.at(flightControlConfig + "/spoilers/manual_actuator_architecture/spoiler@"
                                + num2Str(i) + "/deflection_speed/unit"));
            flightControl.spoilers.back().maxDefSpeed =
                config.at(flightControlConfig + "/spoilers/manual_actuator_architecture/spoiler@"
                             + num2Str(i) + "/deflection_speed/value");
            flightControl.spoilers.back().numberOfActuators =
                config.at(flightControlConfig + "/spoilers/manual_actuator_architecture/spoiler@"
                             + num2Str(i) + "/actuator_layout/number_of_actuators/value");
            if (flightControl.spoilers.back().numberOfActuators < 1) {
                myRuntimeInfo->err << "No actuators for spoilers " << i << " defined in config. Abort!" << std::endl;
                exit(1);
            }
            for (int j(0); j < flightControl.spoilers.back().numberOfActuators; j++) {
                flightControl.spoilers.back().actuators.push_back(flightControlSystem::Actuator());
                flightControl.spoilers.back().actuators.back().powerSource.type =
                    std::string(config.at(flightControlConfig + "/spoilers/manual_actuator_architecture/spoiler@" +
                                             num2Str(i) + "/actuator_layout/actuator@" + num2Str(j) + "/power_source/type/value"));
                flightControl.spoilers.back().actuators.back().powerSource.sourceID =
                    config.at(flightControlConfig + "/spoilers/manual_actuator_architecture/spoiler@"
                                 + num2Str(i) + "/actuator_layout/actuator@" + num2Str(j) + "/power_source/source_ID/value");
                flightControl.spoilers.back().actuators.back().powerSource.operationFactor = 1.;
                flightControl.spoilers.back().actuators.back().operationMode =
                    std::string(config.at(flightControlConfig + "/spoilers/manual_actuator_architecture/spoiler@" +
                                             num2Str(i) + "/actuator_layout/actuator@" + num2Str(j) + "/operation_mode/value"));
                flightControl.spoilers.back().actuators.back().efficiency =
                    config.at(flightControlConfig + "/spoilers/manual_actuator_architecture/spoiler@"
                                 + num2Str(i) + "/actuator_layout/actuator@" + num2Str(j) + "/efficiency/value");
                flightControl.spoilers.back().actuators.back().stdbyPower =
                    config.at(flightControlConfig + "/spoilers/manual_actuator_architecture/spoiler@"
                                 + num2Str(i) + "/actuator_layout/actuator@" + num2Str(j) + "/standby_power/value");
                flightControl.spoilers.back().actuators.back().activePower =
                    config.at(flightControlConfig + "/spoilers/manual_actuator_architecture/spoiler@"
                                 + num2Str(i) + "/actuator_layout/actuator@" + num2Str(
                                     j) + "/active_power/value");
                if (flightControl.flyMission && (!accuracyCheck(flightControl.spoilers.back().actuators.back().activePower, 0., ACCURACY_HIGH)
                                                 || !accuracyCheck(flightControl.spoilers.back().actuators.back().stdbyPower, 0., ACCURACY_HIGH))) {
                    myRuntimeInfo->err << "ActivePower or StandbyPower of the actuator " << j << " of Spoiler " << i << " is greater than zero." << std::endl;
                    myRuntimeInfo->err << "At the same time the CalculateMissionLoads switch is switched on."
                                       << " Manual power inputs in config should only be available if CalculateMissionLoads = 0." << std::endl;
                    myRuntimeInfo->err << "--> Abort!" << std::endl;
                    exit(1);
                }
            }
        }
        //unit of DeflectionSpeed
        if (unit == "deg/s") {
            //conversion in rad
            flightControl.spoilers.back().maxDefSpeed = convertUnit(DEGREE, RADIAN, flightControl.spoilers.back().maxDefSpeed);
        } else if (unit == "rad/s") {
            //no conversion
        } else {
            myRuntimeInfo->err << "The unit of DeflectionSpeed is unknown: " << unit << std::endl;
            exit(1);
        }
        data_->configuration.checkForValidity(flightControl.spoilers.back().maxDefSpeed,
                         convertUnit(DEGREE, RADIAN, 10.0),
                         convertUnit(DEGREE, RADIAN, 120.0),
                         "flightControl.spoilers.at(" + num2Str(i) + ").maxDefSpeed");
    }
}

void conventionalATA27::getElevators(const node& config) {
    std::string unit;
    std::string flightControlConfig = "/module_configuration_file/program_settings/systems_constants/flight_controls";
    //Reading the Elevators
    if (flightControl.defaultElevatorActuator == false) {
        flightControl.numberOfElevators = config.at(flightControlConfig + "/elevators/number_of_control_surfaces/value");
    }
    bool reachedLeftSide(false);
    for (int i(0); i < flightControl.numberOfElevators; i++) {
        flightControl.elevators.push_back(flightControlSystem::ControlDevice());
        flightControl.elevators.back().name = "Elevator" + num2Str(i);
        //Request whether default actuator layout or manual input should be used
        if (flightControl.defaultElevatorActuator == true) {
            // NOTE (aigner#1#): Applies to symmetrical wings only. In case of asymmetry "ManualActuatorArchitecture" must be selected in config.XML.
            if (i <= flightControl.numberOfElevators / 2.)
                flightControl.elevators.back().Side =  "Right";
            else
                flightControl.elevators.back().Side =  "Left";
            unit = std::string(config.at(flightControlConfig + "/elevators/default_actuator_architecture/default_deflection_speed/unit"));
            flightControl.elevators.back().maxDefSpeed =
                config.at(flightControlConfig + "/elevators/default_actuator_architecture/default_deflection_speed/value");
            flightControl.elevators.back().numberOfActuators =
                config.at(flightControlConfig + "/elevators/default_actuator_architecture/number_of_actuators_per_control_surface/value");
            for (int j(0); j < flightControl.elevators.back().numberOfActuators; j++) {
                flightControl.elevators.back().actuators.push_back(flightControl.theDefaultActuator);
            }
        } else {
            flightControl.elevators.back().Side = std::string(config.at(flightControlConfig + "/elevators/manual_actuator_architecture/elevator@"
                                                                        + num2Str(i) + "/side"));

            std::string Side = config.at(flightControlConfig + "/elevators/manual_actuator_architecture/elevator@"
                                            + num2Str(i) + "/side");
            std::string Side_upper = Side;
            std::transform(Side_upper.begin(), Side_upper.end(), Side_upper.begin(), ::toupper);
            if (Side_upper == "LEFT") reachedLeftSide = true;
            if (reachedLeftSide && Side_upper == "RIGHT") {
                myRuntimeInfo->err << "Elevators are not sorted correctly in systemsDesign_conf.xml." << std::endl;
                myRuntimeInfo->err << "For \"ManualActuatorArchitecure\", the elevators must be sorted exactly as given in the AiXML file." << std::endl;
                myRuntimeInfo->err << "Also please specify all right-side elevators first, then all left-side elevators. Please edit systemsDesign_conf.xml!" << std::endl;
                myRuntimeInfo->err << "Abort program!" << std::endl;
                exit(1);
            }
            flightControl.elevators.back().Side = Side;
            unit = std::string(config.at(flightControlConfig + "/elevators/manual_actuator_architecture/elevator@"
                                + num2Str(i) + "/deflection_speed/unit"));
            flightControl.elevators.back().maxDefSpeed =
                config.at(flightControlConfig + "/elevators/manual_actuator_architecture/elevator@"
                             + num2Str(i) + "/deflection_speed/value");
            flightControl.elevators.back().numberOfActuators =
                config.at(flightControlConfig + "/elevators/manual_actuator_architecture/elevator@"
                             + num2Str(i) + "/actuator_layout/number_of_actuators/value");
            if (flightControl.elevators.back().numberOfActuators < 1) {
                myRuntimeInfo->err << "No actuators for elevator " << i << " defined in config. Abort!" << std::endl;
                exit(1);
            }
            for (int j(0); j < flightControl.elevators.back().numberOfActuators; j++) {
                flightControl.elevators.back().actuators.push_back(flightControlSystem::Actuator());
                flightControl.elevators.back().actuators.back().powerSource.type =
                    std::string(config.at(flightControlConfig + "/elevators/manual_actuator_architecture/elevator@"
                                             + num2Str(i) + "/actuator_layout/actuator@" + num2Str(j) + "/power_source/type/value"));
                flightControl.elevators.back().actuators.back().powerSource.sourceID =
                    config.at(flightControlConfig + "/elevators/manual_actuator_architecture/elevator@"
                                 + num2Str(i) + "/actuator_layout/actuator@" + num2Str(j) + "/power_source/source_ID/value");
                flightControl.elevators.back().actuators.back().powerSource.operationFactor = 1.;
                flightControl.elevators.back().actuators.back().operationMode =
                    std::string(config.at(flightControlConfig + "/elevators/manual_actuator_architecture/elevator@"
                                             + num2Str(i) + "/actuator_layout/actuator@" + num2Str(j) + "/operation_mode/value"));
                flightControl.elevators.back().actuators.back().efficiency =
                    config.at(flightControlConfig + "/elevators/manual_actuator_architecture/elevator@"
                                 + num2Str(i) + "/actuator_layout/actuator@" + num2Str(j) + "/efficiency/value");
                flightControl.elevators.back().actuators.back().stdbyPower =
                    config.at(flightControlConfig + "/elevators/manual_actuator_architecture/elevator@"
                                 + num2Str(i) + "/actuator_layout/actuator@" + num2Str(j) + "/standby_power/value");
                flightControl.elevators.back().actuators.back().activePower =
                    config.at(flightControlConfig + "/elevators/manual_actuator_architecture/elevator@"
                                 + num2Str(i) + "/actuator_layout/actuator@" + num2Str(j) + "/active_power/value");
                if (flightControl.flyMission && (!accuracyCheck(flightControl.elevators.back().actuators.back().activePower, 0., ACCURACY_HIGH)
                                                 || !accuracyCheck(flightControl.elevators.back().actuators.back().stdbyPower, 0., ACCURACY_HIGH))) {
                    myRuntimeInfo->err << "ActivePower or StandbyPower of the actuator " << j << " of Elevator " << i << " is greater than zero." << std::endl;
                    myRuntimeInfo->err << "At the same time the CalculateMissionLoads switch is switched on."
                                       << " Manual power inputs in config should only be available if CalculateMissionLoads = 0." << std::endl;
                    myRuntimeInfo->err << "--> Abort!" << std::endl;
                    exit(1);
                }
            }
        }
        //unit of DeflectionSpeed
        if (unit == "deg/s") {
            //conversion in rad
            flightControl.elevators.back().maxDefSpeed = convertUnit(DEGREE, RADIAN, flightControl.elevators.back().maxDefSpeed);
        } else if (unit == "rad/s") {
            //no conversion
        } else {
            myRuntimeInfo->err << "The unit of DeflectionSpeed is unknown: " << unit << std::endl;
            exit(1);
        }
        data_->configuration.checkForValidity(flightControl.elevators.back().maxDefSpeed,
                         convertUnit(DEGREE, RADIAN, 10.0),
                         convertUnit(DEGREE, RADIAN, 120.0),
                         "flightControl.elevators.at(" + num2Str(i) + ").maxDefSpeed");
    }
}

void conventionalATA27::getRudders(const node& config) {
    std::string unit;
    std::string flightControlConfig = "/module_configuration_file/program_settings/systems_constants/flight_controls";
    //Reading the Rudders
    if (flightControl.defaultRudderActuator == false) {
        flightControl.numberOfRudders = config.at(flightControlConfig + "/rudders/number_of_control_surfaces/value");
    }
    for (int i(0); i < flightControl.numberOfRudders; i++) {
        flightControl.rudders.push_back(flightControlSystem::ControlDevice());
        flightControl.rudders.back().name = "Rudder" + num2Str(i);
        //Request whether default actuator layout or manual input should be used
        if (flightControl.defaultRudderActuator == true) {
            unit = std::string(config.at(flightControlConfig + "/rudders/default_actuator_architecture/default_deflection_speed/unit"));
            flightControl.rudders.back().maxDefSpeed =
                config.at(flightControlConfig + "/rudders/default_actuator_architecture/default_deflection_speed/value");
            flightControl.rudders.back().numberOfActuators =
                config.at(flightControlConfig + "/rudders/default_actuator_architecture/number_of_actuators_per_control_surface/value");
            for (int j(0); j < flightControl.rudders.back().numberOfActuators; j++) {
                flightControl.rudders.back().actuators.push_back(flightControl.theDefaultActuator);
            }
        } else {
            unit = std::string(config.at(flightControlConfig + "/rudders/manual_actuator_architecture/rudder@"
                                + num2Str(i) + "/deflection_speed/unit"));
            flightControl.rudders.back().maxDefSpeed =
                config.at(flightControlConfig + "/rudders/manual_actuator_architecture/rudder@"
                             + num2Str(i) + "/deflection_speed/value");
            flightControl.rudders.back().numberOfActuators =
                config.at(flightControlConfig + "/rudders/manual_actuator_architecture/rudder@"
                             + num2Str(i) + "/actuator_layout/number_of_actuators/value");
            if (flightControl.rudders.back().numberOfActuators < 1) {
                myRuntimeInfo->err << "No actuators for Rudder " << i << " defined in config. Abort!" << std::endl;
                exit(1);
            }
            for (int j(0); j < flightControl.rudders.back().numberOfActuators; j++) {
                flightControl.rudders.back().actuators.push_back(flightControlSystem::Actuator());
                flightControl.rudders.back().actuators.back().powerSource.type =
                    std::string(config.at(flightControlConfig + "/rudders/manual_actuator_architecture/rudder@" +
                                             num2Str(i) + "/actuator_layout/actuator@" + num2Str(j) + "/power_source/type/value"));
                flightControl.rudders.back().actuators.back().powerSource.sourceID =
                    config.at(flightControlConfig + "/rudders/manual_actuator_architecture/rudder@"
                                 + num2Str(i) + "/actuator_layout/actuator@" + num2Str(j) + "/power_source/source_ID/value");
                flightControl.rudders.back().actuators.back().powerSource.operationFactor = 1.;
                flightControl.rudders.back().actuators.back().operationMode =
                    std::string(config.at(flightControlConfig + "/rudders/manual_actuator_architecture/rudder@" +
                                             num2Str(i) + "/actuator_layout/actuator@" + num2Str(j) + "/operation_mode/value"));
                flightControl.rudders.back().actuators.back().efficiency =
                    config.at(flightControlConfig + "/rudders/manual_actuator_architecture/rudder@"
                                 + num2Str(i) + "/actuator_layout/actuator@" + num2Str(j) + "/efficiency/value");
                flightControl.rudders.back().actuators.back().stdbyPower =
                    config.at(flightControlConfig + "/rudders/manual_actuator_architecture/rudder@"
                                 + num2Str(i) + "/actuator_layout/actuator@" + num2Str(j) + "/standby_power/value");
                flightControl.rudders.back().actuators.back().activePower =
                    config.at(flightControlConfig + "/rudders/manual_actuator_architecture/rudder@"
                                 + num2Str(i) + "/actuator_layout/actuator@" + num2Str(j) + "/active_power/value");
                if (flightControl.flyMission && (!accuracyCheck(flightControl.rudders.back().actuators.back().activePower, 0., ACCURACY_HIGH)
                                                 || !accuracyCheck(flightControl.rudders.back().actuators.back().stdbyPower, 0., ACCURACY_HIGH))) {
                    myRuntimeInfo->err << "ActivePower or StandbyPower of the actuator " << j << " of Rudder " << i << " is greater than zero." << std::endl;
                    myRuntimeInfo->err << "At the same time the CalculateMissionLoads switch is switched on."
                                       << " Manual power inputs in config should only be available if CalculateMissionLoads = 0." << std::endl;
                    myRuntimeInfo->err << "--> Abort!" << std::endl;
                    exit(1);
                }
            }
        }
        //unit of DeflectionSpeed
        if (unit == "deg/s") {
            //conversion in rad
            flightControl.rudders.back().maxDefSpeed = convertUnit(DEGREE, RADIAN, flightControl.rudders.back().maxDefSpeed);
        } else if (unit == "rad/s") {
            //no conversion
        } else {
            myRuntimeInfo->err << "The unit of DeflectionSpeed is unknown: " << unit << std::endl;
            exit(1);
        }
        data_->configuration.checkForValidity(flightControl.rudders.back().maxDefSpeed, convertUnit(DEGREE, RADIAN, 10.0),
                         convertUnit(DEGREE, RADIAN, 120.0), "flightControl.rudders.at(" + num2Str(i) + ").maxDefSpeed");
    }
}

void conventionalATA27::getTHSAs(const node& config) {
    std::string unit;
    std::string flightControlConfig = "/module_configuration_file/program_settings/systems_constants/flight_controls";
    //Reading the THSAs
    flightControl.numberOfTHSAs = config.at(flightControlConfig + "/trimmable_horizontal_stabilizer/number_of_trimmable_horizontal_stabilizers/value");
    for (int i(0); i < flightControl.numberOfTHSAs; i++) {
        flightControl.THSAs.push_back(flightControlSystem::ControlDevice());
        flightControl.THSAs.back().name = "THSA" + num2Str(i);
        //Request whether default actuator layout or manual input should be used
        if (flightControl.defaultTHSAActuator == true) {
            unit = std::string(
                config.at(flightControlConfig + "/trimmable_horizontal_stabilizer/default_actuator_architecture/default_deflection_speed/unit"));
            flightControl.THSAs.back().maxDefSpeed =
                config.at(flightControlConfig + "/trimmable_horizontal_stabilizer/default_actuator_architecture/default_deflection_speed/value");
            flightControl.THSAs.back().numberOfActuators =
                config.at(flightControlConfig + "/trimmable_horizontal_stabilizer/default_actuator_architecture/number_of_actuators_per_control_surface/value");
            for (int j(0); j < flightControl.THSAs.back().numberOfActuators; j++) {
                flightControl.THSAs.back().actuators.push_back(flightControl.theDefaultActuator);
            }
        } else {
            unit = std::string(config.at(flightControlConfig + "/trimmable_horizontal_stabilizer/manual_actuator_architecture/trimmable_horizontal_stabilizer@"
                                + num2Str(i) + "/deflection_speed/unit"));
            flightControl.THSAs.back().maxDefSpeed =
                config.at(flightControlConfig + "/trimmable_horizontal_stabilizer/manual_actuator_architecture/trimmable_horizontal_stabilizer@"
                             + num2Str(i) + "/deflection_speed/value");
            flightControl.THSAs.back().numberOfActuators =
                config.at(flightControlConfig + "/trimmable_horizontal_stabilizer/manual_actuator_architecture/trimmable_horizontal_stabilizer@"
                             + num2Str(i) + "/actuator_layout/number_of_actuators/value");
            if (flightControl.THSAs.back().numberOfActuators < 1) {
                myRuntimeInfo->err << "No actuators for THSA " << i << " defined in config. Abort!" << std::endl;
                exit(1);
            }
            for (int j(0); j < flightControl.THSAs.back().numberOfActuators; j++) {
                flightControl.THSAs.back().actuators.push_back(flightControlSystem::Actuator());
                flightControl.THSAs.back().actuators.back().powerSource.type =
                    std::string(config.at(flightControlConfig + "/trimmable_horizontal_stabilizer/manual_actuator_architecture/trimmable_horizontal_stabilizer@"
                                             + num2Str(i) + "/actuator_layout/actuator@" + num2Str(j) + "/power_source/type/value"));
                flightControl.THSAs.back().actuators.back().powerSource.sourceID =
                    config.at(flightControlConfig + "/trimmable_horizontal_stabilizer/manual_actuator_architecture/trimmable_horizontal_stabilizer@"
                                 + num2Str(i) + "/actuator_layout/actuator@" + num2Str(j) + "/power_source/source_ID/value");
                flightControl.THSAs.back().actuators.back().powerSource.operationFactor = 1.;
                flightControl.THSAs.back().actuators.back().operationMode =
                    std::string(config.at(flightControlConfig + "/trimmable_horizontal_stabilizer/manual_actuator_architecture/trimmable_horizontal_stabilizer@"
                                             + num2Str(i) + "/actuator_layout/actuator@" + num2Str(j) + "/operation_mode/value"));
                flightControl.THSAs.back().actuators.back().efficiency =
                    config.at(flightControlConfig + "/trimmable_horizontal_stabilizer/manual_actuator_architecture/trimmable_horizontal_stabilizer@"
                                 + num2Str(i) + "/actuator_layout/actuator@" + num2Str(j) + "/efficiency/value");
                flightControl.THSAs.back().actuators.back().stdbyPower =
                    config.at(flightControlConfig + "/trimmable_horizontal_stabilizer/manual_actuator_architecture/trimmable_horizontal_stabilizer@"
                                 + num2Str(i) + "/actuator_layout/actuator@" + num2Str(j) + "/standby_power/value");
                flightControl.THSAs.back().actuators.back().activePower =
                    config.at(flightControlConfig + "/trimmable_horizontal_stabilizer/manual_actuator_architecture/trimmable_horizontal_stabilizer@"
                                 + num2Str(i) + "/actuator_layout/actuator@" + num2Str(j) + "/active_power/value");
                if (flightControl.flyMission && (!accuracyCheck(flightControl.THSAs.back().actuators.back().activePower, 0., ACCURACY_HIGH)
                                                 || !accuracyCheck(flightControl.THSAs.back().actuators.back().stdbyPower, 0., ACCURACY_HIGH))) {
                    myRuntimeInfo->err << "ActivePower or StandbyPower of the actuator " << j << " of THSA " << i << " is greater than zero." << std::endl;
                    myRuntimeInfo->err << "At the same time the CalculateMissionLoads switch is switched on. "
                                       << "Manual power inputs in config should only be available if CalculateMissionLoads = 0." << std::endl;
                    myRuntimeInfo->err << "--> Abort!" << std::endl;
                    exit(1);
                }
            }
        }
        //unit of DeflectionSpeed
        if (unit == "deg/s") {
            //conversion in rad
            flightControl.THSAs.back().maxDefSpeed = convertUnit(DEGREE, RADIAN, flightControl.THSAs.back().maxDefSpeed);
        } else if (unit == "rad/s") {
            //no conversion
        } else {
            myRuntimeInfo->err << "The unit of DeflectionSpeed is unknown: " << unit << std::endl;
            exit(1);
        }
        data_->configuration.checkForValidity(flightControl.THSAs.back().maxDefSpeed, convertUnit(DEGREE, RADIAN, 0.0),
                         convertUnit(DEGREE, RADIAN, 20.0), "flightControl.THSAs.at(" + num2Str(i) + ").maxDefSpeed");
    }
}

void conventionalATA27::getFlapSystems(const node& config) {
    //Flap Systems
    std::string flightControlConfig = "/module_configuration_file/program_settings/systems_constants/flight_controls";
    flightControl.numberOfFlapSystems = config.at(flightControlConfig + "/high_lift_system/trailing_edge/number_of_devices/value");
    for (int i(0); i < flightControl.numberOfFlapSystems; i++) {
        flightControl.highLift.push_back(flightControlSystem::HighLiftSystem());
        flightControl.highLift.back().type = "TE_Device";
        flightControl.highLift.back().numberOfPCUMotors =
            config.at(flightControlConfig + "/high_lift_system/trailing_edge/trailing_edge_device@"
                         + num2Str(i) + "/power_control_unit/number_of_power_control_unit_motors/value");
        if (flightControl.highLift.back().numberOfPCUMotors < 1) {
            myRuntimeInfo->err << "No PCUs defined for Flap in config. Abort!" << std::endl;
            exit(1);
        }
        for (int j(0); j < flightControl.highLift.back().numberOfPCUMotors; j++) {
            flightControl.highLift.back().PCU_Motors.push_back(flightControlSystem::Actuator());
            flightControl.highLift.back().PCU_Motors.back().powerSource.type =
                std::string(config.at(flightControlConfig + "/high_lift_system/trailing_edge/trailing_edge_device@" +
                                         num2Str(i) + "/power_control_unit/power_control_unit_motor@" + num2Str(j) + "/power_source/type/value"));
            flightControl.highLift.back().PCU_Motors.back().powerSource.sourceID =
                config.at(flightControlConfig + "/high_lift_system/trailing_edge/trailing_edge_device@" + num2Str(
                                 i) + "/power_control_unit/power_control_unit_motor@" + num2Str(j) + "/power_source/source_ID/value");
            flightControl.highLift.back().PCU_Motors.back().powerSource.operationFactor = 1.;
            flightControl.highLift.back().PCU_Motors.back().operationMode =
                std::string(config.at(flightControlConfig + "/high_lift_system/trailing_edge/trailing_edge_device@" + num2Str(
                                             i) + "/power_control_unit/power_control_unit_motor@" + num2Str(j) + "/operation_mode/value"));
            flightControl.highLift.back().PCU_Motors.back().stdbyPower =
                config.at(flightControlConfig + "/high_lift_system/trailing_edge/trailing_edge_device@" + num2Str(
                                 i) + "/power_control_unit/power_control_unit_motor@" + num2Str(j) + "/standby_power/value");
            flightControl.highLift.back().PCU_Motors.back().activePower =
                config.at(flightControlConfig + "/high_lift_system/trailing_edge/trailing_edge_device@" + num2Str(
                                 i) + "/power_control_unit/power_control_unit_motor@" + num2Str(j) + "/active_power/value");
            if (flightControl.flyMission && (!accuracyCheck(flightControl.highLift.back().PCU_Motors.back().activePower, 0., ACCURACY_HIGH)
                                             || !accuracyCheck(flightControl.highLift.back().PCU_Motors.back().stdbyPower, 0., ACCURACY_HIGH))) {
                myRuntimeInfo->err << "ActivePower or StandbyPower of the PCU " << j << " of FlapSystem " << i << " is greater than zero." << std::endl;
                myRuntimeInfo->err << "At the same time the CalculateMissionLoads switch is switched on."
                                   << " Manual power inputs in config should only be available if CalculateMissionLoads = 0." << std::endl;
                myRuntimeInfo->err << "--> Abort!" << std::endl;
                exit(1);
            }
        }
    }
}

void conventionalATA27::getSlatSystems(const node& config) {
    //Slat Systems
    std::string flightControlConfig = "/module_configuration_file/program_settings/systems_constants/flight_controls";
    flightControl.numberOfSlatSystems = config.at(flightControlConfig + "/high_lift_system/leading_edge/number_of_devices/value");
    for (int i(0); i < flightControl.numberOfSlatSystems; i++) {
        flightControl.highLift.push_back(flightControlSystem::HighLiftSystem());
        flightControl.highLift.back().type = "LE_Device";
        flightControl.highLift.back().numberOfPCUMotors =
            config.at(flightControlConfig + "/high_lift_system/leading_edge/leading_edge_device@"
                         + num2Str(i) + "/power_control_unit/number_of_power_control_unit_motors/value");
        if (flightControl.highLift.back().numberOfPCUMotors < 1) {
            myRuntimeInfo->err << "No PCU defined for Slat in config. Abort!" << std::endl;
            exit(1);
        }
        for (int j(0); j < flightControl.highLift.back().numberOfPCUMotors; j++) {
            flightControl.highLift.back().PCU_Motors.push_back(flightControlSystem::Actuator());
            flightControl.highLift.back().PCU_Motors.back().powerSource.type =
                std::string(config.at(flightControlConfig + "/high_lift_system/leading_edge/leading_edge_device@"
                                         + num2Str(i) + "/power_control_unit/power_control_unit_motor@" + num2Str(j) + "/power_source/type/value"));
            flightControl.highLift.back().PCU_Motors.back().powerSource.sourceID =
                config.at(flightControlConfig + "/high_lift_system/leading_edge/leading_edge_device@"
                             + num2Str(i) + "/power_control_unit/power_control_unit_motor@" + num2Str(j) + "/power_source/source_ID/value");
            flightControl.highLift.back().PCU_Motors.back().powerSource.operationFactor = 1.;
            flightControl.highLift.back().PCU_Motors.back().operationMode =
                std::string(config.at(flightControlConfig + "/high_lift_system/leading_edge/leading_edge_device@"
                                         + num2Str(i) + "/power_control_unit/power_control_unit_motor@" + num2Str(j) + "/operation_mode/value"));
            flightControl.highLift.back().PCU_Motors.back().stdbyPower =
                config.at(flightControlConfig + "/high_lift_system/leading_edge/leading_edge_device@"
                             + num2Str(i) + "/power_control_unit/power_control_unit_motor@" + num2Str(j) + "/standby_power/value");
            flightControl.highLift.back().PCU_Motors.back().activePower =
                config.at(flightControlConfig + "/high_lift_system/leading_edge/leading_edge_device@"
                             + num2Str(i) + "/power_control_unit/power_control_unit_motor@" + num2Str(j) + "/active_power/value");
            if (flightControl.flyMission && (!accuracyCheck(flightControl.highLift.back().PCU_Motors.back().activePower, 0., ACCURACY_HIGH)
                                             || !accuracyCheck(flightControl.highLift.back().PCU_Motors.back().stdbyPower, 0., ACCURACY_HIGH))) {
                myRuntimeInfo->err << "ActivePower or StandbyPower of the PCU " << j << " of SlatSystem " << i << " is greater than zero." << std::endl;
                myRuntimeInfo->err << "At the same time the CalculateMissionLoads switch is switched on."
                                   << "Manual power inputs in config should only be available if CalculateMissionLoads = 0." << std::endl;
                myRuntimeInfo->err << "--> Abort!" << std::endl;
                exit(1);
            }
        }
    }
}

/* Updating system specific parameters in aircraft xml file */
void conventionalATA27::updateXML() {
    updateMass("flight_controls", Mass, ATA);
}

/* Check User Input */
void conventionalATA27::checkUserInput() {
    int numberOfAilerons_tmp(0);
        int numberOfSpoilers_tmp(0);
        for (size_t i(0); i < data_->theWing.size(); i++) {
            for (size_t j(0); j < data_->theWing.at(i).controlSurface.size(); j++) {
                if (data_->theWing.at(i).controlSurface.at(j).name == "Aileron" || data_->theWing.at(i).controlSurface.at(j).name == "DroopAileron"
                    || data_->theWing.at(i).controlSurface.at(j).name == "aileron" || data_->theWing.at(i).controlSurface.at(j).name == "droop_aileron") {
                    numberOfAilerons_tmp += 1;
                } else if (data_->theWing.at(i).controlSurface.at(j).name == "SpoilerAir" || data_->theWing.at(i).controlSurface.at(j).name == "SpoilerGround"
                           || data_->theWing.at(i).controlSurface.at(j).name == "spoiler_air" || data_->theWing.at(i).controlSurface.at(j).name == "spoiler_ground") {
                    numberOfSpoilers_tmp += 1;
                } else if (data_->theWing.at(i).controlSurface.at(j).name == "Flap" || data_->theWing.at(i).controlSurface.at(j).name == "flap"
                        || data_->theWing.at(i).controlSurface.at(j).name == "SlottedFlap" || data_->theWing.at(i).controlSurface.at(j).name == "slotted_flap"
                        || data_->theWing.at(i).controlSurface.at(j).name == "ADHF"
                        || data_->theWing.at(i).controlSurface.at(j).name == "Fowler" || data_->theWing.at(i).controlSurface.at(j).name == "fowler"
                        || data_->theWing.at(i).controlSurface.at(j).name == "flap_fowler"
                        || data_->theWing.at(i).controlSurface.at(j).name == "DoubleFowler" || data_->theWing.at(i).controlSurface.at(j).name == "double_fowler"
                        || data_->theWing.at(i).controlSurface.at(j).name == "TripleFowler" || data_->theWing.at(i).controlSurface.at(j).name == "triple_fowler"
                        || data_->theWing.at(i).controlSurface.at(j).name == "Special" || data_->theWing.at(i).controlSurface.at(j).name == "special"
                        || data_->theWing.at(i).controlSurface.at(j).name == "morphingTrailingEdge" || data_->theWing.at(i).controlSurface.at(j).name == "morphing_trailing_edge") {
                            //copied from legacy systems design, no method implemented yet for these TE-Devices
                } else if (data_->theWing.at(i).controlSurface.at(j).name == "Slat" || data_->theWing.at(i).controlSurface.at(j).name == "slat"
                        || data_->theWing.at(i).controlSurface.at(j).name == "Krueger" || data_->theWing.at(i).controlSurface.at(j).name == "krueger"
                        || data_->theWing.at(i).controlSurface.at(j).name == "DroopNose" || data_->theWing.at(i).controlSurface.at(j).name == "droop_nose"
                        || data_->theWing.at(i).controlSurface.at(j).name == "morphingDroopNose" || data_->theWing.at(i).controlSurface.at(j).name == "morphing_droop_nose"
                        || data_->theWing.at(i).controlSurface.at(j).name == "Special" || data_->theWing.at(i).controlSurface.at(j).name == "special") {
                    //copied from legacy systems design, no method implemented yet for these LE-Devices
                } else {
                    myRuntimeInfo->err << "Unknown control device type \"" << data_->theWing.at(i).controlSurface.at(j).name << "\"" << std::endl;
                    exit(1);
                }
            }
        }
        //Check of HighLift Systems
    //    if (flightControl.numberOfFlapSystems != numberOfFlapSystems_tmp) {
    //        myRuntimeInfo->err << "Number of Flaps in config ("
    //                           << flightControl.numberOfFlapSystems
    //                           << ") does not correspond to the number in Aircraft-XML ("
    //                           << numberOfFlapSystems_tmp << "). Please adjust!" << std::endl;
    //        myRuntimeInfo->err << "Abort!" << std::endl;
    //        exit(1);
    //    }
    //    if (flightControl.numberOfSlatSystems != numberOfSlatSystems_tmp) {
    //        myRuntimeInfo->err << "Number of Slats in config ("
    //                           << flightControl.numberOfSlatSystems
    //                           << ") does not correspond to the number in Aircraft-XML ("
    //                           << numberOfSlatSystems_tmp << "). Please adjust!" << std::endl;
    //        myRuntimeInfo->err << "Abort!" << std::endl;
    //        exit(1);
    //    }
        int highLiftSystemDevices(flightControl.numberOfFlapSystems + flightControl.numberOfSlatSystems);//Anzahl der Hochauftriebskomponenten
        if (highLiftSystemDevices > 0) {
            for (size_t i(0); i < flightControl.highLift.size(); i++) {
                //Check of the energy sources for the HighLift motors
                for (int j(0); j < flightControl.highLift.at(i).numberOfPCUMotors; j++) {
                    std::string powerSource_tmp = flightControl.highLift.at(i).PCU_Motors.at(j).powerSource.type;
                    int sourceID_tmp = flightControl.highLift.at(i).PCU_Motors.at(j).powerSource.sourceID;
                    if (powerSource_tmp == "Hydraulic") {
                        if (sourceID_tmp > data_->configuration.Hydraulic.numberOfhydraulicCircuits) {
                            myRuntimeInfo->err << "The PowerSourceID of the High Lift motor exceeds the number of hydraulic circuits. Adjust in config!" << std::endl;
                            myRuntimeInfo->err << "Abort!" << std::endl;
                            exit(1);
                        }
                    } else if (powerSource_tmp == "Electric") {
                        if (sourceID_tmp > data_->configuration.Electric.numberOfElectricCircuits) {
                            myRuntimeInfo->err << "The PowerSourceID of the High Lift motor exceeds the number of electrical circuits. Adjust in config!" << std::endl;
                            myRuntimeInfo->err << "Abort!" << std::endl;
                            exit(1);
                        }
                    } else {
                        myRuntimeInfo->err << "unknown PowerSource \"" << powerSource_tmp <<  "\" for High Lift Motor. Adjust in config!" << std::endl;
                        myRuntimeInfo->err << "Abort!" << std::endl;
                        exit(1);
                    }
                }
            }
        }
        //Check of Ailerons
        // number of ailerons from acxml multiplied by two since each side is defined in config but not in acxml
        if (flightControl.defaultAileronActuator == false && flightControl.numberOfAilerons != (numberOfAilerons_tmp * 2)) {
            myRuntimeInfo->err << "Number of Ailerons in config ("
                            << flightControl.numberOfAilerons
                            << ") does not correspond to the number in Aircraft-XML ("
                            << numberOfAilerons_tmp << "). Please adjust!" << std::endl;
            myRuntimeInfo->err << "Abort!" << std::endl;
            exit(1);
        }
        //Check of energy sources for Aileron actuators
        for (int i(0); i < flightControl.numberOfAilerons; i++) {
            for (int j(0); j < flightControl.ailerons.at(i).numberOfActuators; j++) {
                std::string powerSource_tmp = flightControl.ailerons.at(i).actuators.at(j).powerSource.type;
                int sourceID_tmp = flightControl.ailerons.at(i).actuators.at(j).powerSource.sourceID;
                if (powerSource_tmp == "Hydraulic") {
                    if (sourceID_tmp > data_->configuration.Hydraulic.numberOfhydraulicCircuits) {
                        myRuntimeInfo->err << "The PowerSourceID of the Aileron actuator exceeds the number of hydraulic circuits. Adjust in config!" << std::endl;
                        myRuntimeInfo->err << "Abort!" << std::endl;
                        exit(1);
                    }
                } else if (powerSource_tmp == "Electric") {
                    if (sourceID_tmp > data_->configuration.Electric.numberOfElectricCircuits) {
                        myRuntimeInfo->err << "The PowerSourceID of the Aileron actuator exceeds the number of electrical circuits. Adjust in config!" << std::endl;
                        myRuntimeInfo->err << "Abort!" << std::endl;
                        exit(1);
                    }
                } else {
                    myRuntimeInfo->err << "unknown PowerSource \"" << powerSource_tmp <<  "\" for Aileron actuator. Adjust in config!" << std::endl;
                    myRuntimeInfo->err << "Abort!" << std::endl;
                    exit(1);
                }
            }
        }
        //Check of Spoiler
        // number of spoilers from acxml multiplied by two since each side is defined in config but not in acxml
        if (flightControl.defaultSpoilerActuator == false && flightControl.numberOfSpoilers != (numberOfSpoilers_tmp * 2)) {
            myRuntimeInfo->err << "Number of spoilers in config ("
                            << flightControl.numberOfSpoilers
                            << ") does not correspond to the number in Aircraft-XML ("
                            << numberOfSpoilers_tmp << "). Please adjust!" << std::endl;
            myRuntimeInfo->err << "Abort!" << std::endl;
            exit(1);
        }
        //Check of energy sources for spoiler actuators
        for (int i(0); i < flightControl.numberOfSpoilers; i++) {
            for (int j(0); j < flightControl.spoilers.at(i).numberOfActuators; j++) {
                std::string powerSource_tmp = flightControl.spoilers.at(i).actuators.at(j).powerSource.type;
                int sourceID_tmp = flightControl.spoilers.at(i).actuators.at(j).powerSource.sourceID;
                if (powerSource_tmp == "Hydraulic") {
                    if (sourceID_tmp > data_->configuration.Hydraulic.numberOfhydraulicCircuits) {
                        myRuntimeInfo->err << "The PowerSourceID of the spoiler actuator exceeds the number of hydraulic circuits. Adjust in config!" << std::endl;
                        myRuntimeInfo->err << "Abort!" << std::endl;
                        exit(1);
                    }
                } else if (powerSource_tmp == "Electric") {
                    if (sourceID_tmp > data_->configuration.Electric.numberOfElectricCircuits) {
                        myRuntimeInfo->err << "The PowerSourceID of the spoiler actuator exceeds the number of electrical circuits. Adjust in config!" << std::endl;
                        myRuntimeInfo->err << "Abort!" << std::endl;
                        exit(1);
                    }
                } else {
                    myRuntimeInfo->err << "unknown PowerSource \"" << powerSource_tmp <<  "\" for spoiler actuator. Adjust in config!" << std::endl;
                    myRuntimeInfo->err << "Abort!" << std::endl;
                    exit(1);
                }
            }
        }
        //Check of Elevator
        int numberOfElevators_tmp = 0;
        for (const auto& empennage : data_->theEmpennage) {
            numberOfElevators_tmp += std::count_if(empennage.controlSurface.begin(), empennage.controlSurface.end(),
                [](const auto& controlSurface) {
                    return controlSurface.name == "Elevator" || controlSurface.name == "elevator";
                });
        }
        // number of elevators from acxml multiplied by two since each side is defined in config but not in acxml
        if (flightControl.defaultElevatorActuator == false && flightControl.numberOfElevators != (numberOfElevators_tmp * 2)) {
            myRuntimeInfo->err << "Number of elevators in config ("
                            << flightControl.numberOfElevators
                            << ") does not correspond to the number in Aircraft-XML ("
                            << numberOfElevators_tmp
                            << "). Please adjust!" << std::endl;
            myRuntimeInfo->err << "Abort!" << std::endl;
            exit(1);
        }
        //Check of energy sources for elevator actuators
        for (int i(0); i < flightControl.numberOfElevators; i++) {
            for (int j(0); j < flightControl.elevators.at(i).numberOfActuators; j++) {
                std::string powerSource_tmp = flightControl.elevators.at(i).actuators.at(j).powerSource.type;
                int sourceID_tmp = flightControl.elevators.at(i).actuators.at(j).powerSource.sourceID;
                if (powerSource_tmp == "Hydraulic") {
                    if (sourceID_tmp > data_->configuration.Hydraulic.numberOfhydraulicCircuits) {
                        myRuntimeInfo->err << "The PowerSourceID of the Elevator actuator exceeds the number of hydraulic circuits. Adjust in config!" << std::endl;
                        myRuntimeInfo->err << "Abort!" << std::endl;
                        exit(1);
                    }
                } else if (powerSource_tmp == "Electric") {
                    if (sourceID_tmp > data_->configuration.Electric.numberOfElectricCircuits) {
                        myRuntimeInfo->err << "The PowerSourceID of the elevator actuator exceeds the number of electrical circuits. Adjust in config!" << std::endl;
                        myRuntimeInfo->err << "Abort!" << std::endl;
                        exit(1);
                    }
                } else {
                    myRuntimeInfo->err << "unknown PowerSource \"" << powerSource_tmp <<  "\" for Elevator Actuator. Adjust in config!" << std::endl;
                    myRuntimeInfo->err << "Abort!" << std::endl;
                    exit(1);
                }
            }
        }
        //Check of energy sources for THS actuators
        for (int i(0); i < flightControl.numberOfTHSAs; i++) {
            for (int j(0); j < flightControl.THSAs.at(i).numberOfActuators; j++) {
                std::string powerSource_tmp = flightControl.THSAs.at(i).actuators.at(j).powerSource.type;
                int sourceID_tmp = flightControl.THSAs.at(i).actuators.at(j).powerSource.sourceID;
                if (powerSource_tmp == "Hydraulic") {
                    if (sourceID_tmp > data_->configuration.Hydraulic.numberOfhydraulicCircuits) {
                        myRuntimeInfo->err << "The PowerSourceID of the THS actuator exceeds the number of hydraulic circuits. Adjust in config!" << std::endl;
                        myRuntimeInfo->err << "Abort!" << std::endl;
                        exit(1);
                    }
                } else if (powerSource_tmp == "Electric") {
                    if (sourceID_tmp > data_->configuration.Electric.numberOfElectricCircuits) {
                        myRuntimeInfo->err << "The PowerSourceID of the THS actuator exceeds the number of electrical circuits. Adjust in config!" << std::endl;
                        myRuntimeInfo->err << "Abort!" << std::endl;
                        exit(1);
                    }
                } else {
                    myRuntimeInfo->err << "unknown PowerSource \"" << powerSource_tmp <<  "\" for THS actuator. Adjust in config!" << std::endl;
                    myRuntimeInfo->err << "Abort!" << std::endl;
                    exit(1);
                }
            }
        }
        //Check of Rudder
        int numberOfRudders_tmp(0);
        for (size_t i(0); i < data_->theEmpennage.size(); i++) {
            for (size_t j(0); j < data_->theEmpennage.at(i).controlSurface.size(); j++) {
                if (data_->theEmpennage.at(i).controlSurface.at(j).name == "Rudder" || data_->theEmpennage.at(i).controlSurface.at(j).name == "rudder")
                    numberOfRudders_tmp += 1;
            }
        }
        if (flightControl.defaultRudderActuator == false && flightControl.numberOfRudders != numberOfRudders_tmp) {
            myRuntimeInfo->err << "Number of rudders in config ("
                            << flightControl.numberOfRudders << ") does not correspond to the number in Aircraft-XML ("
                            << numberOfRudders_tmp << "). Please adjust!" << std::endl;
            myRuntimeInfo->err << "Abort!" << std::endl;
            exit(1);
        }
        //Check of energy sources for Rudder actuators
        for (int i(0); i < flightControl.numberOfRudders; i++) {
            for (int j(0); j < flightControl.rudders.at(i).numberOfActuators; j++) {
                std::string powerSource_tmp = flightControl.rudders.at(i).actuators.at(j).powerSource.type;
                int sourceID_tmp = flightControl.rudders.at(i).actuators.at(j).powerSource.sourceID;
                if (powerSource_tmp == "Hydraulic") {
                    if (sourceID_tmp > data_->configuration.Hydraulic.numberOfhydraulicCircuits) {
                        myRuntimeInfo->err << "The PowerSourceID of the Rudder actuator exceeds the number of hydraulic circuits. Adjust in config!" << std::endl;
                        myRuntimeInfo->err << "Abort!" << std::endl;
                        exit(1);
                    }
                } else if (powerSource_tmp == "Electric") {
                    if (sourceID_tmp > data_->configuration.Electric.numberOfElectricCircuits) {
                        myRuntimeInfo->err << "The PowerSourceID of the Rudder actuator exceeds the number of electrical circuits. Adjust in config!" << std::endl;
                        myRuntimeInfo->err << "Abort!" << std::endl;
                        exit(1);
                    }
                } else {
                    myRuntimeInfo->err << "unknown PowerSource \"" << powerSource_tmp <<  "\" for Rudder actuator. Adjust in config!" << std::endl;
                    myRuntimeInfo->err << "Abort!" << std::endl;
                    exit(1);
                }
            }
        }
}
/* Calculation methods */

void conventionalATA27::calculateSystemsPower(const powerProfile& allPreviousSystemsBleed,
        const std::vector<powerProfile>& allPreviousSystemsHydraulic,
        const std::vector<powerProfile>& allPreviousSystemsElectric,
        const powerProfile& allPreviousSystemsHeat) {
    /* Define unused function parameters to avoid linter warnings */
    UNUSED(allPreviousSystemsBleed);
    UNUSED(allPreviousSystemsHydraulic);
    UNUSED(allPreviousSystemsElectric);
    UNUSED(allPreviousSystemsHeat);
    /* Calculate Systems Power */
    initializePowerProfile();
    if (Operating == true) {
        getPower();
    }
    calculateMaxMissionPower();
}
void conventionalATA27::calculateSinkMass(const powerProfile& sourceBleed,
        const std::vector<powerProfile>& sourceElectric,
        const std::vector<powerProfile>& sourceHydraulic) {
    /* Define unused function parameters to avoid linter warnings */
    UNUSED(sourceBleed);
    UNUSED(sourceElectric);
    UNUSED(sourceHydraulic);
    /* Calculate Systems Mass */
    Mass = scaling_factor * getMass();
    refPoint = getRefPoint();
}
void conventionalATA27::getPower() {
    double timeHighLift             = flightControl.deploymentHighLiftTime;
    std::vector<systemsIOData::Wing> thisWing = data_->theWing;
    double wingLoading              = data_->data.Mass.WtoS;
    double VMO                      = data_->data.Performance.VMO;
    double MMO                      = data_->data.Performance.MMO;
    // TODO(aigner#2#): Mantis ID 0000206: Ueberpruefen: Stimmen die Formeln mit der die Design Power berechnet wird?
    /* PFCS and SFCS together in one function setDesignPower_FCS(...) */
    setDesignPower_FCS(wingLoading, VMO, MMO, &Electric, &Hydraulic);
    if (flightControl.flyMission == true) {
        getDetailedMissionPower(&bleedAir, &Hydraulic, &Electric, &heatLoad, timeHighLift, thisWing);
    } else {
        getAverageMissionPower(&bleedAir, &Hydraulic, &Electric, &heatLoad, timeHighLift, thisWing);
    }
}

void conventionalATA27::getDetailedMissionPower(powerProfile *myBleedAir, std::vector<powerProfile> *myHydraulic, std::vector<powerProfile> *myElectric,
        powerProfile *myHeatLoad, double timeHighLift, std::vector<systemsIOData::Wing> thisWing) {
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        //Set baseLoads to 0
        myBleedAir->missionPower.at(i).baseLoad     = 0.0;
        for (size_t n(0); n < myHydraulic->size(); n++) {
            myHydraulic->at(n).missionPower.at(i).baseLoad    = 0.0;
        }
        for (size_t n(0); n < myElectric->size(); n++) {
            myElectric->at(n).missionPower.at(i).baseLoad     = 0.0;
        }
        myHeatLoad->missionPower.at(i).baseLoad     = 0.0;
        // NOTE (aigner#1#): Leakage and power to actuate
        //Determination of the hinge moments and power at the individual mission points
        // Aileron
        for (size_t j = 0; j < Aileron.size() ; j++) {
            double hingeMoment = fabs(getHingeMomentPFCS(Aileron.at(j), data_->mission.Steps.at(i).Point.dynamicPressure, data_->mission.Steps.at(i).Point.Mach));
            double requiredPower(hingeMoment * Aileron.at(j).maxDefSpeed);//Required power for deflection
            std::vector<double> actuatorPower;//Necessary actuator power for deflection
            std::vector<double> heatLoad_tmp;//Heat load due to actuator efficiency
            int activeCounter(0);//Counter for actuators in "Active Mode". The required power is split among them.
            for (size_t k(0); k < Aileron.at(j).theActuators.size(); k++) {
                if (Aileron.at(j).theActuators.at(k).operationMode == "active") {
                    activeCounter++;
                }
            }
            for (size_t k(0); k < Aileron.at(j).theActuators.size(); k++) {
                actuatorPower.push_back(0.);
                heatLoad_tmp.push_back(0);
                if (Aileron.at(j).theActuators.at(k).operationMode == "active") {
                    //The power of all active actuators is divided by the number of active actuators (load sharing).
                    actuatorPower.back() += requiredPower / (Aileron.at(j).theActuators.at(k).efficiency * activeCounter);
                    heatLoad_tmp.back() += requiredPower * (1 / Aileron.at(j).theActuators.at(k).efficiency - 1) / activeCounter;
                }
                for (size_t n(0); n < myHydraulic->size(); n++) {
                    if (Aileron.at(j).theActuators.at(k).powerSource == "Hydraulic" && Aileron.at(j).theActuators.at(k).powerSourceID == static_cast<int>(n + 1)) {
                        myHydraulic->at(n).missionPower.at(i).baseLoad += actuatorPower.back();
                    }
                }
                for (size_t n(0); n < myElectric->size(); n++) {
                    if (Aileron.at(j).theActuators.at(k).powerSource == "Electric" && Aileron.at(j).theActuators.at(k).powerSourceID == static_cast<int>(n + 1)) {
                        myElectric->at(n).missionPower.at(i).baseLoad += actuatorPower.back();
                        myHeatLoad->missionPower.at(i).baseLoad += heatLoad_tmp.back();
                    }
                }
            }
        }
        //Spoiler
        for (size_t j = 0; j < Spoiler.size(); j++) {
            double hingeMoment = fabs(getHingeMomentPFCS(Spoiler.at(j), data_->mission.Steps.at(i).Point.dynamicPressure, data_->mission.Steps.at(i).Point.Mach));
            double requiredPower(hingeMoment * Spoiler.at(j).maxDefSpeed);//Required power for deflection
            std::vector<double> actuatorPower;//Necessary actuator power for deflection
            std::vector<double> heatLoad_tmp;//Heat load due to actuator efficiency
            int activeCounter(0);//Counter for actuators in "Active Mode". The required power is split among them.
            for (size_t k(0); k < Spoiler.at(j).theActuators.size(); k++) {
                if (Spoiler.at(j).theActuators.at(k).operationMode == "active") {
                    activeCounter++;
                }
            }
            for (size_t k(0); k < Spoiler.at(j).theActuators.size(); k++) {
                actuatorPower.push_back(0.);
                heatLoad_tmp.push_back(0);
                if (Spoiler.at(j).theActuators.at(k).operationMode == "active") {
                    //The power of all active actuators is divided by the number of active actuators (load sharing).
                    actuatorPower.back() += requiredPower / (Spoiler.at(j).theActuators.at(k).efficiency * activeCounter);
                    heatLoad_tmp.back() += requiredPower * (1 / Spoiler.at(j).theActuators.at(k).efficiency - 1) / activeCounter;
                }
                for (size_t n(0); n < myHydraulic->size(); n++) {
                    if (Spoiler.at(j).theActuators.at(k).powerSource == "Hydraulic" && Spoiler.at(j).theActuators.at(k).powerSourceID == static_cast<int>(n + 1)) {
                        myHydraulic->at(n).missionPower.at(i).baseLoad += actuatorPower.back();
                    }
                }
                for (size_t n(0); n < myElectric->size(); n++) {
                    if (Spoiler.at(j).theActuators.at(k).powerSource == "Electric" && Spoiler.at(j).theActuators.at(k).powerSourceID == static_cast<int>(n + 1)) {
                        myElectric->at(n).missionPower.at(i).baseLoad += actuatorPower.back();
                        myHeatLoad->missionPower.at(i).baseLoad += heatLoad_tmp.back();
                    }
                }
            }
        }
        // Elevator
        for (size_t j = 0; j < Elevator.size(); j++) {
            double hingeMoment = fabs(getHingeMomentPFCS(Elevator.at(j), data_->mission.Steps.at(i).Point.dynamicPressure, data_->mission.Steps.at(i).Point.Mach));
            double requiredPower(hingeMoment * Elevator.at(j).maxDefSpeed);//Required power for deflection
            std::vector<double> actuatorPower;//Necessary actuator power for deflection
            std::vector<double> heatLoad_tmp;//Heat load due to actuator efficiency
            int activeCounter(0);//Counter for actuators in "Active Mode". The required power is split among them.
            for (size_t k(0); k < Elevator.at(j).theActuators.size(); k++) {
                if (Elevator.at(j).theActuators.at(k).operationMode == "active") {
                    activeCounter++;
                }
            }
            for (size_t k(0); k < Elevator.at(j).theActuators.size(); k++) {
                actuatorPower.push_back(0.);
                heatLoad_tmp.push_back(0);
                if (Elevator.at(j).theActuators.at(k).operationMode == "active") {
                    //The power of all active actuators is divided by the number of active actuators (load sharing).
                    actuatorPower.back() += requiredPower / (Elevator.at(j).theActuators.at(k).efficiency * activeCounter);
                    heatLoad_tmp.back() += requiredPower * (1 / Elevator.at(j).theActuators.at(k).efficiency - 1) / activeCounter;
                }
                for (size_t n(0); n < myHydraulic->size(); n++) {
                    if (Elevator.at(j).theActuators.at(k).powerSource == "Hydraulic" && Elevator.at(j).theActuators.at(k).powerSourceID == static_cast<int>(n + 1)) {
                        myHydraulic->at(n).missionPower.at(i).baseLoad += actuatorPower.back();
                    }
                }
                for (size_t n(0); n < myElectric->size(); n++) {
                    if (Elevator.at(j).theActuators.at(k).powerSource == "Electric" && Elevator.at(j).theActuators.at(k).powerSourceID == static_cast<int>(n + 1)) {
                        myElectric->at(n).missionPower.at(i).baseLoad += actuatorPower.back();
                        myHeatLoad->missionPower.at(i).baseLoad += heatLoad_tmp.back();
                    }
                }
            }
        }
        // Rudder
        for (size_t j = 0; j < Rudder.size(); j++) {
            double hingeMoment = fabs(getHingeMomentPFCS(Rudder.at(j), data_->mission.Steps.at(i).Point.dynamicPressure, data_->mission.Steps.at(i).Point.Mach));
            double requiredPower(hingeMoment * Rudder.at(j).maxDefSpeed);//Required power for deflection
            std::vector<double> actuatorPower;//Necessary actuator power for deflection
            std::vector<double> heatLoad_tmp;//Heat load due to actuator efficiency
            int activeCounter(0);//Counter for actuators in "Active Mode". The required power is split among them.
            for (size_t k(0); k < Rudder.at(j).theActuators.size(); k++) {
                if (Rudder.at(j).theActuators.at(k).operationMode == "active") {
                    activeCounter++;
                }
            }
            for (size_t k(0); k < Rudder.at(j).theActuators.size(); k++) {
                actuatorPower.push_back(0.);
                heatLoad_tmp.push_back(0);
                if (Rudder.at(j).theActuators.at(k).operationMode == "active") {
                    //The power of all active actuators is divided by the number of active actuators (load sharing).
                    actuatorPower.back() += requiredPower / (Rudder.at(j).theActuators.at(k).efficiency * activeCounter);
                    heatLoad_tmp.back() += requiredPower * (1 / Rudder.at(j).theActuators.at(k).efficiency - 1) / activeCounter;
                }
                for (size_t n(0); n < myHydraulic->size(); n++) {
                    if (Rudder.at(j).theActuators.at(k).powerSource == "Hydraulic" && Rudder.at(j).theActuators.at(k).powerSourceID == static_cast<int>(n + 1)) {
                        myHydraulic->at(n).missionPower.at(i).baseLoad += actuatorPower.back();
                    }
                }
                for (size_t n(0); n < myElectric->size(); n++) {
                    if (Rudder.at(j).theActuators.at(k).powerSource == "Electric" && Rudder.at(j).theActuators.at(k).powerSourceID == static_cast<int>(n + 1)) {
                        myElectric->at(n).missionPower.at(i).baseLoad += actuatorPower.back();
                        myHeatLoad->missionPower.at(i).baseLoad += heatLoad_tmp.back();
                    }
                }
            }
        }
        // THSA
        for (size_t j = 0; j < THSA.size(); j++) {
            double hingeMoment = fabs(getHingeMomentPFCS(THSA.at(j), data_->mission.Steps.at(i).Point.dynamicPressure, data_->mission.Steps.at(i).Point.Mach));
            double requiredPower(hingeMoment * THSA.at(j).maxDefSpeed);;//Required power for deflection
            std::vector<double> actuatorPower;//Necessary actuator power for deflection
            std::vector<double> heatLoad_tmp;//Heat load due to actuator efficiency
            int activeCounter(0);//Counter for actuators in "Active Mode". The required power is split among them.
            for (size_t k(0); k < THSA.at(j).theActuators.size(); k++) {
                if (THSA.at(j).theActuators.at(k).operationMode == "active") {
                    activeCounter++;
                }
            }
            for (size_t k(0); k < THSA.at(j).theActuators.size(); k++) {
                actuatorPower.push_back(0.);
                heatLoad_tmp.push_back(0);
                if (THSA.at(j).theActuators.at(k).operationMode == "active") {
                    //The power of all active actuators is divided by the number of active actuators (load sharing).
                    actuatorPower.back() += requiredPower / (THSA.at(j).theActuators.at(k).efficiency * activeCounter);
                    heatLoad_tmp.back() += requiredPower * (1 / THSA.at(j).theActuators.at(k).efficiency - 1) / activeCounter;
                }
                for (size_t n(0); n < myHydraulic->size(); n++) {
                    if (THSA.at(j).theActuators.at(k).powerSource == "Hydraulic" && THSA.at(j).theActuators.at(k).powerSourceID == static_cast<int>(n + 1)) {
                        myHydraulic->at(n).missionPower.at(i).baseLoad += actuatorPower.back();
                    }
                }
                for (size_t n(0); n < myElectric->size(); n++) {
                    if (THSA.at(j).theActuators.at(k).powerSource == "Electric" && THSA.at(j).theActuators.at(k).powerSourceID == static_cast<int>(n + 1)) {
                        myElectric->at(n).missionPower.at(i).baseLoad += actuatorPower.back();
                        myHeatLoad->missionPower.at(i).baseLoad += heatLoad_tmp.back();
                    }
                }
            }
        }
        //Constant base load for mission segment (mainly leakage)
        for (size_t n(0); n < myHydraulic->size(); n++) {
            myHydraulic->at(n).missionPower.at(i).baseLoad *= 0.45;
        }
        for (size_t n(0); n < myElectric->size(); n++) {
            myElectric->at(n).missionPower.at(i).baseLoad *= 0.45;
        }
        myHeatLoad->missionPower.at(i).baseLoad *= 0.45;
        // TODO(risse#1#): Mantis ID 0000207: ueberpruefen! im cruise nicht so viele ausschlaege wie in anderen phasen. der verlauf ueber die hingemomente klappt so nicht
        // dies hier ist ein hack: hoehenabhaengigkeit + annahme haelfte an ausschlaegen
        for (size_t n(0); n < myHydraulic->size(); n++) {
            if (data_->mission.Steps.at(i).missionPhase == "cruise") {
                myHydraulic->at(n).missionPower.at(i).baseLoad *= 0.5 * (50000. - data_->mission.Steps.at(i).Point.Altitude) / 50000.;
            } else if (data_->mission.Steps.at(i).missionPhase == "approach"
                       || data_->mission.Steps.at(i).missionPhase == "departure") {
                myHydraulic->at(n).missionPower.at(i).baseLoad *= (50000. - data_->mission.Steps.at(i).Point.Altitude) / 50000.;
            }
        }
        for (size_t n(0); n < myElectric->size(); n++) {
            if (data_->mission.Steps.at(i).missionPhase == "cruise") {
                myElectric->at(n).missionPower.at(i).baseLoad *= 0.5 * (50000. - data_->mission.Steps.at(i).Point.Altitude) / 50000.;
            } else if (data_->mission.Steps.at(i).missionPhase == "approach"
                       || data_->mission.Steps.at(i).missionPhase == "departure") {
                myElectric->at(n).missionPower.at(i).baseLoad *= (50000. - data_->mission.Steps.at(i).Point.Altitude) / 50000.;
            }
        }
        if (data_->mission.Steps.at(i).missionPhase == "cruise") {
            myHeatLoad->missionPower.at(i).baseLoad *= 0.5 * (50000. - data_->mission.Steps.at(i).Point.Altitude) / 50000.;
        } else if (data_->mission.Steps.at(i).missionPhase == "approach"
                   || data_->mission.Steps.at(i).missionPhase == "departure") {
            myHeatLoad->missionPower.at(i).baseLoad *= (50000. - data_->mission.Steps.at(i).Point.Altitude) / 50000.;
        }
        //HL-SYSTEM
        for (size_t j = 0; j < thisWing.size(); j++) {
            //Convert speed to m/s
            double Speed = 0.51444 * data_->mission.Steps.at(i).Point.TAS; //[m/s]
            highLiftSystem myHighLiftSystem(highLiftSystem(thisWing.at(j), data_->configuration.controlDeviceWarning, flightControl.highLift,
                                                           data_->mission.Steps.at(i).Point.density, Speed, data_->data.Mass.MTOM));
            for (size_t n(0); n < myHighLiftSystem.Systems.size(); n++) {
                //Base Load of the HL system (e.g. leakage or constant electric power supply)
                for (size_t p(0); p < myHighLiftSystem.Systems.at(n).PCU.size(); p++) {
                    for (size_t m(0); m < myHighLiftSystem.Systems.at(n).PCU.at(p).PCU_Motor.size(); m++) {
                        for (size_t k(0); k < data_->configuration.Hydraulic.circuits.size(); k++) {
                            if (myHighLiftSystem.Systems.at(n).PCU.at(p).PCU_Motor.at(m).powerSource == "Hydraulic"
                                    && myHighLiftSystem.Systems.at(n).PCU.at(p).PCU_Motor.at(m).powerSourceID == static_cast<int>(k + 1)) {
                                myHydraulic->at(k).missionPower.at(i).baseLoad += myHighLiftSystem.Systems.at(n).PCU.at(p).PCU_Motor.at(m).power;
                            }
                        }
                        for (size_t k(0); k < data_->configuration.Electric.circuits.size(); k++) {
                            if (myHighLiftSystem.Systems.at(n).PCU.at(p).PCU_Motor.at(m).powerSource == "Electric"
                                    && myHighLiftSystem.Systems.at(n).PCU.at(p).PCU_Motor.at(m).powerSourceID == static_cast<int>(k + 1)) {
                                myElectric->at(k).missionPower.at(i).baseLoad += myHighLiftSystem.Systems.at(n).PCU.at(p).PCU_Motor.at(m).power;
                            }
                        }
                    }
                }
                //Peak Loads of the HL-System
                if (i > 0 && i < data_->mission.Steps.size() - 1) {
                    if ((data_->mission.Steps.at(i).configuration == "clean" && data_->mission.Steps.at(i - 1).configuration == "climb") ||
                            (data_->mission.Steps.at(i).configuration == "approach" && data_->mission.Steps.at(i - 1).configuration == "clean")) {
                        for (size_t p(0); p < myHighLiftSystem.Systems.at(n).PCU.size(); p++) {
                            //Even distribution of the power to the PCU motors
                            double HighLiftPower(0.);
                            int activeCounter(0);//Counter for actuators in "Active Mode". The required power is split among them.
                            for (size_t m(0); m < myHighLiftSystem.Systems.at(n).PCU.at(p).PCU_Motor.size(); m++) {
                                if (myHighLiftSystem.Systems.at(n).PCU.at(p).PCU_Motor.at(m).operationMode == "active")
                                    activeCounter++;
                            }
                            for (size_t m(0); m < myHighLiftSystem.Systems.at(n).PCU.at(p).PCU_Motor.size(); m++) {
                                if (myHighLiftSystem.Systems.at(n).PCU.at(p).PCU_Motor.at(m).operationMode == "active") {
                                    HighLiftPower = (myHighLiftSystem.Systems.at(n).requiredPowerPCU + myHighLiftSystem.Systems.at(n).maximumPowerPCU) / activeCounter;
                                }
                                for (size_t k(0); k < data_->configuration.Hydraulic.circuits.size(); k++) {
                                    if (myHighLiftSystem.Systems.at(n).PCU.at(p).PCU_Motor.at(m).powerSource == "Hydraulic"
                                            && myHighLiftSystem.Systems.at(n).PCU.at(p).PCU_Motor.at(m).powerSourceID == static_cast<int>(k + 1)
                                            && HighLiftPower > 0.) {
                                        myHydraulic->at(k).missionPower.at(i).peakLoad.push_back(HighLiftPower);
                                        myHydraulic->at(k).missionPower.at(i).timePeak.push_back(timeHighLift);
                                    }
                                }
                                for (size_t k(0); k < data_->configuration.Electric.circuits.size(); k++) {
                                    if (myHighLiftSystem.Systems.at(n).PCU.at(p).PCU_Motor.at(m).powerSource == "Electric"
                                            && myHighLiftSystem.Systems.at(n).PCU.at(p).PCU_Motor.at(m).powerSourceID == static_cast<int>(k + 1)
                                            && HighLiftPower > 0.) {
                                        myElectric->at(k).missionPower.at(i).peakLoad.push_back(HighLiftPower);
                                        myElectric->at(k).missionPower.at(i).timePeak.push_back(timeHighLift);
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    //Clean up peaks
    for (size_t n(0); n < myHydraulic->size(); n++) {
        cleanPowerPeaks(&myHydraulic->at(n));
    }
    for (size_t n(0); n < myElectric->size(); n++) {
        cleanPowerPeaks(&myElectric->at(n));
    }
    cleanPowerPeaks(myHeatLoad);
    cleanPowerPeaks(myBleedAir);
}

void conventionalATA27::getAverageMissionPower(powerProfile *myBleedAir, std::vector<powerProfile> *myHydraulic, std::vector<powerProfile> *myElectric,
        powerProfile *myHeatLoad, double timeHighLift, std::vector<systemsIOData::Wing> thisWing) {
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        //set baseLoads to 0
        myBleedAir->missionPower.at(i).baseLoad     = 0.0;
        for (size_t n(0); n < myHydraulic->size(); n++) {
            myHydraulic->at(n).missionPower.at(i).baseLoad    = 0.0;
        }
        for (size_t n(0); n < myElectric->size(); n++) {
            myElectric->at(n).missionPower.at(i).baseLoad     = 0.0;
        }
        myHeatLoad->missionPower.at(i).baseLoad     = 0.0;
        // Aileron
        for (size_t j = 0; j < Aileron.size(); j++) {
            for (size_t k(0); k < Aileron.at(j).theActuators.size(); k++) {
                for (size_t n(0); n < myHydraulic->size(); n++) {
                    if (Aileron.at(j).theActuators.at(k).powerSource == "Hydraulic" && Aileron.at(j).theActuators.at(k).powerSourceID == static_cast<int>(n + 1)) {
                        myHydraulic->at(n).missionPower.at(i).baseLoad += Aileron.at(j).theActuators.at(k).power / Aileron.at(j).theActuators.at(k).efficiency;
                    }
                }
                for (size_t n(0); n < myElectric->size(); n++) {
                    if (Aileron.at(j).theActuators.at(k).powerSource == "Electric" && Aileron.at(j).theActuators.at(k).powerSourceID == static_cast<int>(n + 1)) {
                        myElectric->at(n).missionPower.at(i).baseLoad += Aileron.at(j).theActuators.at(k).power / Aileron.at(j).theActuators.at(k).efficiency;
                        myHeatLoad->missionPower.at(i).baseLoad += Aileron.at(j).theActuators.at(k).power * (1 / Aileron.at(j).theActuators.at(k).efficiency - 1);
                    }
                }
            }
        }
        //Spoiler
        for (size_t j = 0; j < Spoiler.size(); j++) {
//            myRuntimeInfo->out << std::endl;
//            myRuntimeInfo->out << "Spoiler No." << j+1 << std::endl;
            for (size_t k(0); k < Spoiler.at(j).theActuators.size(); k++) {
                for (size_t n(0); n < myHydraulic->size(); n++) {
                    if (Spoiler.at(j).theActuators.at(k).powerSource == "Hydraulic" && Spoiler.at(j).theActuators.at(k).powerSourceID == static_cast<int>(n + 1)) {
                        myHydraulic->at(n).missionPower.at(i).baseLoad += Spoiler.at(j).theActuators.at(k).power / Spoiler.at(j).theActuators.at(k).efficiency;
                    }
                }
                for (size_t n(0); n < myElectric->size(); n++) {
                    if (Spoiler.at(j).theActuators.at(k).powerSource == "Electric" && Spoiler.at(j).theActuators.at(k).powerSourceID == static_cast<int>(n + 1)) {
                        myElectric->at(n).missionPower.at(i).baseLoad += Spoiler.at(j).theActuators.at(k).power / Spoiler.at(j).theActuators.at(k).efficiency;
                        myHeatLoad->missionPower.at(i).baseLoad += Spoiler.at(j).theActuators.at(k).power * (1 / Spoiler.at(j).theActuators.at(k).efficiency - 1);
                    }
                }
            }
        }
        // Elevator
        for (size_t j = 0; j < Elevator.size(); j++) {
            for (size_t k(0); k < Elevator.at(j).theActuators.size(); k++) {
                for (size_t n(0); n < myHydraulic->size(); n++) {
                    if (Elevator.at(j).theActuators.at(k).powerSource == "Hydraulic" && Elevator.at(j).theActuators.at(k).powerSourceID == static_cast<int>(n + 1)) {
                        myHydraulic->at(n).missionPower.at(i).baseLoad += Elevator.at(j).theActuators.at(k).power / Elevator.at(j).theActuators.at(k).efficiency;
                    }
                }
                for (size_t n(0); n < myElectric->size(); n++) {
                    if (Elevator.at(j).theActuators.at(k).powerSource == "Electric" && Elevator.at(j).theActuators.at(k).powerSourceID == static_cast<int>(n + 1)) {
                        myElectric->at(n).missionPower.at(i).baseLoad += Elevator.at(j).theActuators.at(k).power / Elevator.at(j).theActuators.at(k).efficiency;
                        myHeatLoad->missionPower.at(i).baseLoad += Elevator.at(j).theActuators.at(k).power * (1 / Elevator.at(j).theActuators.at(k).efficiency - 1);
                    }
                }
            }
        }
        // Rudder
        for (size_t j = 0; j < Rudder.size(); j++) {
            for (size_t k(0); k < Rudder.at(j).theActuators.size(); k++) {
                for (size_t n(0); n < myHydraulic->size(); n++) {
                    if (Rudder.at(j).theActuators.at(k).powerSource == "Hydraulic" && Rudder.at(j).theActuators.at(k).powerSourceID == static_cast<int>(n + 1)) {
                        myHydraulic->at(n).missionPower.at(i).baseLoad += Rudder.at(j).theActuators.at(k).power / Rudder.at(j).theActuators.at(k).efficiency;
                    }
                }
                for (size_t n(0); n < myElectric->size(); n++) {
                    if (Rudder.at(j).theActuators.at(k).powerSource == "Electric" && Rudder.at(j).theActuators.at(k).powerSourceID == static_cast<int>(n + 1)) {
                        myElectric->at(n).missionPower.at(i).baseLoad += Rudder.at(j).theActuators.at(k).power / Rudder.at(j).theActuators.at(k).efficiency;
                        myHeatLoad->missionPower.at(i).baseLoad += Rudder.at(j).theActuators.at(k).power * (1 / Rudder.at(j).theActuators.at(k).efficiency - 1);
                    }
                }
            }
        }
        // THSA
        for (size_t j = 0; j < THSA.size(); j++) {
            for (size_t k(0); k < THSA.at(j).theActuators.size(); k++) {
                for (size_t n(0); n < myHydraulic->size(); n++) {
                    if (THSA.at(j).theActuators.at(k).powerSource == "Hydraulic" && THSA.at(j).theActuators.at(k).powerSourceID == static_cast<int>(n + 1)) {
                        myHydraulic->at(n).missionPower.at(i).baseLoad += THSA.at(j).theActuators.at(k).power / THSA.at(j).theActuators.at(k).efficiency;
                    }
                }
                for (size_t n(0); n < myElectric->size(); n++) {
                    if (THSA.at(j).theActuators.at(k).powerSource == "Electric" && THSA.at(j).theActuators.at(k).powerSourceID == static_cast<int>(n + 1)) {
                        myElectric->at(n).missionPower.at(i).baseLoad += THSA.at(j).theActuators.at(k).power / THSA.at(j).theActuators.at(k).efficiency;
                        myHeatLoad->missionPower.at(i).baseLoad += THSA.at(j).theActuators.at(k).power * (1 / THSA.at(j).theActuators.at(k).efficiency - 1);
                    }
                }
            }
        }
        //HL-SYSTEM
        for (size_t j = 0; j < thisWing.size(); j++) {
            //Convert speed to m/s
            double Speed = 0.51444 * data_->mission.Steps.at(i).Point.TAS;
            highLiftSystem myHighLiftSystem(highLiftSystem(thisWing.at(j), data_->configuration.controlDeviceWarning, flightControl.highLift,
                                                           data_->mission.Steps.at(i).Point.density, Speed, data_->data.Mass.MTOM));
            for (size_t n(0); n < myHighLiftSystem.Systems.size(); n++) {
                //Base Load of the HL Systems (e.g. leakage or constant electric power supply)
                for (size_t p(0); p < myHighLiftSystem.Systems.at(n).PCU.size(); p++) {
                    for (size_t m(0); m < myHighLiftSystem.Systems.at(n).PCU.at(p).PCU_Motor.size(); m++) {
                        for (size_t k(0); k < data_->configuration.Hydraulic.circuits.size(); k++) {
                            if (myHighLiftSystem.Systems.at(n).PCU.at(p).PCU_Motor.at(m).powerSource == "Hydraulic"
                                    && myHighLiftSystem.Systems.at(n).PCU.at(p).PCU_Motor.at(m).powerSourceID == static_cast<int>(k + 1)) {
                                myHydraulic->at(k).missionPower.at(i).baseLoad += myHighLiftSystem.Systems.at(n).PCU.at(p).PCU_Motor.at(m).power;
                            }
                        }
                        for (size_t k(0); k < data_->configuration.Electric.circuits.size(); k++) {
                            if (myHighLiftSystem.Systems.at(n).PCU.at(p).PCU_Motor.at(m).powerSource == "Electric"
                                    && myHighLiftSystem.Systems.at(n).PCU.at(p).PCU_Motor.at(m).powerSourceID == static_cast<int>(k + 1)) {
                                myElectric->at(k).missionPower.at(i).baseLoad += myHighLiftSystem.Systems.at(n).PCU.at(p).PCU_Motor.at(m).power;
                            }
                        }
                    }
                }
                //Peak Loads of the HL-System
                if (i > 0 && i < data_->mission.Steps.size() - 1) {
                    if ((data_->mission.Steps.at(i).configuration == "clean" && data_->mission.Steps.at(i - 1).configuration == "climb") ||
                            (data_->mission.Steps.at(i).configuration == "approach" && data_->mission.Steps.at(i - 1).configuration == "clean")) {
                        for (size_t p(0); p < myHighLiftSystem.Systems.at(n).PCU.size(); p++) {
                            //Even distribution of the power amon the PCu motors
                            double HighLiftPower(0.);
                            int activeCounter(0);//Counter for actuators in "Active Mode". The required power is split among them.
                            for (size_t m(0); m < myHighLiftSystem.Systems.at(n).PCU.at(p).PCU_Motor.size(); m++) {
                                if (myHighLiftSystem.Systems.at(n).PCU.at(p).PCU_Motor.at(m).operationMode == "active")
                                    activeCounter++;
                            }
                            for (size_t m(0); m < myHighLiftSystem.Systems.at(n).PCU.at(p).PCU_Motor.size(); m++) {
                                if (myHighLiftSystem.Systems.at(n).PCU.at(p).PCU_Motor.at(m).operationMode == "active") {
                                    HighLiftPower = (myHighLiftSystem.Systems.at(n).requiredPowerPCU + myHighLiftSystem.Systems.at(n).maximumPowerPCU) / activeCounter;
                                }
                                for (size_t k(0); k < data_->configuration.Hydraulic.circuits.size(); k++) {
                                    if (myHighLiftSystem.Systems.at(n).PCU.at(p).PCU_Motor.at(m).powerSource == "Hydraulic"
                                            && myHighLiftSystem.Systems.at(n).PCU.at(p).PCU_Motor.at(m).powerSourceID == static_cast<int>(k + 1)
                                            && HighLiftPower > 0.) {
                                        myHydraulic->at(k).missionPower.at(i).peakLoad.push_back(HighLiftPower);
                                        myHydraulic->at(k).missionPower.at(i).timePeak.push_back(timeHighLift);
                                    }
                                }
                                for (size_t k(0); k < data_->configuration.Electric.circuits.size(); k++) {
                                    if (myHighLiftSystem.Systems.at(n).PCU.at(p).PCU_Motor.at(m).powerSource == "Electric"
                                            && myHighLiftSystem.Systems.at(n).PCU.at(p).PCU_Motor.at(m).powerSourceID == static_cast<int>(k + 1)
                                            && HighLiftPower > 0.) {
                                        myElectric->at(k).missionPower.at(i).peakLoad.push_back(HighLiftPower);
                                        myElectric->at(k).missionPower.at(i).timePeak.push_back(timeHighLift);
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    //Clean up peaks
    for (size_t n(0); n < myHydraulic->size(); n++) {
        cleanPowerPeaks(&myHydraulic->at(n));
    }
    for (size_t n(0); n < myElectric->size(); n++) {
        cleanPowerPeaks(&myElectric->at(n));
    }
    cleanPowerPeaks(myHeatLoad);
    cleanPowerPeaks(myBleedAir);
}

double conventionalATA27::getMass() {
    /* Calibration Values derived from Airbus Data (Peter Okt 2014) s. Rev 3682*/
    double roll_ElecFC_MassFaktor(1.);
    double yaw_ElecFC_MassFaktor(1.);
    double pitch_ElecFC_MassFaktor(1.);
    double THSA_ElecFC_MassFaktor(1.);
    double Flaps_ElecFC_MassFaktor(1.);
    double Spoilers_ElecFC_MassFaktor(1.);
    double Slats_ElecFC_MassFaktor(1.);
    if (flightControl.electricalFCS) {
        roll_ElecFC_MassFaktor = flightControl.roll_EFCSWeightFactor;
        yaw_ElecFC_MassFaktor = flightControl.yaw_EFCSWeightFactor;
        pitch_ElecFC_MassFaktor = flightControl.pitch_EFCSWeightFactor;
        THSA_ElecFC_MassFaktor = flightControl.THSA_EFCSWeightFactor;
        Flaps_ElecFC_MassFaktor = flightControl.HL_TE_EFCSWeightFactor;
        Spoilers_ElecFC_MassFaktor = flightControl.Spoiler_EFCSWeightFactor;
        Slats_ElecFC_MassFaktor = flightControl.HL_LE_EFCSWeightFactor;
    }
    /* Primary Flight Controls */
    commonInstFCMass = 0.;
    primaryFCSMass = 0.;
    rollFCMass = 0.;
    yawFCMass = 0.;
    pitchFCMass = 0.;
    rollFCMass  = roll_ElecFC_MassFaktor  * getAileronControlMass(Aileron);
    yawFCMass   = yaw_ElecFC_MassFaktor   * getRudderControlMass(Rudder);//no specific Rudder method available
    pitchFCMass = pitch_ElecFC_MassFaktor * getElevatorControlMass(Elevator);
    primaryFCSMass = rollFCMass + yawFCMass + pitchFCMass;
    commonInstFCMass += flightControl.commonInst_FCS_Factor * primaryFCSMass;
    /* Secondary Flight Controls */
    secondaryFCSMass = 0.;
    THSA_FCMass = 0.;
    FlapsFCMass = 0.;
    SpoilersFCMass = 0.;
    SlatsFCMass = 0.;
    THSA_FCMass = THSA_ElecFC_MassFaktor * getTHSAControlMass();
    for (size_t i = 0; i < theHighLiftSystem.size(); i++) {
        for (size_t j = 0; j < theHighLiftSystem.at(i).Systems.size(); j++) {
            if (theHighLiftSystem.at(i).Systems.at(j).Edge == "LE") {
                SlatsFCMass += Slats_ElecFC_MassFaktor * theHighLiftSystem.at(i).Systems.at(j).systemMass;
            } else if (theHighLiftSystem.at(i).Systems.at(j).Edge == "TE") {
                FlapsFCMass += Flaps_ElecFC_MassFaktor * theHighLiftSystem.at(i).Systems.at(j).systemMass;
            } else {
                myRuntimeInfo->err << "Error in Mass estimation of flight Control Systems, Highlift system has no Edge assigend. Abort!" << std::endl;
                exit(1);
            }
        }
    }
    SpoilersFCMass = Spoilers_ElecFC_MassFaktor * getSpoilerControlMass(Spoiler);
    secondaryFCSMass = THSA_FCMass + FlapsFCMass + SpoilersFCMass + SlatsFCMass;
    /* Common Installation of Flight Controls */
    commonInstFCMass += flightControl.commonInst_FCS_Factor * secondaryFCSMass;
    myRuntimeInfo->debug << std::endl;
    myRuntimeInfo->debug << "****** PFCS Masses ******" << std::endl;
    myRuntimeInfo->debug << "aileron:          " << rollFCMass       << " (incl. factor for electric FC: " << roll_ElecFC_MassFaktor << ")" << std::endl;
    myRuntimeInfo->debug << "rudder:           " << yawFCMass        << " (incl. factor for electric FC: " << yaw_ElecFC_MassFaktor << ")" << std::endl;
    myRuntimeInfo->debug << "elevator:         " << pitchFCMass      << " (incl. factor for electric FC: " << pitch_ElecFC_MassFaktor << ")" << std::endl;
    myRuntimeInfo->debug << "-------------------------" << std::endl;
    myRuntimeInfo->debug << "                  " << primaryFCSMass   << std::endl;
    myRuntimeInfo->debug << std::endl;
    myRuntimeInfo->debug << "****** SFCS Masses ******" << std::endl;
    myRuntimeInfo->debug << "THSA:             " << THSA_FCMass      << " (incl. factor for electric FC: " << THSA_ElecFC_MassFaktor << ")" << std::endl;
    myRuntimeInfo->debug << "Flaps:            " << FlapsFCMass      << " (incl. factor for electric FC: " << Flaps_ElecFC_MassFaktor << ")" << std::endl;
    myRuntimeInfo->debug << "Spoilers:         " << SpoilersFCMass   << " (incl. factor for electric FC: " << Spoilers_ElecFC_MassFaktor << ")" << std::endl;
    myRuntimeInfo->debug << "Slat:             " << SlatsFCMass      << " (incl. factor for electric FC: " << Slats_ElecFC_MassFaktor << ")" << std::endl;
    myRuntimeInfo->debug << "-------------------------" << std::endl;
    myRuntimeInfo->debug << "                  " << secondaryFCSMass << std::endl;
    myRuntimeInfo->debug << std::endl;
    myRuntimeInfo->debug << "****** Common ***********" << std::endl;
    myRuntimeInfo->debug << "commonInstFCMass: " << commonInstFCMass << std::endl;
    myRuntimeInfo->debug << std::endl;
    myRuntimeInfo->debug << "****** Total ************" << std::endl;
    myRuntimeInfo->debug << "Total FCS Mass:   " << primaryFCSMass + secondaryFCSMass + commonInstFCMass << std::endl;
    myRuntimeInfo->debug << std::endl;
    return (primaryFCSMass + secondaryFCSMass + commonInstFCMass);
}

Vec3 conventionalATA27::getRefPoint() {
//    /* Annahme:CG-Lagen entsprechend Wing und LWe, gewichtet mit Flaechen **/
//    double x_cg_wing    = aWing.l_mu_right_refPt[0] + 0.5*aWing.l_mu_right;
//    double x_cg_htp     = aHTP.l_mu_right_refPt[0]  + 0.6*aHTP.l_mu_right;
//    double x_cg_vtp     = aVTP.l_mu_refPt[0]        + 0.6*aVTP.l_mu;
//
//    double z_cg_wing    = aWing.l_mu_right_refPt[2];
//    double z_cg_htp     = aHTP.l_mu_right_refPt[2];
//    double z_cg_vtp     = aVTP.l_mu_refPt[2];
//
//    double x_cg_ATA27 = (x_cg_wing*aWing.S+x_cg_htp*aHTP.S+x_cg_vtp*aVTP.S)/(aWing.S+aHTP.S+aVTP.S);
//    double z_cg_ATA27 = (z_cg_wing*aWing.S+z_cg_htp*aHTP.S+z_cg_vtp*aVTP.S)/(aWing.S+aHTP.S+aVTP.S);
//
//    return Vec3(x_cg_ATA27, 0., z_cg_ATA27);
    //alte Annahme: Schwerpunkt bei 45% Rumpflaenge
    // return Vec3(fuselageRef[0]+0.45*fuselageLength, fuselageRef[1], fuselageRef[2]);
    return Vec3(data_->theWing.back().aerodynamicSurface.origin[0] +
                0.5 * data_->theWing.back().aerodynamicSurface.sections.at(0).get_chord_length(),
                data_->theWing.back().aerodynamicSurface.origin[1], data_->theWing.back().aerodynamicSurface.origin[2]);
}

double conventionalATA27::getMaxDeflection(double minDeflection, double maxDeflection) {
    double maxDefl(0.0);
    if (fabs(minDeflection) < fabs(maxDeflection)) {
        maxDefl = fabs(maxDeflection);
    } else {
        maxDefl = fabs(minDeflection);
    }
    // Convert from degree to rad
    return convertUnit(DEGREE, RADIAN, maxDefl);;
}


void conventionalATA27::setHorizontalContrSurfGeo(const geom2::MultisectionSurface<geom2::PolygonSection> &geoContrSurf,
        const geom2::MultisectionSurface<geom2::AirfoilSection> &geoSurface,
        horizontalControlSurface *contrSurf,
        const flightControlSystem::ControlDevice &settingsControlDevice, double posDeflection, double negDeflection) {
    /* control device dimensions are normalized using span and chord lengths of the wing -> need to be multiplied with them for absoulte values */
    // absolute values for position, width, and length
    double controlSurfaceSpanPosition = geoContrSurf.sections.front().origin.z() * 0.5 * geom2::measure::span(geoSurface);
    // transform control surface to absolute coordinates
    geom2::MultisectionSurface<geom2::PolygonSection> absContrSurf = geom2::transform::to_absolute(geoContrSurf, geoSurface);
    contrSurf->l_i_abs              = geom2::measure::width(absContrSurf, absContrSurf.sections.front().origin.z()); // width of first section
    contrSurf->l_a_abs              = geom2::measure::width(absContrSurf, absContrSurf.sections.back().origin.z()); // width of second section
    double taper = contrSurf->l_a_abs / contrSurf->l_i_abs;
    double MAC = 2. / 3. * contrSurf->l_i_abs * (1. + taper + taper * taper) / (1 + taper);
    contrSurf->c_S                  = MAC;
    contrSurf->S                    = geom2::measure::area(absContrSurf);
    contrSurf->type                 = geoContrSurf.name;
    contrSurf->s_element            = geom2::measure::length(absContrSurf);
    contrSurf->y_i_abs              = controlSurfaceSpanPosition;
    /*(relative position in span direction + length of control surface) * span*/
    contrSurf->y_a_abs              = controlSurfaceSpanPosition + contrSurf->s_element;
    /* relative position of trailing edge of control device to wing trailing edge */
    // 1 - relative position trailing edge of control surface
    geom2::Point_3 contrSurfRightPoint_i = geom2::measure::right(geoContrSurf, geoContrSurf.sections.front().origin.z()); // trailing edge point of inner section
    contrSurf->l_i_rel_TE           = 1.0 - contrSurfRightPoint_i.x();
    geom2::Point_3 contrSurfRightPoint_o = geom2::measure::right(geoContrSurf, geoContrSurf.sections.back().origin.z()); // trailing edge point of outer section
    contrSurf->l_a_rel_TE           = 1.0 - contrSurfRightPoint_o.x();
    /* maximum deflection */
    contrSurf->maxDeflection        = getMaxDeflection(negDeflection, posDeflection);
    contrSurf->S_surface                     = geom2::measure::reference_area(geoSurface);
    contrSurf->MAC_surface                   = geom2::measure::mean_aerodynamic_chord(geoSurface);
    contrSurf->localWingChordAtContrSurfIB   = geom2::measure::chord(geoSurface, contrSurf->y_i_abs);
    contrSurf->localWingChordAtContrSurfOB   = geom2::measure::chord(geoSurface, contrSurf->y_a_abs);
    contrSurf->numberOfActuators    = settingsControlDevice.numberOfActuators;
    contrSurf->maxDefSpeed          = settingsControlDevice.maxDefSpeed;
    calcContrSurfSweepAngles(absContrSurf, contrSurf);
    calcContrSurfCenterOfArea(contrSurf);
    calcDistCoA2HingeLine(contrSurf);
    contrSurf->areaMomentAboutHingeLine = contrSurf->S * contrSurf->distCoA2HingeLine;
    //Determine the local reference point of the actuators in the coordinate system of the aerodynamic surface
    //(all are positioned in the center of the control device leading edge)
    Vec3 refPointActuators_local;
    refPointActuators_local[0] = (absContrSurf.sections.front().origin.x() + absContrSurf.sections.back().origin.x()) / 2;
    refPointActuators_local[1] = (absContrSurf.sections.front().origin.y() + absContrSurf.sections.back().origin.y()) / 2;
    refPointActuators_local[2] = (absContrSurf.sections.front().origin.z() + absContrSurf.sections.back().origin.z()) / 2;
    // transform from local aerodynamic surface coordinate system to global aircraft coordinate system
    //reference point in the global coordinate system
    Vec3 refPointActuators;
    refPointActuators[0] = refPointActuators_local[0] + geoSurface.origin.x(); // x local = x global
    refPointActuators[1] = refPointActuators_local[2] + geoSurface.origin.y(); // z local = y global
    refPointActuators[2] = refPointActuators_local[1] + geoSurface.origin.z(); // y local = z global

    for (int m = 0; m <= contrSurf->numberOfActuators - 1; m++) {
        //Assign actuators
        contrSurf->theActuators.push_back(actuatorPFCS());
        contrSurf->theActuators.back().refPoint = refPointActuators;
        contrSurf->theActuators.back().powerSource = settingsControlDevice.actuators.at(contrSurf->theActuators.size() - 1).powerSource.type;
        contrSurf->theActuators.back().powerSourceID = settingsControlDevice.actuators.at(contrSurf->theActuators.size() - 1).powerSource.sourceID;
        contrSurf->theActuators.back().operationMode = settingsControlDevice.actuators.at(contrSurf->theActuators.size() - 1).operationMode;
        contrSurf->theActuators.back().efficiency = settingsControlDevice.actuators.at(contrSurf->theActuators.size() - 1).efficiency;
        contrSurf->theActuators.back().power += settingsControlDevice.actuators.at(contrSurf->theActuators.size() - 1).stdbyPower;
        if (contrSurf->theActuators.back().operationMode == "active") {
            contrSurf->theActuators.back().power += settingsControlDevice.actuators.at(contrSurf->theActuators.size() - 1).activePower;
        }
    }
}

void conventionalATA27::setRudderTemp(const geom2::MultisectionSurface<geom2::PolygonSection> &geoContrSurf,
                                      const geom2::MultisectionSurface<geom2::AirfoilSection> &geoSurface, const std::string& side,
                                      verticalControlSurface *contrSurf,
                                      const flightControlSystem::ControlDevice &settingsControlDevice, double posDeflection, double negDeflection) {
    /* Define unused function parameters to avoid linter warnings */
    UNUSED(side);
    /* control device dimensions are normalized using span and chord lengths of the wing -> need to be multiplied with them for absoulte values */
    // absolute values for position, width, and length
    double controlSurfaceSpanPosition = geoContrSurf.sections.front().origin.z() * geom2::measure::span(geoSurface);
    // convert to absolute coordinates
    geom2::MultisectionSurface<geom2::PolygonSection> absContrSurf = geom2::transform::to_absolute(geoContrSurf, geoSurface);
    contrSurf->l_i_abs              = geom2::measure::width(absContrSurf, absContrSurf.sections.front().origin.z()); // width of first section
    contrSurf->l_a_abs              = geom2::measure::width(absContrSurf, absContrSurf.sections.back().origin.z()); // width of second section
    double taper = contrSurf->l_a_abs / contrSurf->l_i_abs;
    double MAC = 2. / 3. * contrSurf->l_i_abs * (1. + taper + taper * taper) / (1 + taper);
    contrSurf->c_S                  = MAC;
    contrSurf->S                    = geom2::measure::area(absContrSurf);
    contrSurf->type                 = geoContrSurf.name;
    contrSurf->s_element            = geom2::measure::length(absContrSurf);
    contrSurf->y_i_abs              = controlSurfaceSpanPosition;
    /*(relative position in span direction + length of control surface) * span*/
    contrSurf->y_a_abs              = controlSurfaceSpanPosition + contrSurf->s_element;
    contrSurf->l_i_rel_TE           = 0.;
    contrSurf->l_a_rel_TE           = 0.;
    contrSurf->maxDeflection        = getMaxDeflection(negDeflection, posDeflection);
    contrSurf->S_surface                     = geom2::measure::reference_area(geoSurface);
    contrSurf->MAC_surface                   = geom2::measure::mean_aerodynamic_chord(geoSurface);
    contrSurf->localWingChordAtContrSurfIB   = geom2::measure::chord(geoSurface, contrSurf->y_i_abs);
    contrSurf->localWingChordAtContrSurfOB   = geom2::measure::chord(geoSurface, contrSurf->y_a_abs);
    contrSurf->numberOfActuators    = settingsControlDevice.numberOfActuators;
    contrSurf->maxDefSpeed          = settingsControlDevice.maxDefSpeed;
    calcContrSurfSweepAngles(absContrSurf, contrSurf);
    calcContrSurfCenterOfArea(contrSurf);
    calcDistCoA2HingeLine(contrSurf);
    contrSurf->areaMomentAboutHingeLine = contrSurf->S * contrSurf->distCoA2HingeLine;
    //Determine the reference point of the actuators (all are positioned in the center of the control device leading edge)
    Vec3 refPointActuators_local;
    refPointActuators_local[0] = (absContrSurf.sections.front().origin.x() + absContrSurf.sections.back().origin.x()) / 2;
    refPointActuators_local[1] = (absContrSurf.sections.front().origin.y() + absContrSurf.sections.back().origin.y()) / 2;
    refPointActuators_local[2] = (absContrSurf.sections.front().origin.z() + absContrSurf.sections.back().origin.z()) / 2;
    // transform from local aerodynamic surface coordinate system to global aircraft coordinate system
    //reference point in the global coordinate system
    Vec3 refPointActuators;
    refPointActuators[0] = refPointActuators_local[0] + geoSurface.origin.x(); // x local = x global
    refPointActuators[1] = refPointActuators_local[1] + geoSurface.origin.y(); // y local = y global
    refPointActuators[2] = refPointActuators_local[2] + geoSurface.origin.z(); // z local = z global

    for (int m = 0; m <= contrSurf->numberOfActuators - 1; m++) {
        //Assign actuators
        contrSurf->theActuators.push_back(actuatorPFCS());
        contrSurf->theActuators.back().refPoint = refPointActuators;
        contrSurf->theActuators.back().powerSource = settingsControlDevice.actuators.at(contrSurf->theActuators.size() - 1).powerSource.type;
        contrSurf->theActuators.back().powerSourceID = settingsControlDevice.actuators.at(contrSurf->theActuators.size() - 1).powerSource.sourceID;
        contrSurf->theActuators.back().operationMode = settingsControlDevice.actuators.at(contrSurf->theActuators.size() - 1).operationMode;
        contrSurf->theActuators.back().efficiency = settingsControlDevice.actuators.at(contrSurf->theActuators.size() - 1).efficiency;
        contrSurf->theActuators.back().power += settingsControlDevice.actuators.at(contrSurf->theActuators.size() - 1).stdbyPower;
        if (contrSurf->theActuators.back().operationMode == "active") {
            contrSurf->theActuators.back().power += settingsControlDevice.actuators.at(contrSurf->theActuators.size() - 1).activePower;
        }
    }
}

void conventionalATA27::setTHSA_AsControlSurface(const geom2::MultisectionSurface<geom2::AirfoilSection> &geoSurface,
        thsaControlSurface *contrSurf, const flightControlSystem::ControlDevice &settingsControlDevice) {
    contrSurf->c_S                   = geom2::measure::mean_aerodynamic_chord(geoSurface);
    contrSurf->S                     = geom2::measure::reference_area(geoSurface);
    contrSurf->type                  = settingsControlDevice.name;
    contrSurf->S_surface                     = geom2::measure::reference_area(geoSurface);
    contrSurf->MAC_surface                   = geom2::measure::mean_aerodynamic_chord(geoSurface);
    contrSurf->numberOfActuators     = settingsControlDevice.numberOfActuators;
    contrSurf->maxDefSpeed           = settingsControlDevice.maxDefSpeed;
}

void conventionalATA27::setDesignPower_FCS(double wingLoading, double VMO, double MMO, std::vector<powerProfile> *myElectric, std::vector<powerProfile> *myHydraulic) {
    // Aileron
    for (size_t i = 0; i < Aileron.size(); i++) {
        getDesignPowerPFCS(&Aileron.at(i), wingLoading, VMO, MMO);
        /* Loop over hydraulic circuits to assign designPower correctly */
        for (size_t n(0); n < myHydraulic->size(); n++) {
            //Actuator with the largest designPower is dimensioning
            std::vector<double> designPower_tmp;
            double maxDesignPower(0.);
            for (size_t j = 0; j < Aileron.at(i).theActuators.size(); j++) {
                designPower_tmp.push_back(0.);
                if (Aileron.at(i).theActuators.at(j).powerSource == "Hydraulic" && Aileron.at(i).theActuators.at(j).powerSourceID == static_cast<int>(n + 1)) {
                    designPower_tmp.back() = Aileron.at(i).designPowerActuator / Aileron.at(i).theActuators.at(j).efficiency;
                    Aileron.at(i).theActuators.at(j).designPower = designPower_tmp.back();
                }
            }
            maxDesignPower = *max_element(designPower_tmp.begin(), designPower_tmp.end());
            myHydraulic->at(n).designPower += maxDesignPower;
        }
        /* Loop over electric circuits to assign design power correctly*/
        for (size_t n(0); n < myElectric->size(); n++) {
            //Actuator with the largest designPower is dimensioning
            std::vector<double> designPower_tmp;
            double maxDesignPower(0.);
            for (size_t j = 0; j < Aileron.at(i).theActuators.size(); j++) {
                designPower_tmp.push_back(0.);
                if (Aileron.at(i).theActuators.at(j).powerSource == "Electric" && Aileron.at(i).theActuators.at(j).powerSourceID == static_cast<int>(n + 1)) {
                    designPower_tmp.back() = (Aileron.at(i).designPowerActuator) / Aileron.at(i).theActuators.at(j).efficiency;
                    Aileron.at(i).theActuators.at(j).designPower = designPower_tmp.back();
                }
            }
            maxDesignPower = *max_element(designPower_tmp.begin(), designPower_tmp.end());
            myElectric->at(n).designPower += maxDesignPower;
        }
    }
    // Spoiler
    for (size_t i = 0; i < Spoiler.size(); i++) {
        getDesignPowerPFCS(&Spoiler.at(i), wingLoading, VMO, MMO);
        /* Loop over hydraulic circuits to assign designPower correctly */
        for (size_t n(0); n < myHydraulic->size(); n++) {
            //Actuator with the largest designPower is dimensioning
            std::vector<double> designPower_tmp;
            double maxDesignPower(0.);
            for (size_t j = 0; j < Spoiler.at(i).theActuators.size(); j++) {
                designPower_tmp.push_back(0.);
                if (Spoiler.at(i).theActuators.at(j).powerSource == "Hydraulic" && Spoiler.at(i).theActuators.at(j).powerSourceID == static_cast<int>(n + 1)) {
                    designPower_tmp.back() = (Spoiler.at(i).designPowerActuator) / Spoiler.at(i).theActuators.at(j).efficiency;
                    Spoiler.at(i).theActuators.at(j).designPower = designPower_tmp.back();
                }
            }
            maxDesignPower = *max_element(designPower_tmp.begin(), designPower_tmp.end());
            myHydraulic->at(n).designPower += maxDesignPower;
        }
        /* Loop over electric circuits to assign designPower correctly */
        for (size_t n(0); n < myElectric->size(); n++) {
            //Actuator with the largest designPower is dimensioning
            std::vector<double> designPower_tmp;
            double maxDesignPower(0.);
            for (size_t j = 0; j < Spoiler.at(i).theActuators.size(); j++) {
                designPower_tmp.push_back(0.);
                if (Spoiler.at(i).theActuators.at(j).powerSource == "Electric" && Spoiler.at(i).theActuators.at(j).powerSourceID == static_cast<int>(n + 1)) {
                    designPower_tmp.back() = (Spoiler.at(i).designPowerActuator) / Spoiler.at(i).theActuators.at(j).efficiency;
                    Spoiler.at(i).theActuators.at(j).designPower = designPower_tmp.back();
                }
            }
            maxDesignPower = *max_element(designPower_tmp.begin(), designPower_tmp.end());
            myElectric->at(n).designPower += maxDesignPower;
        }
    }
    // Elevator
    for (size_t i = 0; i < Elevator.size(); i++) {
        getDesignPowerPFCS(&Elevator.at(i), wingLoading, VMO, MMO);
        /* Loop over hydraulic circuits to assign designPower correctly */
        for (size_t n(0); n < myHydraulic->size(); n++) {
            //Actuator with the largest designPower is dimensioning
            std::vector<double> designPower_tmp;
            double maxDesignPower(0.);
            for (size_t j = 0; j < Elevator.at(i).theActuators.size(); j++) {
                designPower_tmp.push_back(0.);
                if (Elevator.at(i).theActuators.at(j).powerSource == "Hydraulic" && Elevator.at(i).theActuators.at(j).powerSourceID == static_cast<int>(n + 1)) {
                    designPower_tmp.back() = (Elevator.at(i).designPowerActuator) / Elevator.at(i).theActuators.at(j).efficiency;
                    Elevator.at(i).theActuators.at(j).designPower = designPower_tmp.back();
                }
            }
            maxDesignPower = *max_element(designPower_tmp.begin(), designPower_tmp.end());
            myHydraulic->at(n).designPower += maxDesignPower;
        }
        /* Loop over hydraulic circuits to assign designPower correctly */
        for (size_t n(0); n < myElectric->size(); n++) {
            //Actuator with the largest designPower is dimensioning
            std::vector<double> designPower_tmp;
            double maxDesignPower(0.);
            for (size_t j = 0; j < Elevator.at(i).theActuators.size(); j++) {
                designPower_tmp.push_back(0.);
                if (Elevator.at(i).theActuators.at(j).powerSource == "Electric" && Elevator.at(i).theActuators.at(j).powerSourceID == static_cast<int>(n + 1)) {
                    designPower_tmp.back() = (Elevator.at(i).designPowerActuator) / Elevator.at(i).theActuators.at(j).efficiency;
                    Elevator.at(i).theActuators.at(j).designPower = designPower_tmp.back();
                }
            }
            maxDesignPower = *max_element(designPower_tmp.begin(), designPower_tmp.end());
            myElectric->at(n).designPower += maxDesignPower;
        }
    }
    // Rudder
    for (size_t i = 0; i < Rudder.size(); i++) {
        getDesignPowerPFCS(&Rudder.at(i), wingLoading, VMO, MMO);
        /* Loop over hydraulic circuits to assign designPower correctly */
        for (size_t n(0); n < myHydraulic->size(); n++) {
            //Actuator with the largest designPower is dimensioning
            std::vector<double> designPower_tmp;
            double maxDesignPower(0.);
            for (size_t j = 0; j < Rudder.at(i).theActuators.size(); j++) {
                designPower_tmp.push_back(0.);
                if (Rudder.at(i).theActuators.at(j).powerSource == "Hydraulic" && Rudder.at(i).theActuators.at(j).powerSourceID == static_cast<int>(n + 1)) {
                    // NOTE (Peter#1#): Where does the factor 1.5 come from with the Rudder?!
                    designPower_tmp.back() = (Rudder.at(i).designPowerActuator) / Rudder.at(i).theActuators.at(j).efficiency;
                    Rudder.at(i).theActuators.at(j).designPower = designPower_tmp.back();
                }
            }
            maxDesignPower = *max_element(designPower_tmp.begin(), designPower_tmp.end());
            myHydraulic->at(n).designPower += maxDesignPower;
        }
        /* Loop over electric circuits to assign designPower correctly */
        for (size_t n(0); n < myElectric->size(); n++) {
            //Actuator with the largest designPower is dimensioning
            std::vector<double> designPower_tmp;
            double maxDesignPower(0.);
            for (size_t j = 0; j < Rudder.at(i).theActuators.size(); j++) {
                designPower_tmp.push_back(0.);
                if (Rudder.at(i).theActuators.at(j).powerSource == "Electric" && Rudder.at(i).theActuators.at(j).powerSourceID == static_cast<int>(n + 1)) {
                    designPower_tmp.back() = (Rudder.at(i).designPowerActuator) / Rudder.at(i).theActuators.at(j).efficiency;
                    Rudder.at(i).theActuators.at(j).designPower = designPower_tmp.back();
                }
            }
            maxDesignPower = *max_element(designPower_tmp.begin(), designPower_tmp.end());
            myElectric->at(n).designPower += maxDesignPower;
        }
    }
    //THSA
    for (size_t i = 0; i < THSA.size(); i++) {
        getDesignPowerPFCS(&THSA.at(i), wingLoading, VMO, MMO);
        /* Loop over hydraulic circuits to assign designPower correctly */
        for (size_t n(0); n < myHydraulic->size(); n++) {
            //Actuator with the largest designPower is dimensioning
            std::vector<double> designPower_tmp;
            double maxDesignPower(0.);
            for (size_t j = 0; j < THSA.at(i).theActuators.size(); j++) {
                designPower_tmp.push_back(0.);
                if (THSA.at(i).theActuators.at(j).powerSource == "Hydraulic" && THSA.at(i).theActuators.at(j).powerSourceID == static_cast<int>(n + 1)) {
                    designPower_tmp.back() = (THSA.at(i).designPowerActuator) / THSA.at(i).theActuators.at(j).efficiency;
                    THSA.at(i).theActuators.at(j).designPower = designPower_tmp.back();
                }
            }
            maxDesignPower = *max_element(designPower_tmp.begin(), designPower_tmp.end());
            myHydraulic->at(n).designPower += maxDesignPower;
        }
        /* Loop over electric circuits to assign designPower correctly */
        for (size_t n(0); n < myElectric->size(); n++) {
            //Actuator with the largest designPower is dimensioning
            std::vector<double> designPower_tmp;
            double maxDesignPower(0.);
            for (size_t j = 0; j < THSA.at(i).theActuators.size(); j++) {
                designPower_tmp.push_back(0.);
                if (THSA.at(i).theActuators.at(j).powerSource == "Electric" && THSA.at(i).theActuators.at(j).powerSourceID == static_cast<int>(n + 1)) {
                    designPower_tmp.back() = (THSA.at(i).designPowerActuator) / THSA.at(i).theActuators.at(j).efficiency;
                    THSA.at(i).theActuators.at(j).designPower = designPower_tmp.back();
                }
            }
            maxDesignPower = *max_element(designPower_tmp.begin(), designPower_tmp.end());
            myElectric->at(n).designPower += maxDesignPower;
        }
    }
    //HighLift
    for (size_t i = 0; i < theHighLiftSystem.size(); i++) {
        for (size_t j = 0; j < theHighLiftSystem.at(i).Systems.size(); j++) {
            for (size_t n = 0; n < theHighLiftSystem.at(i).Systems.at(j).PCU.size(); n++) {
                //Even distribution of the power among all PCU motors
                int numberOfPCUMotors(theHighLiftSystem.at(i).Systems.at(j).PCU.at(n).PCU_Motor.size());
                for (size_t m(0); m < theHighLiftSystem.at(i).Systems.at(j).PCU.at(n).PCU_Motor.size(); m++) {
                    /* Loop over hydraulic circuits to assign designPower correctly */
                    for (size_t k(0); k < data_->configuration.Hydraulic.circuits.size(); k++) {
                        if (theHighLiftSystem.at(i).Systems.at(j).PCU.at(n).PCU_Motor.at(m).powerSource == "Hydraulic"
                                && theHighLiftSystem.at(i).Systems.at(j).PCU.at(n).PCU_Motor.at(m).powerSourceID == static_cast<int>(k + 1)) {
                            //Even distribution of the power among all PCU motors
                            myHydraulic->at(k).designPower += theHighLiftSystem.at(i).Systems.at(j).PCU.at(n).installedPower / numberOfPCUMotors;
                        }
                    }
                    /* Loop over electric circuits to assign designPower correctly */
                    for (size_t k(0); k < data_->configuration.Electric.circuits.size(); k++) {
                        if (theHighLiftSystem.at(i).Systems.at(j).PCU.at(n).PCU_Motor.at(m).powerSource == "Electric"
                                && theHighLiftSystem.at(i).Systems.at(j).PCU.at(n).PCU_Motor.at(m).powerSourceID == static_cast<int>(k + 1)) {
                            //Even distribution of the power among all PCU motors
                            myElectric->at(k).designPower += theHighLiftSystem.at(i).Systems.at(j).PCU.at(n).installedPower / numberOfPCUMotors;
                        }
                    }
                }
            }
        }
    }
//    std::cout << "Design Power calculation finished." << std::endl;
}
/* Determining the maximum power per control surface according to [NASA76] */
void conventionalATA27::getDesignPowerPFCS(controlSurfaceATA27 *theControlSurface,
        double theWingLoading, double theVMO, double theMMO) {
    double hingeMoment(0.);
    double maxHingeMoment(0.);
    double dynamicPressure(0.);
    double Mach(0.);
    double C_L_buffet(1.);
    double n_max(2.5);
//    std::cout << "--- " << theControlSurface->type << " ---" << std::endl;
//    std::cout << "WingLoading: " << theWingLoading << " kg/m2" << std::endl;
//    std::cout << "VMO: " << theVMO << " m/s" << std::endl;
//    std::cout << "MMO: " << theMMO << " -" << std::endl;
    // Maneuver Load Control
    dynamicPressure = (n_max * theWingLoading) / C_L_buffet;
    hingeMoment = getHingeMomentPFCS(*theControlSurface, dynamicPressure, 0.6);
    if (fabs(hingeMoment) > maxHingeMoment) {
        maxHingeMoment = fabs(hingeMoment);
    }
//    std::cout << "Maneuver Load Control" << std::endl;
//    std::cout << "q " << dynamicPressure << " kg/m/s2" << std::endl;
//    std::cout << "M " << hingeMoment << " Nm" << std::endl;
    // Gust Load Alleviation (VMO=MMO in 25,000 ft)
    dynamicPressure = 0.5 * data_->mission.myISA.getDensity(convertUnit(FOOT, METER, 25000.))
                    * pow(convertUnit(MACH, TRUEAIRSPEED, convertUnit(FOOT, METER, 25000.), data_->mission.myISA, theMMO), 2.);
    hingeMoment = getHingeMomentPFCS(*theControlSurface, dynamicPressure, theMMO);
    if (fabs(hingeMoment) > maxHingeMoment) {
        maxHingeMoment = fabs(hingeMoment);
    }
//    std::cout << "Gust Load Alleviation" << std::endl;
//    std::cout << "q " << dynamicPressure << " kg/m/s2" << std::endl;
//    std::cout << "M " << hingeMoment << " Nm" << std::endl;
    //Maneuver Load Alleviation (V_dive in 0,000 ft)
    dynamicPressure = 0.5 * data_->mission.myISA.getDensity(convertUnit(FOOT, METER, 0.))
                    * pow((1.2 * theVMO), 2.);
    Mach = 1.2 * convertUnit(TRUEAIRSPEED, MACH, convertUnit(FOOT, METER, 0.), data_->mission.myISA, theVMO);
    hingeMoment = getHingeMomentPFCS(*theControlSurface, dynamicPressure, Mach);
    if (fabs(hingeMoment) > maxHingeMoment) {
        maxHingeMoment = fabs(hingeMoment);
    }
//    std::cout << "Maneuver Load Alleviation" << std::endl;
//    std::cout << "q " << dynamicPressure << " kg/m/s2" << std::endl;
//    std::cout << "M " << hingeMoment << " Nm" << std::endl;
    // Bestimmen des maximalen Leistung nach [NASA76]
    theControlSurface->maxHingeMoment = maxHingeMoment;
    theControlSurface->controlSurfaceDesignPower = maxHingeMoment * theControlSurface->maxDefSpeed;
    theControlSurface->designPowerActuator = theControlSurface->controlSurfaceDesignPower;
}

double conventionalATA27::getHingeMomentPFCS(controlSurfaceATA27 myControlSurface, double q, double theMach) {
    // Determining the Hinge Moment according to [NASA76]
    // Graphs from Fig. 10
    double C_H_0     = getC_H_0(theMach);
    double C_H_gamma = getC_H_gamma(theMach);
    //Calculates factor for changes in hinge moment due to surface chord ratio and AOA, [NASA76]
    double K_alpha   = 2.7918 * myControlSurface.c_S / myControlSurface.MAC_surface + 0.2881;
    //Calculates factor for changes in hinge moment due to surface chord ratio and control surface deflection, [NASA76] **/
    double K_gamma   = 0.9286 * myControlSurface.c_S / myControlSurface.MAC_surface + 0.7633;
    double C_hinge   = K_alpha * (C_H_0 + myControlSurface.deltaC_H_0)
                       + K_gamma * C_H_gamma;
    // Eq. on page 26
    return (q * myControlSurface.S * myControlSurface.c_S * C_hinge);
}

double conventionalATA27::getC_H_0(double M) {
    if (M < 0.55) {
        M = 0.55;
    }
    return (23.9170 * pow(M, 6.)
            - 115.90 * pow(M, 5.)
            + 228.04 * pow(M, 4.)
            - 234.32 * pow(M, 3.)
            + 132.98 * pow(M, 2.)
            - 39.612 * M + 4.6716);
}

double conventionalATA27::getC_H_gamma(double M) {
    if (M < 0.55) {
        M = 0.55;
    }
    if (M <= 0.9) {
        return (-579.11 * pow(M, 6.)
                + 2441.7 * pow(M, 5.)
                - 4274.7 * pow(M, 4.)
                + 3976.3 * pow(M, 3.)
                - 2072.5 * pow(M, 2.)
                + 573.83 * M - 65.975);
    } else {
        return (-4.2822 * pow(M, 2.) + 8.8033 * M - 4.5618);
    }
}

double conventionalATA27::getAileronControlMass(std::vector<horizontalControlSurface> myControlSurfaces) {
    /** [NASA76], S. 42 **/
    double accumulatedWeight(0.);
    double accumulated_k_i_times_S_Ci(0.);
    for (size_t i = 0; i < myControlSurfaces.size(); i++) {
        double k_i  = myControlSurfaces.at(i).numberOfActuators;
        double S_ci = 2. * myControlSurfaces.at(i).areaMomentAboutHingeLine / pow(0.3048, 3.);
        accumulated_k_i_times_S_Ci += 2. * k_i * S_ci;
    }
    accumulatedWeight = 29.35 * pow(2. * accumulated_k_i_times_S_Ci, 0.44) * 0.4536;
    /* Scaling the accumulated weight over all Ailerons, NASA Method can not be split up for each myControlSurfaces for mathematical reasons */
    for (size_t i = 0; i < myControlSurfaces.size(); i++) {
        double k_i  = myControlSurfaces.at(i).numberOfActuators;
        double S_ci = 2. * myControlSurfaces.at(i).areaMomentAboutHingeLine / pow(0.3048, 3.);
        myControlSurfaces.at(i).massControl = accumulatedWeight * (2. * k_i * S_ci / accumulated_k_i_times_S_Ci);
       myRuntimeInfo->debug << std::endl << "***:  " << myControlSurfaces.at(i).type << std::endl;
       //myRuntimeInfo->debug << "side: " << myControlSurfaces.at(i).side << std::endl;
       myRuntimeInfo->debug << "S:    " << myControlSurfaces.at(i).S << std::endl;
       myRuntimeInfo->debug << "distCoA2HingeLine:  " << myControlSurfaces.at(i).distCoA2HingeLine << std::endl;
       myRuntimeInfo->debug << "myControlSurfaces.at(i).numberOfActuators:  " << myControlSurfaces.at(i).numberOfActuators << std::endl;
       myRuntimeInfo->debug << "myControlSurfaces.at(i).massControl: " << myControlSurfaces.at(i).massControl << std::endl;
       myRuntimeInfo->debug << "accumulatedWeight: " << accumulatedWeight << std::endl;
    }
    return accumulatedWeight;
}

double conventionalATA27::getRudderControlMass(std::vector<verticalControlSurface> myControlSurfaces) {
    /* [NASA76], S. 42 **/
    double accumulatedWeight(0.);
    double accumulated_k_i_times_S_Ci(0.);
    for (size_t i = 0; i < myControlSurfaces.size(); i++) {
        double k_i  = myControlSurfaces.at(i).numberOfActuators;
        double S_ci = 2. * myControlSurfaces.at(i).areaMomentAboutHingeLine / pow(0.3048, 3.);
        accumulated_k_i_times_S_Ci += 2. * k_i * S_ci;
    }
    accumulatedWeight = 29.35 * pow(2. * accumulated_k_i_times_S_Ci, 0.44) * 0.4536;
    /* Scaling the accumulated weight over all Ailerons, NASA Method can not be split up for each myControlSurfaces for mathematical reasons */
    for (size_t i = 0; i < myControlSurfaces.size(); i++) {
        double k_i  = myControlSurfaces.at(i).numberOfActuators;
        double S_ci = 2. * myControlSurfaces.at(i).areaMomentAboutHingeLine / pow(0.3048, 3.);
        myControlSurfaces.at(i).massControl = accumulatedWeight * (2. * k_i * S_ci / accumulated_k_i_times_S_Ci);
       myRuntimeInfo->debug << std::endl << "***:  " << myControlSurfaces.at(i).type << std::endl;
       //myRuntimeInfo->debug << "side: " << myControlSurfaces.at(i).side << std::endl;
       myRuntimeInfo->debug << "S:    " << myControlSurfaces.at(i).S << std::endl;
       myRuntimeInfo->debug << "distCoA2HingeLine:  " << myControlSurfaces.at(i).distCoA2HingeLine << std::endl;
       myRuntimeInfo->debug << "myControlSurfaces.at(i).numberOfActuators:  " << myControlSurfaces.at(i).numberOfActuators << std::endl;
       myRuntimeInfo->debug << "myControlSurfaces.at(i).massControl: " << myControlSurfaces.at(i).massControl << std::endl;
       myRuntimeInfo->debug << "accumulatedWeight: " << accumulatedWeight << std::endl;
    }
    return accumulatedWeight;
}

double conventionalATA27::getSpoilerControlMass(std::vector<horizontalControlSurface> myControlSurfaces) {
    /* [NASA76], S. 46 **/
    double accumulatedWeight(0.);
    int number_AirSpoilers(0);
    for (size_t i = 0; i < myControlSurfaces.size(); i++) {
        if (myControlSurfaces.at(i).type == "spoiler_air")
            number_AirSpoilers += 1;
    }
    for (size_t i = 0; i < myControlSurfaces.size(); i++) {
        if (myControlSurfaces.at(i).type == "spoiler_air") {
            double phiSpoiler = myControlSurfaces.at(i).maxDeflection * 180 / 3.14;
            double S_csp = 2. * myControlSurfaces.at(i).areaMomentAboutHingeLine / pow(0.3048, 3.);
            double K_sp = pow(295., 2.) * cos(myControlSurfaces.at(i).phi_LE_CoSu * 3.14 / 180.) * S_csp * sin(phiSpoiler * 3.14 / 180.) / pow(10., 6.);
            myControlSurfaces.at(i).massControl = ((89. + 20.) / number_AirSpoilers + 76.8 * (phiSpoiler / 100.) * pow(K_sp, 0.75)) * 0.4536;
            accumulatedWeight += myControlSurfaces.at(i).massControl;
           myRuntimeInfo->debug << std::endl << "***:  " << myControlSurfaces.at(i).type << std::endl;
           //myRuntimeInfo->debug << "side: " << myControlSurfaces.at(i).side << std::endl;
           myRuntimeInfo->debug << "S:    " << myControlSurfaces.at(i).S << std::endl;
           myRuntimeInfo->debug << "distCoA2HingeLine:  " << myControlSurfaces.at(i).distCoA2HingeLine << std::endl;
           myRuntimeInfo->debug << "S_Ci: " << S_csp << std::endl;
           myRuntimeInfo->debug << "myControlSurfaces.at(i).numberOfActuators:  " << myControlSurfaces.at(i).numberOfActuators << std::endl;
           myRuntimeInfo->debug << "myControlSurfaces.at(i).massControl: " << myControlSurfaces.at(i).massControl << std::endl;
           myRuntimeInfo->debug << "accumulatedWeight: " << accumulatedWeight << std::endl;
        } else {
            /* According to the NASA method ground spoilers should not be considered */
        }
    }
    return accumulatedWeight;
}

double conventionalATA27::getElevatorControlMass(std::vector<horizontalControlSurface> myControlSurfaces) {
    /** [NASA76], S. 48 **/
    double accumulateWeight(0.);
    double S_c_total(0.);
    for (size_t i = 0; i < myControlSurfaces.size() / 2.; i++) {
        // used to be: < myControlSurfaces.size() / 2. Unclear why...
        double S_ci    = 2 * myControlSurfaces.at(i).S * myControlSurfaces.at(i).distCoA2HingeLine;
        S_ci    = convertUnit(VOLUME, NOPREFIX, METER, NOPREFIX, FOOT, S_ci);
        S_c_total += 2 * S_ci;
    }
    for (size_t i = 0; i < myControlSurfaces.size() / 2.; i++) {
        // used to be: < myControlSurfaces.size() / 2. Unclear why...
        double MOA_m(data_->data.Performance.MOA); //in [m]
        double VMO_msTAS(0.);
        atmosphere theISA;
        theISA.setAtmosphere(0., ISA_TEMPERATURE + data_->delta_ISA, ISA_PRESSURE);
        VMO_msTAS = std::min(convertUnit(MACH, TRUEAIRSPEED, MOA_m, theISA, data_->data.Performance.MMO),
                             convertUnit(CALIBRATEDAIRSPEED, TRUEAIRSPEED, MOA_m, theISA, data_->data.Performance.VMO));
        double VMO_ktTAS(convertUnit(METERPERSECOND, KNOTS, VMO_msTAS));
        double S_ci    = 2 * myControlSurfaces.at(i).areaMomentAboutHingeLine;
        S_ci    = convertUnit(VOLUME, NOPREFIX, METER, NOPREFIX, FOOT, S_ci);
        double lambda_e = myControlSurfaces.at(i).phi_LE_CoSu;
        double P_eci = sin(convertUnit(DEGREE, RADIAN, lambda_e)) * pow((VMO_ktTAS / 100.), 2.) * cos(myControlSurfaces.at(i).maxDeflection) * S_ci;
        myControlSurfaces.at(i).massControl = 2 * (119. * S_ci / S_c_total + 0.03 * P_eci) * static_cast<double>(myControlSurfaces.at(i).numberOfActuators);
        myControlSurfaces.at(i).massControl = convertUnit(POUND, KILO, GRAM, myControlSurfaces.at(i).massControl);
        accumulateWeight += myControlSurfaces.at(i).massControl;
       myRuntimeInfo->debug << std::endl << "***:  " << myControlSurfaces.at(i).type << std::endl;
       //myRuntimeInfo->debug << "side: " << myControlSurfaces.at(i).side << std::endl;
       myRuntimeInfo->debug << "S:    " << myControlSurfaces.at(i).S << std::endl;
       myRuntimeInfo->debug << "distCoA2HingeLine:  " << myControlSurfaces.at(i).distCoA2HingeLine << std::endl;
       myRuntimeInfo->debug << "S_Ci: " << S_ci << std::endl;
       myRuntimeInfo->debug << "myControlSurfaces.at(i).numberOfActuators:  " << myControlSurfaces.at(i).numberOfActuators << std::endl;
       myRuntimeInfo->debug << "myControlSurfaces.at(i).massControl: " << myControlSurfaces.at(i).massControl << std::endl;
       myRuntimeInfo->debug << "accumulateWeight: " << accumulateWeight << std::endl;
    }
    return accumulateWeight;
}

double conventionalATA27::getTHSAControlMass() {
    /** [NASA76], S. 47 **/
    double accumulateWeight(0.);
    for (size_t i = 0; i < THSA.size(); i++) {
        double l_PA     = 0.2 * THSA.at(i).c_S * 39.37; // m -> in
        double l_STA    = 0.3 * THSA.at(i).c_S * 39.37; // m -> in
        double maxTspa  = 1.33 * 0.35 * data_->data.Mass.MTOM / 0.4536 * l_PA; //kg -> lb
        double K_sa     = maxTspa / (l_STA * pow(10., 3.));
        THSA.at(i).massControl = (63.6 + 4.1 * K_sa) * 0.4536;   //lb -> kg
        accumulateWeight += THSA.at(i).massControl;
        /*TORENBEEK METHOD [Tor13]*/
//accumulateWeight+= 2.16 * pow(THSA.at(i).S * pow( data_->data.Performance.VMO , 0.5)* sin(15*3.14/180),0.88);
    }
    return accumulateWeight;
}

void conventionalATA27::calcContrSurfCenterOfArea(detailedControlSurface *myContrSurf) {
    double grundseitePlusUeberstand(myContrSurf->s_element * tan(myContrSurf->phi_LE_CoSu * 3.14 / 180.) + myContrSurf->l_a_abs);
    myContrSurf->chordDistIBLE2CenterOfArea = (pow(myContrSurf->l_i_abs, 2.0) - pow(myContrSurf->l_a_abs,
            2.0) + grundseitePlusUeberstand * (myContrSurf->l_i_abs + 2.*myContrSurf->l_a_abs)) / (3.*(myContrSurf->l_i_abs + myContrSurf->l_a_abs));
    myContrSurf->spanDistIBLE2CenterOfArea  = myContrSurf->s_element * (myContrSurf->l_i_abs + 2.*myContrSurf->l_a_abs) / (3.*(myContrSurf->l_i_abs + myContrSurf->l_a_abs));
}

void conventionalATA27::calcDistCoA2HingeLine(detailedControlSurface *myContrSurf) {
    //calculation of chordwise distance from contr surf LE to Center of Area
    double contrSurfChordAtCoA(myContrSurf->l_i_abs - (myContrSurf->l_i_abs - myContrSurf->l_a_abs) * myContrSurf->spanDistIBLE2CenterOfArea / myContrSurf->s_element);
    double chordwiseDistHingeLineCoA(myContrSurf->chordDistIBLE2CenterOfArea
                                     - myContrSurf->spanDistIBLE2CenterOfArea * tan(myContrSurf->phi_LE_CoSu * 3.14 / 180.)
                                     - myContrSurf->relPosOfContrSurfHingeLine * contrSurfChordAtCoA);
    double sweepHingeLingeAtCoASpan(myContrSurf->phi_LE_CoSu -
                                    (myContrSurf->phi_LE_CoSu - myContrSurf->phi_TE_CoSu) * myContrSurf->relPosOfContrSurfHingeLine);
    myContrSurf->distCoA2HingeLine = chordwiseDistHingeLineCoA / cos(sweepHingeLingeAtCoASpan * 3.14 / 180.);
}

void conventionalATA27::calcContrSurfSweepAngles(const geom2::MultisectionSurface<geom2::PolygonSection> &surface, detailedControlSurface *myContrSurf) {
     /* Initialize sweep */
    double sweepLE{0.0};
    double sweepTE{0.0};

    /* Get the front section of the control surface */
    const geom2::PolygonSection section_front = surface.sections.front();
    /* get the 3D point of the leading edge of the first section */
    const geom2::Point_3 front_LE = geom2::transform::to_parent(section_front, *section_front.get_contour(true).left_vertex());
    /* get the 3D point of the trailing edge of the first section */
    const geom2::Point_3 front_TE = geom2::transform::to_parent(section_front, *section_front.get_contour(true).right_vertex());

    /* Get the back section of the control surface */
    const geom2::PolygonSection section_back = surface.sections.back();
    /* get the 3D point of the leading edge of the first section */
    const geom2::Point_3 back_LE = geom2::transform::to_parent(section_back, *section_back.get_contour(true).left_vertex());
    /* get the 3D point of the trailing edge of the first section */
    const geom2::Point_3 back_TE = geom2::transform::to_parent(section_back, *section_back.get_contour(true).right_vertex());

    /* Get the points at the leading edge */
    const geom2::Point_3 first_pointLE{front_LE};
    const geom2::Point_3 second_pointLE{back_LE};
    /* Get the points at trailing edge*/
    const geom2::Point_3 first_pointTE{front_TE};
    const geom2::Point_3 second_pointTE{back_TE};
    /* Calculate the projected angle of the two points with the local z axis */
    const double dx_LE = second_pointLE.x() - first_pointLE.x();
    const double dz_LE = second_pointLE.z() - first_pointLE.z();

    const double dx_TE = second_pointTE.x() - first_pointTE.x();
    const double dz_TE = second_pointTE.z() - first_pointTE.z();

    sweepLE = std::atan2(dx_LE, std::abs(dz_LE)); // The z offset defines the sign of the angle
    sweepTE = std::atan2(dx_TE, std::abs(dz_TE)); // result in [rad]

    /* result is expected to be in [deg] */
    myContrSurf->phi_LE_CoSu = sweepLE * 180 / PI;
    myContrSurf->phi_TE_CoSu = sweepTE * 180 / PI;
}

actuatorPFCS::actuatorPFCS()
    :
    powerSource(""),
    powerSourceID(0),
    operationMode(""),
    refPoint({0., 0., 0.}),
         Mass(0.0),
         designPower(0.0),
         power(0.0),
efficiency(0.0) {}

controlSurfaceATA27::controlSurfaceATA27()
    :
    massControl(0.0),
    c_S(0.0),
    type(""),
    S(0.0),
    maxDefSpeed(0.0),
    designPowerActuator(0.0),
    maxHingeMoment(0.0),
    controlSurfaceDesignPower(0.0),
    S_surface(0.0),
    MAC_surface(0.0),
    deltaC_H_0(0.0),
    numberOfActuators(0) {}

thsaControlSurface::thsaControlSurface():
    controlSurfaceATA27() {
}

detailedControlSurface::detailedControlSurface():
    controlSurfaceATA27(),
    phi_LE_CoSu(0.0),
    phi_TE_CoSu(0.0),
    relPosOfContrSurfHingeLine(0.0),
    spanDistIBLE2CenterOfArea(0.0),
    chordDistIBLE2CenterOfArea(0.0),
    distCoA2HingeLine(0.0),
    areaMomentAboutHingeLine(0.0),
    s_element(0.0),
    y_i_abs(0.0),
    y_a_abs(0.0),
    l_i_abs(0.0),
    l_a_abs(0.0),
    l_i_rel_TE(0.0),
    l_a_rel_TE(0.0),
    maxDeflection(0.0),
    localWingChordAtContrSurfIB(0.0),
    localWingChordAtContrSurfOB(0.0) {
//    Vec3 inboardLEPoint={0.0,0.0,0.0}; /* Point of Control Device Leading Edge Inboard in absolut coordinates **/
//    Vec3 outboardLEPoint={0.0,0.0,0.0}; /* Point of Control Device Leading Edge Outbord in absolut coordinates **/
}

verticalControlSurface::verticalControlSurface():
    detailedControlSurface() {
}

horizontalControlSurface::horizontalControlSurface():
    detailedControlSurface() {
}
