/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef SYSTEMS_DESIGN_SRC_AIRCRAFTSYSTEMS_CONVENTIONALATA30_H_
#define SYSTEMS_DESIGN_SRC_AIRCRAFTSYSTEMS_CONVENTIONALATA30_H_

/* The following statements import necessary libraries */
#include <string>                                                       // Import c++ standard library for using string
#include <vector>                                                       // Import c++ standard library for using vector
#include <memory>                                                       // Import c++ standard library for using memory

/* The following statements import necessary header files */
#include "wingAntiIcing/wingAntiIcing.h"                                // Import of header file from wingAntiIcing folder for wingAntiIcing
#include "../powerProfile.h"                                            // Import of header file from src folder for power profile
#include "../aircraftSystem.h"                                          // Import of header file from src folder for aircraft system
#include "../systemsIOData.h"                                           // Import of header file from systemsIOData for accessing data_. Can't be included in
                                                                        // aircraftSystem.h because that would lead to circular dependencies
#include <svl/SVL.h>                                                    // Import of header file from src folder for SVL

/** \class conventionalATA30
 *\brief class that contains all properties of the ice and rain protection system (ATA30)
 */
class conventionalATA30 : public aircraftSystem {
 public:
    int numberOfPowerSources; /**< number of power sources for ATA30 */
    std::vector<PowerSource> powerSources;/**< vector containing the power sources of ATA30 */
    double operatingAltitude;/**< [m] Flight altitude up to which the wings are de-iced*/
    double engineAntiIce;/**< [kg/s] Bleed Air mass flow for anti-icing per engine nacelle */
    double wingSkinThickness;/**< [m] thickness of the wing's skin **/
    std::string wingRelSpanAntiIcingStart_str;/**< [-] Relative half span width at the start of de-icing.**/
        //If DEFAULT, rel. pos of first LE device after kink (with 5% safety margin) is used.
    double wingSkinHeatConductivity;/**< [W/mK] heat conductivity of wing's skin**/
    double elecPowerDeparture; /**< [W] Electric consumption during Departure**/
    double elecPowerCruise; /**< [W] Electric consumption during Cruise**/
    double elecPowerApproach; /**< [W] Electric consumption during Approach**/
    double elecPowerLand; /**< [W] Electric consumption during Land**/
    bool wingAntiIcingSwitch;/**< Switch for operation wing anti icing [-] **/
    double efficiencyOfElectroThermalSystem; /**< [-] efficiency of the electro thermyl system */
    double designDropDiameter; /**< [micro m] Water drop diameter **/
    double percentageOfOME;/**< Percentage of OME for mass calculation [-] **/
    double temperatureBleedAir;/**< [K] bleed air temperature **/
    double efficiencyFactorBleedAir;/**< [-] bleed air system efficiency factor */

    /**
     * \brief function for reading the system specific parameters from the config
     * 
     * \param config const node&: systemsDesign configuration file
    */
    void readParameters(const node& config) override;
    /**
     * \brief function for updating the system specific values in the aircraft xml
    */
    void updateXML() override;
    /** \brief calculates the power of the previous systems
     * \param allPreviousSystemsBleed powerProfile
     * \param allPreviousSystemsHydraulic vector<powerProfile>
     * \param allPreviousSystemsElectric vector<powerProfile>
     * \param allPreviousSystemsHeat powerProfile
     * \return void
     */
    void calculateSystemsPower(const powerProfile& allPreviousSystemsBleed,
                               const std::vector<powerProfile>& allPreviousSystemsHydraulic,
                               const std::vector<powerProfile>& allPreviousSystemsElectric,
                               const powerProfile& allPreviousSystemsHeat) override;

    /** \brief calculates the mass of the power sinks
     * \param sourceBleed powerProfile
     * \param sourceElectric vector<powerProfile>
     * \param sourceHydraulic vector<powerProfile>
     * \return void
     */
    void calculateSinkMass(const powerProfile& sourceBleed,
                           const std::vector<powerProfile>& sourceElectric,
                           const std::vector<powerProfile>& sourceHydraulic) override;

    /** \brief calculates the power of the ice and rain protection system (ATA30)
     * \return void
     */
    void getPower();

    /** \brief calculates the mass of the ice and rain protection system (ATA30)
     * \return double: mass of the ice and rain protection system [kg]
     */
    double getMass();

    /** \brief calculates the reference point in x, y, and z-direction (global coordinates) of the ice and rain protection system (ATA30)
     * \param fuselageRef Vec3
     * \param fuselageLength double: [m]
     * \return Vec3: [m] coordinates of the system reference point
     */
    Vec3 getRefPoint();

    /** \brief Constructor
     * \param systemName string
     * \param systemIsOperating bool: switch whether system is operating (1) or not(0)
     * \param theSettings const systemsDesignSettings&
     * \param myAircraft aircraft
     */
    conventionalATA30(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO);
    virtual ~conventionalATA30();

 private:
};

#endif // SYSTEMS_DESIGN_SRC_AIRCRAFTSYSTEMS_CONVENTIONALATA30_H_
