/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "conventionalATA32.h"

#include <unitConversion/constants.h>
#include <standardFiles/functions.h>
#include <runtimeInfo/runtimeInfo.h>

conventionalATA32::conventionalATA32(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO)
    :
    aircraftSystem(systemName, systemIsOperating, data, rtIO) {
    ATA = "ATA-32";
    myRuntimeInfo->out << "  Create: " << Name << "  ->  " << ATA << std::endl;
    //ctor
}

conventionalATA32::~conventionalATA32() {
    //dtor
}

/* Reading system specific parameters of configuration xml file */
void conventionalATA32::readParameters(const node& config) {
    scaling_factor = EndnodeReadOnly<double>("/module_configuration_file/program_settings/scaling_factors/landing_gear_system").read(config).value();
    efficiencyFactor = config.at("/module_configuration_file/program_settings/systems_constants/gear/efficiency_factor/value");
    retractionTime = config.at("/module_configuration_file/program_settings/systems_constants/gear/retraction_time/value");
    extensionTime = config.at("/module_configuration_file/program_settings/systems_constants/gear/extension_time/value");
    data_->configuration.checkForValidity(efficiencyFactor, 0.0, 1.0, "Gear.efficiencyFactor");
    numberOfPowerSources = config.at("/module_configuration_file/program_settings/systems_constants/gear/shaft_power_sources/number_of_power_sources/value");
    double operationFactor_gesamt(0.);
    for (int i(0); i < numberOfPowerSources; i++) {
        powerSources.push_back(PowerSource());
        powerSources.back().type =
            std::string(config.at("/module_configuration_file/program_settings/systems_constants/gear/shaft_power_sources/power_source@" + num2Str(i) + "/type/value"));
        powerSources.back().sourceID =
            config.at("/module_configuration_file/program_settings/systems_constants/gear/shaft_power_sources/power_source@" + num2Str(i) + "/source_ID/value");
        powerSources.back().operationFactor =
            config.at("/module_configuration_file/program_settings/systems_constants/gear/shaft_power_sources/power_source@" + num2Str(i) + "/operation_factor/value");
        operationFactor_gesamt += Rounding(powerSources.back().operationFactor, 10);
    }
    if (operationFactor_gesamt < 1.) {
        myRuntimeInfo->err << "The work distribution of the \"PowerSources\" for the LandingGear results together in less than 100% ("
                           << operationFactor_gesamt * 100 << "%). adjust in config. Abort!" <<
                           std::endl;
        exit(1);
    } else if (operationFactor_gesamt > 1.) {
        myRuntimeInfo->err << "The work distribution of the \"PowerSources\" for the LandingGear results together in more than 100% ("
                           << operationFactor_gesamt * 100 << "%). adjust in config. Abort!" <<
                           std::endl;
        exit(1);
    }
}

/* Updating system specific parameters in aircraft xml file */
void conventionalATA32::updateXML() {
    //not written to acxml since calculated by landingGearDesign
}

void conventionalATA32::calculateSystemsPower(const powerProfile& allPreviousSystemsBleed,
        const std::vector<powerProfile>& allPreviousSystemsHydraulic,
        const std::vector<powerProfile>& allPreviousSystemsElectric,
        const powerProfile& allPreviousSystemsHeat) {
    /* Define unused function parameters to avoid linter warnings */
    UNUSED(allPreviousSystemsBleed);
    UNUSED(allPreviousSystemsHydraulic);
    UNUSED(allPreviousSystemsElectric);
    UNUSED(allPreviousSystemsHeat);
    /* Calculate Systems Power */
    initializePowerProfile();
    if (Operating == true) {
        getPower();
    }
    calculateMaxMissionPower();
}
void conventionalATA32::calculateSinkMass(const powerProfile& sourceBleed,
        const std::vector<powerProfile>& sourceElectric,
        const std::vector<powerProfile>& sourceHydraulic) {
    /* Define unused function parameters to avoid linter warnings */
    UNUSED(sourceBleed);
    UNUSED(sourceElectric);
    UNUSED(sourceHydraulic);
    /* Calculate Systems Mass */
    if (data_->data.Mass.Gear > 0.0) {
        Mass = data_->data.Mass.Gear;
        refPoint = data_->data.CofG.landingGear;
    } else {
        Mass = scaling_factor * getMass();
        refPoint = getRefPoint();
    }
}
void conventionalATA32::getPower() {
    //Initializing auxiliary variables
    double Efficiency(efficiencyFactor);
    double gearRetractionTime(retractionTime);
    double gearExtensionTime(extensionTime);
    double massGear(data_->data.Mass.Gear);
    double retractionPower(0.);
    double extensionPower(0.);
    double gearStrutLength(0.);//average strut length of the landing gear
    int mainGearCount(0);//counter for the main gears
    /*Calculation of the strut length as average value of all strut lengths*/
    for (size_t i(0); i < data_->theLandingGear.size(); i++) {
        //Calculation strictly speaking only applies to Main Gear, the most massive part of the landing gear
        //--> Strut Length is only included in the calculation for Main Gear
        if (compareStrings(data_->theLandingGear.at(i).name, "main", false) == true) {
            mainGearCount++;
            gearStrutLength += data_->theLandingGear.at(i).strut_length;
        }
    }
    if (mainGearCount == 0) {
        myRuntimeInfo->warn << "There was no main landing gear found. This might lead to errors in the calculation of hydraulic system power." << std::endl;
    }
    gearStrutLength *= 1 / static_cast<double>(mainGearCount);
    retractionPower =        4. * (2. * G_FORCE * massGear * gearStrutLength) / gearRetractionTime;
    extensionPower  = 0.75 * 4. * (2. * G_FORCE * massGear * gearStrutLength) / gearExtensionTime;;
    for (size_t n(0); n < Hydraulic.size(); n++) {
        double designPower_tmp(0.);
        designPower_tmp = retractionPower / Efficiency;
        for (size_t j(0); j < powerSources.size(); j++) {
            double operationFactor_tmp(powerSources.at(j).operationFactor);
            if (powerSources.at(j).type == "Hydraulic" && powerSources.at(j).sourceID == static_cast<int>(n + 1)) {
                Hydraulic.at(n).designPower += designPower_tmp * operationFactor_tmp;
            }
        }
    }
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        bleedAir.missionPower.at(i).baseLoad     = 0.0;
        for (size_t n(0); n < Hydraulic.size(); n++) {
            Hydraulic.at(n).missionPower.at(i).baseLoad  = 0.0;
            if (i > 0 && i < data_->mission.Steps.size() - 1) {
                // mission segment in which the landing gear is deployed
                double peakLoad_tmp;
                if (data_->mission.Steps.at(i - 1).modeStep == "takeoff") { //should always be deployed in 2nd mission segment
                    peakLoad_tmp = retractionPower / (Efficiency);
                    //Create and calculate peak loads of the hydraulic demand
                    Hydraulic.at(n).missionPower.at(i).peakLoad.push_back(0);
                    Hydraulic.at(n).missionPower.at(i).timePeak.push_back(gearRetractionTime);
                    for (size_t j(0); j < powerSources.size(); j++) {
                        double operationFactor_tmp(powerSources.at(j).operationFactor);
                        if (powerSources.at(j).type == "Hydraulic" && powerSources.at(j).sourceID == static_cast<int>(n + 1)) {
                            Hydraulic.at(n).missionPower.at(i).peakLoad.back() += peakLoad_tmp * operationFactor_tmp;
                        }
                    }
                } else if (data_->mission.Steps.at(i - 1).configuration == "approach" && data_->mission.Steps.at(i).configuration == "approach_landing_gear_out") {
                    peakLoad_tmp = extensionPower / (Efficiency);
                    //Create and calculate peak loads of the hydraulic demand
                    Hydraulic.at(n).missionPower.at(i).peakLoad.push_back(0);
                    Hydraulic.at(n).missionPower.at(i).timePeak.push_back(gearExtensionTime);
                    for (size_t j(0); j < powerSources.size(); j++) {
                        double operationFactor_tmp(powerSources.at(j).operationFactor);
                        if (powerSources.at(j).type == "Hydraulic" && powerSources.at(j).sourceID == static_cast<int>(n + 1)) {
                            Hydraulic.at(n).missionPower.at(i).peakLoad.back() += peakLoad_tmp * operationFactor_tmp;
                        }
                    }
                }
            }
        }
        /*Loop over electric circuits*/
        for (size_t n(0); n < Electric.size(); n++) {
            Electric.at(n).missionPower.at(i).baseLoad     = 0.0;
        }
        heatLoad.missionPower.at(i).baseLoad     = 0.0;
    }
    //Clean up peaks
    for (size_t n(0); n < Hydraulic.size(); n++) {
        cleanPowerPeaks(&Hydraulic.at(n));
    }
    for (size_t n(0); n < Electric.size(); n++) {
        cleanPowerPeaks(&Electric.at(n));
    }
    cleanPowerPeaks(&heatLoad);
    cleanPowerPeaks(&bleedAir);
}
double conventionalATA32::getMass() {
    //TODO(seabrooke): check that this is the landing gear actuation mass
    return  static_cast<double>(0.00271 * pow(data_->data.Mass.MLM, 1.24));
}

Vec3 conventionalATA32::getRefPoint() {
    return Vec3(data_->data.CofG.landingGear);
}
