/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "conventionalATA49.h"

#include <unitConversion/constants.h>
#include <standardFiles/functions.h>
#include <runtimeInfo/runtimeInfo.h>

conventionalATA49::conventionalATA49(const std::string& systemName, bool systemIsOperating,
                                     std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO)
    :
    aircraftSystem(systemName, systemIsOperating, data, rtIO) {
    ATA = "ATA-49";
    myRuntimeInfo->out << "  Create: " << Name << "  ->  " << ATA << std::endl;
    //ctor
}

conventionalATA49::~conventionalATA49() {
    //dtor
}

/* Reading system specific parameters of configuration xml file */
void conventionalATA49::readParameters(const node& config) {
    scaling_factor = EndnodeReadOnly<double>("/module_configuration_file/program_settings/scaling_factors/auxiliary_power_unit").read(config).value();
    std::string APUConfig = "/module_configuration_file/program_settings/systems_constants/auxiliary_power_unit";
    posXFuselage   = config.at(APUConfig + "/position_of_auxiliary_power_unit_in_fuselage_length/value");
    data_->configuration.checkForValidity(posXFuselage, 0.0, 1.0, "APU.posXFuselage");
    percentageBleedAir          = config.at(APUConfig + "/percentage_bleed_air/value");
    data_->configuration.checkForValidity(percentageBleedAir, 0.0, 1.0, "APU.percentageBleedAir");
    percentageElectricPower     = config.at(APUConfig + "/percentage_electrical_power/value");
    data_->configuration.checkForValidity(percentageElectricPower, 0.0, 1.0, "APU.percentageElectricPower");
    percentageHydraulicPower    = config.at(APUConfig + "/percentage_hydraulic_power/value");
    data_->configuration.checkForValidity(percentageHydraulicPower, 0.0, 1.0, "APU.percentageHydraulicPower");
    efficiencyFactorBleedAir    = config.at(APUConfig + "/efficiency_factor_bleed_air/value");
    data_->configuration.checkForValidity(efficiencyFactorBleedAir, 0.0, 1.0, "APU.efficiencyFactorBleedAir");
    designBleedAir          = config.at(APUConfig + "/design_loads/percentage_bleed_air/value");
    data_->configuration.checkForValidity(designBleedAir, 0.0, 1.0, "APU.designBleedAir");
    designElectricPower     = config.at(APUConfig + "/design_loads/percentage_electrical_power/value");
    data_->configuration.checkForValidity(designElectricPower, 0.0, 1.0, "APU.designElectricPower");
    designHydraulicPower    = config.at(APUConfig + "/design_loads/percentage_hydraulic_power/value");
    data_->configuration.checkForValidity(designHydraulicPower, 0.0, 1.0, "APU.designHydraulicPower");
    installationFactor     = config.at(APUConfig + "/installation_factor/value");
    data_->configuration.checkForValidity(installationFactor, 1.0, 3.0, "APU.installationFactor");
}

/* Updating system specific parameters in aircraft xml file */
void conventionalATA49::updateXML() {
    updateMass("auxiliary_power_unit", Mass, ATA);
}

void conventionalATA49::calculateSourcePower(const powerProfile& sourceBleed,
        const std::vector<powerProfile>& sourceElectric,
        const std::vector<powerProfile>& sourceHydraulic) {
    initializePowerProfile();
    getPower(sourceBleed, sourceElectric, sourceHydraulic);
    calculateMaxMissionPower();
}
void conventionalATA49::calculateSourceMass(const powerProfile& sourceBleed,
        const std::vector<powerProfile>& sourceElectric,
        const std::vector<powerProfile>& sourceHydraulic) {
    Mass = scaling_factor * getMass(sourceBleed, sourceElectric, sourceHydraulic);
    refPoint = getRefPoint();
}

void conventionalATA49::getPower(powerProfile theBleed, std::vector<powerProfile> theElectric, std::vector<powerProfile> theHydraulic) {
    // Negating the total power load for energy balance
    powerNegation(&theBleed);
    for (size_t n(0); n < theElectric.size(); n++) {
        powerNegation(&theElectric.at(n));
    }
    /*Loop over hydraulic circuits*/
    for (size_t n(0); n < theHydraulic.size(); n++) {
        powerNegation(&theHydraulic.at(n));
    }
    bleedAir.designPower  = (percentageBleedAir / efficiencyFactorBleedAir) * theBleed.designPower;
    for (size_t n(0); n < theElectric.size(); n++) {
        for (size_t m(0); m < data_->configuration.Electric.circuits.at(n).generators.size(); m++) {
            double genOperationFactor = data_->configuration.Electric.circuits.at(n).generators.at(m).operationFactor;
            double genEfficiency = data_->configuration.Electric.circuits.at(n).generators.at(m).efficiency;
            Electric.at(n).designPower  += genOperationFactor * (percentageElectricPower / genEfficiency) * theElectric.at(n).designPower;
        }
    }
    /*Loop over hydraulic circuits*/
    for (size_t n(0); n < theHydraulic.size(); n++) {
        for (size_t m(0); m < data_->configuration.Hydraulic.circuits.at(n).pumps.size(); m++) {
            double pumpOperationFactor = data_->configuration.Hydraulic.circuits.at(n).pumps.at(m).operationFactor;
            double pumpEfficiency = data_->configuration.Hydraulic.circuits.at(n).pumps.at(m).efficiency;
            Hydraulic.at(n).designPower  += pumpOperationFactor * (percentageHydraulicPower / pumpEfficiency) * theHydraulic.at(n).designPower;
        }
    }
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        // Bleed AiR
        bleedAir.missionPower.at(i).baseLoad = (percentageBleedAir / efficiencyFactorBleedAir) * theBleed.missionPower.at(i).baseLoad;
        for (size_t j = 0; j < theBleed.missionPower.at(i).peakLoad.size(); j++) {
            bleedAir.missionPower.at(i).peakLoad.push_back((percentageBleedAir / efficiencyFactorBleedAir) * theBleed.missionPower.at(i).peakLoad.at(j));
            bleedAir.missionPower.at(i).timePeak.push_back(theBleed.missionPower.at(i).timePeak.at(j));
        }
    }
    for (size_t n(0); n < theElectric.size(); n++) {
        for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
            // Electric
            for (size_t m(0); m < data_->configuration.Electric.circuits.at(n).generators.size(); m++) {
                double genOperationFactor = data_->configuration.Electric.circuits.at(n).generators.at(m).operationFactor;
                double genEfficiency = data_->configuration.Electric.circuits.at(n).generators.at(m).efficiency;
                Electric.at(n).missionPower.at(i).baseLoad += genOperationFactor * (percentageElectricPower / genEfficiency) * theElectric.at(n).missionPower.at(i).baseLoad;
            }
            for (size_t j = 0; j < theElectric.at(n).missionPower.at(i).peakLoad.size(); j++) {
                bool existingStep{false};
                for (size_t k = 0; k < Electric.at(n).missionPower.at(i).peakLoad.size(); k++) {
                    if (accuracyCheck(Electric.at(n).missionPower.at(i).timePeak.at(k), theElectric.at(n).missionPower.at(i).timePeak.at(k), ACCURACY_HIGH)) {
                        existingStep = true;
                        for (size_t m(0); m < data_->configuration.Electric.circuits.at(n).generators.size(); m++) {
                            double genOperationFactor = data_->configuration.Electric.circuits.at(n).generators.at(m).operationFactor;
                            double genEfficiency = data_->configuration.Electric.circuits.at(n).generators.at(m).efficiency;
                            Electric.at(n).missionPower.at(i).peakLoad.at(k) += genOperationFactor * (percentageElectricPower / genEfficiency)
                                    * theElectric.at(n).missionPower.at(i).peakLoad.at(j);
                        }
                        break;
                    } else {
                        existingStep = false;
                    }
                }
                if (existingStep == false) {
                    Electric.at(n).missionPower.at(i).peakLoad.push_back(0.);
                    Electric.at(n).missionPower.at(i).timePeak.push_back(theElectric.at(n).missionPower.at(i).timePeak.at(j));
                    for (size_t m(0); m < data_->configuration.Electric.circuits.at(n).generators.size(); m++) {
                        double genOperationFactor = data_->configuration.Electric.circuits.at(n).generators.at(m).operationFactor;
                        double genEfficiency = data_->configuration.Electric.circuits.at(n).generators.at(m).efficiency;
                        Electric.at(n).missionPower.at(i).peakLoad.back() += genOperationFactor * (percentageElectricPower / genEfficiency)
                                * theElectric.at(n).missionPower.at(i).peakLoad.at(j);
                    }
                }
            }
        }
    }
    for (size_t n(0); n < theHydraulic.size(); n++) {
        for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
            // Hydraulic
            for (size_t m(0); m < data_->configuration.Hydraulic.circuits.at(n).pumps.size(); m++) {
                double pumpOperationFactor = data_->configuration.Hydraulic.circuits.at(n).pumps.at(m).operationFactor;
                double pumpEfficiency = data_->configuration.Hydraulic.circuits.at(n).pumps.at(m).efficiency;
                Hydraulic.at(n).missionPower.at(i).baseLoad += pumpOperationFactor * (percentageHydraulicPower / pumpEfficiency) * theHydraulic.at(n).missionPower.at(i).baseLoad;
            }
            for (size_t j = 0; j < theHydraulic.at(n).missionPower.at(i).peakLoad.size(); j++) {
                bool existingStep{false};
                for (size_t k = 0; k < Hydraulic.at(n).missionPower.at(i).peakLoad.size(); k++) {
                    if (accuracyCheck(Hydraulic.at(n).missionPower.at(i).timePeak.at(k), theHydraulic.at(n).missionPower.at(i).timePeak.at(k), ACCURACY_HIGH)) {
                        existingStep = true;
                        for (size_t m(0); m < data_->configuration.Hydraulic.circuits.at(n).pumps.size(); m++) {
                            double pumpOperationFactor = data_->configuration.Hydraulic.circuits.at(n).pumps.at(m).operationFactor;
                            double pumpEfficiency = data_->configuration.Hydraulic.circuits.at(n).pumps.at(m).efficiency;
                            Hydraulic.at(n).missionPower.at(i).peakLoad.at(k) += pumpOperationFactor * (percentageHydraulicPower / pumpEfficiency)
                                    * theHydraulic.at(n).missionPower.at(i).peakLoad.at(j);
                        }
                        break;
                    } else {
                        existingStep = false;
                    }
                }
                if (existingStep == false) {
                    Hydraulic.at(n).missionPower.at(i).peakLoad.push_back(0.);
                    Hydraulic.at(n).missionPower.at(i).timePeak.push_back(theHydraulic.at(n).missionPower.at(i).timePeak.at(j));
                    for (size_t m(0); m < data_->configuration.Hydraulic.circuits.at(n).pumps.size(); m++) {
                        double pumpOperationFactor = data_->configuration.Hydraulic.circuits.at(n).pumps.at(m).operationFactor;
                        double pumpEfficiency = data_->configuration.Hydraulic.circuits.at(n).pumps.at(m).efficiency;
                        Hydraulic.at(n).missionPower.at(i).peakLoad.back() += pumpOperationFactor * (percentageHydraulicPower / pumpEfficiency)
                                * theHydraulic.at(n).missionPower.at(i).peakLoad.at(j);
                    }
                }
            }
        }
    }
    //Clean up peaks
    for (size_t n(0); n < Hydraulic.size(); n++) {
        cleanPowerPeaks(&Hydraulic.at(n));
    }
    for (size_t n(0); n < Electric.size(); n++) {
        cleanPowerPeaks(&Electric.at(n));
    }
    cleanPowerPeaks(&bleedAir);
}

double conventionalATA49::getMass(const powerProfile& theBleed,
                                  const std::vector<powerProfile>& theElectric,
                                  const std::vector<powerProfile>& theHydraulic) {
    // NOTE (aigner#1#): theHydraulic is temporarily accumulated over the number of hydraulic circuits
    powerProfile theHydraulicSum(getPowerProfileSum(theHydraulic));
    // NOTE (aigner#1#): theElectric is temporarily accumulated over the number of electric circuits
    powerProfile theElectricSum(getPowerProfileSum(theElectric));
// TODO(peter#1#): Mantis ID 0000213: Neue Modellierung der APU-Nutzleistung ...
//- Recherche ob LTH Nutzlastdaten NUR Wellenleistung sind (vgl. reale Datenblaetter)
//- Berechnung der bleedpower nur thermisch (cp * delta T), beruecksichtigung der Kompressionsarbeit fehlt
//- Massenabschaetzung ueberarbeiten
    double bleedPower(0.0); // DA Steinke, p.48, eq. 6.7, with cp = 1005. [kJ/
    bleedPower = theBleed.designPower * 1005. * 200.;
    double effectivePower(0.0);    // DA Steinke, p.48 in [kW]!!!
    effectivePower = (bleedPower * designBleedAir / efficiencyFactorBleedAir
                      +   theElectricSum.designPower * designElectricPower
                      +   theHydraulicSum.designPower * percentageHydraulicPower) / 1000.;
    double APUMass(0.);
    // F. Peter 22.7.4: new regression of the APU power mass [kg/kW] according to LTH data (p. APU-Massenabschaetzung.xls -> sheet LTH_red_2)
    APUMass = installationFactor * (6.8694073751 * pow(effectivePower, -0.4311907202) * effectivePower); //Regression fuer Leistungsgewicht
    // alternatively: linear regression of the mass: APUMass = installationFactor * (0.28198117 * effectivePower + 89.85363477);
    // myRuntimeInfo->debug << std::endl << "APU Design Parameters:" << std::endl;
    // myRuntimeInfo->debug << "theBleed.designPower: " << theBleed.designPower << std::endl;
    // myRuntimeInfo->debug << "bleedPower: " << bleedPower << std::endl;
    // myRuntimeInfo->debug << "bleedPower * designBleedAir/efficiencyFactorBleedAir: " << bleedPower * designBleedAir/efficiencyFactorBleedAir / 1000.<< std::endl;
    // myRuntimeInfo->debug << "theElectric.designPower: " << theElectricSum.designPower << std::endl;
    // myRuntimeInfo->debug << "theElectric.designPower * designElectricPower: " << theElectricSum.designPower * designElectricPower / 1000.<< std::endl;
    // myRuntimeInfo->debug << "theHydraulicSum.designPower: " << theHydraulicSum.designPower << std::endl;
    // myRuntimeInfo->debug << "theHydraulicSum.designPower * percentageHydraulicPower: " << theHydraulicSum.designPower * percentageHydraulicPower / 1000.<< std::endl;
    // myRuntimeInfo->debug << "APU effectivePower in kW: " << effectivePower << std::endl;
    // myRuntimeInfo->debug << "installationFactor: " << installationFactor << std::endl;
    // myRuntimeInfo->debug << "APUMass: " << APUMass << std::endl;
    // myRuntimeInfo->debug << std::endl;
    // myRuntimeInfo->debug << "APU Power in kW (shaft+bleed): " << effectivePower << std::endl;
    // myRuntimeInfo->debug << "APU Mass in kg: " << APUMass << std::endl;
    return  APUMass;
}

Vec3 conventionalATA49::getRefPoint() {
    Vec3 cgPoint;
    cgPoint[0] = data_->theFuselage.back().origin[0] + static_cast<double>(posXFuselage * geom2::measure::length(data_->theFuselage.back()));
    cgPoint[1] = data_->theFuselage.back().origin[1];
    cgPoint[2] = data_->theFuselage.back().origin[2];
    return cgPoint;
}

void conventionalATA49::powerNegation(powerProfile *Power) {
    Power->maxMissionPower = (-1.0) * Power->maxMissionPower;
    Power->designPower = (-1.0) * Power->designPower;
    for (size_t i = 0; i < Power->missionPower.size(); i++) {
        Power->missionPower.at(i).baseLoad = (-1.0) * Power->missionPower.at(i).baseLoad;
        for (size_t j = 0; j < Power->missionPower.at(i).peakLoad.size(); j++) {
            Power->missionPower.at(i).peakLoad.at(j) = (-1.0) * Power->missionPower.at(i).peakLoad.at(j);
        }
    }
}
