/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "ActuationAndSupport.h"

#include <cmath>

#include <runtimeInfo/runtimeInfo.h>
#include <unitConversion/constants.h>

ActuationAndSupport::ActuationAndSupport(const std::string& Type, double _l_m_Panel, double _S_ref, int _numberActuators, double _Density, double _Speed)
    :
    l_m_Panel(_l_m_Panel),
    S_ref(_S_ref),
    numberActuators(_numberActuators),
    Density(_Density),
    Speed(_Speed),
    deploymentXFlap(0.0),
    deploymentZFlap(0.0),
    deploymentTrack(0.0),
    b1(0.0),
    a(0.0),
    deltaXActuator(0.0),
    deltaZActuator(0.0),
    xA(0),
    yA(0),
    xB(0),
    yB(0),
    xC(0),
    yC(0),
    xD(0),
    yD(0),
    delta_rad(0.0),
    r_Pinion(0.0),
    r_zs(0.0),
    schranke(0.000000001),
    actuatorRotationAngle_Begin(0.0),
    actuatorRotationAngle_End(0.0),
    Mass(0.0),
    maxActuationMoment(0.0),
    actuationWork(0.0),
    actuationPower(0.0) {
    //Ebene Plattenloesung
    liftCoefficient.m = 6.14236;
    liftCoefficient.c = 0.0;
    momentCoefficient.m = -1.4333;
    momentCoefficient.c = 0.0;
    dragCoefficient.m = 0.0;
    dragCoefficient.c = 0.0;
    //Assignment of the configurations and kinematics
    //Method for calculating the power of ActuationAndSupport from [Gro09] as of p.32
    if (Type == "Slat") {
        efficiencyFactor = 0.8924;
        rotationSpeedActuator = 0.6354;
        deviceAoA = 13.5 * PI / 180;
        k1 = 0.0;
        k2 = 0.0;
        alphaA = -12 * PI / 180;
        alphaE = -36 * PI / 180;
        gamma = 19 * PI / 180;
        i_Pinion = 25.73;
        timeForDeployment = 22; //[s]
        double scalingX(0.4288);         //Scaling factor for extension in x-direction
        double scalingZ(0.3974);         //Scaling factor for extension in z-direction
        // Bestimmen der Ausfahrwege
        deploymentXFlap = scalingX * l_m_Panel;
        deploymentZFlap = scalingZ * l_m_Panel;
        rotationDirection = "";
        calculateKinematic_Slat();
    } else if (Type == "Fowler") {
        efficiencyFactor = 0.8746;
        rotationSpeedActuator = 0.12852; //[s]
        deviceAoA = -2.8 * PI / 180;
        k1 = 0.6616;
        k2 = -2.125;
        alphaA = 1 * PI / 180;
        alphaE = 35 * PI / 180;
        gamma = 30 * PI / 180;
        i_Pinion = 0.0;
        timeForDeployment = 16; //[s]
        double scalingX(0.3331);        //Scaling factor for extension in X-direction
        double scalingZ(0.07246);       //Scaling factor for extension in z-direction
        double scalingTrack(0.09474);   //Scaling factor distance to flap track
        /*Determination of the extension paths*/
        deploymentXFlap = scalingX * l_m_Panel;
        deploymentZFlap = scalingZ * l_m_Panel;
        deploymentTrack = scalingTrack * l_m_Panel;
        deltaXActuator = k1 * deploymentXFlap;
        deltaZActuator = k2 * deploymentZFlap;
        xC = xA + deltaXActuator;
        yC = yA - deltaZActuator;
        xD = xC + deploymentXFlap;
        yD = yC - deploymentZFlap;
        delta_rad = atan(fabs(yD - yC) / fabs(xD - xC));
        rotationDirection = "right";
        calculateKinematic_Flap();
    } else {
        myRuntimeInfo->err << "This system is unknown. No actuator mass can be determined." << std::endl;
        myRuntimeInfo->err << "System-Type: " << Type << std::endl;
        myRuntimeInfo->err << "Abort!" << std::endl;
        exit(1);
    }
    //ctor
}

ActuationAndSupport::~ActuationAndSupport() {
    //dtor
}

void ActuationAndSupport::getMass(const std::string& deviceType, double Sref, int Stations) {
    // NOTE (aigner#1#): Initially from SA Chavez 2010 S.69-70, entirely new method by Lammering in SVN rev2131 March 2013
    // Calculates the Gearboxes for each station, not the kinematics (this is done in massEstimation)
    if (deviceType == "Slat") {
        Mass = 0.614 * 1.30 * 0.4535924 / pow(0.3048, 2.) * Sref / static_cast<double>(Stations);
    } else if (deviceType == "Fowler") {
        Mass = 0.284 * 2.0 * 0.4535924 / pow(0.3048, 2.) * Sref / static_cast<double>(Stations);
    } else {
        myRuntimeInfo->err << "Dieses System ist nicht bekannt. Kann keine Aktuator-Masse bestimmen!" << std::endl;
        myRuntimeInfo->err << "System-Typ: " << deviceType << std::endl;
        exit(1);
    }
}

void ActuationAndSupport::calculateKinematic_Flap() {
    betaA = gamma - alphaA;
    betaE = gamma - alphaE;
    // TODO(aigner#1#): Comment from Lammering: "//??? Bitte kommentieren, z.B. mit Seitenverweis auf Studienarbeit von Stefan oder Dir"
    double c = xC - xA - sin(betaA) * deploymentTrack;
    double d = cos(betaA);
    double e = yC - yA + cos(betaA) * deploymentTrack;
    double f = sin(betaA);
    double g = xD - xA - sin(betaE) * deploymentTrack;
    double h = cos(betaE);
    double i = yD - yA + cos(betaE) * deploymentTrack;
    double j = sin(betaE);
    // TODO(aigner#1#): Comment from Lammering: "//Was ist b1"
    b1 = (pow(c, 2) + pow(e, 2) - pow(g, 2) - pow(i, 2)) / 2 / (c * d + e * f - g * h - i * j);
    // TODO(aigner#1#): Comment from Lammering: "//Was ist a"
    double a1 = sqrt(pow(c - d * b1, 2) + pow(e - f * b1, 2));
    double a2 = sqrt(pow(g - h * b1, 2) + pow(i - j * b1, 2));
    // TODO(aigner#1#): Comment from Lammering: "//Wodrum geht es hier???"
    if (fabs(a1 - a2) <= schranke) {
        a = a1;
    } else {
        myRuntimeInfo->err << "Bei der Berechnung der Groesse a der Aktuatorkinematik ist ein unbehebbarer Fehler aufgetreten" << std::endl;
        exit(1);
    }
    /*Calculating the possibilities for phiA*/
    double phiA1_rad = asin((c - d * b1) / a);
    double phiA2_rad = asin(-(c - d * b1) / a);
    double phiA3_rad = acos((e - f * b1) / a);
    double phiA4_rad = acos(-(e - f * b1) / a);
    double phiA5_rad;
    double phiA6_rad;
    double phiA7_rad;
    double phiA8_rad;
    // TODO(aigner#1#): Comment from Lammering: "//???"
    if (phiA1_rad < 0)
        phiA5_rad = phiA1_rad + PI;
    else
        phiA5_rad = phiA1_rad - PI;
    if (phiA2_rad < 0)
        phiA6_rad = phiA2_rad + PI;
    else
        phiA6_rad = phiA2_rad - PI;
    if (phiA3_rad < 0)
        phiA7_rad = phiA3_rad + PI;
    else
        phiA7_rad = phiA3_rad - PI;
    if (phiA4_rad < 0)
        phiA8_rad = phiA4_rad + PI;
    else
        phiA8_rad = phiA4_rad - PI;
    /*Calculating the possibilities for phiE*/
    double phiE1_rad = asin((g - h * b1) / a);
    double phiE2_rad = asin(-(g - h * b1) / a);
    double phiE3_rad = acos((i - j * b1) / a);
    double phiE4_rad = acos(-(i - j * b1) / a);
    double phiE5_rad;
    double phiE6_rad;
    double phiE7_rad;
    double phiE8_rad;
    // TODO(aigner#1#): Comment from Lammering: "//???"
    if (phiE1_rad < 0) {
        phiE5_rad = phiE1_rad + PI;
    } else {
            phiE5_rad = phiE1_rad - PI;
    }
    if (phiE2_rad < 0) {
        phiE6_rad = phiE2_rad + PI;
    } else {
        phiE6_rad = phiE2_rad - PI;
    }
    if (phiE3_rad < 0) {
        phiE7_rad = phiE3_rad + PI;
    } else {
        phiE7_rad = phiE3_rad - PI;
    }
    if (phiE4_rad < 0) {
        phiE8_rad = phiE4_rad + PI;
    } else {
        phiE8_rad = phiE4_rad - PI;
    }
    // Chck of the correct angle
    if (angle_test_phiA(phiA1_rad)) {
        actuatorRotationAngle_Begin = phiA1_rad;
    } else {
        if (angle_test_phiA(phiA2_rad)) {
            actuatorRotationAngle_Begin = phiA2_rad;
        } else {
            if (angle_test_phiA(phiA3_rad)) {
                actuatorRotationAngle_Begin = phiA3_rad;
            } else {
                if (angle_test_phiA(phiA4_rad)) {
                    actuatorRotationAngle_Begin = phiA4_rad;
                } else {
                    if (angle_test_phiA(phiA5_rad)) {
                        actuatorRotationAngle_Begin = phiA5_rad;
                    } else {
                        if (angle_test_phiA(phiA6_rad)) {
                            actuatorRotationAngle_Begin = phiA6_rad;
                        } else {
                            if (angle_test_phiA(phiA7_rad)) {
                                actuatorRotationAngle_Begin = phiA7_rad;
                            } else {
                                if (angle_test_phiA(phiA8_rad)) {
                                    actuatorRotationAngle_Begin = phiA8_rad;
                                } else {
                                    myRuntimeInfo->err << "Es konnte kein Winkel phiA gefunden werden, der alle Gleichungen erfuellt." << std::endl;
                                    exit(1);
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    if (angle_test_phiE(phiE1_rad)) {
        actuatorRotationAngle_End = phiE1_rad;
    } else {
        if (angle_test_phiE(phiE2_rad)) {
            actuatorRotationAngle_End = phiE2_rad;
        } else {
            if (angle_test_phiE(phiE3_rad)) {
                actuatorRotationAngle_End = phiE3_rad;
            } else {
                if (angle_test_phiE(phiE4_rad)) {
                    actuatorRotationAngle_End = phiE4_rad;
                } else {
                    if (angle_test_phiE(phiE5_rad)) {
                        actuatorRotationAngle_End = phiE5_rad;
                    } else {
                        if (angle_test_phiE(phiE6_rad)) {
                            actuatorRotationAngle_End = phiE6_rad;
                        } else {
                            if (angle_test_phiE(phiE7_rad)) {
                                actuatorRotationAngle_End = phiE7_rad;
                            } else {
                                if (angle_test_phiE(phiE8_rad)) {
                                    actuatorRotationAngle_End = phiE8_rad;
                                } else {
                                    myRuntimeInfo->err << "No angle phiE could be found that fulfilled all equations. Abort!" << std::endl;
                                    exit(1);
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    /* Adjustment of the angle to the direction of rotation if necessary */
    if (rotationDirection == "right" && actuatorRotationAngle_End < actuatorRotationAngle_Begin) {
        actuatorRotationAngle_End = actuatorRotationAngle_End + 2 * PI;
    }
    if (rotationDirection == "left" && actuatorRotationAngle_End < actuatorRotationAngle_Begin) {
        actuatorRotationAngle_End = actuatorRotationAngle_End - 2 * PI;
    }
}

void ActuationAndSupport::calculateKinematic_Slat() {
    actuatorRotationAngle_Begin = gamma - alphaA;
    actuatorRotationAngle_End = gamma - alphaE;
    if (actuatorRotationAngle_Begin >= actuatorRotationAngle_End) {
        myRuntimeInfo->err << "The angle of the extended slat is greater than that of the retracted slat, which is not permitted by the kinematics. Abort!" << std::endl;
        exit(1);
    }
    r_zs = sqrt(pow(deploymentXFlap, 2) + pow(deploymentZFlap, 2)) / 2 / sin((actuatorRotationAngle_End - actuatorRotationAngle_Begin) / 2);
    if (r_zs <= 0) {
        myRuntimeInfo->err << "The calculated radius of the slat rack is negative. Abort!" << std::endl;
        exit(1);
    } else {
        r_Pinion = r_zs / i_Pinion;
    }
}

// TODO(aigner#1#): Comment from Lammering: "// Bitte Kommentieren, was hier geschieht"
bool ActuationAndSupport::angle_test_phiA(double _phiA) const {
    if (fabs(xC - xA - cos(betaA)*b1 - sin(betaA)*deploymentTrack + sin(_phiA)*a) <= schranke
        && fabs(yC - yA - sin(betaA)*b1 + cos(betaA)*deploymentTrack + cos(_phiA)*a) <= schranke)
        return true;
    else
        return false;
}

bool ActuationAndSupport::angle_test_phiE(double _phiE) const {
    if (fabs(xD - xA - cos(betaE)*b1 - sin(betaE)*deploymentTrack + sin(_phiE)*a) <= schranke
        && fabs(yD - yA - sin(betaE)*b1 + cos(betaE)*deploymentTrack + cos(_phiE)*a) <= schranke)
        return true;
    else
        return false;
}

void ActuationAndSupport::calcActuationMoments(const std::string& deviceType) {
    int steps(100);
    double stepWidth((actuatorRotationAngle_End - actuatorRotationAngle_Begin) / static_cast<double>(steps - 1));
    for (double actuatorAngle = actuatorRotationAngle_Begin; actuatorAngle < actuatorRotationAngle_End; actuatorAngle += stepWidth) {
        actuationMoments.push_back(moment());
        actuationMoments.back().actuationAngle = actuatorAngle;
        if (deviceType == "Slat") {
            actuationMoments.back().actuationMoment = calcMomentSlat(actuatorAngle);
        } else if (deviceType == "Fowler") {
            actuationMoments.back().actuationMoment = calcMomentFlap(actuatorAngle);
        } else {
            myRuntimeInfo->err << "Type is unknown: " << deviceType << std::endl;
            myRuntimeInfo->err << "No actuator torque can be determined. Abort!" << deviceType << std::endl;
            exit(1);
        }
        //myRuntimeInfo->out << actuationMoments.back().actuationAngle*180/PI << " | " << actuationMoments.back().actuationMoment << std::endl;
    }
    maxActuationMoment = getMaxMoment();
    //myRuntimeInfo->out << "Max --> " << maxActuationMoment << std::endl;
    //getchar();
}

void ActuationAndSupport::calcWorkLoad() {
    double tempWork(0.0);
    for (size_t i = 1; i < actuationMoments.size(); i++) {
        tempWork = tempWork + (actuationMoments.at(i).actuationAngle - actuationMoments.at(i - 1).actuationAngle)
                   * 0.5 * (fabs(actuationMoments.at(i).actuationMoment) + fabs(actuationMoments.at(i - 1).actuationMoment));
    }
    actuationWork = tempWork;
}

void ActuationAndSupport::calcPower() {
    actuationPower = actuationWork / timeForDeployment;
}


double ActuationAndSupport::getMaxMoment() {
    double maxMoment(0.0);
    for (size_t i = 0; i < actuationMoments.size(); i++) {
        if (fabs(actuationMoments.at(i).actuationMoment) > maxMoment) {
            maxMoment = fabs(actuationMoments.at(i).actuationMoment);
        }
    }
    return maxMoment;
}

double ActuationAndSupport::calcMomentSlat(double Angle) {
    //double epsilon_rad = deviceAoA;
    double localAoA = deviceAoA - Angle + gamma;
    double drag = (0.5 * Density * pow(Speed, 2) * S_ref * (dragCoefficient.m * localAoA + dragCoefficient.c))
                  / static_cast<double>(numberActuators);
    double lift = (0.5 * Density * pow(Speed, 2) * S_ref * (liftCoefficient.m * localAoA + liftCoefficient.c))
                  / static_cast<double>(numberActuators);
    double moment = (0.5 * Density * pow(Speed, 2) * S_ref * l_m_Panel * (dragCoefficient.m * localAoA + dragCoefficient.c))
                    / static_cast<double>(numberActuators);
    double Rx = cos(deviceAoA) * drag - sin(deviceAoA) * lift;
    double Ry = sin(deviceAoA) * drag + cos(deviceAoA) * lift;
    double RT = cos(Angle) * Rx + sin(Angle) * Ry;
    double m_akt = (RT * r_Pinion - moment * r_Pinion / r_zs);
    return m_akt;
}

double ActuationAndSupport::calcMomentFlap(double Angle) {
    double beta_rad = beta_von_phiFlap(Angle);
    double alpha_rad = gamma - beta_rad;
    double localAoA = deviceAoA + alpha_rad;
    double drag = (0.5 * Density * pow(Speed, 2) * S_ref * (dragCoefficient.m * localAoA + dragCoefficient.c))
                  / static_cast<double>(numberActuators);
    double lift = (0.5 * Density * pow(Speed, 2) * S_ref * (liftCoefficient.m * localAoA + liftCoefficient.c))
                  / static_cast<double>(numberActuators);
    double moment = (0.5 * Density * pow(Speed, 2) * S_ref * l_m_Panel * (dragCoefficient.m * localAoA + dragCoefficient.c))
                    / static_cast<double>(numberActuators);
    double Ex = -sin(deviceAoA) * lift + cos(deviceAoA) * drag;
    double Ey = cos(deviceAoA) * lift + sin(deviceAoA) * drag;
    double _a = cos(beta_rad) * b1;
    double _b = sin(beta_rad) * b1;
    double _c = cos(delta_rad) * (cos(beta_rad) * b1 + sin(beta_rad) * deploymentTrack)
                - sin(delta_rad) * (sin(beta_rad) * b1
                                    - cos(beta_rad) * deploymentTrack);
    double _d = sin(delta_rad);
    double _e = cos(delta_rad);
    double Bx = -Ex + _d * (-_b * Ex + _a * Ey + moment) / _c;
    double By = -Ey + _e * (-_b * Ex + _a * Ey + moment) / _c;
    double m_akt;
    if (rotationDirection == "left")
        m_akt = Bx * a * cos(Angle) - By * a * sin(Angle);
    else
        m_akt = -(Bx * a * cos(Angle) - By * a * sin(Angle));
    return m_akt;
}

double ActuationAndSupport::beta_von_phiFlap(double Phi) {
    double x_B = xA - sin(Phi) * a;
    double y_B = yA - cos(Phi) * a;
    double k = tan(delta_rad) * (x_B - xC) + y_B - yC;
    double l = deploymentTrack - tan(delta_rad) * b1;
    double n = b1 + tan(delta_rad) * deploymentTrack;
    double p = -(2 * l * n) / (pow(n, 2) - pow(k, 2));
    double q = (pow(l, 2) - pow(k, 2)) / (pow(n, 2) - pow(k, 2));
    double m1 = -p / 2 + sqrt(pow(p / 2, 2) - q);
    double m2 = -p / 2 - sqrt(pow(p / 2, 2) - q);
    double m;
    if (test_mFlap(m1, k)) {
        m = m1;
    } else {
        if (test_mFlap(m2, k)) {
            m = m2;
        } else {
            myRuntimeInfo->err << "No gradient m could be determined that satisfies the equation." << std::endl;
            return 0;
        }
    }
    double beta = atan(m);
    return beta;
}

bool ActuationAndSupport::test_mFlap(double _m, double _k) const {
    if (fabs(deploymentTrack - b1 * _m - tan(delta_rad)*b1 - tan(delta_rad)*deploymentTrack * _m - _k * sqrt(1 + pow(_m, 2))) <= schranke)
        return true;
    else
        return false;
}

