/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "geometricalFunctions.h"

#include <algorithm>
#include <numeric>

#include <aircraftGeometry2/airfoil_surface.h>
#include <aircraftGeometry2/fuselage.h>
#include <aircraftGeometry2/processing/measure.h>
#include <aircraftGeometry2/processing/transform.h>

#include <runtimeInfo/runtimeInfo.h>
#include <standardFiles/functions.h>
#include <unitConversion/constants.h>

#include "systemsIOData.h"

geometricalFunctions::geometricalFunctions() {
    //ctor
}

geometricalFunctions::~geometricalFunctions() {
    //dtor
}

double geometricalFunctions::getConductorLengthWingEngines(const geom2::MultisectionSurface<geom2::AirfoilSection>& theWing,
                                        const std::vector<geom2::MultisectionSurface<geom2::PolygonSection>>& theNacelles) const { // cppcheck-suppress unusedFunction
    std::vector<Vec3> enginesOnWing;
    Vec3 innerRefPoint;
    std::vector<double> Length;
    innerRefPoint[0] = theWing.origin[0] + 0.1 * theWing.sections.at(0).get_chord_length();
    innerRefPoint[1] = theWing.origin[1];
    innerRefPoint[2] = theWing.origin[2];
    /* Determining the engine positions on the wing**/
    for (size_t i = 0; i < theNacelles.size(); i++) {
        enginesOnWing.push_back(getWingCoodinatesOfNacelle(theNacelles.at(i).origin[1], theWing));
    }
    /* Sorting the nacelles by y-coordinate **/
    sortCoordinates(&enginesOnWing);
    /* Determination of the individual route lengths **/
    /* Length from the wing root to the engine */
    Length.push_back(getPartialLengthWingConductor(innerRefPoint, enginesOnWing.at(0), theWing));
    /* Length between two engines */
    for (int i = 0; i <= static_cast<int>(enginesOnWing.size()) - 2; i++) {
        Length.push_back(getPartialLengthWingConductor(enginesOnWing.at(i), enginesOnWing.at(i + 1), theWing));
    }
    return accumulate(Length.begin(), Length.end(), static_cast<double>(0.0));
}

double geometricalFunctions::getConductorLengthSurfaceTrailingEdge(const geom2::MultisectionSurface<geom2::AirfoilSection>& theWing) { // cppcheck-suppress unusedFunction
    std::vector<double> Length;
    Vec3 Pos1;
    Vec3 Pos2;
    for (int i = 0; i <= static_cast<int>(theWing.sections.size()) - 2; i++) {
        Pos1[0] = theWing.sections.at(i).origin[0] + (- 0.25 + 0.75) * theWing.sections.at(i).get_chord_length();
        Pos1[1] = theWing.sections.at(i).origin[1];
        Pos1[2] = theWing.sections.at(i).origin[2];
        Pos2[0] = theWing.sections.at(i + 1).origin[0] + (- 0.25 + 0.75) * theWing.sections.at(i + 1).get_chord_length();
        Pos2[1] = theWing.sections.at(i + 1).origin[1];
        Pos2[2] = theWing.sections.at(i + 1).origin[2];
        Length.push_back(sqrt(pow(Pos1[0] - Pos2[0], 2) + pow(Pos1[1] - Pos2[1], 2) + pow(Pos1[2] - Pos2[2], 2)));
    }
    return accumulate(Length.begin(), Length.end(), static_cast<double>(0.0));
}

double geometricalFunctions::getConductorLengthFuselageEngines(const geom2::MultisectionSurface<geom2::PolygonSection>& theFuselage,
                                                               const std::vector<geom2::MultisectionSurface<geom2::PolygonSection>>& theNacelles)
                                                               const { // cppcheck-suppress unusedFunction
    std::vector<double> Length;
    for (size_t i = 0; i < theNacelles.size(); i++) {
        // Assumption: no displacement in x-direction
        Length.push_back(sqrt(pow(theNacelles.at(i).origin[1] - theFuselage.origin[1], 2)
                              + pow(theNacelles.at(i).origin[2] - theFuselage.origin[2], 2)));
    }
    return accumulate(Length.begin(), Length.end(), static_cast<double>(0.0));
}

double geometricalFunctions::getConductorLengthFuselageEnginesToSystem(double Pos1, double Pos2,
                                    const std::vector<geom2::MultisectionSurface<geom2::PolygonSection>>& theNacelles) { // cppcheck-suppress unusedFunction
    std::vector<double> Length;
    for (size_t i = 0; i < theNacelles.size(); i++) {
        if (theNacelles.at(i).origin[0] < Pos1) {
            Length.push_back(fabs(Pos1 - theNacelles.at(i).origin[0]));
        } else if (theNacelles.at(i).origin[0] > Pos2) {
            Length.push_back(fabs(Pos2 - theNacelles.at(i).origin[0]));
        }
    }
    return accumulate(Length.begin(), Length.end(), static_cast<double>(0.0));
}

// double geometricalFunctions::getNoseSectionLength(const geom2::MultisectionSurface<geom2::PolygonSection>& theFuselage) const { // cppcheck-suppress unusedFunction
//     double noseSection(0.0);
//     // assumption: 9th section is the end of the nose segment and the length of the segment = x origin of 9th section
//     // @todo: check this assumption or find different definition
//     geom2::Point_3 noseSection_global = geom2::transform::to_parent(theFuselage, theFuselage.sections.at(9).origin); //transform to global coordinates
//     noseSection = noseSection_global.x();
//     return noseSection;
// }

Vec3 geometricalFunctions::getWingCoodinatesOfNacelle(double yPos, const geom2::MultisectionSurface<geom2::AirfoilSection>& theWing) const {
    double xPos(0.0);
    double zPos(0.0);
    /* Determining the coordinates of the nacelles on the wing **/
    // Determining the segment
    int Segment(0);
    for (size_t i = 0; i < theWing.sections.size(); i++) {
        // transform section origin to global coordinate system
        geom2::Point_3 origin_tmp = geom2::transform::to_parent(theWing, theWing.sections.at(i+1).origin);
        if (origin_tmp.y() < yPos) {
            Segment++;
        } else {
            break;
        }
    }
    double m(0.0);
    //transform section origin points to global coordinate system
    myRuntimeInfo->debug << "*** Wing coordinates of nacelle ***" << std::endl;
    myRuntimeInfo->debug << "Section = " << Segment << std::endl;
    geom2::Point_3 segment1_origin = geom2::transform::to_parent(theWing, theWing.sections.at(Segment).origin);
    geom2::Point_3 segment2_origin = geom2::transform::to_parent(theWing, theWing.sections.at(Segment + 1).origin);
    myRuntimeInfo->debug << "Section 1 origin: " << segment1_origin << std::endl;
    myRuntimeInfo->debug << "Section 2 origin: " << segment2_origin << std::endl;
    m = (segment1_origin.x() - segment2_origin.x()) / (segment1_origin.y() - segment2_origin.y());
    myRuntimeInfo->debug << "m = " << m << std::endl;
    xPos = segment1_origin.x() + m * (yPos - segment1_origin.y());
    // approx. front saver and not quarter line
    double localChord = geom2::measure::chord(theWing, yPos);
    xPos -= (0.25 - 0.10) * localChord;
    m = (segment1_origin.z() - segment2_origin.z()) / (segment1_origin.y() - segment2_origin.y());
    zPos = segment1_origin.z() + m * (yPos - segment1_origin.y());
    return Vec3(xPos, yPos, zPos);
}

void geometricalFunctions::sortCoordinates(std::vector<Vec3> *Coordinates) const {
    std::vector<double> yTemp;
    std::vector<double> xTemp;
    std::vector<double> zTemp;
    for (size_t i = 0; i < Coordinates->size(); i++) {
        yTemp.push_back(Coordinates->at(i)[1]);
        xTemp.push_back(0.0);
        zTemp.push_back(0.0);
    }
    sort(yTemp.begin(), yTemp.end());
    for (size_t i = 0; i < yTemp.size(); i++) {
        for (size_t j = 0; j < Coordinates->size(); j++) {
            if (accuracyCheck(yTemp.at(i), Coordinates->at(j)[1], ACCURACY_HIGH)) {
                xTemp.at(i) = Coordinates->at(j)[0];
                zTemp.at(i) = Coordinates->at(j)[2];
                break;
            }
        }
    }
    Coordinates->clear();
    for (size_t i = 0; i < yTemp.size(); i++) {
        Coordinates->push_back(Vec3(xTemp.at(i), yTemp.at(i), zTemp.at(i)));
    }
}

double geometricalFunctions::getPartialLengthWingConductor(Vec3 Pos1, Vec3 Pos2, const geom2::MultisectionSurface<geom2::AirfoilSection>& theWing) const {
    myRuntimeInfo->info << std::endl << "getPartialLengthWingConductor()" << std::endl;
    myRuntimeInfo->info << "Pos1 " << Pos1 << std::endl;
    myRuntimeInfo->info << "Pos2 " << Pos2 << std::endl;
    double length(0.0);
    int startSegment(0);
    int endSegment(0);
    // transform section origin to global coordinate system
    geom2::Point_3 origin_tmp = geom2::transform::to_parent(theWing, theWing.sections.at(startSegment + 1).origin);
    for (int i = startSegment; i < endSegment; i++) {
        geom2::Point_3 origin_tmp = geom2::transform::to_parent(theWing, theWing.sections.at(startSegment + 1).origin);
        if (origin_tmp.y() < Pos1[1]) {
            startSegment++;
        }
        if (origin_tmp.y() < Pos2[1]) {
            endSegment++;
        }
    }
    if (startSegment == endSegment) { // Both points are in the same segment
        length = sqrt(pow(Pos2[0] - Pos1[0], 2) + pow(Pos2[1] - Pos1[1], 2) + pow(Pos2[2] - Pos1[2], 2));
        myRuntimeInfo->info << "Both points are located in the same segment. length = " << length << std::endl;
    } else {
        //1. Point to segment Limit
        length += sqrt(pow((origin_tmp.x() + (0.25 - 0.1) * theWing.sections.at(startSegment + 1).get_chord_length()) - Pos1[0], 2)
                       + pow((origin_tmp.y()) - Pos1[1], 2)
                       + pow((origin_tmp.z()) - Pos1[2], 2));
        myRuntimeInfo->info << "1st point to Segment limit."  << std::endl;
        myRuntimeInfo->info << "length = " << length << std::endl;
        myRuntimeInfo->info << "theWing.sections.at(startSegment+1).origin[0]         = " << origin_tmp.x() << std::endl;
        myRuntimeInfo->info << "(0.25-0.1)*theWing.sections.at(startSegment+1).get_chord_length()) = "
                            << (0.25 - 0.1)*theWing.sections.at(startSegment + 1).get_chord_length() << std::endl;
        myRuntimeInfo->info << "Pos1[0]                                                      = " << Pos1[0] << std::endl;
        myRuntimeInfo->info << "theWing.sections.at(startSegment+1).origin[1]         = " << origin_tmp.y() << std::endl;
        myRuntimeInfo->info << "Pos1[1]                                                      = " << Pos1[1] << std::endl;
        myRuntimeInfo->info << "theWing.sections.at(startSegment+1).origin[2]         = " << origin_tmp.z() << std::endl;
        myRuntimeInfo->info << "Pos1[2]                                                      = " << Pos1[2] << std::endl << std::endl;
        myRuntimeInfo->info << "startSegment = " << startSegment << std::endl;
        myRuntimeInfo->info << "endSegment   = " << endSegment << std::endl;
        for (int i = startSegment + 1; i < endSegment; i++) {
            geom2::Point_3 origin1 = geom2::transform::to_parent(theWing, theWing.sections.at(i).origin);
            geom2::Point_3 origin2 = geom2::transform::to_parent(theWing, theWing.sections.at(i + 1).origin);
            myRuntimeInfo->info << "i " << i << std::endl;
            length += sqrt(pow((origin2.x() + (0.25 - 0.1) * theWing.sections.at(i + 1).get_chord_length()) - (origin1.x() +
                               (0.25 - 0.1) * theWing.sections.at(i).get_chord_length()), 2)
                           + pow((origin2.y()) - (origin1.y()), 2)
                           + pow((origin2.z()) - (origin1.z()), 2));
            myRuntimeInfo->info << "Through the segments. length = " << length << std::endl;
        }
        //last segment limits up to the 2nd point
        geom2::Point_3 origin_end = geom2::transform::to_parent(theWing, theWing.sections.at(endSegment).origin);
        length += sqrt(pow((origin_end.x() + (0.25 - 0.1) * theWing.sections.at(endSegment).get_chord_length()) - Pos2[0], 2)
                       + pow((origin_end.y()) - Pos2[1], 2)
                       + pow((origin_end.z()) - Pos2[2], 2));
        myRuntimeInfo->info << "last segment limit to 2nd point. length = " << length << std::endl;
    }
    myRuntimeInfo->info << "length " << length << std::endl;
    return (length);
}

//cppcheck-suppress unusedFunction
std::unordered_map<std::string, double> geometricalFunctions::getRelSpanAntiIcing(const systemsIOData::Wing& theWing,
                                                                                  const std::string& wingRelSpanAntiIcingStart_str) {
    std::unordered_map<std::string, double> theRelSpanPositionsAntiIcing;

    theRelSpanPositionsAntiIcing["start"] = getRelSpanAntiIcingStart(theWing, wingRelSpanAntiIcingStart_str);
    theRelSpanPositionsAntiIcing["end"] = getRelSpanAntiIcingEnd(theWing);

    if (theRelSpanPositionsAntiIcing["start"] < 0. || theRelSpanPositionsAntiIcing["start"] > 1. || theRelSpanPositionsAntiIcing["end"] < theRelSpanPositionsAntiIcing["start"]) {
        myRuntimeInfo->err << "wingRelSpanAntiIcingStart = " << theRelSpanPositionsAntiIcing["start"] << "; wingRelSpanAntiIcingEnd = "
                           << theRelSpanPositionsAntiIcing["end"] << std::endl;
        throwError(__FILE__, __func__, __LINE__, "wingRelSpanAntiIcingStart or wingRelSpanAntiIcingEnd not in boundaries. Abort program!");
    }

    return theRelSpanPositionsAntiIcing;
}

double geometricalFunctions::getRelSpanAntiIcingStart(const systemsIOData::Wing& theWing, const std::string& wingRelSpanAntiIcingStart_str) const {
    double wingRelSpanAntiIcingStart(0.);
    if (wingRelSpanAntiIcingStart_str == "DEFAULT") {
        for (uint8_t i(0); i < theWing.controlSurface.size(); ++i) {
            if (theWing.controlSurface.at(i).origin[0] < 0.5) { //check that its a leading edge device
                // Assumption: First LE device, which starts after the kink, is the first being de-iced. This fits the A320 architecture (Fig. 3.1 in [Fre11]).
                // 5% safety margin to ensure correct identification even after small shifts after, e.g., wing modification.
                if (theWing.controlSurface.at(i).origin[2] > 1.05 * theWing.aerodynamicSurface.sections.at(1).origin[2]) { // for now: kink is at the second section of the wing
                // in the future it would be useful to determine the start of anti icing according to the nose radius
                    wingRelSpanAntiIcingStart = theWing.controlSurface.at(i).origin[2];
                    break;
                }
            }
        }
    } else {
        wingRelSpanAntiIcingStart = stod(wingRelSpanAntiIcingStart_str);
    }
    return wingRelSpanAntiIcingStart;
}

double geometricalFunctions::getRelSpanAntiIcingEnd(const systemsIOData::Wing& theWing) const {
    double wingRelSpanAntiIcingEnd(0.);
    // go through the control devices and find the outer most leading edge device
    for (size_t i = 0; i < theWing.controlSurface.size(); i++) {
        /* check if control surface relative chord position is in front half of wing: leading edge device */
        geom2::Point_3 leadingEdgePoint = geom2::measure::left(theWing.controlSurface.at(i), theWing.controlSurface.at(i).sections.front().origin.z());
        if (leadingEdgePoint.x() >= 0. && leadingEdgePoint.x() < 0.5) {
            // check if control surface is further outward than previous one
            if (std::abs(theWing.controlSurface.at(i).sections.back().origin[2]) > wingRelSpanAntiIcingEnd) {
                wingRelSpanAntiIcingEnd = std::abs(theWing.controlSurface.at(i).sections.back().origin[2]);
            }
        }
    }
    // if there is no LE Device, anti icing is applied to the outside edge of the wing (?)
    if (wingRelSpanAntiIcingEnd == 0.0) {
        myRuntimeInfo->warn << "No leading edge control surface was found. Anti Icing will be applied up until the wing tip." << std:: endl;
        wingRelSpanAntiIcingEnd = 1.0;
    }
    return wingRelSpanAntiIcingEnd;
}
