/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef SYSTEMS_DESIGN_SRC_STANDARDDATA_H_
#define SYSTEMS_DESIGN_SRC_STANDARDDATA_H_

#include <tuple>
#include <map>
#include <memory>
#include <string>
#include <vector>
#include "powerProfile.h"
#include "aircraftSystem.h"

namespace fs = std::filesystem;

/**
 * \brief storage class for acxml data and updating
*/
class standardData {
 public:
/* Cabin and Cargo */
    /** \brief contains all cabin properties
     */
    class cabin {
     public:
        int PAX = 0;/**< number of passengers*/

        int cabinCrew = 0;/**< number of cabin crew*/
        int flightCrew = 0;/**< number of flight crew*/
        int numberDecks = 0;/**< number of passenger decks*/
        std::vector<double> lengthDeck;/**<[m] length per pax deck */
        std::vector<double> areaDeck;/**<[m^2] area per pax deck */
        int numberWindows = 0;/**< number of windows*/

        double payload_tube_x_ref = 0;/**<[m] point between cockpit and cabin*/
        std::vector<double> cabinVolume; /**<[m^2] cabin volume of the passenger deck*/
        double cabinVolumeTotal = 0.;/**<[m^3] total cabin volume */

        int cargoDecks = 0;/**< number of cargo decks*/
        std::vector <double> cargoArea;/**<[m^2] cargo area of the deck*/
        std::vector<double> cargoVolume;/**<[m^3] cargo deck volume */
        double cargoVolumeTotal = 0.; /**<[m^3] total cargo volume*/

        double x_CG_PAX_weighted = 0.;/**< weighted center of gravity in x direction of the passengers */

        double massIFE = 0.; /**<[kg] mass of one inflight entertainment system (per seat)*/
        double numberIFE = 0.; /**< number of inflight entertainment systems / seats with IFEs*/

        double galleyPower = 0.; /**<[kW] total power requirement of the galleys*/

        double furnishingMass = 0.; /**<[kg] mass of furnishing without galley and lavatory structure (acc. to Airbus)*/
    };
    cabin Accommodation;/**contains all cabin properties read from the acxml*/

    /*  propulsion **/
    double fuelDensity;/**<[kg/m^3] */
    /**
     * \brief contains all properties of a propulsor
    */
    class propulsor {
     public:
        std::string powertrain;/*powertrain of the propulsor (e.g. turbo, electric, fuel_cell)*/
        std::string type;/*type of main thrust generator: fan or prop*/
        int energyCarrierID;/*energy carrier ID as defined in the energy carrier specification node of the acxml*/
        std::string parentComponent;/*parent component on which the propulsor is mounted*/
    };
    /** \brief contains all properties of propulsionGroup
     */
    class propulsionGroup {
     public:
        std::vector<propulsor> Propulsor;
        int numberEngines;
        double maxShaftPowerOffTakes;/**< [W] max. available shaft power from engines*/
        double maxBleedAirOffTakes;/**< [kg/s] max. available bleed air from engines*/
    };
    propulsionGroup Propulsion;/*Contains propulsion information from the acxml*/

    /* Performance: Speed & Range **/
    /** \brief contains all properties of SpeedRange
     */
    class SpeedRange {
     public:
        double VMO;/**<[m/s] maximum operating speed */
        double MMO;/**<[-] maximum operating Mach number */
        double MOA;/**<[m] maximum operating altitude **/
        double designRange;/**<[m] design range**/
        double maxCertAltitude;/**<[m] max. certified altitude**/
        double designMissionFuel;/**<[kg] design mission fuel mass*/
    };
    SpeedRange Performance;/*Contains performance data from the acxml*/

    /** \brief contains all properties with respect to masses
     */
    class massComponents {
     public:
        double MTOM;/**<[kg] maximum takeoff mass  */
        double WtoS;/**<wing loading  */
        double OME;/**<[kg] operating mass empty  */
        double MME;/**<[kg] manufacturer mass empty  */
        double MLM;/**<[kg] maximum landing mass  */
        //double maxPayload;/**<[kg] maximum payload mass  */
        //double designPayload;/**<[kg] design payload mass  */
        double maxUsableFuel;/**<[kg] maximum usable fuel mass  */
        double totalFuelVolume;/**<[liter]  */   //in Liter
        double Structure;/**<[kg] structural mass  */
        double Wing;/**< Wing mass **/
        double Fuselage;/**<[kg] Fuselage mass **/
        double Empennage;/**<[kg] Empennage **/
        double Gear;/**<[kg] Landing Gear mass **/
        double Pylon;/**<[kg] Mass of all Pylons in kg **/
        double notAllocatedSystems;/**<[kg] */
        double totalGroupAllEngines;/**<[kg]  */
        double furnishingsGroup;/**<[kg]  */
        double systemsGroup;/**<[kg] of the systems group  */
        double totalSystemMass; /**<[kg] mass of all systems excluding masses of accommodation*/
        double waterAndChemicalsMass; /**< [kg] Airbus Ch. 61.1+2: Fluids = TB Potable Water and Toilet Chemicals*/
        double residualFuelMass; /**< [kg] Airbus Ch. 60.1: Unusable Fuel = TB residual fuel*/
        double maxFuel;/**<[kg] Maximum fuel mass that fits in tank and is possible to reach from MTOM (without payload) in kg **/
        double designFuel;/**<[kg] Fuel mass required for the design mission in kg **/
        double fuelAtMaxPayload;/**<[kg] UsableFuel at maxPayload **/
        double payloadAtMaxFuel;/**<[kg] UsablePayload at maxFuel **/
        std::vector<double> Engines;/**<[kg] vector containting the engine masses*/

        massComponents() {
            MTOM = 0.0;
            WtoS = 0.0;
            OME = 0.0;
            MME = 0.0;
            MLM = 0.0;
            //maxPayload = 0.0;
            //designPayload = 0.0;
            maxUsableFuel = 0.0;
            totalFuelVolume = 0.0;
            Structure = 0.0;
            Wing = 0.0;
            Fuselage = 0.0;
            Gear = 0.0;
            Pylon = 0.0;
            notAllocatedSystems = 0.0;
            totalGroupAllEngines = 0.0;
            furnishingsGroup = 0.0;
            systemsGroup = 0.0;
            totalSystemMass = 0.0;
            waterAndChemicalsMass = 0.0;
            residualFuelMass = 0.0;
            maxFuel = 0.0;
            designFuel = 0.0;
            fuelAtMaxPayload = 0.0;
            payloadAtMaxFuel = 0.0;
        }
    };
    massComponents Mass;/*Contains mass information from the acxml*/

    /** \brief contains all properties with respect to centers of gravity. All component reference points are Vec3 as global coordinates in x,y,z-direction
     */
    class cgComponents {
     public:
        Vec3 OME;/**<[m] center of gravity of the aircraft at operating mass empty */
        Vec3 Wing;/**<[m] center of gravity of the wing*/
        Vec3 Fuel;/**<[m] center of gravity of the fuel*/
        Vec3 Fuselage;/**<[m] center of gravity of the fuselage*/
        std::vector<Vec3> Furnishing;/**<[m] center of gravity of furnishing*/
        Vec3 Empennage;/**<[m] center of gravity of the empennage*/
        Vec3 landingGear;/*<[m] center of gravity of the landing gear*/
        // <- Iteration: 1.1*c.g
        std::vector<Vec3> Pylon;/*<[m] centers of gravity of the pylons*/
        Vec3 notAllocatedSystems;/*<[m] center of gravity of the not allocated systems*/
        Vec3 totalGroupAllEngines;/*<[m] center of gravity of all engines*/
        Vec3 furnishingsGroup;/*<[m] center of gravity of the furnishing group*/
        Vec3 systemsGroup;/*<[m] center of gravity of the systems group*/
        Vec3 Payload;/*<[m] center of gravity of the payload*/
        Vec3 Structure;/*<[m] center of gravity of the structural components*/
        std::vector<Vec3> Engines;/*<[m] centers of gravity of each individual engine*/
        Vec3 total; /**<[m] total center of gravity of all systems excluding furnishing (calculated in fuselageDesign) */
        Vec3 waterAndChemicals; /**<[m] center of gravity of Airbus Ch. 61.1+2: Fluids = TB Potable Water and Toilet Chemicals*/
        Vec3 residualFuel; /**<[m] center of gravity of Airbus Ch. 60.1: Unusable Fuel = TB residual fuel*/

        cgComponents() {
            OME = OME.MakeZero();
            Wing = Wing.MakeZero();
            Fuel = Fuel.MakeZero();
            Fuselage = Fuselage.MakeZero();
            Empennage = Empennage.MakeZero();
            landingGear = landingGear.MakeZero();
            notAllocatedSystems = notAllocatedSystems.MakeZero();
            totalGroupAllEngines = totalGroupAllEngines.MakeZero();
            furnishingsGroup = furnishingsGroup.MakeZero();
            systemsGroup = systemsGroup.MakeZero();
            Payload = Payload.MakeZero();
            total = total.MakeZero();
            waterAndChemicals = waterAndChemicals.MakeZero();
            residualFuel = residualFuel.MakeZero();
        }
    };
    cgComponents CofG;/*contains center of gravity information*/

    /** \brief Class containing the accumulated powerProfiles (Bleed, Electric, Hydraulic, heat load)*/
    class accumulatedPowerConsumption {
     public:
        powerProfile Bleed;
        std::vector<powerProfile> Electric;
        std::vector<powerProfile> Hydraulic;
        powerProfile heatLoad;
    };
    accumulatedPowerConsumption sinkConsumption;//accumulated consumption of the sink systems
    accumulatedPowerConsumption transfSinkComnsumption;//Transferred consumer power, stored separately for recalculation of efficiency losses
    accumulatedPowerConsumption ECSConsumption;//accumulated consumption of the environmental control system
    accumulatedPowerConsumption transfECSConsumption;//Transferred power of the ECS, specially stored for recalculation of efficiency losses
    accumulatedPowerConsumption conductorConsumption;//accumulated consumption of the conductor systems
    accumulatedPowerConsumption transfConductorConsumption;//Transferred power of the conductors, specially stored for recalculation of efficiency losses
    accumulatedPowerConsumption totalConsumption;//total accumulated consumption

    double accumulatedSystemsMass;/**< [kg] mass of all systems combined */

    class hydrogenTank {
     public:
        double referencePtx = 0.0;  /**<[m] */
        double referencePty = 0.0;  /**<[m] */
        double referencePtz = 0.0;  /**<[m] */
    };
    hydrogenTank hTank;

    class fuelCell {
     public:
        double volume = 0.0;
    };
    fuelCell fuelCell;

    /* FUNCTIONS FOR WRITING TO THE ACXML FILE */

    /** 
     * @struct NodeInfo
     * @brief Struct which contains the name, id, and description of a new node
     * when inserting the node into the tree.
     */
    struct NodeInfo {
        std::string name{}; /**< The name of the node. */
        std::string id{}; /**< The id of the surface. */
        std::string description{}; /**< The description of the node. */
    };

    /**
     * @brief Create a endnode object and set its properties.
     * 
     * @tparam T The value type of the endnode.
     * @param value The value of the endnode.
     * @param unit The unit of the endnode.
     * @param node_path The location in the parent node of the endnode.
     * @param description The description of the endnode.
     * @return Endnode<T> Return the created endnode.
     */
    template<typename T>
    auto create_endnode(const T value, const std::string& unit, const std::string& node_path, const std::string& description) -> Endnode<T> {
        /* Create the node an set its properties */
        Endnode<T> new_node{node_path, description};
        new_node.set_value(value);
        new_node.set_unit(unit);

        /* Return the new node */
        return new_node;
    }

    /**
     * @brief Insert some reference point of the surface into the specified node.
     * You can specify the name and description of the node which is inserted.
     * The description of the x,y, and z nodes will be set to a default description.
     * @attention The target node will be modified!
     * 
     * @param target The pointer to the target node where the point will be inserted.
     * @param point The point to insert.
     * @param info The name and description of the node to insert.
     */
    void insert_point(node* target, const Vec3 point, const NodeInfo info) {
        /* Create the position node and use its reference to further access it*/
        node& point_node = (*target)[info.name];
        point_node.setAttrib("description", info.description);

        /* Create and add the value end nodes */
        auto x_value = create_endnode(point[0], "m", "x", "x coordinate of point");
        auto y_value = create_endnode(point[1], "m", "y", "y coordinate of point");
        auto z_value = create_endnode(point[2], "m", "z", "z coordinate of point");
        x_value.set_boundaries(-INFINITY, INFINITY);
        y_value.set_boundaries(-INFINITY, INFINITY);
        z_value.set_boundaries(-INFINITY, INFINITY);
        x_value.update(point_node);
        y_value.update(point_node);
        z_value.update(point_node);
    }
};

#endif // SYSTEMS_DESIGN_SRC_STANDARDDATA_H_
