/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef SYSTEMS_DESIGN_SRC_STANDARDSYSTEMSDESIGN_H_
#define SYSTEMS_DESIGN_SRC_STANDARDSYSTEMSDESIGN_H_

#include <memory>
#include <vector>
#include <string>
#include <moduleBasics/strategySelector.h>
#include <moduleBasics/report.h>
#include <moduleBasics/plot.h>
#include "systemsDesign.h"
#include "systemsIOData.h"
#include "standardData.h"
#include "weightsAndCGs.h"
#include "powerProfile.h"


/**
 * \brief Class for standard systems design strategy
 * 
 */
class standardSystemsDesign : public Strategy {
 public:
    explicit standardSystemsDesign(std::shared_ptr<RuntimeIO> rtIO): rtIO_(rtIO), report_(rtIO) {}
    ~standardSystemsDesign() = default;
    /* Inherited functions from class Strategy */

    /** \brief Initializes the IOData pointer with the acxml and config file, reads the geometry information
     * for the fuselage, nacelles, wing, and empennage from the aircraft xml as well as all needed values for 
     * the propulsion system, system energy providers, masses, performance, and accomodation. All functions reading
     * from the acxml belong to the class systemsIOData. 
     * The individual systems are initialized based on the user definiton in the config document.
     * Finally the inputs from both files are checked for consistency.
     * \return void
    */
    void initialize();

    /** \brief Differentiates between the two modes (size and study) of the standard systemsDesign strategy.
     * \return void
    */
    void run();

    /** \brief For the sizing mode this function calls the update functions for the acxml and the mission xml file.
     * In the study mode only the mission xml file is updated.
     * \return void
    */
    void update();

    /** \brief Calls the functions for preparing and generating the plots and reports.
     * \return void
    */
    void report();

    void save();

    /** \brief Runs the sizing mode of systemsDesign. First the performance profile of all systems is determined for
     * the design mission. Then the systems are sized and there masses determined according to the design mission.
     * Finally, the weightsAndCGs class is used to calculated the accumulated masses and centers of gravity.
     * \return void
    */
    void size();

    /** \brief Runs the study mode of systemsDesign. The performance profile in the study mission for all systems is calculated.
     * \attention System and accumulated masses and CofG are not determined in the study mode.
    */
    void study();

    /** \brief Runs the requirements mission mode of systemsDesign. The performance profiles in the requirments mission for all
     * systems are calculated.
     * \attention System and accumulated masses and CofG are not determined in the requirments mission mode.
    */
    void requirements_mission();

    /** \brief Calls the specific implementation of the checkUserInput()-function of the aircraft systems ATA-29,
     * ATA-24, and ATA-27. These functions check user input from the configuration file.
     * \return void
    */
    void checkUserInputForConsistency();

    /** \brief Checks wether the acxml specifies ailerons for the wing and a landing gear. If not it throws an error
     * since these components are required.
     * \return void
    */
    void checkGeometryForConsistency();


    /** \brief Checks if the energy providers specified in the acxml file are implemented in systemsDesign and if they
     * are consistent with those specified in the config file. If not it throws an error. 
     * \todo Implement a check wether the energy source provides all required energy "forms" (electric/bleed air)
     * \return void
    */
    void checkArchitectureForConsistency();

    /** \brief Checks if the given energy source is specified in the config file.
     * \param sourceName name of the source needs to match the name specified in the config file.
     * \return bool is true when given source is defined in config file and false if it is not.
    */
    bool checkEnergySource(std::string sourceName);

    /** \brief Initializes every system based on the inputs from the config file. Based on the system name an
     * instance of the corresponding aircraftSystem is created.
     * \note The systems are grouped into sinks, environmental
     * control system, virtual systems, energy sources and energy conductors, which all have a corresponding vector
     * data_->Systems.energy... . Additionally, there is one vector containing all systems (data_->Systems.allSystems)
     * except the environmental control system.
     * 
     * The environmental control system, despite being a sink system, is kept seperate to allow for iteration between the 
     * power consumption and heat production of the different systems and conductors and the corresponding sizing of the 
     * environmental control system.
     * 
     * For each system the specific instance of the function "readParameters" is called to read the system specific parameter
     * from the config file.
     * \return void
    */
    void initializeSystems();

    /** \brief Calculates the total performance profile of all systems.
     * \note First, the energy consumption of the sink systems is determined. The hydraulic and electric power required
     * by the sink systems is transferred through electric pumps and hydraulic generators. This is calculated with the
     * functions @ref transferElec2HydPower and @ref transferHyd2ElecPower. After this, an iteration is performed between
     * the energy consumption of the environmental control system and the losses of the conductor systems. The average
     * bleed air deviation of the environmental control system is used as a convergence criteria. The iteration stops when
     * this value is below 0.0001 or 50 iterations have taken place. Finally, the energy consumption of the sources is
     * determined, which corresponds to the total power requirement of all other systems.
     * \return void
    */
    void calculatePerformanceProfile();

    /** \brief Calculates the energy consumption (power profiles) of all systems that require energy (are energy sinks)
     * by calling the specific calculateSystemsPower-functions of each sink system and then @ref getAccumulatedPower.
     * After all power consumptions are accumulated the power peaks are cleaned.
     * \note power is split into electric power, hydraulic power, heat load, and bleed air
     * \return void
    */
    void getSinkEnergyConsumption();

    /** \brief Calculates the energy provided by the source systems by getting the total energy consumption of all systems
     * including the environmental control system and then calling the specific implementations of the 
     * calculateSourcePwer-function of each source system.
     * \note power is split into electric power, hydraulic power, heat load, and bleed air
     * \return void
    */
    void getSourceEnergyConsumption();

    /** \brief Calculates the energy consumption (power profiles) of the environmental control system by calling
     * the specific calculateSystemsPower-functions of ECS and then @ref getAccumulatedPower.
     * After all power consumptions are accumulated the power peaks are cleaned.
     * \note power is split into electric power, hydraulic power, heat load, and bleed air
     * \return void
    */
    void getECSEnergyConsumption();

    /** \brief Calculates the energy consumption (power profiles) of systems that conduct energy by calling the
     * specific implementation of the calculateConductingPower-function of each conducting system and then @ref
     * getAccumulatedPower.
     * \note power is split into electric power, hydraulic power, heat load, and bleed air
     * \param aSinkConsumption accumulated energy consumption of the systems that are energy sinks except for the environmental control system
     * \param anECSConsumption accumulated energy consumption of the environmental control system
     * \return void
    */
    void getConductorEnergyConsumption(const standardData::accumulatedPowerConsumption& aSinkConsumption, const standardData::accumulatedPowerConsumption& anECSConsumption);

    /** \brief sets all values to zero of an accumulatedPowerConsumption
     *  \param theConsumption pointer to the consumption that shall be cleared
     * \return void
    */
    void clearInitializePowerConsumption(standardData::accumulatedPowerConsumption *theConsumption);

    /** \brief Accumulates the power demand or provision of a systems through the complete mission resulting in values for the max. mission
     * power, design power, as well as base load, peak load, and time of the peak load in each mission segment.
     * \note power is devided into electric power, hydraulic power, heat load, and bleed air
     * \anchor getAccumulatedPower
     * \param theSystem the system of which the power demand/provision shall be calculated
     * \param theConsumption parameter to store the accumulated power
     * \return void
    */
    void getAccumulatedPower(const aircraftSystem& theSystem, standardData::accumulatedPowerConsumption *theConsumption);

    /** \brief Writes the user defined (config) power requirement of a virtual system into the power profile of that system.
     * \param ID ID of the virtual system
     * \attention virtual systems are completely user defined. The power profile is NOT calculated.
     * \return void
    */
    void calculateVirtualSystemPower(int ID);

    /** \brief transfers electric to hydraulic power via electric pumps taking into account the operation factor and
     * the generator efficiency, which can be defined in the config file.
     * @anchor transferElec2HydPower
     * \param myElectric electric power profile
     * \param myHydraulic hydraulic power profile
     * \param transfHydraulic transfer power profile
     * \return void
    */
    void transferElec2HydPower(std::vector<powerProfile> *myElectric, std::vector<powerProfile> *myHydraulic, std::vector<powerProfile> *transfHydraulic);

    /** \brief transfers electric to hydraulic power via hydraulic generators taking into account the pump opertion factor
     * and the electric pump efficiency, which can be defined in the config file.
     * @anchor transferHyd2ElecPower
     * \param myElectric electric power profile
     * \param myHydraulic hydraulic power profile
     * \param myHeatLoad powerProfile for the heat load created through the transfer
     * \param transfElectric transfer power profile
     * \return void
    */
    void transferHyd2ElecPower(std::vector<powerProfile> *myHydraulic, std::vector<powerProfile> *myElectric, powerProfile *myHeatLoad, std::vector<powerProfile> *transfElectric);

    /** \brief Sums up the total power consumption of all sinks, the environmental control system and the energy conductors
     * for both the design power and the base and peak loads in each mission segment.
     * \anchor getTotalConsumption
     * \param theTotal total accumulated power consumption
     * \param Sinks accumulated power consumption of all sinks
     * \param ECS accumulated power consumption of the environmental control system
     * \param Conductors accumulated power consumption of the energy conductors
     * \return void
    */
    void getTotalConsumption(standardData::accumulatedPowerConsumption *theTotal, standardData::accumulatedPowerConsumption Sinks,
                             standardData::accumulatedPowerConsumption ECS, standardData::accumulatedPowerConsumption Conductors);

    /** \brief Calls the mass calculation function of each individual system.
     * \attention virtual systems are completely user defined and the mass is not calculated but copied from the config file
     * \note source and conductor systems require the total power consumption (bleed air, electric and hydraulic power).
     * Therefore, @ref getTotalConsumption is called before the masses of source and conductor systems is calculated.
     * \return void
     */
    void getSystemsMass();

    /** \brief Calculates the average bleed air consumption over the mission of the environmental control system
     * \param theBleed bleed air power profile of the ECS
     * \return double average bleed air
    */
    double getAvgBleedECS(powerProfile theBleed);

    /* Reporting elements */
    /**
    * \brief Defines the content of the HTML Report body. The content depends on wether the sizing or study mode
    * is used. Since that masses or only calculated in the sizing mode, they are also only written in this mode.
    * \return void
    */
    void setHtmlBody();

   /**
    * \brief Defines the content of the Tex Report body. The content depends on wether the sizing or study mode
    * is used. Since that masses or only calculated in the sizing mode, they are also only written in this mode.
    * \return void
    */
    void setTexBody();

    /* Plotting elements */
    /** \brief Calls all plot generation functions */
    void generatePlotData();

    void generatePlotScripts();

    /** \brief Generats bar plot of aircraft masses (OME, MME, op. items, structure, gear, propulsion, systems,
     * not allocated) */
    void generatePlotMassBreakdown();
    void generatePlotScriptMassBreakdown();

    /** \brief Generates bar plot of individual system masses */
    void generatePlotSystemsMass();

    /** \brief plots the mission data (altitude, temperature, mach, CAS, TAS) over the mission
     * steps. This data is read from the mission file and not altered in systemsDesign. (Plot can
     * serve as a reference when reading mission plots over mission steps of the individual systems)
     */
    void generatePlotDataMission();

    /** \brief Plots the power (hydraulic, electric, bleed air) of the system over the mission
     * steps. Base load and peak loads are plotted.
     * \param acftSystem aircraft system
     * \return void
     */
    void generatePlotMissionEnergy(const aircraftSystem& acftSystem);

    /** \brief Plots the accumulated power demand of all systems including peaks.
     * \param totalBleed pointer to total bleed air power profile
     * \param totalElectric pointer to total electric power profile
     * \param totalHydraulic pointer to total hydraulic power profile
     * \return void
     */
    void generatePlotMissionEnergyAccumulated(const powerProfile& totalBleed, const powerProfile& totalElectric, const powerProfile& totalHydraulic);

    /** \brief bar plot of the maximum power demand (electric, hydraulic and bleed air) of each sink system
     * \return void
     */
    void generatePlotEnergySinksMaxEnergy();

    /** \brief Calculates y-axis scaling for electric / hydraulic power based on the maximum value and including room for a legend
     * \param hydraulicSum pointer to the hydraulic power profile that will be plotted
     * \param electricSum pointer to the electric power profile that will be plotted
     * \return double: maximum value for power y-axis
     */
    double getYAxisScalingEHC(const powerProfile *hydraulicSum, const powerProfile *electricSum);

    /** \brief Calculates y-axis scaling for bleed air mass flow based on the maximum value and including room for a legen
     * \note Depending on values the extra room for the legend is sometimes smaller than that for electric/hydraulic power. It can be useful to add
     * an extra 10-20% of room by increasing the upper y-axis limit by this factor.
     * \param bleedAir pointer to the bleed air power profile
     * \return double: maximum value for mass flow y-axis
     */
    double getYAxisScalingBA(const powerProfile *bleedAir);

    /** \brief Helper function for getYAxisScalingBA and getYAxisScalingEHC
     * \param numToRound number to round (int)
     * \param multiple number to multiply (int)
     * \return int: rounded number
    */
    int roundUp(int numToRound, int multiple);

    /** \brief checks if all values of a vector are zero
     * \param vec vector to be checked
     * \return bool = true if all values of the vector are zero
    */
    bool all_zeros(const std::vector<double>& vec);

    double maxShaftPower = 0.;/**< [W] maximum shaft power offtake of all systems */
    double maxBleedAir = 0.;/**< [kg/s] maximum bleed air offtake of all system*/

 private:
    std::shared_ptr<RuntimeIO> rtIO_; /*<-- RuntimeIO object*/
    std::shared_ptr<systemsIOData> data_; /*<-- IOData object*/
    Report report_; /*Report object*/
    weightsAndCGs *myWeightsPt; /*Pointer to weightsAndCGs object*/
};

#endif //SYSTEMS_DESIGN_SRC_STANDARDSYSTEMSDESIGN_H_
