# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2024 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

# Import standard libraries.
import numpy as np

# Import own modules.
from src.tube_and_wing.empirical.tank_design_tu_berlin.liquid_hydrogen.calculateparameterendcap import calculate_parameter_end_cap
from src.tube_and_wing.empirical.tank_design_tu_berlin.liquid_hydrogen.calculatethicknesswall import calculate_thickness_wall
from src.tube_and_wing.empirical.tank_design_tu_berlin.liquid_hydrogen.calculateparametercylinder import calculate_parameter_cylinder
from src.tube_and_wing.empirical.tank_design_tu_berlin.general.calculatecenterofgravity import calculate_center_of_gravity


def calculate_cylindrical_tank(tank_entity, dict_ac_data, dict_tank_design, runtime_output):
    """_summary_

    :param _type_ tank_entity: _description_
    :param float volume_required: _description_
    :param dict dict_ac_data: Dict containing parameter and values from aircraft exchange and module configuration file
    :param dict dict_tank_design: Dict containing results from tank design
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict dict_tank_design: Dictionary containing results from tank design for tube-and-wing aircraft
    """
    # Extract values from 'dict_ac_data'.
    mass_pumps = dict_ac_data['mass_pumps']
    mass_baffle = dict_ac_data['mass_baffle']
    type_end_cap = dict_ac_data['type_end_cap']
    density_wall = dict_ac_data['density_wall']
    material_wall = dict_ac_data['material_wall']
    internal_pressure = dict_ac_data['internal_pressure']
    density_insulation = dict_ac_data['density_insulation']
    factor_volume_allowance = dict_ac_data['factor_volume_allowance']
    thickness_wall_calculation_method = dict_ac_data['thickness_wall_calculation_method']

    energy_required = dict_tank_design[tank_entity]['energy_required']
    mass_vapor_barrier = 24  # TODO: Extract from tank_design_xml?
    thickness_wall = 0.003
    thickness_insulation = 0.20
    thickness_wall_and_insulation = thickness_wall + thickness_insulation
    tensile_stress_allowable = 172  # in N/mm TODO: Read from aircraft exchange file/lib/...
    thickness_wall_minimum_aluminum = 0.003  # AD-2000: minimum wall thickness AL-2219 is 3 mmm.

    volume_required = energy_required/dict_ac_data['liquid_hydrogen_volumetric_energy_density']

    """ Calculation. """
    # Note: The diameters are actual sizes. This means that the thicknesses for the tank wall and insulation have
    # already been taken into account.
    # Assumption: Maximum outer tank diameter equals maximum useable fuselage diameter minus twice the
    # thickness of insulation and wall. The calculated maximum outer diameter is assumed to be the actual tank
    # diameter.
    maximum_tank_diameter = (dict_tank_design[tank_entity]['geometry']['box_dimensions']['maximum_diameter_front']
                             - 2*thickness_wall_and_insulation)

    """ Find position of end caps."""
    # Set iteration values.
    increment = 0.01
    boundary = 100
    distance = 0

    while boundary > distance:
        boundary = calculate_parameter_end_cap(maximum_tank_diameter, type_end_cap, thickness_wall,
                                                        thickness_insulation, density_wall, density_insulation)[3]
        distance += increment

    """ Calculate wall thickness."""
    maximum_diameter = maximum_tank_diameter
    thickness_wall = calculate_thickness_wall(maximum_diameter, internal_pressure, tensile_stress_allowable,
                                              thickness_wall_calculation_method)
    if (material_wall == 'aluminum') and (thickness_wall < thickness_wall_minimum_aluminum):
        thickness_wall = thickness_wall_minimum_aluminum

    """ Calculate volumes, masses, and lengths of tank parts."""
    # Note that the lengths of the end caps already take into account the tank wall and the insulation thickness.
    # 1) Front end cap (equals rear end cap).
    (volume_tank_end_cap, mass_wall_end_cap, mass_insulation_end_cap,
     length_end_cap) = calculate_parameter_end_cap(maximum_tank_diameter, type_end_cap, thickness_wall,
                                                         thickness_insulation, density_wall, density_insulation)
    # 2) Calculate how much volume the cylindrical part has to hold.
    volume_end_caps_available = 2*volume_tank_end_cap
    # energy_end_caps_available = volume_end_caps_available * dict_ac_data['liquid_hydrogen_volumetric_energy_density']
    # energy_remaining = energy_required - energy_end_caps_available
    volume_remaining = volume_required - volume_end_caps_available

    # 3) The remaining energy must be stored in the cylindrical part of the tank (iterative approach).
    length_cylindrical_section = 0
    volume_cylindrical_section = 0
    volume_cylindrical_section_usable = 0
    while volume_cylindrical_section_usable < volume_remaining:
        (volume_cylindrical_section, mass_wall_cylindrical_section,
         mass_insulation_cylindrical_section) = calculate_parameter_cylinder(
                    maximum_diameter, length_cylindrical_section, thickness_wall, thickness_insulation,
                    density_wall, density_insulation)
        volume_cylindrical_section_usable = volume_cylindrical_section/(1 + factor_volume_allowance)
        length_cylindrical_section += 0.01

    # 4) Total tank.
    volume_tank_available = volume_end_caps_available + volume_cylindrical_section_usable
    energy_hydrogen_available = volume_tank_available * dict_ac_data['liquid_hydrogen_volumetric_energy_density']
    mass_hydrogen_available = energy_hydrogen_available/dict_ac_data['liquid_hydrogen_gravimetric_energy_density']

    """ Calculate overall values."""
    # Calculate mass of tank wall.
    mass_wall = 2*mass_wall_end_cap + mass_wall_cylindrical_section
    # Calculate structural mass of tank.
    mass_tank_structure = mass_wall + mass_pumps + mass_baffle
    # Calculate mass of insulation.
    mass_insulation = mass_insulation_end_cap + mass_insulation_cylindrical_section + mass_vapor_barrier
    # Calculate overall tank mass.
    mass_tank_overall = mass_tank_structure + mass_insulation
    # Calculate overall tank length.
    length_tank_overall = 2*length_end_cap + length_cylindrical_section
    # Calculate gravimetric efficiency tank.
    gravimetric_efficiency = mass_hydrogen_available/(mass_hydrogen_available + mass_tank_overall)

    """ Calculate center of gravity."""
    (x_coordinate_center_of_gravity, y_coordinate_center_of_gravity, z_coordinate_center_of_gravity
     ) = calculate_center_of_gravity(dict_tank_design[tank_entity]['tank_shape'],
         dict_tank_design[tank_entity]['geometry']['box_dimensions']['maximum_diameter_front'],
         dict_tank_design[tank_entity]['geometry']['box_dimensions']['maximum_diameter_front'],
         length_cylindrical_section)

    # Prepare data output.
    dict_tank_design[tank_entity]['energy_available'] = float(energy_hydrogen_available)
    dict_tank_design[tank_entity]['masses'] = {}
    dict_tank_design[tank_entity]['masses']['mass_insulation'] = float(mass_insulation)
    dict_tank_design[tank_entity]['masses']['mass_tank_overall'] = float(mass_tank_overall)
    dict_tank_design[tank_entity]['masses']['mass_tank_structure'] = float(mass_tank_structure)
    dict_tank_design[tank_entity]['geometry']['length_tank_overall'] = float(length_tank_overall)
    dict_tank_design[tank_entity]['geometry']['center_of_gravity'] = {}
    dict_tank_design[tank_entity]['geometry']['center_of_gravity']['x'] = float(x_coordinate_center_of_gravity)
    dict_tank_design[tank_entity]['geometry']['center_of_gravity']['y'] = float(y_coordinate_center_of_gravity)
    dict_tank_design[tank_entity]['geometry']['center_of_gravity']['z'] = float(z_coordinate_center_of_gravity)
    dict_tank_design[tank_entity]['gravimetric_efficiency'] = float(gravimetric_efficiency)

    return dict_tank_design
