# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2024 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

# Import standard libraries.
import math
from numpy import pi as PI


def calculate_parameter_end_cap(diameter, type_end_cap, thickness_wall=0.003, thickness_insulation=0.20,
                               density_material_wall=2850, density_material_insulation=32):
    """Calculate parameters of tank end cap.

    This function calculates the following parameters of the tank end cap:
        * Volume
        * Mass of the tank wall
        * Mass of the insulation
        * Length of the end cap
    The end cap can either be of a
        a) hemispherical or
        b) torispherical
    type.
    a) Hemispherical end cap
        The calculation of the volume is based on geometric volume estimates. The volume of the tank wall
        and insulation are determined based on a hollow sphere. The mass can then be determined using the
        corresponding material density.
    b) Torispherical end cap
        Torispherical end caps are calculated according to DIN 28011. It also provides a formula for
        determining the mass using the corresponding material density.

    Note: Default values are based on aluminum (Al-2219) as the tank wall material and polyurethane foam as insulation.

    :param float diameter: Tank diameter in m
    :param str type: Type of end cap ('hemispherical' or 'torispherical')
    :param float thickness_wall: Wall thickness in m, defaults to 0.003
    :param float thickness_insulation: Insulation thickness in m, defaults to 0.20
    :param float density_material_wall: Density of tank wall material in kg/m^3, defaults to 2850
    :param float density_material_insulation: Density of insulation material in kg/m^3, defaults to 32
    :return float: Volume of tank in m^3, mass of wall in kg, mass of insulation in kg, length of end cap in m
    """
    radius_excl_wall = diameter*0.5
    radius_incl_wall = radius_excl_wall + thickness_wall
    radius_incl_insulation = radius_incl_wall + thickness_insulation
    match type_end_cap:
        case 'hemispherical':
            # Volume calculation for tank, based on standard sphere volume.
            volume_tank = (4/3*PI*radius_excl_wall**3)*0.5
            # Volume and mass calculation for tank wall, based on volume calculation for a hollow sphere.
            volume_wall = ((4*PI*(radius_incl_wall**3 - radius_excl_wall**3))/(3))
            mass_wall = volume_wall*density_material_wall
            # Volume and mass calculation for tank insulation, based on volume calculation for a hollow sphere.
            volume_insulation = ((4*PI*(radius_incl_insulation**3 - radius_incl_wall**3))/(3))
            mass_insulation = volume_insulation*density_material_insulation
            # Total length calculation.
            length = radius_incl_insulation
        case 'torispherical':
            # Note: Calculation and naming according to DIN 28011.
            # [naming acc. to DIN 28011] -> [naming in implementation]
            # d_a -> outer_base_diameter (äußerer Bodendurchmesser) in mm
            # Sr_1 -> radius_of_curvature (Wölbungsradius)
            # r_2 -> corner_radius (Eckradius)
            # s -> thickness_wall in mm
            # h_1 -> cylindrical_board_height (zylindrische Bordhöhe)
            # h_2 -> camber_height (Wölbungshöhe)
            # h_3 -> inner_total_floor_height (gesamte Bodenhöhe innen)
            thickness_wall = thickness_wall*1000
            thickness_insulation = thickness_insulation*1000
            outer_base_diameter = (radius_incl_wall*2)*1000
            outer_base_diameter_incl_insulation = (radius_incl_insulation*2)*1000
            radius_of_curvature = outer_base_diameter
            cylindrical_board_height = 3.5*thickness_wall  # acc. to 'Tabelle 1'
            camber_height = 0.1935*outer_base_diameter - 0.455*thickness_wall  # acc. to '3.1 Geometrische Beziehungen'
            # Volume and shell surface of cylindrical board.
            volume_board = PI*(outer_base_diameter*0.5)**2*cylindrical_board_height
            shell_surface_board = 2*PI*outer_base_diameter*cylindrical_board_height
            # Volume and shell surface of curvature.
            volume_curvature = 0.1*(outer_base_diameter-2*thickness_wall)**3
            # Total tank volume.
            volume_tank = (volume_board + volume_curvature)/1e9
            # Calculate mass of wall.
            density_material_wall = density_material_wall*0.001
            D_R = 1.128*math.sqrt(0.99*outer_base_diameter**2 + outer_base_diameter*PI*cylindrical_board_height)
            mass_wall = ((D_R**2)/4*PI*thickness_wall*1.1*density_material_wall)/1000000
            # Calculate mass of insulation.
            density_material_insulation = density_material_insulation*0.001
            D_R = 1.128*math.sqrt(0.99*outer_base_diameter_incl_insulation**2
                                  + outer_base_diameter_incl_insulation*PI*cylindrical_board_height)
            mass_insulation = ((D_R**2)/4*PI*thickness_insulation*1.1*density_material_insulation)/1000000
            # Total length calculation.
            length = (cylindrical_board_height + camber_height + thickness_wall + thickness_insulation)/1000
        case 'ellipsoidal':
            print('Ellipsoidal end cap not yet implemented.')

    return volume_tank, mass_wall, mass_insulation, length
