/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include <aircraftGeometry2/processing/measure.h>
#include <aircraftGeometry2/processing/transform.h>
#include <moduleBasics/html.h>

#include "cantileverWingDesignTaw.h"

namespace taw {
namespace cantilever {
void Wing::set_html_body() {
  /* Add box data */
  reporter.htmlReportStream() << "<div class=\"box data\">\n";
  /* Add headline Data*/
  reporter.htmlReportStream() << "<h2>Data</h2>\n";
  /* Table 1 - General parameters */
  reporter.htmlReportStream()
      << "<table class=\"content-table\">\n"
      << "<caption>General parameters</caption>\n"
      << "<thead>\n"
      << "<tr>\n"
      << "<th>parameter</th><th>symbol</th><th>unit</th><th>value</th>\n"
      << "</tr>\n"
      << "</thead>\n"
      << "<tbody>\n"
      << "<tr>\n"
      << std::format("<td>Reference area</td><td>S<sub>ref</sub></td><td>m<sup>2</sup></td><td>{:.2f}</td>\n",
                     geom2::measure::reference_area(data->wing))
      << "</tr>\n"
      << "<tr>\n"
      << std::format("<td>Mean aero. chord</td><td>MAC</td><td>m</td><td>{:.2f}</td>\n",
                     geom2::measure::mean_aerodynamic_chord(data->wing))
      << "</tr>\n"
      << "<tr>\n"
      << std::format("<td>Span</td><td>b</td><td>m</td><td>{:.2f}</td>\n", geom2::measure::span(data->wing))
      << "</tr>\n"
      << "<tr>\n"
      << std::format("<td>Aspect ratio</td><td>AR</td><td>1</td><td>{:.2f}</td>\n",
                     geom2::measure::aspect_ratio(data->wing))
      << "</tr>\n"
      << "<tr>\n"
      << std::format("<td>Taper ratio</td><td>&lambda;</td><td>1</td><td>{:.2f}</td>\n",
                     geom2::measure::taper_ratio(data->wing))
      << "</tr>\n"
      << "</tbody>\n"
      << "</table>\n";

  /* Table 2 - Section parameters*/
  reporter.htmlReportStream() << "<table class=\"content-table\">\n"
                              << "<caption>Section parameters</caption>\n"
                              << "<thead>\n"
                              << "<tr>\n"
                              << "<th>parameter</th><th>symbol</th><th>unit</th>";
  for (size_t i = 0; i < data->wing.sections.size(); ++i) {
    reporter.htmlReportStream() << std::format("<th>S<sub>{:02d}</sub></th>", i);
  }
  reporter.htmlReportStream() << "\n</tr>\n"
                              << "</thead>\n"
                              << "<tbody>\n"
                              /* Dimensionless half span - eta*/
                              << "<tr>\n"
                              << "<td>Dimensionless half span</td><td>&eta;</td><td>1</td>";
  for (auto section : data->wing.sections) {
    reporter.htmlReportStream() << std::format("<td>{:.2f}</td>",
                                               -2 * section.origin.z() / geom2::measure::span(data->wing));
  }
  /* Spanwise coordinate - y*/
  reporter.htmlReportStream() << "\n</tr>\n"
                              << "<td>Spanwise coordinate</td><td>y</td><td>m</td>";
  for (auto section : data->wing.sections) {
    reporter.htmlReportStream() << std::format("<td>{:.2f}</td>", -section.origin.z());
  }
  reporter.htmlReportStream() << "\n</tr>\n"
                              /* Chord - c*/
                              << "<tr>\n"
                              << "<td>Chord</td><td>c</td><td>m</td>";
  for (auto section : data->wing.sections) {
    reporter.htmlReportStream() << std::format("<td>{:.2f}</td>", section.get_chord_length());
  }
  reporter.htmlReportStream() << "\n</tr>\n"
                              /* Sweep - Leading edge - phi_LE*/
                              << "<tr>\n"
                              << "<td>Sweep leading edge</td><td>&phi;<sub>LE</sub></td><td>&deg;</td>";
  for (auto section : data->wing.sections) {
    reporter.htmlReportStream() << std::format(
        "<td>{:.2f}</td>", -geom2::detail::to_degrees * geom2::measure::sweep(data->wing, section.origin.z(), 0.0));
  }
  reporter.htmlReportStream() << "\n</tr>\n"
                              /* Sweep - 25 percent - phi_25%*/
                              << "<tr>\n"
                              << "<td>Sweep quarter chord</td><td>&phi;<sub>25</sub></td><td>&deg;</td>";
  for (auto section : data->wing.sections) {
    reporter.htmlReportStream() << std::format(
        "<td>{:.2f}</td>", -geom2::detail::to_degrees * geom2::measure::sweep(data->wing, section.origin.z(), 0.25));
  }
  reporter.htmlReportStream() << "\n</tr>\n"
                              /* Sweep - Trailing edge percent - phi_TE*/
                              << "<tr>\n"
                              << "<td>Sweep Trailing edge</td><td>&phi;<sub>TE</sub></td><td>&deg;</td>";
  for (auto section : data->wing.sections) {
    reporter.htmlReportStream() << std::format(
        "<td>{:.2f}</td>", -geom2::detail::to_degrees * geom2::measure::sweep(data->wing, section.origin.z(), 1.0));
  }
  reporter.htmlReportStream() << "\n</tr>\n"
                              /* Dihedral - nu*/
                              << "<tr>\n"
                              << "<td>Dihedral</td><td>&nu;</td><td>&deg;</td>";
  for (auto section : data->wing.sections) {
    reporter.htmlReportStream() << std::format(
        "<td>{:.2f}</td>", geom2::detail::to_degrees * geom2::measure::dihedral(data->wing, section.origin.z()));
  }
  reporter.htmlReportStream() << "\n</tr>\n"
                              /* Twist - epsilon*/
                              << "<tr>\n"
                              << "<td>Twist angle</td><td>&epsilon;</td><td>&deg;</td>";
  for (auto section : data->wing.sections) {
    reporter.htmlReportStream() << std::format("<td>{:.2f}</td>", geom2::detail::to_degrees * section.rotation_z);
  }
  reporter.htmlReportStream() << "\n</tr>\n"
                              /* Thickness to chord - epsilon*/
                              << "<tr>\n"
                              << "<td>Thickness ratio</td><td>t/c</td><td>%</td>";
  for (auto section : data->wing.sections) {
    reporter.htmlReportStream() << std::format(
        "<td>{:.2f}</td>", 100 * geom2::measure::thickness_max(section) / section.get_chord_length());
  }
  reporter.htmlReportStream() << "\n</tr>\n"
                              /* Airfoil - none */
                              << "<tr>\n"
                              << "<td>Airfoil</td><td>-</td><td>-</td>";
  for (auto section : data->wing.sections) {
    reporter.htmlReportStream() << std::format("<td>{}</td>", section.name);
  }
  reporter.htmlReportStream() << "\n</tr>\n"
                              << "</tbody>\n"
                              << "</table>\n";

  /* Table 3 - Spar parameters*/
  for (size_t spar_id = 0; spar_id < data->spars.size(); ++spar_id) {
    reporter.htmlReportStream() << "<table class=\"content-table\">\n";
    if (spar_id == 0) {
      reporter.htmlReportStream() << "<caption>Spar parameters</caption>\n";
    }
    reporter.htmlReportStream() << "<thead>\n"
                                << "<tr>\n"
                                << std::format("<th>{:<10}</th><th>&eta; [1]<th>x/c [%]</th>",
                                               data->spars.at(spar_id).name)
                                << "\n</tr>\n"
                                << "</thead>\n"
                                << "<tbody>\n";
    /* Spar name - eta, x/c*/
    for (size_t i = 0; i < data->spars.at(spar_id).sections.size(); ++i) {
      reporter.htmlReportStream() << "<tr>\n";
      std::string tag = "";
      if (i == 0) {
        tag = "from";
      } else if (i == data->spars.at(spar_id).sections.size() - 1) {
        tag = "to";
      }
      reporter.htmlReportStream() << std::format("<td>{}</td><td>{:.2f}</td><td>{:.2f}</td>\n", tag,
                                                 data->spars.at(spar_id).sections.at(i).origin.z(),
                                                 data->spars.at(spar_id).sections.at(i).get_contour().vertex(0).x());
      reporter.htmlReportStream() << "\n</tr>\n";
    }
    /* Spanwise coordinate - y*/
    reporter.htmlReportStream() << "</tbody>\n"
                                << "</table>\n";
  }

  /* Sort devices to leading, trailing and other devices*/
  /* device data - name, eta_from, rel_chord_from_start, rel_chord_from_end, eta_to, rel_chord_to_start,
   * rel_chord_to_end */
  using device_data = std::tuple<std::string, double, double, double, double, double, double>;
  std::vector<device_data> leading_edge_devices;
  std::vector<device_data> trailing_edge_devices;
  std::vector<device_data> other_devices;
  for (auto device : data->control_devices) {
    /* Device has only two sections (from + to) spanwise */
    double eta_from = device.sections.at(0).origin.z();
    double rel_chord_from_start = device.sections.at(0).get_contour().vertex(0).x(),
           rel_chord_from_end = device.sections.at(0).get_contour().vertex(1).x();
    double eta_to = device.sections.at(1).origin.z();
    double rel_chord_to_start = device.sections.at(1).get_contour().vertex(0).x(),
           rel_chord_to_end = device.sections.at(1).get_contour().vertex(1).x();

    if (std::fabs(rel_chord_from_start - rel_chord_to_start) < ACCURACY_MEDIUM &&
        std::fabs(rel_chord_from_start) < ACCURACY_MEDIUM) {
      leading_edge_devices.push_back({device.name, eta_from, rel_chord_from_start, rel_chord_from_end, eta_to,
                                      rel_chord_to_start, rel_chord_to_end});
    } else if (std::fabs(rel_chord_from_end - rel_chord_to_end) < ACCURACY_MEDIUM &&
               std::fabs(rel_chord_from_end - 1) < ACCURACY_MEDIUM) {
      trailing_edge_devices.push_back({device.name, eta_from, rel_chord_from_start, rel_chord_from_end, eta_to,
                                       rel_chord_to_start, rel_chord_to_end});
    } else {
      other_devices.push_back({device.name, eta_from, rel_chord_from_start, rel_chord_from_end, eta_to,
                               rel_chord_to_start, rel_chord_to_end});
    }
  }

  /* Leading edge table*/
  if (!leading_edge_devices.empty()) {
    reporter.htmlReportStream()
        << "<table class=\"content-table\">\n"
        << "<caption>Leading edge control devices</caption>\n"
        << "<thead>\n"
        << "<tr>\n"
        << "<th>Device</th><th>&eta;<sub>from</sub></th><th>x/c<sub>from,start</sub></th><th>x/c<sub>from,end</sub></"
           "th><th>&eta;<sub>to</sub></th><th>x/c<sub>to,start</sub></th><th>x/c<sub>to,end</sub></th>\n"
        << "</tr>\n"
        << "</thead>\n"
        << "<tbody>\n";
    for (auto device : leading_edge_devices) {
      const auto [name, eta_from, rel_chord_from_start, rel_chord_from_end, eta_to, rel_chord_to_start,
                  rel_chord_to_end] = device;
      reporter.htmlReportStream() << "<tr>\n";
      reporter.htmlReportStream() << std::format(
          "<td>{}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td>\n",
          name, eta_from, rel_chord_from_start, rel_chord_from_end, eta_to, rel_chord_to_start,
          rel_chord_to_end);
      reporter.htmlReportStream() << "\n</tr>\n";
    }
    /* Spanwise coordinate - y*/
    reporter.htmlReportStream() << "</tbody>\n"
                                << "</table>\n";
  }

  if (!trailing_edge_devices.empty()) {
    reporter.htmlReportStream()
        << "<table class=\"content-table\">\n"
        << "<caption>Trailing edge control devices</caption>\n"
        << "<thead>\n"
        << "<tr>\n"
        << "<th>Device</th><th>&eta;<sub>from</sub></th><th>x/c<sub>from,start</sub></th><th>x/c<sub>from,end</sub></"
           "th><th>&eta;<sub>to</sub></th><th>x/c<sub>to,start</sub></th><th>x/c<sub>to,end</sub></th>\n"
        << "</tr>\n"
        << "</thead>\n"
        << "<tbody>\n";
    for (auto device : trailing_edge_devices) {
      const auto [name, eta_from, rel_chord_from_start, rel_chord_from_end, eta_to, rel_chord_to_start,
                  rel_chord_to_end] = device;
      reporter.htmlReportStream() << "<tr>\n";
      reporter.htmlReportStream() << std::format(
          "<td>{}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td>\n",
          name, eta_from, rel_chord_from_start, rel_chord_from_end, eta_to, rel_chord_to_start,
          rel_chord_to_end);
      reporter.htmlReportStream() << "\n</tr>\n";
    }
    reporter.htmlReportStream() << "</tbody>\n"
                                << "</table>\n";
  }

  if (!other_devices.empty()) {
    reporter.htmlReportStream()
        << "<table class=\"content-table\">\n"
        << "<caption>Other control devices</caption>\n"
        << "<thead>\n"
        << "<tr>\n"
        << "<th>Device</th><th>&eta;<sub>from</sub></th><th>x/c<sub>from,start</sub></th><th>x/c<sub>from,end</sub></"
           "th><th>&eta;<sub>to</sub></th><th>x/c<sub>to,start</sub></th><th>x/c<sub>to,end</sub></th>\n"
        << "</tr>\n"
        << "</thead>\n"
        << "<tbody>\n";
    for (auto device : other_devices) {
      const auto [name, eta_from, rel_chord_from_start, rel_chord_from_end, eta_to, rel_chord_to_start,
                  rel_chord_to_end] = device;
      reporter.htmlReportStream() << "<tr>\n";
      reporter.htmlReportStream() << std::format(
          "<td>{}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td><td>{:.2f}</td>\n",
          name, eta_from, rel_chord_from_start, rel_chord_from_end, eta_to, rel_chord_to_start,
          rel_chord_to_end);
      reporter.htmlReportStream() << "\n</tr>\n";
    }
    /* Spanwise coordinate - y*/
    reporter.htmlReportStream() << "</tbody>\n"
                                << "</table>\n";
  }

  reporter.htmlReportStream() << "<table class=\"content-table\">\n"
                              << "<caption>Mass and CoG</caption>\n"
                              << "<thead>\n"
                              << "<tr>\n"
                              << "<th>Mass</th><th>CoG<sub>x</sub></th><th>CoG<sub>y</sub></th><th>CoG<sub>z</sub></th>\n"
                              << "</tr>\n"
                              << "</thead>\n"
                              << "<tbody>\n"
                              << "<tr>\n"
                              << std::format("<td>{:.2f}kg</td><td>{:.2f}m</td><td>{:.2f}m</td><td>{:.2f}m</td>",
                                             data->wing_mass_properties.data["mass"].value(),
                                             data->wing_mass_properties.data["x"].value(),
                                             data->wing_mass_properties.data["y"].value(),
                                             data->wing_mass_properties.data["z"].value())
                              << "\n<tr>\n"
                              << "</tbody>\n"
                              << "</table>\n";
  /* close box data div */
  reporter.htmlReportStream() << "</div>\n";

  /* open box plot div */
  reporter.htmlReportStream() << "<div class=\"box plot\">\n";
  /* Add headline Data*/
  reporter.htmlReportStream() << "<h2>Plots</h2>\n";

  reporter.htmlReportStream() << "<h3>Geometry</h3>\n";
  auto planform_plot_path = plot_wing_planform();
  reporter.htmlReportStream() << std::format("<img class=\"image-plot\" src=\"{}\"/>\n",planform_plot_path.string());

  auto airfoil_plot_paths = plot_airfoils();
  reporter.htmlReportStream() << "<h3>Airfoils</h3>\n";
  for( auto airfoil_path : airfoil_plot_paths) {
    reporter.htmlReportStream() << std::format("<img class=\"image-plot\" src=\"{}\"/>\n", airfoil_path.string());
  }
  reporter.htmlReportStream() << "<h3>Thickness & Geometric Twist</h3>\n";
  auto thickness_and_twist_plot_path = plot_thickness_and_twist_distribution();
  reporter.htmlReportStream() << std::format("<img class=\"image-plot\" src=\"{}\"/>\n", thickness_and_twist_plot_path.string());

  /* close box plot */
  reporter.htmlReportStream() << "</div>\n";
}
}  // namespace cantilever
}  // namespace taw
